Package: multiscape
Type: Package
Version: 1.0.7
Title: Multi-Objective Spatial Planning
Description: Provides a modular framework for exact multi-objective spatial
    planning using mixed-integer programming. The package supports the
    definition of planning problems through planning units, features,
    management actions, action effects, spatial relations, targets,
    constraints, and objective functions. It enables the optimisation of
    spatial planning portfolios under considerations such as boundary
    structure, connectivity, and fragmentation. Supported multi-objective
    methods include weighted-sum aggregation, epsilon-constraint, and the
    augmented epsilon-constraint method. Problems can be solved with several
    commercial and open-source optimisation solvers. Optional solver backends 
    include the 'gurobi' R package, which is distributed with the Gurobi Optimizer 
    installation <https://docs.gurobi.com/projects/optimizer/en/13.0/reference/r/setup.html>, 
    and the 'rcbc' R package, available from GitHub at 
    <https://github.com/dirkschumacher/rcbc>. For background on multi-objective 
    optimisation methods, see Halffmann et al.
    (2022) <doi:10.1002/mcda.1780>; for the augmented epsilon-constraint method,
    see Mavrotas (2009) <doi:10.1016/j.amc.2009.03.037>.
Authors@R: c(
    person('José', 'Salgado-Rojas',
        email='jose.salgroj@gmail.com', 
        role = c('aut','cre'),
        comment = c(ORCID = "0000-0001-9817-2625")
    ),
    person('Matías', 'Moreno-Faguett',
        email='matias.moreno.faguett@gmail.com', 
        role = c('aut'),
        comment = c(ORCID = "0009-0003-6561-234X")
    ),
    person('Núria', 'Aquilué',
        email='nuria.aquilue@ctfc.cat', 
        role = c('aut'),
        comment = c(ORCID = "0000-0001-7911-3144")
    ))
Depends: R (>= 4.1.0)
Imports: assertthat (>= 0.2.0), Matrix, proto, magrittr, dplyr, Rcpp,
        cli, sf, methods, RANN, exactextractr, ggplot2, terra, ggrepel
Suggests: roxygen2, prioritizr, Rsymphony (>= 0.1-31), Rcplex, slam,
        rlang, testthat (>= 3.0.0), raster, tmap, sp, viridis,
        markdown, data.table, purrr, readr, tibble, reshape2, rcbc,
        gurobi
LinkingTo: Rcpp, RcppArmadillo (>= 0.10.1.0.0), BH
Encoding: UTF-8
LazyData: true
License: GPL (>= 3)
Language: en-US
RoxygenNote: 7.3.3
URL: https://josesalgr.github.io/multiscape/
BugReports: https://github.com/josesalgr/multiscape/issues
Collate: 'RcppExports.R' 'internal.R' 'add_action_max_per_pu.R'
        'add_actions.R' 'add_constraint_area.R'
        'add_constraint_budget.R' 'add_constraint_locked_actions.R'
        'add_constraint_locked_pu.R' 'add_constraint_targets.R'
        'add_effects.R' 'add_objectives.R' 'add_profit.R'
        'add_spatial_relations.R' 'build_model.R' 'compile_model.R'
        'create_problem.R' 'data-sim.R' 'internalMO.R' 'get_solution.R'
        'globals.R' 'load_sim.R' 'package.R' 'plot.R' 'print.R'
        'problem-class.R' 'set_method_augmecon.R'
        'set_method_epsilon_constraint.R' 'set_method_weighted_sum.R'
        'set_solver.R' 'solutionset-class.R' 'solution-class.R'
        'show.R' 'solve.R' 'utils-pipe.R' 'zzz.R'
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-04-29 18:14:57 UTC; Y6026159N
Author: José Salgado-Rojas [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9817-2625>),
  Matías Moreno-Faguett [aut] (ORCID:
    <https://orcid.org/0009-0003-6561-234X>),
  Núria Aquilué [aut] (ORCID: <https://orcid.org/0000-0001-7911-3144>)
Maintainer: José Salgado-Rojas <jose.salgroj@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-30 08:30:19 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-01 18:22:33 UTC; unix
Archs: multiscape.so.dSYM

Package: multiscape
Type: Package
Version: 1.0.7
Title: Multi-Objective Spatial Planning
Description: Provides a modular framework for exact multi-objective spatial
    planning using mixed-integer programming. The package supports the
    definition of planning problems through planning units, features,
    management actions, action effects, spatial relations, targets,
    constraints, and objective functions. It enables the optimisation of
    spatial planning portfolios under considerations such as boundary
    structure, connectivity, and fragmentation. Supported multi-objective
    methods include weighted-sum aggregation, epsilon-constraint, and the
    augmented epsilon-constraint method. Problems can be solved with several
    commercial and open-source optimisation solvers. Optional solver backends 
    include the 'gurobi' R package, which is distributed with the Gurobi Optimizer 
    installation <https://docs.gurobi.com/projects/optimizer/en/13.0/reference/r/setup.html>, 
    and the 'rcbc' R package, available from GitHub at 
    <https://github.com/dirkschumacher/rcbc>. For background on multi-objective 
    optimisation methods, see Halffmann et al.
    (2022) <doi:10.1002/mcda.1780>; for the augmented epsilon-constraint method,
    see Mavrotas (2009) <doi:10.1016/j.amc.2009.03.037>.
Authors@R: c(
    person('José', 'Salgado-Rojas',
        email='jose.salgroj@gmail.com', 
        role = c('aut','cre'),
        comment = c(ORCID = "0000-0001-9817-2625")
    ),
    person('Matías', 'Moreno-Faguett',
        email='matias.moreno.faguett@gmail.com', 
        role = c('aut'),
        comment = c(ORCID = "0009-0003-6561-234X")
    ),
    person('Núria', 'Aquilué',
        email='nuria.aquilue@ctfc.cat', 
        role = c('aut'),
        comment = c(ORCID = "0000-0001-7911-3144")
    ))
Depends: R (>= 4.1.0)
Imports: assertthat (>= 0.2.0), Matrix, proto, magrittr, dplyr, Rcpp,
        cli, sf, methods, RANN, exactextractr, ggplot2, terra, ggrepel
Suggests: roxygen2, prioritizr, Rsymphony (>= 0.1-31), Rcplex, slam,
        rlang, testthat (>= 3.0.0), raster, tmap, sp, viridis,
        markdown, data.table, purrr, readr, tibble, reshape2, rcbc,
        gurobi
LinkingTo: Rcpp, RcppArmadillo (>= 0.10.1.0.0), BH
Encoding: UTF-8
LazyData: true
License: GPL (>= 3)
Language: en-US
RoxygenNote: 7.3.3
URL: https://josesalgr.github.io/multiscape/
BugReports: https://github.com/josesalgr/multiscape/issues
Collate: 'RcppExports.R' 'internal.R' 'add_action_max_per_pu.R'
        'add_actions.R' 'add_constraint_area.R'
        'add_constraint_budget.R' 'add_constraint_locked_actions.R'
        'add_constraint_locked_pu.R' 'add_constraint_targets.R'
        'add_effects.R' 'add_objectives.R' 'add_profit.R'
        'add_spatial_relations.R' 'build_model.R' 'compile_model.R'
        'create_problem.R' 'data-sim.R' 'internalMO.R' 'get_solution.R'
        'globals.R' 'load_sim.R' 'package.R' 'plot.R' 'print.R'
        'problem-class.R' 'set_method_augmecon.R'
        'set_method_epsilon_constraint.R' 'set_method_weighted_sum.R'
        'set_solver.R' 'solutionset-class.R' 'solution-class.R'
        'show.R' 'solve.R' 'utils-pipe.R' 'zzz.R'
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-04-29 18:14:57 UTC; Y6026159N
Author: José Salgado-Rojas [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9817-2625>),
  Matías Moreno-Faguett [aut] (ORCID:
    <https://orcid.org/0009-0003-6561-234X>),
  Núria Aquilué [aut] (ORCID: <https://orcid.org/0000-0001-7911-3144>)
Maintainer: José Salgado-Rojas <jose.salgroj@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-30 08:30:19 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 21:37:59 UTC; unix
Archs: multiscape.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(compile_model,Problem)
S3method(print,Problem)
S3method(print,Solution)
S3method(print,SolutionSet)
S3method(solve,Problem)
export("%>%")
export(Problem)
export(Solution)
export(SolutionSet)
export(add_actions)
export(add_benefits)
export(add_constraint_area)
export(add_constraint_budget)
export(add_constraint_locked_actions)
export(add_constraint_locked_pu)
export(add_constraint_targets_absolute)
export(add_constraint_targets_relative)
export(add_effects)
export(add_losses)
export(add_objective_max_benefit)
export(add_objective_max_net_profit)
export(add_objective_max_profit)
export(add_objective_min_cost)
export(add_objective_min_fragmentation_action)
export(add_objective_min_fragmentation_pu)
export(add_objective_min_intervention_impact)
export(add_objective_min_loss)
export(add_profit)
export(add_spatial_boundary)
export(add_spatial_distance)
export(add_spatial_knn)
export(add_spatial_queen)
export(add_spatial_relations)
export(add_spatial_rook)
export(compile_model)
export(create_problem)
export(get_actions)
export(get_features)
export(get_pu)
export(get_solution_vector)
export(get_targets)
export(load_sim_features_raster)
export(plot_spatial)
export(plot_spatial_actions)
export(plot_spatial_features)
export(plot_spatial_pu)
export(plot_tradeoff)
export(set_method_augmecon)
export(set_method_epsilon_constraint)
export(set_method_weighted_sum)
export(set_solver)
export(set_solver_cbc)
export(set_solver_cplex)
export(set_solver_gurobi)
export(set_solver_symphony)
export(solve)
exportMethods(create_problem)
importFrom(Rcpp,evalCpp)
importFrom(magrittr,"%>%")
useDynLib(multiscape, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# multiscape: Multi-objective spatial planning in R <img src="man/figures/logo.png" align="right" width=15% />

<!-- badges: start -->

[![CRAN status](https://www.r-pkg.org/badges/version/multiscape)](https://CRAN.R-project.org/package=multiscape)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/multiscape)](https://cran.r-project.org/package=multiscape)
[![lifecycle](https://img.shields.io/badge/Lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check](https://github.com/josesalgr/multiscape/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/josesalgr/multiscape/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/josesalgr/multiscape/graph/badge.svg)](https://app.codecov.io/gh/josesalgr/multiscape)
<!-- badges: end -->

## Overview

`multiscape` is an exact optimisation framework for **multi-objective
spatial planning** in R. It is designed for planning problems in which
spatial data, ecological or socioeconomic features, constraints, and
multiple competing objectives must be considered simultaneously within a
single decision-support workflow. The package is built around
**mixed-integer linear programming (MILP)** formulations, allowing users
to represent spatial planning problems explicitly as optimisation models
and solve them with exact methods. This makes `multiscape` especially
suitable for applications where transparent model structure,
reproducibility, and rigorous trade-off analysis are important.
`multiscape` supports both general spatial planning formulations and
action-based formulations in which decisions are expressed as
**management actions** applied across planning units. With it, users can
build planning problems from tabular or spatial inputs, define feasible
actions and their effects, add targets and other constraints, register
multiple objectives such as cost, benefit, profit, or fragmentation, and
explore exact trade-offs using multi-objective methods such as
weighted-sum, epsilon-constraint, and AUGMECON.

## Installation

You can install the stable version of `multiscape` from CRAN:

```r
install.packages("multiscape")
```

And, the current development version can be installed from GitHub:

``` r
if (!requireNamespace("remotes", quietly = TRUE)) {
  install.packages("remotes")
}

remotes::install_github("josesalgr/multiscape")
```

## Why `multiscape`?

`multiscape` brings together several key components of spatial planning
in a single optimisation framework:

- **exact MILP-based modelling** for rigorous spatial decision support,
- **multi-objective optimisation** to analyse trade-offs instead of a
  single optimum,
- **spatially explicit formulations** with targets, constraints, and
  spatial relations,
- **flexible decision structures**, including both planning-unit and
  action-based formulations,
- and a **reproducible R workflow** for building and solving
  optimisation problems.

## Core workflow

A typical `multiscape` workflow has five steps.

<figure>
<img src="man/figures/overview_multiscape.png"
alt="Conceptual overview of the multiscape workflow, from problem definition and actions to constraints, objectives, optimization, and outputs." />
<figcaption aria-hidden="true">Conceptual overview of the multiscape
workflow, from problem definition and actions to constraints,
objectives, optimization, and outputs.</figcaption>
</figure>

First, the user creates a `Problem` object from planning units,
features, and baseline feature amounts. Second, the user adds feasible
actions and, when relevant, defines how those actions affect the
features of interest. Third, the user adds targets, constraints, and
spatial relations. Fourth, the user registers one or more atomic
objectives. Finally, the user solves the problem in single-objective or
multi-objective mode.

In other words, `multiscape` separates problem definition from
optimization. The `Problem` object stores the planning specification,
and the solving stage later translates that specification into one or
more exact optimization runs.

## Multi-objective methods

`multiscape` currently supports several exact multi-objective workflows.

The **weighted-sum** method combines multiple registered objectives into
a single scalar objective through user-defined weights. It is simple and
useful for preference-driven exploration of trade-offs.

The **epsilon-constraint** method optimizes one objective directly while
transforming the remaining objectives into explicit performance
constraints. This is especially useful when one objective should be
prioritised while the others are controlled through thresholds.

The **AUGMECON** method extends epsilon-constraint with an augmented
formulation designed to reduce weakly efficient solutions and improve
Pareto-front generation.

Together, these methods allow users to move beyond a single objective
solution and instead analyse sets of efficient trade-off solutions.

## Worked example

The example below illustrates a simple **multi-objective spatial action
planning** problem. Some planning units can only be managed through
**conservation**, whereas others can only be managed through
**restoration**. The aim is to identify efficient trade-offs between
**total cost** and **impact on ecosystem services** while satisfying a
minimum area requirement.

### Load the package and example data

``` r
library(multiscape)
library(dplyr)
library(terra)

data("sim_pu_sf", package = "multiscape")
sim_features <- load_sim_features_raster()

n_pu <- nrow(sim_pu_sf)
area_pu <- sim_pu_sf$area[1] 
```

This example considers a landscape with 30,496 planning units and four
ecosystem-service layers represented in `sim_features`. For each unit,
the model can assign one of two management actions, `conservation` or
`restoration`, whenever that action is feasible. The goal is to explore
the trade-off between minimizing total implementation cost and
minimizing impact on the baseline ecosystem services. Instead of
returning a single solution, the workflow uses an epsilon-constraint
method to generate several efficient alternatives that reveal the
cost-impact trade-off.

### Visualise the ecosystem-service layers

These four raster layers represent the baseline ecosystem-service
conditions used by the optimisation model. They define the spatial
distribution of the features that are later affected by conservation and
restoration actions.

``` r
terra::plot(
  sim_features,
  main = c(
    "Rodent and lagomorph control",
    "Carbon stock",
    "Seasonal water yield",
    "Sediment delivery ratio"
  )
)
```

<img src="man/figures/README-unnamed-chunk-2-1.png" width="100%" />

### Build the base problem

We first create the base `Problem` object from the planning units and
the feature layers.

``` r
p <- create_problem(
  pu = sim_pu_sf,
  features = sim_features,
  cost = "cost"
)

p
#> A multiscape object (<Problem>)
#> ├─data
#> │├─planning units: <data.frame> (30496 total)
#> │├─costs: min: 0, max: 779.43
#> │└─features: 4 total ("L1_control_inv", "L1_stock_inv", "L1_rend_inv", ...)
#> └─actions and effects
#> │├─actions: none
#> │├─feasible action pairs: none
#> │├─effect data: none
#> │└─profit data: none
#> └─spatial
#> │├─geometry: sf (30496 rows)
#> │├─coordinates: 30496 rows (x: 344825.53946..350775.13398, y:
#> 6225654.6982..6229659.6982)
#> │└─relations: none
#> └─targets and constraints
#> │├─targets: none
#> │├─area constraints: none
#> │├─budget constraints: none
#> │├─planning-unit locks: none
#> │└─action locks: none
#> └─model
#> │├─status: not built yet (will build in solve())
#> │├─objectives: none
#> │├─method: single-objective
#> │├─solver: not set (auto)
#> │└─checks: incomplete (no objective registered)
#> # ℹ Use `x$data` to inspect stored tables and model snapshots.
```

This call defines the spatial domain of the optimisation problem and
tells `multiscape` which planning-unit attribute should be interpreted
as cost. Importantly, this is still only a **problem definition**. The
model does not yet know which actions are available, where they are
admissible, how they affect the features, or which objectives should be
optimised. Those elements are added explicitly in the following
sections.

Printing the `Problem` object at this stage is useful because it
provides a compact summary of the current planning specification. It
lets the user verify that the planning units, features, and baseline
inputs have been correctly registered before actions, effects,
constraints, and objectives are added.

### Define the action set

We now specify the catalogue of actions that the model is allowed to
assign.

``` r
actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)
```

Here, the action set is deliberately simple and includes only two
interventions: **conservation** and **restoration**. In more complex
applications, the same structure can be extended to represent larger
management portfolios, alternative treatment intensities, or multiple
intervention types.

### Derive action eligibility

Not every action is allowed in every planning unit. In this example, the
variable `locked_in` is used to derive a simple action-eligibility rule
for the illustrative dataset: one subset of planning units is only
eligible for **conservation**, whereas the complementary subset is only
eligible for **restoration**.

``` r
feasible_forest <- sim_pu_sf |>
  sf::st_drop_geometry() |>
  select(id, locked_in) |>
  filter(locked_in) |>
  transmute(pu = id, action = "conservation")

feasible_agro <- sim_pu_sf |>
  sf::st_drop_geometry() |>
  select(id, locked_in) |>
  filter(!locked_in) |>
  transmute(pu = id, action = "restoration")

include_df <- bind_rows(feasible_forest, feasible_agro) |>
  distinct()
```

The resulting table defines the feasible planning unit-action
combinations that the optimizer is allowed to use. This is an important
aspect of the action-based formulation: feasibility is not inferred
automatically by the solver, but is instead encoded directly by the user
as part of the planning specification. In other words, the optimisation
model can only allocate actions where they have been declared
admissible.

### Specify action effects

We next specify how actions affect the features.

``` r
effects_mult <- data.frame(
  action = rep(c("conservation", "restoration"), each = 4),
  feature = rep(
    c("L1_control_inv", "L1_stock_inv", "L1_rend_inv", "L1_reten_inv"),
    times = 2
  ),
  multiplier = 1
)
```

This table assigns an effect to every `action Ã— feature` combination.
Because `effect_type = "delta"` will be used later, the values are
interpreted as **action-induced changes** relative to the baseline
feature representation. In the present example, the same multiplier
(`1`) is assigned to all combinations, so the ecological response is
intentionally kept simple: both conservation and restoration are assumed
to contribute with the same relative magnitude across all four features.

This means that, in this README example, the difference between the two
actions comes mainly from **where they are eligible** and **how much
they cost**, rather than from different ecological effect strengths. In
a more realistic application, one would often use different multipliers
by action and feature, for example making restoration stronger for some
features and weaker, neutral, or even negative for others.

### Define action-specific costs

Costs are attached to planning unit-action combinations so that the
model can compare interventions not only by their ecological
consequences, but also by their local implementation burden.

``` r
action_cost_df <- sim_pu_sf |>
  sf::st_drop_geometry() |>
  transmute(
    pu = id,
    action = ifelse(locked_in, "conservation", "restoration"),
    cost = cost
  ) |>
  semi_join(include_df, by = c("pu", "action"))
```

This representation is useful in action-based planning because the same
planning unit may imply different costs depending on the intervention
that is permitted or selected. By storing costs at the
`planning unit Ã— action` level, the optimisation problem can evaluate
trade-offs in a more realistic way than with a single uniform
planning-unit cost.

### Add constraints and objectives

Once actions, eligibility, effects, and costs have been defined, we can
complete the optimisation model.

``` r
target_area <- n_pu * area_pu * 0.20

p <- p |>
  add_actions(actions = actions_df, 
              cost = action_cost_df, 
              include = include_df) |>
  add_effects(effects = effects_mult, 
              effect_type = "delta") |>
  add_objective_min_cost(alias = "Total cost", 
                         include_pu_cost = FALSE, 
                         include_action_cost = TRUE) |>
  add_objective_min_intervention_impact(alias = "Total impact") |>
  add_constraint_targets_relative(0) |>
  add_constraint_area(area = target_area, 
                      sense = "min") |>
  add_constraint_locked_pu(locked_in = "locked_in", 
                           locked_out = "locked_out") 
```

The call to `add_actions()` links the action catalogue to the admissible
allocation table and the action-specific cost table. The call to
`add_effects()` then tells the model how these actions modify the
tracked features. After that, the optimisation problem is completed by
combining constraints and objectives.

The minimum-area constraint imposes a lower bound on the amount of land
to be selected, ensuring that the solution represents a non-trivial
intervention effort. In this example, the area threshold is expressed in
area units rather than as a simple count of planning units, which is why
the code multiplies the target number of selected units by the area
represented by one planning unit in the example dataset. The registered
objectives define the trade-off of interest: one objective minimizes
total action cost, while the other minimizes intervention impact. The
epsilon-constraint method is then used to generate a set of efficient
alternatives rather than a single solution.

### Configure the epsilon-constraint method and solver

``` r
p <- p |>
  set_method_epsilon_constraint(
    primary = "Total impact",
    n_points = 4,
    mode = "auto",
    lexicographic = TRUE
  ) |>
  set_solver_cbc(gap_limit = 0.1)

p
#> A multiscape object (<Problem>)
#> ├─data
#> │├─planning units: <data.frame> (30496 total)
#> │├─costs: min: 0, max: 779.43
#> │└─features: 4 total ("L1_control_inv", "L1_stock_inv", "L1_rend_inv", ...)
#> └─actions and effects
#> │├─actions: 2 total ("conservation", "restoration")
#> │├─feasible action pairs: 30496 feasible rows
#> │├─action costs: min: 0, max: 779.43
#> │├─effect data: 105537 rows
#> │├─effect mode: benefit only
#> │└─profit data: none
#> └─spatial
#> │├─geometry: sf (30496 rows)
#> │├─coordinates: 30496 rows (x: 344825.53946..350775.13398, y:
#> 6225654.6982..6229659.6982)
#> │└─relations: none
#> └─targets and constraints
#> │├─targets: 4 rows
#> │├─target preview: "L1_control_inv" >= 0, "L1_stock_inv" >= 0, "L1_rend_inv" >=
#> 0
#> │├─area constraints: 1 ([all] min)
#> │├─budget constraints: none
#> │├─planning-unit locks: 5407 units (3988 locked-in, 1419 locked-out)
#> │└─action locks: none
#> └─model
#> │├─status: not built yet (will build in solve())
#> │├─objectives: 2 registered (Total cost, Total impact)
#> │├─method: epsilon_constraint (primary: Total impact)
#> │├─solver: cbc
#> │└─checks: ok
#> # ℹ Use `x$data` to inspect stored tables and model snapshots.
```

Printing `p` again after adding actions, effects, constraints, and
objectives helps confirm that the problem has moved from a base spatial
specification to a fully defined optimisation model ready to be solved.

### Solve and inspect the results

``` r
res <- solve(p, process = TRUE)
plot_tradeoff(res, label_runs = TRUE)
```

<img src="man/figures/README-unnamed-chunk-10-1.png" width="100%" />

The resulting `SolutionSet` can be used to compare efficient
alternatives both numerically and spatially. The trade-off plot helps
assess how the solutions balance the competing objectives, whereas the
spatial action maps show where conservation and restoration are
allocated in each efficient solution. This is particularly valuable
because different Pareto-efficient solutions may achieve similar totals
while distributing actions very differently across the landscape.

If desired, the spatial outputs can also be rendered with a consistent
manual palette to improve visual comparison across solutions. For
example:

``` r
plot_spatial_actions(
  res,
  fill_values = c(
    restoration  = "#f38b3d",
    conservation = "#97df83"
  ),
  runs = c(1, 2, 3, 4),
  base_alpha = 0.0
)
```

<img src="man/figures/README-unnamed-chunk-11-1.png" width="100%" />

The resulting `SolutionSet` can be used to compare efficient
alternatives both numerically and spatially. The trade-off plot helps
assess how the solutions balance the competing objectives, whereas the
spatial action maps show where conservation and restoration are
allocated in each efficient solution. This is particularly valuable
because different Pareto-efficient solutions may achieve similar totals
while distributing actions very differently across the landscape.

## Why this example matters

This example illustrates one of the main strengths of `multiscape`:
spatial planning problems can be formulated in terms of **management
actions**, **action-specific effects**, **feasibility rules**,
**constraints**, and **multiple competing objectives**. Rather than
returning only a single optimum, the package can generate and analyse
alternative efficient solutions that reveal trade-offs relevant for
decision-making.

## Learn more

To explore the package further, see the function reference on the
package website and the documentation of key functions such as
`create_problem()`, `add_actions()`, `add_effects()`,
`add_constraint_targets_relative()`, and `solve()`.

For multi-objective workflows, the most relevant functions are
`set_method_weighted_sum()`, `set_method_epsilon_constraint()`, and
`set_method_augmecon()`.

If you find a bug or want to suggest an improvement, please open an
issue at:

<https://github.com/josesalgr/multiscape/issues>

#' @include internal.R
NULL

#' @title Add locked planning units to a problem
#'
#' @description
#' Define planning units that must be included in, or excluded from, the
#' optimization problem.
#'
#' This function updates the planning-unit table stored in the \code{Problem}
#' object by creating or replacing the logical columns \code{locked_in} and
#' \code{locked_out}. These columns are later used by the model builder when
#' translating the problem into optimization constraints.
#'
#' Lock information may be supplied either directly as logical vectors, as
#' vectors of planning-unit ids, or by referencing columns in the raw
#' planning-unit data originally passed to \code{\link{create_problem}}.
#'
#' @details
#' Use this function when whole planning units must be forced into or out of
#' the solution, regardless of which action may later be selected in them.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and let
#' \eqn{w_i \in \{0,1\}} denote the binary variable indicating whether planning
#' unit \eqn{i \in \mathcal{I}} is selected by the model.
#'
#' This function defines two subsets:
#' \itemize{
#'   \item \eqn{\mathcal{I}^{in} \subseteq \mathcal{I}}, the planning units that
#'   must be included,
#'   \item \eqn{\mathcal{I}^{out} \subseteq \mathcal{I}}, the planning units
#'   that must be excluded.
#' }
#'
#' Conceptually, these sets correspond to the following conditions:
#' \itemize{
#'   \item if \eqn{i \in \mathcal{I}^{in}}, then \eqn{w_i = 1},
#'   \item if \eqn{i \in \mathcal{I}^{out}}, then \eqn{w_i = 0}.
#' }
#'
#' These constraints are not imposed immediately by this function; instead, they
#' are stored in the planning-unit table and enforced later when building the
#' optimization model.
#'
#' \strong{Philosophy}
#'
#' The role of \code{\link{create_problem}} is to construct and normalize the basic
#' inputs of the planning problem. Locking planning units is treated as a
#' separate modelling step so that users can define or revise selection
#' restrictions after the \code{Problem} object has already been created.
#'
#' In contrast, \code{\link{add_constraint_locked_actions}} is used to fix
#' specific feasible \code{(pu, action)} decisions rather than whole planning
#' units.
#'
#' \strong{Supported input formats}
#'
#' For both \code{locked_in} and \code{locked_out}, the function accepts:
#' \itemize{
#'   \item \code{NULL}, meaning that no planning units are locked on that side,
#'   \item a single character string, interpreted as a column name in the raw
#'   planning-unit data,
#'   \item a logical vector of length \code{nrow(pu)},
#'   \item a vector of planning-unit ids.
#' }
#'
#' When a column name is supplied, the referenced column is coerced to logical.
#' Numeric values are interpreted as non-zero = \code{TRUE}; character and
#' factor values are interpreted using common logical strings such as
#' \code{"true"}, \code{"t"}, \code{"1"}, \code{"yes"}, and \code{"y"}.
#' Missing values are treated as \code{FALSE}.
#'
#' \strong{Replacement behaviour}
#'
#' Each call to \code{add_constraint_locked_pu()} replaces any existing
#' \code{locked_in} and \code{locked_out} columns in the planning-unit table. In
#' other words, the function defines the complete current set of locked planning
#' units; it does not merge new values with previous ones.
#'
#' \strong{Consistency checks}
#'
#' The function checks that no planning unit is simultaneously assigned to both
#' \code{locked_in} and \code{locked_out}. If such conflicts are found, an error
#' is raised.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}.
#'
#' @param locked_in Optional locked-in specification. It may be \code{NULL}, a
#'   column name in the raw planning-unit data, a logical vector, or a vector of
#'   planning-unit ids.
#'
#' @param locked_out Optional locked-out specification. It may be \code{NULL}, a
#'   column name in the raw planning-unit data, a logical vector, or a vector of
#'   planning-unit ids.
#'
#' @return An updated \code{Problem} object in which the planning-unit table
#'   contains logical columns \code{locked_in} and \code{locked_out}.
#'
#' @examples
#' pu <- data.frame(
#'   id = 1:5,
#'   cost = c(2, 3, 1, 4, 2),
#'   lock_col = c(TRUE, FALSE, FALSE, TRUE, FALSE),
#'   out_col = c(FALSE, FALSE, FALSE, FALSE, TRUE)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' # 1) Lock by planning-unit ids
#' p1 <- add_constraint_locked_pu(
#'   x = p,
#'   locked_in = c(1, 3),
#'   locked_out = c(5)
#' )
#'
#' p1$data$pu[, c("id", "locked_in", "locked_out")]
#'
#' # 2) Read lock information from raw PU data columns
#' p2 <- add_constraint_locked_pu(
#'   x = p,
#'   locked_in = "lock_col",
#'   locked_out = "out_col"
#' )
#'
#' p2$data$pu[, c("id", "locked_in", "locked_out")]
#'
#' # 3) Use logical vectors
#' p3 <- add_constraint_locked_pu(
#'   x = p,
#'   locked_in = c(TRUE, FALSE, TRUE, FALSE, FALSE),
#'   locked_out = c(FALSE, FALSE, FALSE, TRUE, FALSE)
#' )
#'
#' p3$data$pu[, c("id", "locked_in", "locked_out")]
#'
#' @seealso
#' \code{\link{create_problem}},
#' \code{\link{add_actions}},
#' \code{\link{add_constraint_locked_actions}}
#'
#' @export
add_constraint_locked_pu <- function(
    x,
    locked_in = NULL,
    locked_out = NULL
) {
  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)

  pu <- x$data$pu %||% NULL
  if (is.null(pu) || !inherits(pu, "data.frame") || nrow(pu) == 0L) {
    stop("x$data$pu is missing or empty.", call. = FALSE)
  }
  if (!("id" %in% names(pu))) {
    stop("x$data$pu must contain column 'id'.", call. = FALSE)
  }

  pu_raw <- x$data$pu_data_raw %||% NULL
  n_pu <- nrow(pu)
  pu_ids <- as.integer(pu$id)

  .as_logical_from_column <- function(colname, source_df, what) {
    if (is.null(source_df) || !inherits(source_df, "data.frame")) {
      stop(
        what, " was provided as a column name, but x$data$pu_data_raw is missing.",
        call. = FALSE
      )
    }

    colname <- as.character(colname)[1]
    if (is.na(colname) || !nzchar(colname)) {
      stop(what, " column name must be a non-empty string.", call. = FALSE)
    }
    if (!(colname %in% names(source_df))) {
      stop(
        what, " column '", colname, "' was not found in x$data$pu_data_raw.",
        call. = FALSE
      )
    }

    v <- source_df[[colname]]

    if (is.logical(v)) {
      out <- v
    } else if (is.numeric(v) || is.integer(v)) {
      out <- v != 0
    } else if (is.factor(v)) {
      v <- as.character(v)
      vv <- tolower(trimws(v))
      out <- vv %in% c("true", "t", "1", "yes", "y")
    } else if (is.character(v)) {
      vv <- tolower(trimws(v))
      out <- vv %in% c("true", "t", "1", "yes", "y")
    } else {
      stop(
        what, " column '", colname, "' could not be coerced to logical.",
        call. = FALSE
      )
    }

    out <- as.logical(out)
    out[is.na(out)] <- FALSE

    if (length(out) != n_pu) {
      stop(
        what, " column '", colname, "' has length ", length(out),
        " but expected ", n_pu, ".",
        call. = FALSE
      )
    }

    out
  }

  .resolve_lock_spec <- function(spec, what) {
    if (is.null(spec)) return(rep(FALSE, n_pu))

    # case 1: single character -> column name in pu_data_raw
    if (is.character(spec) && length(spec) == 1L && !spec %in% as.character(pu_ids)) {
      return(.as_logical_from_column(spec, pu_raw, what))
    }

    # case 2: logical vector
    if (is.logical(spec)) {
      if (length(spec) != n_pu) {
        stop(
          what, " logical vector must have length ", n_pu, ".",
          call. = FALSE
        )
      }
      spec <- as.logical(spec)
      spec[is.na(spec)] <- FALSE
      return(spec)
    }

    # case 3: PU ids
    ids <- suppressWarnings(as.integer(spec))
    if (length(ids) == 0L || all(is.na(ids))) {
      stop(
        what, " must be NULL, a column name, a logical vector, or a vector of PU ids.",
        call. = FALSE
      )
    }
    ids <- unique(ids[!is.na(ids)])

    bad <- ids[!ids %in% pu_ids]
    if (length(bad) > 0L) {
      stop(
        what, " contains PU ids not present in x$data$pu$id: ",
        paste(bad, collapse = ", "),
        call. = FALSE
      )
    }

    pu_ids %in% ids
  }

  new_in  <- .resolve_lock_spec(locked_in,  "locked_in")
  new_out <- .resolve_lock_spec(locked_out, "locked_out")

  if (any(new_in & new_out, na.rm = TRUE)) {
    bad_ids <- pu_ids[new_in & new_out]
    stop(
      "Some planning units are both locked_in and locked_out: ",
      paste(utils::head(bad_ids, 20), collapse = ", "),
      if (length(bad_ids) > 20) " ..." else "",
      call. = FALSE
    )
  }

  x$data$pu$locked_in <- as.logical(new_in)
  x$data$pu$locked_out <- as.logical(new_out)

  if (!is.null(x$data$model_ptr)) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

test_that("single-objective solve returns a Solution and respects locked units", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()
  toy$pu$locked_in[1] <- 1
  toy$pu$locked_out[4] <- 1

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_constraint_locked_pu(locked_in = "locked_in", locked_out = "locked_out") |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_s3_class(s, "Solution")

  acts <- multiscape::get_actions(s, only_selected = TRUE)
  expect_true(any(acts$pu == 1))
  expect_false(any(acts$pu == 4))
})

#' @title Set the weighted-sum multi-objective method
#'
#' @description
#' Configure a \code{Problem} object to be solved with a weighted-sum
#' multi-objective method.
#'
#' In the weighted-sum method, several registered atomic objectives are combined
#' into a single scalar objective using a weighted linear combination. This
#' function stores that configuration in \code{x$data$method} so that it can be
#' used later by \code{\link{solve}}.
#'
#' @details
#' Use this method when several registered objectives should be combined into a
#' single scalar optimization problem through explicit preference weights.
#'
#' \strong{General idea}
#'
#' Suppose that a set of atomic objectives has already been registered in the
#' problem under aliases \eqn{k \in \mathcal{K}}. Let \eqn{f_k(x)} denote the
#' scalar value of objective \eqn{k}, and let \eqn{w_k} denote its user-supplied
#' weight.
#'
#' The weighted-sum method combines them into a single scalar objective of the
#' form:
#'
#' \deqn{
#' \sum_{k \in \mathcal{K}} w_k \, f_k(x).
#' }
#'
#' In practice, the exact sign convention used internally depends on the sense
#' of each registered atomic objective, for example whether it is a
#' minimization-type or maximization-type objective. The solving layer is
#' therefore responsible for constructing a solver-ready scalar objective from
#' the stored objective specifications and the requested weights.
#'
#' \strong{Atomic objectives requirement}
#'
#' The weighted-sum method can only be used with atomic objectives that have
#' already been registered under aliases. These aliases are typically created by
#' calling objective setters with an \code{alias} argument, for example:
#' \preformatted{
#' x <- x |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_min_fragmentation(alias = "frag")
#' }
#'
#' Internally, each atomic objective is stored in
#' \code{x$data$objectives[[alias]]} together with its metadata, such as:
#' \itemize{
#'   \item \code{objective_id},
#'   \item \code{model_type},
#'   \item \code{sense},
#'   \item \code{objective_args}.
#' }
#'
#' The \code{aliases} argument passed to this function selects which of those
#' registered atomic objectives are included in the weighted combination.
#'
#' \strong{Weight normalization}
#'
#' If \code{normalize_weights = TRUE}, the supplied weights are rescaled to sum
#' to 1:
#'
#' \deqn{
#' \tilde{w}_k = \frac{w_k}{\sum_{j \in \mathcal{K}} w_j}.
#' }
#'
#' This normalization does not change the optimizer's solution in a pure
#' weighted-sum formulation as long as all weights are multiplied by the same
#' positive constant, but it can improve interpretability and numerical
#' conditioning.
#'
#' If \code{normalize_weights = FALSE}, the supplied weights are stored exactly
#' as provided.
#'
#' \strong{Objective scaling}
#'
#' If \code{objective_scaling = TRUE}, the solving layer is expected to scale the
#' participating objectives before combining them. The purpose of scaling is to
#' reduce distortions caused by objectives being measured on very different
#' numerical ranges.
#'
#' Conceptually, if \eqn{R_k} denotes a scale or range associated with objective
#' \eqn{k}, then a scaled weighted sum may be interpreted as:
#'
#' \deqn{
#' \sum_{k \in \mathcal{K}} w_k \, \frac{f_k(x)}{R_k}.
#' }
#'
#' The exact scaling rule is implemented later in the solving layer. This
#' function only stores whether objective scaling has been requested.
#'
#' \strong{Mixed objective senses}
#'
#' Weighted sums are straightforward when all participating objectives have the
#' same optimization sense. When minimization and maximization objectives are
#' mixed, the solving layer must standardize them internally before building the
#' scalar objective.
#'
#' This function validates that the requested aliases exist, but it does not
#' itself resolve objective-sense compatibility. That logic is delegated to the
#' downstream solving layer.
#'
#' \strong{Theoretical limitation}
#'
#' The weighted-sum method typically recovers only \emph{supported} efficient
#' solutions, that is, solutions lying on the convex hull of the Pareto front in
#' objective space. In non-convex multi-objective problems, especially mixed
#' integer problems, some efficient solutions cannot be obtained by any weighted
#' combination. In such cases, methods such as
#' \code{\link{set_method_epsilon_constraint}} or
#' \code{\link{set_method_augmecon}} may be preferable.
#'
#' \strong{Stored configuration}
#'
#' This function stores the method definition in \code{x$data$method} with:
#' \itemize{
#'   \item \code{name = "weighted"},
#'   \item \code{aliases},
#'   \item \code{weights},
#'   \item \code{normalize_weights},
#'   \item \code{objective_scaling}.
#' }
#'
#' The actual scalarization is performed later by \code{\link{solve}}.
#'
#' @param x A \code{Problem} object.
#' @param aliases Character vector of objective aliases to combine. Each alias
#'   must correspond to a previously registered atomic objective.
#' @param weights Numeric vector of weights, with the same length and order as
#'   \code{aliases}. Weights must be finite. If \code{normalize_weights = TRUE},
#'   they are rescaled to sum to 1 before being stored.
#' @param normalize_weights Logical. If \code{TRUE}, normalize the supplied
#'   weights to sum to 1 before storing them.
#' @param objective_scaling Logical. If \code{TRUE}, request scaling of the
#'   participating objectives before weighted aggregation in the solving layer.
#'
#' @return The updated \code{Problem} object with the weighted-sum method
#'   configuration stored in \code{x$data$method}.
#'
#' @examples
#' # Small toy problem
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' effects_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   feature = c(1, 1, 1, 1, 2, 2, 2, 2),
#'   benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
#'   loss = c(0, 0, 1, 0, 0, 1, 0, 0)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_effects(effects_df) |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_max_benefit(alias = "benefit")
#'
#' x <- set_method_weighted_sum(
#'   x,
#'   aliases = c("cost", "benefit"),
#'   weights = c(0.4, 0.6),
#'   normalize_weights = FALSE
#' )
#'
#' x$data$method
#'
#' # Normalize weights before storing
#' x2 <- set_method_weighted_sum(
#'   x,
#'   aliases = c("cost", "benefit"),
#'   weights = c(2, 3),
#'   normalize_weights = TRUE
#' )
#'
#' x2$data$method
#'
#' # Request objective scaling
#' x3 <- set_method_weighted_sum(
#'   x,
#'   aliases = c("cost", "benefit"),
#'   weights = c(0.7, 0.3),
#'   objective_scaling = TRUE
#' )
#'
#' x3$data$method
#'
#' @seealso
#' \code{\link{set_method_epsilon_constraint}},
#' \code{\link{set_method_augmecon}},
#' \code{\link{solve}}
#'
#' @export
set_method_weighted_sum <- function(x,
                                aliases,
                                weights,
                                normalize_weights = FALSE,
                                objective_scaling = FALSE) {

  # ---- promote (handles both Problem and MOProblem)
  #x <- .pamo_as_mo(x)

  # ---- validate aliases
  if (!is.character(aliases) || length(aliases) == 0L || anyNA(aliases)) {
    stop("`aliases` must be a non-empty character vector without NA.", call. = FALSE)
  }
  aliases <- as.character(aliases)

  if (any(!nzchar(aliases))) {
    stop("`aliases` must not contain empty strings.", call. = FALSE)
  }
  if (anyDuplicated(aliases) != 0L) {
    dups <- unique(aliases[duplicated(aliases)])
    stop("`aliases` must not contain duplicates: ", paste(dups, collapse = ", "), call. = FALSE)
  }

  # ---- validate weights
  if (!is.numeric(weights) || length(weights) != length(aliases) || anyNA(weights)) {
    stop("`weights` must be a numeric vector, same length as `aliases`, without NA.", call. = FALSE)
  }
  weights <- as.numeric(weights)

  if (any(!is.finite(weights))) stop("`weights` must be finite.", call. = FALSE)
  #if (any(weights < 0)) stop("`weights` must be non-negative.", call. = FALSE)

  # ---- normalize_weights flag
  if (!is.logical(normalize_weights) || length(normalize_weights) != 1L || is.na(normalize_weights)) {
    stop("`normalize_weights` must be TRUE or FALSE.", call. = FALSE)
  }
  normalize_weights <- isTRUE(normalize_weights)

  # ---- objective_scaling flag
  if (!is.logical(objective_scaling) || length(objective_scaling) != 1L || is.na(objective_scaling)) {
    stop("`objective_scaling` must be TRUE or FALSE.", call. = FALSE)
  }
  objective_scaling <- isTRUE(objective_scaling)

  if (normalize_weights) {
    s <- sum(weights)
    if (!is.finite(s) || s <= 0) {
      stop("`normalize_weights = TRUE` requires sum(weights) > 0.", call. = FALSE)
    }
    weights <- weights / s
  }

  # ---- validate that objective aliases exist (reads x$base$data$objectives)
  # This gives you nicer errors and also future-proofs (e.g., checks objective_id/sense)
  .pamo_get_objective_specs(x, aliases)

  # ---- store method configuration
  x$data$method <- list(
    name = "weighted",
    aliases = aliases,
    weights = weights,
    normalize_weights = normalize_weights,
    objective_scaling = objective_scaling
  )

  x
}

#' @include internal.R
#'
#' @title Add profit to a planning problem
#'
#' @description
#' Define economic profit values for feasible planning unit--action pairs and
#' store them in a profit table.
#'
#' Profit is stored separately from ecological effects. In particular,
#' \code{profit} is not the same as ecological \code{benefit} or
#' \code{loss} as represented in \code{\link{add_effects}}. This separation
#' allows the package to distinguish economic returns from ecological
#' consequences when building objectives, constraints, and reporting summaries.
#'
#' @details
#' \strong{When to use \code{add_profit()}.}
#'
#' Use this function when economic returns, penalties, or other action-specific
#' financial values are part of the planning problem. Typical downstream uses
#' include objectives such as \code{\link{add_objective_max_profit}} and
#' \code{\link{add_objective_max_net_profit}}.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and
#' \eqn{\mathcal{A}} the set of actions. Let
#' \eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
#' feasible planning unit--action pairs currently stored in the problem.
#'
#' This function assigns to each feasible pair \eqn{(i,a) \in \mathcal{D}} a
#' numeric profit value \eqn{\pi_{ia} \in \mathbb{R}} and stores the result in a
#' profit table.
#'
#' Thus, the stored table can be interpreted as a mapping
#' \deqn{
#' \pi : \mathcal{D} \to \mathbb{R},
#' }
#' where \eqn{\pi_{ia}} represents the economic return associated with selecting
#' action \eqn{a} in planning unit \eqn{i}.
#'
#' Profit values may be positive, zero, or negative. Positive values represent
#' gains or revenues, zero represents no net profit contribution, and negative
#' values can be used to encode penalties or net economic losses.
#'
#' The stored table contains:
#' \itemize{
#'   \item \code{pu}: external planning-unit id,
#'   \item \code{action}: action id,
#'   \item \code{profit}: numeric profit value,
#'   \item \code{internal_pu}: internal planning-unit index,
#'   \item \code{internal_action}: internal action index.
#' }
#'
#' \strong{Supported input formats}
#'
#' The \code{profit} argument may be specified in several ways:
#' \itemize{
#'   \item \code{NULL}: assign profit 0 to all feasible \code{(pu, action)}
#'   pairs,
#'   \item a numeric scalar: assign the same profit value to all feasible pairs,
#'   \item a named numeric vector: names are action ids, assigning one global
#'   profit value per action,
#'   \item a \code{data.frame(action, profit)}: assign one global profit value
#'   per action,
#'   \item a \code{data.frame(pu, action, profit)}: assign pair-specific profit
#'   values.
#' }
#'
#' When action-level profit is supplied, the same profit value is assigned to
#' all feasible planning units for that action. When pair-specific profit is
#' supplied, only the listed \code{(pu, action)} pairs receive explicit values;
#' unmatched feasible pairs are interpreted as zero-profit pairs.
#'
#' \strong{Storage behaviour}
#'
#' This function stores only rows with non-zero profit values. Feasible pairs
#' whose final profit is zero are omitted from the stored profit table.
#' Missing values produced during matching or joins are treated as zero before
#' this filtering step. Therefore, the resulting table is a sparse
#' representation of economic returns over the feasible decision space.
#'
#' \strong{Data-only behaviour}
#'
#' This function is purely data-oriented. It does not build or modify the
#' optimization model, and it does not change feasibility. It simply assigns
#' profit values to rows already present in the feasible action table.
#'
#' In particular:
#' \itemize{
#'   \item it does not add new feasible \code{(pu, action)} pairs,
#'   \item it does not remove infeasible pairs,
#'   \item it does not apply solver-side filtering such as dropping locked-out
#'   decisions,
#'   \item it does not modify ecological effect tables.
#' }
#'
#' Any such filtering is expected to occur later when model-ready tables are
#' prepared, typically during the build stage invoked by \code{solve()}.
#'
#' \strong{Use in optimization}
#'
#' Profit values stored by this function can later be used in objectives such as
#' \code{\link{add_objective_max_profit}} or
#' \code{\link{add_objective_max_net_profit}}, in derived budget expressions, or
#' in reporting and summary functions.
#'
#' For example, if \eqn{x_{ia} \in \{0,1\}} denotes whether action \eqn{a} is
#' selected in planning unit \eqn{i}, then a profit-maximization objective
#' typically takes the form
#' \deqn{
#' \max \sum_{(i,a) \in \mathcal{D}} \pi_{ia} x_{ia}.
#' }
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}. It
#'   must already contain feasible actions and an action catalogue; run
#'   \code{\link{add_actions}} first.
#'
#' @param profit Profit specification. One of:
#' \itemize{
#'   \item \code{NULL}: profit is set to 0 for all feasible
#'   \code{(pu, action)} pairs,
#'   \item a numeric scalar: recycled to all feasible pairs,
#'   \item a named numeric vector: names are action ids and values define
#'   action-level profit,
#'   \item a \code{data.frame(action, profit)} defining action-level profit,
#'   \item a \code{data.frame(pu, action, profit)} defining pair-specific
#'   profit.
#' }
#'
#' @return An updated \code{Problem} object with a stored profit table created
#'   or replaced. The stored table contains columns \code{pu}, \code{action},
#'   \code{profit}, \code{internal_pu}, and \code{internal_action}, and
#'   includes only rows with non-zero profit.
#'
#' @examples
#' # Minimal problem
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(2, 3, 1, 4)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   x = p,
#'   actions = data.frame(id = c("harvest", "restoration"))
#' )
#'
#' # 1) Constant profit for every feasible (pu, action)
#' p1 <- add_profit(p, profit = 10)
#' p1$data$dist_profit
#'
#' # 2) Profit per action using a named vector
#' pr <- c(harvest = 50, restoration = -5)
#' p2 <- add_profit(p, profit = pr)
#' p2$data$dist_profit
#'
#' # 3) Profit per action using a data frame
#' pr_df <- data.frame(
#'   action = c("harvest", "restoration"),
#'   profit = c(40, 15)
#' )
#' p3 <- add_profit(p, profit = pr_df)
#' p3$data$dist_profit
#'
#' # 4) Profit per (pu, action) pair
#' pr_pair <- data.frame(
#'   pu = c(1, 2, 3),
#'   action = c("harvest", "harvest", "restoration"),
#'   profit = c(100, 80, 30)
#' )
#' p4 <- add_profit(p, profit = pr_pair)
#' p4$data$dist_profit
#'
#' @seealso
#' \code{\link{add_actions}},
#' \code{\link{add_objective_max_profit}},
#' \code{\link{add_objective_max_net_profit}},
#' \code{\link{add_effects}}
#'
#' @export
add_profit <- function(
    x,
    profit = NULL
) {

  # ---- checks: x
  assertthat::assert_that(!is.null(x), msg = "x is NULL")
  assertthat::assert_that(!is.null(x$data), msg = "x does not look like a mulstiscape Problem object")
  assertthat::assert_that(!is.null(x$data$pu), msg = "x$data$pu is missing. Run create_problem() first.")
  assertthat::assert_that(!is.null(x$data$dist_actions), msg = "No actions found. Run add_actions() first.")
  assertthat::assert_that(!is.null(x$data$actions), msg = "No action catalog found. Run add_actions() first.")

  x <- .pa_clone_data(x)
  pu   <- x$data$pu
  da   <- x$data$dist_actions
  acts <- x$data$actions

  # pu must have id; internal_id can be created if missing (defensive)
  assertthat::assert_that("id" %in% names(pu), msg = "x$data$pu must contain column 'id'.")
  if (!("internal_id" %in% names(pu))) {
    pu$internal_id <- seq_len(nrow(pu))
    x$data$pu <- pu
  }

  assertthat::assert_that(all(c("pu", "action") %in% names(da)), msg = "x$data$dist_actions must contain columns 'pu' and 'action'.")
  assertthat::assert_that("id" %in% names(acts), msg = "x$data$actions must contain column 'id'.")

  # enforce internal_id for actions (defensive)
  if (!("internal_id" %in% names(acts))) {
    acts$internal_id <- seq_len(nrow(acts))
    x$data$actions <- acts
  }

  pu_ids     <- pu$id
  action_ids <- as.character(acts$id)

  # base skeleton: all (pu, action) pairs currently in dist_actions
  base <- da[, c("pu", "action"), drop = FALSE]
  base$action <- as.character(base$action)

  # default profit
  base$profit <- 0

  # ---- fill profit from spec
  if (is.null(profit)) {

    # keep default 0

  } else if (is.numeric(profit) && length(profit) == 1) {

    base$profit <- as.numeric(profit)

  } else if (is.numeric(profit) && !is.null(names(profit))) {

    # named vector by action
    if (!all(names(profit) %in% action_ids)) {
      bad <- setdiff(names(profit), action_ids)
      stop("profit contains unknown action ids: ", paste(bad, collapse = ", "), call. = FALSE)
    }
    base$profit <- as.numeric(profit[base$action])
    base$profit[is.na(base$profit)] <- 0

  } else if (inherits(profit, "data.frame")) {

    p <- profit

    # normalize legacy column naming
    if ("id" %in% names(p) && !("action" %in% names(p))) names(p)[names(p) == "id"] <- "action"

    # Case A: (action, profit)
    if (all(c("action", "profit") %in% names(p)) && !("pu" %in% names(p))) {

      p$action <- as.character(p$action)
      p$profit <- as.numeric(p$profit)

      if (!all(p$action %in% action_ids)) {
        bad <- unique(p$action[!p$action %in% action_ids])
        stop("profit contains unknown action ids: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      if (anyDuplicated(p$action)) {
        stop("profit (action,profit) must have unique action rows.", call. = FALSE)
      }

      base <- dplyr::left_join(base, p, by = "action", suffix = c("", ".new"))
      if ("profit.new" %in% names(base)) {
        base$profit.new[is.na(base$profit.new)] <- 0
        base$profit <- base$profit.new
        base$profit.new <- NULL
      } else {
        base$profit <- as.numeric(base$profit)
      }

      # Case B: (pu, action, profit)
    } else if (all(c("pu", "action", "profit") %in% names(p))) {

      p$pu     <- as.numeric(p$pu)
      p$action <- as.character(p$action)
      p$profit <- as.numeric(p$profit)

      if (!all(p$pu %in% pu_ids)) {
        bad <- unique(p$pu[!p$pu %in% pu_ids])
        stop("profit contains unknown pu ids: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      if (!all(p$action %in% action_ids)) {
        bad <- unique(p$action[!p$action %in% action_ids])
        stop("profit contains unknown action ids: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      tmp <- p[, c("pu", "action")]
      if (nrow(dplyr::distinct(tmp)) != nrow(tmp)) {
        stop("profit has duplicate (pu, action) rows.", call. = FALSE)
      }

      base <- dplyr::left_join(base, p, by = c("pu", "action"), suffix = c("", ".new"))
      if ("profit.new" %in% names(base)) {
        base$profit.new[is.na(base$profit.new)] <- 0
        base$profit <- base$profit.new
        base$profit.new <- NULL
      } else {
        base$profit <- as.numeric(base$profit)
      }

    } else {
      stop("Unsupported profit data.frame format. Use (action,profit) or (pu,action,profit).", call. = FALSE)
    }

  } else {
    stop("Unsupported type for 'profit'. Use NULL, numeric scalar, named numeric vector, or a data.frame.", call. = FALSE)
  }

  # ---- cleanup / validation
  base$profit <- as.numeric(base$profit)
  base$profit[is.na(base$profit)] <- 0

  if (!all(is.finite(base$profit))) stop("profit values must be finite.", call. = FALSE)

  # store only non-zero rows
  base <- base[base$profit != 0, , drop = FALSE]

  # ---- add internal ids
  pu_map   <- x$data$pu[, c("id", "internal_id")]
  acts_map <- x$data$actions[, c("id", "internal_id")]

  base$internal_pu     <- pu_map$internal_id[match(base$pu, pu_map$id)]
  base$internal_action <- acts_map$internal_id[match(base$action, acts_map$id)]

  dist_profit <- base[, c("pu", "action", "profit", "internal_pu", "internal_action"), drop = FALSE]

  # store
  x$data$dist_profit <- dist_profit

  x
}

#' @include internal.R
#'
#' @title Build optimization model from Problem
#'
#' @description
#' Materializes (builds) the optimization model using the current state of the `Problem` object:
#' prepared data tables, stored objective settings, and stored constraints (e.g., targets).
#'
#' @param x Problem object (class "Problem") created with create_problem().
#'
#' @return Updated `Problem` object with model pointer and model snapshot.
#' @keywords internal
.pa_build_model <- function(x) {

  stopifnot(inherits(x, "Problem"))
  stopifnot(!is.null(x$data$pu))
  stopifnot(!is.null(x$data$dist_features))

  # ------------------------------------------------------------
  # local helpers (keep minimal here; you can move to internal.R)
  # ------------------------------------------------------------
  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)
  .pa_has_rows <- function(df) !is.null(df) && inherits(df, "data.frame") && nrow(df) > 0

  .pa_log_init <- function(x) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$build_log <- x$data$meta$build_log %||% list(
      dropped_status3 = 0L,
      dropped_missing_internal = 0L,
      dropped_bad_cost = 0L,
      filtered_effects_rows = 0L,
      filtered_profit_rows = 0L
    )
    x
  }
  .pa_log_add <- function(x, key, n) {
    x$data$meta$build_log[[key]] <- as.integer(x$data$meta$build_log[[key]] %||% 0L) + as.integer(n)
    x
  }
  .pa_emit_build_warnings <- function(x) {
    log <- x$data$meta$build_log %||% list()
    msgs <- character(0)

    if (isTRUE(log$dropped_status3 > 0L)) {
      msgs <- c(msgs, paste0("Dropped ", log$dropped_status3, " dist_actions rows with status==3 (locked-out)."))
    }
    if (isTRUE(log$dropped_missing_internal > 0L)) {
      msgs <- c(msgs, paste0("Dropped ", log$dropped_missing_internal, " dist_actions rows with missing internal_pu/internal_action."))
    }
    if (isTRUE(log$dropped_bad_cost > 0L)) {
      msgs <- c(msgs, paste0("Dropped ", log$dropped_bad_cost, " dist_actions rows with missing/non-finite cost."))
    }
    if (isTRUE(log$filtered_effects_rows > 0L)) {
      msgs <- c(msgs, paste0("Filtered ", log$filtered_effects_rows, " dist_effects rows not matching feasible dist_actions."))
    }
    if (isTRUE(log$filtered_profit_rows > 0L)) {
      msgs <- c(msgs, paste0("Filtered ", log$filtered_profit_rows, " dist_profit rows not matching feasible dist_actions."))
    }

    if (length(msgs) > 0) {
      warning(paste(msgs, collapse = "\n"), call. = FALSE, immediate. = TRUE)
    }
    x
  }

  .pa_check_single_objective <- function(args) {
    if (isTRUE(args$mo_mode)) return(invisible(TRUE))

    if (is.null(args)) return(invisible(TRUE))

    if (!is.null(args$objectives)) {
      nobj <- if (is.data.frame(args$objectives)) nrow(args$objectives) else length(args$objectives)
      if (is.finite(nobj) && nobj > 1) {
        .pa_abort(
          "Multiple objectives detected (", nobj, "). multiscape builds a single-objective MILP.\n",
          "Use a multiobjective method for epsilon-constraint/AUGMECON/interactive methods (e.g., set_method_weighted)."
        )
      }
    }

    if (is.vector(args$model_type) && length(args$model_type) > 1) {
      .pa_abort(
        "Multiple model_type values detected: ",
        paste(args$model_type, collapse = ", "),
        ". Choose exactly one objective."
      )
    }

    invisible(TRUE)
  }

  # ------------------------------------------------------------
  # init build log and args
  # ------------------------------------------------------------
  x <- .pa_log_init(x)

  input_format <- x$data$meta$input_format %||% "new"

  if (is.null(x$data$model_args)) x$data$model_args <- list()
  .pa_check_single_objective(x$data$model_args)

  # objective required for new pipeline (legacy can default)
  if (is.null(x$data$model_args$model_type)) {
    if (identical(input_format, "legacy")) {
      x$data$model_args$model_type <- "minimizeCosts"
    } else {
      .pa_abort(
        "No objective configured. Use add_objective_*() (new pipeline)."
      )
    }
  }

  # defaults for later printing / adapters
  if (is.null(x$data$model_args$budget))   x$data$model_args$budget   <- 0
  if (is.null(x$data$model_args$blm))      x$data$model_args$blm      <- 0
  if (is.null(x$data$model_args$curve))    x$data$model_args$curve    <- 1L
  if (is.null(x$data$model_args$segments)) x$data$model_args$segments <- 3L

  # ---- NEW: initialize + set needs flags (single-objective safe, MO-safe)
  x <- .pa_build_model_set_needs_from_objective(x)

  # ------------------------------------------------------------
  # legacy adapters
  # ------------------------------------------------------------
  # if (identical(input_format, "legacy")) {
  #   x <- .pa_build_model_apply_legacy_adapters(x)
  # }

  # ------------------------------------------------------------
  # ensure tables exist (allow "no actions" case)
  # ------------------------------------------------------------
  x <- .pa_build_model_ensure_tables(x)

  # ------------------------------------------------------------
  # post-adapter validation
  # ------------------------------------------------------------
  x <- .pa_build_model_validate_pipeline_state(x, input_format = input_format)

  # ------------------------------------------------------------
  # prepare model-ready tables (filters + joins)
  # ------------------------------------------------------------
  x <- .pa_build_model_prepare_tables(x)

  x <- .pa_build_model_validate_locked_in_action_feasibility(x)

  # ------------------------------------------------------------
  # early validation: objective dependencies
  # ------------------------------------------------------------
  x <- .pa_build_model_validate_objective_requirements(x)

  # ------------------------------------------------------------
  # build C++ model pointer + base vars + minimal constraints
  # ------------------------------------------------------------
  x <- .pa_build_model_build_cpp_core(x)

  # ------------------------------------------------------------
  # prepare auxiliary variables/constraints required by needs
  # ------------------------------------------------------------
  x <- .pa_build_model_prepare_needs_cpp(x)

  # ------------------------------------------------------------
  # objective (C++ side)
  # ------------------------------------------------------------
  x <- .pa_build_model_set_objective_cpp(x)

  # ------------------------------------------------------------
  # apply stored targets/constraints
  # ------------------------------------------------------------
  x <- .pa_build_model_apply_constraints(x)

  # ------------------------------------------------------------
  # store metadata + snapshot + warnings
  # ------------------------------------------------------------
  x <- .pa_build_model_finalize(x)
  x <- .pa_emit_build_warnings(x)

  x$data$has_model <- TRUE
  x
}

# -------------------------------------------------------------------------
# Helpers (recommended to live in internal.R, but ok here as noRd)
# -------------------------------------------------------------------------

.pa_build_model_ensure_tables <- function(x) {

  if (is.null(x$data$dist_actions)) {
    x$data$dist_actions <- data.frame(
      pu = integer(0),
      action = character(0),
      cost = numeric(0),
      status = integer(0),
      internal_pu = integer(0),
      internal_action = integer(0),
      stringsAsFactors = FALSE
    )
  }
  if (is.null(x$data$actions)) {
    x$data$actions <- data.frame(
      id = character(0),
      internal_id = integer(0),
      stringsAsFactors = FALSE
    )
  }
  x
}

.pa_build_model_validate_pipeline_state <- function(x, input_format) {

  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)

  mtype_check <- x$data$model_args$model_type %||% "minimizeCosts"

  if (!identical(input_format, "legacy") && identical(mtype_check, "minimizeCosts")) {
    if (is.null(x$data$targets) || !inherits(x$data$targets, "data.frame") || nrow(x$data$targets) == 0) {
      .pa_abort(
        "model_type='minimizeCosts' requires targets, but x$data$targets is empty.\n",
        "Run add_constraint_target_*() before solve()."
      )
    }
  }

  x
}

.pa_build_model_prepare_tables <- function(x) {

  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)
  .has_rows <- function(df) !is.null(df) && inherits(df, "data.frame") && nrow(df) > 0

  da <- x$data$dist_actions
  if (is.null(da)) da <- data.frame()

  if (!inherits(da, "data.frame")) {
    .pa_abort("dist_actions must be a data.frame (or NULL).")
  }

  n0 <- nrow(da)

  if (n0 > 0) {

    # status filter
    if ("status" %in% names(da)) {
      keep <- da$status != 3L
      dropped <- sum(!keep, na.rm = TRUE)
      if (dropped > 0) x <- .pa_log_add(x, "dropped_status3", dropped)
      da <- da[keep, , drop = FALSE]
    }

    # internal ids
    if (all(c("internal_pu", "internal_action") %in% names(da))) {
      keep <- !is.na(da$internal_pu) & !is.na(da$internal_action)
      dropped <- sum(!keep)
      if (dropped > 0) x <- .pa_log_add(x, "dropped_missing_internal", dropped)
      da <- da[keep, , drop = FALSE]
    }

    # costs
    if ("cost" %in% names(da)) {
      da$cost <- as.numeric(da$cost)
      keep <- is.finite(da$cost) & !is.na(da$cost)
      dropped <- sum(!keep)
      if (dropped > 0) x <- .pa_log_add(x, "dropped_bad_cost", dropped)
      da <- da[keep, , drop = FALSE]
    }

    # normalize types
    if ("pu" %in% names(da)) da$pu <- as.integer(da$pu)
    if ("action" %in% names(da)) da$action <- as.character(da$action)
    if ("internal_pu" %in% names(da)) da$internal_pu <- as.integer(da$internal_pu)
    if ("internal_action" %in% names(da)) da$internal_action <- as.integer(da$internal_action)

    # deterministic order
    if (all(c("internal_pu","internal_action") %in% names(da)) && nrow(da) > 0) {
      da <- da[order(da$internal_pu, da$internal_action), , drop = FALSE]
    }
  }

  # internal_row
  da$internal_row <- seq_len(nrow(da))

  # validate internal id ranges (if present)
  if (.has_rows(da) && all(c("internal_pu","internal_action") %in% names(da))) {
    n_pu <- nrow(x$data$pu)
    if (any(da$internal_pu < 1L | da$internal_pu > n_pu, na.rm = TRUE)) {
      .pa_abort("dist_actions contains internal_pu indices out of range (1..n_pu).")
    }
    if (!is.null(x$data$actions) && inherits(x$data$actions, "data.frame") && nrow(x$data$actions) > 0) {
      n_a <- nrow(x$data$actions)
      if (any(da$internal_action < 1L | da$internal_action > n_a, na.rm = TRUE)) {
        .pa_abort("dist_actions contains internal_action indices out of range (1..n_actions).")
      }
    }
  }

  # duplicates check (pu, action)
  if (.has_rows(da) && all(c("pu","action") %in% names(da))) {
    key <- paste(da$pu, da$action, sep = "||")
    if (anyDuplicated(key)) {
      .pa_abort(
        "dist_actions has duplicated (pu, action) pairs. Please aggregate/resolve duplicates before building the model."
      )
    }
  }

  x$data$dist_actions_model <- da

  # dist_effects model-ready
  de <- x$data$dist_effects

  if (is.null(de)) {
    de <- data.frame(
      pu = integer(0),
      action = character(0),
      feature = integer(0),
      benefit = numeric(0),
      loss = numeric(0),
      internal_pu = integer(0),
      internal_action = integer(0),
      internal_feature = integer(0),
      stringsAsFactors = FALSE
    )
  }

  if (!inherits(de, "data.frame")) {
    .pa_abort("dist_effects must be a data.frame (or NULL).")
  }

  if (.has_rows(de) && .has_rows(da)) {

    n_before <- nrow(de)

    if ("pu" %in% names(de))      de$pu      <- as.integer(de$pu)
    if ("action" %in% names(de))  de$action  <- as.character(de$action)
    if ("feature" %in% names(de)) de$feature <- as.integer(de$feature)

    de <- dplyr::inner_join(
      de,
      da[, c("pu", "action"), drop = FALSE],
      by = c("pu", "action")
    )

    n_after <- nrow(de)
    dropped <- n_before - n_after
    if (dropped > 0) x <- .pa_log_add(x, "filtered_effects_rows", dropped)

    pu_map <- x$data$pu[, c("id", "internal_id")]
    act_map <- x$data$actions[, c("id", "internal_id")]
    feat_map <- x$data$features[, c("id", "internal_id")]

    if (!("internal_pu" %in% names(de))) {
      de$internal_pu <- pu_map$internal_id[match(de$pu, pu_map$id)]
    } else {
      de$internal_pu <- as.integer(de$internal_pu)
    }

    if (!("internal_action" %in% names(de))) {
      de$internal_action <- act_map$internal_id[match(de$action, act_map$id)]
    } else {
      de$internal_action <- as.integer(de$internal_action)
    }

    if (!("internal_feature" %in% names(de))) {
      if (!("feature" %in% names(de))) {
        .pa_abort("dist_effects must contain column 'feature' (feature id) to derive internal_feature.")
      }
      de$internal_feature <- feat_map$internal_id[match(de$feature, feat_map$id)]
    } else {
      de$internal_feature <- as.integer(de$internal_feature)
    }

    if (anyNA(de$internal_pu)) {
      .pa_abort("dist_effects has pu ids not found in x$data$pu$id (cannot derive internal_pu).")
    }
    if (anyNA(de$internal_action)) {
      .pa_abort("dist_effects has action ids not found in x$data$actions$id (cannot derive internal_action).")
    }
    if (anyNA(de$internal_feature)) {
      .pa_abort("dist_effects has feature ids not found in x$data$features$id (cannot derive internal_feature).")
    }
  }

  if (.has_rows(de)) {
    de <- .pa_add_feature_labels(
      df = de,
      features_df = x$data$features,
      feature_col = "feature",
      internal_feature_col = "internal_feature",
      out_col = "feature_name"
    )

    de <- .pa_add_action_labels(
      df = de,
      actions_df = x$data$actions,
      action_col = "action",
      internal_action_col = "internal_action",
      out_col = "action_name"
    )
  }

  x$data$dist_effects_model <- de



  # dist_benefit_model derived (now has internal_* too)
  db <- NULL
  if (.has_rows(de) && "benefit" %in% names(de)) {
    db <- de
    db$benefit <- as.numeric(db$benefit)
  }
  x$data$dist_benefit_model <- db

  # dist_profit model-ready (MUST carry internal_pu/internal_action)
  # dist_profit model-ready (MUST carry internal_pu/internal_action)
  dp <- x$data$dist_profit
  if (.has_rows(dp) && .has_rows(da)) {

    n_before <- nrow(dp)

    # tipos base
    if ("pu" %in% names(dp)) dp$pu <- as.integer(dp$pu)
    if ("action" %in% names(dp)) dp$action <- as.character(dp$action)

    # filtra a (pu, action) factibles
    dp <- dplyr::inner_join(
      dp,
      da[, c("pu", "action"), drop = FALSE],
      by = c("pu", "action")
    )

    n_after <- nrow(dp)
    dropped <- n_before - n_after
    if (dropped > 0) x <- .pa_log_add(x, "filtered_profit_rows", dropped)

    # --- NEW: asegurar columnas internas (igual que en dist_effects)
    pu_map  <- x$data$pu[, c("id", "internal_id")]
    act_map <- x$data$actions[, c("id", "internal_id")]

    pu_map$id  <- as.integer(pu_map$id)
    act_map$id <- as.character(act_map$id)

    if (!("internal_pu" %in% names(dp))) {
      dp$internal_pu <- pu_map$internal_id[match(dp$pu, pu_map$id)]
    } else {
      dp$internal_pu <- as.integer(dp$internal_pu)
    }

    if (!("internal_action" %in% names(dp))) {
      dp$internal_action <- act_map$internal_id[match(dp$action, act_map$id)]
    } else {
      dp$internal_action <- as.integer(dp$internal_action)
    }

    if (anyNA(dp$internal_pu)) {
      .pa_abort("dist_profit has pu ids not found in x$data$pu$id (cannot derive internal_pu).")
    }
    if (anyNA(dp$internal_action)) {
      .pa_abort("dist_profit has action ids not found in x$data$actions$id (cannot derive internal_action).")
    }

    # opcional: orden determinista
    dp <- dp[order(dp$internal_pu, dp$internal_action), , drop = FALSE]
  }

  x$data$dist_profit_model <- dp


  x
}

.pa_build_model_validate_objective_requirements <- function(x) {

  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)
  .has_rows <- function(df) !is.null(df) && inherits(df, "data.frame") && nrow(df) > 0

  args <- x$data$model_args %||% list()

  if (is.null(args$model_type) || !nzchar(as.character(args$model_type)[1])) {
    .pa_abort("No active objective has been defined in x$data$model_args$model_type.")
  }

  mtype <- as.character(args$model_type)[1]
  oargs <- args$objective_args %||% list()

  has_actions_model <- .has_rows(x$data$dist_actions_model)
  has_effects_model <- .has_rows(x$data$dist_effects_model)
  has_targets <- !is.null(x$data$targets) &&
    inherits(x$data$targets, "data.frame") &&
    nrow(x$data$targets) > 0


  needs_actions <- mtype %in% c(
    "maximizeBenefits",
    "minimizeLosses",
    "maximizeProfit",
    "maximizeNetProfit",
    "minimizeActionFragmentation",
    "minimizeInterventionFragmentation",
    "minimizeInterventionImpact"
  )

  if (needs_actions && !isTRUE(has_actions_model)) {
    .pa_abort(
      "Objective '", mtype, "' requires actions, but no actions are available.\n",
      "Run add_actions() (and ensure feasible dist_actions rows exist) before solve()."
    )
  }

  # ------------------------------------------------------------
  # Actions + targets over features require effects
  # ------------------------------------------------------------
  if (isTRUE(has_actions_model) && isTRUE(has_targets) && !isTRUE(has_effects_model)) {
    .pa_abort(
      "This problem includes actions and feature targets, but no action effects were provided.\n",
      "Because targets are defined on features, multiscape needs to know how each action affects each feature.\n",
      "Run add_effects() after add_actions() before solve()."
    )
  }

  if (identical(mtype, "maximizeBenefits")) {

    if (!.has_rows(x$data$dist_effects_model)) {
      .pa_abort(
        "Objective 'maximizeBenefits' requires effects/benefits, but dist_effects is empty.\n",
        "Run add_benefits() / add_effects() after add_actions()."
      )
    }

    bcol <- as.character(oargs$benefit_col %||% "benefit")[1]

    if (!(bcol %in% names(x$data$dist_effects_model))) {
      .pa_abort(
        "Objective 'maximizeBenefits' requires column '", bcol, "' in dist_effects."
      )
    }

    de <- x$data$dist_effects_model

    # Apply the same action/feature filters that are later used by
    # .pa_build_model_set_objective_cpp().
    acts <- oargs$actions %||% NULL
    if (!is.null(acts)) {
      de <- de[de$internal_action %in% as.integer(acts), , drop = FALSE]
    }

    feats <- oargs$features %||% NULL
    if (!is.null(feats)) {
      de <- de[de$internal_feature %in% as.integer(feats), , drop = FALSE]
    }

    obj_alias <- .pa_active_objective_alias(x) %||% "maximizeBenefits"

    if (!.has_rows(de)) {

      feature_msg <- ""
      if (!is.null(feats)) {
        feature_names <- .pa_feature_names_from_internal_ids(x, as.integer(feats))
        feature_msg <- paste0(
          "\nSelected feature(s): ",
          paste(feature_names, collapse = ", "),
          "."
        )
      }

      action_msg <- ""
      if (!is.null(acts)) {
        action_names <- .pa_action_names_from_internal_ids(x, as.integer(acts))
        action_msg <- paste0(
          "\nSelected action(s): ",
          paste(action_names, collapse = ", "),
          "."
        )
      }

      .pa_abort(
        "Objective '", obj_alias, "' has no matching effect rows after applying ",
        "the selected action/feature filters.",
        feature_msg,
        action_msg,
        "\nBenefit objectives are based on action effects, not directly on baseline feature values.",
        "\nCheck that the selected feature(s) appear in add_effects() with positive non-zero effects."
      )
    }

    bb <- as.numeric(de[[bcol]])

    if (!any(is.finite(bb) & bb > .Machine$double.eps, na.rm = TRUE)) {

      feature_msg <- ""
      if (!is.null(feats)) {
        feature_names <- .pa_feature_names_from_internal_ids(x, as.integer(feats))
        feature_msg <- paste0(
          "\nSelected feature(s): ",
          paste(feature_names, collapse = ", "),
          "."
        )
      }

      action_msg <- ""
      if (!is.null(acts)) {
        action_names <- .pa_action_names_from_internal_ids(x, as.integer(acts))
        action_msg <- paste0(
          "\nSelected action(s): ",
          paste(action_names, collapse = ", "),
          "."
        )
      }

      .pa_abort(
        "Objective '", obj_alias, "' has no positive non-zero benefit coefficients.",
        feature_msg,
        action_msg,
        "\nThis objective cannot be used as a benefit objective because all selected effects are zero, missing, or non-positive.",
        "\nFor add_objective_max_benefit(), the selected feature(s) must have positive action effects in add_effects()."
      )
    }
  }

  if (identical(mtype, "minimizeLosses")) {
    if (!.has_rows(x$data$dist_effects_model)) {
      .pa_abort(
        "Objective 'minimizeLosses' requires effects/losses, but dist_effects is empty.\n",
        "Run add_losses() / add_effects() after add_actions()."
      )
    }
    lcol <- as.character(oargs$loss_col %||% "loss")[1]
    if (!(lcol %in% names(x$data$dist_effects_model))) {
      .pa_abort(
        "Objective 'minimizeLosses' requires column '", lcol, "' in dist_effects."
      )
    }
  }

  if (mtype %in% c("maximizeProfit", "maximizeNetProfit")) {
    if (!.has_rows(x$data$dist_profit_model)) {
      .pa_abort(
        "Objective '", mtype, "' requires dist_profit, but dist_profit is empty.\n",
        "Run add_profit() after add_actions()."
      )
    }
    pcol <- as.character(oargs$profit_col %||% "profit")[1]
    if (!(pcol %in% names(x$data$dist_profit_model))) {
      .pa_abort(
        "Objective '", mtype, "' requires column '", pcol, "' in dist_profit."
      )
    }
  }

  # ------------------------------------------------------------
  # Fragmentation objectives: require spatial relation
  # ------------------------------------------------------------
  if (mtype %in% c(
    "minimizeFragmentation",
    "minimizeActionFragmentation",
    "minimizeInterventionFragmentation"
  )) {

    rel_name <- as.character(oargs$relation_name %||% "boundary")[1]
    rels <- x$data$spatial_relations

    if (is.null(rels) || !is.list(rels) || is.null(rels[[rel_name]])) {
      .pa_abort(
        "Objective '", mtype, "' requires a spatial relation named '", rel_name, "'.\n",
        "But x$data$spatial_relations[['", rel_name, "']] is missing."
      )
    }

    rel <- rels[[rel_name]]
    if (!inherits(rel, "data.frame") || nrow(rel) == 0) {
      .pa_abort("Spatial relation '", rel_name, "' exists but is empty or not a data.frame.")
    }

    need_cols <- c("internal_pu1", "internal_pu2", "weight")
    if (!all(need_cols %in% names(rel))) {
      .pa_abort(
        "Spatial relation '", rel_name, "' must contain columns: ",
        paste(need_cols, collapse = ", "), "."
      )
    }

    n_pu <- nrow(x$data$pu)
    if (anyNA(rel$internal_pu1) || anyNA(rel$internal_pu2)) {
      .pa_abort("Spatial relation '", rel_name, "' contains NA in internal_pu1/internal_pu2.")
    }
    if (any(rel$internal_pu1 < 1L | rel$internal_pu1 > n_pu) ||
        any(rel$internal_pu2 < 1L | rel$internal_pu2 > n_pu)) {
      .pa_abort("Spatial relation '", rel_name, "' has internal PU indices out of range (1..n_pu).")
    }

    ww <- as.numeric(rel$weight)
    # if (any(!is.finite(ww)) || any(ww < 0)) {
    #   .pa_abort("Spatial relation '", rel_name, "' has non-finite or negative weights.")
    # }

    if (identical(mtype, "minimizeActionFragmentation")) {
      if (is.null(x$data$actions) || !inherits(x$data$actions, "data.frame") || nrow(x$data$actions) == 0) {
        .pa_abort("Objective 'minimizeActionFragmentation' requires x$data$actions to exist (non-empty).")
      }
      if (!is.null(oargs$actions_to_use)) {
        a <- as.integer(oargs$actions_to_use)
        if (anyNA(a)) .pa_abort("actions_to_use contains NA.")
        if (any(a < 1L | a > nrow(x$data$actions))) {
          .pa_abort("actions_to_use has internal_action ids out of range (1..n_actions).")
        }
      }
      # if (!is.null(oargs$action_weights)) {
      #   aw <- as.numeric(oargs$action_weights)
      #   if (any(!is.finite(aw)) || any(aw < 0)) {
      #     .pa_abort("action_weights must be finite and >= 0.")
      #   }
      # }
    }
  }

  # ------------------------------------------------------------
  # Intervention impact: does NOT require spatial relation
  # ------------------------------------------------------------
  if (identical(mtype, "minimizeInterventionImpact")) {

    if (!.has_rows(x$data$dist_features)) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact' requires dist_features, but dist_features is empty."
      )
    }

    icol <- as.character(oargs$impact_col %||% "amount")[1]
    if (!(icol %in% names(x$data$dist_features))) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact' requires column '", icol, "' in dist_features."
      )
    }

    if (is.null(x$data$actions) || !inherits(x$data$actions, "data.frame") || nrow(x$data$actions) == 0) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact' requires x$data$actions to exist (non-empty)."
      )
    }

    acts <- oargs$actions %||% NULL
    if (is.null(acts)) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact' requires `actions=` in objective_args.\n",
        "Example: add_objective_min_intervention_impact(actions = c('restoration'))."
      )
    }

    act_subset <- tryCatch(
      .pa_resolve_action_subset(x, subset = acts),
      error = function(e) .pa_abort(conditionMessage(e))
    )

    if (!inherits(act_subset, "data.frame") || nrow(act_subset) == 0) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact': action subset resolved to zero actions."
      )
    }

    feats <- oargs$features %||% NULL
    if (!is.null(feats)) {
      feat_subset <- tryCatch(
        .pa_resolve_feature_subset(x, features = feats),
        error = function(e) .pa_abort(conditionMessage(e))
      )

      if (!inherits(feat_subset, "data.frame") || nrow(feat_subset) == 0) {
        .pa_abort(
          "Objective 'minimizeInterventionImpact': feature subset resolved to zero features."
        )
      }
    }
  }

  x
}

.pa_build_model_validate_locked_in_action_feasibility <- function(x) {

  stopifnot(inherits(x, "Problem"))

  pu <- x$data$pu
  da <- x$data$dist_actions_model

  if (is.null(pu) || !inherits(pu, "data.frame") || nrow(pu) == 0) {
    return(x)
  }

  if (!("locked_in" %in% names(pu)) || !any(pu$locked_in, na.rm = TRUE)) {
    return(x)
  }

  locked_in_pu <- as.integer(pu$internal_id[isTRUE(pu$locked_in) | (!is.na(pu$locked_in) & pu$locked_in)])

  if (length(locked_in_pu) == 0) return(x)

  feasible_pu <- integer(0)
  if (!is.null(da) && inherits(da, "data.frame") && nrow(da) > 0 && "internal_pu" %in% names(da)) {
    feasible_pu <- unique(as.integer(da$internal_pu))
  }

  bad <- setdiff(locked_in_pu, feasible_pu)

  if (length(bad) > 0) {
    bad_ext <- pu$id[match(bad, pu$internal_id)]
    stop(
      "Some locked-in planning units have no feasible actions, but the model requires w_i = 1 ",
      "to imply at least one action.\n",
      "Affected PU ids: ", paste(utils::head(bad_ext, 20), collapse = ", "),
      if (length(bad_ext) > 20) paste0(" ... (", length(bad_ext), " total)") else "",
      "\nFix this by either:\n",
      "- adding at least one feasible action to those units,\n",
      "- removing locked_in from those units, or\n",
      "- making 'Conservation' an explicit feasible action there.",
      call. = FALSE
    )
  }

  x
}

.pa_build_model_build_cpp_core <- function(x) {


  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)

  if (!exists("rcpp_new_optimization_problem", mode = "function")) {
    .pa_abort("Missing rcpp_new_optimization_problem() in the package.")
  }
  if (!exists("rcpp_add_base_variables", mode = "function")) {
    .pa_abort("Missing rcpp_add_base_variables() in the package.")
  }

  # ---- NEW: read needs flags (default-safe)
  args  <- x$data$model_args %||% list()
  needs <- args$needs %||% list()
  need_z <- isTRUE(needs$z)

  op <- rcpp_new_optimization_problem()

  # registry placeholder for future MO updates (constraint/objective IDs)
  x$data$model_registry <- list(
    cons = list(),
    vars = list(),
    obj_templates = list(),
    objective = list()
  )

  idx <- rcpp_add_base_variables(
    op,
    pu_data            = x$data$pu,
    dist_actions_data  = x$data$dist_actions_model,
    dist_features_data = x$data$dist_features,
    add_z              = need_z            # <- NEW
  )

  # structural constraints
  if (!is.null(x$data$dist_actions_model) && nrow(x$data$dist_actions_model) > 0) {

    if (!exists("rcpp_add_linking_x_le_w", mode = "function")) {
      .pa_abort("Missing rcpp_add_linking_x_le_w() in the package.")
    }
    res_locks <- rcpp_add_linking_x_le_w(op, x$data$dist_actions_model)
    x$data$model_registry$cons$x_le_w <- res_locks

    # NEW: w_i <= sum_a x_ia
    if (!exists("rcpp_add_linking_w_le_sum_x", mode = "function")) {
      .pa_abort("Missing rcpp_add_linking_w_le_sum_x() in the package.")
    }
    res_locks <- rcpp_add_linking_w_le_sum_x(op, x$data$dist_actions_model)
    x$data$model_registry$cons$w_le_sum_x <- res_locks
  }

  # only add z <= w if z exists
  if (isTRUE(need_z) && exists("rcpp_add_linking_z_le_w", mode = "function")) {
    res_locks <- rcpp_add_linking_z_le_w(op, x$data$dist_features)
    x$data$model_registry$cons$z_le_w <- res_locks
  }

  if (exists("rcpp_add_pu_locks", mode = "function")) {
    res_locks <- rcpp_add_pu_locks(op, x$data$pu)
    x$data$model_registry$cons$pu_locks <- res_locks
  }

  if (!is.null(x$data$dist_actions_model) && nrow(x$data$dist_actions_model) > 0 &&
      exists("rcpp_add_action_locks", mode = "function")) {

    res_locks <- rcpp_add_action_locks(op, x$data$dist_actions_model)
    x$data$model_registry$cons$action_locks <- res_locks
  }

  x$data$model_ptr   <- op
  x$data$model_index <- idx

  x
}


.pa_build_model_set_objective_cpp <- function(x) {

  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)

  .pa_prepare_relation_model <- function(rel) {
    rel <- rel[, c(
      "internal_pu1","internal_pu2","weight",
      intersect(names(rel), c("distance","source","relation_name"))
    ), drop = FALSE]
    rel$internal_pu1 <- as.integer(rel$internal_pu1)
    rel$internal_pu2 <- as.integer(rel$internal_pu2)
    rel$weight <- as.numeric(rel$weight)
    rel <- rel[order(rel$internal_pu1, rel$internal_pu2), , drop = FALSE]
    rel$internal_edge <- seq_len(nrow(rel))
    rel
  }

  op <- x$data$model_ptr
  if (is.null(op)) .pa_abort("model_ptr is NULL. Build core model first.")

  args  <- x$data$model_args %||% list()
  mtype <- args$model_type %||% "minimizeCosts"
  oargs <- args$objective_args %||% list()

  # ------------------------------------------------------------
  # Decide modelsense for the atomic objective we are activating
  # ------------------------------------------------------------
  modelsense <- if (mtype %in% c("minimizeCosts",
                                 "minimizeLosses",
                                 "minimizeFragmentation",
                                 "minimizeActionFragmentation",
                                 "minimizeInterventionFragmentation",
                                 "minimizeInterventionImpact")) "min" else "max"

  if (!exists("rcpp_reset_objective", mode = "function")) {
    .pa_abort("Missing rcpp_reset_objective() in the package.")
  }
  rcpp_reset_objective(op, modelsense)


  # ------------------------------------------------------------
  # ONE atomic objective (prepare + add)
  # ------------------------------------------------------------
  if (identical(mtype, "minimizeCosts")) {

    if (!exists("rcpp_prepare_objective_min_cost", mode = "function")) .pa_abort("Missing rcpp_prepare_objective_min_cost().")
    if (!exists("rcpp_add_objective_min_cost",     mode = "function")) .pa_abort("Missing rcpp_add_objective_min_cost().")

    rcpp_prepare_objective_min_cost(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      include_pu_cost = isTRUE(oargs$include_pu_cost %||% TRUE),
      include_action_cost = isTRUE(oargs$include_action_cost %||% TRUE),
      block_name = "objective_min_cost",
      tag = as.character(oargs$tag %||% "")[1]
    )

    res <- rcpp_add_objective_min_cost(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      include_pu_cost = isTRUE(oargs$include_pu_cost %||% TRUE),
      include_action_cost = isTRUE(oargs$include_action_cost %||% TRUE),
      weight = 1.0
    )

    objective_id <- "min_cost"

  } else if (identical(mtype, "maximizeBenefits")) {

    if (!exists("rcpp_prepare_objective_max_benefit", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_max_benefit().")
    }
    if (!exists("rcpp_add_objective_max_benefit", mode = "function")) {
      .pa_abort("Missing rcpp_add_objective_max_benefit().")
    }

    de <- x$data$dist_effects_model
    if (is.null(de) || !inherits(de, "data.frame")) {
      .pa_abort("Missing x$data$dist_effects_model for objective 'maximizeBenefits'.")
    }

    for (nm in c("internal_pu", "internal_action", "internal_feature", "benefit")) {
      if (!(nm %in% names(de))) {
        .pa_abort("dist_effects_model must contain column '", nm, "'.")
      }
    }

    # filter action subset if requested
    acts <- oargs$actions %||% NULL
    if (!is.null(acts)) {
      de <- de[de$internal_action %in% as.integer(acts), , drop = FALSE]
    }

    # filter feature subset if requested
    feats <- oargs$features %||% NULL
    if (!is.null(feats)) {
      de <- de[de$internal_feature %in% as.integer(feats), , drop = FALSE]
    }

    prep <- rcpp_prepare_objective_max_benefit(
      op,
      dist_actions_data = x$data$dist_actions_model,
      dist_effects_data = de,
      block_name = "objective_max_benefit",
      tag = as.character(oargs$tag %||% "")[1]
    )

    coef_x <- as.numeric(prep$coef_x)

    if (is.null(coef_x) ||
        length(coef_x) == 0L ||
        !any(is.finite(coef_x) & abs(coef_x) > .Machine$double.eps, na.rm = TRUE)) {

      obj_alias <- .pa_active_objective_alias(x) %||% "maximizeBenefits"

      .pa_abort(
        "Objective '", obj_alias, "' produced an empty or zero objective vector.",
        "\nThis usually means that the selected feature(s) have no positive non-zero effects ",
        "for the feasible action set.",
        "\nCheck add_effects(), especially the `feature` column and the selected `features=` argument."
      )
    }

    res <- rcpp_add_objective_max_benefit(
      op,
      coef_x = coef_x,
      weight = 1.0,
      block_name = "objective_max_benefit",
      tag = as.character(oargs$tag %||% "")[1]
    )

    objective_id <- "max_benefit"

  } else if (identical(mtype, "minimizeLosses")) {

    if (!exists("rcpp_prepare_objective_min_loss", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_min_loss().")
    }
    if (!exists("rcpp_add_objective_min_loss", mode = "function")) {
      .pa_abort("Missing rcpp_add_objective_min_loss().")
    }

    de <- x$data$dist_effects_model
    if (is.null(de) || !inherits(de, "data.frame")) {
      .pa_abort("Missing x$data$dist_effects_model for objective 'minimizeLosses'.")
    }

    for (nm in c("internal_pu", "internal_action", "internal_feature", "loss")) {
      if (!(nm %in% names(de))) {
        .pa_abort("dist_effects_model must contain column '", nm, "'.")
      }
    }

    # filter action subset if requested
    acts <- oargs$actions %||% NULL
    if (!is.null(acts)) {
      de <- de[de$internal_action %in% as.integer(acts), , drop = FALSE]
    }

    # filter feature subset if requested
    feats <- oargs$features %||% NULL
    if (!is.null(feats)) {
      de <- de[de$internal_feature %in% as.integer(feats), , drop = FALSE]
    }

    prep <- rcpp_prepare_objective_min_loss(
      x = op,
      dist_actions_data = x$data$dist_actions_model,
      dist_effects_data = de,
      block_name = "objective_min_loss",
      tag = as.character(oargs$tag %||% "")[1]
    )

    res <- rcpp_add_objective_min_loss(
      x = op,
      coef_x = as.numeric(prep$coef_x),
      weight = 1.0,
      block_name = "objective_min_loss",
      tag = as.character(oargs$tag %||% "")[1]
    )

    objective_id <- "min_loss"

  } else if (identical(mtype, "maximizeProfit")) {

    if (!exists("rcpp_prepare_objective_max_profit", mode = "function")) .pa_abort("Missing rcpp_prepare_objective_max_profit().")
    if (!exists("rcpp_add_objective_max_profit",     mode = "function")) .pa_abort("Missing rcpp_add_objective_max_profit().")

    pcol <- as.character(oargs$profit_col %||% "profit")[1]

    rcpp_prepare_objective_max_profit(
      op,
      dist_actions_data = x$data$dist_actions_model,
      dist_profit_data = x$data$dist_profit_model,
      profit_col = pcol,
      block_name = "objective_max_profit",
      tag = as.character(oargs$tag %||% "")[1]
    )

    res <- rcpp_add_objective_max_profit(
      op,
      dist_actions_data = x$data$dist_actions_model,
      dist_profit_data = x$data$dist_profit_model,
      profit_col = pcol,
      weight = 1.0
    )

    objective_id <- "max_profit"

  } else if (identical(mtype, "maximizeNetProfit")) {

    if (!exists("rcpp_prepare_objective_max_net_profit", mode = "function")) .pa_abort("Missing rcpp_prepare_objective_max_net_profit().")
    if (!exists("rcpp_add_objective_max_net_profit",     mode = "function")) .pa_abort("Missing rcpp_add_objective_max_net_profit().")

    pcol <- as.character(oargs$profit_col %||% "profit")[1]

    rcpp_prepare_objective_max_net_profit(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      dist_profit_data = x$data$dist_profit_model,
      profit_col = pcol,
      include_pu_cost = isTRUE(oargs$include_pu_cost %||% TRUE),
      include_action_cost = isTRUE(oargs$include_action_cost %||% TRUE),
      block_name = "objective_max_net_profit",
      tag = as.character(oargs$tag %||% "")[1]
    )

    res <- rcpp_add_objective_max_net_profit(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      dist_profit_data = x$data$dist_profit_model,
      profit_col = pcol,
      include_pu_cost = isTRUE(oargs$include_pu_cost %||% TRUE),
      include_action_cost = isTRUE(oargs$include_action_cost %||% TRUE),
      weight = 1.0
    )

    objective_id <- "max_net_profit"


  } else if (identical(mtype, "minimizeActionFragmentation")) {

    if (!exists("rcpp_add_objective_min_fragmentation_actions", mode = "function")) {
      .pa_abort("Missing rcpp_add_objective_min_fragmentation_actions().")
    }

    rel_name  <- as.character(oargs$relation_name %||% "boundary")[1]
    rel       <- x$data$spatial_relations[[rel_name]]
    if (is.null(rel)) .pa_abort("Missing spatial relation: ", rel_name)

    rel_model <- .pa_prepare_relation_model(rel)
    x$data$spatial_relations_model <- x$data$spatial_relations_model %||% list()
    x$data$spatial_relations_model[[rel_name]] <- rel_model

    subset_actions_raw <- oargs$actions_to_use %||% oargs$actions %||% NULL

    act_subset <- NULL
    actions_to_use <- NULL
    if (!is.null(subset_actions_raw)) {
      act_subset <- .pa_resolve_action_subset(x, subset = subset_actions_raw)
      actions_to_use <- as.integer(act_subset$internal_id)
    }

    aw_vec <- .pa_action_weights_vector(
      actions_df = x$data$actions,
      action_weights = oargs$action_weights %||% NULL,
      subset_actions = actions_to_use,
      default_weight = 1
    )

    res <- rcpp_add_objective_min_fragmentation_actions(
      op,
      dist_actions_data = x$data$dist_actions_model,
      relation_data     = rel_model,
      actions_to_use    = actions_to_use,
      action_weights    = aw_vec,
      weight_multiplier = as.numeric(oargs$weight_multiplier %||% 1)[1]
    )

    objective_id <- "min_action_fragmentation"

  } else if (identical(mtype, "minimizeFragmentation")) {

    if (!exists("rcpp_add_objective_min_fragmentation", mode = "function")) .pa_abort("Missing rcpp_add_objective_min_fragmentation().")

    rel_name  <- as.character(oargs$relation_name %||% "boundary")[1]
    rel       <- x$data$spatial_relations[[rel_name]]
    if (is.null(rel)) .pa_abort("Missing spatial relation: ", rel_name)

    rel_model <- x$data$spatial_relations_model[[rel_name]] %||% .pa_prepare_relation_model(rel)
    x$data$spatial_relations_model <- x$data$spatial_relations_model %||% list()
    x$data$spatial_relations_model[[rel_name]] <- rel_model

    # IMPORTANT: prepare of aux vars (idempotent)
    if (exists("rcpp_prepare_objective_min_fragmentation", mode = "function")) {
      rcpp_prepare_objective_min_fragmentation(op, rel_model)
    }

    res <- rcpp_add_objective_min_fragmentation(
      op,
      relation_data = rel_model,
      weight = 1.0 ,
      weight_multiplier = as.numeric(oargs$weight_multiplier %||% 1)[1]
    )

    objective_id <- "min_fragmentation"

  } else if (identical(mtype, "minimizeInterventionImpact")) {

    if (!exists("rcpp_prepare_objective_min_intervention_impact", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_min_intervention_impact().")
    }
    if (!exists("rcpp_add_objective_min_intervention_impact", mode = "function")) {
      .pa_abort("Missing rcpp_add_objective_min_intervention_impact().")
    }

    icol <- as.character(oargs$impact_col %||% "amount")[1]

    # ---- feature subset
    feats <- oargs$features %||% NULL
    feats_internal <- integer(0)
    if (!is.null(feats)) {
      m <- match(feats, x$data$features$id)
      if (anyNA(m)) {
        .pa_abort(
          "minimizeInterventionImpact: some features not in x$data$features$id: ",
          paste(feats[is.na(m)], collapse = ", ")
        )
      }
      feats_internal <- as.integer(x$data$features$internal_id[m])
    }

    # ---- action subset (REQUIRED for option 2)
    acts <- oargs$actions %||% NULL
    if (is.null(acts)) {
      .pa_abort(
        "Objective 'minimizeInterventionImpact' now requires `actions=`.\n",
        "This objective is defined over intervention-group variables linked to a subset of actions."
      )
    }

    act_subset <- .pa_resolve_action_subset(x, subset = acts)
    acts_internal <- as.integer(act_subset$internal_id)

    if (length(acts_internal) == 0) {
      .pa_abort("minimizeInterventionImpact: action subset resolved to zero actions.")
    }

    # optional identifier for the action subset/group
    subset_key <- .pa_subset_to_string(act_subset$id)

    prep <- rcpp_prepare_objective_min_intervention_impact(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      dist_features_data = x$data$dist_features,
      actions_to_use = acts_internal,
      subset_key = subset_key,
      impact_col = icol,
      features_to_use = feats_internal,
      internal_feature_col = "internal_feature",
      block_name = "objective_min_intervention_impact",
      tag = as.character(oargs$tag %||% "")[1]
    )

    res <- rcpp_add_objective_min_intervention_impact(
      op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      dist_features_data = x$data$dist_features,
      actions_to_use = acts_internal,
      subset_key = subset_key,
      impact_col = icol,
      features_to_use = feats_internal,
      internal_feature_col = "internal_feature",
      weight = 1.0
    )

    x$data$model_registry$vars$u_intervention_impact <- prep
    objective_id <- "min_intervention_impact"
  } else {
    .pa_abort("Unknown model_type: ", mtype)
  }

  x$data$model_args$modelsense   <- modelsense
  x$data$model_args$objective_id <- objective_id

  x$data$model_registry$objective <- list(type = mtype, id = objective_id)
  x$data$model_registry$objective$cpp <- res

  x
}

.pa_build_model_apply_constraints <- function(x) {

  if (!is.null(x$data$targets) &&
      inherits(x$data$targets, "data.frame") &&
      nrow(x$data$targets) > 0) {

    if (!exists(".pa_apply_targets_if_present", mode = "function")) {
      stop(".pa_apply_targets_if_present() is missing.", call. = FALSE)
    }

    x <- .pa_apply_targets_if_present(x)
  }

  x <- .pa_apply_action_max_per_pu_default(x)

  if (exists(".pa_apply_area_constraints_if_present", mode = "function")) {
    x <- .pa_apply_area_constraints_if_present(x)
  }

  if (exists(".pa_apply_budget_constraints_if_present", mode = "function")) {
    x <- .pa_apply_budget_constraints_if_present(x)
  }

  x
}


.pa_build_model_finalize <- function(x) {

  if (!exists(".pa_refresh_model_snapshot", mode = "function")) {
    stop(".pa_refresh_model_snapshot() is missing.", call. = FALSE)
  }

  x <- .pa_refresh_model_snapshot(x)
  x
}


.pa_apply_action_max_per_pu_default <- function(x) {

  stopifnot(inherits(x, "Problem"))

  da <- x$data$dist_actions_model
  if (is.null(da) || !inherits(da, "data.frame") || nrow(da) == 0) {
    return(x)
  }

  if (!exists("rcpp_add_action_max_per_pu", mode = "function")) {
    stop("Missing rcpp_add_action_max_per_pu().", call. = FALSE)
  }

  res <- rcpp_add_action_max_per_pu(
    x$data$model_ptr,
    dist_actions_data = da,
    max_per_pu = 1L,
    internal_pu_ids = integer(),
    internal_action_ids = integer()
  )

  x$data$model_registry$cons$action_max_per_pu <- res
  x
}

#' @include internal.R
#'
#' @title Add area constraint
#'
#' @description
#' Add an area constraint to a planning problem.
#'
#' This function stores one area-constraint specification in the
#' \code{Problem} object so that it can later be incorporated when the
#' optimization model is assembled. Multiple area constraints can be added by
#' calling this function repeatedly, provided that no duplicated combination of
#' action subset and constraint sense is introduced.
#'
#' @details
#' Use this function when area requirements must be imposed either on the total
#' selected landscape or on the subset of selected decisions associated with
#' specific actions.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and let
#' \eqn{a_i \ge 0} be the area associated with planning unit
#' \eqn{i \in \mathcal{I}}
#'
#' When \code{actions = NULL}, the constraint refers to the total selected area
#' in the problem. In that case, let \eqn{w_i \in \{0,1\}} denote the binary
#' variable indicating whether planning unit \eqn{i} is selected by at least one
#' decision in the model.
#'
#' Depending on \code{sense}, this function stores one of the following
#' constraints:
#'
#' If \code{sense = "min"}:
#' \deqn{
#' \sum_{i \in \mathcal{I}} a_i w_i \ge A
#' }
#'
#' If \code{sense = "max"}:
#' \deqn{
#' \sum_{i \in \mathcal{I}} a_i w_i \le A
#' }
#'
#' If \code{sense = "equal"} and \code{tolerance = 0}:
#' \deqn{
#' \sum_{i \in \mathcal{I}} a_i w_i = A
#' }
#'
#' If \code{sense = "equal"} and \code{tolerance > 0}, the equality is stored as
#' a two-sided band:
#' \deqn{
#' A - \tau \le \sum_{i \in \mathcal{I}} a_i w_i \le A + \tau
#' }
#' where \eqn{\tau} is the value supplied through \code{tolerance}.
#'
#' When \code{actions} is not \code{NULL}, the constraint is applied only to the
#' selected decisions associated with the specified subset of actions. Let
#' \eqn{\mathcal{A}^\star \subseteq \mathcal{A}} denote that subset and let
#' \eqn{x_{ia} \in \{0,1\}} denote the binary variable indicating whether action
#' \eqn{a \in \mathcal{A}^\star} is selected in planning unit
#' \eqn{i \in \mathcal{I}}. In that case, the constrained quantity is
#' \deqn{
#' \sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}^\star} a_i x_{ia}.
#' }
#'
#' Under formulations where at most one action can be selected per planning
#' unit, this coincides with the area allocated to that subset of actions.
#'
#' Areas are obtained from the planning-unit table. If \code{area_col} is
#' provided, that column is used. Otherwise, the model builder later determines
#' the default area source according to the internal rules of the package. The
#' value of \code{area_unit} indicates the unit in which \code{area} and
#' \code{tolerance} are expressed and therefore how the stored threshold should
#' be interpreted.
#'
#' This function only stores the constraint specification; it does not validate
#' the feasibility of the threshold against the available planning units at this
#' stage.
#'
#' Multiple area constraints can be stored in a \code{Problem} object. However,
#' at most one can be stored for the same combination of action subset and
#' constraint sense. Attempting to add a duplicated
#' \code{actions}--\code{sense} combination results in an error.
#'
#' @param x A \code{Problem} object.
#'
#' @param area Numeric scalar greater than or equal to zero. Target value for
#'   the constrained area.
#'
#' @param sense Character string indicating the type of area constraint. Must be
#'   one of \code{"min"}, \code{"max"}, or \code{"equal"}.
#'
#' @param tolerance Numeric scalar greater than or equal to zero. Only used when
#'   \code{sense = "equal"}. In that case, equality is interpreted as a band
#'   around \code{area} with half-width \code{tolerance}. Ignored otherwise.
#'
#' @param area_col Optional character string giving the name of the area column
#'   in \code{x$data$pu}. If \code{NULL}, the area source is resolved later by
#'   the model builder.
#'
#' @param area_unit Character string indicating the unit of \code{area} and
#'   \code{tolerance}. Must be one of \code{"m2"}, \code{"ha"}, or
#'   \code{"km2"}.
#'
#' @param actions Optional subset of actions to which the constraint applies.
#'   If \code{NULL}, the constraint applies to the total selected area in the
#'   problem through the planning-unit selection variables. Otherwise, it applies
#'   to the selected decision variables associated with the specified subset of
#'   actions. This argument is resolved using the package's standard action
#'   subset parser.
#'
#' @param name Optional character string used as the label of the stored linear
#'   constraint when it is later added to the optimization model. If
#'   \code{NULL}, a default name is generated.
#'
#' @return An updated \code{Problem} object with the new area-constraint
#'   specification appended to \code{x$data$constraints$area}.
#'
#' @seealso
#' \code{\link{create_problem}}
#'
#' @examples
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(2, 3, 1, 4),
#'   area_ha = c(10, 15, 8, 20)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' actions <- data.frame(
#'   id = c("conservation", "restoration")
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   p,
#'   actions = actions,
#'   cost = c(conservation = 1, restoration = 2)
#' )
#'
#' p <- add_constraint_area(
#'   x = p,
#'   area = 25,
#'   sense = "min",
#'   area_col = "area_ha",
#'   area_unit = "ha"
#' )
#'
#' p <- add_constraint_area(
#'   x = p,
#'   area = 15,
#'   sense = "max",
#'   area_col = "area_ha",
#'   area_unit = "ha",
#'   actions = "restoration"
#' )
#'
#' p <- add_constraint_area(
#'   x = p,
#'   area = 5,
#'   sense = "min",
#'   area_col = "area_ha",
#'   area_unit = "ha",
#'   actions = "restoration"
#' )
#'
#' p$data$constraints$area
#'
#' @export
add_constraint_area <- function(x,
                                area,
                                sense,
                                tolerance = 0,
                                area_col = NULL,
                                area_unit = c("m2", "ha", "km2"),
                                actions = NULL,
                                name = NULL) {

  stopifnot(inherits(x, "Problem"))

  if (missing(sense) || is.null(sense)) {
    stop(
      "`sense` must be explicitly provided. Use one of: 'min', 'max', or 'equal'.",
      call. = FALSE
    )
  }

  sense <- match.arg(sense, c("min", "max", "equal"))
  area_unit <- match.arg(area_unit)

  assertthat::assert_that(
    assertthat::is.number(area),
    is.finite(area),
    area >= 0
  )

  assertthat::assert_that(
    assertthat::is.number(tolerance),
    is.finite(tolerance),
    tolerance >= 0
  )

  if (!is.null(area_col)) {
    assertthat::assert_that(
      is.character(area_col),
      length(area_col) == 1,
      !is.na(area_col),
      nzchar(area_col)
    )
  }

  if (!is.null(name)) {
    if (!is.character(name) || length(name) != 1 || is.na(name) || !nzchar(name)) {
      stop("`name` must be NULL or a non-empty character string.", call. = FALSE)
    }
  }

  if (sense != "equal" && tolerance > 0) {
    warning(
      "`tolerance` is only used when sense = 'equal'. It will be ignored.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  x <- .pa_clone_data(x)

  actions_txt <- NA_character_

  if (!is.null(actions)) {
    act_subset <- .pa_resolve_action_subset(x, subset = actions)

    if (!is.data.frame(act_subset) || nrow(act_subset) == 0) {
      stop("`actions` did not resolve to any valid actions.", call. = FALSE)
    }

    actions_txt <- .pa_subset_to_string(act_subset$id)
  }

  area_name <- if (is.null(name)) {
    if (is.na(actions_txt)) {
      paste0("area_", sense)
    } else {
      paste0("area_", sense, "_", gsub("\\|", "_", actions_txt))
    }
  } else {
    as.character(name)[1]
  }

  area_df <- data.frame(
    type = "area",
    sense = sense,
    value = as.numeric(area),
    tolerance = if (sense == "equal") as.numeric(tolerance) else 0,
    unit = area_unit,
    area_col = if (is.null(area_col)) NA_character_ else as.character(area_col)[1],
    actions = actions_txt,
    name = area_name,
    stringsAsFactors = FALSE
  )

  x <- .pa_store_area_constraints(x, area_df)

  if (!is.null(x$data$model_ptr)) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

test_that("add_constraint_area stores one area constraint row", {
  toy <- toy_multiaction_semantics()
  toy$pu$area_ha <- c(1, 2, 3, 4, 5)

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  p <- multiscape::add_constraint_area(
    p,
    area = 10,
    sense = "max",
    area_col = "area_ha",
    area_unit = "ha"
  )

  expect_true(is.data.frame(p$data$constraints$area))
  expect_equal(nrow(p$data$constraints$area), 1L)

  expect_true(all(c(
    "type", "sense", "value", "tolerance",
    "unit", "area_col", "actions", "name"
  ) %in% names(p$data$constraints$area)))

  expect_equal(p$data$constraints$area$type[1], "area")
  expect_equal(p$data$constraints$area$sense[1], "max")
  expect_equal(p$data$constraints$area$value[1], 10)
  expect_equal(p$data$constraints$area$tolerance[1], 0)
  expect_equal(p$data$constraints$area$unit[1], "ha")
  expect_equal(p$data$constraints$area$area_col[1], "area_ha")
  expect_true(is.na(p$data$constraints$area$actions[1]))
})

test_that("add_constraint_area appends constraints for different subsets or senses", {
  toy <- toy_multiaction_semantics()
  toy$pu$area_ha <- c(1, 2, 3, 4, 5)

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  p <- p |>
    multiscape::add_constraint_area(
      area = 10,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha"
    ) |>
    multiscape::add_constraint_area(
      area = 4,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha",
      actions = "restoration"
    ) |>
    multiscape::add_constraint_area(
      area = 1,
      sense = "min",
      area_col = "area_ha",
      area_unit = "ha",
      actions = "restoration"
    )

  expect_true(is.data.frame(p$data$constraints$area))
  expect_equal(nrow(p$data$constraints$area), 3L)
})

test_that("add_constraint_area rejects duplicated actions-sense combinations", {
  toy <- toy_multiaction_semantics()
  toy$pu$area_ha <- c(1, 2, 3, 4, 5)

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    ) |>
    multiscape::add_constraint_area(
      area = 4,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha",
      actions = "restoration"
    )

  expect_error(
    p |>
      multiscape::add_constraint_area(
        area = 5,
        sense = "max",
        area_col = "area_ha",
        area_unit = "ha",
        actions = "restoration"
      ),
    "same subset of actions|same combination"
  )

  expect_no_error(
    p |>
      multiscape::add_constraint_area(
        area = 1,
        sense = "min",
        area_col = "area_ha",
        area_unit = "ha",
        actions = "restoration"
      )
  )
})

test_that("add_constraint_area validates inputs", {
  toy <- toy_multiaction_semantics()
  toy$pu$area_ha <- c(1, 2, 3, 4, 5)

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  expect_error(
    multiscape::add_constraint_area(
      p,
      area = -1,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha"
    )
  )

  expect_error(
    multiscape::add_constraint_area(
      p,
      area = 1,
      sense = "equal",
      tolerance = -1,
      area_col = "area_ha",
      area_unit = "ha"
    )
  )

  expect_error(
    multiscape::add_constraint_area(
      p,
      area = 1,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha",
      name = ""
    ),
    "`name`"
  )

  expect_error(
    multiscape::add_constraint_area(
      p,
      area = 1,
      sense = "max",
      area_col = "area_ha",
      area_unit = "ha",
      actions = "does_not_exist"
    ),
    "subset did not match any action ids"
  )
})

test_that("compile_model works with stored area constraints", {
  skip_if_no_cbc()

  toy <- toy_multiaction_semantics()
  toy$pu$area_ha <- c(1, 1, 1, 1, 1)

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    ) |>
    multiscape::add_effects(
      effects = toy$effects,
      effect_type = "after"
    ) |>
    multiscape::add_constraint_targets_relative(0.2) |>
    multiscape::add_constraint_area(
      area = 6,
      sense = "max",
      area_col = "cost",
      area_unit = "m2"
    ) |>
    multiscape::add_constraint_area(
      area = 3,
      sense = "max",
      area_col = "cost",
      area_unit = "m2",
      actions = "restoration"
    ) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  expect_no_error(multiscape::compile_model(p))
})

#' Compile the optimization model stored in a Problem
#'
#' @description
#' Materializes the optimization model represented by a \code{Problem} object
#' without solving it. This is an advanced function mainly intended for
#' debugging, inspection, and explicit model preparation.
#'
#' In standard workflows, users normally do not need to call this function,
#' because \code{\link{solve}} compiles the model automatically when needed.
#'
#' @details
#' Use this function when you want to prepare the optimization model explicitly
#' before solving, inspect compiled model structures, or verify that the problem
#' compiles successfully.
#'
#' Conceptually, a \code{Problem} object stores a declarative optimization
#' specification: planning data, actions, effects, targets, constraints,
#' objectives, spatial relations, and optional method or solver settings.
#' \code{compile_model()} transforms that stored specification into an internal
#' compiled model representation that can later be reused by the solving layer.
#'
#' The exact compiled representation is implementation-specific, but it may
#' include indexed variables, prepared constraint blocks, objective structures,
#' and internal model snapshots or pointers.
#'
#' Compilation does not solve the optimization problem. Therefore, a problem may
#' compile successfully and still later be infeasible, numerically difficult, or
#' otherwise fail during solver execution.
#'
#' @param x A \code{Problem} object.
#' @param force Logical. If \code{TRUE}, rebuild the model even if a current
#'   compiled model already exists.
#' @param ... Reserved for future extensions.
#'
#' @return
#' A \code{Problem} object with compiled model structures stored internally.
#'
#' @examples
#' \dontrun{
#' x <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' x <- x |>
#'   add_constraint_targets_relative(0.3) |>
#'   add_objective_min_cost(alias = "cost")
#'
#' x <- compile_model(x)
#'
#' # Force recompilation
#' x <- compile_model(x, force = TRUE)
#' }
#'
#' @seealso
#' \code{\link{solve}}
#'
#' @export
compile_model <- function(x, force = FALSE, ...) {
  UseMethod("compile_model")
}

#' @rdname compile_model
#' @export
compile_model.Problem <- function(x, force = FALSE, ...) {
  assertthat::assert_that(inherits(x, "Problem"))

  force <- isTRUE(force)

  objs <- x$data$objectives %||% list()
  n_obj <- if (is.list(objs)) length(objs) else 0L

  active_model_type <- x$data$model_args$model_type %||% NULL
  has_active_objective <- !is.null(active_model_type) &&
    nzchar(as.character(active_model_type)[1])

  method <- x$data$method %||% NULL
  has_method <- is.list(method) && length(method) > 0L

  if (!force && .pa_model_is_current(x)) {
    return(x)
  }

  if (has_method) {
    method_name <- as.character(method$type %||% method$name %||% NA_character_)[1]

    if (is.na(method_name) || !nzchar(method_name)) {
      stop(
        "Invalid multi-objective method configuration: missing method name.",
        call. = FALSE
      )
    }

    .pamo_validate_objectives(x)

    x <- .pamo_compile_problem(
      x = x,
      method_name = method_name,
      ...
    )

    if (is.null(x$data$model_ptr)) {
      stop(
        "Internal error: multi-objective compilation did not create model_ptr.",
        call. = FALSE
      )
    }

    if (is.null(x$data$model_list) || isTRUE(x$data$meta$model_dirty)) {
      x <- .pa_refresh_model_snapshot(x)
    }

    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- FALSE
    x$data$has_model <- TRUE

    return(x)
  }

  if (n_obj > 1L) {
    stop(
      "Multiple objectives are registered but no multi-objective method was selected.\n",
      "Use set_method_weighted(), set_method_epsilon_constraint(), etc.",
      call. = FALSE
    )
  }

  if (n_obj < 1L && !isTRUE(has_active_objective)) {
    stop(
      "No objective configured. Add an objective before compiling the model.",
      call. = FALSE
    )
  }

  x <- .pa_compile_problem(x, ...)

  if (is.null(x$data$model_ptr)) {
    stop(
      "Internal error: model compilation did not create model_ptr.",
      call. = FALSE
    )
  }

  if (is.null(x$data$model_list) || isTRUE(x$data$meta$model_dirty)) {
    x <- .pa_refresh_model_snapshot(x)
  }

  x$data$meta <- x$data$meta %||% list()
  x$data$meta$model_dirty <- FALSE
  x$data$has_model <- TRUE

  x
}

test_that("compile_model can be called twice on an already compiled single-objective problem", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost")

  p1 <- multiscape::compile_model(p)
  p2 <- multiscape::compile_model(p1)

  expect_s3_class(p1, "Problem")
  expect_s3_class(p2, "Problem")

  expect_false(is.null(p1$data$model_ptr))
  expect_false(is.null(p2$data$model_ptr))

  expect_true(is.list(p1$data$model_list))
  expect_true(is.list(p2$data$model_list))

  expect_false(isTRUE(p1$data$meta$model_dirty))
  expect_false(isTRUE(p2$data$meta$model_dirty))
})

test_that("compile_model(force = TRUE) recompiles a single-objective problem", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost")

  p <- multiscape::compile_model(p)
  p <- multiscape::compile_model(p, force = TRUE)

  expect_s3_class(p, "Problem")
  expect_false(is.null(p$data$model_ptr))
  expect_true(is.list(p$data$model_list))
  expect_false(isTRUE(p$data$meta$model_dirty))
})

test_that("add_constraint_budget stores one global budget constraint row", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  p <- multiscape::add_constraint_budget(
    p,
    budget = 10,
    sense = "max",
    include_pu_cost = TRUE,
    include_action_cost = TRUE
  )

  expect_true(is.data.frame(p$data$constraints$budget))
  expect_equal(nrow(p$data$constraints$budget), 1L)

  expect_true(all(c(
    "type", "sense", "value", "tolerance", "actions",
    "include_pu_cost", "include_action_cost", "name"
  ) %in% names(p$data$constraints$budget)))

  expect_equal(p$data$constraints$budget$type[1], "budget")
  expect_equal(p$data$constraints$budget$sense[1], "max")
  expect_equal(p$data$constraints$budget$value[1], 10)
  expect_equal(p$data$constraints$budget$tolerance[1], 0)
  expect_true(is.na(p$data$constraints$budget$actions[1]))
  expect_true(p$data$constraints$budget$include_pu_cost[1])
  expect_true(p$data$constraints$budget$include_action_cost[1])
})

test_that("add_constraint_budget appends constraints for different subsets or senses", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  p <- p |>
    multiscape::add_constraint_budget(
      budget = 10,
      sense = "max",
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    ) |>
    multiscape::add_constraint_budget(
      budget = 4,
      sense = "max",
      actions = "restoration",
      include_pu_cost = FALSE,
      include_action_cost = TRUE
    ) |>
    multiscape::add_constraint_budget(
      budget = 1,
      sense = "min",
      actions = "restoration",
      include_pu_cost = FALSE,
      include_action_cost = TRUE
    )

  expect_true(is.data.frame(p$data$constraints$budget))
  expect_equal(nrow(p$data$constraints$budget), 3L)
})

test_that("add_constraint_budget rejects duplicated actions-sense combinations", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    ) |>
    multiscape::add_constraint_budget(
      budget = 4,
      sense = "max",
      actions = "restoration",
      include_pu_cost = FALSE,
      include_action_cost = TRUE
    )

  expect_error(
    p |>
      multiscape::add_constraint_budget(
        budget = 5,
        sense = "max",
        actions = "restoration",
        include_pu_cost = FALSE,
        include_action_cost = TRUE
      ),
    "same subset of actions|same combination"
  )

  expect_no_error(
    p |>
      multiscape::add_constraint_budget(
        budget = 1,
        sense = "min",
        actions = "restoration",
        include_pu_cost = FALSE,
        include_action_cost = TRUE
      )
  )
})

test_that("add_constraint_budget validates cost-component settings", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = 5,
      sense = "max",
      include_pu_cost = FALSE,
      include_action_cost = FALSE
    ),
    "At least one"
  )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = 5,
      sense = "max",
      actions = "restoration",
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    ),
    "include_pu_cost = TRUE.*actions = NULL|not action-specific"
  )
})

test_that("add_constraint_budget validates other inputs", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = -1,
      sense = "max",
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    )
  )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = 1,
      sense = "equal",
      tolerance = -1,
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    )
  )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = 1,
      sense = "max",
      include_pu_cost = TRUE,
      include_action_cost = TRUE,
      name = ""
    ),
    "`name`"
  )

  expect_error(
    multiscape::add_constraint_budget(
      p,
      budget = 1,
      sense = "max",
      actions = "does_not_exist",
      include_pu_cost = FALSE,
      include_action_cost = TRUE
    ),
    "subset did not match any action ids"
  )
})

test_that("compile_model works with stored budget constraints", {
  skip_if_no_cbc()

  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(
      actions = toy$actions,
      cost = c(conservation = 1, restoration = 2)
    ) |>
    multiscape::add_effects(
      effects = toy$effects,
      effect_type = "after"
    ) |>
    multiscape::add_constraint_targets_relative(0.2) |>
    multiscape::add_constraint_budget(
      budget = 12,
      sense = "max",
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    ) |>
    multiscape::add_constraint_budget(
      budget = 4,
      sense = "max",
      actions = "restoration",
      include_pu_cost = FALSE,
      include_action_cost = TRUE
    ) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  expect_no_error(multiscape::compile_model(p))
})

test_that("compile_model errors when no objective is configured", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5)

  expect_error(
    multiscape::compile_model(p),
    "No objective configured"
  )
})

.pkgenv <- new.env(parent = emptyenv())

.onUnload <- function(libpath) {
  library.dynam.unload("multiscape", libpath)
}

.onLoad <- function(libname, pkgname) {
  requireNamespace("Rcpp", quietly = TRUE)
}

# .onAttach <- function(libname, pkgname) {
#   # define message generator function
#   msg <- function() {
#     packageStartupMessage(paste(rep("-", 30), collapse = ""))
#     packageStartupMessage(
#       "You have loaded both prioriactions and prioritizr - ",
#       "this is likely to cause serious issues.\n",
#       "You should only have one of these packages loaded at a time,\n",
#       "please unload one prioriactions or prioritizr using one of the commands below:\n",
#       "  detach(\"package:prioriactions\", unload = TRUE) # unload prioriactions package\n",
#       "  detach(\"package:prioritizr\", unload = TRUE) # unload prioritizr",
#       "package\n",
#       "and then reload the desired package."
#     )
#     packageStartupMessage(paste(rep("-", 30), collapse = ""))
#   }
#   # print message if prioritizr already loaded
#   if ("prioritizr" %in% .packages()) {
#     msg()
#   }
# }

#' @include internal.R
#'
#' @title Add management actions to a planning problem
#'
#' @description
#' Define the action catalogue, the set of feasible planning unit--action pairs,
#' and their implementation costs.
#'
#' This function adds two core components to a \code{Problem} object. First, it
#' stores the action catalogue. Second, it creates the feasible planning
#' unit--action table, including implementation costs, status codes, and
#' internal indices used by the optimization backend.
#'
#' Conceptually, if \eqn{\mathcal{I}} is the set of planning units and
#' \eqn{\mathcal{A}} is the set of actions, this function determines which
#' pairs \eqn{(i,a) \in \mathcal{I} \times \mathcal{A}} are feasible decisions
#' and assigns a non-negative implementation cost to each feasible pair.
#'
#' @details
#' \strong{When to use \code{add_actions()}.}
#'
#' Use this function when you want to move from a planning problem defined only
#' by planning units and features to a problem in which decisions are explicitly
#' represented as actions applied in planning units.
#'
#' \strong{Action catalogue.}
#'
#' The \code{actions} argument must be a \code{data.frame} with a unique
#' \code{id} column identifying each action. If a column named \code{action} is
#' supplied instead, it is renamed internally to \code{id}. Additional columns
#' are preserved. If no \code{name} column is provided, action labels are taken
#' from \code{id}. If an \code{action_set} column is present, it is also
#' preserved and can later be used to refer to groups of actions.
#'
#' Actions are stored sorted by \code{id} to ensure reproducible internal
#' indexing.
#'
#' \strong{Feasible planning unit--action pairs.}
#'
#' Feasibility is controlled through \code{include_pairs} and
#' \code{exclude_pairs}.
#'
#' If \code{include_pairs = NULL}, all possible \code{(pu, action)} pairs are
#' initially considered feasible, that is, all pairs
#' \eqn{(i,a) \in \mathcal{I} \times \mathcal{A}}.
#'
#' If \code{include_pairs} is supplied, only those pairs are retained. If
#' \code{exclude_pairs} is also supplied, matching pairs are removed
#' afterwards.
#'
#' More precisely, let \eqn{\mathcal{D}^{\mathrm{inc}}} denote the set of
#' included planning unit--action pairs and let
#' \eqn{\mathcal{D}^{\mathrm{exc}}} denote the set of excluded pairs.
#'
#' If \code{include_pairs = NULL}, the feasible decision set is:
#' \deqn{
#' \{(i,a) : i \in \mathcal{I},\ a \in \mathcal{A}\} \setminus \mathcal{D}^{\mathrm{exc}}.
#' }
#'
#' If \code{include_pairs} is supplied, the feasible decision set is:
#' \deqn{
#' \mathcal{D}^{\mathrm{inc}} \setminus \mathcal{D}^{\mathrm{exc}}.
#' }
#'
#' Both \code{include_pairs} and \code{exclude_pairs} can be specified as:
#' \itemize{
#'   \item \code{NULL},
#'   \item a \code{data.frame} with columns \code{pu} and \code{action},
#'   \item or a named list whose names are action ids.
#' }
#'
#' When supplied as a \code{data.frame}, the object must contain columns
#' \code{pu} and \code{action}. An optional logical-like column
#' \code{feasible} may also be provided; only rows with \code{feasible = TRUE}
#' are retained. Missing values in \code{feasible} are treated as
#' \code{FALSE}.
#'
#' When supplied as a named list, names must match action ids. Each element may
#' contain either:
#' \itemize{
#'   \item a vector of planning-unit ids, or
#'   \item an \code{sf} object defining the spatial zone where the action is
#'   feasible.
#' }
#'
#' In the spatial case, feasible planning units are identified using
#' \code{sf::st_intersects()} against the stored planning-unit geometry.
#'
#' \strong{Feasibility versus decision fixing.}
#'
#' This function only determines whether a pair \eqn{(i,a)} exists in the model.
#' It does not force a feasible action to be selected or forbidden beyond
#' structural infeasibility. Fixed decisions should instead be imposed later
#' with \code{\link{add_constraint_locked_actions}}.
#'
#' \strong{Costs.}
#'
#' Costs can be supplied in several ways:
#' \itemize{
#'   \item If \code{cost = NULL}, all feasible pairs receive a default cost of
#'   \code{1}.
#'   \item If \code{cost} is a scalar, that value is assigned to all feasible
#'   pairs.
#'   \item If \code{cost} is a named numeric vector, names must match action ids
#'   and costs are assigned by action.
#'   \item If \code{cost} is a \code{data.frame}, it must define either:
#'   \itemize{
#'     \item action-level costs through columns \code{action} and \code{cost}, or
#'     \item pair-specific costs through columns \code{pu}, \code{action}, and
#'     \code{cost}.
#'   }
#' }
#'
#' In all cases, costs must be finite and non-negative.
#'
#' In practice, a scalar cost is useful when all actions cost the same
#' everywhere, a named vector is useful when cost depends only on action type,
#' and a \code{(pu, action, cost)} table is useful when cost varies by both
#' planning unit and action.
#'
#' \strong{Status values.}
#'
#' Internally, all feasible pairs are initialized with \code{status = 0},
#' meaning that the decision is free. If planning units have already been marked
#' as locked out, then all feasible actions in those planning units are assigned
#' \code{status = 3}. This preserves consistency with planning-unit exclusions
#' already stored in the problem.
#'
#' \strong{Replacement behaviour.}
#'
#' Calling \code{add_actions()} replaces any previous action catalogue and
#' feasible action table stored in the problem object.
#'
#' After defining actions, typical next steps include adding effects, optional
#' decision-fixing constraints, objectives, and solver settings before calling
#' \code{solve()}.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}.
#'
#' @param actions A \code{data.frame} defining the action catalogue. It must
#'   contain a unique \code{id} column. A column named \code{action} is also
#'   accepted and automatically renamed to \code{id}.
#'
#' @param include_pairs Optional specification of feasible \code{(pu, action)}
#'   pairs. It can be \code{NULL}, a \code{data.frame} with columns
#'   \code{pu} and \code{action} (optionally also \code{feasible}), or a named
#'   list whose names are action ids and whose elements are vectors of planning
#'   unit ids or \code{sf} objects.
#'
#' @param exclude_pairs Optional specification of infeasible \code{(pu, action)}
#'   pairs. It uses the same formats as \code{include_pairs} and removes
#'   matching pairs from the feasible set.
#'
#' @param cost Optional cost specification for feasible pairs. It may be
#'   \code{NULL}, a scalar numeric value, a named numeric vector indexed by
#'   action id, or a \code{data.frame} with columns \code{action, cost} or
#'   \code{pu, action, cost}.
#'
#' @return An updated \code{Problem} object with:
#' \describe{
#'   \item{\code{actions}}{The action catalogue, including a unique integer
#'   \code{internal_id} for each action.}
#'   \item{\code{dist_actions}}{The feasible planning unit--action table with
#'   columns \code{pu}, \code{action}, \code{cost}, \code{status},
#'   \code{internal_pu}, and \code{internal_action}.}
#'   \item{\code{pu index}}{A mapping from user-supplied planning-unit ids to
#'   internal integer ids.}
#'   \item{\code{action index}}{A mapping from action ids to internal integer
#'   ids.}
#' }
#'
#' @seealso
#' \code{\link{create_problem}},
#' \code{\link{add_constraint_locked_actions}}
#'
#' @examples
#' # ------------------------------------------------------
#' # Minimal planning problem
#' # ------------------------------------------------------
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(2, 3, 1, 4)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' actions <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("Conservation", "Restoration")
#' )
#'
#' # Example 1: all actions feasible in all planning units
#' p1 <- add_actions(
#'   x = p,
#'   actions = actions,
#'   cost = c(conservation = 5, restoration = 12)
#' )
#'
#' print(p1)
#' utils::head(p1$data$dist_actions)
#'
#' # Example 2: specify feasible pairs explicitly
#' include_df <- data.frame(
#'   pu = c(1, 2, 3, 4),
#'   action = c("conservation", "conservation", "restoration", "restoration")
#' )
#'
#' p2 <- add_actions(
#'   x = p,
#'   actions = actions,
#'   include_pairs = include_df,
#'   cost = 10
#' )
#'
#' p2$data$dist_actions
#'
#' # Example 3: remove selected pairs after full expansion
#' exclude_df <- data.frame(
#'   pu = c(2, 4),
#'   action = c("restoration", "conservation")
#' )
#'
#' p3 <- add_actions(
#'   x = p,
#'   actions = actions,
#'   exclude_pairs = exclude_df,
#'   cost = c(conservation = 3, restoration = 8)
#' )
#'
#' p3$data$dist_actions
#'
#' @export
add_actions <- function(
    x,
    actions,
    include_pairs = NULL,
    exclude_pairs = NULL,
    cost = NULL
) {

  .as_int_id <- function(v, what) {
    if (is.factor(v)) v <- as.character(v)
    if (is.character(v)) {
      if (any(grepl("[^0-9\\-]", v))) {
        stop(what, " must be numeric/integer ids (got non-numeric strings).", call. = FALSE)
      }
      v <- as.integer(v)
    } else {
      v <- as.integer(v)
    }
    if (anyNA(v)) stop(what, " contains NA after coercion to integer.", call. = FALSE)
    v
  }

  .normalize_feasible_col <- function(df, what) {
    if (!("feasible" %in% names(df))) {
      df$feasible <- TRUE
      return(df)
    }

    f <- df$feasible

    if (is.logical(f)) {
      # keep
    } else if (is.numeric(f) || is.integer(f)) {
      f <- f != 0
    } else if (is.factor(f)) {
      f <- as.character(f)
    }

    if (is.character(f)) {
      w <- tolower(trimws(f))
      f <- w %in% c("true", "t", "1", "yes", "y")
    } else {
      f <- as.logical(f)
    }

    f[is.na(f)] <- FALSE
    df$feasible <- as.logical(f)
    df
  }

  .spec_to_pairs <- function(spec, what, action_ids, pu_ids, pu_sf, as_int_id_fun) {
    if (is.null(spec)) return(NULL)

    if (inherits(spec, "data.frame")) {
      assertthat::assert_that(nrow(spec) > 0, msg = paste0(what, " is an empty data.frame."))

      if ("id" %in% names(spec) && !("action" %in% names(spec))) {
        names(spec)[names(spec) == "id"] <- "action"
      }
      assertthat::assert_that(
        assertthat::has_name(spec, "pu"),
        assertthat::has_name(spec, "action"),
        msg = paste0(what, " must have columns 'pu' and 'action'.")
      )

      spec$pu <- as_int_id_fun(spec$pu, paste0(what, "$pu"))
      spec$action <- as.character(spec$action)
      spec <- .normalize_feasible_col(spec, what)

      if (!all(spec$pu %in% pu_ids)) {
        bad <- unique(spec$pu[!spec$pu %in% pu_ids])
        stop(what, " contains PU ids not present in x: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      if (!all(spec$action %in% action_ids)) {
        bad <- unique(spec$action[!spec$action %in% action_ids])
        stop(what, " contains action ids not present in actions: ", paste(bad, collapse = ", "), call. = FALSE)
      }

      tmp <- spec[, c("pu", "action")]
      if (nrow(dplyr::distinct(tmp)) != nrow(tmp)) {
        stop(what, " has duplicate (pu, action) rows. Please de-duplicate.", call. = FALSE)
      }

      out <- spec[spec$feasible, c("pu", "action"), drop = FALSE]
      return(out)
    }

    if (is.list(spec)) {
      if (is.null(names(spec)) || any(names(spec) == "")) {
        stop("If '", what, "' is a list, it must be a named list with names = action ids.", call. = FALSE)
      }
      if (!all(names(spec) %in% action_ids)) {
        bad <- setdiff(names(spec), action_ids)
        stop(what, " list contains unknown actions: ", paste(bad, collapse = ", "), call. = FALSE)
      }

      idx_first <- which(!vapply(spec, is.null, logical(1)))[1]
      first <- spec[[idx_first]]
      is_sf_list <- inherits(first, "sf")

      if (is_sf_list) {
        if (!requireNamespace("sf", quietly = TRUE)) {
          stop(what, " provided as sf layers requires the 'sf' package.", call. = FALSE)
        }
        if (is.null(pu_sf) || !inherits(pu_sf, "sf")) {
          stop("To use '", what, "' as sf layers, the problem object must contain x$data$pu_sf (sf planning unit geometry).", call. = FALSE)
        }
        if (!("id" %in% names(pu_sf))) {
          stop("x$data$pu_sf is missing an 'id' column.", call. = FALSE)
        }

        pu_sf2 <- pu_sf
        pu_sf2$id <- as_int_id_fun(pu_sf2$id, "x$data$pu_sf$id")

        out <- vector("list", length(spec))
        names(out) <- names(spec)

        for (a in names(spec)) {
          zone <- spec[[a]]
          if (is.null(zone)) {
            out[[a]] <- NULL
            next
          }
          if (!inherits(zone, "sf")) {
            stop(what, "[[", a, "]] must be an sf object.", call. = FALSE)
          }

          hits <- sf::st_intersects(pu_sf2, zone, sparse = TRUE)
          feasible_ids <- pu_sf2$id[lengths(hits) > 0]

          if (length(feasible_ids) == 0) {
            out[[a]] <- NULL
          } else {
            out[[a]] <- data.frame(
              pu = feasible_ids,
              action = a,
              stringsAsFactors = FALSE
            )
          }
        }

        out_df <- dplyr::bind_rows(out)
        if (nrow(out_df) == 0) return(out_df)
        out_df$pu <- as_int_id_fun(out_df$pu, paste0(what, "$pu"))
        out_df <- dplyr::distinct(out_df)
        return(out_df)
      }

      out <- vector("list", length(spec))
      names(out) <- names(spec)

      for (a in names(spec)) {
        ids <- spec[[a]]
        if (is.null(ids)) {
          out[[a]] <- NULL
          next
        }
        ids <- unique(as_int_id_fun(ids, paste0(what, "[['", a, "']]")))
        if (!all(ids %in% pu_ids)) {
          bad <- ids[!ids %in% pu_ids]
          stop(what, "[[", a, "]] contains PU ids not present in x: ", paste(bad, collapse = ", "), call. = FALSE)
        }
        out[[a]] <- data.frame(
          pu = ids,
          action = a,
          stringsAsFactors = FALSE
        )
      }

      out_df <- dplyr::bind_rows(out)
      if (nrow(out_df) == 0) return(out_df)
      out_df$pu <- as_int_id_fun(out_df$pu, paste0(what, "$pu"))
      out_df <- dplyr::distinct(out_df)
      return(out_df)
    }

    stop("Unsupported type for '", what, "'. Use NULL, data.frame, or a named list.", call. = FALSE)
  }

  # ---- checks: x
  assertthat::assert_that(!is.null(x), msg = "x is NULL")
  assertthat::assert_that(!is.null(x$data), msg = "x does not look like a multiscape Problem object")
  assertthat::assert_that(
    !is.null(x$data$pu),
    !is.null(x$data$features),
    !is.null(x$data$dist_features),
    msg = "x must be created with create_problem()"
  )

  if (is.null(x$data$pu$internal_id)) {
    x$data$pu$internal_id <- seq_len(nrow(x$data$pu))
  }

  x <- .pa_clone_data(x)
  x$data$pu$id <- .as_int_id(x$data$pu$id, "x$data$pu$id")

  pu_ids <- x$data$pu$id
  pu_index <- stats::setNames(x$data$pu$internal_id, as.character(x$data$pu$id))

  # ---- actions catalog
  assertthat::assert_that(inherits(actions, "data.frame"), nrow(actions) > 0)

  if ("action" %in% names(actions) && !("id" %in% names(actions))) {
    warning("actions has column 'action'. Renaming it to 'id'.", call. = FALSE, immediate. = TRUE)
    names(actions)[names(actions) == "action"] <- "id"
  }

  assertthat::assert_that(assertthat::has_name(actions, "id"), assertthat::noNA(actions$id))
  actions$id <- as.character(actions$id)
  if (anyDuplicated(actions$id) != 0) {
    stop("actions$id must be unique.", call. = FALSE)
  }

  if (!("name" %in% names(actions))) {
    actions$name <- as.character(actions$id)
  } else {
    actions$name <- as.character(actions$name)
    if (anyNA(actions$name) || any(!nzchar(actions$name))) {
      stop("actions$name cannot contain NA or empty strings.", call. = FALSE)
    }
  }

  if ("action_set" %in% names(actions)) {
    actions$action_set <- as.character(actions$action_set)
    if (anyNA(actions$action_set) || any(!nzchar(actions$action_set))) {
      stop("actions$action_set cannot contain NA or empty strings.", call. = FALSE)
    }
  }

  actions <- actions[order(actions$id), , drop = FALSE]

  if (!("internal_id" %in% names(actions))) {
    actions$internal_id <- seq_len(nrow(actions))
  } else {
    actions$internal_id <- as.integer(actions$internal_id)
    if (base::anyNA(actions$internal_id)) {
      stop("actions$internal_id contains NA.", call. = FALSE)
    }
    if (anyDuplicated(actions$internal_id) != 0) {
      stop("actions$internal_id must be unique if provided.", call. = FALSE)
    }
  }

  action_ids <- actions$id
  action_index <- stats::setNames(actions$internal_id, actions$id)

  if (is.null(x$data$index) || !is.list(x$data$index)) x$data$index <- list()
  x$data$index$pu <- pu_index
  x$data$index$action <- action_index

  # ---- build feasible pairs
  pu_sf <- x$data$pu_sf
  include_df <- .spec_to_pairs(include_pairs, "include_pairs", action_ids, pu_ids, pu_sf, .as_int_id)
  exclude_df <- .spec_to_pairs(exclude_pairs, "exclude_pairs", action_ids, pu_ids, pu_sf, .as_int_id)

  if (is.null(include_df)) {
    dist_actions <- base::expand.grid(
      pu = pu_ids,
      action = action_ids,
      KEEP.OUT.ATTRS = FALSE,
      stringsAsFactors = FALSE
    )
  } else {
    dist_actions <- include_df
  }

  if (!is.null(exclude_df) && nrow(exclude_df) > 0) {
    key_da <- paste(dist_actions$pu, dist_actions$action)
    key_ex <- paste(exclude_df$pu, exclude_df$action)
    keep <- !(key_da %in% key_ex)
    dist_actions <- dist_actions[keep, , drop = FALSE]
  }

  if (nrow(dist_actions) == 0) {
    stop("No feasible (pu, action) pairs were created after applying include_pairs/exclude_pairs.", call. = FALSE)
  }

  dist_actions$pu <- .as_int_id(dist_actions$pu, "dist_actions$pu")
  dist_actions$action <- as.character(dist_actions$action)

  # ---- costs
  dist_actions$cost <- 1

  if (is.null(cost)) {

    # keep default

  } else if (is.numeric(cost) && length(cost) == 1) {

    dist_actions$cost <- as.numeric(cost)

  } else if (is.numeric(cost) && !is.null(names(cost))) {

    if (!all(names(cost) %in% action_ids)) {
      bad <- setdiff(names(cost), action_ids)
      stop("cost contains unknown actions: ", paste(bad, collapse = ", "), call. = FALSE)
    }
    dist_actions$cost <- as.numeric(cost[dist_actions$action])

  } else if (inherits(cost, "data.frame")) {

    if ("id" %in% names(cost) && !("action" %in% names(cost))) {
      names(cost)[names(cost) == "id"] <- "action"
    }

    if (all(c("action", "cost") %in% names(cost)) && !("pu" %in% names(cost))) {

      cost$action <- as.character(cost$action)
      if (!all(cost$action %in% action_ids)) {
        bad <- unique(cost$action[!cost$action %in% action_ids])
        stop("cost contains unknown actions: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      if (nrow(dplyr::distinct(cost[, c("action")])) != nrow(cost)) {
        stop("cost (action,cost) must have unique action rows.", call. = FALSE)
      }

      m <- match(dist_actions$action, cost$action)
      dist_actions$cost <- cost$cost[m]

    } else if (all(c("pu", "action", "cost") %in% names(cost))) {

      cost$pu <- .as_int_id(cost$pu, "cost$pu")
      cost$action <- as.character(cost$action)

      if (!all(cost$pu %in% pu_ids)) stop("cost contains unknown pu ids.", call. = FALSE)
      if (!all(cost$action %in% action_ids)) stop("cost contains unknown actions.", call. = FALSE)

      tmp <- cost[, c("pu", "action")]
      if (nrow(dplyr::distinct(tmp)) != nrow(tmp)) {
        stop("cost has duplicate (pu, action) rows.", call. = FALSE)
      }

      key_da <- paste(dist_actions$pu, dist_actions$action)
      key_c  <- paste(cost$pu, cost$action)
      m <- match(key_da, key_c)
      hit <- !is.na(m)
      dist_actions$cost[hit] <- cost$cost[m[hit]]

    } else {
      stop("Unsupported cost data.frame format. Use (action,cost) or (pu,action,cost).", call. = FALSE)
    }

  } else {
    stop("Unsupported type for 'cost'.", call. = FALSE)
  }

  assertthat::assert_that(is.numeric(dist_actions$cost))
  if (any(!is.finite(dist_actions$cost) | is.na(dist_actions$cost))) {
    stop("Some feasible (pu, action) pairs have missing/invalid costs after processing 'cost'.", call. = FALSE)
  }
  if (any(dist_actions$cost < 0, na.rm = TRUE)) {
    stop("Action costs must be non-negative.", call. = FALSE)
  }

  # ---- initialize status as free
  dist_actions$status <- 0L

  # ---- enforce PU locked_out
  # if ("locked_out" %in% names(x$data$pu)) {
  #   pu_locked_out <- x$data$pu$locked_out
  #   pu_locked_out[is.na(pu_locked_out)] <- FALSE
  #   pu_locked_out <- as.logical(pu_locked_out)
  #
  #   locked_out_pus <- x$data$pu$id[pu_locked_out]
  #
  #   if (length(locked_out_pus) > 0) {
  #     idx_pu_lo <- dist_actions$pu %in% locked_out_pus
  #     dist_actions$status[idx_pu_lo] <- 3L
  #   }
  # }

  # ---- add internal ids
  dist_actions$internal_pu <- unname(pu_index[as.character(dist_actions$pu)])
  dist_actions$internal_action <- unname(action_index[as.character(dist_actions$action)])

  dist_actions <- dist_actions[order(dist_actions$internal_pu, dist_actions$internal_action), , drop = FALSE]

  if (anyNA(dist_actions$internal_pu)) {
    stop("Internal error: could not map pu -> internal_pu.", call. = FALSE)
  }
  if (anyNA(dist_actions$internal_action)) {
    stop("Internal error: could not map action -> internal_action.", call. = FALSE)
  }

  x$data$actions <- actions
  x$data$dist_actions <- dist_actions

  x
}

test_that("solve selects at most one action per planning unit", {
  skip_if_no_cbc()

  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.2, features = 1) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_s3_class(s, "Solution")
  expect_true(is.data.frame(s$summary$actions))

  acts <- s$summary$actions
  acts_sel <- acts[acts$selected > 0.5, , drop = FALSE]

  if (nrow(acts_sel) > 0) {
    n_by_pu <- table(acts_sel$pu)
    expect_true(all(as.integer(n_by_pu) <= 1L))
  } else {
    succeed()
  }
})

test_that("add_actions creates one feasible row per pu-action pair", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  p <- multiscape::add_actions(
    p,
    actions = toy$actions,
    cost = 0
  )

  expect_true(is.data.frame(p$data$actions))
  expect_true(is.data.frame(p$data$dist_actions))

  expect_equal(nrow(p$data$actions), 1)
  expect_equal(nrow(p$data$dist_actions), nrow(toy$pu) * nrow(toy$actions))

  expect_true(all(c("pu", "action") %in% names(p$data$dist_actions)))
  expect_equal(anyDuplicated(p$data$dist_actions[, c("pu", "action")]), 0)

  expect_true("internal_action" %in% names(p$data$dist_actions))
  expect_true("internal_pu" %in% names(p$data$dist_actions))
})

test_that("add_actions stores a constant action cost correctly", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  p <- multiscape::add_actions(
    p,
    actions = toy$actions,
    cost = 7
  )

  expect_true("cost" %in% names(p$data$dist_actions))
  expect_equal(unique(p$data$dist_actions$cost), 7)
})

test_that("add_effects with delta separates benefit and loss correctly", {
  toy <- toy_equivalent_basic()

  effects_delta <- data.frame(
    pu = c(1, 2, 3, 4),
    action = c("conservation", "conservation", "conservation", "conservation"),
    feature = c(1, 1, 2, 2),
    delta = c(5, -2, 0, -7)
  )

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = effects_delta, effect_type = "delta")

  de <- p$data$dist_effects

  expect_true(all(c("benefit", "loss") %in% names(de)))

  row1 <- de[de$pu == 1 & de$feature == 1, ]
  row2 <- de[de$pu == 2 & de$feature == 1, ]
  row4 <- de[de$pu == 4 & de$feature == 2, ]

  expect_equal(row1$benefit, 5)
  expect_equal(row1$loss, 0)

  expect_equal(row2$benefit, 0)
  expect_equal(row2$loss, 2)

  expect_equal(row4$benefit, 0)
  expect_equal(row4$loss, 7)
})

test_that("add_effects with after computes delta from baseline amounts", {
  toy <- toy_equivalent_basic()

  effects_after <- data.frame(
    pu = c(1, 2),
    action = c("conservation", "conservation"),
    feature = c(1, 2),
    after = c(10, 1)
  )

  # baseline: pu1-feature1 = 8 ; pu2-feature2 = 2
  # delta esperado: +2 y -1

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = effects_after, effect_type = "after")

  de <- p$data$dist_effects

  r1 <- de[de$pu == 1 & de$feature == 1, ]
  r2 <- de[de$pu == 2 & de$feature == 2, ]

  expect_equal(r1$benefit, 2)
  expect_equal(r1$loss, 0)

  expect_equal(r2$benefit, 0)
  expect_equal(r2$loss, 1)
})

#' @include internal.R
#'
#' @export
if (!methods::isClass("Problem")) methods::setOldClass("Problem")
NULL

#' @name problem-class
#' @aliases Problem
#' @title Problem class
#'
#' @description
#' The \code{Problem} class is the central container used by \pkg{multiscape} to
#' represent a planning problem before, during, and after model construction.
#'
#' A \code{Problem} object stores the full problem specification in a modular
#' way. This includes the baseline planning data, optional spatial metadata,
#' action definitions, effects, profit, targets, constraints, objective
#' registrations, solver settings, and, when available, a built optimization
#' model or model snapshot.
#'
#' In other words, \code{Problem} is the persistent object that connects the full
#' \code{multiscape} workflow:
#' \preformatted{
#' create_problem()
#' -> add_*() / set_*()
#' -> solve()
#' }
#'
#' @details
#' \strong{Conceptual role}
#'
#' The \code{Problem} class is designed for a data-first and modular workflow.
#' User-facing functions do not usually modify a solver object directly. Instead,
#' they enrich the \code{Problem} object by storing new specifications in its
#' internal \code{data} field.
#'
#' Thus, a \code{Problem} object should be understood as a structured container
#' for the mathematical planning problem, not necessarily as a built
#' optimization model.
#'
#' Before \code{\link{solve}} is called, the object may contain only input data
#' and user specifications. During or after solving, it may additionally contain
#' a built model pointer, model metadata, and solver-related information.
#'
#' \strong{Core mathematical interpretation}
#'
#' At a high level, the \code{Problem} object stores the ingredients required to
#' define an optimization problem over:
#' \itemize{
#'   \item planning units \eqn{i \in \mathcal{P}},
#'   \item features \eqn{f \in \mathcal{F}},
#'   \item actions \eqn{a \in \mathcal{A}},
#'   \item optional spatial relations over planning units,
#'   \item and user-defined objectives and constraints.
#' }
#'
#' The baseline ecological state is typically stored through a planning
#' unit--feature table of amounts:
#' \deqn{
#' a_{if} \ge 0,
#' }
#' where \eqn{a_{if}} is the baseline amount of feature \eqn{f} in planning unit
#' \eqn{i}.
#'
#' Subsequent functions then add action feasibility, effects, profit, targets,
#' spatial relations, and optimization settings to this baseline representation.
#'
#' \strong{How objects are created}
#'
#' \code{Problem} objects are usually created by \code{\link{create_problem}}.
#'
#' After creation, downstream functions such as \code{\link{add_actions}},
#' \code{\link{add_effects}}, \code{\link{add_profit}},
#' \code{\link{add_constraint_targets_absolute}}, \code{\link{add_constraint_targets_relative}},
#' spatial relation constructors, objective setters, and solver setters extend
#' the internal \code{data} list.
#'
#' \strong{Internal storage}
#'
#' The class contains a single field:
#' \describe{
#'   \item{\code{data}}{A named \code{list} storing the full problem
#'   specification, metadata, and, when available, built-model information.}
#' }
#'
#' Common entries of \code{data} include:
#' \describe{
#'   \item{\code{pu}}{Planning-unit table.}
#'   \item{\code{features}}{Feature table.}
#'   \item{\code{actions}}{Action catalog.}
#'   \item{\code{dist_features}}{Planning unit--feature baseline amounts.}
#'   \item{\code{dist_actions}}{Feasible planning unit--action pairs.}
#'   \item{\code{dist_effects}}{Action effects by planning unit, action, and
#'   feature.}
#'   \item{\code{dist_profit}}{Profit by planning unit--action pair.}
#'   \item{\code{pu_sf}}{Planning-unit geometry when available.}
#'   \item{\code{pu_coords}}{Planning-unit coordinates when available.}
#'   \item{\code{spatial_relations}}{Registered spatial relations.}
#'   \item{\code{targets}}{Stored target specifications.}
#'   \item{\code{constraints}}{Stored user-defined constraints.}
#'   \item{\code{objectives}}{Registered atomic objectives for single- or
#'   multi-objective workflows.}
#'   \item{\code{method}}{Stored multi-objective method configuration, when
#'   applicable.}
#'   \item{\code{solve_args}}{Stored solver settings.}
#'   \item{\code{model_ptr}}{Pointer to a built optimization model, when
#'   available.}
#'   \item{\code{model_args}}{Metadata describing model construction.}
#'   \item{\code{model_list}}{Optional exported model snapshot or representation.}
#'   \item{\code{meta}}{Auxiliary metadata, including model-dirty flags and other
#'   bookkeeping fields.}
#' }
#'
#' Not every \code{Problem} object contains all of these entries. The content of
#' \code{data} depends on how far the workflow has progressed.
#'
#' \strong{Lifecycle}
#'
#' A \code{Problem} object typically moves through the following stages:
#' \enumerate{
#'   \item input stage: baseline planning units, features, and feature
#'   distributions are stored,
#'   \item specification stage: actions, effects, targets, objectives,
#'   constraints, and spatial relations are added,
#'   \item model stage: an optimization model is built from the stored
#'   specification,
#'   \item solve stage: the model is solved and results are returned in a
#'   separate \code{Solution} or \code{SolutionSet} object.
#' }
#'
#' The \code{Problem} object itself is not the solution. It is the structured
#' problem definition from which a solution can be obtained.
#'
#' @section Methods:
#' \describe{
#'   \item{\code{print()}}{Print a structured summary of the stored problem
#'   specification, including data, actions and effects, spatial inputs, targets
#'   and constraints, and model status. If a model has already been built,
#'   additional dimensions and auxiliary-variable information are displayed.}
#'
#'   \item{\code{show()}}{Alias of \code{print()}.}
#'
#'   \item{\code{repr()}}{Return a short one-line representation of the object.}
#'
#'   \item{\code{getData(name)}}{Return a named entry from \code{self$data}.}
#'
#'   \item{\code{getPlanningUnitsAmount()}}{Return the number of planning units
#'   stored in \code{x$data$pu}.}
#'
#'   \item{\code{getMonitoringCosts()}}{Return the planning-unit cost vector,
#'   typically taken from \code{x$data$pu$cost}.}
#'
#'   \item{\code{getFeatureAmount()}}{Return the number of stored features.}
#'
#'   \item{\code{getFeatureNames()}}{Return feature names from
#'   \code{x$data$features$name}, or feature ids if names are unavailable.}
#'
#'   \item{\code{getActionCosts()}}{Return action-level costs from
#'   \code{x$data$dist_actions$cost} when available.}
#'
#'   \item{\code{getActionsAmount()}}{Return the number of stored actions.}
#' }
#'
#' @section Printing and diagnostics:
#' The \code{print()} method is intended as a quick diagnostic summary. It helps
#' users understand:
#' \itemize{
#'   \item what data have already been loaded,
#'   \item whether actions, effects, spatial relations, targets, and constraints
#'   have been registered,
#'   \item whether objectives and methods have been configured,
#'   \item whether a model has already been built,
#'   \item and whether the object appears ready to be solved.
#' }
#'
#' In particular, the model section of the printed output summarizes whether the
#' current problem specification is incomplete, ready, or already materialized as
#' a built optimization model.
#'
#' @return No return value. This page documents the \code{Problem} class.
#'
#' @seealso
#' \code{\link{create_problem}},
#' \code{\link{add_actions}},
#' \code{\link{add_effects}},
#' \code{\link{add_constraint_targets_absolute}},
#' \code{\link{solve}}
NULL

.pa_data_status_text <- function(self) {
  if (.pa_has_model(self)) {
    "built"
  } else {
    "not built yet (will build in solve())"
  }
}

.pa_data_solver_text <- function(self) {
  sa <- self$data$solve_args %||% list()
  solver <- sa$solver %||% NULL
  if (is.null(solver) || !nzchar(as.character(solver)[1])) {
    return("not set (auto)")
  }
  as.character(solver)[1]
}

.pa_method_summary <- function(self) {
  obj_sum <- .pa_objectives_summary(self)
  method <- self$data$method %||% NULL

  if (is.list(method) && length(method) > 0L) {
    mtype <- as.character(method$type %||% method$name %||% "configured")[1]

    if (identical(mtype, "epsilon_constraint")) {
      primary <- as.character(method$primary %||% NA_character_)[1]
      if (!is.na(primary) && nzchar(primary)) {
        return(list(
          text = paste0("epsilon_constraint (primary: ", primary, ")"),
          is_set = TRUE
        ))
      }
    }

    if (identical(mtype, "weighted")) {
      aliases <- as.character(method$aliases %||% character(0))
      if (length(aliases) > 0L) {
        return(list(
          text = paste0("weighted (", paste(aliases, collapse = ", "), ")"),
          is_set = TRUE
        ))
      }
    }

    return(list(
      text = mtype,
      is_set = TRUE
    ))
  }

  if (obj_sum$n <= 1L) {
    return(list(
      text = "single-objective",
      is_set = TRUE
    ))
  }

  list(
    text = "not set",
    is_set = FALSE
  )
}

.pa_model_checks_text <- function(self) {
  obj_sum <- .pa_objectives_summary(self)
  met_sum <- .pa_method_summary(self)

  if (obj_sum$n == 0L) {
    return("incomplete (no objective registered)")
  }

  if (obj_sum$n > 1L && !isTRUE(met_sum$is_set)) {
    return("incomplete (multiple objectives registered but no MO method selected)")
  }

  "ok"
}

.pa_preview_text <- function(x, max_show = 3L, quote = TRUE) {
  x <- as.character(x)
  x <- x[!is.na(x) & nzchar(x)]

  if (length(x) == 0L) {
    return("none")
  }

  prev <- utils::head(x, max_show)
  if (isTRUE(quote)) {
    prev <- paste0('"', prev, '"')
  }

  txt <- paste(prev, collapse = ", ")
  if (length(x) > max_show) {
    txt <- paste0(txt, ", ...")
  }
  txt
}

.pa_features_summary <- function(self, max_show = 3L) {
  feats <- self$data$features
  if (is.null(feats) || !inherits(feats, "data.frame") || nrow(feats) == 0) {
    return(list(
      n = 0L,
      preview = "none"
    ))
  }

  nm <- if ("name" %in% names(feats)) as.character(feats$name) else as.character(feats$id)

  list(
    n = nrow(feats),
    preview = .pa_preview_text(nm, max_show = max_show, quote = TRUE)
  )
}

.pa_actions_summary <- function(self, max_show = 3L) {
  acts <- self$data$actions

  if (is.null(acts) || !inherits(acts, "data.frame") || nrow(acts) == 0) {
    return(list(
      n = 0L,
      preview = "none"
    ))
  }

  nm <- if ("name" %in% names(acts)) {
    as.character(acts$name)
  } else {
    as.character(acts$id)
  }

  list(
    n = nrow(acts),
    preview = .pa_preview_text(nm, max_show = max_show, quote = TRUE)
  )
}

.pa_effects_summary <- function(self) {
  de <- self$data$dist_effects
  dp <- self$data$dist_profit

  out <- list(
    n_effects = .pa_nrow0(de),
    n_profit = .pa_nrow0(dp),
    effect_mode = "none"
  )

  if (!is.null(de) && inherits(de, "data.frame") && nrow(de) > 0) {
    has_b <- "benefit" %in% names(de) && any(de$benefit > 0, na.rm = TRUE)
    has_l <- "loss" %in% names(de) && any(de$loss > 0, na.rm = TRUE)

    out$effect_mode <- if (has_b && has_l) {
      "benefit + loss"
    } else if (has_b) {
      "benefit only"
    } else if (has_l) {
      "loss only"
    } else {
      "all zero"
    }
  }

  out
}

.pa_targets_summary <- function(self, max_show = 3L) {
  t <- self$data$targets

  if (is.null(t) || !inherits(t, "data.frame") || nrow(t) == 0) {
    return(list(
      n_targets = 0L,
      preview = NULL
    ))
  }

  lab <- if ("feature_name" %in% names(t)) {
    as.character(t$feature_name)
  } else if ("feature" %in% names(t)) {
    as.character(t$feature)
  } else {
    rep("?", nrow(t))
  }

  val <- if ("target_value" %in% names(t)) {
    as.numeric(t$target_value)
  } else {
    rep(NA_real_, nrow(t))
  }

  sense <- if ("sense" %in% names(t)) as.character(t$sense) else rep("ge", nrow(t))
  sense_txt <- ifelse(sense %in% c("ge", ">="), ">=", ifelse(sense %in% c("le", "<="), "<=", sense))

  prev_n <- min(nrow(t), max_show)
  prev <- paste0('"', lab[seq_len(prev_n)], '" ', sense_txt[seq_len(prev_n)], " ", signif(val[seq_len(prev_n)], 4))

  list(
    n_targets = nrow(t),
    preview = prev
  )
}

.pa_constraints_summary <- function(self) {
  out <- list(
    area_constraints = 0L,
    budget_constraints = 0L,
    pu_locked_in = 0L,
    pu_locked_out = 0L,
    action_locked_in = 0L,
    action_locked_out = 0L
  )

  cons <- self$data$constraints %||% list()

  # area constraints
  if (is.list(cons) && !is.null(cons$area)) {
    area <- cons$area

    if (is.data.frame(area) && nrow(area) > 0L) {
      out$area_constraints <- as.integer(nrow(area))

      if ("sense" %in% names(area)) {
        s <- unique(as.character(area$sense))
        s <- s[!is.na(s) & nzchar(s)]
        out$area_sense <- if (length(s) > 0L) paste(s, collapse = ", ") else NA_character_
      }

      if ("actions" %in% names(area)) {
        a <- as.character(area$actions)
        a_lab <- ifelse(is.na(a) | !nzchar(a), "all", a)
        out$area_actions <- a_lab
      }

      if (all(c("actions", "sense") %in% names(area))) {
        a <- as.character(area$actions)
        a_lab <- ifelse(is.na(a) | !nzchar(a), "all", a)
        s <- as.character(area$sense)
        out$area_labels <- paste0("[", a_lab, "] ", s)
      }

    } else if (is.list(area)) {
      # backward compatibility with older single-constraint format
      out$area_constraints <- 1L
      out$area_sense <- as.character(area$sense %||% NA_character_)[1]

      area_actions <- area$actions %||% NA_character_
      area_actions <- as.character(area_actions)[1]
      out$area_actions <- if (is.na(area_actions) || !nzchar(area_actions)) {
        "all"
      } else {
        area_actions
      }
      out$area_labels <- paste0(
        "[", out$area_actions, "] ",
        as.character(area$sense %||% NA_character_)[1]
      )
    }
  }

  # budget constraints
  if (is.list(cons) && !is.null(cons$budget)) {
    budget <- cons$budget

    if (is.data.frame(budget) && nrow(budget) > 0L) {
      out$budget_constraints <- as.integer(nrow(budget))

      if ("sense" %in% names(budget)) {
        s <- unique(as.character(budget$sense))
        s <- s[!is.na(s) & nzchar(s)]
        out$budget_sense <- if (length(s) > 0L) paste(s, collapse = ", ") else NA_character_
      }

      if ("actions" %in% names(budget)) {
        a <- as.character(budget$actions)
        a_lab <- ifelse(is.na(a) | !nzchar(a), "all", a)
        out$budget_actions <- a_lab
      }

      if (all(c("actions", "sense") %in% names(budget))) {
        a <- as.character(budget$actions)
        a_lab <- ifelse(is.na(a) | !nzchar(a), "all", a)
        s <- as.character(budget$sense)
        out$budget_labels <- paste0("[", a_lab, "] ", s)
      }

    } else if (is.list(budget)) {
      # backward compatibility with older single-constraint format
      out$budget_constraints <- 1L
      out$budget_sense <- as.character(budget$sense %||% NA_character_)[1]

      budget_actions <- budget$actions %||% NA_character_
      budget_actions <- as.character(budget_actions)[1]
      out$budget_actions <- if (is.na(budget_actions) || !nzchar(budget_actions)) {
        "all"
      } else {
        budget_actions
      }
      out$budget_labels <- paste0(
        "[", out$budget_actions, "] ",
        as.character(budget$sense %||% NA_character_)[1]
      )
    }
  }

  # planning-unit locks
  pu <- self$data$pu
  if (!is.null(pu) && inherits(pu, "data.frame")) {
    if ("locked_in" %in% names(pu)) {
      out$pu_locked_in <- sum(
        isTRUE(pu$locked_in) | (!is.na(pu$locked_in) & pu$locked_in),
        na.rm = TRUE
      )
    }
    if ("locked_out" %in% names(pu)) {
      out$pu_locked_out <- sum(
        isTRUE(pu$locked_out) | (!is.na(pu$locked_out) & pu$locked_out),
        na.rm = TRUE
      )
    }
  }

  # action locks
  da <- self$data$dist_actions
  if (!is.null(da) && inherits(da, "data.frame") && "status" %in% names(da)) {
    # ajusta estos códigos si tu convención final cambia
    out$action_locked_in <- sum(da$status %in% c(1, 2), na.rm = TRUE)
    out$action_locked_out <- sum(da$status %in% c(3), na.rm = TRUE)
  }

  out
}

.pa_model_aux_summary <- function(self) {
  if (!.pa_has_model(self)) return(NULL)

  d <- .pa_model_dims(self)
  frag <- .pa_model_frag_vars_summary(self)

  out <- list(
    n_con = d$n_con %||% 0L,
    n_var = d$n_var %||% 0L,
    nnz   = d$nnz %||% 0L,
    n_z = 0L,
    n_y_pu = 0L,
    n_y_actions = 0L,
    n_y_interventions = 0L
  )

  op_list <- tryCatch(
    .pa_model_from_ptr(
      self$data$model_ptr,
      args = self$data$model_args %||% list(),
      drop_triplets = TRUE
    ),
    error = function(e) NULL
  )

  if (!is.null(op_list)) {
    out$n_z <- as.integer(op_list$n_z %||% 0L)
  }

  if (!is.null(frag)) {
    out$n_y_pu <- as.integer(frag$n_y_pu %||% 0L)
    out$n_y_actions <- as.integer(frag$n_y_actions %||% 0L)
    out$n_y_interventions <- as.integer(frag$n_y_interventions %||% 0L)
  }

  out
}

.pa_objectives_summary <- function(self, max_show = 4L) {
  objs <- self$data$objectives %||% NULL

  if (is.null(objs) || !is.list(objs) || length(objs) == 0L) {
    return(list(
      n = 0L,
      aliases = character(0),
      preview = "none"
    ))
  }

  aliases <- names(objs)

  if (is.null(aliases) || !any(nzchar(aliases))) {
    aliases <- vapply(
      objs,
      function(o) as.character(o$objective_id %||% "objective")[1],
      character(1)
    )
  }

  aliases <- aliases[!is.na(aliases) & nzchar(aliases)]

  list(
    n = length(aliases),
    aliases = aliases,
    preview = .pa_preview_text(aliases, max_show = max_show, quote = FALSE)
  )
}

.pa_print_model_section <- function(self, ch) {
  status_txt <- .pa_data_status_text(self)
  obj_sum <- .pa_objectives_summary(self)
  met_sum <- .pa_method_summary(self)
  solver_txt <- .pa_data_solver_text(self)
  checks_txt <- .pa_model_checks_text(self)

  cli::cli_text("{ch$l}{ch$b}{.h model}", .envir = environment())

  if (.pa_has_model(self)) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}status:         {.ok {status_txt}}", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}status:         {.warn {status_txt}}", .envir = environment())
  }

  if (obj_sum$n == 0L) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}objectives:     {.muted none}", .envir = environment())
  } else if (obj_sum$n == 1L) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}objectives:     1 registered ({obj_sum$preview})", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}objectives:     {obj_sum$n} registered ({obj_sum$preview})", .envir = environment())
  }

  if (identical(met_sum$text, "not set")) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}method:         {.warn {met_sum$text}}", .envir = environment())
  } else if (identical(met_sum$text, "single-objective")) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}method:         {.ok {met_sum$text}}", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}method:         {met_sum$text}", .envir = environment())
  }

  if (identical(solver_txt, "not set (auto)")) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}solver:         {.muted {solver_txt}}", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}solver:         {solver_txt}", .envir = environment())
  }

  if (identical(checks_txt, "ok")) {
    cli::cli_text(" {ch$v}{ch$l}{ch$b}checks:         {.ok {checks_txt}}", .envir = environment())
  } else if (startsWith(checks_txt, "incomplete")) {
    cli::cli_text(" {ch$v}{ch$l}{ch$b}checks:         {.warn {checks_txt}}", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$l}{ch$b}checks:         {.bad {checks_txt}}", .envir = environment())
  }
}

.pa_print_built_model_section <- function(self, ch) {
  if (!.pa_has_model(self)) return(invisible(NULL))

  sm <- .pa_model_aux_summary(self)
  if (is.null(sm)) return(invisible(NULL))

  cli::cli_text("{ch$l}{ch$b}{.h built model details}", .envir = environment())
  cli::cli_text(" {ch$v}{ch$j}{ch$b}dimensions:     {sm$n_con} constraints, {sm$n_var} vars, {sm$nnz} nnz",
                .envir = environment())

  if (sm$n_z > 0) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}z:              {sm$n_z} vars", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}z:              {.muted none}", .envir = environment())
  }

  if (sm$n_y_pu > 0) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}y_pu:           {sm$n_y_pu} vars", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}y_pu:           {.muted none}", .envir = environment())
  }

  if (sm$n_y_actions > 0) {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}y_actions:      {sm$n_y_actions} vars", .envir = environment())
  } else {
    cli::cli_text(" {ch$v}{ch$j}{ch$b}y_actions:      {.muted none}", .envir = environment())
  }

  if (sm$n_y_interventions > 0) {
    cli::cli_text(
      " {ch$v}{ch$l}{ch$b}y_interventions: {sm$n_y_interventions} vars",
      .envir = environment()
    )
  } else {
    cli::cli_text(
      " {ch$v}{ch$l}{ch$b}y_interventions: {.muted none}",
      .envir = environment()
    )
  }

  invisible(NULL)
}

#' @export
Problem <- pproto(
  "Problem",
  data = list(),

  print = function(self) {
    ch <- .pa_cli_box_chars()
    div_id <- cli::cli_div(theme = .pa_cli_theme())

    cli::cli_text("A multiscape object ({.cls Problem})")

    pu <- self$data$pu
    n_pu <- .pa_nrow0(pu)
    unit_rng <- .pa_safe_range(.pa_get_cost_vec(pu))

    feat_sum <- .pa_features_summary(self, max_show = 3L)
    act_sum <- .pa_actions_summary(self, max_show = 3L)

    pu_coords <- self$data$pu_coords
    has_coords <- .pa_has_coords(self)

    pu_sf <- self$data$pu_sf
    has_sf <- !is.null(pu_sf) && inherits(pu_sf, "sf")

    rels_sum <- .pa_spatial_relations_summary(self)
    n_rels <- if (is.null(rels_sum)) 0L else nrow(rels_sum)

    actions <- self$data$actions
    n_act <- .pa_nrow0(actions)

    dist_actions <- self$data$dist_actions
    n_dist_act <- .pa_nrow0(dist_actions)
    act_cost_rng <- .pa_safe_range(.pa_get_action_cost_vec(self))

    eff_sum <- .pa_effects_summary(self)
    tgt_sum <- .pa_targets_summary(self)
    cons_sum <- .pa_constraints_summary(self)

    # ---- PROBLEM SECTION
    cli::cli_text("{ch$j}{ch$b}{.h data}", .envir = environment())

    pu_cls <- if (is.null(pu)) "NULL" else class(pu)[1]
    cli::cli_text(
      "{ch$v}{ch$j}{ch$b}planning units:  {.cls {pu_cls}} ({n_pu} total)",
      .envir = environment()
    )

    if (is.null(unit_rng)) {
      cli::cli_text("{ch$v}{ch$j}{ch$b}costs:           {.muted (not available)}",
                    .envir = environment())
    } else {
      mn <- unit_rng[[1]]
      mx <- unit_rng[[2]]
      cli::cli_text("{ch$v}{ch$j}{ch$b}costs:           min: {mn}, max: {mx}",
                    .envir = environment())
    }

    cli::cli_text(
      "{ch$v}{ch$l}{ch$b}features:        {feat_sum$n} total ({feat_sum$preview})",
      .envir = environment()
    )

    # ---- ACTIONS AND EFFECTS SECTION
    cli::cli_text("{ch$l}{ch$b}{.h actions and effects}", .envir = environment())

    if (n_act == 0) {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}actions:         {.muted none}",
                    .envir = environment())
      cli::cli_text(" {ch$v}{ch$j}{ch$b}feasible action pairs:    {.muted none}",
                    .envir = environment())
    } else {
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}actions:         {act_sum$n} total ({act_sum$preview})",
        .envir = environment()
      )
      cli::cli_text(" {ch$v}{ch$j}{ch$b}feasible action pairs:    {n_dist_act} feasible rows",
                    .envir = environment())

      if (is.null(act_cost_rng)) {
        cli::cli_text(" {ch$v}{ch$j}{ch$b}action costs:    {.muted (not available)}",
                      .envir = environment())
      } else {
        mn <- act_cost_rng[[1]]
        mx <- act_cost_rng[[2]]
        cli::cli_text(" {ch$v}{ch$j}{ch$b}action costs:    min: {mn}, max: {mx}",
                      .envir = environment())
      }
    }

    if (eff_sum$n_effects == 0) {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}effect data:    {.muted none}",
                    .envir = environment())
    } else {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}effect data:    {eff_sum$n_effects} rows",
                    .envir = environment())
      cli::cli_text(" {ch$v}{ch$j}{ch$b}effect mode:     {eff_sum$effect_mode}",
                    .envir = environment())
    }

    if (eff_sum$n_profit == 0) {
      cli::cli_text(" {ch$v}{ch$l}{ch$b}profit data:     {.muted none}",
                    .envir = environment())
    } else {
      cli::cli_text(" {ch$v}{ch$l}{ch$b}profit data:     {eff_sum$n_profit} rows",
                    .envir = environment())
    }

    # ---- SPATIAL SECTION
    cli::cli_text("{ch$l}{ch$b}{.h spatial}", .envir = environment())

    if (has_sf) {
      n_sf <- nrow(pu_sf)
      cli::cli_text(" {ch$v}{ch$j}{ch$b}geometry:        sf ({n_sf} rows)",
                    .envir = environment())
    } else {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}geometry:        {.muted none}",
                    .envir = environment())
    }

    if (has_coords) {
      n_c <- nrow(pu_coords)
      xr <- .pa_safe_range(pu_coords$x)
      yr <- .pa_safe_range(pu_coords$y)
      if (is.null(xr) || is.null(yr)) {
        cli::cli_text(" {ch$v}{ch$j}{ch$b}coordinates:      {n_c} rows",
                      .envir = environment())
      } else {
        cli::cli_text(" {ch$v}{ch$j}{ch$b}coordinates:      {n_c} rows (x: {xr[[1]]}..{xr[[2]]}, y: {yr[[1]]}..{yr[[2]]})",
                      .envir = environment())
      }
    } else {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}coordinates:      {.muted none}",
                    .envir = environment())
    }

    if (n_rels == 0) {
      cli::cli_text(" {ch$v}{ch$l}{ch$b}relations:       {.muted none}",
                    .envir = environment())
    } else if (n_rels == 1L) {
      nm <- rels_sum$name[1]
      ed <- rels_sum$edges[1]
      w1 <- rels_sum$w_min[1]
      w2 <- rels_sum$w_max[1]

      if (is.na(w1) || is.na(w2)) {
        cli::cli_text(" {ch$v}{ch$l}{ch$b}relations:       {nm} ({ed} edges)",
                      .envir = environment())
      } else {
        cli::cli_text(" {ch$v}{ch$l}{ch$b}relations:       {nm} ({ed} edges, w: {w1}..{w2})",
                      .envir = environment())
      }
    } else {
      cli::cli_text(" {ch$v}{ch$j}{ch$b}relations:       {n_rels} registered",
                    .envir = environment())

      max_show <- 6L
      show_sum <- rels_sum[seq_len(min(n_rels, max_show)), , drop = FALSE]

      for (i in seq_len(nrow(show_sum))) {
        nm <- show_sum$name[i]
        ed <- show_sum$edges[i]
        w1 <- show_sum$w_min[i]
        w2 <- show_sum$w_max[i]
        if (is.na(w1) || is.na(w2)) {
          cli::cli_text(" {ch$v}{ch$j}{ch$b}- {nm}: {ed} edges",
                        .envir = environment())
        } else {
          cli::cli_text(" {ch$v}{ch$j}{ch$b}- {nm}: {ed} edges (w: {w1}..{w2})",
                        .envir = environment())
        }
      }

      if (n_rels > max_show) {
        extra <- n_rels - max_show
        cli::cli_text(" {ch$v}{ch$l}{ch$b}{.muted ... +{extra} more relation(s)}",
                      .envir = environment())
      }
    }

    # ---- TARGETS AND CONSTRAINTS SECTION
    cli::cli_text("{ch$l}{ch$b}{.h targets and constraints}", .envir = environment())

    if (tgt_sum$n_targets == 0L) {
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}targets:          {.muted none}",
        .envir = environment()
      )
    } else {
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}targets:          {tgt_sum$n_targets} rows",
        .envir = environment()
      )

      if (!is.null(tgt_sum$preview) && length(tgt_sum$preview) > 0) {
        prev_txt <- paste(tgt_sum$preview, collapse = ", ")
        cli::cli_text(
          " {ch$v}{ch$j}{ch$b}target preview:   {prev_txt}",
          .envir = environment()
        )
      }
    }

    if (cons_sum$area_constraints == 0L) {

      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}area constraints: {.muted none}",
        .envir = environment()
      )

    } else if (cons_sum$area_constraints == 1L) {

      area_lab <- cons_sum$area_labels %||% NA_character_
      area_lab <- as.character(area_lab)[1]

      if (is.na(area_lab) || !nzchar(area_lab)) {
        area_lab <- cons_sum$area_sense %||% "unknown"
      }

      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}area constraints: 1 ({area_lab})",
        .envir = environment()
      )

    } else {

      n_area <- as.integer(cons_sum$area_constraints)
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}area constraints: {n_area} registered",
        .envir = environment()
      )

      labs <- cons_sum$area_labels %||% character(0)
      labs <- as.character(labs)
      labs <- labs[!is.na(labs) & nzchar(labs)]

      max_show <- 4L
      if (length(labs) > 0L) {
        show_labs <- labs[seq_len(min(length(labs), max_show))]

        for (lab in show_labs) {
          cli::cli_text(
            " {ch$v}{ch$j}{ch$b}- {lab}",
            .envir = environment()
          )
        }

        if (length(labs) > max_show) {
          extra <- length(labs) - max_show
          cli::cli_text(
            " {ch$v}{ch$j}{ch$b}{.muted ... +{extra} more area constraint(s)}",
            .envir = environment()
          )
        }
      }
    }

    if (cons_sum$budget_constraints == 0L) {

      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}budget constraints: {.muted none}",
        .envir = environment()
      )

    } else if (cons_sum$budget_constraints == 1L) {

      budget_lab <- cons_sum$budget_labels %||% NA_character_
      budget_lab <- as.character(budget_lab)[1]

      if (is.na(budget_lab) || !nzchar(budget_lab)) {
        budget_lab <- cons_sum$budget_sense %||% "unknown"
      }

      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}budget constraints: 1 ({budget_lab})",
        .envir = environment()
      )

    } else {

      n_budget <- as.integer(cons_sum$budget_constraints)
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}budget constraints: {n_budget} registered",
        .envir = environment()
      )

      labs <- cons_sum$budget_labels %||% character(0)
      labs <- as.character(labs)
      labs <- labs[!is.na(labs) & nzchar(labs)]

      max_show <- 4L
      if (length(labs) > 0L) {
        show_labs <- labs[seq_len(min(length(labs), max_show))]

        for (lab in show_labs) {
          cli::cli_text(
            " {ch$v}{ch$j}{ch$b}- {lab}",
            .envir = environment()
          )
        }

        if (length(labs) > max_show) {
          extra <- length(labs) - max_show
          cli::cli_text(
            " {ch$v}{ch$j}{ch$b}{.muted ... +{extra} more budget constraint(s)}",
            .envir = environment()
          )
        }
      }
    }

    pu_lock_total <- cons_sum$pu_locked_in + cons_sum$pu_locked_out
    if (pu_lock_total > 0L) {
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}planning-unit locks: {pu_lock_total} units ({cons_sum$pu_locked_in} locked-in, {cons_sum$pu_locked_out} locked-out)",
        .envir = environment()
      )
    } else {
      cli::cli_text(
        " {ch$v}{ch$j}{ch$b}planning-unit locks: {.muted none}",
        .envir = environment()
      )
    }

    act_lock_total <- cons_sum$action_locked_in + cons_sum$action_locked_out
    if (act_lock_total > 0L) {
      cli::cli_text(
        " {ch$v}{ch$l}{ch$b}action locks:       {act_lock_total} rows ({cons_sum$action_locked_in} locked-in, {cons_sum$action_locked_out} locked-out)",
        .envir = environment()
      )
    } else {
      cli::cli_text(
        " {ch$v}{ch$l}{ch$b}action locks:       {.muted none}",
        .envir = environment()
      )
    }

    # ---- MODEL SECTION
    .pa_print_model_section(self, ch)

    # ---- BUILT MODEL DETAILS
    .pa_print_built_model_section(self, ch)

    info_sym <- cli::symbol$info
    if (is.function(info_sym)) info_sym <- info_sym()
    cli::cli_text(cli::col_grey(
      paste0("# ", info_sym, " Use {.code x$data} to inspect stored tables and model snapshots.")
    ))

    cli::cli_end(div_id)
    invisible(TRUE)
  },

  show = function(self) self$print(),

  repr = function(self) {
    pu_n <- .pa_nrow0(self$data$pu)
    ft_n <- .pa_nrow0(self$data$features)
    paste0("<Problem> ", pu_n, " planning units, ", ft_n, " features")
  },

  getData = function(self, x) {
    assertthat::assert_that(assertthat::is.string(x))
    if (!x %in% names(self$data)) return("x object do not found")
    self$data[[x]]
  },

  getPlanningUnitsAmount = function(self) {
    pu <- self$data$pu
    if (is.null(pu)) return(0L)
    nrow(pu)
  },

  getMonitoringCosts = function(self) {
    pu <- self$data$pu
    if (is.null(pu) || nrow(pu) == 0) return(numeric(0))
    if ("cost" %in% names(pu)) return(as.numeric(pu$cost))
    numeric(0)
  },

  getFeatureAmount = function(self) {
    feats <- self$data$features
    if (is.null(feats)) return(0L)
    nrow(feats)
  },

  getFeatureNames = function(self) {
    feats <- self$data$features
    if (is.null(feats) || nrow(feats) == 0) return(character(0))
    if ("name" %in% names(feats)) return(as.character(feats$name))
    as.character(feats$id)
  },

  getActionCosts = function(self) {
    if (!is.null(self$data$dist_actions) &&
        inherits(self$data$dist_actions, "data.frame") &&
        nrow(self$data$dist_actions) > 0 &&
        "cost" %in% names(self$data$dist_actions)) {
      return(as.numeric(self$data$dist_actions$cost))
    }
    numeric(0)
  },

  getActionsAmount = function(self) {
    if (!is.null(self$data$actions) && inherits(self$data$actions, "data.frame")) {
      return(nrow(self$data$actions))
    }
    0L
  }
)

#' @include internal.R
NULL

#' @title Print
#'
#' @description Displays information about an object.
#'
#' @param x Any object.
#' @param ... Not used.
#'
#' @name print
#'
#' @return None.
#'
#' @seealso [base::print()].
#'
#' @aliases print
#' @keywords internal

NULL

#' @rdname print
#' @method print Problem
#' @export
print.Problem <- function(x, ...) x$print()


#' @rdname print
#' @method print Solution
#' @export
print.Solution <- function(x, ...) x$print()

#' @rdname print
#' @method print SolutionSet
#' @export
print.SolutionSet <- function(x, ...) x$print()

#' @title Set the AUGMECON multi-objective method
#'
#' @description
#' Configure a \code{Problem} object to be solved with the augmented
#' epsilon-constraint method (AUGMECON).
#'
#' AUGMECON is an exact multi-objective optimization method in which one
#' objective is treated as the primary objective and the remaining objectives
#' are converted into \eqn{\varepsilon}-constraints. In the augmented formulation,
#' each secondary objective is associated with a non-negative slack variable,
#' and the primary objective is augmented with a small reward term based on the
#' normalized slacks. This augmentation is used to avoid weakly efficient
#' solutions, following Mavrotas (2009).
#'
#' This function does not solve the problem directly. It stores the AUGMECON
#' configuration in \code{x$data$method}, to be used later by
#' \code{\link{solve}}.
#'
#' @details
#' Use this method when one objective should be optimized directly, the
#' remaining objectives should be controlled through epsilon levels, and weakly
#' efficient solutions should be reduced through the augmented formulation.
#'
#' \strong{General idea}
#'
#' Suppose that \eqn{m \ge 2} objective functions have already been registered
#' in the problem:
#' \deqn{
#' f_1(x), f_2(x), \dots, f_m(x).
#' }
#'
#' AUGMECON selects one of them as the primary objective, say
#' \eqn{f_p(x)}, and treats the remaining \eqn{m - 1} objectives as secondary
#' objectives.
#'
#' For a fixed combination of epsilon levels, the method solves a sequence of
#' single-objective subproblems of the form:
#'
#' \deqn{
#' \max \; f_p(x) + \rho \sum_{k \in \mathcal{S}} \frac{s_k}{R_k}
#' }
#'
#' subject to
#'
#' \deqn{
#' f_k(x) - s_k = \varepsilon_k, \qquad k \in \mathcal{S},
#' }
#'
#' \deqn{
#' s_k \ge 0, \qquad k \in \mathcal{S},
#' }
#'
#' together with all original feasibility constraints of the planning problem.
#'
#' Here:
#' \itemize{
#'   \item \eqn{f_p(x)} is the primary objective,
#'   \item \eqn{\mathcal{S}} is the set of secondary objectives,
#'   \item \eqn{\varepsilon_k} is the imposed level for secondary objective
#'   \eqn{k},
#'   \item \eqn{s_k} is a non-negative slack variable,
#'   \item \eqn{R_k} is the payoff-table range used to normalize objective
#'   \eqn{k},
#'   \item \eqn{\rho > 0} is a small augmentation coefficient.
#' }
#'
#' In the original AUGMECON formulation of Mavrotas (2009), the augmentation
#' term ensures that, among solutions with the same primary objective value, the
#' solver prefers those with larger normalized slack, thereby avoiding weakly
#' efficient points and improving Pareto-front generation.
#'
#' \strong{Secondary-objective equalities and slacks}
#'
#' The key difference between standard epsilon-constraint and AUGMECON is that
#' the secondary objectives are written as equalities with slacks rather than as
#' simple inequalities. For a maximization-type secondary objective, this takes
#' the form:
#'
#' \deqn{
#' f_k(x) - s_k = \varepsilon_k, \qquad s_k \ge 0.
#' }
#'
#' This implies:
#' \deqn{
#' f_k(x) \ge \varepsilon_k,
#' }
#'
#' while explicitly measuring the excess above the imposed epsilon level through
#' \eqn{s_k}. The augmentation term then rewards such excess in normalized form.
#'
#' In implementation terms, the exact sign convention for each objective depends
#' on whether it is internally treated as a minimization or maximization
#' objective, but the method always preserves the same AUGMECON principle:
#' \itemize{
#'   \item one objective is optimized directly,
#'   \item all others are turned into constrained objectives,
#'   \item non-negative slacks measure controlled deviation from the imposed
#'   epsilon levels,
#'   \item the primary objective is augmented with a small slack-based reward.
#' }
#'
#' \strong{Manual and automatic epsilon grids}
#'
#' AUGMECON requires a grid of epsilon levels for each secondary objective.
#'
#' If \code{grid} is supplied, it must be a named list with one numeric vector
#' per secondary objective. Each vector defines the exact epsilon levels to be
#' explored for that objective.
#'
#' If \code{grid = NULL}, the grid is generated automatically later during
#' \code{\link{solve}}. In that case, the method first computes extreme points
#' and payoff-table ranges for the secondary objectives, and then generates
#' \code{n_points} levels for each one.
#'
#' If \code{include_extremes = TRUE}, the automatic grid includes the extreme
#' values of each secondary objective.
#'
#' If \code{lexicographic = TRUE}, extreme points are computed using
#' lexicographic anchoring, which can improve payoff-table quality when
#' objectives are tightly competing. The tolerance used for lexicographic
#' anchoring is controlled by \code{lexicographic_tol}.
#'
#' \strong{Normalization and augmentation}
#'
#' The augmentation term is scaled using the payoff-table ranges of the
#' secondary objectives. If \eqn{R_k} denotes the range of secondary objective
#' \eqn{k}, then the effective coefficient applied to the slack is:
#'
#' \deqn{
#' \frac{\rho}{R_k},
#' }
#'
#' where \eqn{\rho = \code{augmentation}}.
#'
#' This normalization is important because different objectives may be measured
#' on very different numerical scales. Without normalization, a slack belonging
#' to a large-scale objective could dominate the augmentation term simply due to
#' units.
#'
#' In this implementation, the user supplies \code{augmentation} as the base
#' coefficient \eqn{\rho}, while the normalized slack coefficients are computed
#' internally at solve time using the corresponding payoff-table ranges.
#'
#' \strong{Stored configuration}
#'
#' This function stores the method definition in \code{x$data$method} with:
#' \itemize{
#'   \item \code{name = "augmecon"},
#'   \item the primary objective alias,
#'   \item the full set of participating aliases,
#'   \item the set of secondary aliases,
#'   \item either a manual grid or the information required to generate one
#'   automatically,
#'   \item augmentation and slack-bound parameters.
#' }
#'
#' The actual payoff table, grid construction, and subproblem solution loop are
#' performed later by \code{\link{solve}}.
#'
#' @param x A \code{Problem} object.
#' @param primary Character string giving the alias of the primary objective,
#'   that is, the objective optimized directly in the AUGMECON formulation.
#' @param aliases Optional character vector of objective aliases to include in
#'   the method. If \code{NULL}, all registered objective aliases are used. The
#'   value of \code{primary} must be included in \code{aliases}.
#' @param grid Optional named list defining manual epsilon levels for the
#'   secondary objectives. Each name must correspond to a secondary objective
#'   alias, and each element must be a non-empty numeric vector of finite
#'   values. If \code{NULL}, the grid is generated automatically.
#' @param n_points Number of automatically generated epsilon levels per
#'   secondary objective when \code{grid = NULL}. Must be at least 2. Ignored
#'   when \code{grid} is supplied.
#' @param include_extremes Logical. If \code{TRUE}, automatically generated
#'   grids include extreme values of each secondary objective.
#' @param lexicographic Logical. If \code{TRUE}, use lexicographic anchoring
#'   when computing extreme points for automatic grid construction.
#' @param lexicographic_tol Non-negative numeric tolerance used in
#'   lexicographic anchoring.
#' @param augmentation Positive numeric augmentation coefficient
#'   \eqn{\rho}. The effective coefficient of each secondary slack is computed
#'   internally as \eqn{\rho / R_k}, where \eqn{R_k} is the payoff-table range
#'   of the corresponding secondary objective.
#' @param slack_upper_bound Positive numeric upper bound imposed on the explicit
#'   non-negative slack variables associated with the secondary objectives.
#'
#' @return The updated \code{Problem} object with the AUGMECON method
#'   configuration stored in \code{x$data$method}.
#'
#' @references
#' Mavrotas, G. (2009). Effective implementation of the
#' \eqn{\varepsilon}-constraint method in multi-objective mathematical programming
#' problems. \emph{Applied Mathematics and Computation}, 213(2), 455--465.
#'
#' @examples
#' # Small toy problem
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' effects_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   feature = c(1, 1, 1, 1, 2, 2, 2, 2),
#'   benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
#'   loss = c(0, 0, 1, 0, 0, 1, 0, 0)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_effects(effects_df) |>
#'   add_objective_max_benefit(alias = "benefit") |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_min_loss(alias = "loss")
#'
#' # Automatic epsilon grids generated later during solve()
#' x1 <- set_method_augmecon(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost"),
#'   n_points = 5,
#'   include_extremes = TRUE,
#'   lexicographic = TRUE,
#'   augmentation = 1e-3
#' )
#'
#' x1$data$method
#'
#' # Manual epsilon grids for two secondary objectives
#' x2 <- set_method_augmecon(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost", "loss"),
#'   grid = list(
#'     cost = c(4, 6, 8),
#'     loss = c(0, 1)
#'   ),
#'   augmentation = 1e-3,
#'   slack_upper_bound = 1e6
#' )
#'
#' x2$data$method
#'
#' @seealso
#' \code{\link{set_method_epsilon_constraint}},
#' \code{\link{set_method_weighted_sum}},
#' \code{\link{solve}}
#'
#' @export
set_method_augmecon <- function(
    x,
    primary,
    aliases = NULL,
    grid = NULL,
    n_points = 10,
    include_extremes = TRUE,
    lexicographic = TRUE,
    lexicographic_tol = 1e-9,
    augmentation = 1e-3,
    slack_upper_bound = 1e6
) {

  # ---- primary
  if (!is.character(primary) || length(primary) != 1L || is.na(primary) || !nzchar(primary)) {
    stop("`primary` must be a non-empty character string.", call. = FALSE)
  }
  primary <- as.character(primary)

  # ---- aliases
  if (is.null(aliases)) {
    aliases <- .pamo_get_specs(x)
    obj_alias <- names(aliases)
    aliases <- obj_alias
  } else {
    if (!is.character(aliases) || length(aliases) == 0L || anyNA(aliases)) {
      stop("`aliases` must be NULL or a non-empty character vector without NA.", call. = FALSE)
    }
    aliases <- as.character(aliases)

    if (any(!nzchar(aliases))) {
      stop("`aliases` must not contain empty strings.", call. = FALSE)
    }

    if (anyDuplicated(aliases) != 0L) {
      dups <- unique(aliases[duplicated(aliases)])
      stop("`aliases` must not contain duplicates: ", paste(dups, collapse = ", "), call. = FALSE)
    }
  }

  # validate aliases exist
  .pamo_get_objective_specs(x, aliases)

  if (!(primary %in% aliases)) {
    stop("`primary` must be included in `aliases`.", call. = FALSE)
  }

  if (length(aliases) < 2L) {
    stop("AUGMECON requires at least 2 objectives.", call. = FALSE)
  }

  secondary <- setdiff(aliases, primary)
  if (length(secondary) == 0L) {
    stop("AUGMECON requires at least one secondary objective.", call. = FALSE)
  }

  # ---- common arguments
  if (!is.logical(include_extremes) || length(include_extremes) != 1L || is.na(include_extremes)) {
    stop("`include_extremes` must be TRUE or FALSE.", call. = FALSE)
  }
  include_extremes <- isTRUE(include_extremes)

  if (!is.logical(lexicographic) || length(lexicographic) != 1L || is.na(lexicographic)) {
    stop("`lexicographic` must be TRUE or FALSE.", call. = FALSE)
  }
  lexicographic <- isTRUE(lexicographic)

  if (!is.numeric(lexicographic_tol) || length(lexicographic_tol) != 1L ||
      is.na(lexicographic_tol) || !is.finite(lexicographic_tol) || lexicographic_tol < 0) {
    stop("`lexicographic_tol` must be a single finite non-negative number.", call. = FALSE)
  }
  lexicographic_tol <- as.numeric(lexicographic_tol)

  if (!is.numeric(augmentation) || length(augmentation) != 1L ||
      is.na(augmentation) || !is.finite(augmentation) || augmentation <= 0) {
    stop("`augmentation` must be a single finite positive number.", call. = FALSE)
  }
  augmentation <- as.numeric(augmentation)

  if (!is.numeric(slack_upper_bound) || length(slack_upper_bound) != 1L ||
      is.na(slack_upper_bound) || !is.finite(slack_upper_bound) || slack_upper_bound <= 0) {
    stop("`slack_upper_bound` must be a single finite positive number.", call. = FALSE)
  }
  slack_upper_bound <- as.numeric(slack_upper_bound)

  # ---- automatic grid
  if (is.null(grid)) {

    if (!is.numeric(n_points) || length(n_points) != 1L ||
        is.na(n_points) || !is.finite(n_points) || n_points < 2) {
      stop("`n_points` must be a single number >= 2 when `grid = NULL`.", call. = FALSE)
    }
    n_points <- as.integer(n_points)

  } else {

    # ---- manual grid
    if (is.atomic(grid) && !is.list(grid)) {
      stop("`grid` must be NULL or a named list.", call. = FALSE)
    }

    if (!is.list(grid) || length(grid) == 0L || is.null(names(grid)) || any(!nzchar(names(grid)))) {
      stop("`grid` must be a named non-empty list when supplied.", call. = FALSE)
    }

    gnames <- names(grid)

    extra <- setdiff(gnames, secondary)
    miss  <- setdiff(secondary, gnames)

    if (length(extra) > 0L) {
      stop(
        "`grid` contains names not corresponding to secondary objectives: ",
        paste(extra, collapse = ", "),
        call. = FALSE
      )
    }

    if (length(miss) > 0L) {
      stop(
        "`grid` is missing secondary objectives: ",
        paste(miss, collapse = ", "),
        call. = FALSE
      )
    }

    grid <- grid[secondary]

    grid <- lapply(seq_along(grid), function(i) {
      nm <- secondary[i]
      v  <- grid[[i]]

      if (!is.numeric(v) || length(v) == 0L || anyNA(v) || any(!is.finite(v))) {
        stop("`grid[['", nm, "']]` must be a non-empty numeric vector of finite values.", call. = FALSE)
      }

      sort(unique(as.numeric(v)))
    })
    names(grid) <- secondary

    n_points <- NULL
  }

  x$data$method <- list(
    name = "augmecon",
    primary = primary,
    aliases = aliases,
    secondary = secondary,
    grid = grid,
    n_points = n_points,
    include_extremes = include_extremes,
    lexicographic = lexicographic,
    lexicographic_tol = lexicographic_tol,
    augmentation = augmentation,
    slack_upper_bound = slack_upper_bound
  )

  x
}

#' @title Set the epsilon-constraint multi-objective method
#'
#' @description
#' Configure a \code{Problem} object to be solved with the
#' epsilon-constraint multi-objective method.
#'
#' In this method, one objective is designated as the \emph{primary} objective
#' and is optimized directly, while the remaining objectives are transformed
#' into \eqn{\varepsilon}-constraints.
#'
#' Two operating modes are supported:
#' \itemize{
#'   \item \strong{manual mode}: the user supplies the \eqn{\varepsilon}-levels
#'   explicitly,
#'   \item \strong{automatic mode}: the \eqn{\varepsilon}-levels are generated
#'   later during \code{\link{solve}} from extreme-point or payoff information.
#' }
#'
#' This function does not solve the problem. It stores the method configuration
#' in \code{x$data$method}, to be used later by \code{\link{solve}}.
#'
#' @details
#' Use this method when one objective should be optimized directly while the
#' remaining objectives are controlled through explicit performance thresholds.
#'
#' \strong{General idea}
#'
#' Suppose that \eqn{m \ge 2} objective functions have already been registered
#' in the problem:
#' \deqn{
#' f_1(x), f_2(x), \dots, f_m(x).
#' }
#'
#' The epsilon-constraint method selects one of them as the primary objective,
#' say \eqn{f_p(x)}, and treats the remaining objectives as constrained
#' objectives.
#'
#' For a fixed vector of epsilon levels, the method solves subproblems in which
#' the primary objective is optimized directly and the remaining objectives are
#' imposed through epsilon constraints.
#'
#' A representative formulation is:
#'
#' \deqn{
#' \max \; f_p(x)
#' }
#'
#' subject to
#'
#' \deqn{
#' f_k(x) \ge \varepsilon_k, \qquad k \in \mathcal{C},
#' }
#'
#' together with all original feasibility constraints of the planning problem,
#' where \eqn{\mathcal{C}} is the set of constrained objectives.
#'
#' Depending on objective sense, the internal implementation may transform
#' minimization objectives into equivalent constrained forms, but the method
#' always follows the same principle:
#' \itemize{
#'   \item one objective is optimized directly,
#'   \item all remaining objectives are imposed through
#'   \eqn{\varepsilon}-constraints.
#' }
#'
#' By solving the problem repeatedly for different epsilon levels, the method
#' generates a set of efficient trade-off solutions.
#'
#' \strong{Atomic objectives requirement}
#'
#' The epsilon-constraint method can only be used with atomic objectives that
#' have already been registered under aliases. These aliases are typically
#' created by calling objective setters with an \code{alias} argument, for
#' example:
#' \preformatted{
#' x <- x |>
#'   add_objective_max_benefit(alias = "benefit") |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_min_fragmentation(alias = "frag")
#' }
#'
#' The \code{primary} argument selects which registered objective is optimized
#' directly. The remaining aliases are treated as constrained objectives.
#'
#' \strong{Manual mode}
#'
#' In \code{mode = "manual"}, the user must provide \code{eps}.
#'
#' The \code{eps} argument can be supplied as:
#' \itemize{
#'   \item a named numeric vector, defining a single run,
#'   \item or a named list of numeric vectors, defining a grid of runs.
#' }
#'
#' The names of \code{eps} must correspond exactly to the constrained objective
#' aliases, that is, to all aliases in \code{aliases} except \code{primary}.
#'
#' If the constrained objectives are
#' \eqn{\mathcal{C} = \{c_1, \dots, c_q\}}, then manual mode creates a design
#' grid containing all combinations of the supplied epsilon levels for the
#' constrained objectives.
#'
#' Each row of this grid defines one subproblem to be solved later.
#'
#' \strong{Important:} manual mode supports \strong{two or more objectives}.
#' In particular, it can be used with:
#' \itemize{
#'   \item 2 objectives: 1 primary + 1 constrained objective,
#'   \item 3 or more objectives: 1 primary + multiple constrained objectives.
#' }
#'
#' Thus, manual mode is the general way to use the epsilon-constraint method
#' when more than two objectives are involved.
#'
#' In manual mode, the generated design grid is stored immediately in
#' \code{x$data$method$runs}. Its epsilon columns are named
#' \code{eps_<alias>}, for example \code{eps_frag}.
#'
#' \strong{Automatic mode}
#'
#' In \code{mode = "auto"}, the user omits \code{eps} and instead supplies
#' \code{n_points}.
#'
#' In this case, the epsilon grid is not built immediately. Instead, it is
#' constructed later during \code{\link{solve}} using extreme-point or
#' payoff-table information.
#'
#' In the current implementation, automatic mode supports
#' \strong{exactly two objectives only}:
#' \itemize{
#'   \item one primary objective,
#'   \item one constrained objective.
#' }
#'
#' Therefore, if \code{mode = "auto"}, then \code{aliases} must contain exactly
#' two objective aliases. Problems with three or more objectives must use
#' \code{mode = "manual"}.
#'
#' If \code{include_extremes = TRUE}, the automatically generated grid includes
#' the extreme values of the constrained objective. Otherwise, only interior
#' values are used.
#'
#' If \code{lexicographic = TRUE}, the extreme points used to generate the grid
#' are computed lexicographically. In that case, one objective is optimized
#' first, and then the second objective is optimized while constraining the
#' first to remain within \code{lexicographic_tol} of its optimum.
#'
#' \strong{Stored configuration}
#'
#' The configured method stores:
#' \itemize{
#'   \item \code{name = "epsilon_constraint"},
#'   \item \code{mode},
#'   \item \code{primary},
#'   \item \code{aliases},
#'   \item \code{constrained},
#'   \item epsilon design information,
#'   \item lexicographic configuration.
#' }
#'
#' In manual mode, \code{x$data$method$runs} contains the explicit design grid.
#'
#' In automatic mode, \code{x$data$method$runs} is initially \code{NULL} and is
#' generated later during \code{\link{solve}}.
#'
#' For more than two objectives, automatic grid generation is currently
#' unavailable because the number of epsilon combinations grows rapidly and
#' requires explicit user control.
#'
#' @param x A \code{Problem} object.
#' @param primary Character string giving the alias of the primary objective to
#'   optimize directly.
#' @param eps Optional epsilon specification used only in \code{mode = "manual"}.
#'   It may be:
#'   \itemize{
#'     \item a named numeric vector, defining epsilon values for a single run,
#'     \item or a named list of numeric vectors, defining epsilon values for a
#'     grid of runs.
#'   }
#'   Names must correspond exactly to the constrained objective aliases.
#' @param aliases Optional character vector of objective aliases to include.
#'   By default, all registered objective aliases are used. The value of
#'   \code{primary} must be included in \code{aliases}.
#' @param mode Character string. Must be either \code{"manual"} or
#'   \code{"auto"}.
#' @param n_points Integer scalar used only in \code{mode = "auto"}. Number of
#'   epsilon points to generate automatically for the constrained objective.
#'   Must be at least 2.
#' @param include_extremes Logical scalar used only in \code{mode = "auto"}. If
#'   \code{TRUE}, include extreme epsilon values in the automatically generated
#'   grid.
#' @param lexicographic Logical scalar used only in \code{mode = "auto"}. If
#'   \code{TRUE}, compute extreme points lexicographically.
#' @param lexicographic_tol Numeric scalar \eqn{\ge 0}. Tolerance used in
#'   lexicographic extreme-point computation.
#'
#' @return An updated \code{Problem} object with the epsilon-constraint method
#'   configuration stored in \code{x$data$method}.
#'
#' In manual mode, \code{x$data$method$runs} contains the explicit
#' epsilon-design grid.
#'
#' In automatic mode, \code{x$data$method$runs} is \code{NULL} until the grid is
#' generated later during \code{\link{solve}}.
#'
#' @examples
#' # Small toy problem
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' effects_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   feature = c(1, 1, 1, 1, 2, 2, 2, 2),
#'   benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
#'   loss = c(0, 0, 1, 0, 0, 1, 0, 0)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_effects(effects_df) |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_max_benefit(alias = "benefit") |>
#'   add_objective_min_loss(alias = "loss")
#'
#' # Manual mode with one constrained objective
#' x1 <- set_method_epsilon_constraint(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost"),
#'   mode = "manual",
#'   eps = c(cost = 5)
#' )
#'
#' x1$data$method
#'
#' # Manual mode with multiple epsilon values
#' x2 <- set_method_epsilon_constraint(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost"),
#'   mode = "manual",
#'   eps = list(cost = c(4, 6, 8))
#' )
#'
#' x2$data$method$runs
#'
#' # Manual mode with more than two objectives
#' x3 <- set_method_epsilon_constraint(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost", "loss"),
#'   mode = "manual",
#'   eps = list(
#'     cost = c(4, 6),
#'     loss = c(0, 1)
#'   )
#' )
#'
#' x3$data$method$runs
#'
#' # Automatic mode currently supports exactly two objectives
#' x4 <- set_method_epsilon_constraint(
#'   x,
#'   primary = "benefit",
#'   aliases = c("benefit", "cost"),
#'   mode = "auto",
#'   n_points = 5,
#'   include_extremes = TRUE,
#'   lexicographic = TRUE,
#'   lexicographic_tol = 1e-8
#' )
#'
#' x4$data$method
#'
#' @seealso
#' \code{\link{set_method_augmecon}},
#' \code{\link{set_method_weighted_sum}},
#' \code{\link{solve}}
#'
#' @export
set_method_epsilon_constraint <- function(
    x,
    primary,
    eps = NULL,
    aliases = NULL,
    mode = c("manual", "auto"),
    n_points = 10,
    include_extremes = TRUE,
    lexicographic = TRUE,
    lexicographic_tol = 1e-8
) {
  #x <- .pamo_as_mo(x)
  stopifnot(inherits(x, "Problem"))

  mode <- match.arg(mode)

  primary <- as.character(primary)[1]
  if (is.na(primary) || !nzchar(primary)) {
    stop("primary must be a non-empty string.", call. = FALSE)
  }

  .pamo_validate_objectives(x)

  specs_all <- .pamo_get_specs(x)
  obj_alias <- names(specs_all)
  if (!primary %in% obj_alias) {
    stop("primary alias not found: '", primary, "'.", call. = FALSE)
  }

  if (is.null(aliases)) aliases <- obj_alias
  aliases <- as.character(aliases)

  if (any(!aliases %in% obj_alias)) {
    bad <- aliases[!aliases %in% obj_alias]
    stop("Unknown aliases: ", paste(bad, collapse = ", "), call. = FALSE)
  }

  if (!primary %in% aliases) {
    stop("primary must be included in aliases.", call. = FALSE)
  }

  constrained <- setdiff(aliases, primary)

  if (length(constrained) == 0L) {
    stop("At least one constrained objective is required.", call. = FALSE)
  }

  if (!is.logical(lexicographic) || length(lexicographic) != 1L || is.na(lexicographic)) {
    stop("lexicographic must be TRUE or FALSE.", call. = FALSE)
  }

  lexicographic_tol <- as.numeric(lexicographic_tol)[1]
  if (!is.finite(lexicographic_tol) || lexicographic_tol < 0) {
    stop("lexicographic_tol must be a finite non-negative number.", call. = FALSE)
  }

  if (mode == "manual") {
    if (is.null(eps)) {
      stop("In mode='manual', eps must be provided.", call. = FALSE)
    }

    if (is.numeric(eps) && !is.null(names(eps))) {
      eps_list <- as.list(eps)
      eps_list <- lapply(eps_list, function(v) c(as.numeric(v)[1]))
    } else if (is.list(eps) && length(eps) > 0 && !is.null(names(eps))) {
      eps_list <- lapply(eps, function(v) as.numeric(v))
    } else {
      stop("eps must be a named numeric vector or a named list of numeric vectors.", call. = FALSE)
    }

    miss <- setdiff(constrained, names(eps_list))
    if (length(miss) > 0) {
      stop("eps must include all constrained objectives. Missing: ", paste(miss, collapse = ", "), call. = FALSE)
    }

    extra <- setdiff(names(eps_list), constrained)
    if (length(extra) > 0) {
      stop(
        "eps contains aliases that are not constrained objectives: ",
        paste(extra, collapse = ", "),
        call. = FALSE
      )
    }

    bad_empty <- names(eps_list)[vapply(eps_list, length, integer(1)) == 0L]
    if (length(bad_empty) > 0) {
      stop("eps contains empty vectors for: ", paste(bad_empty, collapse = ", "), call. = FALSE)
    }

    bad_nonfinite <- names(eps_list)[vapply(eps_list, function(v) any(!is.finite(v)), logical(1))]
    if (length(bad_nonfinite) > 0) {
      stop("eps contains non-finite values for: ", paste(bad_nonfinite, collapse = ", "), call. = FALSE)
    }

    grid <- expand.grid(
      eps_list[constrained],
      KEEP.OUT.ATTRS = FALSE,
      stringsAsFactors = FALSE
    )
    if (nrow(grid) == 0) {
      stop("Empty epsilon grid.", call. = FALSE)
    }

    names(grid) <- paste0("eps_", names(grid))
    grid$run_id <- seq_len(nrow(grid))

    x$data$method <- list(
      name = "epsilon_constraint",
      mode = "manual",
      primary = primary,
      aliases = aliases,
      constrained = constrained,
      eps = eps_list,
      runs = grid,
      lexicographic = isTRUE(lexicographic),
      lexicographic_tol = lexicographic_tol
    )

    return(x)
  }

  if (length(aliases) != 2L) {
    stop(
      "set_method_epsilon_constraint(mode='auto') currently supports exactly 2 objectives.\n",
      "Use mode='manual' for 3+ objectives.",
      call. = FALSE
    )
  }

  n_points <- as.integer(n_points)[1]
  if (!is.finite(n_points) || is.na(n_points) || n_points < 2L) {
    stop("n_points must be an integer >= 2.", call. = FALSE)
  }

  x$data$method <- list(
    name = "epsilon_constraint",
    mode = "auto",
    primary = primary,
    aliases = aliases,
    constrained = constrained,
    n_points = n_points,
    include_extremes = isTRUE(include_extremes),
    lexicographic = isTRUE(lexicographic),
    lexicographic_tol = lexicographic_tol,
    runs = NULL
  )

  x
}

#' @include internal.R
#'
#' @name spatial_relations
#' @title Spatial relations infrastructure
#'
#' @description
#' Define a unified internal representation for spatial relations between
#' planning units and provide constructors for building such relations from
#' common spatial inputs.
#'
#' Spatial relations in \code{multiscape} are graph-like structures defined over
#' the set of planning units. They are used to represent adjacency, shared
#' boundary length, proximity, neighbourhood graphs, or other pairwise spatial
#' relationships that may later be used by objectives, constraints, or
#' diagnostics.
#'
#' Relations are stored inside the \code{Problem} object as
#' \code{x$data$spatial_relations[[name]]}.
#'
#' @details
#' \strong{Internal representation}
#'
#' Each spatial relation is stored as a \code{data.frame} containing, at a
#' minimum, the columns:
#' \describe{
#'   \item{\code{internal_pu1}}{Integer index of the first planning unit.}
#'   \item{\code{internal_pu2}}{Integer index of the second planning unit.}
#'   \item{\code{weight}}{Non-negative edge weight. Its interpretation depends
#'   on the relation type.}
#' }
#'
#' Additional columns may also be present, such as \code{pu1}, \code{pu2},
#' \code{distance}, \code{source}, or other diagnostic fields. These extra
#' columns are preserved when possible.
#'
#' Thus, a spatial relation can be interpreted as a weighted graph
#' \eqn{G = (\mathcal{I}, E, \omega)}, where:
#' \itemize{
#'   \item \eqn{\mathcal{I}} is the set of planning units,
#'   \item \eqn{E} is the set of stored edges,
#'   \item \eqn{\omega_{ij} \ge 0} is the weight associated with edge
#'   \eqn{(i,j)}.
#' }
#'
#' \strong{Directed and undirected relations}
#'
#' Some relation constructors create undirected relations by default. In that
#' case, each unordered pair is typically stored once, usually with
#' \eqn{i < j}, unless self-edges are explicitly allowed.
#'
#' \strong{Boundary relations and diagonal terms}
#'
#' The function \code{\link{add_spatial_boundary}} supports boundary-length
#' relations derived either from a boundary table or from polygon geometry.
#'
#' If \code{include_self = TRUE}, diagonal entries \eqn{(i,i)} may be added to
#' represent exposed boundary. These diagonal weights are intended for
#' boundary-based compactness or fragmentation formulations, where perimeter
#' exposed to the outside of the selected solution should contribute to the
#' objective.
#'
#' \strong{Geometry safety}
#'
#' All \code{sf}-based constructors operate only through spatial predicates,
#' topological relations, or boundary-length calculations. They never subdivide,
#' cut, or alter planning-unit geometries. Planning units are aligned to the
#' internal planning-unit order before spatial relations are computed.
#'
#' \strong{Available constructors}
#'
#' The following functions build and register spatial relations:
#' \itemize{
#'   \item \code{\link{add_spatial_relations}}: register a precomputed relation,
#'   \item \code{\link{add_spatial_boundary}}: shared boundary length relation,
#'   \item \code{\link{add_spatial_rook}}: rook adjacency from polygons,
#'   \item \code{\link{add_spatial_queen}}: queen adjacency from polygons,
#'   \item \code{\link{add_spatial_knn}}: k-nearest-neighbours graph from
#'   coordinates,
#'   \item \code{\link{add_spatial_distance}}: distance-threshold graph from
#'   coordinates.
#' }
#'
#' @keywords internal
NULL

#' @title Add spatial relations
#'
#' @description
#' Register an externally computed spatial relation inside a \code{Problem}
#' object using the unified internal representation adopted by \code{multiscape}.
#'
#' Most users will typically prefer one of the convenience constructors such as
#' \code{\link{add_spatial_boundary}}, \code{\link{add_spatial_rook}},
#' \code{\link{add_spatial_queen}}, \code{\link{add_spatial_knn}}, or
#' \code{\link{add_spatial_distance}}. This function is the advanced low-level
#' entry point for adding an already computed relation.
#'
#' @details
#' Use this function when the spatial relation has already been computed
#' externally and should be registered directly in the \code{Problem} object.
#'
#' The input relation may be provided either in terms of external planning-unit
#' identifiers or in terms of internal planning-unit indices.
#'
#' Specifically, the input \code{relations} table must contain either:
#' \itemize{
#'   \item \code{pu1}, \code{pu2}, and \code{weight}, or
#'   \item \code{internal_pu1}, \code{internal_pu2}, and \code{weight}.
#' }
#'
#' If external ids are supplied, they are mapped to internal indices using the
#' planning-unit identifiers stored in the problem.
#'
#' Let \eqn{G = (\mathcal{I}, E, \omega)} denote the supplied relation, where
#' \eqn{E} corresponds to the rows of \code{relations}. If
#' \code{directed = FALSE}, each edge is treated as undirected, so pairs
#' \eqn{(i,j)} and \eqn{(j,i)} are interpreted as the same edge. In that case,
#' duplicated undirected edges are collapsed automatically using the maximum
#' weight observed for each unordered pair.
#'
#' If \code{directed = TRUE}, edges are preserved as ordered pairs, so
#' \eqn{(i,j)} and \eqn{(j,i)} are distinct unless the user provides both.
#'
#' Self-edges \eqn{(i,i)} are permitted only if \code{allow_self = TRUE}.
#'
#' The final relation is stored in \code{x$data$spatial_relations[[name]]}.
#'
#' If a relation with the same \code{name} already exists, it is replaced.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}.
#' @param relations A \code{data.frame} describing relation edges. It must
#'   contain either:
#'   \itemize{
#'     \item \code{pu1}, \code{pu2}, and \code{weight}, using external
#'     planning-unit ids, or
#'     \item \code{internal_pu1}, \code{internal_pu2}, and \code{weight}, using
#'     internal planning-unit indices.
#'   }
#'   Extra columns such as \code{distance} or \code{source} are allowed and are
#'   preserved when possible.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param directed Logical. If \code{FALSE}, treat edges as undirected and
#'   collapse duplicate unordered pairs. If \code{TRUE}, keep edges as directed
#'   ordered pairs.
#' @param allow_self Logical. If \code{TRUE}, allow self-edges
#'   \eqn{(i,i)}. Default is \code{FALSE}.
#'
#' @return An updated \code{Problem} object with the relation stored in
#'   \code{x$data$spatial_relations[[name]]}.
#'
#' @examples
#' pu <- data.frame(id = 1:3, cost = c(1, 2, 3))
#'
#' features <- data.frame(
#'   id = 1,
#'   name = "sp1"
#' )
#'
#' dist_features <- data.frame(
#'   pu = 1:3,
#'   feature = 1,
#'   amount = c(1, 1, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' rel <- data.frame(
#'   pu1 = c(1, 1, 2),
#'   pu2 = c(2, 3, 3),
#'   weight = c(1, 1, 2)
#' )
#'
#' p <- add_spatial_relations(
#'   x = p,
#'   relations = rel,
#'   name = "my_relation"
#' )
#'
#' p$data$spatial_relations$my_relation
#'
#' @seealso
#' \code{\link{add_spatial_boundary}},
#' \code{\link{add_spatial_rook}},
#' \code{\link{add_spatial_queen}},
#' \code{\link{add_spatial_knn}},
#' \code{\link{add_spatial_distance}}
#'
#' @export
add_spatial_relations <- function(x,
                                  relations,
                                  name,
                                  directed = FALSE,
                                  allow_self = FALSE) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)
  n_pu <- nrow(x$data$pu)

  if (!is.character(name) || length(name) != 1 || is.na(name) || !nzchar(name)) {
    stop("name must be a non-empty character string.", call. = FALSE)
  }

  stopifnot(inherits(relations, "data.frame"), nrow(relations) > 0)
  rel <- relations

  # map pu ids -> internal ids if needed
  if (all(c("pu1", "pu2", "weight") %in% names(rel)) &&
      !all(c("internal_pu1", "internal_pu2") %in% names(rel))) {

    rel$pu1 <- as.integer(rel$pu1)
    rel$pu2 <- as.integer(rel$pu2)
    rel$weight <- as.numeric(rel$weight)

    idx <- x$data$index$pu
    rel$internal_pu1 <- unname(idx[as.character(rel$pu1)])
    rel$internal_pu2 <- unname(idx[as.character(rel$pu2)])

    if (anyNA(rel$internal_pu1) || anyNA(rel$internal_pu2)) {
      stop("Some pu1/pu2 ids were not found in x$data$pu$id.", call. = FALSE)
    }
  }

  keep_cols <- c("internal_pu1","internal_pu2","weight","pu1","pu2","distance","source")
  rel <- rel[, intersect(keep_cols, names(rel)), drop = FALSE]

  if (!all(c("internal_pu1","internal_pu2","weight") %in% names(rel))) {
    stop(
      "relations must contain either (pu1, pu2, weight) or (internal_pu1, internal_pu2, weight).",
      call. = FALSE
    )
  }

  .pa_rbind_samecols <- function(a, b) {
    cols <- union(names(a), names(b))
    a2 <- a; b2 <- b
    for (cc in setdiff(cols, names(a2))) a2[[cc]] <- NA
    for (cc in setdiff(cols, names(b2))) b2[[cc]] <- NA
    a2 <- a2[, cols, drop = FALSE]
    b2 <- b2[, cols, drop = FALSE]
    base::rbind(a2, b2)
  }

  rel$internal_pu1 <- as.integer(rel$internal_pu1)
  rel$internal_pu2 <- as.integer(rel$internal_pu2)
  rel$weight <- as.numeric(rel$weight)

  if (any(rel$internal_pu1 < 1L | rel$internal_pu1 > n_pu)) stop("internal_pu1 out of range.", call. = FALSE)
  if (any(rel$internal_pu2 < 1L | rel$internal_pu2 > n_pu)) stop("internal_pu2 out of range.", call. = FALSE)
  if (!allow_self && any(rel$internal_pu1 == rel$internal_pu2)) stop("Self-edges are not allowed.", call. = FALSE)
  #if (any(!is.finite(rel$weight)) || any(rel$weight < 0)) stop("weight must be finite and >= 0.", call. = FALSE)

  if (!directed) {
    self_edges <- rel[rel$internal_pu1 == rel$internal_pu2, , drop = FALSE]
    off_edges  <- rel[rel$internal_pu1 != rel$internal_pu2, , drop = FALSE]

    if (nrow(off_edges) > 0) {
      a <- pmin(off_edges$internal_pu1, off_edges$internal_pu2)
      b <- pmax(off_edges$internal_pu1, off_edges$internal_pu2)
      off_edges$internal_pu1 <- a
      off_edges$internal_pu2 <- b

      key <- paste(off_edges$internal_pu1, off_edges$internal_pu2, sep = "_")
      wmax <- tapply(off_edges$weight, key, max)

      parts <- strsplit(names(wmax), "_", fixed = TRUE)
      i1 <- as.integer(vapply(parts, `[`, "", 1))
      j1 <- as.integer(vapply(parts, `[`, "", 2))

      rel_u <- data.frame(
        internal_pu1 = i1,
        internal_pu2 = j1,
        weight = as.numeric(wmax),
        stringsAsFactors = FALSE
      )

      extra_cols <- intersect(names(off_edges), c("pu1","pu2","distance","source"))
      if (length(extra_cols) > 0) {
        rep_idx <- match(names(wmax), key)
        extras <- off_edges[rep_idx, extra_cols, drop = FALSE]
        rel_u <- cbind(rel_u, extras)
      }
    } else {
      rel_u <- off_edges
    }

    if (nrow(self_edges) > 0) {
      key <- as.character(self_edges$internal_pu1)
      wself <- tapply(self_edges$weight, key, max)

      self_fix <- data.frame(
        internal_pu1 = as.integer(names(wself)),
        internal_pu2 = as.integer(names(wself)),
        weight = as.numeric(wself),
        stringsAsFactors = FALSE
      )

      extra_cols <- intersect(names(self_edges), c("pu1","pu2","distance","source"))
      if (length(extra_cols) > 0) {
        rep_idx <- match(names(wself), as.character(self_edges$internal_pu1))
        extras <- self_edges[rep_idx, extra_cols, drop = FALSE]
        self_fix <- cbind(self_fix, extras)
      }

      rel <- if (nrow(rel_u) == 0) self_fix else .pa_rbind_samecols(rel_u, self_fix)
    } else {
      rel <- rel_u
    }
  }

  rel$relation_name <- name
  x <- .pa_store_relation(x, rel, name)
  x
}

.sum_incident_by_pu <- function(i, j, w, n_pu) {
  i <- as.integer(i); j <- as.integer(j); w <- as.numeric(w)

  s1 <- rowsum(w, i, reorder = FALSE)
  s2 <- rowsum(w, j, reorder = FALSE)

  out <- numeric(n_pu)
  out[as.integer(rownames(s1))] <- out[as.integer(rownames(s1))] + s1[, 1]
  out[as.integer(rownames(s2))] <- out[as.integer(rownames(s2))] + s2[, 1]
  out
}

#' @title Add spatial boundary-length relations
#'
#' @description
#' Build and register a boundary-length spatial relation between planning units.
#'
#' Boundary relations represent shared edge length between adjacent polygons. In
#' contrast to queen adjacency, they only account for boundary segments of
#' positive length and ignore point-only contacts.
#'
#' @details
#' Use this function when spatial structure should be represented through shared
#' boundary length rather than binary contiguity or coordinate-based proximity.
#'
#' Two input modes are supported:
#' \enumerate{
#'   \item \strong{Boundary-table mode.} If \code{boundary} is supplied, it is
#'   interpreted as a boundary table, for example a Marxan-style
#'   \code{bound.dat}.
#'
#'   \item \strong{Geometry mode.} If \code{boundary = NULL}, boundary lengths
#'   are derived from polygon geometry using \code{geometry} or
#'   \code{x$data$pu_sf}.
#' }
#'
#' Let \eqn{\omega_{ij} \ge 0} denote the shared boundary length between
#' planning units \eqn{i} and \eqn{j}, multiplied by
#' \code{weight_multiplier}.
#'
#' For off-diagonal entries \eqn{i \neq j}, the stored weight is:
#' \deqn{
#' \omega_{ij} = \mathrm{\gamma} \times b_{ij},
#' }
#' where \eqn{b_{ij}} is the shared boundary length and \eqn{\gamma} is
#' the user-supplied \code{weight_multiplier}.
#'
#' If \code{include_self = TRUE}, diagonal entries are also created. These are
#' not geometric self-neighbours in the graph sense; instead, they represent the
#' effective boundary exposed to the outside of the solution.
#'
#' Let \eqn{p_i} be the total perimeter of planning unit \eqn{i}, and let
#' \eqn{\sum_{j \neq i} \omega_{ij}} be the total incident shared boundary
#' recorded for that planning unit. Then the exposed boundary is represented by
#' a diagonal term derived from the difference between total perimeter and shared
#' boundary, scaled by \code{edge_factor}.
#'
#' These diagonal terms are useful in boundary-based compactness or
#' fragmentation objectives, because they encode the portion of each planning
#' unit's perimeter that would remain exposed if the unit were selected.
#'
#' \strong{Boundary-table mode}
#'
#' If \code{boundary} is provided, accepted formats are:
#' \itemize{
#'   \item \code{(id1, id2, boundary)}, or
#'   \item \code{(pu1, pu2, weight)}.
#' }
#'
#' If the table contains diagonal rows \eqn{(i,i)}, these are interpreted as
#' total perimeter values in boundary-table mode.
#'
#' \strong{Geometry mode}
#'
#' If \code{boundary = NULL}, shared boundary lengths are derived directly from
#' polygon geometry. Only positive-length intersections are retained. Point
#' touches are ignored.
#'
#' \strong{Storage}
#'
#' The final relation is stored through \code{\link{add_spatial_relations}},
#' typically as an undirected relation with optional diagonal entries.
#'
#' @param x A \code{Problem} object.
#' @param boundary Optional \code{data.frame} describing boundary lengths.
#'   Accepted formats are:
#'   \itemize{
#'     \item \code{(id1, id2, boundary)}, or
#'     \item \code{(pu1, pu2, weight)}.
#'   }
#' @param geometry Optional \code{sf} object with planning-unit polygons and an
#'   \code{id} column. If \code{NULL}, \code{x$data$pu_sf} is used.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param weight_col Optional character string giving the name of the weight
#'   column in \code{boundary}. If \code{NULL}, the function tries to infer it
#'   from \code{"boundary"} or \code{"weight"}.
#' @param weight_multiplier Positive numeric scalar applied to all boundary
#'   weights.
#' @param include_self Logical. If \code{TRUE}, include diagonal entries
#'   representing exposed boundary.
#' @param edge_factor Numeric scalar greater than or equal to zero. Multiplier
#'   applied to exposed boundary when constructing diagonal entries.
#'
#' @return An updated \code{Problem} object with the stored relation in
#'   \code{x$data$spatial_relations[[name]]}.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' bound_df <- data.frame(
#'   id1 = c(1, 1, 2, 1, 2, 3, 4),
#'   id2 = c(1, 2, 2, 3, 4, 4, 4),
#'   boundary = c(4, 1, 4, 1, 1, 1, 4)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' p <- add_spatial_boundary(
#'   x = p,
#'   boundary = bound_df,
#'   name = "boundary",
#'   include_self = TRUE,
#'   edge_factor = 1
#' )
#'
#' p$data$spatial_relations$boundary
#'
#' @seealso
#' \code{\link{add_spatial_relations}},
#' \code{\link{add_objective_min_fragmentation_pu}},
#' \code{\link{add_objective_min_fragmentation_action}}
#'
#' @export
add_spatial_boundary <- function(x,
                                 boundary = NULL,
                                 geometry = NULL,
                                 name = "boundary",
                                 weight_col = NULL,
                                 weight_multiplier = 1,
                                 include_self = TRUE,
                                 edge_factor = 1) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)

  edge_factor <- as.numeric(edge_factor)[1]
  if (!is.finite(edge_factor) || edge_factor < 0) {
    stop("edge_factor must be a finite number >= 0.", call. = FALSE)
  }

  weight_multiplier <- as.numeric(weight_multiplier)[1]
  if (!is.finite(weight_multiplier) || weight_multiplier <= 0) {
    stop("weight_multiplier must be a positive finite number.", call. = FALSE)
  }

  # ------------------------------------------------------------
  # Case A) boundary table provided
  # ------------------------------------------------------------
  if (!is.null(boundary)) {

    stopifnot(inherits(boundary, "data.frame"), nrow(boundary) > 0)
    b <- boundary

    if (all(c("id1", "id2") %in% names(b)) && !all(c("pu1", "pu2") %in% names(b))) {
      b$pu1 <- b$id1
      b$pu2 <- b$id2
    }

    if (is.null(weight_col)) {
      if ("boundary" %in% names(b)) weight_col <- "boundary"
      else if ("weight" %in% names(b)) weight_col <- "weight"
      else stop("Could not find a weight column. Provide weight_col.", call. = FALSE)
    }
    if (!(weight_col %in% names(b))) {
      stop("weight_col not found in boundary table.", call. = FALSE)
    }

    rel0 <- data.frame(
      pu1 = as.integer(b$pu1),
      pu2 = as.integer(b$pu2),
      weight = as.numeric(b[[weight_col]]) * weight_multiplier,
      source = "boundary_table",
      stringsAsFactors = FALSE
    )

    x <- .pa_ensure_pu_index(x)
    idx <- x$data$index$pu
    rel0$internal_pu1 <- unname(idx[as.character(rel0$pu1)])
    rel0$internal_pu2 <- unname(idx[as.character(rel0$pu2)])

    if (anyNA(rel0$internal_pu1) || anyNA(rel0$internal_pu2)) {
      stop("Some pu1/pu2 ids were not found in x$data$pu$id.", call. = FALSE)
    }

    n_pu <- nrow(x$data$pu)

    diag_rows <- rel0$internal_pu1 == rel0$internal_pu2

    total <- numeric(n_pu)
    names(total) <- as.character(seq_len(n_pu))
    if (any(diag_rows)) {
      tt <- tapply(rel0$weight[diag_rows], rel0$internal_pu1[diag_rows], sum)
      total[names(tt)] <- tt
    }

    off <- rel0[!diag_rows, , drop = FALSE]

    if (nrow(off) > 0) {
      a <- pmin(off$internal_pu1, off$internal_pu2)
      b2 <- pmax(off$internal_pu1, off$internal_pu2)
      off$internal_pu1 <- a
      off$internal_pu2 <- b2

      key <- paste(off$internal_pu1, off$internal_pu2, sep = "_")
      wmax <- tapply(off$weight, key, max)

      parts <- strsplit(names(wmax), "_", fixed = TRUE)
      i1 <- as.integer(vapply(parts, `[`, "", 1))
      j1 <- as.integer(vapply(parts, `[`, "", 2))

      off_u <- data.frame(
        internal_pu1 = i1,
        internal_pu2 = j1,
        weight = as.numeric(wmax),
        source = "boundary_table_shared",
        stringsAsFactors = FALSE
      )
    } else {
      off_u <- off
    }

    incident <- numeric(n_pu)
    if (nrow(off_u) > 0) {
      incident <- .sum_incident_by_pu(
        i = off_u$internal_pu1,
        j = off_u$internal_pu2,
        w = off_u$weight,
        n_pu = n_pu
      )
    }

    if (!isTRUE(include_self)) {
      if (nrow(off_u) == 0) {
        stop("Boundary table has no off-diagonal edges and include_self=FALSE.", call. = FALSE)
      }
      rel <- off_u
      rel$pu1 <- x$data$pu$id[rel$internal_pu1]
      rel$pu2 <- x$data$pu$id[rel$internal_pu2]
      return(add_spatial_relations(
        x, rel, name = name,
        directed = FALSE,
        allow_self = FALSE
      ))
    }

    has_useful_diagonal <- any(total > 0)

    if (has_useful_diagonal) {
      diag_eff <- edge_factor * (total - incident)
      diag_source <- "boundary_table_diag_effective"
    } else {
      diag_eff <- -incident
      diag_source <- "boundary_table_diag_algebraic"
    }

    rel_self <- data.frame(
      internal_pu1 = seq_len(n_pu),
      internal_pu2 = seq_len(n_pu),
      weight = as.numeric(diag_eff),
      source = diag_source,
      stringsAsFactors = FALSE
    )

    rel <- if (nrow(off_u) == 0) rel_self else rbind(off_u, rel_self)
    rel$pu1 <- x$data$pu$id[rel$internal_pu1]
    rel$pu2 <- x$data$pu$id[rel$internal_pu2]

    return(add_spatial_relations(
      x, rel, name = name,
      directed = FALSE,
      allow_self = TRUE
    ))
  }

  # ------------------------------------------------------------
  # Case B) derive from sf polygons
  # ------------------------------------------------------------
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("add_spatial_boundary(boundary=NULL) requires the 'sf' package.", call. = FALSE)
  }

  if (is.null(geometry)) geometry <- x$data$pu_sf
  if (is.null(geometry)) stop("boundary is NULL and geometry is missing.", call. = FALSE)
  if (!inherits(geometry, "sf")) stop("geometry must be an sf object.", call. = FALSE)
  if (!("id" %in% names(geometry))) stop("geometry must contain an 'id' column.", call. = FALSE)

  geometry$id <- as.integer(geometry$id)
  ord <- match(x$data$pu$id, geometry$id)
  if (anyNA(ord)) stop("geometry$id does not match x$data$pu$id.", call. = FALSE)
  geometry <- geometry[ord, , drop = FALSE]

  geom <- sf::st_make_valid(sf::st_geometry(geometry))
  geometry <- sf::st_set_geometry(geometry, geom)
  geom <- sf::st_geometry(geometry)

  nb <- sf::st_intersects(geometry, geometry, sparse = TRUE)

  pu1 <- integer(0)
  pu2 <- integer(0)
  w <- numeric(0)
  n <- length(nb)

  for (i in seq_len(n)) {
    js <- nb[[i]]
    js <- js[js > i]
    if (!length(js)) next

    for (j in js) {
      inter <- suppressWarnings(sf::st_intersection(geom[i], geom[j]))
      if (length(inter) == 0) next

      gtype <- as.character(sf::st_geometry_type(inter, by_geometry = TRUE))

      if (all(gtype %in% c("LINESTRING", "MULTILINESTRING"))) {
        inter_line <- inter
      } else if (any(gtype == "GEOMETRYCOLLECTION")) {
        inter_line <- suppressWarnings(
          sf::st_collection_extract(inter, "LINESTRING")
        )
        if (length(inter_line) == 0) next
      } else {
        next
      }

      inter_line <- suppressWarnings(sf::st_cast(inter_line, "MULTILINESTRING"))
      len <- suppressWarnings(sf::st_length(inter_line))
      len <- sum(as.numeric(len), na.rm = TRUE)

      if (is.finite(len) && len > 0) {
        pu1 <- c(pu1, geometry$id[i])
        pu2 <- c(pu2, geometry$id[j])
        w   <- c(w, len)
      }
    }
  }

  w <- round(w, 6)

  if (!length(pu1) && !isTRUE(include_self)) {
    stop("No shared-boundary adjacencies found.", call. = FALSE)
  }

  rel_off <- data.frame(
    pu1 = as.integer(pu1),
    pu2 = as.integer(pu2),
    weight = as.numeric(w) * weight_multiplier,
    source = "boundary_sf_shared_length",
    stringsAsFactors = FALSE
  )

  if (!isTRUE(include_self)) {
    return(add_spatial_relations(
      x, rel_off, name = name,
      directed = FALSE,
      allow_self = FALSE
    ))
  }

  total <- as.numeric(sf::st_length(sf::st_boundary(sf::st_geometry(geometry))))
  total <- round(total, 6) * weight_multiplier

  incident <- numeric(nrow(geometry))
  if (nrow(rel_off) > 0) {
    idx_map <- seq_len(nrow(geometry))
    names(idx_map) <- as.character(geometry$id)

    i_int <- idx_map[as.character(rel_off$pu1)]
    j_int <- idx_map[as.character(rel_off$pu2)]

    incident <- .sum_incident_by_pu(
      i = i_int,
      j = j_int,
      w = rel_off$weight,
      n_pu = nrow(geometry)
    )
  }

  has_useful_diagonal <- any(total > 0)

  if (has_useful_diagonal) {
    diag_eff <- edge_factor * (total - incident)
    diag_source <- "boundary_sf_diag_effective"
  } else {
    diag_eff <- -incident
    diag_source <- "boundary_sf_diag_algebraic"
  }

  rel_self <- data.frame(
    pu1 = geometry$id,
    pu2 = geometry$id,
    weight = as.numeric(diag_eff),
    source = diag_source,
    stringsAsFactors = FALSE
  )

  rel <- if (nrow(rel_off) == 0) rel_self else rbind(rel_off, rel_self)

  add_spatial_relations(
    x, rel, name = name,
    directed = FALSE,
    allow_self = TRUE
  )
}

#' @title Add rook adjacency from polygons
#'
#' @description
#' Build and register a rook adjacency relation from planning-unit polygons.
#'
#' Two planning units are rook-adjacent if they share a boundary segment of
#' positive length. Corner-only contact does not count as rook adjacency.
#'
#' @details
#' Use this function when neighbourhood should be defined by shared polygon
#' edges rather than by point-touching or coordinate-based proximity.
#'
#' This constructor derives an adjacency graph from polygon geometry using a
#' rook criterion. If planning units \eqn{i} and \eqn{j} share a common edge of
#' non-zero length, then an edge \eqn{(i,j)} is added to the relation.
#'
#' Let \eqn{G = (\mathcal{I}, E)} denote the resulting graph. Then:
#' \deqn{
#' (i,j) \in E \quad \Longleftrightarrow \quad
#' \mathrm{length}(\partial i \cap \partial j) > 0.
#' }
#'
#' All edges receive the same user-supplied weight.
#'
#' The resulting relation is stored as an undirected spatial relation.
#'
#' @param x A \code{Problem} object created with
#'   \code{\link{create_problem}} or another object containing aligned
#'   planning-unit polygons.
#' @param geometry Optional \code{sf} object with planning-unit polygons and an
#'   \code{id} column. If \code{NULL}, \code{x$data$pu_sf} is used.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param weight Numeric scalar giving the edge weight assigned to each rook
#'   adjacency.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' \donttest{
#' library(terra)
#'
#' data("sim_pu_sf", package = "multiscape")
#' sim_features <- load_sim_features_raster()
#'
#' p <- create_problem(
#'   pu = sim_pu_sf,
#'   features = sim_features,
#'   cost = "cost"
#' )
#'
#' p <- add_spatial_rook(
#'   x = p,
#'   geometry = sim_pu_sf,
#'   name = "rook",
#'   weight = 1
#' )
#'
#' head(p$data$spatial_relations$rook)
#' }
#'
#' @seealso
#' \code{\link{add_spatial_queen}},
#' \code{\link{add_spatial_boundary}}
#'
#' @export
add_spatial_rook <- function(x,
                             geometry = NULL,
                             name = "rook",
                             weight = 1) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)
  if (!.pa_has_sf()) stop("add_spatial_rook requires the 'sf' package.", call. = FALSE)

  geometry <- .pa_get_pu_sf_aligned(x, pu_sf = geometry, arg_name = "geometry")
  weight <- as.numeric(weight)[1]
  if (!is.finite(weight) || weight < 0) stop("weight must be finite and >= 0.", call. = FALSE)

  nb <- sf::st_relate(geometry, geometry, pattern = "F***1****", sparse = TRUE)

  edges <- vector("list", length(nb))
  for (i in seq_along(nb)) {
    js <- nb[[i]]
    if (!length(js)) next
    edges[[i]] <- data.frame(
      pu1 = geometry$id[i],
      pu2 = geometry$id[js],
      weight = weight,
      source = "rook_sf",
      stringsAsFactors = FALSE
    )
  }

  rel <- do.call(rbind, edges)
  if (is.null(rel) || nrow(rel) == 0) stop("No rook adjacencies found.", call. = FALSE)

  add_spatial_relations(
    x, rel, name = name,
    directed = FALSE, allow_self = FALSE
  )
}

#' @title Add queen adjacency from polygons
#'
#' @description
#' Build and register a queen adjacency relation from planning-unit polygons.
#'
#' Two planning units are queen-adjacent if their boundaries touch, either along
#' a shared edge or at a shared vertex.
#'
#' @details
#' Use this function when neighbourhood should include both shared edges and
#' corner-touching polygon contacts.
#'
#' This constructor derives an adjacency graph from polygon geometry using a
#' queen criterion. If planning units \eqn{i} and \eqn{j} touch at any boundary
#' point, then an edge \eqn{(i,j)} is added to the relation.
#'
#' Let \eqn{G = (\mathcal{I}, E)} denote the resulting graph. Then:
#' \deqn{
#' (i,j) \in E \quad \Longleftrightarrow \quad
#' \partial i \cap \partial j \neq \varnothing.
#' }
#'
#' Thus, queen adjacency includes all rook neighbours plus corner-touching
#' neighbours.
#'
#' All edges receive the same user-supplied weight.
#'
#' The resulting relation is stored as an undirected spatial relation.
#'
#' @param x A \code{Problem} object created with
#'   \code{\link{create_problem}} or another object containing aligned
#'   planning-unit polygons.
#' @param geometry Optional \code{sf} object with planning-unit polygons and an
#'   \code{id} column. If \code{NULL}, \code{x$data$pu_sf} is used.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param weight Numeric scalar giving the edge weight assigned to each queen
#'   adjacency.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' \donttest{
#' library(terra)
#'
#' data("sim_pu_sf", package = "multiscape")
#' sim_features <- load_sim_features_raster()
#'
#' p <- create_problem(
#'   pu = sim_pu_sf,
#'   features = sim_features,
#'   cost = "cost"
#' )
#'
#' p <- add_spatial_queen(
#'   x = p,
#'   geometry = sim_pu_sf,
#'   name = "queen",
#'   weight = 1
#' )
#'
#' head(p$data$spatial_relations$queen)
#' }
#'
#' @seealso
#' \code{\link{add_spatial_rook}},
#' \code{\link{add_spatial_boundary}}
#'
#' @export
add_spatial_queen <- function(x,
                              geometry = NULL,
                              name = "queen",
                              weight = 1) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)
  if (!.pa_has_sf()) stop("add_spatial_queen requires the 'sf' package.", call. = FALSE)

  geometry <- .pa_get_pu_sf_aligned(x, pu_sf = geometry, arg_name = "geometry")
  weight <- as.numeric(weight)[1]
  if (!is.finite(weight) || weight < 0) stop("weight must be finite and >= 0.", call. = FALSE)

  nb <- sf::st_touches(geometry, geometry, sparse = TRUE)

  edges <- vector("list", length(nb))
  for (i in seq_along(nb)) {
    js <- nb[[i]]
    js <- js[js > i]
    if (!length(js)) next
    edges[[i]] <- data.frame(
      pu1 = geometry$id[i],
      pu2 = geometry$id[js],
      weight = weight,
      source = "queen_sf",
      stringsAsFactors = FALSE
    )
  }

  rel <- do.call(rbind, edges)
  if (is.null(rel) || nrow(rel) == 0) stop("No queen adjacencies found.", call. = FALSE)

  add_spatial_relations(
    x, rel, name = name,
    directed = FALSE, allow_self = FALSE
  )
}

#' @title Add k-nearest-neighbours spatial relations
#'
#' @description
#' Build and register a k-nearest-neighbours graph between planning units using
#' coordinates.
#'
#' This constructor does not require polygon geometry. It uses planning-unit
#' coordinates supplied explicitly or stored in the \code{Problem} object.
#'
#' @details
#' Use this function when neighbourhood should be defined by a fixed number of
#' nearby planning units rather than by polygon topology or a fixed distance
#' threshold.
#'
#' Let \eqn{s_i = (x_i, y_i)} denote the coordinates of planning unit
#' \eqn{i}. For each planning unit, this function identifies the \code{k}
#' nearest distinct planning units under Euclidean distance.
#'
#' If \eqn{d_{ij}} denotes the Euclidean distance between units \eqn{i} and
#' \eqn{j}, then the k-nearest-neighbours relation is constructed by adding an
#' edge from \eqn{i} to each of its \code{k} nearest neighbours.
#'
#' Edge weights are then assigned according to \code{weight_mode}:
#' \itemize{
#'   \item \code{"constant"}:
#'   \deqn{\omega_{ij} = 1,}
#'   \item \code{"inverse"}:
#'   \deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)},}
#'   \item \code{"inverse_sq"}:
#'   \deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)^2},}
#' }
#' where \eqn{\varepsilon} = \code{distance_eps} is a small constant to avoid
#' division by zero.
#'
#' The raw k-nearest-neighbours structure is directional by construction, but
#' the stored relation is registered as undirected by default through
#' \code{\link{add_spatial_relations}}, which collapses duplicate unordered
#' pairs.
#'
#' If the \pkg{RANN} package is available, it is used for efficient nearest
#' neighbour search. Otherwise, a full distance matrix is computed.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}.
#' @param coords Optional coordinates specification. This may be:
#'   \itemize{
#'     \item a \code{data.frame(id, x, y)}, or
#'     \item a numeric matrix with two columns \code{(x, y)} aligned to the
#'     order of planning units.
#'   }
#'   If \code{NULL}, coordinates are taken from \code{x$data$pu_coords} or from
#'   columns \code{x$data$pu$x} and \code{x$data$pu$y}.
#' @param k Integer giving the number of neighbours per planning unit. Must be
#'   at least 1 and strictly less than the number of planning units.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param weight_mode Character string indicating how distance is converted to
#'   weight. Must be one of \code{"constant"}, \code{"inverse"}, or
#'   \code{"inverse_sq"}.
#' @param distance_eps Small positive numeric constant used to avoid division by
#'   zero in inverse-distance weighting.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4),
#'   x = c(0, 1, 0, 1),
#'   y = c(0, 0, 1, 1)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' p <- add_spatial_knn(
#'   x = p,
#'   k = 2,
#'   name = "knn2",
#'   weight_mode = "constant"
#' )
#'
#' p$data$spatial_relations$knn2
#'
#' @seealso
#' \code{\link{add_spatial_distance}},
#' \code{\link{add_spatial_relations}}
#'
#' @export
add_spatial_knn <- function(x,
                            coords = NULL,
                            k = 8,
                            name = "knn",
                            weight_mode = c("constant", "inverse", "inverse_sq"),
                            distance_eps = 1e-9) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)
  weight_mode <- match.arg(weight_mode)
  k <- as.integer(k)
  if (k < 1) stop("k must be >= 1.", call. = FALSE)

  C <- .pa_coords_from_input(x, coords = coords)
  C <- C[match(x$data$pu$id, C$id), , drop = FALSE]
  if (anyNA(C$id)) stop("coords does not cover all pu ids.", call. = FALSE)

  X <- as.matrix(C[, c("x", "y")])
  n <- nrow(X)
  if (k >= n) stop("k must be < number of PUs.", call. = FALSE)

  if (requireNamespace("RANN", quietly = TRUE)) {
    nn <- RANN::nn2(X, k = k + 1)
    idx_mat <- nn$nn.idx[, -1, drop = FALSE]
    dist_mat <- nn$nn.dists[, -1, drop = FALSE]
  } else {
    D <- as.matrix(stats::dist(X))
    diag(D) <- Inf
    idx_mat <- t(apply(D, 1, function(d) order(d)[seq_len(k)]))
    dist_mat <- matrix(D[cbind(rep(seq_len(n), each = k), as.vector(t(idx_mat)))], nrow = n, byrow = TRUE)
  }

  pu1 <- rep(C$id, each = k)
  pu2 <- C$id[as.vector(t(idx_mat))]
  dist <- as.vector(t(dist_mat))

  w <- switch(
    weight_mode,
    constant = rep(1, length(dist)),
    inverse = 1 / pmax(dist, distance_eps),
    inverse_sq = 1 / pmax(dist, distance_eps)^2
  )

  rel <- data.frame(
    pu1 = as.integer(pu1),
    pu2 = as.integer(pu2),
    weight = as.numeric(w),
    distance = as.numeric(dist),
    source = paste0("knn_", weight_mode),
    stringsAsFactors = FALSE
  )

  add_spatial_relations(x, rel, name = name, directed = FALSE, allow_self = FALSE)
}

#' @title Add distance-threshold spatial relations
#'
#' @description
#' Build and register a spatial relation connecting planning units whose
#' Euclidean distance is less than or equal to a user-defined threshold.
#'
#' This constructor does not require polygon geometry and instead uses
#' planning-unit coordinates.
#'
#' @details
#' Use this function when neighbourhood should be defined by a fixed distance
#' radius rather than by polygon topology or a fixed number of neighbours.
#'
#' Let \eqn{s_i = (x_i, y_i)} denote the coordinates of planning unit
#' \eqn{i}. Let \eqn{d_{ij}} be the Euclidean distance between planning units
#' \eqn{i} and \eqn{j}.
#'
#' For a user-supplied threshold \eqn{d_{\max}}, this constructor creates an
#' edge between \eqn{i} and \eqn{j} whenever:
#' \deqn{
#' d_{ij} \le d_{\max}.
#' }
#'
#' Edge weights are assigned according to \code{weight_mode}:
#' \itemize{
#'   \item \code{"constant"}:
#'   \deqn{\omega_{ij} = 1,}
#'   \item \code{"inverse"}:
#'   \deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)},}
#'   \item \code{"inverse_sq"}:
#'   \deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)^2},}
#' }
#' where \eqn{\varepsilon} = \code{distance_eps} is a small constant.
#'
#' The implementation computes an \eqn{O(n^2)} distance matrix and is therefore
#' best suited to small or moderate numbers of planning units. For large
#' problems, \code{\link{add_spatial_knn}} is often more scalable.
#'
#' The resulting relation is registered as undirected.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}.
#' @param coords Optional coordinates specification, following the same rules as
#'   in \code{\link{add_spatial_knn}}.
#' @param max_distance Positive numeric scalar giving the maximum distance for
#'   an edge.
#' @param name Character string giving the key under which the relation is
#'   stored.
#' @param weight_mode Character string indicating how distance is converted to
#'   weight. Must be one of \code{"constant"}, \code{"inverse"}, or
#'   \code{"inverse_sq"}.
#' @param distance_eps Small positive numeric constant used to avoid division by
#'   zero in inverse-distance weighting.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4),
#'   x = c(0, 1, 0, 1),
#'   y = c(0, 0, 1, 1)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' p <- add_spatial_distance(
#'   x = p,
#'   max_distance = 1.01,
#'   name = "within_1",
#'   weight_mode = "constant"
#' )
#'
#' p$data$spatial_relations$within_1
#'
#' @seealso
#' \code{\link{add_spatial_knn}},
#' \code{\link{add_spatial_relations}}
#'
#' @export
add_spatial_distance <- function(x,
                                 coords = NULL,
                                 max_distance,
                                 name = "distance",
                                 weight_mode = c("constant", "inverse", "inverse_sq"),
                                 distance_eps = 1e-9) {

  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  x <- .pa_ensure_pu_index(x)
  weight_mode <- match.arg(weight_mode)
  max_distance <- as.numeric(max_distance)
  if (!is.finite(max_distance) || max_distance <= 0) {
    stop("max_distance must be a positive finite number.", call. = FALSE)
  }

  C <- .pa_coords_from_input(x, coords = coords)
  C <- C[match(x$data$pu$id, C$id), , drop = FALSE]
  if (anyNA(C$id)) stop("coords does not cover all pu ids.", call. = FALSE)

  X <- as.matrix(C[, c("x", "y")])
  n <- nrow(X)

  D <- as.matrix(stats::dist(X))
  diag(D) <- Inf
  which_edges <- which(D <= max_distance, arr.ind = TRUE)

  if (nrow(which_edges) == 0) stop("No edges found under max_distance. Try a larger threshold.", call. = FALSE)

  pu1 <- C$id[which_edges[, 1]]
  pu2 <- C$id[which_edges[, 2]]
  dist <- D[which_edges]

  w <- switch(
    weight_mode,
    constant = rep(1, length(dist)),
    inverse = 1 / pmax(dist, distance_eps),
    inverse_sq = 1 / pmax(dist, distance_eps)^2
  )

  rel <- data.frame(
    pu1 = as.integer(pu1),
    pu2 = as.integer(pu2),
    weight = as.numeric(w),
    distance = as.numeric(dist),
    source = paste0("distance_", weight_mode),
    stringsAsFactors = FALSE
  )

  add_spatial_relations(x, rel, name = name, directed = FALSE, allow_self = FALSE)
}

test_that("add_spatial_boundary stores non-diagonal relations correctly", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  p <- multiscape::add_spatial_boundary(
    p,
    boundary = toy$boundary,
    include_self = FALSE
  )

  expect_true("spatial_relations" %in% names(p$data))
  expect_true("boundary" %in% names(p$data$spatial_relations))

  bnd <- p$data$spatial_relations$boundary
  expect_true(is.data.frame(bnd))
  expect_true(all(c("pu1", "pu2") %in% names(bnd)))

  expect_false(any(bnd$pu1 == bnd$pu2))
})

test_that("add_spatial_boundary with include_self = TRUE adds diagonal rows", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  p <- multiscape::add_spatial_boundary(
    p,
    boundary = toy$boundary,
    include_self = TRUE
  )

  bnd <- p$data$spatial_relations$boundary
  expect_true(any(bnd$pu1 == bnd$pu2))
})

#' @include internal.R
#'
#' @export
if (!methods::isClass("SolutionSet")) methods::setOldClass("SolutionSet")
NULL

#' @name solutionset-class
#' @aliases SolutionSet
#' @title SolutionSet class
#'
#' @description
#' The \code{SolutionSet} class stores the result of solving a
#' \code{\link{Problem}} object when multiple runs are produced.
#'
#' A \code{SolutionSet} object is the multi-run counterpart of
#' \code{\link{solution-class}}. It contains the original problem, run-level
#' design and outcome tables, the list of individual \code{Solution} objects,
#' diagnostics summarizing the full set of runs, method metadata, and additional
#' metadata.
#'
#' Objects of this class are typically created by \code{\link{solve}} in
#' workflows that generate more than one optimization run, such as weighted-sum
#' scans, \eqn{\epsilon}-constraint designs, or AUGMECON grids.
#'
#' @details
#' \strong{Conceptual role}
#'
#' The \code{SolutionSet} class represents the result of a multi-run solving
#' workflow:
#' \preformatted{
#' Problem -> solve() -> SolutionSet
#' }
#'
#' Each run corresponds to one specific optimization subproblem, parameter
#' setting, or trade-off configuration. The \code{SolutionSet} object keeps all
#' these runs together in a structured form.
#'
#' Thus, a \code{SolutionSet} is not a single solution with multiple labels, but
#' a collection of distinct \code{Solution} objects linked to a shared problem
#' and a shared multi-run design.
#'
#' \strong{Typical use cases}
#'
#' A \code{SolutionSet} is typically returned when the chosen solution method
#' generates multiple runs, for example:
#' \itemize{
#'   \item several weighted-sum configurations,
#'   \item a grid of \eqn{\epsilon}-constraint runs,
#'   \item an AUGMECON exploration of multiple
#'   \eqn{\epsilon}-combinations,
#'   \item or any other workflow producing more than one optimizer call.
#' }
#'
#' \strong{Internal structure}
#'
#' A \code{SolutionSet} object separates information into several layers:
#'
#' \describe{
#'   \item{\code{problem}}{The original \code{Problem} object shared by all
#'   runs.}
#'
#'   \item{\code{solution}}{A named \code{list} containing the core multi-run
#'   optimization outputs. This typically includes the run design, the run
#'   summary table, and the list of individual \code{Solution} objects.}
#'
#'   \item{\code{summary}}{A named \code{list} containing user-facing summaries
#'   aggregated across runs when such summaries are available.}
#'
#'   \item{\code{diagnostics}}{A named \code{list} containing diagnostics about
#'   the solution set as a whole.}
#'
#'   \item{\code{method}}{A named \code{list} describing the multi-run
#'   optimization method used.}
#'
#'   \item{\code{meta}}{A named \code{list} containing additional metadata.}
#'
#'   \item{\code{name}}{A \code{character(1)} identifier for the solution set.}
#' }
#'
#' \strong{Run-level content}
#'
#' The \code{solution} field is the main entry point for run-level information.
#' Typical entries include:
#' \describe{
#'   \item{\code{design}}{A \code{data.frame} describing the experimental or
#'   optimization design, for example weights or \eqn{\epsilon}-levels.}
#'
#'   \item{\code{runs}}{A \code{data.frame} summarizing the outcome of each run.
#'   This typically includes run identifiers, solver status, runtime, gap, and
#'   objective values.}
#'
#'   \item{\code{solutions}}{A list of individual \code{Solution} objects, one
#'   per run.}
#' }
#'
#' In many workflows, the \code{runs} table is the most important compact
#' representation of the solution set, while \code{solutions[[i]]} provides the
#' full detailed output for run \eqn{i}.
#'
#' \strong{Objective values and run tables}
#'
#' In multi-objective workflows, the run table often stores objective values in
#' columns named \code{value_<alias>}, where \code{<alias>} is the alias of a
#' registered objective. For example:
#' \itemize{
#'   \item \code{value_cost},
#'   \item \code{value_frag},
#'   \item \code{value_benefit}.
#' }
#'
#' This naming convention is used by downstream functions such as
#' \code{\link{plot_tradeoff}}.
#'
#' Depending on the solving method, the run table may also contain design
#' columns such as:
#' \itemize{
#'   \item \code{weight_<alias>} for weighted-sum runs,
#'   \item \code{eps_<alias>} for \eqn{\epsilon}-constraint or AUGMECON runs.
#' }
#'
#' \strong{Relationship with \code{Solution}}
#'
#' A \code{SolutionSet} is conceptually a collection of \code{Solution} objects.
#' The individual runs are typically stored in:
#' \deqn{
#' \code{x$solution$solutions[[i]]}
#' }
#'
#' where each element is itself a full \code{Solution} object.
#'
#' Therefore:
#' \itemize{
#'   \item use \code{SolutionSet} when working with the full set of runs,
#'   \item use an individual \code{Solution} when inspecting one particular run
#'   in detail.
#' }
#'
#' \strong{Diagnostics}
#'
#' The \code{diagnostics} field stores metadata about the multi-run solve
#' process. Depending on the implementation, it may summarize:
#' \itemize{
#'   \item number of runs,
#'   \item status frequencies,
#'   \item runtime ranges,
#'   \item gap ranges,
#'   \item and other aggregate information about the set of runs.
#' }
#'
#' \strong{Printing}
#'
#' The \code{print()} method provides a concise summary of the full solution set.
#' It reports:
#' \itemize{
#'   \item the optimization method name,
#'   \item the participating objective aliases,
#'   \item the number of design rows, runs, and stored solutions,
#'   \item run-level status summaries,
#'   \item runtime and gap ranges,
#'   \item and the names of design and objective-value columns when available.
#' }
#'
#' This printed output is intended as a quick overview. Detailed inspection
#' should use:
#' \itemize{
#'   \item \code{x$solution$runs},
#'   \item \code{x$solution$design},
#'   \item \code{x$solution$solutions[[i]]},
#'   \item or the accessor methods documented below.
#' }
#'
#' @section Fields:
#' \describe{
#'   \item{\code{problem}}{The \code{Problem} object used to generate the full
#'   solution set.}
#'   \item{\code{solution}}{A named \code{list} containing the core multi-run
#'   outputs, typically including \code{design}, \code{runs}, and
#'   \code{solutions}.}
#'   \item{\code{summary}}{A named \code{list} containing user-facing summaries
#'   associated with the solution set.}
#'   \item{\code{diagnostics}}{A named \code{list} containing diagnostics about
#'   the solution set as a whole.}
#'   \item{\code{method}}{A named \code{list} describing the optimization method
#'   used.}
#'   \item{\code{meta}}{A named \code{list} of additional metadata.}
#'   \item{\code{name}}{A \code{character(1)} name for the solution set
#'   object.}
#' }
#'
#' @section Methods:
#' \describe{
#'   \item{\code{print()}}{Print a concise summary of the solution set,
#'   including method name, number of runs, and run-level diagnostics.}
#'
#'   \item{\code{show()}}{Alias of \code{print()}.}
#'
#'   \item{\code{repr()}}{Return a short one-line representation of the
#'   solution set.}
#'
#'   \item{\code{getMethod()}}{Return the method specification stored in
#'   \code{self$method}.}
#'
#'   \item{\code{getDesign()}}{Return the design table stored in
#'   \code{self$solution$design}.}
#'
#'   \item{\code{getRuns()}}{Return the run summary table stored in
#'   \code{self$solution$runs}.}
#'
#'   \item{\code{getSolutions()}}{Return the list of individual
#'   \code{Solution} objects stored in \code{self$solution$solutions}.}
#' }
#'
#' @return No return value. This page documents the \code{SolutionSet} class.
#'
#' @seealso
#' \code{\link{solution-class}},
#' \code{\link{plot_tradeoff}},
#' \code{\link{get_pu}},
#' \code{\link{get_actions}},
#' \code{\link{get_features}},
#' \code{\link{get_targets}},
#' \code{\link{solve}}
NULL

.pa_solutionset_method_label <- function(method) {
  if (is.null(method) || !is.list(method)) return("unknown")
  as.character(method$name %||% method$type %||% "unknown")[1]
}

.pa_solutionset_aliases <- function(method) {
  if (is.null(method) || !is.list(method)) return(character(0))
  aliases <- as.character(method$aliases %||% character(0))
  aliases[!is.na(aliases) & nzchar(aliases)]
}

.pa_solutionset_preview_runs <- function(runs, max_show = 6L) {
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L) {
    return(NULL)
  }

  val_cols <- grep("^value_", names(runs), value = TRUE)

  if (length(val_cols) == 0L) {
    return(NULL)
  }

  rr <- runs[1, , drop = FALSE]

  txt <- paste0(
    sub("^value_", "", val_cols),
    ": ",
    vapply(rr[val_cols], .pa_num_text, character(1)),
    collapse = ", "
  )

  txt
}

.pa_solutionset_status_summary <- function(runs) {
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L || !("status" %in% names(runs))) {
    return("none")
  }

  st <- as.character(runs$status)
  st <- st[!is.na(st) & nzchar(st)]

  if (length(st) == 0L) return("none")

  tb <- table(st)
  paste0(names(tb), ": ", as.integer(tb), collapse = ", ")
}

.pa_solutionset_range_text <- function(x, digits = 3) {
  if (is.null(x)) return("none")
  x <- as.numeric(x)
  x <- x[is.finite(x) & !is.na(x)]
  if (length(x) == 0L) return("none")
  if (length(x) == 1L) return(as.character(round(x, digits)))
  paste0(round(min(x), digits), "..", round(max(x), digits))
}

.pa_solutionset_run_columns <- function(runs) {
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L) {
    return(list(design = character(0), values = character(0), core = character(0)))
  }

  nm <- names(runs)

  core_cols <- intersect(c("run_id", "status", "runtime", "gap"), nm)
  design_cols <- grep("^(weight_|eps_)", nm, value = TRUE)
  value_cols <- grep("^value_", nm, value = TRUE)

  list(
    core = core_cols,
    design = design_cols,
    values = value_cols
  )
}

.pa_solutionset_preview_table <- function(runs, max_rows = 6L, digits = 3) {
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L) {
    return(NULL)
  }

  cols <- .pa_solutionset_run_columns(runs)
  keep <- c(cols$core, cols$design, cols$values)
  keep <- unique(keep[keep %in% names(runs)])

  if (length(keep) == 0L) return(NULL)

  out <- runs[seq_len(min(nrow(runs), max_rows)), keep, drop = FALSE]

  for (nm in names(out)) {
    if (is.numeric(out[[nm]])) {
      out[[nm]] <- round(out[[nm]], digits)
    }
  }

  out
}

.pa_solutionset_best_run <- function(runs, method = NULL) {
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L) {
    return(NULL)
  }

  ok <- rep(TRUE, nrow(runs))
  if ("status" %in% names(runs)) {
    ok <- runs$status %in% c("optimal", "time_limit_feasible", "solution_limit")
    ok[is.na(ok)] <- FALSE
  }

  rr <- runs[ok, , drop = FALSE]
  if (nrow(rr) == 0L) return(NULL)

  mname <- .pa_solutionset_method_label(method)

  if (identical(mname, "weighted")) {
    val_cols <- grep("^value_", names(rr), value = TRUE)
    if (length(val_cols) > 0L) {
      idx <- 1L
      out <- rr[idx, , drop = FALSE]
      return(out)
    }
  }

  rr[1, , drop = FALSE]
}

#' @export
SolutionSet <- pproto(
  "SolutionSet",
  problem = NULL,
  solution = list(),
  summary = list(),
  diagnostics = list(),
  method = list(),
  meta = list(),
  name = "solset",

  print = function(self) {
    ch <- .pa_cli_box_chars()
    div_id <- cli::cli_div(theme = .pa_cli_theme())

    method_name <- .pa_solutionset_method_label(self$method)
    aliases <- .pa_solutionset_aliases(self$method)

    design <- self$solution$design %||% NULL
    runs <- self$solution$runs %||% NULL
    sols <- self$solution$solutions %||% list()

    n_design <- if (inherits(design, "data.frame")) nrow(design) else 0L
    n_runs <- if (inherits(runs, "data.frame")) nrow(runs) else 0L
    n_solutions <- length(sols)

    alias_txt <- if (length(aliases) == 0L) {
      "none"
    } else {
      .pa_preview_text(aliases, max_show = 6L, quote = FALSE)
    }

    status_txt <- .pa_solutionset_status_summary(runs)

    runtime_txt <- if (!is.null(runs) && "runtime" %in% names(runs)) {
      .pa_solutionset_range_text(runs$runtime, digits = 3)
    } else {
      "none"
    }

    gap_txt <- if (!is.null(runs) && "gap" %in% names(runs)) {
      .pa_solutionset_range_text(runs$gap, digits = 4)
    } else {
      "none"
    }

    run_cols <- .pa_solutionset_run_columns(runs)
    design_txt <- if (length(run_cols$design) == 0L) "none" else paste(run_cols$design, collapse = ", ")
    value_txt  <- if (length(run_cols$values) == 0L) "none" else paste(run_cols$values, collapse = ", ")

    cli::cli_text("A multiscape solution set ({.cls SolutionSet})")

    # ---- METHOD
    cli::cli_text("{ch$j}{ch$b}{.h method}", .envir = environment())
    cli::cli_text(
      "{ch$v}{ch$j}{ch$b}name:            {.code {method_name}}",
      .envir = environment()
    )
    cli::cli_text(
      "{ch$v}{ch$l}{ch$b}objectives:      {length(aliases)} ({alias_txt})",
      .envir = environment()
    )

    # ---- CONTENT
    cli::cli_text("{ch$l}{ch$b}{.h content}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}design rows:     {n_design}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}runs:            {n_runs}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$l}{ch$b}solutions:       {n_solutions}", .envir = environment())

    # ---- RUN SUMMARY
    cli::cli_text("{ch$l}{ch$b}{.h run summary}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}statuses:        {status_txt}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}runtime:         {runtime_txt}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}gap:             {gap_txt}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$j}{ch$b}design cols:     {design_txt}", .envir = environment())
    cli::cli_text(" {ch$v}{ch$l}{ch$b}value cols:      {value_txt}", .envir = environment())

    info_sym <- cli::symbol$info
    if (is.function(info_sym)) info_sym <- info_sym()
    cli::cli_text(
      cli::col_grey(
        paste0(
          "# ", info_sym,
          " Use {.code x$solution$runs}, {.code x$solution$design}, ",
          "and {.code x$solution$solutions[[i]]} to inspect details."
        )
      )
    )

    cli::cli_end(div_id)
    invisible(TRUE)
  },

  show = function(self) self$print(),

  repr = function(self) {
    method_name <- .pa_solutionset_method_label(self$method)
    runs <- self$solution$runs %||% NULL
    n_runs <- if (inherits(runs, "data.frame")) nrow(runs) else 0L
    paste0("<SolutionSet> method=", method_name, ", runs=", n_runs)
  },

  getMethod = function(self) self$method,
  getDesign = function(self) self$solution$design %||% NULL,
  getRuns = function(self) self$solution$runs %||% NULL,
  getSolutions = function(self) self$solution$solutions %||% list()
)

#' @include internal.R
#'
#' @title Limit the number of actions per planning unit (maximum)
#' @name add_action_max_per_pu
#'
#' @description
#' Store a constraint that limits the number of actions that can be selected within each
#' planning unit (PU). The constraint has the form:
#' \deqn{\sum_{a \in A} x_{pu,a} \le \mathrm{max}}
#' for each PU in the selected set.
#'
#' This function is \strong{data-only}: it records the constraint specification in the
#' \code{Problem} object but does not build or modify the optimization model. The constraint
#' is later translated into linear constraints by the model builder (e.g.,
#' \code{.pa_build_model_apply_constraints()}).
#'
#' @details
#' The constraint can be applied to:
#' \itemize{
#'   \item all planning units and all actions (default), or
#'   \item a subset of planning units via \code{pu}, and/or
#'   \item a subset of actions via \code{actions}.
#' }
#'
#' Note that the subset is applied over the existing feasible \code{(pu, action)} pairs
#' in \code{x$data$dist_actions}. If the requested \code{pu}/\code{actions} subset yields
#' zero feasible pairs, the function stops with an informative error.
#'
#' If \code{x$data$model_ptr} is already present (a model has been built previously),
#' the function marks the model as dirty by setting \code{x$data$meta$model_dirty <- TRUE},
#' signalling that the model should be rebuilt before solving.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}. Must already contain actions (i.e., run
#'   \code{\link{add_actions}} first).
#' @param max Integer scalar \eqn{\ge 0}. Maximum number of actions allowed per PU.
#'   Default is \code{1L}.
#' @param pu Optional integer vector of planning unit ids (external ids, i.e.,
#'   \code{x$data$pu$id}) to which the constraint is applied. If \code{NULL} (default),
#'   the constraint applies to all PUs.
#' @param actions Optional character vector of action ids (i.e., \code{x$data$actions$id})
#'   to include in the sum. If \code{NULL} (default), all actions are included.
#' @param overwrite Logical. If \code{TRUE}, replace any existing \code{action_max_per_pu}
#'   constraint stored in \code{x$data$constraints}. If \code{FALSE} (default) and a constraint
#'   already exists, an error is raised.
#'
#' @return The updated \code{Problem} object with \code{x$data$constraints$action_max_per_pu}
#'   set to a list containing the constraint specification.
#'
#' @seealso \code{\link{add_actions}}
#' @keywords internal
#'
NULL

add_action_max_per_pu <- function(
    x,
    max = 1L,
    pu = NULL,
    actions = NULL,
    overwrite = FALSE
) {
  # ---- checks: x
  assertthat::assert_that(!is.null(x), msg = "x is NULL")
  assertthat::assert_that(!is.null(x$data), msg = "x does not look like a multiscape Problem object")
  assertthat::assert_that(!is.null(x$data$pu), msg = "x$data$pu is missing. Run create_problem() first.")
  assertthat::assert_that(!is.null(x$data$dist_actions), msg = "No actions found. Run add_actions() first.")
  assertthat::assert_that(!is.null(x$data$actions), msg = "No action catalog found. Run add_actions() first.")

  # ---- validate max
  if (length(max) != 1 || is.na(max)) stop("'max' must be a single non-NA integer.", call. = FALSE)
  max <- as.integer(max)
  if (max < 0L) stop("'max' must be >= 0.", call. = FALSE)

  pu_df   <- x$data$pu
  da      <- x$data$dist_actions
  acts_df <- x$data$actions

  # required cols
  assertthat::assert_that("id" %in% names(pu_df), msg = "x$data$pu must have column 'id'.")
  assertthat::assert_that(all(c("pu","action") %in% names(da)),
                          msg = "x$data$dist_actions must have columns 'pu' and 'action'.")
  assertthat::assert_that("id" %in% names(acts_df), msg = "x$data$actions must have column 'id'.")

  # normalize types
  x <- .pa_clone_data(x)
  pu_ids_all <- as.integer(pu_df$id)
  act_ids_all <- as.character(acts_df$id)

  # ---- subset of PUs
  if (!is.null(pu)) {
    pu <- as.integer(pu)
    if (anyNA(pu)) stop("'pu' contains NA after coercion.", call. = FALSE)
    bad <- setdiff(unique(pu), pu_ids_all)
    if (length(bad) > 0) stop("Unknown PU id(s) in 'pu': ", paste(bad, collapse = ", "), call. = FALSE)
    pu_ids_use <- unique(pu)
  } else {
    pu_ids_use <- pu_ids_all
  }

  # ---- subset of actions
  if (!is.null(actions)) {
    actions <- as.character(actions)
    if (anyNA(actions)) stop("'actions' contains NA.", call. = FALSE)
    bad <- setdiff(unique(actions), act_ids_all)
    if (length(bad) > 0) stop("Unknown action id(s) in 'actions': ", paste(bad, collapse = ", "), call. = FALSE)
    action_ids_use <- unique(actions)
  } else {
    action_ids_use <- act_ids_all
  }

  # ---- sanity: does it select at least one feasible (pu,action) row?
  da2 <- da
  da2$pu <- as.integer(da2$pu)
  da2$action <- as.character(da2$action)

  da2 <- da2[da2$pu %in% pu_ids_use & da2$action %in% action_ids_use, , drop = FALSE]
  if (nrow(da2) == 0) {
    stop(
      "The (pu, actions) subset produces zero feasible (pu, action) pairs.\n",
      "Check that 'pu' and 'actions' match what exists in dist_actions.",
      call. = FALSE
    )
  }

  # ---- store constraint spec (data-only)
  spec <- list(
    type = "action_max_per_pu",
    max = max,
    pu = pu_ids_use,
    actions = action_ids_use
  )

  if (is.null(x$data$constraints) || !is.list(x$data$constraints)) x$data$constraints <- list()

  if (!isTRUE(overwrite) && !is.null(x$data$constraints$action_max_per_pu)) {
    stop(
      "An action_max_per_pu constraint already exists. Use overwrite=TRUE to replace it.",
      call. = FALSE
    )
  }

  x$data$constraints$action_max_per_pu <- spec

  # mark model dirty if already built
  if (!is.null(x$data$model_ptr)) {
    if (is.null(x$data$meta) || !is.list(x$data$meta)) x$data$meta <- list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

skip_if_no_prioritizr <- function() {
  testthat::skip_if_not_installed("prioritizr")
}

build_prioritizr_basic <- function(toy, target = 0.5, boundary_matrix = NULL) {
  skip_if_no_prioritizr()

  pu <- toy$pu
  feat_wide <- reshape(
    toy$dist_features,
    idvar = "pu",
    timevar = "feature",
    direction = "wide"
  )

  dat <- merge(pu, feat_wide, by.x = "id", by.y = "pu", all.x = TRUE, sort = FALSE)
  dat[is.na(dat)] <- 0

  names(dat) <- sub("^amount\\.", "feature_", names(dat))

  p <- prioritizr::problem(
    dat,
    features = paste0("feature_", toy$features$id),
    cost_column = "cost"
  ) |>
    prioritizr::add_min_set_objective() |>
    prioritizr::add_relative_targets(target)

  if (!is.null(boundary_matrix)) {
    p <- p |>
      prioritizr::add_boundary_penalties(
        penalty = 1,
        data = boundary_matrix
      )
  }

  p
}

solve_prioritizr <- function(x, ...) {
  get("solve", envir = asNamespace("prioritizr"))(x, ...)
}

test_that("multiscape matches prioritizr on a one-action min-cost problem", {
  skip_if_no_cbc()
  skip_if_no_prioritizr()

  toy <- toy_equivalent_basic()

  p_multiscape <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s_multiscape <- multiscape::solve(p_multiscape)

  acts_multiscape <- multiscape::get_actions(s_multiscape, only_selected = TRUE)
  sel_multiscape <- sort(unique(acts_multiscape$pu))
  cost_multiscape <- sum(toy$pu$cost[toy$pu$id %in% sel_multiscape])

  p_prio <- build_prioritizr_basic(toy, target = 0.5)

  s_prio <- solve_prioritizr(p_prio)

  sel_prio <- sort(which(s_prio$solution_1 > 0.5))
  cost_prio <- sum(toy$pu$cost[toy$pu$id %in% sel_prio])

  expect_equal(cost_multiscape, cost_prio)
  expect_equal(sel_multiscape, sel_prio)
})

#' @include internal.R
NULL

#' @title Add locked action decisions to a planning problem
#'
#' @description
#' Fix feasible planning unit--action decisions to be selected or excluded.
#'
#' This function modifies the status of existing feasible
#' \code{(pu, action)} pairs stored in the feasible action table. It does not
#' create new feasible action pairs and therefore must be used only after
#' \code{\link{add_actions}} has been called.
#'
#' @details
#' Use this function when only specific feasible \code{(pu, action)} decisions
#' must be forced in or out of the solution, rather than whole planning units.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and
#' \eqn{\mathcal{A}} the set of actions. Let
#' \eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
#' feasible planning unit--action pairs already defined in the problem.
#'
#' This function allows the user to define two subsets:
#' \itemize{
#'   \item \eqn{\mathcal{D}^{in} \subseteq \mathcal{D}}, the set of feasible
#'   pairs that must be selected,
#'   \item \eqn{\mathcal{D}^{out} \subseteq \mathcal{D}}, the set of feasible
#'   pairs that must not be selected.
#' }
#'
#' These sets are encoded by updating the \code{status} column of the feasible
#' action table. The function validates that all requested locked-in and
#' locked-out pairs are already feasible. Therefore, it cannot be used to
#' introduce new planning unit--action combinations into the problem.
#'
#' In optimization terms, if \eqn{x_{ia}} denotes the decision variable
#' associated with planning unit \eqn{i} and action \eqn{a}, then:
#' \itemize{
#'   \item locked-in pairs conceptually impose \eqn{x_{ia} = 1},
#'   \item locked-out pairs conceptually impose \eqn{x_{ia} = 0}.
#' }
#'
#' The exact translation into solver-side constraints occurs later when the
#' model is built.
#'
#' In contrast, \code{\link{add_constraint_locked_pu}} fixes whole planning
#' units through the unit-selection variables, whereas this function fixes only
#' specific feasible \code{(pu, action)} decisions.
#'
#' \strong{Accepted formats}
#'
#' Both \code{locked_in} and \code{locked_out} accept the same formats:
#' \itemize{
#'   \item \code{NULL},
#'   \item a \code{data.frame} with columns \code{pu} and \code{action},
#'   optionally including a \code{feasible} column used as a filter,
#'   \item a named list whose names are action ids and whose elements are either
#'   vectors of planning unit ids or \code{sf} objects.
#' }
#'
#' If a \code{feasible} column is supplied in a \code{data.frame}, only rows
#' with \code{feasible = TRUE} are used. Missing values in \code{feasible} are
#' treated as \code{FALSE}.
#'
#' If an \code{sf} specification is supplied, the problem object must contain
#' planning-unit geometry, and planning units are matched spatially using
#' \code{sf::st_intersects()}.
#'
#' \strong{Conflict checking}
#'
#' A given \code{(pu, action)} pair cannot be simultaneously requested in both
#' \code{locked_in} and \code{locked_out}. Such overlaps are rejected.
#'
#' In addition, if a planning unit is already marked as locked out at the
#' planning-unit level, then all feasible actions in that planning unit are
#' forced to \code{status = 3}. Any attempt to lock in an action within such a
#' planning unit raises an error.
#'
#' \strong{Order of precedence}
#'
#' User-supplied locked-in and locked-out action requests are first applied to
#' the feasible action table. Afterwards, any planning-unit-level
#' \code{locked_out} flag is enforced, overriding action-level status and
#' ensuring consistency with planning-unit exclusions.
#'
#' @param x A \code{Problem} object with action feasibility already defined via
#'   \code{\link{add_actions}}.
#'
#' @param locked_in Optional specification of feasible \code{(pu, action)} pairs
#'   that must be selected. It may be \code{NULL}, a \code{data.frame}, or a
#'   named list.
#'
#' @param locked_out Optional specification of feasible \code{(pu, action)}
#'   pairs that must not be selected. It may be \code{NULL}, a
#'   \code{data.frame}, or a named list.
#'
#' @return An updated \code{Problem} object in which the status column of the
#'   feasible action table has been modified to reflect locked-in and locked-out
#'   decisions.
#'
#' @examples
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(2, 3, 1, 4)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   x = p,
#'   actions = data.frame(id = c("conservation", "restoration")),
#'   cost = c(conservation = 3, restoration = 8)
#' )
#'
#' # Lock a few feasible decisions
#' p <- add_constraint_locked_actions(
#'   x = p,
#'   locked_in = data.frame(
#'     pu = c(1, 2),
#'     action = c("conservation", "restoration")
#'   ),
#'   locked_out = data.frame(
#'     pu = c(4),
#'     action = c("conservation")
#'   )
#' )
#'
#' p$data$dist_actions
#'
#' # Named-list interface
#' p2 <- add_constraint_locked_actions(
#'   x = p,
#'   locked_in = list(
#'     conservation = c(1, 3)
#'   ),
#'   locked_out = list(
#'     restoration = c(2)
#'   )
#' )
#'
#' p2$data$dist_actions
#'
#' @seealso
#' \code{\link{add_actions}},
#' \code{\link{add_constraint_locked_pu}}
#'
#' @export
add_constraint_locked_actions <- function(
    x,
    locked_in = NULL,
    locked_out = NULL
) {

  .as_int_id <- function(v, what) {
    if (is.factor(v)) v <- as.character(v)
    if (is.character(v)) {
      if (any(grepl("[^0-9\\-]", v))) {
        stop(what, " must be numeric/integer ids (got non-numeric strings).", call. = FALSE)
      }
      v <- as.integer(v)
    } else {
      v <- as.integer(v)
    }
    if (anyNA(v)) stop(what, " contains NA after coercion to integer.", call. = FALSE)
    v
  }

  .normalize_feasible_col <- function(df, what) {
    if (!("feasible" %in% names(df))) {
      df$feasible <- TRUE
      return(df)
    }

    f <- df$feasible

    if (is.logical(f)) {
      # keep
    } else if (is.numeric(f) || is.integer(f)) {
      f <- f != 0
    } else if (is.factor(f)) {
      f <- as.character(f)
    }

    if (is.character(f)) {
      w <- tolower(trimws(f))
      f <- w %in% c("true", "t", "1", "yes", "y")
    } else {
      f <- as.logical(f)
    }

    # if (na_is_infeasible)
    f[is.na(f)] <- FALSE
    df$feasible <- as.logical(f)
    df
  }

  .spec_to_pairs <- function(spec, what, action_ids, pu_ids, pu_sf, as_int_id_fun) {
    if (is.null(spec)) return(NULL)

    if (inherits(spec, "data.frame")) {
      assertthat::assert_that(nrow(spec) > 0, msg = paste0(what, " is an empty data.frame."))

      if ("id" %in% names(spec) && !("action" %in% names(spec))) {
        names(spec)[names(spec) == "id"] <- "action"
      }
      assertthat::assert_that(
        assertthat::has_name(spec, "pu"),
        assertthat::has_name(spec, "action"),
        msg = paste0(what, " must have columns 'pu' and 'action'.")
      )

      spec$pu <- as_int_id_fun(spec$pu, paste0(what, "$pu"))
      spec$action <- as.character(spec$action)
      spec <- .normalize_feasible_col(spec, what)

      if (!all(spec$pu %in% pu_ids)) {
        bad <- unique(spec$pu[!spec$pu %in% pu_ids])
        stop(what, " contains PU ids not present in x: ", paste(bad, collapse = ", "), call. = FALSE)
      }
      if (!all(spec$action %in% action_ids)) {
        bad <- unique(spec$action[!spec$action %in% action_ids])
        stop(what, " contains action ids not present in actions: ", paste(bad, collapse = ", "), call. = FALSE)
      }

      tmp <- spec[, c("pu", "action")]
      if (nrow(dplyr::distinct(tmp)) != nrow(tmp)) {
        stop(what, " has duplicate (pu, action) rows. Please de-duplicate.", call. = FALSE)
      }

      out <- spec[spec$feasible, c("pu", "action"), drop = FALSE]
      return(out)
    }

    if (is.list(spec)) {
      if (is.null(names(spec)) || any(names(spec) == "")) {
        stop("If '", what, "' is a list, it must be a named list with names = action ids.", call. = FALSE)
      }
      if (!all(names(spec) %in% action_ids)) {
        bad <- setdiff(names(spec), action_ids)
        stop(what, " list contains unknown actions: ", paste(bad, collapse = ", "), call. = FALSE)
      }

      idx_first <- which(!vapply(spec, is.null, logical(1)))[1]
      first <- spec[[idx_first]]
      is_sf_list <- inherits(first, "sf")

      if (is_sf_list) {
        if (!requireNamespace("sf", quietly = TRUE)) {
          stop(what, " provided as sf layers requires the 'sf' package.", call. = FALSE)
        }
        if (is.null(pu_sf) || !inherits(pu_sf, "sf")) {
          stop("To use '", what, "' as sf layers, the problem object must contain x$data$pu_sf (sf planning unit geometry).", call. = FALSE)
        }
        if (!("id" %in% names(pu_sf))) {
          stop("x$data$pu_sf is missing an 'id' column.", call. = FALSE)
        }

        pu_sf2 <- pu_sf
        pu_sf2$id <- as_int_id_fun(pu_sf2$id, "x$data$pu_sf$id")

        out <- vector("list", length(spec))
        names(out) <- names(spec)

        for (a in names(spec)) {
          zone <- spec[[a]]
          if (is.null(zone)) {
            out[[a]] <- NULL
            next
          }
          if (!inherits(zone, "sf")) {
            stop(what, "[[", a, "]] must be an sf object.", call. = FALSE)
          }

          hits <- sf::st_intersects(pu_sf2, zone, sparse = TRUE)
          ids <- pu_sf2$id[lengths(hits) > 0]

          if (length(ids) == 0) {
            out[[a]] <- NULL
          } else {
            out[[a]] <- data.frame(
              pu = ids,
              action = a,
              stringsAsFactors = FALSE
            )
          }
        }

        out_df <- dplyr::bind_rows(out)
        if (nrow(out_df) == 0) return(out_df)
        out_df$pu <- as_int_id_fun(out_df$pu, paste0(what, "$pu"))
        out_df <- dplyr::distinct(out_df)
        return(out_df)
      }

      out <- vector("list", length(spec))
      names(out) <- names(spec)

      for (a in names(spec)) {
        ids <- spec[[a]]
        if (is.null(ids)) {
          out[[a]] <- NULL
          next
        }
        ids <- unique(as_int_id_fun(ids, paste0(what, "[['", a, "']]")))
        if (!all(ids %in% pu_ids)) {
          bad <- ids[!ids %in% pu_ids]
          stop(what, "[[", a, "]] contains PU ids not present in x: ", paste(bad, collapse = ", "), call. = FALSE)
        }
        out[[a]] <- data.frame(
          pu = ids,
          action = a,
          stringsAsFactors = FALSE
        )
      }

      out_df <- dplyr::bind_rows(out)
      if (nrow(out_df) == 0) return(out_df)
      out_df$pu <- as_int_id_fun(out_df$pu, paste0(what, "$pu"))
      out_df <- dplyr::distinct(out_df)
      return(out_df)
    }

    stop("Unsupported type for '", what, "'. Use NULL, data.frame, or a named list.", call. = FALSE)
  }

  .validate_lock_pairs_exist <- function(lock_pairs, dist_actions, what) {
    if (is.null(lock_pairs) || nrow(lock_pairs) == 0) return(invisible(TRUE))

    key_da <- paste(dist_actions$pu, dist_actions$action)
    key_lk <- paste(lock_pairs$pu, lock_pairs$action)

    miss <- !(key_lk %in% key_da)
    if (any(miss)) {
      bad <- lock_pairs[miss, , drop = FALSE]
      ex <- utils::head(paste0("(", bad$pu, ", ", bad$action, ")"), 8)
      stop(
        what, " contains pairs that are not feasible in x$data$dist_actions.\n",
        "Examples: ", paste(ex, collapse = ", "),
        if (nrow(bad) > 8) paste0(" ... (", nrow(bad), " total)") else "",
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  stopifnot(inherits(x, "Problem"))

  if (is.null(x$data$dist_actions) || !inherits(x$data$dist_actions, "data.frame")) {
    stop("No action feasibility found. Run add_actions() before add_constraint_locked_actions().", call. = FALSE)
  }

  x <- .pa_clone_data(x)

  da <- x$data$dist_actions
  pu <- x$data$pu
  actions <- x$data$actions
  pu_sf <- x$data$pu_sf %||% NULL

  if (is.null(pu) || !inherits(pu, "data.frame") || nrow(pu) == 0) {
    stop("x$data$pu is missing or empty.", call. = FALSE)
  }
  if (is.null(actions) || !inherits(actions, "data.frame") || nrow(actions) == 0) {
    stop("x$data$actions is missing or empty.", call. = FALSE)
  }

  pu_ids <- .as_int_id(pu$id, "x$data$pu$id")
  action_ids <- as.character(actions$id)

  locked_in_pairs <- .spec_to_pairs(locked_in, "locked_in", action_ids, pu_ids, pu_sf, .as_int_id)
  locked_out_pairs <- .spec_to_pairs(locked_out, "locked_out", action_ids, pu_ids, pu_sf, .as_int_id)

  .validate_lock_pairs_exist(locked_in_pairs, da, "locked_in")
  .validate_lock_pairs_exist(locked_out_pairs, da, "locked_out")

  if (!is.null(locked_in_pairs) && !is.null(locked_out_pairs) &&
      nrow(locked_in_pairs) > 0 && nrow(locked_out_pairs) > 0) {
    key_li <- paste(locked_in_pairs$pu, locked_in_pairs$action)
    key_lo <- paste(locked_out_pairs$pu, locked_out_pairs$action)
    overlap <- intersect(key_li, key_lo)

    if (length(overlap) > 0) {
      ex <- utils::head(overlap, 8)
      stop(
        "Some (pu, action) pairs are simultaneously locked_in and locked_out.\n",
        "Examples: ", paste(ex, collapse = ", "),
        if (length(overlap) > 8) paste0(" ... (", length(overlap), " total)") else "",
        call. = FALSE
      )
    }
  }

  if (is.null(da$status)) {
    da$status <- 0L
  }
  da$status <- as.integer(da$status)
  da$status[is.na(da$status)] <- 0L

  if (!is.null(locked_in_pairs) && nrow(locked_in_pairs) > 0) {
    key_da <- paste(da$pu, da$action)
    key_li <- paste(locked_in_pairs$pu, locked_in_pairs$action)
    idx_li <- key_da %in% key_li
    da$status[idx_li] <- 2L
  }

  if (!is.null(locked_out_pairs) && nrow(locked_out_pairs) > 0) {
    key_da <- paste(da$pu, da$action)
    key_lo <- paste(locked_out_pairs$pu, locked_out_pairs$action)
    idx_lo <- key_da %in% key_lo
    da$status[idx_lo] <- 3L
  }

  # enforce PU locked_out at the end
  if ("locked_out" %in% names(pu)) {
    pu_locked_out <- as.logical(pu$locked_out)
    pu_locked_out[is.na(pu_locked_out)] <- FALSE

    locked_out_pus <- pu$id[pu_locked_out]

    if (length(locked_out_pus) > 0) {
      idx_pu_lo <- da$pu %in% locked_out_pus

      if (any(idx_pu_lo & da$status == 2L, na.rm = TRUE)) {
        bad <- da[idx_pu_lo & da$status == 2L, c("pu", "action"), drop = FALSE]
        ex <- utils::head(paste0("(", bad$pu, ", ", bad$action, ")"), 8)
        stop(
          "Some actions are locked_in inside planning units that are locked_out.\n",
          "Examples: ", paste(ex, collapse = ", "),
          if (nrow(bad) > 8) paste0(" ... (", nrow(bad), " total)") else "",
          call. = FALSE
        )
      }

      da$status[idx_pu_lo] <- 3L
    }
  }

  x$data$dist_actions <- da
  x
}

#' @include internal.R
#'
#' @title Register an atomic objective (internal)
#' @name register_atomic_objective_internal
#'
#' @description
#' Internal helper used by objective setter functions to define the active
#' single-objective configuration of a \code{Problem} object and, optionally,
#' register that objective as an atomic objective for multi-objective workflows.
#'
#' @details
#' In \code{multiscape}, an \emph{atomic objective} is a fully specified
#' objective definition that can later be reused by a multi-objective method
#' such as a weighted-sum formulation, an \eqn{\epsilon}-constraint method,
#' AUGMECON, or other objective-orchestration procedures.
#'
#' Each atomic objective is identified by:
#' \itemize{
#'   \item a stable internal identifier \code{objective_id},
#'   \item a solver-facing model label \code{model_type},
#'   \item a list of objective-specific arguments stored in
#'   \code{objective_args},
#'   \item an optimization sense, either \code{"min"} or \code{"max"},
#'   \item and, optionally, a user-facing identifier \code{alias}.
#' }
#'
#' If \code{alias} is not \code{NULL}, the objective is stored in
#' \code{x$data$objectives[[alias]]}. This makes it possible to refer to the
#' same objective later by a stable user-facing name. For example, a user may
#' register objectives under aliases such as \code{"cost"},
#' \code{"benefit"}, or \code{"frag"} and then pass those aliases to a
#' multi-objective method.
#'
#' If \code{alias} is \code{NULL}, no atomic-objective entry is created in
#' \code{x$data$objectives}. In that case, the calling function still defines
#' the currently active single-objective configuration through
#' \code{x$data$model_args}, but no reusable multi-objective registration is
#' created.
#'
#' Thus, this helper supports two complementary modes:
#' \itemize{
#'   \item \strong{single-objective mode}: only the active objective is stored,
#'   \item \strong{multi-objective-ready mode}: the active objective is stored
#'   and also registered under an alias for later reuse.
#' }
#'
#' Conceptually, if an objective function is denoted by \eqn{f(x)}, this helper
#' does not itself define the mathematical form of \eqn{f}; rather, it stores
#' the metadata required so that downstream code can reconstruct the correct
#' objective expression, its direction of optimization, and its user-visible
#' identity.
#'
#' @param x A \code{Problem} object.
#' @param alias Optional character scalar used to register the objective as an
#'   atomic objective. If \code{NULL}, no registration entry is created.
#' @param objective_id Character string giving the stable internal identifier of
#'   the objective, for example \code{"min_cost"} or \code{"max_benefit"}.
#' @param model_type Character string giving the model-builder label associated
#'   with this objective, for example \code{"minimizeCosts"}.
#' @param objective_args A list of objective-specific arguments to be stored with
#'   the objective definition.
#' @param sense Character string giving the optimization direction. Must be
#'   either \code{"min"} or \code{"max"}.
#'
#' @return An updated \code{Problem} object.
#'
#' @keywords internal
NULL

#' @title Add objective: minimize cost
#'
#' @description
#' Define an objective that minimizes the total cost of the solution.
#'
#' Depending on the function arguments, the objective may include planning-unit
#' costs, action costs, or both. Action costs can optionally be restricted to a
#' subset of actions.
#'
#' @details
#' Use this function when the planning problem is framed primarily as a
#' cost-minimization problem, with costs arising from planning-unit selection,
#' action implementation, or both.
#'
#' Let \eqn{\mathcal{I}} be the set of planning units and let
#' \eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
#' feasible planning unit--action decisions.
#'
#' Let:
#' \itemize{
#'   \item \eqn{w_i \in \{0,1\}} denote whether planning unit \eqn{i} is
#'   selected,
#'   \item \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected
#'   in planning unit \eqn{i},
#'   \item \eqn{c_i^{PU} \ge 0} denote the planning-unit cost of unit \eqn{i},
#'   \item \eqn{c_{ia}^{A} \ge 0} denote the cost of selecting action
#'   \eqn{a} in planning unit \eqn{i}.
#' }
#'
#' The most general form of this objective is:
#'
#' \deqn{
#' \min \left(
#' \sum_{i \in \mathcal{I}} c_i^{PU} w_i
#' +
#' \sum_{(i,a) \in \mathcal{D}^{\star}} c_{ia}^{A} x_{ia}
#' \right),
#' }
#'
#' where \eqn{\mathcal{D}^{\star}} denotes the subset of feasible decisions
#' whose action contributes to the action-cost term.
#'
#' If \code{include_pu_cost = FALSE}, the planning-unit cost term is omitted.
#'
#' If \code{include_action_cost = FALSE}, the action-cost term is omitted.
#'
#' If \code{actions = NULL}, all feasible actions contribute to the action-cost
#' term. If \code{actions} is supplied, only the selected subset contributes to
#' that term. Planning-unit costs are never subset by \code{actions}; they are
#' always global whenever \code{include_pu_cost = TRUE}.
#'
#' @param x A \code{Problem} object.
#' @param include_pu_cost Logical. If \code{TRUE}, include planning-unit costs
#'   in the objective.
#' @param include_action_cost Logical. If \code{TRUE}, include action costs in
#'   the objective.
#' @param actions Optional subset of actions to include in the action-cost
#'   component. Values may match \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}. If \code{NULL}, all feasible actions are
#'   included in the action-cost term.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2))
#'
#' p1 <- add_objective_min_cost(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_min_cost(
#'   p,
#'   include_pu_cost = FALSE,
#'   include_action_cost = TRUE
#' )
#' p2$data$model_args
#'
#' p3 <- add_objective_min_cost(
#'   p,
#'   actions = "restoration"
#' )
#' p3$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_max_profit}},
#' \code{\link{add_objective_max_net_profit}}
#'
#' @export
add_objective_min_cost <- function(
    x,
    include_pu_cost = TRUE,
    include_action_cost = TRUE,
    actions = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  args <- list(
    include_pu_cost = isTRUE(include_pu_cost),
    include_action_cost = isTRUE(include_action_cost),
    actions = if (is.null(action_subset)) NULL else as.character(action_subset$id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "minimizeCosts",
    objective_id = "min_cost",
    objective_args = args,
    sense = "min",
    alias = alias
  )
}

#' @title Add objective: maximize benefit
#'
#' @description
#' Define an objective that maximizes the total positive effects generated by
#' selected actions on selected features.
#'
#' This objective is based on the canonical effects table and uses only the
#' non-negative \code{benefit} component.
#'
#' @details
#' Use this function when positive ecological gains should be maximized
#' explicitly, without offsetting them against harmful effects.
#'
#' Let \eqn{b_{iaf} \ge 0} denote the stored benefit associated with planning
#' unit \eqn{i}, action \eqn{a}, and feature \eqn{f}. Since the effects table is
#' already expressed in canonical form, \eqn{b_{iaf}} represents the positive
#' part of the net effect associated with the corresponding selected action
#' decision.
#'
#' If no subsets are supplied, the objective can be written as:
#'
#' \deqn{
#' \max \sum_{(i,a,f) \in \mathcal{R}} b_{iaf} \, x_{ia},
#' }
#'
#' where \eqn{\mathcal{R}} denotes the set of stored benefit rows and
#' \eqn{x_{ia} \in \{0,1\}} indicates whether action \eqn{a} is selected in
#' planning unit \eqn{i}.
#'
#' If \code{actions} is provided, only rows whose action belongs to the selected
#' subset contribute to the objective.
#'
#' If \code{features} is provided, only rows whose feature belongs to the
#' selected subset contribute to the objective.
#'
#' More generally, letting \eqn{\mathcal{R}^{\star}} be the subset induced by
#' the selected actions and features, the objective is:
#'
#' \deqn{
#' \max \sum_{(i,a,f) \in \mathcal{R}^{\star}} b_{iaf} \, x_{ia}.
#' }
#'
#' This objective maximizes gains only. It does not subtract losses. If harmful
#' effects should also be accounted for, they must be handled separately through
#' additional objectives or constraints.
#'
#' @param x A \code{Problem} object.
#' @param actions Optional subset of actions to include in the objective. Values
#'   may match \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}. If \code{NULL}, all actions are included.
#' @param features Optional subset of features to include in the objective.
#'   Values may match \code{x$data$features$id} and, if present,
#'   \code{x$data$features$name}. If \code{NULL}, all features are included.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#' effects_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   feature = c(1, 1, 1, 1, 2, 2, 2, 2),
#'   benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
#'   loss = c(0, 0, 1, 0, 0, 1, 0, 0)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_effects(effects_df)
#'
#' p1 <- add_objective_max_benefit(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_max_benefit(
#'   p,
#'   actions = "restoration"
#' )
#' p2$data$model_args
#'
#' p3 <- add_objective_max_benefit(
#'   p,
#'   features = 1
#' )
#' p3$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_min_loss}},
#' \code{\link{add_effects}}
#'
#' @export
add_objective_max_benefit <- function(
    x,
    actions = NULL,
    features = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  feature_subset <- NULL

  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  if (!is.null(features)) {
    feature_subset <- .pa_resolve_feature_subset(x, features)
  }

  args <- list(
    actions = if (is.null(action_subset)) NULL else as.integer(action_subset$internal_id),
    features = if (is.null(feature_subset)) NULL else as.integer(feature_subset$internal_id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "maximizeBenefits",
    objective_id = "max_benefit",
    objective_args = args,
    sense = "max",
    alias = alias
  )
}

#' @title Add objective: minimize loss
#'
#' @description
#' Define an objective that minimizes the total negative effects generated by
#' selected actions on selected features.
#'
#' This objective is based on the canonical effects table and uses only the
#' non-negative \code{loss} component.
#'
#' @details
#' Use this function when harmful ecological effects should be minimized
#' explicitly, without offsetting them against beneficial effects.
#'
#' Let \eqn{\ell_{iaf} \ge 0} denote the stored loss associated with planning
#' unit \eqn{i}, action \eqn{a}, and feature \eqn{f}.
#'
#' If no subsets are supplied, the objective can be written as:
#'
#' \deqn{
#' \min \sum_{(i,a,f) \in \mathcal{R}} \ell_{iaf} \, x_{ia}.
#' }
#'
#' where \eqn{\mathcal{R}} denotes the set of stored loss rows and
#' \eqn{x_{ia} \in \{0,1\}} indicates whether action \eqn{a} is selected in
#' planning unit \eqn{i}.
#'
#' If \code{actions} is provided, only rows whose action belongs to the selected
#' subset contribute to the objective.
#'
#' If \code{features} is provided, only rows whose feature belongs to the
#' selected subset contribute to the objective.
#'
#' More generally, letting \eqn{\mathcal{R}^{\star}} be the subset induced by
#' the selected actions and features, the objective is:
#'
#' \deqn{
#' \min \sum_{(i,a,f) \in \mathcal{R}^{\star}} \ell_{iaf} \, x_{ia}.
#' }
#'
#' This objective minimizes harmful effects only. It does not offset losses
#' against benefits unless benefits are handled elsewhere through additional
#' objectives or constraints.
#'
#' @param x A \code{Problem} object.
#' @param actions Optional subset of actions to include in the objective. Values
#'   may match \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}. If \code{NULL}, all actions are included.
#' @param features Optional subset of features to include in the objective.
#'   Values may match \code{x$data$features$id} and, if present,
#'   \code{x$data$features$name}. If \code{NULL}, all features are included.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#' effects_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   feature = c(1, 1, 1, 1, 2, 2, 2, 2),
#'   benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
#'   loss = c(0, 0, 1, 0, 0, 1, 0, 0)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_effects(effects_df)
#'
#' p1 <- add_objective_min_loss(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_min_loss(
#'   p,
#'   actions = "restoration"
#' )
#' p2$data$model_args
#'
#' p3 <- add_objective_min_loss(
#'   p,
#'   features = 1
#' )
#' p3$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_max_benefit}},
#' \code{\link{add_effects}}
#'
#' @export
add_objective_min_loss <- function(
    x,
    actions = NULL,
    features = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  feature_subset <- NULL

  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  if (!is.null(features)) {
    feature_subset <- .pa_resolve_feature_subset(x, features)
  }

  args <- list(
    actions = if (is.null(action_subset)) NULL else as.integer(action_subset$internal_id),
    features = if (is.null(feature_subset)) NULL else as.integer(feature_subset$internal_id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "minimizeLosses",
    objective_id = "min_loss",
    objective_args = args,
    sense = "min",
    alias = alias
  )
}

#' @title Add objective: maximize profit
#'
#' @description
#' Define an objective that maximizes total profit from selected planning
#' unit--action decisions.
#'
#' @details
#' Use this function when the objective is to maximize gross economic return,
#' without subtracting planning-unit or action costs.
#'
#' Let \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected in
#' planning unit \eqn{i}, and let \eqn{\pi_{ia}} denote the profit associated
#' with that decision, as taken from column \code{profit_col} in the stored
#' profit table.
#'
#' If all actions are included, the objective is:
#'
#' \deqn{
#' \max \sum_{(i,a) \in \mathcal{D}} \pi_{ia} x_{ia},
#' }
#'
#' where \eqn{\mathcal{D}} denotes the set of feasible planning unit--action
#' decisions.
#'
#' If \code{actions} is provided, only the selected subset contributes to the
#' objective. Letting \eqn{\mathcal{D}^{\star}} denote the feasible decisions
#' whose action belongs to the selected subset, the objective becomes:
#'
#' \deqn{
#' \max \sum_{(i,a) \in \mathcal{D}^{\star}} \pi_{ia} x_{ia}.
#' }
#'
#' This objective considers profit only. It does not subtract planning-unit
#' costs or action costs. For a net-profit formulation, use
#' \code{\link{add_objective_max_net_profit}}.
#'
#' @param x A \code{Problem} object.
#' @param profit_col Character string giving the profit column in the stored
#'   profit table.
#' @param actions Optional subset of actions to include. Values may match
#'   \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}. If \code{NULL}, all actions are included.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#' profit_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   profit = c(5, 4, 3, 2, 8, 7, 6, 5)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_profit(profit_df)
#'
#' p1 <- add_objective_max_profit(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_max_profit(
#'   p,
#'   actions = "restoration"
#' )
#' p2$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_min_cost}},
#' \code{\link{add_objective_max_net_profit}}
#'
#' @export
add_objective_max_profit <- function(
    x,
    profit_col = "profit",
    actions = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  args <- list(
    profit_col = as.character(profit_col)[1],
    actions = if (is.null(action_subset)) NULL else as.character(action_subset$id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "maximizeProfit",
    objective_id = "max_profit",
    objective_args = args,
    sense = "max",
    alias = alias
  )
}

#' @title Add objective: maximize net profit
#'
#' @description
#' Define an objective that maximizes net profit by combining profits with
#' optional planning-unit and action-cost penalties.
#'
#' @details
#' Use this function when decisions generate returns and the objective should
#' optimize the resulting net balance after subtracting selected cost
#' components.
#'
#' Let:
#' \itemize{
#'   \item \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected
#'   in planning unit \eqn{i},
#'   \item \eqn{w_i \in \{0,1\}} denote whether planning unit \eqn{i} is
#'   selected,
#'   \item \eqn{\pi_{ia}} denote the profit associated with decision
#'   \eqn{(i,a)},
#'   \item \eqn{c_i^{PU} \ge 0} denote the planning-unit cost,
#'   \item \eqn{c_{ia}^{A} \ge 0} denote the action cost.
#' }
#'
#' In its most general form, the objective is:
#'
#' \deqn{
#' \max \left(
#' \sum_{(i,a) \in \mathcal{D}^{\star}} \pi_{ia} x_{ia}
#' -
#' \sum_{i \in \mathcal{I}} c_i^{PU} w_i
#' -
#' \sum_{(i,a) \in \mathcal{D}^{\star}} c_{ia}^{A} x_{ia}
#' \right),
#' }
#'
#' where \eqn{\mathcal{D}^{\star}} denotes the subset of feasible
#' planning unit--action decisions included in the objective.
#'
#' If \code{actions = NULL}, all feasible actions contribute to both the profit
#' term and the action-cost term.
#'
#' If \code{actions} is provided, the profit term and the action-cost term are
#' restricted to that subset. The planning-unit cost term, if included, remains
#' global.
#'
#' If \code{include_pu_cost = FALSE}, the planning-unit cost term is omitted.
#'
#' If \code{include_action_cost = FALSE}, the action-cost term is omitted.
#'
#' @param x A \code{Problem} object.
#' @param profit_col Character string giving the profit column in the stored
#'   profit table.
#' @param include_pu_cost Logical. If \code{TRUE}, subtract planning-unit costs.
#' @param include_action_cost Logical. If \code{TRUE}, subtract action costs.
#' @param actions Optional subset of actions to include in the profit and
#'   action-cost terms. Values may match \code{x$data$actions$id} and, if
#'   present, \code{x$data$actions$action_set}.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#' profit_df <- data.frame(
#'   pu = c(1, 2, 3, 4, 1, 2, 3, 4),
#'   action = c("conservation", "conservation", "conservation", "conservation",
#'              "restoration", "restoration", "restoration", "restoration"),
#'   profit = c(5, 4, 3, 2, 8, 7, 6, 5)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
#'   add_profit(profit_df)
#'
#' p1 <- add_objective_max_net_profit(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_max_net_profit(
#'   p,
#'   include_pu_cost = FALSE,
#'   include_action_cost = TRUE
#' )
#' p2$data$model_args
#'
#' p3 <- add_objective_max_net_profit(
#'   p,
#'   actions = "restoration"
#' )
#' p3$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_max_profit}},
#' \code{\link{add_objective_min_cost}}
#'
#' @export
add_objective_max_net_profit <- function(
    x,
    profit_col = "profit",
    include_pu_cost = TRUE,
    include_action_cost = TRUE,
    actions = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  args <- list(
    profit_col = as.character(profit_col)[1],
    include_pu_cost = isTRUE(include_pu_cost),
    include_action_cost = isTRUE(include_action_cost),
    actions = if (is.null(action_subset)) NULL else as.character(action_subset$id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "maximizeNetProfit",
    objective_id = "max_net_profit",
    objective_args = args,
    sense = "max",
    alias = alias
  )
}

#' @title Add objective: minimize fragmentation
#'
#' @description
#' Define an objective that minimizes planning-unit fragmentation over a stored
#' spatial relation.
#'
#' This objective acts on the planning-unit selection pattern through the binary
#' planning-unit variables \eqn{w_i}. It is therefore appropriate when spatial
#' cohesion is to be encouraged at the level of the selected planning-unit set
#' as a whole.
#'
#' @details
#' Use this function when spatial cohesion should be encouraged at the level of
#' the selected planning-unit set as a whole.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and let
#' \eqn{w_i \in \{0,1\}} indicate whether planning unit \eqn{i \in \mathcal{I}}
#' is selected.
#'
#' Let the chosen spatial relation define a set of weighted pairs with weights
#' \eqn{\omega_{ij} \ge 0}. These relation weights are interpreted by the model
#' builder after scaling by \eqn{\lambda =} \code{weight_multiplier}.
#'
#' The internal preparation step constructs one auxiliary variable
#' \eqn{y_{ij} \in [0,1]} for each unique non-diagonal undirected edge
#' \eqn{(i,j)} with \eqn{i < j}. The intended semantics is:
#' \deqn{
#' y_{ij} = w_i \land w_j.
#' }
#'
#' This is enforced by the standard linearization:
#' \deqn{
#' y_{ij} \le w_i,
#' }
#' \deqn{
#' y_{ij} \le w_j,
#' }
#' \deqn{
#' y_{ij} \ge w_i + w_j - 1.
#' }
#'
#' Thus, \eqn{y_{ij}=1} if and only if both planning units \eqn{i} and
#' \eqn{j} are selected, and \eqn{y_{ij}=0} otherwise.
#'
#' The exact objective coefficients are assembled later by the model builder
#' from:
#' \itemize{
#'   \item the planning-unit variables \eqn{w_i},
#'   \item the edge-conjunction variables \eqn{y_{ij}},
#'   \item the stored relation weights \eqn{\omega_{ij}},
#'   \item and the multiplier \eqn{\lambda}.
#' }
#'
#' Conceptually, the resulting objective is a boundary- or relation-based
#' compactness functional that penalizes exposed or weakly connected selected
#' patterns while rewarding adjacency among selected planning units.
#'
#' In the common case where \code{relation_name = "boundary"} and the relation
#' was built with \code{\link{add_spatial_boundary}}, the objective corresponds
#' to a boundary-length-style fragmentation penalty.
#'
#' Setting \code{weight_multiplier = 0} removes the contribution of the spatial
#' relation from the objective after scaling.
#'
#' This objective does not distinguish between different actions within the same
#' planning unit. If action-specific spatial cohesion is required, use
#' \code{\link{add_objective_min_fragmentation_action}} instead.
#'
#' @param x A \code{Problem} object.
#' @param relation_name Character string giving the name of the spatial relation
#'   to use. The relation must already exist in
#'   \code{x$data$spatial_relations}.
#' @param weight_multiplier Numeric scalar greater than or equal to zero. Global
#'   multiplier applied to the relation weights when the objective is built.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' bound_df <- data.frame(
#'   id1 = c(1, 1, 2, 1, 2, 3, 4),
#'   id2 = c(1, 2, 2, 3, 4, 4, 4),
#'   boundary = c(4, 1, 4, 1, 1, 1, 4)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2))
#'
#' p <- add_spatial_boundary(
#'   x = p,
#'   boundary = bound_df,
#'   name = "boundary",
#'   include_self = TRUE,
#'   edge_factor = 1
#' )
#'
#' p <- add_objective_min_fragmentation_pu(
#'   p,
#'   relation_name = "boundary"
#' )
#'
#' p$data$model_args
#'
#' @seealso
#' \code{\link{add_spatial_boundary}},
#' \code{\link{add_spatial_relations}},
#' \code{\link{add_objective_min_fragmentation_action}}
#'
#' @export
add_objective_min_fragmentation_pu <- function(
    x,
    relation_name = "boundary",
    weight_multiplier = 1,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  relation_name <- as.character(relation_name)[1]
  weight_multiplier <- as.numeric(weight_multiplier)[1]

  if (!is.finite(weight_multiplier) || weight_multiplier < 0) {
    stop("weight_multiplier must be a finite number >= 0.", call. = FALSE)
  }

  rels <- x$data$spatial_relations
  if (is.null(rels) || !is.list(rels) || is.null(rels[[relation_name]])) {
    stop(
      "Spatial relation '", relation_name, "' not found in x$data$spatial_relations. ",
      "Add it first.",
      call. = FALSE
    )
  }

  args <- list(
    relation_name = relation_name,
    weight_multiplier = weight_multiplier
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "minimizeFragmentation",
    objective_id = "min_fragmentation",
    objective_args = args,
    sense = "min",
    alias = alias
  )
}

#' @title Add objective: minimize action fragmentation
#'
#' @description
#' Define an objective that minimizes fragmentation at the action level over a
#' stored spatial relation.
#'
#' Unlike \code{\link{add_objective_min_fragmentation_pu}}, which acts on the
#' selected planning-unit set through \eqn{w_i}, this objective acts on the
#' spatial arrangement of individual action decisions through the action
#' variables \eqn{x_{ia}}.
#'
#' @details
#' Use this function when spatial cohesion should be encouraged separately for
#' each selected action pattern.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and let
#' \eqn{\mathcal{A}} denote the set of actions.
#'
#' Let \eqn{x_{ia} \in \{0,1\}} indicate whether action \eqn{a \in \mathcal{A}}
#' is selected in planning unit \eqn{i \in \mathcal{I}}.
#'
#' Let the chosen spatial relation define weighted pairs with
#' weights \eqn{\omega_{ij} \ge 0}, and let
#' \eqn{\lambda =} \code{weight_multiplier} be the global scaling factor applied
#' to these weights.
#'
#' If \code{actions} is supplied, only the selected subset
#' \eqn{\mathcal{A}^{\star} \subseteq \mathcal{A}} contributes to the final
#' objective. If \code{actions = NULL}, all actions are included.
#'
#' The internal preparation step constructs one auxiliary variable
#' \eqn{b_{ija} \in [0,1]} for each unique non-diagonal undirected edge
#' \eqn{(i,j)} with \eqn{i < j} and for each action \eqn{a}. The intended
#' semantics is:
#' \deqn{
#' b_{ija} = x_{ia} \land x_{ja}.
#' }
#'
#' Whenever both decision variables \eqn{x_{ia}} and \eqn{x_{ja}} exist in the
#' model, this conjunction is enforced by the linearization:
#' \deqn{
#' b_{ija} \le x_{ia},
#' }
#' \deqn{
#' b_{ija} \le x_{ja},
#' }
#' \deqn{
#' b_{ija} \ge x_{ia} + x_{ja} - 1.
#' }
#'
#' If one of the two action variables does not exist because the corresponding
#' \code{(pu, action)} pair is not feasible, the auxiliary variable is forced to
#' zero.
#'
#' Therefore, \eqn{b_{ija}=1} if and only if action \eqn{a} is selected in both
#' adjacent planning units \eqn{i} and \eqn{j}; otherwise \eqn{b_{ija}=0}.
#'
#' The exact objective coefficients are assembled later by the model builder
#' from:
#' \itemize{
#'   \item the action decision variables \eqn{x_{ia}},
#'   \item the edge-conjunction variables \eqn{b_{ija}},
#'   \item the relation weights \eqn{\omega_{ij}},
#'   \item the multiplier \eqn{\lambda},
#'   \item and, if supplied, the action-specific weights.
#' }
#'
#' If action-specific weights are provided, let \eqn{\alpha_a \ge 0} denote the
#' weight associated with action \eqn{a}. Then the resulting objective can be
#' interpreted as an action-wise compactness or fragmentation functional of the
#' form:
#' \deqn{
#' \min \sum_{a \in \mathcal{A}^{\star}} \alpha_a \,
#' F_a(x_{\cdot a}, b_{\cdot\cdot a}; \lambda \omega),
#' }
#' where \eqn{F_a} is the fragmentation expression induced by the selected
#' relation and the internal coefficient construction for action \eqn{a}.
#'
#' In practical terms, this objective penalizes solutions in which the same
#' action is spatially scattered or broken into separate patches, while allowing
#' different actions to form different spatial patterns.
#'
#' This differs from planning-unit fragmentation:
#' \itemize{
#'   \item \code{add_objective_min_fragmentation_pu()} encourages cohesion of the
#'   union of selected planning units,
#'   \item \code{add_objective_min_fragmentation_action()} encourages cohesion of
#'   each selected action pattern separately.
#' }
#'
#' Setting \code{weight_multiplier = 0} removes the contribution of the spatial
#' relation from the objective after scaling.
#'
#' @param x A \code{Problem} object.
#' @param relation_name Character string giving the name of the spatial relation
#'   to use. The relation must already exist in
#'   \code{x$data$spatial_relations}.
#' @param weight_multiplier Numeric scalar greater than or equal to zero. Global
#'   multiplier applied to the relation weights when the objective is built.
#' @param action_weights Optional action weights. Either a named numeric vector
#'   with names equal to action ids, or a \code{data.frame} with columns
#'   \code{action} and \code{weight}. These weights scale the contribution of
#'   each action to the final objective.
#' @param actions Optional subset of actions to include. Values may match
#'   \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}. If \code{NULL}, all actions are included.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' bound_df <- data.frame(
#'   id1 = c(1, 1, 2, 1, 2, 3, 4),
#'   id2 = c(1, 2, 2, 3, 4, 4, 4),
#'   boundary = c(4, 1, 4, 1, 1, 1, 4)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2))
#'
#' p <- add_spatial_boundary(
#'   x = p,
#'   boundary = bound_df,
#'   name = "boundary",
#'   include_self = TRUE,
#'   edge_factor = 1
#' )
#'
#' p <- add_objective_min_fragmentation_action(
#'   p,
#'   relation_name = "boundary",
#'   actions = "restoration",
#'   weight_multiplier = 1
#' )
#'
#' p$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_min_fragmentation_pu}},
#' \code{\link{add_spatial_boundary}},
#' \code{\link{add_spatial_relations}}
#'
#' @export
add_objective_min_fragmentation_action <- function(
    x,
    relation_name = "boundary",
    weight_multiplier = 1,
    action_weights = NULL,
    actions = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  action_subset <- NULL
  if (!is.null(actions)) {
    action_subset <- .pa_resolve_action_subset(x, actions)
  }

  args <- list(
    relation_name = as.character(relation_name)[1],
    weight_multiplier = as.numeric(weight_multiplier)[1],
    action_weights = action_weights,
    actions = if (is.null(action_subset)) NULL else as.character(action_subset$id)
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "minimizeActionFragmentation",
    objective_id = "min_action_fragmentation",
    objective_args = args,
    sense = "min",
    alias = alias
  )
}

#' @title Add objective: minimize intervention impact
#'
#' @description
#' Define an objective that minimizes the impact associated with selecting
#' planning units for intervention.
#'
#' This objective uses planning-unit selection variables rather than summing the
#' same impact repeatedly over multiple actions. As a result, each planning unit
#' contributes at most once to the objective, regardless of how many feasible
#' actions exist in that unit.
#'
#' @details
#' Use this function when intervention itself has a baseline ecological, social,
#' or operational burden that should be minimized independently of the detailed
#' effects of particular actions.
#'
#' Let \eqn{w_i \in \{0,1\}} denote whether planning unit \eqn{i} is selected
#' for intervention. Let \eqn{q_{if}} denote the impact amount associated with
#' planning unit \eqn{i} and feature \eqn{f}, taken from column
#' \code{impact_col} in the feature-distribution table.
#'
#' If all selected features are included, the objective can be interpreted as:
#'
#' \deqn{
#' \min \sum_{i \in \mathcal{I}} \left( \sum_{f \in \mathcal{F}^{\star}} q_{if}
#' \right) w_i,
#' }
#'
#' where \eqn{\mathcal{F}^{\star}} denotes the selected subset of features.
#'
#' Thus, the coefficient attached to \eqn{w_i} is the aggregated impact of the
#' selected features in planning unit \eqn{i}.
#'
#' The role of \code{actions} in this objective is not to make impact additive
#' over actions, but to restrict the notion of intervention to planning units
#' that are relevant for the selected subset of actions in downstream model
#' construction. Even when multiple feasible actions exist in a planning unit,
#' the planning unit contributes at most once through \eqn{w_i}.
#'
#' This objective is useful when intervention itself has a baseline ecological,
#' social, or operational impact that should be minimized independently of the
#' detailed gain or loss generated by particular actions.
#'
#' @param x A \code{Problem} object.
#' @param impact_col Character string giving the column in the
#'   feature-distribution table that contains the per-\code{(pu, feature)}
#'   impact amount. The default is \code{"amount"}.
#' @param features Optional subset of features to include. Values may match
#'   \code{x$data$features$id} and, if present, \code{x$data$features$name}.
#' @param actions Optional subset of actions used to define the intervention
#'   context. Values may match \code{x$data$actions$id} and, if present,
#'   \code{x$data$actions$action_set}.
#' @param alias Optional identifier used to register this objective for
#'   multi-objective workflows.
#'
#' @return An updated \code{Problem} object.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#' actions_df <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("conservation", "restoration")
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(actions_df, cost = c(conservation = 1, restoration = 2))
#'
#' p1 <- add_objective_min_intervention_impact(p)
#' p1$data$model_args
#'
#' p2 <- add_objective_min_intervention_impact(
#'   p,
#'   features = 1
#' )
#' p2$data$model_args
#'
#' p3 <- add_objective_min_intervention_impact(
#'   p,
#'   actions = "restoration"
#' )
#' p3$data$model_args
#'
#' @seealso
#' \code{\link{add_objective_max_benefit}},
#' \code{\link{add_objective_min_loss}}
#'
#' @export
add_objective_min_intervention_impact <- function(
    x,
    impact_col = "amount",
    features = NULL,
    actions = NULL,
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))

  impact_col <- as.character(impact_col)[1]
  if (is.na(impact_col) || !nzchar(impact_col)) {
    stop("impact_col must be a non-empty string.", call. = FALSE)
  }

  if (is.null(x$data$dist_features) ||
      !inherits(x$data$dist_features, "data.frame") ||
      nrow(x$data$dist_features) == 0) {
    stop("x$data$dist_features is missing or empty.", call. = FALSE)
  }

  if (!(impact_col %in% names(x$data$dist_features))) {
    stop(
      "impact_col '", impact_col, "' was not found in x$data$dist_features.",
      call. = FALSE
    )
  }

  feat_subset <- .pa_resolve_feature_subset(x, features = features)
  act_subset  <- .pa_resolve_action_subset(x, subset = actions)

  args <- list(
    impact_col = impact_col,
    features = feat_subset$id,
    actions = act_subset$id
  )

  .pa_set_active_and_register_objective(
    x = x,
    model_type = "minimizeInterventionImpact",
    objective_id = "min_intervention_impact",
    objective_args = args,
    sense = "min",
    alias = alias
  )
}

test_that("compile_model builds MO superset with boundary auxiliaries when fragmentation is included", {
  toy <- toy_equivalent_basic()

  p2 <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_spatial_boundary(boundary = toy$boundary, include_self = TRUE) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag") |>
    multiscape::set_method_weighted_sum(
      aliases = c("cost", "frag"),
      weights = c(1, 1)
    )

  p2 <- multiscape::compile_model(p2)

  expect_s3_class(p2, "Problem")
  expect_false(is.null(p2$data$model_ptr))
  expect_true(is.list(p2$data$model_list))
  expect_false(isTRUE(p2$data$meta$model_dirty))

  n_y_pu <- as.integer(p2$data$model_list$n_y_pu %||% 0L)
  expect_gt(n_y_pu, 0)
})

test_that("get_actions returns selected actions for Solution", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  a_all <- multiscape::get_actions(s)
  expect_true(is.data.frame(a_all))
  expect_gt(nrow(a_all), 0)

  a_sel <- multiscape::get_actions(s, only_selected = TRUE)
  expect_true(is.data.frame(a_sel))
})

test_that("get_actions returns actions for SolutionSet run", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_spatial_boundary(boundary = toy$boundary, include_self = TRUE) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag") |>
    multiscape::set_method_weighted_sum(
      aliases = c("cost", "frag"),
      weights = c(1, 1)
    ) |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  a <- multiscape::get_actions(s, run = 1)
  expect_true(is.data.frame(a))
})

test_that("compile_model errors if multiple objectives are present without a MO method", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_spatial_boundary(boundary = toy$boundary, include_self = TRUE) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag")

  expect_error(
    multiscape::compile_model(p),
    "Multiple objectives are registered"
  )
})

test_that("weighted method returns a SolutionSet with runs", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_spatial_boundary(boundary = toy$boundary, include_self = TRUE) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag") |>
    multiscape::set_method_weighted_sum(
      aliases = c("cost", "frag"),
      weights = c(1, 1)
    ) |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_s3_class(s, "SolutionSet")

  expect_true("solution" %in% names(s))
  expect_true(is.list(s$solution))

  expect_true("runs" %in% names(s$solution))
  expect_true("solutions" %in% names(s$solution))

  expect_true(is.data.frame(s$solution$runs))
  expect_gte(nrow(s$solution$runs), 1)

  expect_true(is.list(s$solution$solutions))
  expect_gte(length(s$solution$solutions), 1)
})

test_that("printing Problem does not fail", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  expect_no_error(print(p))
})

test_that("printing Solution does not fail", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_no_error(print(s))
})

test_that("printing SolutionSet does not fail", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_spatial_boundary(boundary = toy$boundary, include_self = TRUE) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag") |>
    multiscape::set_method_weighted_sum(
      aliases = c("cost", "frag"),
      weights = c(1, 1)
    ) |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_no_error(print(s))
})


test_that("printing Problem with area and budget constraints does not fail", {
  toy <- toy_multiaction_semantics()

  p <- multiscape::create_problem(
    pu = transform(toy$pu, area_ha = c(1, 1, 1, 1, 1)),
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = c(conservation = 1, restoration = 2)) |>
    multiscape::add_constraint_area(area = 2, sense = "max", area_col = "area_ha") |>
    multiscape::add_constraint_budget(
      budget = 10,
      sense = "max",
      include_pu_cost = TRUE,
      include_action_cost = TRUE
    )

  expect_no_error(print(p))
})

test_that("set_method_weighted_sum errors when aliases and weights have different lengths", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_spatial_boundary(
      boundary = toy$boundary,
      weight_col = "boundary",
      include_self = TRUE
    ) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag")

  expect_error(
    multiscape::set_method_weighted_sum(
      p,
      aliases = c("cost", "frag"),
      weights = c(1)
    ),
    "length"
  )
})

test_that("set_method_weighted_sum errors when aliases contain duplicates", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_spatial_boundary(
      boundary = toy$boundary,
      weight_col = "boundary",
      include_self = TRUE
    ) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_min_fragmentation_pu(alias = "frag")

  expect_error(
    multiscape::set_method_weighted_sum(
      p,
      aliases = c("cost", "cost"),
      weights = c(1, 1)
    )
  )
})

test_that("epsilon-constraint returns a SolutionSet with runs", {
  skip_if_no_cbc()

  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  ) |>
    multiscape::add_actions(actions = toy$actions, cost = 0) |>
    multiscape::add_effects(effects = toy$effects, effect_type = "after") |>
    multiscape::add_constraint_targets_relative(0.5) |>
    multiscape::add_objective_min_cost(alias = "cost") |>
    multiscape::add_objective_max_benefit(alias = "benefit") |>
    multiscape::set_method_epsilon_constraint(
      primary = "cost",
      aliases = c("cost", "benefit"),
      mode = "manual",
      eps = data.frame(
        benefit = c(0, 1)
      )
    ) |>
    multiscape::set_solver_cbc(gap_limit = 0, verbose = FALSE)

  s <- multiscape::solve(p)

  expect_s3_class(s, "SolutionSet")
  expect_true(is.list(s$solution))
  expect_true("runs" %in% names(s$solution))
  expect_true("solutions" %in% names(s$solution))
  expect_gte(nrow(s$solution$runs), 1)
  expect_gte(length(s$solution$solutions), 1)
})

#' Example feature raster
#'
#' Load the example feature raster shipped with the package.
#'
#' @return A `terra::SpatRaster`.
#' @export
load_sim_features_raster <- function() {
  path <- system.file("extdata", "sim_features_raster.tif", package = "multiscape")
  terra::rast(path)
}

#' @title Add absolute targets
#'
#' @description
#' Add feature-level absolute targets to a planning problem.
#'
#' These targets are stored in the problem object and later translated into
#' linear constraints when the optimization model is built. Absolute targets
#' are interpreted directly in the same units as the feature
#' contributions used by the model. Each call appends one or more target definitions to the problem. This makes
#' it possible to combine multiple target rules, including targets associated
#' with different action subsets.
#'
#' @details
#' Use this function when target requirements are naturally expressed in the
#' original units of the modelled feature contributions, rather than as
#' proportions of current baseline totals.
#'
#' Let \eqn{\mathcal{F}} denote the set of features. For each targeted feature
#' \eqn{f \in \mathcal{F}}, this function stores an absolute target threshold
#' \eqn{T_f \ge 0}.
#'
#' When the optimization model is built, each such target is interpreted as a
#' lower-bound constraint of the form:
#' \deqn{
#' \sum_{(i,a) \in \mathcal{D}_f^{\star}} c_{iaf} x_{ia} \ge T_f,
#' }
#' where:
#' \itemize{
#'   \item \eqn{i \in \mathcal{I}} indexes planning units,
#'   \item \eqn{a \in \mathcal{A}} indexes actions,
#'   \item \eqn{x_{ia}} indicates whether action \eqn{a} is selected in planning
#'   unit \eqn{i},
#'   \item \eqn{c_{iaf}} is the contribution of that action to feature
#'   \eqn{f},
#'   \item \eqn{\mathcal{D}_f^{\star}} is the subset of planning unit--action
#'   pairs allowed to count toward the target for feature \eqn{f}.
#' }
#'
#' In the absolute case, the stored target threshold is simply:
#' \deqn{
#' T_f = t_f,
#' }
#' where \eqn{t_f} is the user-supplied target value for feature \eqn{f}.
#'
#' The \code{actions} argument restricts which actions may contribute toward
#' achievement of the target, but it does not modify the value of \eqn{T_f}
#' itself.
#'
#' The \code{targets} argument is parsed by \code{.pa_parse_targets()} and may be
#' supplied in several equivalent forms, including:
#' \itemize{
#'   \item a single numeric value recycled to all selected features,
#'   \item a numeric vector aligned to \code{features},
#'   \item a named numeric vector where names identify features,
#'   \item a \code{data.frame} with \code{feature} and \code{target} columns.
#' }
#'
#' If \code{targets} does not explicitly identify features:
#' \itemize{
#'   \item if \code{features = NULL}, the target is applied to all features,
#'   \item if \code{features} is supplied, the target values are interpreted with
#'   respect to that feature set.
#' }
#'
#' Repeated calls append new target rules rather than replacing previous ones.
#' This allows cumulative target modelling, including multiple rules on the same
#' feature with different contributing action subsets.
#'
#' @param x A \code{Problem} object.
#' @param targets Target specification. This is interpreted as an absolute target
#'   value in the same units as the modelled feature contributions. It may be a
#'   scalar, vector, named vector, or \code{data.frame}. See Details.
#' @param features Optional feature specification indicating which features the
#'   supplied target values refer to when \code{targets} does not identify
#'   features explicitly. If \code{NULL}, all features are targeted.
#' @param actions Optional character vector indicating which actions count toward
#'   target achievement. Entries may match action ids, \code{action_set} labels,
#'   or both. If \code{NULL}, all actions count.
#' @param label Optional character string stored with the targets for reporting
#'   and bookkeeping.
#'
#' @return An updated \code{Problem} object with absolute targets appended to
#'   the stored target table.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(data.frame(id = "conservation", name = "conservation"), cost = 0)
#'
#' # Same absolute target for all features
#' p1 <- add_constraint_targets_absolute(p, 3)
#' p1$data$targets
#'
#' # Different targets by feature
#' p2 <- add_constraint_targets_absolute(
#'   p,
#'   c("1" = 4, "2" = 2)
#' )
#' p2$data$targets
#'
#' # Restrict which actions count toward target achievement
#' p3 <- add_constraint_targets_absolute(
#'   p,
#'   2,
#'   actions = "conservation"
#' )
#' p3$data$targets
#'
#' @seealso
#' \code{\link{add_constraint_targets_relative}}
#'
#' @export
add_constraint_targets_absolute <- function(x, targets,
                                 features = NULL,
                                 actions = NULL,
                                 label = NULL) {
  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  dt <- .pa_parse_targets(x, targets, features = features)

  actions_txt <- .pa_subset_to_string(actions)

  out <- data.frame(
    feature      = as.numeric(dt$feature),
    type         = "actions",
    sense        = "ge",
    target_unit  = "absolute",
    target_raw   = as.numeric(dt$target_raw),
    basis_total  = NA_real_,
    target_value = as.numeric(dt$target_raw),
    actions      = actions_txt,
    label        = if (is.null(label)) NA_character_ else as.character(label),
    created_at   = as.character(Sys.time()),
    stringsAsFactors = FALSE
  )

  out <- .pa_add_feature_labels(
    df = out,
    features_df = x$data$features,
    feature_col = "feature",
    internal_feature_col = "internal_feature",
    out_col = "feature_name"
  )

  x$data$meta$model_dirty <- TRUE
  .pa_store_targets(x, out)
}

#' @title Add relative targets
#'
#' @description
#' Add feature-level relative targets to a planning problem.
#'
#' These targets are stored in the problem object and later translated into
#' linear constraints when the optimization model is built. Relative targets are
#' supplied as proportions in \eqn{[0,1]} and are converted internally into
#' absolute thresholds using the current total amount of each
#' feature in the landscape.
#'
#' Each call appends one or more target definitions to the problem. This makes
#' it possible to combine multiple target rules, including targets associated
#' with different action subsets.
#'
#' @details
#' Use this function when target requirements are naturally expressed as
#' proportions of current baseline feature totals rather than in original
#' feature units.
#'
#' Let \eqn{\mathcal{F}} denote the set of features. For each targeted feature
#' \eqn{f \in \mathcal{F}}, let \eqn{B_f} denote the current baseline total
#' amount of that feature in the landscape, as computed by
#' \code{.pa_feature_totals()}.
#'
#' If the user supplies a relative target \eqn{r_f \in [0,1]}, then this
#' function converts it to an absolute threshold:
#' \deqn{
#' T_f = r_f \times B_f.
#' }
#'
#' The absolute threshold \eqn{T_f} is stored in \code{target_value}, while:
#' \itemize{
#'   \item the original user-supplied proportion \eqn{r_f} is stored in
#'   \code{target_raw},
#'   \item the baseline total \eqn{B_f} is stored in \code{basis_total}.
#' }
#'
#' When the optimization model is built, the resulting target is interpreted as:
#' \deqn{
#' \sum_{(i,a) \in \mathcal{D}_f^{\star}} c_{iaf} x_{ia} \ge T_f,
#' }
#' where:
#' \itemize{
#'   \item \eqn{i \in \mathcal{I}} indexes planning units,
#'   \item \eqn{a \in \mathcal{A}} indexes actions,
#'   \item \eqn{x_{ia}} indicates whether action \eqn{a} is selected in planning
#'   unit \eqn{i},
#'   \item \eqn{c_{iaf}} is the contribution of that action to feature
#'   \eqn{f},
#'   \item \eqn{\mathcal{D}_f^{\star}} is the subset of planning unit--action
#'   pairs allowed to count toward the target for feature \eqn{f}.
#' }
#'
#' The \code{actions} argument restricts which actions may contribute toward
#' target achievement, but it does not affect the baseline amount \eqn{B_f} used
#' to compute the threshold. In other words, relative targets are always scaled
#' against the current full landscape baseline.
#'
#' Therefore, \code{actions} changes who may satisfy the target, but not how the
#' threshold itself is scaled.
#'
#' The \code{targets} argument is parsed by \code{.pa_parse_targets()} and may be
#' supplied in several equivalent forms, including:
#' \itemize{
#'   \item a single numeric value recycled to all selected features,
#'   \item a numeric vector aligned to \code{features},
#'   \item a named numeric vector where names identify features,
#'   \item a \code{data.frame} with \code{feature} and \code{target} columns.
#' }
#'
#' If \code{targets} does not explicitly identify features:
#' \itemize{
#'   \item if \code{features = NULL}, the target is applied to all features,
#'   \item if \code{features} is supplied, the target values are interpreted with
#'   respect to that feature set.
#' }
#'
#' Relative targets must lie in \eqn{[0,1]}.
#'
#' Repeated calls append new target rules rather than replacing previous ones.
#' This allows cumulative target modelling, including multiple rules on the same
#' feature with different contributing action subsets.
#'
#' @param x A \code{Problem} object.
#' @param targets Target specification as proportions in \eqn{[0,1]}. It may be
#'   a scalar, vector, named vector, or \code{data.frame}. See Details.
#' @param features Optional feature specification indicating which features the
#'   supplied target values refer to when \code{targets} does not identify
#'   features explicitly. If \code{NULL}, all features are targeted.
#' @param actions Optional character vector indicating which actions count toward
#'   target achievement. Entries may match action ids, \code{action_set} labels,
#'   or both. If \code{NULL}, all actions count.
#' @param label Optional character string stored with the targets for reporting
#'   and bookkeeping.
#'
#' @return An updated \code{Problem} object with relative targets appended to
#'   the stored target table.
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' p <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' ) |>
#'   add_actions(data.frame(id = "conservation", name = "conservation"), cost = 0)
#'
#' # Require 30% of the baseline total for all features
#' p1 <- add_constraint_targets_relative(p, 0.3)
#' p1$data$targets
#'
#' # Require 20% for one selected feature
#' p2 <- add_constraint_targets_relative(
#'   p,
#'   0.2,
#'   features = 1
#' )
#' p2$data$targets
#'
#' # Restrict which actions count toward target achievement
#' p3 <- add_constraint_targets_relative(
#'   p,
#'   0.2,
#'   actions = "conservation"
#' )
#' p3$data$targets
#'
#' @seealso
#' \code{\link{add_constraint_targets_absolute}}
#'
#' @export
add_constraint_targets_relative <- function(x, targets,
                                 features = NULL,
                                 actions = NULL,
                                 label = NULL) {
  stopifnot(inherits(x, "Problem"))

  x <- .pa_clone_data(x)
  dt <- .pa_parse_targets(x, targets, features = features)

  rel <- as.numeric(dt$target_raw)
  if (any(rel < 0 | rel > 1, na.rm = TRUE)) {
    stop("Relative targets must be between 0 and 1.", call. = FALSE)
  }

  basis <- .pa_feature_totals(x)
  basis_v <- basis[as.character(dt$feature)]
  basis_v[is.na(basis_v)] <- 0
  abs_target <- rel * as.numeric(basis_v)

  actions_txt <- .pa_subset_to_string(actions)

  out <- data.frame(
    feature      = as.numeric(dt$feature),
    type         = "actions",
    sense        = "ge",
    target_unit  = "relative_baseline",
    target_raw   = rel,
    basis_total  = as.numeric(basis_v),
    target_value = as.numeric(abs_target),
    actions      = actions_txt,
    label        = if (is.null(label)) NA_character_ else as.character(label),
    created_at   = as.character(Sys.time()),
    stringsAsFactors = FALSE
  )

  out <- .pa_add_feature_labels(
    df = out,
    features_df = x$data$features,
    feature_col = "feature",
    internal_feature_col = "internal_feature",
    out_col = "feature_name"
  )

  x$data$meta$model_dirty <- TRUE
  .pa_store_targets(x, out)
}

test_that("add_targets_relative computes absolute targets from total amounts", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  p <- multiscape::add_constraint_targets_relative(p, 0.5)

  expect_true(is.data.frame(p$data$targets))
  expect_true(all(c("feature", "target_value") %in% names(p$data$targets)))

  total_f1 <- sum(toy$dist_features$amount[toy$dist_features$feature == 1])
  total_f2 <- sum(toy$dist_features$amount[toy$dist_features$feature == 2])

  tgt <- p$data$targets
  tgt_f1 <- tgt$target_value[tgt$feature == 1]
  tgt_f2 <- tgt$target_value[tgt$feature == 2]

  expect_equal(tgt_f1, 0.5 * total_f1)
  expect_equal(tgt_f2, 0.5 * total_f2)

  expect_true(all(tgt$type == "actions"))
  expect_true(all(tgt$sense == "ge"))
  expect_true(all(tgt$target_unit == "relative_baseline"))
  expect_equal(tgt$target_raw[tgt$feature == 1], 0.5)
  expect_equal(tgt$basis_total[tgt$feature == 1], total_f1)
})

#' @include internal.R
#'
#' @title Add budget constraint
#'
#' @description
#' Add a budget constraint to a planning problem.
#'
#' This function stores one budget-constraint specification in the
#' \code{Problem} object so that it can later be incorporated when the
#' optimization model is assembled. Multiple budget constraints can be added by
#' calling this function repeatedly, provided that no duplicated combination of
#' action subset and constraint sense is introduced.
#'
#' @details
#' Use this function when spending limits or minimum spending requirements must
#' be imposed either on the full problem or on the subset of selected decisions
#' associated with specific actions.
#'
#' Let \eqn{\mathcal{I}} denote the set of planning units and let
#' \eqn{\mathcal{A}} denote the set of actions. Let
#' \eqn{w_i \in \{0,1\}} denote the binary variable indicating whether planning
#' unit \eqn{i \in \mathcal{I}} is selected by at least one decision in the
#' model, and let \eqn{x_{ia} \in \{0,1\}} denote the binary variable
#' indicating whether action \eqn{a \in \mathcal{A}} is selected in planning
#' unit \eqn{i \in \mathcal{I}}.
#'
#' The total constrained budget can include two cost components:
#'
#' \itemize{
#'   \item planning-unit costs, associated with \eqn{w_i};
#'   \item action costs, associated with \eqn{x_{ia}}.
#' }
#'
#' The arguments \code{include_pu_cost} and \code{include_action_cost}
#' determine which of these components are included in the stored budget
#' constraint.
#'
#' When \code{actions = NULL}, the constraint refers to the total budget across
#' the whole problem. In that case, depending on the values of
#' \code{include_pu_cost} and \code{include_action_cost}, the constrained
#' quantity is one of the following:
#'
#' If only planning-unit costs are included:
#' \deqn{
#' \sum_{i \in \mathcal{I}} c_i^{pu} w_i
#' }
#'
#' If only action costs are included:
#' \deqn{
#' \sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}} c_{ia}^{act} x_{ia}
#' }
#'
#' If both components are included:
#' \deqn{
#' \sum_{i \in \mathcal{I}} c_i^{pu} w_i +
#' \sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}} c_{ia}^{act} x_{ia}
#' }
#'
#' Depending on \code{sense}, this function stores one of the following
#' constraints:
#'
#' If \code{sense = "min"}:
#' \deqn{
#' C \ge B
#' }
#'
#' If \code{sense = "max"}:
#' \deqn{
#' C \le B
#' }
#'
#' If \code{sense = "equal"} and \code{tolerance = 0}:
#' \deqn{
#' C = B
#' }
#'
#' If \code{sense = "equal"} and \code{tolerance > 0}, the equality is stored as
#' a two-sided band:
#' \deqn{
#' B - \tau \le C \le B + \tau
#' }
#'
#' where \eqn{C} denotes the selected cost expression and \eqn{\tau} is the
#' value supplied through \code{tolerance}.
#'
#' When \code{actions} is not \code{NULL}, only action costs can be included.
#' In that case, let \eqn{\mathcal{A}^\star \subseteq \mathcal{A}} denote the
#' selected subset of actions, and the constrained quantity is
#' \deqn{
#' \sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}^\star} c_{ia}^{act} x_{ia}.
#' }
#'
#' Action-specific budget constraints only support action costs. Therefore,
#' \code{include_pu_cost = TRUE} is only allowed when \code{actions = NULL},
#' because planning-unit costs are not action-specific.
#'
#' This function only stores the constraint specification; it does not validate
#' the feasibility of the threshold against the available cost data at this
#' stage.
#'
#' Multiple budget constraints can be stored in a \code{Problem} object.
#' However, at most one can be stored for the same combination of action subset
#' and constraint sense. Attempting to add a duplicated
#' \code{actions}--\code{sense} combination results in an error.
#'
#' @param x A \code{Problem} object.
#'
#' @param budget Numeric scalar greater than or equal to zero. Target value for
#'   the constrained budget.
#'
#' @param sense Character string indicating the type of budget constraint. Must
#'   be one of \code{"min"}, \code{"max"}, or \code{"equal"}.
#'
#' @param tolerance Numeric scalar greater than or equal to zero. Only used when
#'   \code{sense = "equal"}. In that case, equality is interpreted as a band
#'   around \code{budget} with half-width \code{tolerance}. Ignored otherwise.
#'
#' @param actions Optional subset of actions to which the constraint applies.
#'   If \code{NULL}, the constraint applies to the whole problem. Otherwise, it
#'   applies only to the selected decision variables associated with the
#'   specified subset of actions. This argument is resolved using the package's
#'   standard action subset parser.
#'
#' @param include_pu_cost Logical scalar indicating whether planning-unit costs
#'   should be included in the constrained budget. This is only supported when
#'   \code{actions = NULL}.
#'
#' @param include_action_cost Logical scalar indicating whether action costs
#'   should be included in the constrained budget.
#'
#' @param name Optional character string used as the label of the stored linear
#'   constraint when it is later added to the optimization model. If
#'   \code{NULL}, a default name is generated.
#'
#' @return An updated \code{Problem} object with the new budget-constraint
#'   specification appended to the stored budget-constraint table.
#'
#' @seealso
#' \code{\link{create_problem}}
#'
#' @examples
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(2, 3, 1, 4)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4, 4),
#'   feature = c(1, 2, 1, 2, 1, 2),
#'   amount = c(1, 2, 1, 3, 2, 1)
#' )
#'
#' actions <- data.frame(
#'   id = c("conservation", "restoration")
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   p,
#'   actions = actions,
#'   cost = c(conservation = 1, restoration = 2)
#' )
#'
#' p <- add_constraint_budget(
#'   x = p,
#'   budget = 10,
#'   sense = "max",
#'   include_pu_cost = TRUE,
#'   include_action_cost = TRUE
#' )
#'
#' p <- add_constraint_budget(
#'   x = p,
#'   budget = 4,
#'   sense = "max",
#'   actions = "restoration",
#'   include_pu_cost = FALSE,
#'   include_action_cost = TRUE
#' )
#'
#' p <- add_constraint_budget(
#'   x = p,
#'   budget = 1,
#'   sense = "min",
#'   actions = "restoration",
#'   include_pu_cost = FALSE,
#'   include_action_cost = TRUE
#' )
#'
#' p$data$constraints$budget
#'
#' @export
add_constraint_budget <- function(x,
                                  budget,
                                  sense,
                                  tolerance = 0,
                                  actions = NULL,
                                  include_pu_cost = TRUE,
                                  include_action_cost = TRUE,
                                  name = NULL) {

  stopifnot(inherits(x, "Problem"))

  if (missing(sense) || is.null(sense)) {
    stop(
      "`sense` must be explicitly provided. Use one of: 'min', 'max', or 'equal'.",
      call. = FALSE
    )
  }

  sense <- match.arg(sense, c("min", "max", "equal"))

  assertthat::assert_that(
    assertthat::is.number(budget),
    is.finite(budget),
    budget >= 0
  )

  assertthat::assert_that(
    assertthat::is.number(tolerance),
    is.finite(tolerance),
    tolerance >= 0
  )

  assertthat::assert_that(
    is.logical(include_pu_cost),
    length(include_pu_cost) == 1,
    !is.na(include_pu_cost)
  )

  assertthat::assert_that(
    is.logical(include_action_cost),
    length(include_action_cost) == 1,
    !is.na(include_action_cost)
  )

  if (!include_pu_cost && !include_action_cost) {
    stop(
      "At least one of `include_pu_cost` or `include_action_cost` must be TRUE.",
      call. = FALSE
    )
  }

  if (!is.null(actions) && isTRUE(include_pu_cost)) {
    stop(
      "`include_pu_cost = TRUE` is only supported when `actions = NULL`, because planning-unit costs are not action-specific.",
      call. = FALSE
    )
  }

  if (!is.null(name)) {
    if (!is.character(name) || length(name) != 1 || is.na(name) || !nzchar(name)) {
      stop("`name` must be NULL or a non-empty character string.", call. = FALSE)
    }
  }

  if (sense != "equal" && tolerance > 0) {
    warning(
      "`tolerance` is only used when sense = 'equal'. It will be ignored.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  x <- .pa_clone_data(x)

  actions_txt <- NA_character_

  if (!is.null(actions)) {
    act_subset <- .pa_resolve_action_subset(x, subset = actions)

    if (!is.data.frame(act_subset) || nrow(act_subset) == 0) {
      stop("`actions` did not resolve to any valid actions.", call. = FALSE)
    }

    actions_txt <- .pa_subset_to_string(act_subset$id)
  }

  budget_name <- if (is.null(name)) {
    prefix <- if (isTRUE(include_pu_cost) && isTRUE(include_action_cost)) {
      "budget_total"
    } else if (isTRUE(include_pu_cost)) {
      "budget_pu"
    } else {
      "budget_action"
    }

    if (is.na(actions_txt)) {
      paste0(prefix, "_", sense)
    } else {
      paste0(prefix, "_", sense, "_", gsub("\\|", "_", actions_txt))
    }
  } else {
    as.character(name)[1]
  }

  budget_df <- data.frame(
    type = "budget",
    sense = sense,
    value = as.numeric(budget),
    tolerance = if (sense == "equal") as.numeric(tolerance) else 0,
    actions = actions_txt,
    include_pu_cost = include_pu_cost,
    include_action_cost = include_action_cost,
    name = budget_name,
    stringsAsFactors = FALSE
  )

  x <- .pa_store_budget_constraints(x, budget_df)

  if (!is.null(x$data$model_ptr)) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

#' @include internal.R
#'
#' @title Add action effects to a planning problem
#'
#' @description
#' Define the effects of management actions on features across planning units.
#'
#' Effects are stored in a canonical representation in a effects table, with one
#' row per \code{(pu, action, feature)} triple and two non-negative columns:
#' \itemize{
#'   \item \code{benefit}: the positive component of the effect,
#'   \item \code{loss}: the magnitude of the negative component of the effect.
#' }
#'
#' The net effect is therefore interpreted as
#' \deqn{
#' \Delta_{iaf} = \mathrm{benefit}_{iaf} - \mathrm{loss}_{iaf},
#' }
#' where \eqn{i} indexes planning units, \eqn{a} indexes actions, and
#' \eqn{f} indexes features.
#'
#' Under the semantics adopted by this package, each
#' \code{(pu, action, feature)} triple represents a single net effect.
#' Consequently, after validation and aggregation, a stored row cannot have both
#' \code{benefit > 0} and \code{loss > 0} at the same time.
#'
#' @details
#' \strong{When to use \code{add_effects()}.}
#'
#' Use this function when you want to specify what feasible actions do to
#' features. It is the stage at which an action-based decision space is linked
#' to feature-level ecological or functional consequences.
#'
#' This function provides a unified interface for specifying action effects from
#' several input formats while enforcing a single internal representation.
#' Regardless of how the user supplies the effects, the stored output always
#' follows the same canonical structure based on non-negative
#' \code{benefit}/\code{loss} components.
#'
#' Let \eqn{i \in \mathcal{I}} index planning units,
#' \eqn{a \in \mathcal{A}} index actions, and
#' \eqn{f \in \mathcal{F}} index features.
#' Let \eqn{b_{if}} denote the baseline amount of feature \eqn{f} in planning
#' unit \eqn{i}, as given by the feature-distribution table. Let
#' \eqn{\Delta_{iaf}} denote the net change caused by applying action
#' \eqn{a} in planning unit \eqn{i} to feature \eqn{f}. The canonical stored
#' representation is:
#'
#' \deqn{
#' \mathrm{benefit}_{iaf} = \max(\Delta_{iaf}, 0),
#' }
#'
#' \deqn{
#' \mathrm{loss}_{iaf} = \max(-\Delta_{iaf}, 0).
#' }
#'
#' Hence:
#' \itemize{
#'   \item if \eqn{\Delta_{iaf} > 0}, then \code{benefit > 0} and \code{loss = 0},
#'   \item if \eqn{\Delta_{iaf} < 0}, then \code{benefit = 0} and \code{loss > 0},
#'   \item if \eqn{\Delta_{iaf} = 0}, then both are zero.
#' }
#'
#' \strong{Why split effects into benefit and loss?}
#'
#' This representation avoids ambiguity in downstream optimization models. It
#' allows the package to support, for example, objectives that maximize
#' beneficial effects, minimize damages, impose no-net-loss conditions, or
#' combine both components differently in multi-objective formulations.
#'
#' \strong{Supported effect specifications}
#'
#' The \code{effects} argument may be provided in one of the following forms:
#'
#' \enumerate{
#'   \item \code{NULL}. An empty effects table is stored.
#'
#'   \item A \code{data.frame(action, feature, multiplier)}. In this case,
#'   effects are constructed by multiplying baseline feature amounts by the
#'   supplied multiplier:
#'   \deqn{
#'   \Delta_{iaf} = b_{if} \times m_{af},
#'   }
#'   where \eqn{m_{af}} is the multiplier associated with action \eqn{a} and
#'   feature \eqn{f}. This specification is expanded over all feasible
#'   \code{(pu, action)} pairs.
#'
#'   \item A \code{data.frame(pu, action, feature, ...)} giving explicit effects
#'   for individual triples. The table may contain:
#'   \itemize{
#'     \item \code{delta} or \code{effect}: interpreted as signed net changes,
#'     \item \code{after}: interpreted as after-action amounts if
#'     \code{effect_type = "after"},
#'     \item \code{benefit} and/or \code{loss}: explicit non-negative split
#'     components,
#'     \item legacy signed \code{benefit} without \code{loss}: interpreted as a
#'     signed net effect for backwards compatibility.
#'   }
#'
#'   \item A named list of \code{terra::SpatRaster} objects, one per action. In
#'   this case, names must match action ids, and each raster must contain one
#'   layer per feature. Raster values are aggregated to planning-unit level
#'   using \code{effect_aggregation}.
#' }
#'
#' \strong{Interpretation of \code{effect_type}}
#'
#' If \code{effect_type = "delta"}, supplied values are interpreted as net
#' changes directly.
#'
#' If \code{effect_type = "after"}, supplied values are interpreted as
#' after-action amounts and converted internally to net effects using:
#'
#' \deqn{
#' \Delta_{iaf} = \mathrm{after}_{iaf} - b_{if}.
#' }
#'
#' Missing baseline values are treated as zero.
#'
#' \strong{Feasibility and locked-out decisions}
#'
#' Effects are only retained for feasible \code{(pu, action)} pairs. Thus,
#' \code{add_actions()} must be called first. Pairs marked as locked out
#' (\code{status == 3}) are removed before storing the final effects table.
#'
#' This function does not define the action-decision layer itself; it builds on
#' the feasible \code{(pu, action)} pairs already stored in the problem.
#'
#' \strong{Duplicate rows and semantic validation}
#'
#' If multiple rows are supplied for the same \code{(pu, action, feature)}
#' triple, they are aggregated by summing \code{benefit} and \code{loss}
#' separately. The resulting triple must still respect the package semantics,
#' namely that both components cannot be strictly positive simultaneously.
#' Inputs violating this rule are rejected.
#'
#' \strong{Component filtering}
#'
#' After canonicalization and validation, rows can be restricted to:
#' \itemize{
#'   \item \code{component = "any"}: keep all non-zero effects,
#'   \item \code{component = "benefit"}: keep only rows with \code{benefit > 0},
#'   \item \code{component = "loss"}: keep only rows with \code{loss > 0}.
#' }
#'
#' Rows with \code{benefit = 0} and \code{loss = 0} are always removed.
#'
#' \strong{Raster handling}
#'
#' When effects are supplied as rasters, they are automatically aligned to the
#' planning-unit raster or geometry when needed before extraction or zonal
#' aggregation.
#'
#' \strong{Stored output}
#'
#' The resulting effects table contains user-facing ids, internal integer ids,
#' and optional labels for actions and features. Metadata describing the stored
#' representation and input interpretation are written to an effects metadata
#' field.
#'
#' After defining effects, typical next steps include adding objectives that use
#' beneficial or harmful effects, and then solving the configured problem.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}}. It
#'   must already contain feasible actions; run \code{\link{add_actions}} first.
#'
#' @param effects Effect specification. One of:
#' \itemize{
#'   \item \code{NULL}, to store an empty effects table,
#'   \item a \code{data.frame(action, feature, multiplier)},
#'   \item a \code{data.frame(pu, action, feature, ...)} with explicit effects,
#'   \item a named list of \code{terra::SpatRaster} objects, one per action.
#' }
#'
#' @param effect_type Character string indicating how supplied effect values are
#'   interpreted. Must be one of:
#'   \itemize{
#'     \item \code{"delta"}: values represent signed net changes,
#'     \item \code{"after"}: values represent after-action amounts and are
#'     converted to net changes relative to baseline feature amounts.
#'   }
#'
#' @param effect_aggregation Character string giving the aggregation used when
#'   converting raster values to planning-unit level. Must be one of
#'   \code{"sum"} or \code{"mean"}.
#'
#' @param component Character string controlling which component of the
#'   canonical effects table is retained. Must be one of:
#'   \itemize{
#'     \item \code{"any"}: keep all non-zero rows,
#'     \item \code{"benefit"}: keep only rows with \code{benefit > 0},
#'     \item \code{"loss"}: keep only rows with \code{loss > 0}.
#'   }
#'
#' @return An updated \code{Problem} object containing:
#' \describe{
#'   \item{\code{dist_effects}}{A canonical effects table with columns
#'   \code{pu}, \code{action}, \code{feature}, \code{benefit}, \code{loss},
#'   \code{internal_pu}, \code{internal_action}, \code{internal_feature}, and
#'   optional labels such as \code{feature_name} and \code{action_name}.}
#'   \item{\code{effects_meta}}{Metadata describing how effects were
#'   interpreted and stored.}
#' }
#'
#' @examples
#' # ------------------------------------------------------
#' # Minimal problem with actions
#' # ------------------------------------------------------
#' pu <- data.frame(
#'   id = 1:3,
#'   cost = c(1, 2, 3)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3),
#'   feature = c(1, 2, 1, 2),
#'   amount = c(10, 5, 8, 4)
#' )
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' actions <- data.frame(
#'   id = c("conservation", "restoration"),
#'   name = c("Conservation", "Restoration")
#' )
#'
#' p <- add_actions(
#'   x = p,
#'   actions = actions,
#'   cost = c(conservation = 2, restoration = 4)
#' )
#'
#' print(p)
#'
#' # ------------------------------------------------------
#' # 1) Empty effects
#' # ------------------------------------------------------
#' p0 <- add_effects(p, effects = NULL)
#' p0$data$dist_effects
#'
#' # ------------------------------------------------------
#' # 2) Multipliers by action and feature
#' # delta = baseline amount * multiplier
#' # ------------------------------------------------------
#' mult <- data.frame(
#'   action = c("conservation", "conservation",
#'              "restoration", "restoration"),
#'   feature = c("sp1", "sp2", "sp1", "sp2"),
#'   multiplier = c(0.10, 0.00, -0.20, 0.25)
#' )
#'
#' p1 <- add_effects(
#'   x = p,
#'   effects = mult,
#'   effect_type = "delta"
#' )
#'
#' p1$data$dist_effects
#'
#' # ------------------------------------------------------
#' # 3) Explicit net effects using signed deltas
#' # ------------------------------------------------------
#' eff_delta <- data.frame(
#'   pu = c(1, 2, 3, 3),
#'   action = c("conservation", "restoration", "restoration", "conservation"),
#'   feature = c(1, 1, 2, 2),
#'   delta = c(2, -3, 1, 0.5)
#' )
#'
#' p2 <- add_effects(
#'   x = p,
#'   effects = eff_delta
#' )
#'
#' p2$data$dist_effects
#'
#' # ------------------------------------------------------
#' # 4) Explicit after-action amounts
#' # delta = after - baseline
#' # ------------------------------------------------------
#' eff_after <- data.frame(
#'   pu = c(1, 2, 3),
#'   action = c("conservation", "restoration", "restoration"),
#'   feature = c(1, 1, 2),
#'   after = c(12, 5, 6)
#' )
#'
#' p3 <- add_effects(
#'   x = p,
#'   effects = eff_after,
#'   effect_type = "after"
#' )
#'
#' p3$data$dist_effects
#'
#' # ------------------------------------------------------
#' # 5) Keep only beneficial effects
#' # ------------------------------------------------------
#' p4 <- add_effects(
#'   x = p,
#'   effects = eff_delta,
#'   component = "benefit"
#' )
#'
#' p4$data$dist_effects
#'
#' @seealso
#' \code{\link{add_actions}},
#' \code{\link{add_benefits}},
#' \code{\link{add_losses}}
#'
#' @export
add_effects <- function(
    x,
    effects = NULL,
    effect_type = c("delta", "after"),
    effect_aggregation = c("sum", "mean"),
    component = c("any", "benefit", "loss")
) {

  effect_type <- match.arg(effect_type)
  effect_aggregation <- match.arg(effect_aggregation)
  component <- match.arg(component)

  align_rasters <- TRUE
  na_to_zero <- TRUE

  # ---- checks: x
  assertthat::assert_that(!is.null(x), msg = "x is NULL")
  assertthat::assert_that(!is.null(x$data), msg = "x does not look like a multiscape Problem object")
  assertthat::assert_that(
    !is.null(x$data$pu), !is.null(x$data$features), !is.null(x$data$dist_features),
    msg = "x must be created with create_problem()"
  )
  assertthat::assert_that(!is.null(x$data$dist_actions), msg = "No actions found. Run add_actions() first.")

  x <- .pa_clone_data(x)

  pu    <- x$data$pu
  feats <- x$data$features
  df    <- x$data$dist_features
  da    <- x$data$dist_actions
  acts  <- x$data$actions

  # required columns
  assertthat::assert_that(all(c("id", "internal_id") %in% names(pu)))
  assertthat::assert_that(all(c("id", "internal_id") %in% names(feats)))
  assertthat::assert_that(all(c("id") %in% names(acts)))
  assertthat::assert_that(all(c("pu", "feature", "amount") %in% names(df)))
  assertthat::assert_that(all(c("pu", "action", "cost") %in% names(da)))

  # defensive: enforce action internal_id
  if (!("internal_id" %in% names(acts))) {
    acts$internal_id <- seq_len(nrow(acts))
    x$data$actions <- acts
    acts <- x$data$actions
  }

  # ---- defensive coercions (consistent ids)
  pu$id    <- as.integer(pu$id)
  feats$id <- as.integer(feats$id)

  pu_ids     <- pu$id
  feat_ids   <- feats$id
  feat_names <- if ("name" %in% names(feats)) as.character(feats$name) else paste0("feature.", feat_ids)
  action_ids <- as.character(acts$id)

  # ---- helper: normalize feature column (id OR name)
  .normalize_feature <- function(feature_col, feats_df) {
    if (is.null(feature_col)) return(feature_col)

    if (is.factor(feature_col)) feature_col <- as.character(feature_col)

    if (is.character(feature_col)) {
      if (!("name" %in% names(feats_df))) {
        stop("features have no 'name' column, cannot match features by name.", call. = FALSE)
      }
      m <- match(feature_col, as.character(feats_df$name))
      bad <- unique(feature_col[is.na(m)])
      if (length(bad) > 0) {
        stop(
          "Unknown feature name(s) in effects$feature: ",
          paste0("'", bad, "'", collapse = ", "),
          ". Valid names are: ",
          paste0("'", as.character(feats_df$name), "'", collapse = ", "),
          call. = FALSE
        )
      }
      return(as.integer(feats_df$id[m]))
    }

    if (is.numeric(feature_col) || is.integer(feature_col)) {
      feature_col <- as.integer(feature_col)
      bad <- unique(feature_col[!feature_col %in% feats_df$id])
      if (length(bad) > 0) {
        stop(
          "Unknown feature id(s) in effects$feature: ",
          paste(bad, collapse = ", "),
          ". Valid ids are: ",
          paste(feats_df$id, collapse = ", "),
          call. = FALSE
        )
      }
      return(feature_col)
    }

    stop("effects$feature must be either numeric ids or character feature names.", call. = FALSE)
  }

  # ---- baseline lookup for (pu, feature) -> amount (missing treated as 0)
  df$pu      <- as.integer(df$pu)
  df$feature <- as.integer(df$feature)
  df$amount  <- as.numeric(df$amount)

  base_key <- paste(df$pu, df$feature, sep = "||")
  base_amt <- df$amount
  names(base_amt) <- base_key

  .baseline_amount <- function(pu_vec, feat_vec) {
    k <- paste(as.integer(pu_vec), as.integer(feat_vec), sep = "||")
    out <- unname(base_amt[k])
    out[is.na(out)] <- 0
    out
  }

  # ---- helper: split signed delta into benefit/loss (both >=0)
  .split_delta <- function(delta) {
    delta <- as.numeric(delta)
    delta[is.na(delta)] <- 0
    benefit <- pmax(delta, 0)
    loss    <- pmax(-delta, 0)
    list(benefit = benefit, loss = loss)
  }

  # ---- helper: validate split effects semantics
  .validate_split_effects <- function(tbl, context = "effects") {
    if (!("benefit" %in% names(tbl)) || !("loss" %in% names(tbl))) {
      stop("Internal error: .validate_split_effects() requires 'benefit' and 'loss' columns.", call. = FALSE)
    }

    tbl$benefit <- as.numeric(tbl$benefit)
    tbl$loss    <- as.numeric(tbl$loss)

    tbl$benefit[is.na(tbl$benefit)] <- 0
    tbl$loss[is.na(tbl$loss)] <- 0

    if (any(tbl$benefit < 0, na.rm = TRUE) || any(tbl$loss < 0, na.rm = TRUE)) {
      stop(
        context,
        ": 'benefit' and 'loss' must be non-negative.",
        call. = FALSE
      )
    }

    bad <- which(tbl$benefit > 0 & tbl$loss > 0)
    if (length(bad) > 0) {
      ex <- tbl[bad[1], intersect(c("pu", "action", "feature", "benefit", "loss"), names(tbl)), drop = FALSE]
      msg <- paste0(
        context,
        ": a single (pu, action, feature) effect cannot have both positive 'benefit' and positive 'loss'."
      )
      if (nrow(ex) == 1) {
        msg <- paste0(
          msg,
          " Example offending row -> pu=", ex$pu,
          ", action='", ex$action,
          "', feature=", ex$feature,
          ", benefit=", ex$benefit,
          ", loss=", ex$loss, "."
        )
      }
      stop(msg, call. = FALSE)
    }

    tbl
  }

  # always drop locked out actions
  if ("status" %in% names(da)) {
    da <- da[da$status != 3L, , drop = FALSE]
    if (nrow(da) == 0) stop("All (pu, action) pairs are locked_out (status=3).", call. = FALSE)
  }

  # helper: align raster to a template
  .align_to <- function(r, template) {
    if (!is.na(terra::crs(r)) && !is.na(terra::crs(template)) && terra::crs(r) != terra::crs(template)) {
      r <- terra::project(r, template)
    }
    if (!terra::compareGeom(r, template, stopOnError = FALSE)) {
      r <- terra::resample(r, template)
    }
    r
  }

  # helper: build effects from action rasters using zonal or extract
  .effects_from_rasters <- function(x, effects_list) {

    if (!requireNamespace("terra", quietly = TRUE)) {
      stop("Raster effects require the 'terra' package.", call. = FALSE)
    }

    has_pu_raster <- !is.null(x$data$pu_raster_id) && inherits(x$data$pu_raster_id, "SpatRaster")
    has_pu_sf     <- !is.null(x$data$pu_sf) && inherits(x$data$pu_sf, "sf")

    if (!has_pu_raster && !has_pu_sf) {
      stop(
        "To use raster effects, the object must contain either x$data$pu_raster_id (SpatRaster) or x$data$pu_sf (sf).",
        call. = FALSE
      )
    }

    if (is.null(names(effects_list)) || any(names(effects_list) == "")) {
      stop("If effects is a list of rasters, it must be a named list with names = action ids.", call. = FALSE)
    }
    if (!all(names(effects_list) %in% action_ids)) {
      bad <- setdiff(names(effects_list), action_ids)
      stop("effects list contains unknown action ids: ", paste(bad, collapse = ", "), call. = FALSE)
    }

    # baseline dense matrix once for after->delta
    baseline_mat <- NULL
    if (identical(effect_type, "after")) {
      baseline_mat <- matrix(0, nrow = length(pu_ids), ncol = length(feat_ids))
      pu_pos <- match(df$pu, pu_ids)
      ft_pos <- match(df$feature, feat_ids)
      ok <- !(is.na(pu_pos) | is.na(ft_pos))
      if (any(ok)) {
        idx <- (ft_pos[ok] - 1L) * length(pu_ids) + pu_pos[ok]
        baseline_mat[idx] <- df$amount[ok]
      }
    }

    out_list <- vector("list", length(effects_list))
    k <- 0L

    for (a in names(effects_list)) {
      r <- effects_list[[a]]
      if (is.null(r)) next
      if (!inherits(r, "SpatRaster")) stop("effects[['", a, "']] must be a terra::SpatRaster.", call. = FALSE)

      if (terra::nlyr(r) != nrow(feats)) {
        stop(
          "effects[['", a, "']] has ", terra::nlyr(r), " layers but x$data$features has ", nrow(feats), " features. ",
          "Provide one layer per feature.",
          call. = FALSE
        )
      }

      try(names(r) <- feat_names, silent = TRUE)

      if (has_pu_raster) {
        z <- x$data$pu_raster_id
        r2 <- .align_to(r, z)
        zb <- terra::zonal(r2, z, fun = effect_aggregation, na.rm = TRUE)
        zb <- zb[match(pu_ids, zb[[1]]), , drop = FALSE]
        mat <- as.matrix(zb[, -1, drop = FALSE])
      } else {
        pu_sf <- x$data$pu_sf
        pu_v  <- terra::vect(pu_sf)
        fun <- switch(
          effect_aggregation,
          sum  = function(v) sum(v, na.rm = TRUE),
          mean = function(v) mean(v, na.rm = TRUE)
        )
        ex <- terra::extract(r, pu_v, fun = fun, na.rm = TRUE)
        ex <- ex[match(pu_ids, ex[[1]]), , drop = FALSE]
        mat <- as.matrix(ex[, -1, drop = FALSE])
      }

      # convert after->delta if needed (delta can be signed)
      if (identical(effect_type, "after")) {
        mat <- mat - baseline_mat
      }

      # flatten signed delta
      delta_vec <- as.vector(t(mat))
      delta_vec[is.na(delta_vec)] <- 0

      sp <- .split_delta(delta_vec)

      k <- k + 1L
      out_list[[k]] <- data.frame(
        pu      = rep(pu_ids, times = ncol(mat)),
        action  = rep(a,    times = length(pu_ids) * ncol(mat)),
        feature = rep(feat_ids, each = length(pu_ids)),
        benefit = sp$benefit,
        loss    = sp$loss,
        stringsAsFactors = FALSE
      )
    }

    out_list <- out_list[seq_len(k)]
    out <- dplyr::bind_rows(out_list)

    # keep only feasible (pu, action)
    out <- dplyr::inner_join(
      out,
      da[, c("pu", "action"), drop = FALSE],
      by = c("pu", "action")
    )

    out
  }

  # ---- compute effects depending on 'effects'
  if (is.list(effects) && !inherits(effects, "data.frame")) {

    base <- .effects_from_rasters(x, effects)

  } else if (is.null(effects)) {

    # recommended default: empty table
    base <- da[0, c("pu", "action"), drop = FALSE]
    base$feature <- integer(0)
    base$benefit <- numeric(0)
    base$loss <- numeric(0)

  } else if (inherits(effects, "data.frame")) {

    b <- effects
    if ("id" %in% names(b) && !("action" %in% names(b))) names(b)[names(b) == "id"] <- "action"
    if ("action" %in% names(b)) b$action <- as.character(b$action)
    if ("feature" %in% names(b)) b$feature <- .normalize_feature(b$feature, feats)

    # accept delta/effect as signed values
    if ("effect" %in% names(b) && !("delta" %in% names(b))) b$delta <- b$effect

    # Case A: (action, feature, multiplier) => signed delta = amount * multiplier
    if (all(c("action", "feature", "multiplier") %in% names(b)) &&
        !("pu" %in% names(b)) &&
        !any(c("delta","benefit","loss","after") %in% names(b))) {

      b$multiplier <- as.numeric(b$multiplier)

      assertthat::assert_that(assertthat::noNA(b$action), assertthat::noNA(b$feature), assertthat::noNA(b$multiplier))
      assertthat::assert_that(all(b$action %in% action_ids), msg = "Unknown action id(s) in effects.")
      assertthat::assert_that(all(b$feature %in% feat_ids), msg = "Unknown feature id(s) in effects.")

      df2 <- df[, c("pu", "feature", "amount"), drop = FALSE]
      tmp <- dplyr::inner_join(da[, c("pu","action"), drop = FALSE], df2, by = "pu", relationship = "many-to-many")
      if (nrow(tmp) == 0) stop("No (pu, action, feature) triples were created. Check dist_actions/dist_features.", call. = FALSE)

      tmp <- dplyr::left_join(tmp, b, by = c("action","feature"))
      tmp$multiplier[is.na(tmp$multiplier)] <- 0

      delta <- as.numeric(tmp$amount) * as.numeric(tmp$multiplier)
      sp <- .split_delta(delta)

      base <- tmp[, c("pu","action","feature")]
      base$benefit <- sp$benefit
      base$loss    <- sp$loss

    } else {

      # Case B: explicit rows (pu, action, feature, ...)
      assertthat::assert_that(all(c("pu","action","feature") %in% names(b)))

      b$pu <- as.integer(b$pu)
      b$feature <- as.integer(b$feature)

      assertthat::assert_that(assertthat::noNA(b$pu), assertthat::noNA(b$action), assertthat::noNA(b$feature))
      assertthat::assert_that(all(b$pu %in% pu_ids), msg = "Unknown pu id(s) in effects.")
      assertthat::assert_that(all(b$action %in% action_ids), msg = "Unknown action id(s) in effects.")
      assertthat::assert_that(all(b$feature %in% feat_ids), msg = "Unknown feature id(s) in effects.")

      tmp <- dplyr::inner_join(
        b,
        da[, c("pu","action"), drop = FALSE],
        by = c("pu","action")
      )
      if (nrow(tmp) == 0) stop("No rows in effects match feasible (pu, action) pairs.", call. = FALSE)

      # Input mode:
      # 1) already split benefit/loss (non-negative)
      # 2) signed delta (delta/effect/legacy benefit)
      has_any_split <- any(c("benefit","loss") %in% names(tmp))

      if (has_any_split && !("delta" %in% names(tmp)) && !("after" %in% names(tmp)) && !("effect" %in% names(tmp))) {
        # split-only path
        if (!("benefit" %in% names(tmp))) tmp$benefit <- 0
        if (!("loss" %in% names(tmp)))    tmp$loss <- 0

        tmp <- .validate_split_effects(
          tmp,
          context = "When providing explicit benefit/loss columns"
        )

        base <- tmp[, c("pu","action","feature","benefit","loss")]

      } else {
        # signed path (delta/effect/after or legacy benefit)
        if (!("delta" %in% names(tmp))) {
          if ("effect" %in% names(tmp)) tmp$delta <- tmp$effect
          else if ("after" %in% names(tmp)) tmp$delta <- tmp$after
          else if ("benefit" %in% names(tmp) && !("loss" %in% names(tmp))) tmp$delta <- tmp$benefit
          else stop("effects data.frame must include 'delta'/'effect'/'after' (signed) or 'benefit/loss' (non-negative).", call. = FALSE)
        }

        tmp$delta <- as.numeric(tmp$delta)
        tmp$delta[is.na(tmp$delta)] <- 0

        if (identical(effect_type, "after")) {
          base_amount <- .baseline_amount(tmp$pu, tmp$feature)
          tmp$delta <- tmp$delta - base_amount
        }

        sp <- .split_delta(tmp$delta)
        base <- tmp[, c("pu","action","feature")]
        base$benefit <- sp$benefit
        base$loss    <- sp$loss
      }
    }

  } else {
    stop("Unsupported type for 'effects'. Use NULL, a data.frame, or a named list of SpatRaster.", call. = FALSE)
  }

  # ---- aggregate duplicates and enforce one net effect per (pu, action, feature)
  if (nrow(base) > 0) {
    base <- stats::aggregate(
      cbind(benefit, loss) ~ pu + action + feature,
      data = base,
      FUN = sum
    )
  }

  # ---- cleanup / validation / filtering
  base$pu      <- as.integer(base$pu)
  base$feature <- as.integer(base$feature)
  base$benefit <- as.numeric(base$benefit)
  base$loss    <- as.numeric(base$loss)

  base$benefit[is.na(base$benefit)] <- 0
  base$loss[is.na(base$loss)] <- 0

  base <- .validate_split_effects(base, context = "Validated effects")

  if (identical(component, "benefit")) base <- base[base$benefit > 0, , drop = FALSE]
  if (identical(component, "loss"))    base <- base[base$loss > 0, , drop = FALSE]

  # always drop zero rows
  base <- base[!(base$benefit == 0 & base$loss == 0), , drop = FALSE]
  if (nrow(base) == 0) warning("All effect values are zero after filtering.", call. = FALSE, immediate. = TRUE)

  # ---- add internal ids
  pu_map    <- pu[, c("id", "internal_id")]
  feats_map <- feats[, c("id", "internal_id")]
  acts_map  <- x$data$actions[, c("id", "internal_id")]

  base$internal_pu      <- pu_map$internal_id[match(base$pu, pu_map$id)]
  base$internal_feature <- feats_map$internal_id[match(base$feature, feats_map$id)]
  base$internal_action  <- acts_map$internal_id[match(base$action, acts_map$id)]

  dist_effects <- base[, c("pu","action","feature","benefit","loss",
                           "internal_pu","internal_action","internal_feature"), drop = FALSE]

  dist_effects <- .pa_add_feature_labels(
    df = dist_effects,
    features_df = feats,
    feature_col = "feature",
    internal_feature_col = "internal_feature",
    out_col = "feature_name"
  )

  dist_effects <- .pa_add_action_labels(
    df = dist_effects,
    actions_df = acts,
    action_col = "action",
    internal_action_col = "internal_action",
    out_col = "action_name"
  )

  x$data$dist_effects <- dist_effects

  x$data$effects_meta <- list(
    stored_as = "benefit_loss",
    input_interpretation = effect_type,
    component = component
  )

  x
}

#' @title Add benefits
#'
#' @description
#' Convenience wrapper around \code{\link{add_effects}} that keeps only positive
#' effects, that is, rows with \code{benefit > 0}.
#'
#' This function is useful when the workflow is focused only on beneficial
#' consequences of actions and a gain-only wrapper is more convenient than
#' calling \code{add_effects(..., component = "benefit")} directly.  Internally,
#' it calls \code{add_effects()} with \code{component = "benefit"}.
#' The canonical stored result therefore contains only rows whose effect
#' satisfies \eqn{\mathrm{benefit}_{iaf} > 0}.
#'
#' For backwards compatibility, a mirror table containing only the benefit
#' component is also written to \code{x$data$dist_benefit}. The canonical
#' filtered representation remains the one stored by \code{add_effects()}.
#'
#' @inheritParams add_effects
#' @param benefits Alias of \code{effects}, kept for backwards compatibility.
#'
#' @return An updated \code{Problem} object containing:
#' \describe{
#'   \item{\code{dist_effects}}{The canonical filtered effects table, containing
#'   only rows with \code{benefit > 0}.}
#'   \item{\code{dist_benefit}}{A backwards-compatible mirror table containing
#'   only the benefit component.}
#' }
#'
#' @examples
#' pu <- data.frame(id = 1:2, cost = c(1, 2))
#' features <- data.frame(id = 1, name = "sp1")
#' dist_features <- data.frame(pu = 1:2, feature = 1, amount = c(5, 10))
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   p,
#'   data.frame(id = "restoration")
#' )
#'
#' eff <- data.frame(
#'   pu = c(1, 2),
#'   action = c("restoration", "restoration"),
#'   feature = c(1, 1),
#'   delta = c(2, -1)
#' )
#'
#' p <- add_benefits(p, benefits = eff)
#' p$data$dist_benefit
#'
#' @seealso
#' \code{\link{add_effects}},
#' \code{\link{add_losses}},
#' \code{\link{add_objective_max_benefit}}
#'
#' @export
add_benefits <- function(
    x,
    benefits = NULL,
    effect_type = c("delta", "after"),
    effect_aggregation = c("sum", "mean")
) {
  effects <- benefits

  x <- add_effects(
    x = x,
    effects = effects,
    effect_type = effect_type,
    effect_aggregation = effect_aggregation,
    component = "benefit"
  )

  # backward-compatible mirror: dist_benefit (ONLY benefit rows/col)
  if (!is.null(x$data$dist_effects) && inherits(x$data$dist_effects, "data.frame")) {
    db <- x$data$dist_effects
    if ("loss" %in% names(db)) db$loss <- NULL
    x$data$dist_benefit <- db
  } else {
    x$data$dist_benefit <- x$data$dist_effects
  }

  x
}

#' @title Add losses
#'
#' @description
#' Convenience wrapper around \code{\link{add_effects}} that keeps only negative
#' effects, represented by rows with \code{loss > 0}.
#'
#' This function is useful when the workflow is focused only on damaging
#' consequences of actions and a loss-only wrapper is more convenient than
#' calling \code{add_effects(..., component = "loss")} directly. Internally,
#' it calls \code{add_effects()} with \code{component = "loss"}.
#' The canonical stored result therefore contains only rows whose effect
#' satisfies \eqn{\mathrm{loss}_{iaf} > 0}.
#'
#' In addition, a mirror table containing only the loss component is stored in
#' \code{x$data$dist_loss}.
#'
#' @inheritParams add_effects
#' @param losses Alias of \code{effects}, used for symmetry with
#'   \code{add_benefits()}.
#'
#' @return An updated \code{Problem} object containing:
#' \describe{
#'   \item{\code{dist_effects}}{The canonical filtered effects table,
#'   containing only rows with \code{loss > 0}.}
#'   \item{\code{dist_loss}}{A convenience table containing only the loss
#'   component.}
#'   \item{\code{losses_meta}}{Metadata for the stored loss table.}
#' }
#'
#' @examples
#' pu <- data.frame(id = 1:2, cost = c(1, 2))
#' features <- data.frame(id = 1, name = "sp1")
#' dist_features <- data.frame(pu = 1:2, feature = 1, amount = c(5, 10))
#'
#' p <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' )
#'
#' p <- add_actions(
#'   p,
#'   data.frame(id = "harvest")
#' )
#'
#' eff <- data.frame(
#'   pu = c(1, 2),
#'   action = c("harvest", "harvest"),
#'   feature = c(1, 1),
#'   delta = c(2, -4)
#' )
#'
#' p <- add_losses(p, losses = eff)
#' p$data$dist_loss
#'
#' @seealso
#' \code{\link{add_effects}},
#' \code{\link{add_benefits}},
#' \code{\link{add_objective_min_loss}}
#'
#' @export
add_losses <- function(
    x,
    losses = NULL,
    effect_type = c("delta", "after"),
    effect_aggregation = c("sum", "mean")
) {
  effects <- losses

  x <- add_effects(
    x = x,
    effects = effects,
    effect_type = effect_type,
    effect_aggregation = effect_aggregation,
    component = "loss"
  )

  if (!is.null(x$data$dist_effects) && inherits(x$data$dist_effects, "data.frame")) {
    dl <- x$data$dist_effects
    if ("benefit" %in% names(dl)) dl$benefit <- NULL
    x$data$dist_loss <- dl
    x$data$losses_meta <- list(
      stored_as = "loss",
      input_interpretation = x$data$effects_meta$input_interpretation
    )
  }

  x
}

toy_equivalent_basic <- function() {
  pu <- data.frame(
    id = 1:4,
    cost = c(10, 12, 30, 40),
    locked_in = c(0, 0, 0, 0),
    locked_out = c(0, 0, 0, 0)
  )

  features <- data.frame(
    id = 1:2,
    name = c("f1", "f2")
  )

  dist_features <- data.frame(
    pu = c(1, 2, 3, 4, 1, 2, 3, 4),
    feature = c(1, 1, 1, 1, 2, 2, 2, 2),
    amount = c(8, 6, 3, 1, 1, 2, 8, 7)
  )

  actions <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  effects <- expand.grid(
    action = "conservation",
    feature = features$id,
    stringsAsFactors = FALSE
  )
  effects$multiplier <- 1

  boundary <- data.frame(
    pu1 = c(1, 1, 2, 3),
    pu2 = c(2, 3, 4, 4),
    boundary = c(1, 1, 1, 1)
  )

  list(
    pu = pu,
    features = features,
    dist_features = dist_features,
    actions = actions,
    effects = effects,
    boundary = boundary
  )
}

toy_multiaction_semantics <- function() {
  pu <- data.frame(
    id = 1:5,
    cost = c(5, 6, 8, 9, 20),
    locked_in = c(0, 0, 0, 0, 0),
    locked_out = c(0, 0, 0, 0, 0)
  )

  features <- data.frame(
    id = 1:2,
    name = c("common", "restoration_only")
  )

  dist_features <- data.frame(
    pu = c(1, 2, 3, 4, 5, 1, 2, 3, 4, 5),
    feature = c(1, 1, 1, 1, 1, 2, 2, 2, 2, 2),
    amount = c(5, 5, 5, 5, 5, 0.5, 0.5, 0.5, 0.5, 0.5)
  )

  actions <- data.frame(
    id = c("conservation", "restoration"),
    name = c("conservation", "restoration")
  )

  # conservation mejora solo feature 1; restoration mejora mucho feature 2
  effects <- data.frame(
    action = c(
      "conservation", "conservation",
      "restoration",  "restoration"
    ),
    feature = c(1, 2, 1, 2),
    multiplier = c(0.2, 0, 0.1, 2)
  )

  boundary <- data.frame(
    pu1 = c(1, 2, 3, 4),
    pu2 = c(2, 3, 4, 5),
    boundary = c(1, 1, 1, 1)
  )

  list(
    pu = pu,
    features = features,
    dist_features = dist_features,
    actions = actions,
    effects = effects,
    boundary = boundary
  )
}


toy_boundary_matrix <- function(toy) {
  ids <- sort(unique(c(toy$boundary$pu1, toy$boundary$pu2)))
  n <- max(ids)

  mat <- Matrix::Matrix(0, nrow = n, ncol = n, sparse = TRUE)

  for (i in seq_len(nrow(toy$boundary))) {
    a <- toy$boundary$pu1[i]
    b <- toy$boundary$pu2[i]
    w <- toy$boundary$boundary[i]

    mat[a, b] <- w
    mat[b, a] <- w
  }

  mat
}

test_that("create_problem builds the core problem tables correctly", {
  toy <- toy_equivalent_basic()

  p <- multiscape::create_problem(
    pu = toy$pu,
    features = toy$features,
    dist_features = toy$dist_features,
    cost = "cost"
  )

  expect_s3_class(p, "Problem")

  expect_true(is.data.frame(p$data$pu))
  expect_true(is.data.frame(p$data$features))
  expect_true(is.data.frame(p$data$dist_features))

  expect_equal(nrow(p$data$pu), 4)
  expect_equal(nrow(p$data$features), 2)
  expect_equal(nrow(p$data$dist_features), 8)

  expect_true(all(c("id", "cost") %in% names(p$data$pu)))
  expect_true("id" %in% names(p$data$features))
  expect_true(all(c("pu", "feature", "amount") %in% names(p$data$dist_features)))

  expect_true("internal_id" %in% names(p$data$pu))
  expect_true("internal_id" %in% names(p$data$features))

  expect_identical(anyDuplicated(p$data$pu$id), 0L)
  expect_identical(anyDuplicated(p$data$features$id), 0L)
})

test_that("create_problem fails if required feature distribution columns are missing", {
  toy <- toy_equivalent_basic()
  bad_dist <- toy$dist_features
  bad_dist$amount <- NULL

  expect_error(
    multiscape::create_problem(
      pu = toy$pu,
      features = toy$features,
      dist_features = bad_dist,
      cost = "cost"
    )
  )
})

#' No extra arguments
#'
#' Check that no additional unused arguments have been supplied to a function
#' through the `...`.
#'
#' @param ... arguments that are not used.
#'
#' @return `logical` indicating success.
#'
#' @noRd
no_extra_arguments <- function(...) {
  return(length(list(...)) == 0)
}

assertthat::on_failure(no_extra_arguments) <- function(call, env) {
  "unused arguments"
}

#' Verify if assertion is met
#'
#' Verify if an assertion is met and throw a [base::warning()] if it
#' is not. This function is equivalent to [assertthat::assert_that()]
#' except that it throws warnings and not errors.
#'
#' @param x `logical` condition.
#'
#' @return `logical` if assertion is met and a `warning` if it is not.
#'
#' @noRd
verify_that <- function(..., env = parent.frame()) {
  res <- assertthat::validate_that(..., env = env)
  if (isTRUE(res)) {
    return(TRUE)
  }
  warning(res, immediate. = TRUE)
  FALSE
}

#' Atomic representation
#'
#' Return a pretty character representation of an object with elements and
#  names.
#'
#' @param x `object`.
#'
#' @return `character` object.
#'
#' @examples
#' repr_atomic(letters)
#' repr_atomic(letters, "characters")
#' @noRd
repr_atomic <- function(x, description = "") {
  n <- length(x)
  if (nchar(description) > 0) {
    description <- paste0(" ", description)
  }
  if (length(x) <= 4) {
    x <- x[seq_len(min(length(x), 4))]
  } else {
    x <- c(x[seq_len(min(length(x), 3))], "...")
  }
  paste0(paste(x, collapse = ", "), " (", n, description, ")")
}

#' Create a new `pproto` object
#'
#' Construct a new object with `pproto`. This object system is inspired
#' from the `ggproto` system used in the `ggplot2` package.
#'
#' @param _class Class name to assign to the object. This is stored as the class
#'   attribute of the object. This is optional: if `NULL` (the default),
#'   no class name will be added to the object.
#'
#' @param _inherit `pproto` object to inherit from. If `NULL`, don"t
#'   inherit from any object.
#'
#' @param ... A list of members to add to the new `pproto` object.
#'
#' @examples
#' Adder <- pproto("Adder",
#'   x = 0,
#'   add = function(self, n) {
#'     self$x <- self$x + n
#'     self$x
#'   }
#' )
#'
#' Adder$add(10)
#' Adder$add(10)
#'
#' Abacus <- pproto("Abacus", Adder,
#'   subtract = function(self, n) {
#'     self$x <- self$x - n
#'     self$x
#'   }
#' )
#' Abacus$add(10)
#' Abacus$subtract(10)
#' @noRd
pproto <- function(`_class` = NULL, `_inherit` = NULL, ...) {
  assertthat::assert_that(
    assertthat::is.string(`_class`) || is.null(`_class`),
    inherits(`_inherit`, "pproto") || is.null(`_inherit`)
  )
  # copy objects from one proto to another proto
  assign_fields <- function(p1, p2) {
    if (!inherits(p2, "proto")) {
      return()
    }
    for (i in p2$ls()) {
      if (inherits(p2[[i]], "proto")) {
        p1[[i]] <- proto::proto()
        class(p1[[i]]) <- class(p2[[i]])
        assign_fields(p1[[i]], p2[[i]])
      } else {
        p1[[i]] <- p2[[i]]
      }
    }
    assign_fields(p1, p2$.super)
  }
  # create new proto
  p <- proto::proto()
  if (!is.null(`_inherit`)) {
    # assign inherited members
    assign_fields(p, `_inherit`)
    # assign inherited classes
    class(p) <- class(`_inherit`)
  } else {
    # assign pproto class
    class(p) <- c("pproto", class(p))
  }
  # assign members to new proto
  assign_fields(p, proto::proto(...))
  # assign new class if specified
  if (!is.null(`_class`)) {
    class(p) <- c(`_class`, class(p))
  }
  # return value
  p
}


#' Get status
#'
#' Construct a new object with `pproto`. This object system is inspired
#' from the `ggproto` system used in the `ggplot2` package.
#'
#' @param _class Class name to assign to the object. This is stored as the class
#'   attribute of the object. This is optional: if `NULL` (the default),
#'   no class name will be added to the object.
#'
#' @param _inherit `pproto` object to inherit from. If `NULL`, don"t
#'   inherit from any object.
#'
#' @param ... A list of members to add to the new `pproto` object.
#'
#' @examples
#' # set seed for reproducibility
#' set.seed(14)
#'
#' ## Load data
#' data(sim_pu_data, sim_features_data, sim_dist_features_data,
#' sim_threats_data, sim_dist_threats_data, sim_sensitivity_data,
#' sim_boundary_data)
#'
#' ## Create data instance
#' problem_data <- create_problem(
#'   pu = sim_pu_data, features = sim_features_data, dist_features = sim_dist_features_data,
#'   threats = sim_threats_data, dist_threats = sim_dist_threats_data,
#'   sensitivity = sim_sensitivity_data, boundary = sim_boundary_data
#' )
#'
#' ## Create optimization model
#' problem_model <- problem(x = problem_data, blm = 1)
#'
#' ## Solve the optimization model
#' s <- solve(a = problem_model, solver = "gurobi", gap_limit = 0.01, output_file = FALSE)
#'
#' ## get status of solution
#' getStatus(s)
#'
#' @noRd
getStatus <- function(x) {
  # Preferred structure: Solution with diagnostics
  d <- NULL

  if (!is.null(x$diagnostics)) {
    d <- x$diagnostics
  } else {
    d <- x
  }

  code <- d$status_code %||% d$status %||% NA_integer_
  code <- as.integer(code)[1]

  if (is.na(code)) return("unknown")

  switch(
    as.character(code),
    "0"   = "optimal",
    "1"   = "infeasible_or_unbounded",
    "2"   = "time_limit_feasible",
    "3"   = "time_limit_no_solution",
    "4"   = "solution_limit",
    "999" = "unknown",
    paste0("unknown(", code, ")")
  )
}


#' Check if solvers are working
#'
#' Provides the status of solver. Being TRUE if it's working fine and FALSE in otherwise.
#'
#' @param package `character` object. Posible values: "gurobi", "cplex", and "symphony".
#'
#' @examples
#' available_to_solve("cplex")
#'

#' @noRd
available_to_solve <- function(package = ""){

  # define primitive data
  nPlants     <- 1
  nWarehouses <- 1
  # Warehouse demand in thousands of units
  Demand      <- c(10)
  # Plant capacity in thousands of units
  Capacity    <- c(20)
  # Fixed costs for each plant
  FixedCosts  <- c(100)
  # Transportation costs per thousand units
  TransCosts  <- c(100)

  flowidx <- function(w, p) {nPlants * (w-1) + p}

   # Build model
  model <- list()
  model$modelname <- 'facility'
  model$modelsense <- 'min'

  # initialize data for variables
  model$lb       <- 0
  model$ub       <- c(rep(1, nPlants),   rep(Inf, nPlants * nWarehouses))
  model$vtype    <- c(rep('B', nPlants), rep('C', nPlants * nWarehouses))
  model$obj      <- c(FixedCosts, TransCosts)
  model$varnames <- c(paste0(rep('Open',nPlants),1:nPlants),
                      sprintf('Trans%d,%d',
                              c(mapply(rep,1:nWarehouses,nPlants)),
                              1:nPlants))

  # build production constraint matrix
  A1 <- Matrix::spMatrix(nPlants, nPlants, i = c(1:nPlants), j = (1:nPlants), x = -Capacity)
  A2 <- Matrix::spMatrix(nPlants, nPlants * nWarehouses,
                 i = c(mapply(rep, 1:nPlants, nWarehouses)),
                 j = mapply(flowidx,1:nWarehouses,c(mapply(rep,1:nPlants,nWarehouses))),
                 x = rep(1, nWarehouses * nPlants))
  A3 <- Matrix::spMatrix(nWarehouses, nPlants)
  A4 <- Matrix::spMatrix(nWarehouses, nPlants * nWarehouses,
                 i = c(mapply(rep, 1:nWarehouses, nPlants)),
                 j = mapply(flowidx,c(mapply(rep,1:nWarehouses,nPlants)),1:nPlants),
                 x = rep(1, nPlants * nWarehouses))
  model$A           <- rbind(cbind(A1, A2), cbind(A3, A4))
  model$rhs         <- c(rep(0, nPlants),   Demand)
  model$sense       <- c(rep('<=', nPlants), rep('==', nWarehouses))
  model$constrnames <- c(sprintf('Capacity%d',1:nPlants),
                         sprintf('Demand%d',1:nWarehouses))


  if(package == "gurobi"){
    # set parameters
    params <- list()
    params$TimeLimit <- 0.01
    params$LogToConsole <- 0
    model$sense <- replace(model$sense, model$sense == "==", "=")

    sol <- invisible(try(gurobi::gurobi(model, params), silent = TRUE))
  }
  else if(package == "cbc"){
    # set parameters
    cbc_args <- list()
    cbc_args$sec <- "0.01"
    cbc_args$log <- "0"

    constraints_minus_equal <- which(model$sense != "<=")
    constraints_plus_equal <- which(model$sense == "<=")
    row_ub <- model$rhs
    row_ub[constraints_minus_equal] <- Inf
    row_lb <- model$rhs
    row_lb[constraints_plus_equal] <- -Inf

    sol <- invisible(try(rcbc::cbc_solve(obj = model$obj,
                                         mat = model$A,
                                         is_integer = ifelse(model$vtype == "B", TRUE, FALSE),
                                         row_ub = row_ub,
                                         row_lb = row_lb,
                                         col_lb = rep(0, length(model$vtype)),
                                         col_ub = model$ub,
                                         max = ifelse(model$modelsense == "min", FALSE, TRUE),
                                         cbc_args = cbc_args), silent = TRUE))
  }
  else if(package == "cplex"){
    model$sense[model$sense == ">="] <- "G"
    model$sense[model$sense == "=="] <- "E"
    model$sense[model$sense == "<="] <- "L"

    # set parameters
    params <- list()
    params$tilim <- 0.01
    params$trace <- 0

    sol <- try(Rcplex::Rcplex(cvec = model$obj,
                              Amat = model$A,
                              bvec = model$rhs,
                              lb = model$lb,
                              ub = model$ub,
                              objsense = model$modelsense,
                              sense = model$sense,
                              vtype = model$vtype,
                              control = params),
               silent = TRUE)
  }
  else if(package == "symphony"){
    model$mat <- model$A
    model$dir <- model$sense
    model$max <- ifelse(model$modelsense == "min", FALSE, TRUE)
    model$types <- model$vtype

    sol <- invisible(try(Rsymphony::Rsymphony_solve_LP(model$obj,
                                             model$mat,
                                             model$dir,
                                             model$rhs,
                                             model$bounds,
                                             model$types,
                                             model$max,
                                             gap_limit = 100,
                                             time_limit = 0.01),
               silent = TRUE))

  }
  if(inherits(sol, "try-error")){
    return(FALSE)
  }
  else{
    return(TRUE)
  }
}


# -------------------------------------------------------------------------
# Internal helpers TARGETS
# -------------------------------------------------------------------------
.pa_parse_targets <- function(x, targets, features = NULL) {
  feats <- x$data$features
  if (is.null(feats) || nrow(feats) == 0) {
    stop("x$data$features is missing/empty.", call. = FALSE)
  }
  if (!("id" %in% names(feats))) {
    stop("x$data$features must contain column 'id'.", call. = FALSE)
  }

  feat_ids <- as.numeric(feats$id)
  feat_names <- if ("name" %in% names(feats)) as.character(feats$name) else rep(NA_character_, length(feat_ids))

  .feature_name_from_id <- function(fid) {
    fid <- as.numeric(fid)
    feat_names[match(fid, feat_ids)]
  }

  # helper: map a feature identifier (id or name) to numeric id
  .map_feature <- function(v) {
    if (is.null(v)) return(NULL)
    if (is.factor(v)) v <- as.character(v)

    if (is.numeric(v) || is.integer(v)) {
      v <- as.numeric(v)
      bad <- v[!v %in% feat_ids]
      if (length(bad) > 0) {
        stop("Unknown feature id(s): ", paste(unique(bad), collapse = ", "), call. = FALSE)
      }
      return(v)
    }

    if (is.character(v)) {
      sup <- suppressWarnings(as.numeric(v))
      is_numlike <- !is.na(sup) & nzchar(v)
      out <- rep(NA_real_, length(v))

      if (any(is_numlike)) {
        vv <- sup[is_numlike]
        bad <- vv[!vv %in% feat_ids]
        if (length(bad) == 0) out[is_numlike] <- vv
      }

      need_name <- is.na(out)
      if (any(need_name)) {
        if (!("name" %in% names(feats))) {
          stop("Targets reference feature names, but x$data$features has no 'name' column.", call. = FALSE)
        }
        m <- match(v[need_name], feat_names)
        bad <- v[need_name][is.na(m)]
        if (length(bad) > 0) {
          stop(
            "Unknown feature name(s): ", paste0("'", unique(bad), "'", collapse = ", "),
            ". Valid names include: ", paste0("'", utils::head(feat_names, 10), "'", collapse = ", "),
            if (length(feat_names) > 10) " ..." else "",
            call. = FALSE
          )
        }
        out[need_name] <- feat_ids[m]
      }

      return(as.numeric(out))
    }

    stop("Unsupported feature identifier type.", call. = FALSE)
  }

  .resolve_target_features <- function(features) {
    if (is.null(features)) return(feat_ids)
    f <- .map_feature(features)
    if (length(f) == 0) {
      stop("features resolved to an empty set.", call. = FALSE)
    }
    f <- unique(as.numeric(f))
    f
  }

  .finalize_targets <- function(f, t) {
    dt <- data.frame(
      feature = as.numeric(f),
      feature_name = .feature_name_from_id(f),
      target_raw = as.numeric(t),
      stringsAsFactors = FALSE
    )
    dt <- dt[order(dt$feature), , drop = FALSE]
    if (anyDuplicated(dt$feature)) {
      stop("Duplicate targets for the same feature. Please de-duplicate.", call. = FALSE)
    }
    dt
  }

  selected_features <- .resolve_target_features(features)

  # 1) data.frame(feature, target)
  if (inherits(targets, "data.frame")) {
    if (!all(c("feature", "target") %in% names(targets))) {
      stop("If targets is a data.frame it must contain columns 'feature' and 'target'.", call. = FALSE)
    }
    if (!is.null(features)) {
      warning(
        "'features' was supplied but targets is a data.frame with an explicit 'feature' column. ",
        "Ignoring 'features'.",
        call. = FALSE
      )
    }
    f <- .map_feature(targets$feature)
    t <- as.numeric(targets$target)
    if (any(is.na(t))) stop("Targets contain NA values.", call. = FALSE)
    return(.finalize_targets(f, t))
  }

  # 2) matrix (one column) with optional rownames
  if (is.matrix(targets)) {
    if (ncol(targets) != 1) {
      stop("If targets is a matrix it must have exactly 1 column.", call. = FALSE)
    }
    t <- as.numeric(targets[, 1])
    if (any(is.na(t))) stop("Targets contain NA values.", call. = FALSE)

    rn <- rownames(targets)
    if (!is.null(rn) && length(rn) == length(t)) {
      if (!is.null(features)) {
        warning(
          "'features' was supplied but targets is a matrix with rownames. Ignoring 'features'.",
          call. = FALSE
        )
      }
      f <- .map_feature(rn)
      return(.finalize_targets(f, t))
    }

    if (length(t) == 1) {
      return(.finalize_targets(selected_features, rep(t, length(selected_features))))
    }

    if (length(t) != length(selected_features)) {
      stop(
        "Matrix targets without rownames must have nrow = 1 or nrow = number of targeted features (",
        length(selected_features), ").",
        call. = FALSE
      )
    }

    return(.finalize_targets(selected_features, t))
  }

  # 3) scalar numeric -> recycle
  if (is.numeric(targets) && length(targets) == 1) {
    t <- as.numeric(targets)
    if (is.na(t)) stop("Target is NA.", call. = FALSE)
    return(.finalize_targets(selected_features, rep(t, length(selected_features))))
  }

  # 4) numeric vector
  if (is.numeric(targets) && length(targets) > 1) {
    t <- as.numeric(targets)
    if (any(is.na(t))) stop("Targets contain NA values.", call. = FALSE)

    nm <- names(targets)
    if (!is.null(nm) && any(nzchar(nm))) {
      if (!is.null(features)) {
        warning(
          "'features' was supplied but targets is a named vector. Ignoring 'features'.",
          call. = FALSE
        )
      }
      f <- .map_feature(nm)
      return(.finalize_targets(f, t))
    }

    if (length(t) != length(selected_features)) {
      stop(
        "Un-named numeric targets must have length 1 or length equal to the number of targeted features (",
        length(selected_features), ").",
        call. = FALSE
      )
    }

    return(.finalize_targets(selected_features, t))
  }

  # 5) character
  if (is.character(targets)) {
    if (length(targets) == 1 && grepl("^\\s*[-+]?[0-9]*\\.?[0-9]+\\s*$", targets)) {
      t <- as.numeric(targets)
      return(.finalize_targets(selected_features, rep(t, length(selected_features))))
    }

    if (any(grepl("=", targets, fixed = TRUE))) {
      if (!is.null(features)) {
        warning(
          "'features' was supplied but targets uses 'feature=target' syntax. Ignoring 'features'.",
          call. = FALSE
        )
      }
      parts <- strsplit(targets, "=", fixed = TRUE)
      feat_part <- vapply(parts, `[[`, character(1), 1)
      val_part  <- vapply(parts, `[[`, character(1), 2)
      f <- .map_feature(trimws(feat_part))
      t <- as.numeric(trimws(val_part))
      if (any(is.na(t))) {
        stop("Could not parse numeric target values from character input.", call. = FALSE)
      }
      return(.finalize_targets(f, t))
    }

    stop(
      "Unsupported character targets format. Use 'feature=target' pairs or a single numeric string.",
      call. = FALSE
    )
  }

  stop("Unsupported targets format.", call. = FALSE)
}

.pa_store_targets <- function(x, targets_df) {
  stopifnot(inherits(x, "Problem"))
  stopifnot(inherits(targets_df, "data.frame"))
  stopifnot(all(c("feature", "type", "target_value") %in% names(targets_df)))

  # normalize minimal fields
  targets_df$feature <- as.integer(targets_df$feature)
  targets_df$type <- as.character(targets_df$type)
  targets_df$target_value <- as.numeric(targets_df$target_value)

  # ensure actions column exists
  if (!("actions" %in% names(targets_df))) {
    targets_df$actions <- NA_character_
  }
  targets_df$actions <- as.character(targets_df$actions)

  valid_types <- c("actions")
  bad_type <- setdiff(unique(targets_df$type), valid_types)
  if (length(bad_type) > 0) {
    stop("Unknown target types: ", paste(bad_type, collapse = ", "), call. = FALSE)
  }

  # initialize if empty
  if (is.null(x$data$targets) || !inherits(x$data$targets, "data.frame") || nrow(x$data$targets) == 0) {
    x$data$targets <- targets_df

    if (!is.null(x$data$model_ptr)) {
      x$data$meta <- x$data$meta %||% list()
      x$data$meta$model_dirty <- TRUE
    }

    return(x)
  }

  old <- x$data$targets
  old$feature <- as.integer(old$feature)
  old$type <- as.character(old$type)
  old$target_value <- as.numeric(old$target_value)

  # backward compatibility: migrate old subset column if present
  if (!("actions" %in% names(old))) {
    if ("subset" %in% names(old)) {
      old$actions <- as.character(old$subset)
    } else {
      old$actions <- NA_character_
    }
  }
  old$actions <- as.character(old$actions)

  # optional informative warning for repeated keys
  key_old <- paste0(old$feature, "||", old$type, "||", old$actions)
  key_new <- paste0(targets_df$feature, "||", targets_df$type, "||", targets_df$actions)

  overlap <- intersect(key_old, key_new)
  if (length(overlap) > 0) {
    warning(
      "Additional targets were added for existing (feature, type, actions) combinations. ",
      "These target rows remain stored simultaneously and will be handled downstream. ",
      "Example key: ", overlap[1],
      call. = FALSE
    )
  }

  x$data$targets <- rbind(old, targets_df)

  if (!is.null(x$data$model_ptr)) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

.pa_feature_totals <- function(x) {
  df <- x$data$dist_features
  if (is.null(df) || nrow(df) == 0) {
    stop("Need x$data$dist_features to compute baseline totals.", call. = FALSE)
  }
  stopifnot(all(c("feature","amount") %in% names(df)))
  f <- as.numeric(df$feature)
  a <- as.numeric(df$amount)
  a[!is.finite(a) | is.na(a)] <- 0
  out <- tapply(a, f, sum, na.rm = TRUE)
  out[is.na(out)] <- 0
  out
}


.pa_feature_potential <- function(x, value_col = "benefit") {
  de <- x$data$dist_effects_model %||% x$data$dist_effects
  if (is.null(de) || nrow(de) == 0) {
    stop("Need dist_effects to compute potential. Run add_effects()/add_benefits() first.", call. = FALSE)
  }
  stopifnot(all(c("feature", value_col) %in% names(de)))

  f <- as.numeric(de$feature)
  v <- as.numeric(de[[value_col]])

  # potential is sum of positive contributions only
  v_pos <- pmax(v, 0)
  v_pos[!is.finite(v_pos) | is.na(v_pos)] <- 0

  out <- tapply(v_pos, f, sum, na.rm = TRUE)
  out[is.na(out)] <- 0
  out
}

# -----------------------------------------------------------------------------
# Apply targets to the active C++ model (helper)
# -----------------------------------------------------------------------------
# This helper is meant to be called INSIDE problem() after:
#   - base variables have been added
#   - structural constraints have been added
#   - objective has been set (optional)
#
# It will:
#   - read x$data$targets (data.frame)
#   - decide per-feature whether we have conservation-only, recovery-only, or mixed
#   - call the corresponding Rcpp function(s)
#   - store a lightweight "applied_targets" summary in x$data$model_args
#
# Assumptions (consistent with your design):
#   - x$data$model_ptr exists and is an XPtr<OptimizationProblem>
#   - target_value is already ABSOLUTE in x$data$targets
#   - conservation-only uses z variables + dist_features amounts
#   - recovery-only uses x variables + dist_benefit deltas (can be + or -)
#   - mixed (option 1) uses total = (conservation baseline) + (action deltas),
#     with target = Tc + Tr if both targets exist for a feature.
# -----------------------------------------------------------------------------
# helper: get model-ready benefit table from dist_effects
.get_dist_benefit_model_from_effects <- function(x, benefit_col = "benefit") {
  de <- x$data$dist_effects_model %||% x$data$dist_effects

  if (is.null(de) || !inherits(de, "data.frame") || nrow(de) == 0) {
    return(NULL)
  }

  benefit_col <- as.character(benefit_col)[1]
  if (is.na(benefit_col) || !nzchar(benefit_col)) {
    stop("`benefit_col` must be a non-empty string.", call. = FALSE)
  }

  if (!(benefit_col %in% names(de))) {
    stop("dist_effects is missing column '", benefit_col, "'.", call. = FALSE)
  }

  out <- de

  if ("feature" %in% names(out))          out$feature <- as.integer(out$feature)
  if ("internal_pu" %in% names(out))      out$internal_pu <- as.integer(out$internal_pu)
  if ("internal_action" %in% names(out))  out$internal_action <- as.integer(out$internal_action)
  if ("internal_feature" %in% names(out)) out$internal_feature <- as.integer(out$internal_feature)

  out$benefit <- as.numeric(out[[benefit_col]])

  if (anyNA(out$benefit) || any(!is.finite(out$benefit))) {
    stop(
      "Column '", benefit_col, "' in dist_effects contains NA or non-finite values.",
      call. = FALSE
    )
  }

  out
}

.get_dist_effects_model <- function(x, mode = c("benefit", "loss", "delta")) {
  mode <- match.arg(mode)

  de <- x$data$dist_effects_model %||% x$data$dist_effects
  if (is.null(de) || !inherits(de, "data.frame") || nrow(de) == 0) {
    return(NULL)
  }

  out <- de

  if ("feature" %in% names(out))          out$feature <- as.integer(out$feature)
  if ("internal_pu" %in% names(out))      out$internal_pu <- as.integer(out$internal_pu)
  if ("internal_action" %in% names(out))  out$internal_action <- as.integer(out$internal_action)
  if ("internal_feature" %in% names(out)) out$internal_feature <- as.integer(out$internal_feature)

  if (mode == "benefit") {
    if (!("benefit" %in% names(out))) {
      stop("dist_effects is missing column 'benefit'.", call. = FALSE)
    }

    out$benefit <- as.numeric(out$benefit)
    if (anyNA(out$benefit) || any(!is.finite(out$benefit))) {
      stop("dist_effects$benefit contains NA or non-finite values.", call. = FALSE)
    }

    out$effect <- out$benefit
    return(out)
  }

  if (mode == "loss") {
    if (!("loss" %in% names(out))) {
      stop("dist_effects is missing column 'loss'.", call. = FALSE)
    }

    out$loss <- as.numeric(out$loss)
    if (anyNA(out$loss) || any(!is.finite(out$loss))) {
      stop("dist_effects$loss contains NA or non-finite values.", call. = FALSE)
    }

    out$effect <- out$loss
    return(out)
  }

  # mode == "delta"
  if (!("benefit" %in% names(out)) && !("loss" %in% names(out))) {
    stop(
      "dist_effects must contain at least one of 'benefit' or 'loss' to compute delta.",
      call. = FALSE
    )
  }

  if (!("benefit" %in% names(out))) out$benefit <- 0
  if (!("loss" %in% names(out)))    out$loss <- 0

  out$benefit <- as.numeric(out$benefit)
  out$loss    <- as.numeric(out$loss)

  if (anyNA(out$benefit) || any(!is.finite(out$benefit))) {
    stop("dist_effects$benefit contains NA or non-finite values.", call. = FALSE)
  }
  if (anyNA(out$loss) || any(!is.finite(out$loss))) {
    stop("dist_effects$loss contains NA or non-finite values.", call. = FALSE)
  }

  out$effect <- out$benefit - out$loss
  out
}


.pa_apply_targets_if_present <- function(x,
                                         zero_tol = 1e-12) {

  stopifnot(inherits(x, "Problem"))

  if (is.null(x$data$targets) || !inherits(x$data$targets, "data.frame") || nrow(x$data$targets) == 0) {
    return(invisible(x))
  }
  if (is.null(x$data$model_ptr)) {
    stop("No active model pointer found in x$data$model_ptr. Run add_constraint_targets_*() before applying targets.", call. = FALSE)
  }

  t <- x$data$targets
  req <- c("feature", "type", "target_value")
  miss <- setdiff(req, names(t))
  if (length(miss) > 0) {
    stop("x$data$targets is missing required columns: ", paste(miss, collapse = ", "), call. = FALSE)
  }

  # backward compatibility: migrate subset -> actions
  if (!("actions" %in% names(t))) {
    if ("subset" %in% names(t)) {
      t$actions <- as.character(t$subset)
    } else {
      t$actions <- NA_character_
    }
  }

  t$feature <- as.integer(t$feature)
  t$type <- as.character(t$type)
  t$actions <- as.character(t$actions)
  t$target_value <- as.numeric(t$target_value)

  bad_type <- setdiff(unique(t$type), c("actions"))
  if (length(bad_type) > 0) {
    stop("Unknown target types in x$data$targets: ", paste(bad_type, collapse = ", "), call. = FALSE)
  }

  # ------------------------------------------------------------------
  # 1) exact duplicates by (feature, type, actions) are NOT allowed
  # ------------------------------------------------------------------
  key_exact <- paste0(
    t$feature, "||",
    t$type, "||",
    ifelse(is.na(t$actions), "", t$actions)
  )

  dup_exact <- duplicated(key_exact) | duplicated(key_exact, fromLast = TRUE)

  if (any(dup_exact)) {
    bad <- t[dup_exact, c("feature", "type", "actions", "target_value"), drop = FALSE]
    ex <- utils::head(
      paste0(
        "(feature=", bad$feature,
        ", type='", bad$type,
        "', actions='", ifelse(is.na(bad$actions) | !nzchar(bad$actions), "ALL", bad$actions),
        "', target_value=", bad$target_value, ")"
      ),
      8
    )

    stop(
      "Multiple targets were defined for the same (feature, type, actions) combination.\n",
      "This is ambiguous and is not allowed.\n",
      "Examples: ", paste(ex, collapse = ", "),
      if (nrow(bad) > 8) paste0(" ... (", nrow(bad), " rows)") else "",
      call. = FALSE
    )
  }

  # ------------------------------------------------------------------
  # 2) same (feature, actions) with different type: allowed, but warn
  # ------------------------------------------------------------------
  key_feature_actions <- paste0(
    t$feature, "||",
    ifelse(is.na(t$actions), "", t$actions)
  )

  split_fa <- split(t$type, key_feature_actions, drop = TRUE)
  multi_type_keys <- names(split_fa)[vapply(split_fa, function(z) length(unique(z)) > 1, logical(1))]

  if (length(multi_type_keys) > 0) {
    ex_keys <- utils::head(multi_type_keys, 8)
    warning(
      "Multiple targets were defined for the same (feature, actions) combination with different target types.\n",
      "They will be applied as separate constraints, and the strongest one may dominate in practice.\n",
      "Examples: ", paste(ex_keys, collapse = ", "),
      if (length(multi_type_keys) > 8) paste0(" ... (", length(multi_type_keys), " combinations)") else "",
      call. = FALSE
    )
  }

  # ------------------------------------------------------------------
  # 3) drop near-zero targets
  # ------------------------------------------------------------------
  idx_zero <- is.finite(t$target_value) & abs(t$target_value) <= zero_tol
  if (any(idx_zero)) {
    z <- t[idx_zero, , drop = FALSE]
    warning(
      "Skipping ", nrow(z), " target(s) with target_value ~ 0 (|rhs| <= ", format(zero_tol), "). ",
      "They are redundant and can obscure infeasibility diagnostics.",
      call. = FALSE
    )
    t <- t[!idx_zero, , drop = FALSE]
  }

  if (nrow(t) == 0) {
    x$data$model_args$targets_applied <- TRUE
    x$data$model_args$targets_counts <- list(actions = 0L)
    return(invisible(x))
  }

  op <- x$data$model_ptr

  .mk_targets_df <- function(features, values, colname) {
    df <- data.frame(internal_id = as.integer(features), stringsAsFactors = FALSE)
    df[[colname]] <- as.numeric(values)
    df
  }

  .check_feature_present_in_df <- function(features, df, feature_col, what) {
    features <- as.integer(features)
    if (length(features) == 0) return(invisible(TRUE))

    present <- unique(as.integer(df[[feature_col]]))
    miss <- features[!features %in% present]

    if (length(miss) > 0) {
      ex <- paste(utils::head(sort(unique(miss)), 8), collapse = ", ")
      stop(
        "Infeasible targets detected: some features in ", what, " targets have no contributions in the model tables.\n",
        "Missing feature ids: ", ex,
        if (length(unique(miss)) > 8) paste0(" ... (", length(unique(miss)), " total)") else "",
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  .filter_effects_by_actions <- function(de, actions_string) {
    if (is.na(actions_string) || !nzchar(actions_string)) {
      return(de)
    }

    actions_vals <- strsplit(actions_string, "\\|")[[1]]
    actions_vals <- unique(actions_vals[nzchar(actions_vals)])

    matched <- .pa_resolve_action_subset(x, actions_vals)
    keep_internal <- as.integer(matched$internal_id)

    de[de$internal_action %in% keep_internal, , drop = FALSE]
  }

  dbm <- .get_dist_benefit_model_from_effects(x, benefit_col = "benefit")
  if (is.null(dbm) || nrow(dbm) == 0) {
    stop("Targets present but no benefit column available in dist_effects.", call. = FALSE)
  }

  if (!("feature" %in% names(dbm))) {
    stop("dist_benefit_data must contain a 'feature' column for validation.", call. = FALSE)
  }

  if (!exists("rcpp_add_target_recovery", mode = "function")) {
    stop("Missing rcpp_add_target_recovery() in the package.", call. = FALSE)
  }

  # ------------------------------------------------------------------
  # 4) Apply each row as a separate constraint
  # ------------------------------------------------------------------
  n_applied <- 0L

  for (ii in seq_len(nrow(t))) {
    tt <- t[ii, , drop = FALSE]
    actions_string <- tt$actions[1]

    dbm_sub <- .filter_effects_by_actions(dbm, actions_string)

    if (nrow(dbm_sub) == 0) {
      actions_lab <- if (is.na(actions_string) || !nzchar(actions_string)) "ALL actions" else actions_string
      stop(
        "Target for feature ", tt$feature[1], " and actions subset '", actions_lab,
        "' cannot be applied because no matching effect rows remain after filtering.",
        call. = FALSE
      )
    }

    .check_feature_present_in_df(tt$feature, dbm_sub, "feature", what = "actions")

    rcpp_add_target_recovery(
      op,
      features_data     = .mk_targets_df(tt$feature, tt$target_value, "target_actions"),
      dist_actions_data = x$data$dist_actions_model,
      dist_benefit_data = dbm_sub,
      target_col        = "target_actions"
    )

    n_applied <- n_applied + 1L
  }

  x$data$model_args$targets_applied <- TRUE
  x$data$model_args$targets_counts <- list(actions = n_applied)

  invisible(x)
}


# -------------------------------------------------------------------------
# Internal helpers MATHEMATICAL MODEL
# -------------------------------------------------------------------------
.pa_refresh_model_snapshot <- function(x, drop_triplets = TRUE) {
  stopifnot(inherits(x, "Problem"))

  if (is.null(x$data$model_ptr)) return(x)

  if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop("Package 'Matrix' is required to build sparse constraint matrix.", call. = FALSE)
  }

  model_list <- rcpp_optimization_problem_as_list(x$data$model_ptr)

  # Build sparse A defensively (A_i/A_j are 0-based in C++ -> +1 in R)
  if (all(c("A_i", "A_j", "A_x") %in% names(model_list)) &&
      length(model_list$A_i) == length(model_list$A_j) &&
      length(model_list$A_i) == length(model_list$A_x)) {

    model_list$A <- Matrix::sparseMatrix(
      i = as.integer(model_list$A_i) + 1L,
      j = as.integer(model_list$A_j) + 1L,
      x = as.numeric(model_list$A_x)
    )

    if (isTRUE(drop_triplets)) {
      model_list$A_i <- NULL
      model_list$A_j <- NULL
      model_list$A_x <- NULL
    }

  } else {
    # fallback: empty sparse matrix if missing triplets
    n_con <- if (!is.null(model_list$rhs)) length(model_list$rhs) else 0L
    n_var <- if (!is.null(model_list$obj)) length(model_list$obj) else 0L
    model_list$A <- Matrix::sparseMatrix(
      i = integer(0), j = integer(0), x = numeric(0),
      dims = c(n_con, n_var)
    )
  }

  # Keep args attached if you use them downstream (optional but handy)
  model_list$args <- x$data$model_args %||% list()

  x$data$model_list <- model_list

  # Optional: store dims for fast printing
  x$data$model_args <- x$data$model_args %||% list()
  x$data$model_args$n_constraints <- if (!is.null(model_list$rhs)) length(model_list$rhs) else NA_integer_
  x$data$model_args$n_variables   <- if (!is.null(model_list$obj)) length(model_list$obj) else NA_integer_
  x$data$model_args$nnz           <- Matrix::nnzero(model_list$A)

  x
}


.pa_apply_action_max_per_pu_if_present <- function(x) {
  stopifnot(inherits(x, "Problem"))

  spec <- x$data$constraints$action_max_per_pu %||% NULL
  if (is.null(spec)) return(invisible(x))

  if (is.null(x$data$model_ptr)) {
    stop("No active model pointer found in x$data$model_ptr.", call. = FALSE)
  }

  if (!exists("rcpp_add_action_max_per_pu", mode = "function")) {
    stop("Missing rcpp_add_action_max_per_pu() in the package.", call. = FALSE)
  }

  da <- x$data$dist_actions_model
  if (is.null(da) || !inherits(da, "data.frame") || nrow(da) == 0) {
    stop("action_max_per_pu requires non-empty x$data$dist_actions_model.", call. = FALSE)
  }

  # defensive: required columns
  need <- c("internal_pu", "internal_action", "internal_row")
  miss <- setdiff(need, names(da))
  if (length(miss) > 0) {
    stop("dist_actions_model must contain columns: ", paste(miss, collapse = ", "), call. = FALSE)
  }

  maxv <- as.integer(spec$max %||% 1L)
  if (!is.finite(maxv) || is.na(maxv) || maxv < 0L) stop("Invalid max in action_max_per_pu.", call. = FALSE)

  # Convert external ids -> internal ids
  pu_ext <- spec$pu %||% x$data$pu$id
  pu_ext <- as.integer(pu_ext)
  pu_map <- x$data$pu[, c("id","internal_id")]
  pu_int <- pu_map$internal_id[match(pu_ext, pu_map$id)]
  pu_int <- pu_int[!is.na(pu_int)]
  if (length(pu_int) == 0) stop("action_max_per_pu: no valid PUs after mapping to internal ids.", call. = FALSE)

  act_ext <- spec$actions %||% x$data$actions$id
  act_ext <- as.character(act_ext)
  act_map <- x$data$actions[, c("id","internal_id")]
  act_int <- act_map$internal_id[match(act_ext, act_map$id)]
  act_int <- act_int[!is.na(act_int)]
  if (length(act_int) == 0) stop("action_max_per_pu: no valid actions after mapping to internal ids.", call. = FALSE)

  # call C++
  res <- rcpp_add_action_max_per_pu(
    x = x$data$model_ptr,
    dist_actions_data = da,
    max_per_pu = maxv,
    internal_pu_ids = as.integer(pu_int),
    internal_action_ids = as.integer(act_int)
  )

  # optional registry
  x$data$model_registry <- x$data$model_registry %||% list(cons = list(), vars = list(), obj_templates = list(), objective = list())
  x$data$model_registry$cons$action_max_per_pu <- res

  invisible(x)
}

.pa_apply_area_constraints_if_present <- function(x) {
  stopifnot(inherits(x, "Problem"))

  cons <- x$data$constraints %||% list()
  if (length(cons) == 0L) return(x)

  specs <- cons$area %||% NULL
  if (is.null(specs)) return(x)

  if (is.null(x$data$model_ptr)) {
    stop("Model pointer is missing while applying area constraints.", call. = FALSE)
  }

  x <- .pa_refresh_model_snapshot(x)

  ml <- x$data$model_list
  if (is.null(ml)) {
    stop("Model snapshot is missing while applying area constraints.", call. = FALSE)
  }

  n_pu <- as.integer(ml$n_pu %||% 0L)
  if (n_pu <= 0L) {
    stop("Model has n_pu=0; cannot apply area constraints.", call. = FALSE)
  }

  if (!is.data.frame(specs)) {
    stop(
      "Stored area constraints must be a data.frame.",
      call. = FALSE
    )
  }

  required_cols <- c(
    "type", "sense", "value", "tolerance",
    "unit", "area_col", "actions", "name"
  )

  miss <- setdiff(required_cols, names(specs))
  if (length(miss) > 0) {
    stop(
      "Stored area constraints are malformed. Missing columns: ",
      paste(miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  w0 <- as.integer(ml$w_offset %||% 0L)
  j0_w <- w0 + (0:(n_pu - 1L))

  da <- x$data$dist_actions_model %||% NULL
  if (is.null(da) || !is.data.frame(da) || nrow(da) == 0L) {
    stop(
      "Action-level area constraints require `x$data$dist_actions_model`.",
      call. = FALSE
    )
  }

  x0 <- as.integer(ml$x_offset %||% 0L)

  for (k in seq_len(nrow(specs))) {
    spec <- specs[k, , drop = FALSE]

    sense <- as.character(spec$sense)[1]
    rhs <- as.numeric(spec$value)[1]
    tol <- as.numeric(spec$tolerance)[1]
    nm <- as.character(spec$name)[1]
    area_col <- as.character(spec$area_col)[1]
    area_unit <- as.character(spec$unit)[1]
    actions_txt <- as.character(spec$actions)[1]

    if (!sense %in% c("min", "max", "equal")) {
      stop("Unknown area constraint sense: ", sense, call. = FALSE)
    }
    if (!is.finite(rhs) || is.na(rhs) || rhs < 0) {
      stop("Invalid area constraint value in stored area constraints.", call. = FALSE)
    }
    if (!is.finite(tol) || is.na(tol) || tol < 0) {
      stop("Invalid area constraint tolerance in stored area constraints.", call. = FALSE)
    }
    if (is.na(nm) || !nzchar(nm)) {
      stop("Stored area constraint has invalid `name`.", call. = FALSE)
    }

    a <- .pa_get_area_vec(
      x,
      area_col = if (is.na(area_col)) NULL else area_col,
      area_unit = area_unit %||% "m2"
    )

    if (length(a) != n_pu) {
      stop(
        "Area vector length (", length(a), ") != n_pu (", n_pu,
        ") while applying area constraint `", nm, "`.",
        call. = FALSE
      )
    }

    if (is.na(actions_txt) || !nzchar(actions_txt)) {

      var_index <- j0_w
      coeff <- a

    } else {

      actions_chr <- strsplit(actions_txt, "\\|", fixed = FALSE)[[1]]

      act_subset <- .pa_resolve_action_subset(x, subset = actions_chr)
      if (!is.data.frame(act_subset) || nrow(act_subset) == 0L) {
        stop(
          "Stored area constraint `", nm,
          "` refers to an invalid or empty action subset.",
          call. = FALSE
        )
      }

      keep <- da$action %in% act_subset$id

      if (!any(keep)) {
        stop(
          "Stored area constraint `", nm,
          "` does not match any rows in `dist_actions_model`.",
          call. = FALSE
        )
      }

      if (!("internal_row" %in% names(da))) {
        stop(
          "`dist_actions_model` must contain column `internal_row`.",
          call. = FALSE
        )
      }

      if (!("internal_pu" %in% names(da))) {
        stop(
          "`dist_actions_model` must contain column `internal_pu`.",
          call. = FALSE
        )
      }

      var_index <- x0 + (as.integer(da$internal_row[keep]) - 1L)
      coeff <- a[as.integer(da$internal_pu[keep])]

      if (length(var_index) != length(coeff) || length(coeff) == 0L) {
        stop(
          "Failed to assemble coefficients for area constraint `", nm, "`.",
          call. = FALSE
        )
      }
    }

    if (identical(sense, "min")) {

      x <- .pa_add_linear_constraint(
        x,
        var_index_0based = var_index,
        coeff = coeff,
        sense = ">=",
        rhs = rhs,
        name = nm
      )

    } else if (identical(sense, "max")) {

      x <- .pa_add_linear_constraint(
        x,
        var_index_0based = var_index,
        coeff = coeff,
        sense = "<=",
        rhs = rhs,
        name = nm
      )

    } else if (identical(sense, "equal")) {

      if (tol == 0) {

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = "==",
          rhs = rhs,
          name = nm
        )

      } else {

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = ">=",
          rhs = rhs - tol,
          name = paste0(nm, "_lower")
        )

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = "<=",
          rhs = rhs + tol,
          name = paste0(nm, "_upper")
        )
      }
    }
  }

  x
}

.pa_model_from_ptr <- function(op, args = list(), drop_triplets = TRUE) {

  if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop("Package 'Matrix' is required to build sparse constraint matrix.", call. = FALSE)
  }

  model_list <- rcpp_optimization_problem_as_list(op)

  # n_con / n_var siempre desde rhs/obj
  n_con <- if (!is.null(model_list$rhs)) length(model_list$rhs) else 0L
  n_var <- if (!is.null(model_list$obj)) length(model_list$obj) else 0L

  # build sparse A defensively (A_i/A_j are 0-based -> +1)
  if (all(c("A_i", "A_j", "A_x") %in% names(model_list)) &&
      length(model_list$A_i) == length(model_list$A_j) &&
      length(model_list$A_i) == length(model_list$A_x) &&
      length(model_list$A_i) > 0L) {

    model_list$A <- Matrix::sparseMatrix(
      i = as.integer(model_list$A_i) + 1L,
      j = as.integer(model_list$A_j) + 1L,
      x = as.numeric(model_list$A_x),
      dims = c(n_con, n_var)   # <<< CRÍTICO: fija dimensiones reales
    )

    if (isTRUE(drop_triplets)) {
      model_list$A_i <- NULL
      model_list$A_j <- NULL
      model_list$A_x <- NULL
    }

  } else {

    model_list$A <- Matrix::sparseMatrix(
      i = integer(0), j = integer(0), x = numeric(0),
      dims = c(n_con, n_var)
    )
  }

  # attach model args (needed for curve/segments and any other solver params)
  model_list$args <- args %||% list()

  # registry: ya viene desde C++ como data.frame en model_list$registry.
  # No lo toques aquí; solo asegúrate de no borrarlo por error.
  # (si quisieras, podrías validar que exista)
  # if (is.null(model_list$registry)) model_list$registry <- data.frame()

  model_list
}



# -------------------------------------------------------------------------
# Internal helpers CLASS
# -------------------------------------------------------------------------
.pa_cli_theme <- function() {
  list(
    .h     = list("font-weight" = "bold", color = "#569746"),
    .cls   = list("font-weight" = "bold", color = "#4C78A8"),
    .code  = list(color = "#3A7D44"),
    .muted = list(color = "grey60"),
    .ok    = list(color = "#569746", "font-weight" = "bold"),
    .warn  = list(color = "#C78A00", "font-weight" = "bold"),
    .bad   = list(color = "#C0392B", "font-weight" = "bold")
  )
}

.pa_cli_box_chars <- function() {
  use_unicode <- l10n_info()[["UTF-8"]]
  if (isTRUE(use_unicode)) {
    list(
      j = "\u251C", # ├
      l = "\u2514", # └
      v = "\u2502", # │
      b = "\u2500"  # ─
    )
  } else {
    list(
      j = "+",
      l = "\\",
      v = "|",
      b = "-"
    )
  }
}

.pa_safe_range <- function(x, digits = 5) {
  x <- as.numeric(x)
  x <- x[is.finite(x) & !is.na(x)]
  if (length(x) == 0) return(NULL)
  base::round(range(x), digits)
}

.pa_repr_atomic <- function(x, label = NULL, max_items = 6) {
  x <- as.character(x)
  x <- x[!is.na(x) & nzchar(x)]
  if (length(x) == 0) {
    if (is.null(label)) return("{.muted (none)}")
    return(paste0("{.muted (0 ", label, ")}"))
  }
  if (length(x) <= max_items) {
    paste0(paste0('"', x, '"', collapse = ", "))
  } else {
    head <- x[seq_len(max_items)]
    paste0(
      paste0('"', head, '"', collapse = ", "),
      ", {.muted ... (", length(x), " total)}"
    )
  }
}

.pa_get_cost_vec <- function(pu) {
  if (is.null(pu) || nrow(pu) == 0) return(numeric(0))
  if ("cost" %in% names(pu)) return(as.numeric(pu$cost))
  if ("monitoring_cost" %in% names(pu)) return(as.numeric(pu$monitoring_cost))
  numeric(0)
}

.pa_get_action_cost_vec <- function(self) {
  # new field (preferred)
  if (!is.null(self$data$dist_actions) &&
      inherits(self$data$dist_actions, "data.frame") &&
      nrow(self$data$dist_actions) > 0 &&
      "cost" %in% names(self$data$dist_actions)) {
    return(as.numeric(self$data$dist_actions$cost))
  }
  numeric(0)
}

.pa_nrow0 <- function(x) {
  if (is.null(x)) return(0L)
  if (!inherits(x, "data.frame")) return(0L)
  nrow(x)
}

.pa_has_model <- function(self) {
  !is.null(self$data$model_list) && is.list(self$data$model_list)
}

.pa_model_dims <- function(self) {
  # Returns list(n_con, n_var, nnz) safely
  ml <- self$data$model_list

  # If model_list doesn't exist yet, return zeros
  if (is.null(ml) || !is.list(ml)) {
    return(list(n_con = 0L, n_var = 0L, nnz = 0L))
  }

  # Best effort: support both A (dgCMatrix) and triplet (A_i/A_j/A_x)
  n_con <- 0L
  n_var <- 0L
  nnz   <- 0L

  if (!is.null(ml$A) && inherits(ml$A, "Matrix")) {
    n_con <- nrow(ml$A)
    n_var <- ncol(ml$A)
    nnz   <- length(ml$A@x)
    return(list(n_con = as.integer(n_con), n_var = as.integer(n_var), nnz = as.integer(nnz)))
  }

  if (!is.null(ml$A_i) && !is.null(ml$A_j)) {
    # We may not know nrow/ncol if not stored; infer n_con/n_var if present
    nnz <- length(ml$A_i)
    if (!is.null(ml$rhs)) n_con <- length(ml$rhs)
    if (!is.null(ml$obj)) n_var <- length(ml$obj)
    return(list(n_con = as.integer(n_con), n_var = as.integer(n_var), nnz = as.integer(nnz)))
  }

  # Fallback from rhs/obj if present
  if (!is.null(ml$rhs)) n_con <- length(ml$rhs)
  if (!is.null(ml$obj)) n_var <- length(ml$obj)

  list(n_con = as.integer(n_con), n_var = as.integer(n_var), nnz = as.integer(nnz))
}

.pa_model_blocks <- function(self) {
  # Returns list(n_w, n_x, n_z) if index mapping exists; otherwise NULL.
  # This expects you to store something like idx from rcpp_add_base_variables()
  # in self$data$model_index. Your idx currently returns w_index/x_index and
  # could also return z_index (recommended).
  idx <- self$data$model_index
  if (is.null(idx) || !is.list(idx)) return(NULL)

  n_w <- if (!is.null(idx$w_index)) length(idx$w_index) else NA_integer_
  n_x <- if (!is.null(idx$x_index)) length(idx$x_index) else NA_integer_
  n_z <- if (!is.null(idx$z_index)) length(idx$z_index) else NA_integer_

  # if none exist, don't print a misleading line
  if (is.na(n_w) && is.na(n_x) && is.na(n_z)) return(NULL)

  list(
    n_w = ifelse(is.na(n_w), 0L, as.integer(n_w)),
    n_x = ifelse(is.na(n_x), 0L, as.integer(n_x)),
    n_z = ifelse(is.na(n_z), 0L, as.integer(n_z))
  )
}

.pa_model_args <- function(self) {
  # Return a standardized args list if present; otherwise NULL
  a <- self$data$model_args
  if (is.null(a) || !is.list(a)) return(NULL)

  # Standardize expected fields (do not error if missing)
  out <- list(
    model_type    = if (!is.null(a$model_type)) a$model_type else NA_character_,
    modelsense    = if (!is.null(a$modelsense)) a$modelsense else NA_character_,
    objective_id  = if (!is.null(a$objective_id)) a$objective_id else NA_character_,
    budget        = if (!is.null(a$budget)) a$budget else NA_real_,
    blm           = if (!is.null(a$blm)) a$blm else NA_real_,
    benefit_exponent         = if (!is.null(a$benefit_exponent)) a$benefit_exponent else NA_integer_,
    curve_segments      = if (!is.null(a$curve_segments)) a$curve_segments else NA_integer_
  )

  out
}


# -------------------------------------------------------------------------
# Internal helpers DATA
# -------------------------------------------------------------------------
`%||%` <- function(a, b) if (is.null(a)) b else a


# internal helper: detect tabular vs spatial vs invalid mixes
.pa_detect_input_mode <- function(pu, features, dist_features) {

  # ---- local helpers (NO imports; only type checks)
  .is_df <- function(x) inherits(x, "data.frame")

  .is_spatraster <- function(x) inherits(x, "SpatRaster")
  .is_spatvector <- function(x) inherits(x, "SpatVector")
  .is_sf <- function(x) inherits(x, "sf")

  .is_raster_path <- function(x) {
    is.character(x) && length(x) == 1L &&
      grepl("\\.(tif|tiff|grd|asc|nc)$", tolower(x))
  }

  .is_vector_path <- function(x) {
    is.character(x) && length(x) == 1L &&
      grepl("\\.(shp|gpkg|geojson|json|fgb)$", tolower(x))
  }

  .is_spatial_pu <- function(x) .is_spatraster(x) || .is_spatvector(x) || .is_sf(x) || .is_raster_path(x) || .is_vector_path(x)
  .is_spatial_features <- function(x) .is_spatraster(x) || .is_raster_path(x)

  dist_missing <- missing(dist_features) || is.null(dist_features)

  # ------------------------------------------------------------
  # Rule 1: If dist_features is provided as data.frame -> TABULAR
  # ------------------------------------------------------------
  if (!dist_missing) {
    if (!.is_df(dist_features)) {
      stop("`dist_features` must be a data.frame in tabular mode, or NULL/missing in spatial mode.", call. = FALSE)
    }

    # dist_features df => require pu + features df (avoid ambiguity)
    if (!.is_df(pu)) {
      stop("Tabular mode detected because `dist_features` is a data.frame, but `pu` is not a data.frame.", call. = FALSE)
    }
    if (!.is_df(features)) {
      stop("Tabular mode detected because `dist_features` is a data.frame, but `features` is not a data.frame.", call. = FALSE)
    }

    return(list(mode = "tabular"))
  }

  # ------------------------------------------------------------
  # Rule 2: dist_features missing/NULL -> could be SPATIAL or invalid
  # ------------------------------------------------------------
  # spatial mode requires at least one spatial-ish input
  if (.is_spatial_features(features) || .is_spatial_pu(pu)) {
    return(list(mode = "spatial"))
  }

  # dist_features missing + no spatial inputs -> invalid (user forgot dist_features)
  stop(
    "Could not determine input style. Provide tabular inputs (data.frames: `pu`, `features`, `dist_features`) ",
    "or spatial inputs (e.g., `pu` and/or `features` as rasters/vectors/paths) with `dist_features = NULL`.",
    call. = FALSE
  )
}

# helpers
.read_rast <- function(x) {
  if (is.null(x)) return(NULL)
  if (inherits(x, "SpatRaster")) return(x)
  if (is.character(x)) return(terra::rast(x))
  stop("Unsupported raster input.", call. = FALSE)
}
.read_vect <- function(x) {
  if (is.null(x)) return(NULL)
  if (inherits(x, "SpatVector")) return(x)
  if (inherits(x, "sf")) return(terra::vect(x))
  if (is.character(x)) return(terra::vect(x))
  stop("Unsupported vector input.", call. = FALSE)
}
.is_raster_path <- function(x) {
  is.character(x) && grepl("\\.(tif|tiff|grd|asc|nc)$", tolower(x))
}

.fun_from_name <- function(name) {
  switch(
    name,
    mean = function(v, ...) mean(v, na.rm = TRUE),
    sum  = function(v, ...) sum(v, na.rm = TRUE),
    stop("Unknown aggregation function: ", name, call. = FALSE)
  )
}


.pa_has_coords <- function(x) {
  is.list(x$data) && !is.null(x$data$pu_coords) &&
    inherits(x$data$pu_coords, "data.frame") &&
    all(c("id","x","y") %in% names(x$data$pu_coords)) &&
    nrow(x$data$pu_coords) > 0
}

.pa_spatial_relations_summary <- function(x) {
  rels <- x$data$spatial_relations
  if (is.null(rels) || !is.list(rels) || length(rels) == 0) return(NULL)

  nm <- names(rels)
  if (is.null(nm)) nm <- rep("", length(rels))
  nm[nm == ""] <- paste0("rel_", seq_along(nm))

  out <- lapply(seq_along(rels), function(i) {
    r <- rels[[i]]
    if (!inherits(r, "data.frame") || nrow(r) == 0) {
      return(data.frame(name = nm[i], edges = 0L, w_min = NA_real_, w_max = NA_real_))
    }
    w <- suppressWarnings(as.numeric(r$weight))
    rng <- .pa_safe_range(w)
    data.frame(
      name  = nm[i],
      edges = as.integer(nrow(r)),
      w_min = if (is.null(rng)) NA_real_ else rng[[1]],
      w_max = if (is.null(rng)) NA_real_ else rng[[2]]
    )
  })

  do.call(rbind, out)
}


# -------------------------------------------------------------------------
# Internal helpers SOLVER
# -------------------------------------------------------------------------
.pa_get_solve_args <- function(x,
                               solver = NULL,
                               gap_limit = NULL,
                               time_limit = NULL,
                               solution_limit = NULL,
                               cores = NULL,
                               verbose = NULL,
                               name_output_file = NULL,
                               output_file = NULL,
                               solver_params = NULL) {

  stopifnot(inherits(x, "Problem"))

  defaults <- list(
    solver = "auto",
    gap_limit = 0.0,
    time_limit = .Machine$integer.max,
    solution_limit = FALSE,
    cores = 2L,
    verbose = TRUE,
    name_output_file = "output",
    output_file = TRUE,
    solver_params = list()
  )

  stored <- x$data$solve_args %||% list()
  if (!is.list(stored)) stored <- list()

  out <- utils::modifyList(defaults, stored)

  # ---- explicit overrides (optional)
  if (!is.null(gap_limit)) out$gap_limit <- gap_limit
  if (!is.null(time_limit)) out$time_limit <- time_limit
  if (!is.null(solution_limit)) out$solution_limit <- solution_limit
  if (!is.null(cores)) out$cores <- cores
  if (!is.null(verbose)) out$verbose <- verbose
  if (!is.null(name_output_file)) out$name_output_file <- name_output_file
  if (!is.null(output_file)) out$output_file <- output_file

  # solver override (NULL = not specified)
  if (!is.null(solver)) {
    solver <- as.character(solver)[1]
    if (!identical(solver, "")) out$solver <- solver
  }

  # solver_params override/merge
  if (!is.null(solver_params)) {
    if (!is.list(solver_params)) stop("solver_params must be a list.", call. = FALSE)
    out$solver_params <- utils::modifyList(out$solver_params %||% list(), solver_params)
  }

  # ---- normalize
  out$solver <- as.character(out$solver)[1]
  if (!out$solver %in% c("auto", "gurobi", "cplex", "cbc", "symphony")) {
    stop("Unknown solver in x$data$solve_args$solver: ", out$solver,
         ". Use one of: auto, gurobi, cplex, cbc, symphony.", call. = FALSE)
  }

  out$gap_limit <- round(as.numeric(out$gap_limit), 3)
  if (!is.finite(out$gap_limit) || out$gap_limit < 0 || out$gap_limit > 1) {
    stop("gap_limit must be a finite number in [0, 1].", call. = FALSE)
  }

  out$time_limit <- round(as.numeric(out$time_limit), 3)
  if (!is.finite(out$time_limit) || out$time_limit < 0) {
    stop("time_limit must be a finite number >= 0.", call. = FALSE)
  }

  out$cores <- as.integer(out$cores)
  if (!is.finite(out$cores) || out$cores < 1) {
    stop("cores must be an integer >= 1.", call. = FALSE)
  }

  # cap cores defensively (avoid hard error inside solve)
  max_cores <- parallel::detectCores(TRUE)
  if (is.finite(max_cores) && out$cores > max_cores) {
    out$cores <- as.integer(max_cores)
  }

  out$solution_limit <- isTRUE(out$solution_limit)
  out$verbose <- isTRUE(out$verbose)
  out$output_file <- isTRUE(out$output_file)

  out$name_output_file <- as.character(out$name_output_file)[1]
  if (!nzchar(out$name_output_file)) out$name_output_file <- "output"

  if (is.null(out$solver_params) || !is.list(out$solver_params)) out$solver_params <- list()

  out
}


# -------------------------------------------------------------------------
# Internal helpers SOLUTIONS
# -------------------------------------------------------------------------
.pa_extract_solution_tables <- function(x, sol, threshold = 0.5) {
  stopifnot(inherits(x, "Problem"))
  stopifnot(is.numeric(sol))

  # --- get model list (snapshot) or from ptr
  ml <- x$data$model_list %||% NULL
  if (is.null(ml)) {
    stopifnot(!is.null(x$data$model_ptr))
    ml <- rcpp_optimization_problem_as_list(x$data$model_ptr)
  }

  # --- fallbacks if model metadata missing
  pu_df <- x$data$pu
  da_df <- x$data$dist_actions_model %||% x$data$dist_actions
  df_df <- x$data$dist_features

  n_pu <- as.integer(ml$n_pu %||% nrow(pu_df) %||% 0L)
  n_x  <- as.integer(ml$n_x  %||% (if (inherits(da_df, "data.frame")) nrow(da_df) else 0L))
  n_z  <- as.integer(ml$n_z  %||% (if (inherits(df_df, "data.frame")) nrow(df_df) else 0L))

  w_off <- as.integer(ml$w_offset %||% 0L)
  x_off <- as.integer(ml$x_offset %||% 0L)
  z_off <- as.integer(ml$z_offset %||% 0L)

  slice_safe <- function(v, start1, end1) {
    if (length(v) == 0) return(numeric(0))
    if (is.na(start1) || is.na(end1) || start1 > end1) return(numeric(0))
    end1 <- min(end1, length(v))
    start1 <- min(max(1L, start1), end1)
    v[start1:end1]
  }

  # --- slices
  w  <- if (n_pu > 0L) slice_safe(sol, w_off + 1L, w_off + n_pu) else numeric(0)
  xv <- if (n_x  > 0L) slice_safe(sol, x_off + 1L, x_off + n_x ) else numeric(0)
  zv <- if (n_z  > 0L) slice_safe(sol, z_off + 1L, z_off + n_z ) else numeric(0)

  # -------------------------
  # 1) PU selection table
  # -------------------------
  pu_out <- pu_df

  if (!inherits(pu_out, "data.frame") || nrow(pu_out) == 0) {
    pu_out <- data.frame()
  } else {
    if (length(w) != nrow(pu_out)) {
      warning(
        "Mismatch: nrow(pu) = ", nrow(pu_out),
        " but length(w slice) = ", length(w),
        ". Setting pu$selected = NA.",
        call. = FALSE
      )
      pu_out$selected <- rep(NA_integer_, nrow(pu_out))
    } else {
      pu_out$selected <- as.integer(w > threshold)
    }
  }

  # -------------------------
  # 2) Actions selection table
  # -------------------------
  da_out <- da_df

  if (!inherits(da_out, "data.frame") || nrow(da_out) == 0) {
    da_out <- data.frame()
  } else {
    if (length(xv) != nrow(da_out)) {
      warning(
        "Mismatch: nrow(dist_actions_model) = ", nrow(da_out),
        " but length(x slice) = ", length(xv),
        ". Setting actions$selected = NA.",
        call. = FALSE
      )
      da_out$selected <- rep(NA_integer_, nrow(da_out))
    } else {
      da_out$selected <- as.integer(xv > threshold)
    }

    if (all(c("internal_pu", "internal_action") %in% names(da_out))) {
      key_da <- paste0(da_out$internal_pu, ":", da_out$internal_action)
      if (anyDuplicated(key_da)) {
        stop(
          "dist_actions_model has duplicated (internal_pu, internal_action) pairs. ",
          "This breaks mapping of effects->x. Fix upstream (ensure unique feasible pairs).",
          call. = FALSE
        )
      }
    }
  }

  # -------------------------
  # 3) Features achieved
  # -------------------------
  # -------------------------
  # 3) Features summary
  # -------------------------
  feats <- x$data$features
  de <- x$data$dist_effects_model %||% x$data$dist_effects

  # total available in the landscape (independent of the solution)
  base_by_feat <- data.frame(
    internal_feature = integer(0),
    total_available = numeric(0)
  )

  if (!is.null(df_df) && inherits(df_df, "data.frame") && nrow(df_df) > 0 &&
      all(c("internal_feature", "amount") %in% names(df_df))) {

    df2 <- df_df
    df2$amount <- as.numeric(df2$amount)

    base_by_feat <- stats::aggregate(
      total_available ~ internal_feature,
      data = data.frame(
        internal_feature = as.integer(df2$internal_feature),
        total_available = as.numeric(df2$amount)
      ),
      FUN = sum
    )
  }

  # map x rows to effects table
  de_with_x <- NULL
  if (!is.null(de) && inherits(de, "data.frame") && nrow(de) > 0 &&
      length(xv) > 0 &&
      all(c("internal_pu", "internal_action", "internal_feature") %in% names(de)) &&
      inherits(da_out, "data.frame") && nrow(da_out) > 0 &&
      all(c("internal_pu", "internal_action") %in% names(da_out))) {

    key_da <- paste0(da_out$internal_pu, ":", da_out$internal_action)
    map <- stats::setNames(seq_len(nrow(da_out)), key_da)

    key_de <- paste0(de$internal_pu, ":", de$internal_action)
    xrow <- unname(map[key_de])

    ok <- !is.na(xrow)
    if (!all(ok)) {
      warning(
        sum(!ok),
        " rows in dist_effects have no matching (pu,action) in dist_actions_model and were ignored.",
        call. = FALSE
      )
      de <- de[ok, , drop = FALSE]
      xrow <- xrow[ok]
    }

    ok2 <- xrow >= 1L & xrow <= length(xv)
    if (!all(ok2)) {
      warning(
        sum(!ok2),
        " mapped x rows fall outside the x slice (length(x) = ", length(xv), "); ignored.",
        call. = FALSE
      )
      de <- de[ok2, , drop = FALSE]
      xrow <- xrow[ok2]
    }

    de_with_x <- de
    de_with_x$x_value <- as.numeric(xv[xrow])

    if (!("benefit" %in% names(de_with_x))) de_with_x$benefit <- 0
    if (!("loss" %in% names(de_with_x))) de_with_x$loss <- 0

    de_with_x$benefit <- as.numeric(de_with_x$benefit)
    de_with_x$loss <- as.numeric(de_with_x$loss)

    de_with_x$selected_benefit <- de_with_x$benefit * de_with_x$x_value
    de_with_x$selected_loss <- de_with_x$loss * de_with_x$x_value
    de_with_x$selected_net <- de_with_x$selected_benefit - de_with_x$selected_loss
  }

  benefit_by_feat <- data.frame(
    internal_feature = integer(0),
    benefit = numeric(0)
  )

  loss_by_feat <- data.frame(
    internal_feature = integer(0),
    loss = numeric(0)
  )

  net_by_feat <- data.frame(
    internal_feature = integer(0),
    net = numeric(0)
  )

  if (!is.null(de_with_x) && nrow(de_with_x) > 0) {
    benefit_by_feat <- stats::aggregate(
      benefit ~ internal_feature,
      data = data.frame(
        internal_feature = as.integer(de_with_x$internal_feature),
        benefit = as.numeric(de_with_x$selected_benefit)
      ),
      FUN = sum
    )

    loss_by_feat <- stats::aggregate(
      loss ~ internal_feature,
      data = data.frame(
        internal_feature = as.integer(de_with_x$internal_feature),
        loss = as.numeric(de_with_x$selected_loss)
      ),
      FUN = sum
    )

    net_by_feat <- stats::aggregate(
      net ~ internal_feature,
      data = data.frame(
        internal_feature = as.integer(de_with_x$internal_feature),
        net = as.numeric(de_with_x$selected_net)
      ),
      FUN = sum
    )
  }

  # main feature table
  if (is.null(feats) || !inherits(feats, "data.frame") || nrow(feats) == 0) {
    feat_tbl <- data.frame()
  } else {
    feat_tbl <- data.frame(
      feature = as.integer(feats$internal_id),
      feature_name = as.character(feats$name),
      stringsAsFactors = FALSE
    )

    feat_tbl <- dplyr::left_join(feat_tbl, base_by_feat,    by = c("feature" = "internal_feature"))
    feat_tbl <- dplyr::left_join(feat_tbl, benefit_by_feat, by = c("feature" = "internal_feature"))
    feat_tbl <- dplyr::left_join(feat_tbl, loss_by_feat,    by = c("feature" = "internal_feature"))
    feat_tbl <- dplyr::left_join(feat_tbl, net_by_feat,     by = c("feature" = "internal_feature"))

    feat_tbl$total_available[is.na(feat_tbl$total_available)] <- 0
    feat_tbl$benefit[is.na(feat_tbl$benefit)] <- 0
    feat_tbl$loss[is.na(feat_tbl$loss)] <- 0
    feat_tbl$net[is.na(feat_tbl$net)] <- 0

    feat_tbl$total <- feat_tbl$total_available + feat_tbl$net
  }

  # -------------------------
  # 4) Targets summary
  # -------------------------
  tgt <- x$data$targets
  tgt_out <- NULL

  if (inherits(tgt, "data.frame") && nrow(tgt) > 0 &&
      all(c("feature", "type", "target_value") %in% names(tgt)) &&
      inherits(feat_tbl, "data.frame") && nrow(feat_tbl) > 0) {

    tgt2 <- tgt
    tgt2$feature <- as.integer(tgt2$feature)
    tgt2$type <- as.character(tgt2$type)
    tgt2$target_value <- as.numeric(tgt2$target_value)

    if (!("subset" %in% names(tgt2))) {
      tgt2$subset <- NA_character_
    }
    tgt2$subset <- as.character(tgt2$subset)

    # default achieved = NA
    tgt2$achieved <- NA_real_

    # helper to compute achieved by subset
    .achieved_for_subset <- function(feature_ids, subset_string) {
      if (is.null(de_with_x) || nrow(de_with_x) == 0) {
        return(stats::setNames(rep(0, length(feature_ids)), feature_ids))
      }

      dd <- de_with_x

      if (!is.na(subset_string) && nzchar(subset_string)) {
        matched <- .pa_resolve_action_subset(x, strsplit(subset_string, "\\|")[[1]])
        keep_actions <- as.integer(matched$internal_id)
        dd <- dd[dd$internal_action %in% keep_actions, , drop = FALSE]
      }

      if (nrow(dd) == 0) {
        return(stats::setNames(rep(0, length(feature_ids)), feature_ids))
      }

      tmp <- data.frame(
        feature = as.integer(dd$internal_feature),
        achieved = as.numeric(dd$selected_benefit)
      )
      agg <- stats::aggregate(achieved ~ feature, data = tmp, FUN = sum)

      out <- stats::setNames(rep(0, length(feature_ids)), feature_ids)
      hit <- match(feature_ids, agg$feature)
      ok <- !is.na(hit)
      out[ok] <- agg$achieved[hit[ok]]
      out
    }

    split_key <- ifelse(is.na(tgt2$subset) | !nzchar(tgt2$subset), "__ALL__", tgt2$subset)
    idx_split <- split(seq_len(nrow(tgt2)), split_key)

    for (nm in names(idx_split)) {
      ii <- idx_split[[nm]]
      subset_string <- tgt2$subset[ii][1]
      feats_i <- tgt2$feature[ii]
      achieved_i <- .achieved_for_subset(feats_i, subset_string)
      tgt2$achieved[ii] <- as.numeric(achieved_i[as.character(feats_i)])
    }

    tgt2$gap <- tgt2$achieved - tgt2$target_value
    tgt_out <- tgt2
  }

  list(
    pu = pu_out,
    actions = da_out,
    features = feat_tbl,
    targets = tgt_out
  )
}

# -------------------------------------------------------------------------
# Internal helpers AREAS
# -------------------------------------------------------------------------
.pa_get_area_vec <- function(x, area_col = NULL, area_unit = c("m2", "ha", "km2")) {
  stopifnot(inherits(x, "Problem"))
  area_unit <- match.arg(area_unit)

  pu <- x$data$pu
  if (is.null(pu) || !inherits(pu, "data.frame")) {
    stop("x$data$pu is missing or is not a data.frame.", call. = FALSE)
  }

  n_pu <- nrow(pu)
  if (n_pu <= 0L) {
    stop("x$data$pu has zero rows; cannot retrieve planning unit areas.", call. = FALSE)
  }

  area <- NULL
  area_source <- NULL
  area_col_used <- NULL

  # ---------------------------------------------------------
  # 1) explicit area_col
  # ---------------------------------------------------------
  if (!is.null(area_col)) {
    area_col <- as.character(area_col)[1]
    if (is.na(area_col) || !nzchar(area_col)) {
      stop("area_col must be a non-empty string or NULL.", call. = FALSE)
    }

    # 1a) from x$data$pu
    if (area_col %in% names(pu)) {
      area <- as.numeric(pu[[area_col]])
      area_source <- "pu"
      area_col_used <- area_col

      # 1b) from x$data$pu_sf
    } else if (!is.null(x$data$pu_sf) && inherits(x$data$pu_sf, "sf") &&
               area_col %in% names(x$data$pu_sf)) {
      area <- as.numeric(x$data$pu_sf[[area_col]])
      area_source <- "pu_sf"
      area_col_used <- area_col

    } else {
      stop(
        "area_col '", area_col, "' was not found in x$data$pu",
        if (!is.null(x$data$pu_sf) && inherits(x$data$pu_sf, "sf")) " or x$data$pu_sf." else ".",
        " Provide a valid area column or spatial information.",
        call. = FALSE
      )
    }
  }

  # ---------------------------------------------------------
  # 2) automatic lookup if area_col is NULL
  # ---------------------------------------------------------
  if (is.null(area)) {
    cand <- c("area", "Area", "AREA", "surf", "surface", "area_m2", "area_ha", "area_km2")
    hit_pu <- cand[cand %in% names(pu)][1]

    if (!is.na(hit_pu) && length(hit_pu) == 1L) {
      area <- as.numeric(pu[[hit_pu]])
      area_source <- "pu"
      area_col_used <- hit_pu
    }
  }

  if (is.null(area) && !is.null(x$data$pu_sf) && inherits(x$data$pu_sf, "sf")) {
    cand <- c("area", "Area", "AREA", "surf", "surface", "area_m2", "area_ha", "area_km2")
    hit_sf <- cand[cand %in% names(x$data$pu_sf)][1]

    if (!is.na(hit_sf) && length(hit_sf) == 1L) {
      area <- as.numeric(x$data$pu_sf[[hit_sf]])
      area_source <- "pu_sf"
      area_col_used <- hit_sf
    }
  }

  # ---------------------------------------------------------
  # 3) derive from geometry if possible
  # ---------------------------------------------------------
  if (is.null(area)) {
    if (!is.null(x$data$pu_sf) && inherits(x$data$pu_sf, "sf")) {
      area <- as.numeric(sf::st_area(x$data$pu_sf))
      area_source <- "geometry_sf"
      area_col_used <- NA_character_

      # future raster branch could go here
      # else if (...) {
      #   area <- ...
      #   area_source <- "geometry_raster"
      # }

    } else {
      stop(
        "Could not retrieve planning unit area.\n",
        "No usable area column was found in x$data$pu, and no spatial geometry is available ",
        "to derive area internally.\n",
        "Provide `area_col` explicitly or build the Problem object with spatial information.",
        call. = FALSE
      )
    }
  }

  # ---------------------------------------------------------
  # 4) validate length and values
  # ---------------------------------------------------------
  if (length(area) != n_pu) {
    stop(
      "Area vector length (", length(area), ") != n_pu (", n_pu, "). ",
      "Area source: ", area_source,
      if (!is.null(area_col_used) && !is.na(area_col_used)) paste0(" [column='", area_col_used, "']") else "",
      ".",
      call. = FALSE
    )
  }

  if (anyNA(area) || any(!is.finite(area))) {
    stop("Area vector contains NA or non-finite values.", call. = FALSE)
  }

  if (any(area < 0)) {
    stop("Area vector contains negative values.", call. = FALSE)
  }

  # ---------------------------------------------------------
  # 5) convert to requested output units
  # assume raw geometry/st_area is m2
  # if column name suggests ha/km2, convert accordingly
  # ---------------------------------------------------------
  area_m2 <- area

  if (!is.null(area_col_used) && !is.na(area_col_used)) {
    nm <- tolower(area_col_used)

    if (grepl("km2", nm, fixed = TRUE)) {
      area_m2 <- area * 1e6
    } else if (grepl("ha", nm, fixed = TRUE)) {
      area_m2 <- area * 1e4
    } else {
      # assume already in m2 for generic names like area/surf/surface
      area_m2 <- area
    }
  }

  out <- switch(
    area_unit,
    m2  = area_m2,
    ha  = area_m2 / 1e4,
    km2 = area_m2 / 1e6
  )

  out
}

# helper: ensure model built and model_list snapshot available
.pa_ensure_model_snapshot <- function(x) {
  stopifnot(inherits(x, "Problem"))

  if (is.null(x$data$model_ptr)) {
    x <- .pa_build_model(x)
  }
  if (is.null(x$data$model_list) || isTRUE(x$data$meta$model_dirty)) {
    x <- .pa_refresh_model_snapshot(x)
    x$data$meta$model_dirty <- FALSE
  }
  x
}


# helper: add 1 linear constraint sum(coeff_j * x_j) (sense) rhs
# NOTE: replace this with YOUR real C++ bridge that appends rows/triplets/rhs/sense.
.pa_add_linear_constraint <- function(
    x,
    var_index_0based,
    coeff,
    sense = c(">=", "<=", "=="),
    rhs,
    name = NULL,
    block_name = "linear_constraint",
    tag = "",
    drop_zeros = TRUE,
    merge_duplicates = TRUE,
    refresh_snapshot = FALSE
) {
  sense <- match.arg(sense)

  stopifnot(inherits(x, "Problem"))
  if (is.null(x$data$model_ptr)) {
    stop("Model pointer missing. Build model first.", call. = FALSE)
  }

  # ---- validate inputs
  if (!is.numeric(var_index_0based) || !is.numeric(coeff)) {
    stop("var_index_0based and coeff must be numeric.", call. = FALSE)
  }
  if (length(var_index_0based) != length(coeff)) {
    stop("Length mismatch: var_index_0based and coeff must have the same length.", call. = FALSE)
  }
  if (!is.numeric(rhs) || length(rhs) != 1L || !is.finite(rhs)) {
    stop("rhs must be a single finite numeric value.", call. = FALSE)
  }

  j0 <- as.integer(var_index_0based)
  a  <- as.numeric(coeff)

  # ---- drop NA / non-finite / zero
  keep <- !(is.na(j0) | is.na(a) | !is.finite(a))
  if (isTRUE(drop_zeros)) keep <- keep & (a != 0)

  j0 <- j0[keep]
  a  <- a[keep]

  if (length(j0) == 0L) {
    stop("Linear constraint has no non-zero coefficients after filtering.", call. = FALSE)
  }

  # ---- optionally merge duplicates (important if you build vectors by rbind/cbind)
  if (isTRUE(merge_duplicates) && length(j0) > 1L) {
    o <- order(j0)
    j0 <- j0[o]; a <- a[o]
    # aggregate by j0
    u <- unique(j0)
    a2 <- numeric(length(u))
    idx <- match(j0, u)
    for (k in seq_along(a)) a2[idx[k]] <- a2[idx[k]] + a[k]
    j0 <- u
    a  <- a2
    if (isTRUE(drop_zeros)) {
      keep2 <- a != 0
      j0 <- j0[keep2]
      a  <- a[keep2]
    }
    if (length(j0) == 0L) {
      stop("Linear constraint became empty after merging duplicates.", call. = FALSE)
    }
  }

  name <- as.character(name %||% "")[1]
  block_name <- as.character(block_name %||% "linear_constraint")[1]
  tag <- as.character(tag %||% "")[1]

  # ---- call C++
  out <- rcpp_add_linear_constraint(
    model_ptr  = x$data$model_ptr,
    j0         = j0,
    x          = a,
    sense      = sense,
    rhs        = as.numeric(rhs),
    name       = name,
    block_name = block_name,
    tag        = tag
  )

  if (isTRUE(refresh_snapshot)) {
    x <- .pa_refresh_model_snapshot(x)
  }

  # store last-added info (optional, but handy for debugging)
  x$data$model_registry <- x$data$model_registry %||% list()
  x$data$model_registry$last_linear_constraint <- out

  x
}




# -------------------------------------------------------------------------
# Internal helpers spatial relations
# -------------------------------------------------------------------------
.pa_validate_relation <- function(rel, n_pu, allow_self = FALSE,
                                  dup_agg = c("sum", "max", "min", "mean")) {

  stopifnot(is.data.frame(rel))
  dup_agg <- match.arg(dup_agg)

  req <- c("internal_pu1", "internal_pu2", "weight")
  miss <- setdiff(req, names(rel))
  if (length(miss) > 0) stop("Relation is missing columns: ", paste(miss, collapse = ", "), call. = FALSE)

  rel$internal_pu1 <- as.integer(rel$internal_pu1)
  rel$internal_pu2 <- as.integer(rel$internal_pu2)
  rel$weight <- as.numeric(rel$weight)

  if (anyNA(rel$internal_pu1) || anyNA(rel$internal_pu2) || anyNA(rel$weight)) {
    stop("Relation has NA in internal_pu1/internal_pu2/weight.", call. = FALSE)
  }
  if (any(rel$internal_pu1 < 1L | rel$internal_pu1 > n_pu)) stop("internal_pu1 out of range.", call. = FALSE)
  if (any(rel$internal_pu2 < 1L | rel$internal_pu2 > n_pu)) stop("internal_pu2 out of range.", call. = FALSE)
  if (!allow_self && any(rel$internal_pu1 == rel$internal_pu2)) stop("Self-edges are not allowed.", call. = FALSE)
  if (any(!is.finite(rel$weight)) || any(rel$weight < 0)) stop("weight must be finite and >= 0.", call. = FALSE)

  # undirected canonical ordering
  a <- pmin(rel$internal_pu1, rel$internal_pu2)
  b <- pmax(rel$internal_pu1, rel$internal_pu2)
  rel$internal_pu1 <- a
  rel$internal_pu2 <- b

  # columns to keep
  core <- c("internal_pu1", "internal_pu2", "weight")
  extra_cols <- intersect(names(rel), c("pu1","pu2","distance","source"))
  keep <- c(core, extra_cols)

  # aggregate duplicates
  key <- paste(rel$internal_pu1, rel$internal_pu2)
  if (anyDuplicated(key) != 0) {
    FUN <- switch(
      dup_agg,
      sum  = sum,
      max  = max,
      min  = min,
      mean = mean
    )

    # aggregate ONLY weight (core), then optionally restore extras deterministically
    agg <- stats::aggregate(
      weight ~ internal_pu1 + internal_pu2,
      data = rel[, core, drop = FALSE],
      FUN = FUN
    )

    # if extras exist, attach a deterministic representative per edge (first row)
    if (length(extra_cols) > 0) {
      # representative rows: first occurrence per key
      rep_idx <- match(paste(agg$internal_pu1, agg$internal_pu2),
                       paste(rel$internal_pu1, rel$internal_pu2))
      extras <- rel[rep_idx, extra_cols, drop = FALSE]
      agg <- cbind(agg, extras)
    }

    rel <- agg
  } else {
    rel <- rel[, keep, drop = FALSE]
  }

  rel
}



.pa_action_weights_vector <- function(actions_df,
                                      action_weights = NULL,
                                      subset_actions = NULL,
                                      default_weight = 1) {

  stopifnot(is.data.frame(actions_df), nrow(actions_df) > 0)

  n_actions <- nrow(actions_df)

  w <- rep(as.numeric(default_weight)[1], n_actions)

  if (!is.null(subset_actions)) {
    subset_actions <- as.integer(subset_actions)
    if (anyNA(subset_actions)) stop("subset_actions contains NA.", call. = FALSE)
    if (any(subset_actions < 1L | subset_actions > n_actions)) {
      stop("subset_actions out of range (1..n_actions).", call. = FALSE)
    }

    # si se especifica subset y NO se da vector completo, el default para acciones fuera del subset suele ser 0
    w[-subset_actions] <- 0
  }

  if (!is.null(action_weights)) {
    action_weights <- as.numeric(action_weights)
    if (any(!is.finite(action_weights)) || any(action_weights < 0)) {
      stop("action_weights must be finite and >= 0.", call. = FALSE)
    }

    if (length(action_weights) == n_actions) {
      w <- action_weights

    } else if (!is.null(subset_actions) && length(action_weights) == length(subset_actions)) {
      w[subset_actions] <- action_weights

    } else {
      stop(
        "action_weights must have length n_actions (= ", n_actions, ") ",
        "or length(subset_actions) (= ", length(subset_actions) %||% 0, ").",
        call. = FALSE
      )
    }
  }

  w
}


.pa_model_frag_vars_summary <- function(x) {
  if (!inherits(x, "Problem")) return(NULL)
  ml <- x$data$model_list
  if (is.null(ml) || !is.list(ml)) return(NULL)

  # defensivo: algunos campos pueden no existir
  get_int0 <- function(nm) {
    v <- ml[[nm]]
    if (is.null(v)) return(0L)
    as.integer(v)[1]
  }

  out <- list(
    n_y_pu            = get_int0("n_y_pu"),
    n_y_actions       = get_int0("n_y_actions"),
    n_y_interventions = get_int0("n_y_interventions"),
    y_pu_offset            = get_int0("y_pu_offset"),
    y_actions_offset       = get_int0("y_actions_offset"),
    y_interventions_offset = get_int0("y_interventions_offset")
  )

  # si no hay ninguna y*, devuelve NULL para no ensuciar print
  if ((out$n_y_pu + out$n_y_actions + out$n_y_interventions) == 0L) return(NULL)
  out
}


# -------------------------------------------------------------------------
# Internal helpers multi objectives
# -------------------------------------------------------------------------

.pa_abort <- function(...) stop(paste0(...), call. = FALSE)

.pa_apply_runtime_updates_to_model <- function(model, x) {

  upd <- x$data$runtime_updates %||% list()
  reg <- x$data$model_registry %||% list()

  # nada que hacer
  if (length(upd) == 0) return(model)

  bigM <- as.numeric(upd$bigM %||% 1e15)[1]

  # Helper: relajar una fila para que sea redundante
  relax_row <- function(i) {
    s <- model$sense[i]
    if (identical(s, ">=")) {
      model$rhs[i] <<- -bigM
    } else if (identical(s, "<=")) {
      model$rhs[i] <<-  bigM
    } else if (identical(s, "==") || identical(s, "=")) {
      model$sense[i] <<- "<="
      model$rhs[i]   <<-  bigM
    } else {
      # por seguridad: si aparece algo raro, lo relajo como <= bigM
      model$sense[i] <<- "<="
      model$rhs[i]   <<-  bigM
    }
    invisible(NULL)
  }

  # 1) activar/desactivar constraints por grupos (si el registry lo soporta)
  # upd$deactivate_constraints puede ser vector de ids lógicos (por nombre) o filas directas
  if (!is.null(upd$deactivate_rows)) {
    rows <- as.integer(upd$deactivate_rows)
    rows <- rows[rows >= 1L & rows <= length(model$rhs)]
    for (i in rows) relax_row(i)
  }

  if (!is.null(upd$deactivate_groups) && !is.null(reg$cons)) {
    gs <- as.character(upd$deactivate_groups)
    for (g in gs) {
      rows <- reg$cons[[g]] %||% integer(0)
      rows <- as.integer(rows)
      rows <- rows[rows >= 1L & rows <= length(model$rhs)]
      for (i in rows) relax_row(i)
    }
  }

  # 2) epsilons: update RHS de constraints epsilon (sin cambiar estructura)
  # ejemplo: upd$epsilon <- list(f2 = 0.25, f3 = 10)
  if (!is.null(upd$epsilon) && !is.null(reg$cons$epsilon)) {
    eps_list <- upd$epsilon
    for (nm in names(eps_list)) {
      rows <- reg$cons$epsilon[[nm]] %||% integer(0)
      rows <- as.integer(rows)
      if (length(rows) == 0) next
      rhs <- as.numeric(eps_list[[nm]])[1]
      # aquí asumes que la fila es del tipo f(x) >= epsilon
      model$rhs[rows] <- rhs
    }
  }

  # 3A) objetivo: override directo (vector obj completo)
  # Esto es lo que necesita multiscape (weighted, etc.)
  if (!is.null(upd$obj)) {
    obj_new <- as.numeric(upd$obj)

    if (length(obj_new) != length(model$obj)) {
      stop(
        "runtime_updates$obj length mismatch: ",
        length(obj_new), " != ", length(model$obj),
        call. = FALSE
      )
    }

    model$obj <- obj_new

    if (!is.null(upd$modelsense)) {
      ms <- as.character(upd$modelsense)[1]
      if (!ms %in% c("min", "max")) stop("runtime_updates$modelsense must be 'min' or 'max'.", call. = FALSE)
      model$modelsense <- ms
    }
  }
  else if (!is.null(upd$objective_template) && !is.null(reg$obj_templates)) {
    tpl <- reg$obj_templates[[upd$objective_template]]
    if (!is.null(tpl)) {
      if (length(tpl$obj) != length(model$obj)) {
        stop("Objective template length mismatch.", call. = FALSE)
      }
      model$obj <- tpl$obj
      if (!is.null(tpl$modelsense)) model$modelsense <- tpl$modelsense
    }
  }

  # 4) pesos AUGMECON: por ejemplo coef de slacks en obj (sin reconstruir)
  if (!is.null(upd$slack_weight) && !is.null(reg$vars$slack)) {
    w <- as.numeric(upd$slack_weight)[1]
    slack_ids <- as.integer(reg$vars$slack)
    slack_ids <- slack_ids[slack_ids >= 1L & slack_ids <= length(model$obj)]
    model$obj[slack_ids] <- w
  }

  model
}


# -------------------------------------------------------------------------
# Internal helpers spatial relations
# -------------------------------------------------------------------------
.pa_create_problem_tabular_impl <- function(pu, features, dist_features, boundary = NULL, ...) {

  dots <- list(...)
  `%||%` <- function(a, b) if (!is.null(a)) a else b

  pu_coords <- NULL

  # helper: coerce ids to integer safely
  .as_int_id <- function(x, what) {
    if (is.factor(x)) x <- as.character(x)
    if (is.character(x)) {
      if (any(grepl("[^0-9\\-]", x))) {
        stop(what, " must be numeric/integer ids (got non-numeric strings).", call. = FALSE)
      }
      x <- as.integer(x)
    } else {
      x <- as.integer(x)
    }
    if (anyNA(x)) stop(what, " contains NA after coercion to integer.", call. = FALSE)
    x
  }

  # keep raw PU table for downstream helpers such as add_locked_pu()
  pu_raw <- pu

  # =========================
  # PU: validate + normalize
  # =========================
  assertthat::assert_that(
    inherits(pu, "data.frame"),
    assertthat::has_name(pu, "id"),
    nrow(pu) > 0,
    assertthat::noNA(pu$id)
  )

  pu$id <- .as_int_id(pu$id, "pu$id")
  if (anyDuplicated(pu$id) != 0) stop("pu$id must be unique.", call. = FALSE)

  # accept cost or monitoring_cost -> normalize to cost
  if ("cost" %in% names(pu)) {
    assertthat::assert_that(is.numeric(pu$cost), assertthat::noNA(pu$cost))
  } else if ("monitoring_cost" %in% names(pu)) {
    assertthat::assert_that(is.numeric(pu$monitoring_cost), assertthat::noNA(pu$monitoring_cost))
    pu$cost <- pu$monitoring_cost
  } else {
    stop("pu must contain either a 'cost' column or a 'monitoring_cost' column.", call. = FALSE)
  }

  # IMPORTANT:
  # internal contract: pu always carries locked_in / locked_out,
  # defaulting to FALSE when they are not provided.
  if (!("locked_in" %in% names(pu)))  pu$locked_in  <- FALSE
  if (!("locked_out" %in% names(pu))) pu$locked_out <- FALSE

  pu$locked_in  <- as.logical(pu$locked_in)
  pu$locked_out <- as.logical(pu$locked_out)

  pu$locked_in[is.na(pu$locked_in)] <- FALSE
  pu$locked_out[is.na(pu$locked_out)] <- FALSE

  if (any(pu$locked_in & pu$locked_out, na.rm = TRUE)) {
    stop("Some planning units are both locked_in and locked_out. Please fix pu input.", call. = FALSE)
  }

  # optional: store coordinates if present (tabular users)
  if (all(c("x", "y") %in% names(pu))) {
    pu_coords <- data.frame(
      id = pu$id,
      x  = as.numeric(pu$x),
      y  = as.numeric(pu$y),
      stringsAsFactors = FALSE
    )
    if (any(!is.finite(pu_coords$x) | !is.finite(pu_coords$y))) {
      stop("pu$x/pu$y contain non-finite values.", call. = FALSE)
    }
  }

  # internal PU table keeps the core fields needed by the model
  pu <- pu[, c("id", "cost", "locked_in", "locked_out"), drop = FALSE]
  pu <- pu[order(pu$id), , drop = FALSE]

  # raw PU table aligned to normalized id order
  pu_raw$id <- .as_int_id(pu_raw$id, "pu_raw$id")
  pu_raw <- pu_raw[match(pu$id, pu_raw$id), , drop = FALSE]

  # internal ids + lookup
  pu$internal_id <- seq_len(nrow(pu))
  pu_index <- stats::setNames(pu$internal_id, as.character(pu$id))

  # =========================
  # FEATURES: validate + normalize
  # =========================
  assertthat::assert_that(
    inherits(features, "data.frame"),
    assertthat::has_name(features, "id"),
    nrow(features) > 0,
    assertthat::noNA(features$id)
  )

  features$id <- .as_int_id(features$id, "features$id")
  if (anyDuplicated(features$id) != 0) stop("features$id must be unique.", call. = FALSE)

  if (!("name" %in% names(features))) {
    features$name <- paste0("feature.", seq_len(nrow(features)))
  } else {
    features$name <- as.character(features$name)
    assertthat::assert_that(assertthat::noNA(features$name))
    if (anyDuplicated(features$name) != 0) stop("features$name must be unique.", call. = FALSE)
  }

  features <- features[, c("id", "name", setdiff(names(features), c("id", "name"))), drop = FALSE]
  features <- features[order(features$id), , drop = FALSE]

  features$internal_id <- seq_len(nrow(features))
  feature_index <- stats::setNames(features$internal_id, as.character(features$id))

  # =========================
  # DIST_FEATURES: validate + normalize
  # =========================
  assertthat::assert_that(
    inherits(dist_features, "data.frame"),
    assertthat::has_name(dist_features, "pu"),
    assertthat::has_name(dist_features, "feature"),
    assertthat::has_name(dist_features, "amount"),
    nrow(dist_features) > 0,
    assertthat::noNA(dist_features$pu),
    assertthat::noNA(dist_features$feature),
    assertthat::noNA(dist_features$amount),
    is.numeric(dist_features$amount),
    all(dist_features$amount >= 0)
  )

  dist_features$pu      <- .as_int_id(dist_features$pu, "dist_features$pu")
  dist_features$feature <- .as_int_id(dist_features$feature, "dist_features$feature")

  if (!all(dist_features$pu %in% pu$id)) {
    bad <- unique(dist_features$pu[!dist_features$pu %in% pu$id])
    stop("dist_features contains unknown PU ids: ", paste(bad, collapse = ", "), call. = FALSE)
  }
  if (!all(dist_features$feature %in% features$id)) {
    bad <- unique(dist_features$feature[!dist_features$feature %in% features$id])
    stop("dist_features contains unknown feature ids: ", paste(bad, collapse = ", "), call. = FALSE)
  }

  dist_features <- dist_features[dist_features$amount != 0, , drop = FALSE]

  key <- paste(dist_features$pu, dist_features$feature, sep = "||")
  if (anyDuplicated(key) != 0) stop("There are duplicate (pu, feature) pairs in dist_features.", call. = FALSE)

  dist_features$internal_pu      <- unname(pu_index[as.character(dist_features$pu)])
  dist_features$internal_feature <- unname(feature_index[as.character(dist_features$feature)])

  dist_features <- dist_features[order(dist_features$internal_pu, dist_features$internal_feature), , drop = FALSE]
  dist_features$internal_row <- seq_len(nrow(dist_features))

  # =========================
  # BOUNDARY: validate + normalize
  # =========================
  assertthat::assert_that(is.null(boundary) || inherits(boundary, "data.frame"))
  if (inherits(boundary, "data.frame")) {
    assertthat::assert_that(
      assertthat::has_name(boundary, "id1"),
      assertthat::has_name(boundary, "id2"),
      assertthat::has_name(boundary, "boundary"),
      assertthat::noNA(boundary$id1),
      assertthat::noNA(boundary$id2),
      assertthat::noNA(boundary$boundary),
      is.numeric(boundary$boundary)
    )

    boundary$id1 <- .as_int_id(boundary$id1, "boundary$id1")
    boundary$id2 <- .as_int_id(boundary$id2, "boundary$id2")

    if (!all(boundary$id1 %in% pu$id) || !all(boundary$id2 %in% pu$id)) {
      warning("boundary contains PU ids not present in pu; they will be removed.", call. = FALSE, immediate. = TRUE)
      keep <- boundary$id1 %in% pu$id & boundary$id2 %in% pu$id
      boundary <- boundary[keep, , drop = FALSE]
    }

    if (nrow(boundary) == 0) boundary <- NULL
  }

  # =========================
  # useful warnings
  # =========================
  dif_pu <- setdiff(unique(pu$id), unique(dist_features$pu))
  if (length(dif_pu) != 0L) {
    warning(
      paste0("The following pu's do not contain features: ", paste(dif_pu, collapse = " ")),
      call. = FALSE, immediate. = TRUE
    )
  }

  dif_features <- setdiff(unique(features$id), unique(dist_features$feature))
  if (length(dif_features) != 0L) {
    warning(
      paste0("The following features are not represented in dist_features: ", paste(dif_features, collapse = " ")),
      call. = FALSE, immediate. = TRUE
    )
  }

  # =========================
  # build Problem object
  # =========================
  x <- pproto(
    NULL, Problem,
    data = list(
      pu = pu,
      pu_data_raw = pu_raw,
      features = features,
      dist_features = dist_features,

      # ---- spatial / auxiliary storage
      pu_coords = pu_coords,
      spatial_relations = list(),

      index = list(
        pu = pu_index,
        feature = feature_index,
        feature_name_to_id = stats::setNames(features$id, features$name)
      ),

      meta = list(
        dist_features_meaning = "baseline_amount",
        dist_benefit_meaning  = "delta_by_default"
      ),

      # workflow placeholders
      actions = NULL,
      dist_actions = NULL,
      dist_benefit = NULL,
      locked_actions = NULL,
      targets = NULL,
      objective = NULL,
      decisions = NULL,
      solver = NULL
    )
  )

  # =========================
  # boundary -> spatial relation ("boundary")
  # =========================
  if (!is.null(boundary) && inherits(boundary, "data.frame") && nrow(boundary) > 0) {

    rel <- data.frame(
      internal_pu1 = unname(pu_index[as.character(boundary$id1)]),
      internal_pu2 = unname(pu_index[as.character(boundary$id2)]),
      weight       = as.numeric(boundary$boundary),
      source       = "boundary_table",
      stringsAsFactors = FALSE
    )

    if (anyNA(rel$internal_pu1) || anyNA(rel$internal_pu2)) {
      stop("boundary contains PU ids not present in pu (after filtering).", call. = FALSE)
    }

    rel <- .pa_validate_relation(rel, n_pu = nrow(pu), allow_self = FALSE, dup_agg = "sum")
    rel$relation_name <- "boundary"

    if (is.null(x$data$spatial_relations) || !is.list(x$data$spatial_relations)) {
      x$data$spatial_relations <- list()
    }
    x$data$spatial_relations[["boundary"]] <- rel
  }

  x
}

# -------------------------------------------------------------------------
# Internal helpers objective relations
# -------------------------------------------------------------------------
.pa_register_objective <- function(
    x,
    alias,
    objective_id,
    model_type,
    objective_args = list(),
    sense = c("min", "max")
) {
  stopifnot(inherits(x, "Problem"))

  if (is.null(alias)) return(x)

  alias <- as.character(alias)[1]
  if (is.na(alias) || !nzchar(alias)) {
    stop("alias must be a non-empty string.", call. = FALSE)
  }

  objective_id <- as.character(objective_id)[1]
  if (is.na(objective_id) || !nzchar(objective_id)) {
    stop("objective_id must be a non-empty string.", call. = FALSE)
  }

  model_type <- as.character(model_type)[1]
  if (is.na(model_type) || !nzchar(model_type)) {
    stop("model_type must be a non-empty string.", call. = FALSE)
  }

  sense <- match.arg(sense)

  if (is.null(objective_args)) {
    objective_args <- list()
  }
  if (!is.list(objective_args)) {
    stop("objective_args must be a list.", call. = FALSE)
  }

  if (is.null(x$data$objectives) || !is.list(x$data$objectives)) {
    x$data$objectives <- list()
  }

  # prevent duplicated aliases
  if (!is.null(x$data$objectives[[alias]])) {
    stop(
      "Objective alias '", alias, "' already exists. ",
      "Use a different alias.",
      call. = FALSE
    )
  }

  x$data$objectives[[alias]] <- list(
    alias = alias,
    objective_id = objective_id,
    model_type = model_type,
    objective_args = objective_args,
    sense = sense,
    created_at = as.character(Sys.time())
  )

  # keep deterministic ordering by alias
  x$data$objectives <- x$data$objectives[sort(names(x$data$objectives))]

  # if a model already exists, mark dirty
  if (!is.null(x$data$model_ptr)) {
    x$data$meta <- x$data$meta %||% list()
    x$data$meta$model_dirty <- TRUE
  }

  x
}

# internal registry for MO objectives inside Problem
.pa_init_objectives <- function(x) {
  if (is.null(x$data$objectives) || !is.list(x$data$objectives)) {
    x$data$objectives <- list()
  }
  x
}


.pa_get_objective_specs <- function(x) {
  if (is.null(x$data$objectives)) return(list())
  x$data$objectives
}



.pa_set_objective_linear <- function(x, obj, modelsense = c("min", "max")) {
  stopifnot(inherits(x, "Problem"))
  modelsense <- match.arg(modelsense)

  # guarda runtime update para que .pa_apply_runtime_updates_to_model() lo aplique
  x$data$runtime_updates <- x$data$runtime_updates %||% list()
  x$data$runtime_updates$obj <- as.numeric(obj)
  x$data$runtime_updates$modelsense <- modelsense

  # marca "dirty" para que se regenere model_list desde ptr si aplica (o solo runtime update)
  x$data$meta$model_dirty <- TRUE
  x
}

.pa_mark_mo_needs <- function(x, needs) {
  stopifnot(inherits(x, "Problem"))
  x$data$mo <- x$data$mo %||% list()
  x$data$mo$needs <- utils::modifyList(x$data$mo$needs %||% list(), needs)
  x$data$meta$model_dirty <- TRUE
  x
}


.pa_fast_extract <- function(x, y, fun = c("sum", "mean", "max")) {
  fun <- match.arg(fun)

  stopifnot(inherits(x, "terra::SpatRaster") || inherits(x, "SpatRaster"))
  stopifnot(inherits(y, "terra::SpatVector") || inherits(y, "SpatVector") || inherits(y, "sf"))

  if (inherits(y, "SpatVector")) {
    if (!requireNamespace("sf", quietly = TRUE)) stop("Need 'sf' to convert SpatVector -> sf.", call. = FALSE)
    y <- sf::st_as_sf(y)
  }

  out <- matrix(NA_real_, nrow = nrow(y), ncol = terra::nlyr(x))
  colnames(out) <- make.unique(names(x))

  geomc <- as.character(sf::st_geometry_type(y, by_geometry = TRUE))

  # points
  point_idx <- grepl("POINT", geomc, fixed = TRUE)
  if (any(point_idx)) {
    ex <- terra::extract(x = x, y = sf::st_coordinates(y[point_idx, , drop = FALSE]))
    out[point_idx, ] <- as.matrix(ex[, -1, drop = FALSE]) # drop ID column
  }

  # lines
  line_idx <- grepl("LINE", geomc, fixed = TRUE)
  if (any(line_idx)) {
    fun2 <- switch(fun, mean = mean, sum = sum, max = max)
    out[line_idx, ] <- as.matrix(
      terra::extract(
        x = x,
        y = terra::vect(y[line_idx, , drop = FALSE]),
        ID = FALSE,
        touches = TRUE,
        fun = fun2,
        na.rm = TRUE
      )
    )
  }

  # polygons (prioritizr-like)
  poly_idx <- grepl("POLYGON", geomc, fixed = TRUE)
  if (any(poly_idx)) {
    if (!requireNamespace("exactextractr", quietly = TRUE)) {
      stop("Polygon feature extraction requires 'exactextractr'.", call. = FALSE)
    }

    # IMPORTANT: avoid mutating user's objects
    y2 <- y[poly_idx, , drop = FALSE]
    x2 <- x

    # (this is just to silence CRS/proj warnings like prioritizr; extraction uses coordinates)
    sf::st_crs(y2) <- sf::st_crs(NA_character_)
    terra::crs(x2) <- NA_character_

    out[poly_idx, ] <- as.matrix(
      exactextractr::exact_extract(
        x2,
        y2,
        fun = fun,
        progress = FALSE
      )
    )
  }

  out[!is.finite(out)] <- 0
  out[abs(out) < 1e-10] <- 0
  out
}


# -------------------------------------------------------------------------
# Internal helpers SPATIAL relations
# -------------------------------------------------------------------------
.pa_has_sf <- function() requireNamespace("sf", quietly = TRUE)

.pa_ensure_pu_index <- function(x) {
  stopifnot(inherits(x, "Problem"))
  if (is.null(x$data$pu) || !inherits(x$data$pu, "data.frame")) {
    stop("x$data$pu is missing. Create the problem with create_problem().", call. = FALSE)
  }
  if (is.null(x$data$pu$internal_id)) x$data$pu$internal_id <- seq_len(nrow(x$data$pu))
  if (is.null(x$data$pu$id)) {
    stop("x$data$pu must contain column 'id' (planning unit id).", call. = FALSE)
  }
  x$data$pu$id <- as.integer(x$data$pu$id)
  x$data$pu$internal_id <- as.integer(x$data$pu$internal_id)
  if (anyNA(x$data$pu$id) || anyNA(x$data$pu$internal_id)) {
    stop("x$data$pu$id/internal_id contain NA after coercion.", call. = FALSE)
  }
  if (anyDuplicated(x$data$pu$internal_id) != 0) stop("x$data$pu$internal_id must be unique.", call. = FALSE)
  if (anyDuplicated(x$data$pu$id) != 0) stop("x$data$pu$id must be unique.", call. = FALSE)
  if (is.null(x$data$index) || !is.list(x$data$index)) x$data$index <- list()
  x$data$index$pu <- stats::setNames(x$data$pu$internal_id, as.character(x$data$pu$id))
  x
}

.pa_store_relation <- function(x, rel, name) {
  stopifnot(inherits(x, "Problem"))
  if (is.null(x$data$spatial_relations) || !is.list(x$data$spatial_relations)) {
    x$data$spatial_relations <- list()
  }
  x$data$spatial_relations[[name]] <- rel
  x
}


.pa_coords_from_input <- function(x, coords = NULL) {
  stopifnot(inherits(x, "Problem"))

  if (!is.null(coords)) {
    if (inherits(coords, "data.frame")) {
      if (!all(c("id", "x", "y") %in% names(coords))) {
        stop("coords data.frame must contain columns id, x, y.", call. = FALSE)
      }
      out <- coords[, c("id", "x", "y")]
      out$id <- as.integer(out$id)
      out$x  <- as.numeric(out$x)
      out$y  <- as.numeric(out$y)
      return(out)
    }
    if (is.matrix(coords)) {
      if (ncol(coords) < 2) stop("coords matrix must have at least 2 columns (x,y).", call. = FALSE)
      out <- data.frame(id = x$data$pu$id, x = coords[, 1], y = coords[, 2])
      out$id <- as.integer(out$id)
      out$x  <- as.numeric(out$x)
      out$y  <- as.numeric(out$y)
      return(out)
    }
    stop("Unsupported coords type. Use data.frame(id,x,y) or a matrix with 2 columns.", call. = FALSE)
  }

  # NEW fallback: x$data$pu_coords
  if (!is.null(x$data$pu_coords) && inherits(x$data$pu_coords, "data.frame")) {
    pc <- x$data$pu_coords
    if (all(c("id", "x", "y") %in% names(pc))) {
      out <- pc[, c("id", "x", "y")]
      out$id <- as.integer(out$id)
      out$x  <- as.numeric(out$x)
      out$y  <- as.numeric(out$y)
      return(out)
    }
  }

  # fallback: try x$data$pu columns
  if (all(c("x", "y") %in% names(x$data$pu))) {
    out <- data.frame(id = x$data$pu$id, x = x$data$pu$x, y = x$data$pu$y)
    out$id <- as.integer(out$id)
    out$x  <- as.numeric(out$x)
    out$y  <- as.numeric(out$y)
    return(out)
  }

  stop(
    "No coordinates available. Provide coords=data.frame(id,x,y) or store x$data$pu_coords (id,x,y), or add x$data$pu$x/y.",
    call. = FALSE
  )
}


.pa_get_pu_sf_aligned <- function(x, pu_sf = NULL, arg_name = "pu_sf") {
  stopifnot(inherits(x, "Problem"))
  x <- .pa_ensure_pu_index(x)

  if (is.null(pu_sf)) pu_sf <- x$data$pu_sf

  if (is.null(pu_sf)) {
    stop(
      arg_name, " is NULL and x$data$pu_sf is missing.\n",
      "Provide ", arg_name, " (sf polygons with an 'id' column) or make sure create_problem() stored x$data$pu_sf.",
      call. = FALSE
    )
  }

  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("This function requires the 'sf' package.", call. = FALSE)
  }

  if (!inherits(pu_sf, "sf")) stop(arg_name, " must be an sf object.", call. = FALSE)
  if (!("id" %in% names(pu_sf))) stop(arg_name, " must contain an 'id' column.", call. = FALSE)

  pu_sf$id <- as.integer(pu_sf$id)

  # align to x$data$pu$id (critical)
  ord <- match(x$data$pu$id, pu_sf$id)
  if (anyNA(ord)) {
    missing_ids <- x$data$pu$id[is.na(ord)]
    stop(
      arg_name, "$id does not match x$data$pu$id (some ids missing). Missing: ",
      paste(utils::head(missing_ids, 20), collapse = ", "),
      if (length(missing_ids) > 20) " ..." else "",
      call. = FALSE
    )
  }

  pu_sf[ord, , drop = FALSE]
}

.pa_rbind_fill <- function(a, b) {
  stopifnot(inherits(a, "data.frame"), inherits(b, "data.frame"))
  cols <- union(names(a), names(b))
  for (cc in setdiff(cols, names(a))) a[[cc]] <- NA
  for (cc in setdiff(cols, names(b))) b[[cc]] <- NA
  a <- a[, cols, drop = FALSE]
  b <- b[, cols, drop = FALSE]
  rbind(a, b)
}

.pa_swap_edges <- function(rel) {
  out <- rel
  out$internal_pu1 <- rel$internal_pu2
  out$internal_pu2 <- rel$internal_pu1
  if ("pu1" %in% names(rel) && "pu2" %in% names(rel)) {
    out$pu1 <- rel$pu2
    out$pu2 <- rel$pu1
  }
  out
}


# -------------------------------------------------------------------------
# Needs flags (feature requirements) for build
# -------------------------------------------------------------------------

.pa_needs_default <- function() {
  list(
    z = FALSE,                # z vars (representation)
    y_pu = FALSE,             # PU fragmentation auxiliaries
    y_action = FALSE,         # action fragmentation auxiliaries
    y_intervention = FALSE,   # intervention fragmentation auxiliaries
    u_intervention = FALSE    # intervention extra vars (if applicable)
  )
}

.pa_build_model_init_needs <- function(x) {
  stopifnot(inherits(x, "Problem"))
  if (is.null(x$data$model_args) || !is.list(x$data$model_args)) x$data$model_args <- list()

  needs <- x$data$model_args$needs
  if (is.null(needs) || !is.list(needs)) needs <- list()

  # merge defaults without overwriting existing values
  def <- .pa_needs_default()
  for (nm in names(def)) {
    if (is.null(needs[[nm]])) needs[[nm]] <- def[[nm]]
  }

  x$data$model_args$needs <- needs
  x
}

.pa_build_model_set_needs_from_objective <- function(x) {
  stopifnot(inherits(x, "Problem"))

  args  <- x$data$model_args %||% list()
  mtype <- args$model_type %||% "minimizeCosts"
  oargs <- args$objective_args %||% list()
  raw_needs <- args$needs %||% list()

  needs <- .pa_needs_default()

  # decisión de diseño actual: z siempre disponible
  needs$z <- TRUE

  # primero respetamos overrides explícitos si vienen dados
  if (!is.null(raw_needs$z))              needs$z <- isTRUE(raw_needs$z)
  if (!is.null(raw_needs$y_pu))           needs$y_pu <- isTRUE(raw_needs$y_pu)
  if (!is.null(raw_needs$y_action))       needs$y_action <- isTRUE(raw_needs$y_action)
  if (!is.null(raw_needs$y_intervention)) needs$y_intervention <- isTRUE(raw_needs$y_intervention)
  if (!is.null(raw_needs$u_intervention)) needs$u_intervention <- isTRUE(raw_needs$u_intervention)

  # luego inferimos solo lo que no vino explícito
  if (is.null(raw_needs$y_pu)) {
    needs$y_pu <- identical(mtype, "minimizeFragmentation")
  }

  if (is.null(raw_needs$y_action)) {
    needs$y_action <- identical(mtype, "minimizeActionFragmentation")
  }

  if (is.null(raw_needs$y_intervention)) {
    needs$y_intervention <- identical(mtype, "minimizeInterventionFragmentation")
  }

  if (is.null(raw_needs$u_intervention)) {
    needs$u_intervention <- identical(mtype, "minimizeInterventionImpact")
  }

  # propagación explícita de metadatos estructurales necesarios
  if (isTRUE(needs$y_pu) || isTRUE(needs$y_action) || isTRUE(needs$y_intervention)) {
    if (is.null(raw_needs$relation_name)) {
      needs$relation_name <- as.character(oargs$relation_name %||% "boundary")[1]
    }
  }

  if (isTRUE(needs$y_action)) {
    if (is.null(raw_needs$actions_to_use)) {
      needs$actions_to_use <- oargs$actions_to_use %||% oargs$actions %||% NULL
    }
  }

  if (isTRUE(needs$u_intervention)) {
    if (is.null(raw_needs$u_intervention_actions)) {
      needs$u_intervention_actions <- oargs$actions %||% NULL
    }
  }

  # preserva metadatos extra útiles para MO
  extra_fields <- setdiff(names(raw_needs), names(needs))
  if (length(extra_fields) > 0L) {
    for (nm in extra_fields) {
      needs[[nm]] <- raw_needs[[nm]]
    }
  }

  needs$z              <- isTRUE(needs$z)
  needs$y_pu           <- isTRUE(needs$y_pu)
  needs$y_action       <- isTRUE(needs$y_action)
  needs$y_intervention <- isTRUE(needs$y_intervention)
  needs$u_intervention <- isTRUE(needs$u_intervention)

  x$data$model_args$needs <- needs
  x
}





.pa_build_model_prepare_needs_cpp <- function(x) {
  stopifnot(inherits(x, "Problem"))
  .pa_abort <- function(...) stop(paste0(...), call. = FALSE)

  op    <- x$data$model_ptr
  args  <- x$data$model_args %||% list()
  needs <- args$needs %||% list()
  oargs <- args$objective_args %||% list()

  mo_mode <- isTRUE(args$mo_mode)

  if (isTRUE(mo_mode) && isTRUE(needs$y_intervention)) {
    .pa_abort(
      "MO phase 1 does not support preparing y_intervention in the shared superset yet."
    )
  }

  if (!isTRUE(needs$y_pu) &&
      !isTRUE(needs$y_action) &&
      !isTRUE(needs$y_intervention) &&
      !isTRUE(needs$u_intervention)) {
    return(x)
  }

  x$data$model_registry <- x$data$model_registry %||% list()
  x$data$model_registry$vars <- x$data$model_registry$vars %||% list()

  .pa_prepare_relation_model <- function(rel) {
    rel <- rel[, c(
      "internal_pu1","internal_pu2","weight",
      intersect(names(rel), c("distance","source","relation_name"))
    ), drop = FALSE]
    rel$internal_pu1 <- as.integer(rel$internal_pu1)
    rel$internal_pu2 <- as.integer(rel$internal_pu2)
    rel$weight <- as.numeric(rel$weight)
    rel <- rel[order(rel$internal_pu1, rel$internal_pu2), , drop = FALSE]
    rel$internal_edge <- seq_len(nrow(rel))
    rel
  }

  rel_model <- NULL
  if (isTRUE(needs$y_pu) ||
      isTRUE(needs$y_action) ||
      isTRUE(needs$y_intervention)) {

    rel_name <- as.character(needs$relation_name %||% oargs$relation_name %||% "boundary")[1]

    rels <- x$data$spatial_relations
    if (is.null(rels) || is.null(rels[[rel_name]])) {
      .pa_abort("prepare_needs: missing spatial relation '", rel_name, "'.")
    }

    x$data$spatial_relations_model <- x$data$spatial_relations_model %||% list()
    if (is.null(x$data$spatial_relations_model[[rel_name]])) {
      x$data$spatial_relations_model[[rel_name]] <- .pa_prepare_relation_model(rels[[rel_name]])
    }
    rel_model <- x$data$spatial_relations_model[[rel_name]]
  }

  # ---- y_pu
  if (isTRUE(needs$y_pu)) {
    if (!exists("rcpp_prepare_objective_min_fragmentation", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_min_fragmentation() in the package.")
    }

    res <- rcpp_prepare_objective_min_fragmentation(
      x = op,
      relation_data = rel_model
    )
    x$data$model_registry$vars$y_pu <- res
  }

  # ---- y_action
  if (isTRUE(needs$y_action)) {
    if (!exists("rcpp_prepare_objective_min_fragmentation_actions", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_min_fragmentation_actions() in the package.")
    }

    if (is.null(x$data$dist_actions_model) || !inherits(x$data$dist_actions_model, "data.frame")) {
      .pa_abort("prepare_needs: y_action requires x$data$dist_actions_model (model-ready).")
    }
    if (nrow(x$data$dist_actions_model) == 0) {
      .pa_abort("prepare_needs: y_action requires action variables, but dist_actions_model has 0 rows.")
    }

    actions_to_use_raw <- needs$actions_to_use %||% oargs$actions_to_use %||% oargs$actions %||% NULL

    actions_to_use <- NULL
    if (!is.null(actions_to_use_raw)) {
      act_subset <- .pa_resolve_action_subset(x, subset = actions_to_use_raw)
      actions_to_use <- as.integer(act_subset$internal_id)
    }

    res <- rcpp_prepare_objective_min_fragmentation_actions(
      x = op,
      dist_actions_data = x$data$dist_actions_model,
      relation_data     = rel_model,
      actions_to_use    = actions_to_use,
      block_name        = "fragmentation_actions",
      tag               = ""
    )

    x$data$model_registry$vars$y_action <- res
  }

  # ---- u_intervention for min_intervention_impact
  if (isTRUE(needs$u_intervention)) {
    if (!exists("rcpp_prepare_objective_min_intervention_impact", mode = "function")) {
      .pa_abort("Missing rcpp_prepare_objective_min_intervention_impact() in the package.")
    }

    if (is.null(x$data$dist_actions_model) || !inherits(x$data$dist_actions_model, "data.frame")) {
      .pa_abort("prepare_needs: u_intervention requires x$data$dist_actions_model (model-ready).")
    }
    if (nrow(x$data$dist_actions_model) == 0) {
      .pa_abort("prepare_needs: u_intervention requires action variables, but dist_actions_model has 0 rows.")
    }

    if (is.null(x$data$dist_features) || !inherits(x$data$dist_features, "data.frame")) {
      .pa_abort("prepare_needs: u_intervention requires x$data$dist_features.")
    }
    if (nrow(x$data$dist_features) == 0) {
      .pa_abort("prepare_needs: u_intervention requires non-empty x$data$dist_features.")
    }

    actions_chr <- needs$u_intervention_actions %||% oargs$actions %||% NULL
    if (is.null(actions_chr) || length(actions_chr) == 0) {
      .pa_abort(
        "prepare_needs: u_intervention requires a non-empty shared action subset.\n",
        "For MO phase 1, all intervention_impact objectives must use the same actions."
      )
    }

    act_subset <- .pa_resolve_action_subset(x, subset = actions_chr)
    acts_internal <- as.integer(act_subset$internal_id)

    if (length(acts_internal) == 0) {
      .pa_abort("prepare_needs: u_intervention action subset resolved to zero actions.")
    }

    subset_key <- .pa_subset_to_string(act_subset$id)

    impact_col <- as.character(oargs$impact_col %||% "amount")[1]
    if (!(impact_col %in% names(x$data$dist_features))) {
      impact_col <- "amount"
      if (!("amount" %in% names(x$data$dist_features))) {
        .pa_abort(
          "prepare_needs: could not find a valid impact column in x$data$dist_features."
        )
      }
    }

    res <- rcpp_prepare_objective_min_intervention_impact(
      x = op,
      pu_data = x$data$pu,
      dist_actions_data = x$data$dist_actions_model,
      dist_features_data = x$data$dist_features,
      subset_key = subset_key,
      impact_col = impact_col,
      features_to_use = integer(0),
      actions_to_use = acts_internal,
      internal_feature_col = "internal_feature",
      block_name = "objective_min_intervention_impact",
      tag = if (isTRUE(mo_mode)) "from=prepare_needs_cpp;mode=MO" else ""
    )

    x$data$model_registry$vars$u_intervention_impact <- res
  }

  x
}


.pa_deepcopy_data <- function(d) {
  unserialize(serialize(d, NULL))
}

.pa_clone_data <- function(x, drop_model = TRUE) {
  stopifnot(inherits(x, "Problem"))

  # crear un NUEVO proto heredando de x
  y <- pproto(NULL, x)

  # ahora sí, sustituir data por una copia profunda
  y$data <- .pa_deepcopy_data(x$data)

  if (isTRUE(drop_model)) {
    y$data$model_ptr   <- NULL
    y$data$model_index <- NULL
    y$data$model_list  <- NULL
    y$data$has_model   <- FALSE

    if (is.null(y$data$meta) || !is.list(y$data$meta)) {
      y$data$meta <- list()
    }
    y$data$meta$model_dirty <- TRUE
  }

  y
}



.pa_log_add <- function(x, key, value) {
  stopifnot(inherits(x, "Problem"))

  key <- as.character(key)[1]
  if (is.na(key) || !nzchar(key)) {
    stop("`key` must be a non-empty string.", call. = FALSE)
  }

  value <- as.numeric(value)[1]
  if (!is.finite(value)) {
    stop("`value` must be a finite numeric scalar.", call. = FALSE)
  }

  if (is.null(x$data$build_log) || !is.list(x$data$build_log)) {
    x$data$build_log <- list()
  }

  old <- x$data$build_log[[key]] %||% 0
  x$data$build_log[[key]] <- old + value

  x
}



.pa_resolve_action_subset <- function(x, subset = NULL) {
  stopifnot(inherits(x, "Problem"))

  acts <- x$data$actions
  if (is.null(acts) || !inherits(acts, "data.frame") || nrow(acts) == 0) {
    stop("No actions available in x$data$actions.", call. = FALSE)
  }
  if (!all(c("id", "internal_id") %in% names(acts))) {
    stop("x$data$actions must contain columns 'id' and 'internal_id'.", call. = FALSE)
  }

  # NULL = all actions
  if (is.null(subset)) {
    out <- acts[, c("id", "internal_id"), drop = FALSE]
    if ("action_set" %in% names(acts)) out$action_set <- acts$action_set
    return(out)
  }

  subset <- as.character(subset)
  subset <- unique(subset[!is.na(subset) & nzchar(subset)])
  if (length(subset) == 0) {
    stop("subset must contain at least one non-empty value, or be NULL.", call. = FALSE)
  }

  hit_id <- acts$id %in% subset
  hit_set <- rep(FALSE, nrow(acts))
  if ("action_set" %in% names(acts)) {
    hit_set <- acts$action_set %in% subset
  }

  keep <- hit_id | hit_set

  if (!any(keep)) {
    stop(
      "subset did not match any action ids",
      if ("action_set" %in% names(acts)) " or action_set values" else "",
      ".",
      call. = FALSE
    )
  }

  out <- acts[keep, c("id", "internal_id"), drop = FALSE]
  if ("action_set" %in% names(acts)) out$action_set <- acts$action_set[keep]
  rownames(out) <- NULL
  out
}

#' @include internal.R
NULL

# -------------------------------------------------------------------------
# Internal helpers for normalized atomic objectives
# -------------------------------------------------------------------------

.pa_set_active_and_register_objective <- function(
    x,
    model_type,
    objective_id,
    objective_args = list(),
    sense = c("min", "max"),
    alias = NULL
) {
  stopifnot(inherits(x, "Problem"))
  sense <- match.arg(sense)

  x <- .pa_clone_data(x)

  if (is.null(x$data$model_args) || !is.list(x$data$model_args)) {
    x$data$model_args <- list()
  }

  x$data$model_args$model_type <- as.character(model_type)[1]
  x$data$model_args$objective_id <- as.character(objective_id)[1]
  x$data$model_args$objective_args <- objective_args

  x <- .pa_register_objective(
    x = x,
    alias = alias,
    objective_id = objective_id,
    model_type = model_type,
    objective_args = objective_args,
    sense = sense
  )

  x
}

.pa_resolve_feature_subset <- function(x, features = NULL) {
  stopifnot(inherits(x, "Problem"))

  feats <- x$data$features
  if (is.null(feats) || !inherits(feats, "data.frame") || nrow(feats) == 0) {
    stop("No features available in x$data$features.", call. = FALSE)
  }
  if (!all(c("id", "internal_id") %in% names(feats))) {
    stop("x$data$features must contain columns 'id' and 'internal_id'.", call. = FALSE)
  }

  if (is.null(features)) {
    out <- feats[, c("id", "internal_id"), drop = FALSE]
    if ("name" %in% names(feats)) out$name <- feats$name
    return(out)
  }

  vals <- unique(features)
  vals <- vals[!is.na(vals)]

  if (length(vals) == 0) {
    stop("features must contain at least one non-missing value, or be NULL.", call. = FALSE)
  }

  hit_id <- feats$id %in% vals
  hit_name <- rep(FALSE, nrow(feats))
  if ("name" %in% names(feats)) {
    hit_name <- as.character(feats$name) %in% as.character(vals)
  }

  keep <- hit_id | hit_name

  if (!any(keep)) {
    stop(
      "features did not match any feature ids",
      if ("name" %in% names(feats)) " or feature names" else "",
      ".",
      call. = FALSE
    )
  }

  out <- feats[keep, c("id", "internal_id"), drop = FALSE]
  if ("name" %in% names(feats)) out$name <- feats$name[keep]
  rownames(out) <- NULL
  out
}

.pa_subset_to_string <- function(subset) {
  if (is.null(subset)) return(NA_character_)

  subset <- as.character(subset)
  subset <- unique(subset[!is.na(subset) & nzchar(subset)])

  if (length(subset) == 0) return(NA_character_)

  paste(sort(subset), collapse = "|")
}

.pa_add_feature_labels <- function(df,
                                   features_df,
                                   feature_col = "feature",
                                   internal_feature_col = "internal_feature",
                                   out_col = "feature_name") {
  stopifnot(is.data.frame(df), is.data.frame(features_df))

  if (nrow(df) == 0) {
    df[[out_col]] <- character(0)
    return(df)
  }

  if (!("name" %in% names(features_df))) {
    df[[out_col]] <- rep(NA_character_, nrow(df))
    return(df)
  }

  out <- rep(NA_character_, nrow(df))

  # Prefer external feature id
  if (feature_col %in% names(df) && "id" %in% names(features_df)) {
    m <- match(df[[feature_col]], features_df$id)
    ok <- !is.na(m)
    out[ok] <- as.character(features_df$name[m[ok]])
  }

  # Fill remaining using internal id
  if (anyNA(out) &&
      internal_feature_col %in% names(df) &&
      "internal_id" %in% names(features_df)) {
    miss <- which(is.na(out))
    m2 <- match(df[[internal_feature_col]][miss], features_df$internal_id)
    ok2 <- !is.na(m2)
    out[miss[ok2]] <- as.character(features_df$name[m2[ok2]])
  }

  df[[out_col]] <- out
  df
}

.pa_add_action_labels <- function(df,
                                  actions_df,
                                  action_col = "action",
                                  internal_action_col = "internal_action",
                                  out_col = "action_name") {
  stopifnot(is.data.frame(df), is.data.frame(actions_df))

  if (nrow(df) == 0) {
    df[[out_col]] <- character(0)
    return(df)
  }

  if (!("name" %in% names(actions_df))) {
    df[[out_col]] <- rep(NA_character_, nrow(df))
    return(df)
  }

  out <- rep(NA_character_, nrow(df))

  # Prefer external action id
  if (action_col %in% names(df) && "id" %in% names(actions_df)) {
    m <- match(df[[action_col]], actions_df$id)
    ok <- !is.na(m)
    out[ok] <- as.character(actions_df$name[m[ok]])
  }

  # Fill remaining using internal id
  if (anyNA(out) &&
      internal_action_col %in% names(df) &&
      "internal_id" %in% names(actions_df)) {
    miss <- which(is.na(out))
    m2 <- match(df[[internal_action_col]][miss], actions_df$internal_id)
    ok2 <- !is.na(m2)
    out[miss[ok2]] <- as.character(actions_df$name[m2[ok2]])
  }

  df[[out_col]] <- out
  df
}



.pa_solve_single_problem <- function(x, ...) {
  assertthat::assert_that(inherits(x, "Problem"))

  # ---- gather stored solve args (defaults + stored)
  dots <- list(...)

  # consideramos “ya configurado” si solve_args existe y tiene al menos un campo no nulo
  has_stored_args <- !is.null(x$data$solve_args) &&
    length(x$data$solve_args) > 0 &&
    any(!vapply(x$data$solve_args, is.null, logical(1)))

  sa <- do.call(.pa_get_solve_args, c(list(x = x), dots))

  solver         <- sa$solver %||% "auto"
  gap_limit      <- sa$gap_limit
  time_limit     <- sa$time_limit
  solution_limit <- sa$solution_limit
  cores          <- sa$cores
  verbose        <- sa$verbose
  name_output_file <- sa$name_output_file
  output_file    <- sa$output_file
  solver_params_user <- sa$solver_params %||% list()

  # ---- autodetect solver if requested
  available_gurobi   <- available_to_solve("gurobi")
  available_cplex    <- available_to_solve("cplex")
  available_symphony <- available_to_solve("symphony")
  available_cbc      <- available_to_solve("cbc")

  if (identical(solver, "auto") || identical(solver, "") || is.null(solver)) {
    if (requireNamespace("Rcplex", quietly = TRUE) && available_cplex) {
      solver <- "cplex"
    } else if (requireNamespace("gurobi", quietly = TRUE) && available_gurobi) {
      solver <- "gurobi"
    } else if (requireNamespace("rcbc", quietly = TRUE) && available_cbc) {
      solver <- "cbc"
    } else if (requireNamespace("Rsymphony", quietly = TRUE) && available_symphony) {
      solver <- "symphony"
    } else {
      stop("No optimization problem solvers available on this system.", call. = FALSE)
    }
  } else {
    if (!solver %in% c("gurobi", "cbc", "symphony", "cplex")) {
      stop("Solver not supported: ", solver, call. = FALSE)
    }
    if (identical(solver, "gurobi") && (!requireNamespace("gurobi", quietly = TRUE) || !available_gurobi)) {
      stop("Gurobi solver not available (package/license not found).", call. = FALSE)
    }
    if (identical(solver, "cbc") && (!requireNamespace("rcbc", quietly = TRUE) || !available_cbc)) {
      stop("CBC solver not available (rcbc not installed or CBC not available).", call. = FALSE)
    }
    if (identical(solver, "symphony") && (!requireNamespace("Rsymphony", quietly = TRUE) || !available_symphony)) {
      stop("SYMPHONY solver not available (Rsymphony not installed).", call. = FALSE)
    }
    if (identical(solver, "cplex") && (!requireNamespace("Rcplex", quietly = TRUE) || !available_cplex)) {
      stop("CPLEX solver not available (Rcplex not installed/licensed).", call. = FALSE)
    }
  }


  # ---- ensure model is built
  if (is.null(x$data$model_ptr) || isTRUE(x$data$meta$model_dirty)) {
    x <- .pa_build_model(x)
    x$data$meta$model_dirty <- FALSE
  }

  # ---- model list from pointer
  model <- .pa_model_from_ptr(
    x$data$model_ptr,
    args = x$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  # ---- APPLY SUPERSET RUNTIME UPDATES (solver-agnostic)
  model <- .pa_apply_runtime_updates_to_model(model, x)

  # ---- pack args into Solution metadata
  solve_args <- list(
    solver = solver,
    gap = gap_limit,
    timelimit = time_limit,
    cores = cores,
    verbose = verbose,
    solution_limit = solution_limit,
    name_output_file = name_output_file,
    output_file = output_file,
    solver_params = solver_params_user
  )

  # ------------------------------------------------------------
  # 1) Call solver -> return unified payload:
  #    objval, solvec, gap, status_code, runtime
  # ------------------------------------------------------------
  objval <- NA_real_
  solvec <- numeric(0)
  gap    <- NA_real_
  status_code <- 999L
  runtime <- NA_real_

  if (solver == "gurobi") {

    model$sense <- replace(model$sense, model$sense == "==", "=")
    model$lb <- model$bounds$lower$val
    model$ub <- model$bounds$upper$val

    params <- list(
      Threads = cores,
      LogToConsole = as.integer(verbose),
      NodefileStart = 0.5,
      MIPGap = gap_limit,
      TimeLimit = time_limit
    )
    if (isTRUE(output_file)) params$LogFile <- paste0(name_output_file, "_log.txt")
    if (isTRUE(solution_limit)) params$SolutionLimit <- 1

    if (!is.null(model$args$curve) && !is.null(model$args$segments) && model$args$curve != 1) {
      params$FuncPieces <- 1
      params$FuncPieceLength <- round(1 / as.numeric(model$args$segments), digits = 1)
    }

    params <- utils::modifyList(params, solver_params_user)

    sol <- gurobi::gurobi(model, params)

    # status_code <- dplyr::case_when(
    #   sol$status == "OPTIMAL" ~ 0L,
    #   sol$status %in% c("INF_OR_UNBD", "INFEASIBLE", "UNBOUNDED") ~ 1L,
    #   (sol$status == "TIME_LIMIT" && !is.null(sol$objval)) ~ 2L,
    #   (sol$status == "TIME_LIMIT" && is.null(sol$objval)) ~ 3L,
    #   sol$status == "SOLUTION_LIMIT" ~ 4L,
    #   TRUE ~ 999L
    # )

    status_chr <- as.character(sol$status %||% NA_character_)[1]

    status_code <- if (identical(status_chr, "OPTIMAL")) {
      0L
    } else if (status_chr %in% c("INF_OR_UNBD", "INFEASIBLE", "UNBOUNDED")) {
      1L
    } else if (identical(status_chr, "TIME_LIMIT") && !is.null(sol$objval)) {
      2L
    } else if (identical(status_chr, "TIME_LIMIT") && is.null(sol$objval)) {
      3L
    } else if (identical(status_chr, "SOLUTION_LIMIT")) {
      4L
    } else {
      999L
    }

    objval  <- sol$objval %||% NA_real_
    solvec  <- sol$x %||% numeric(0)
    gap     <- sol$mipgap %||% NA_real_
    runtime <- sol$runtime %||% NA_real_

  } else if (solver == "cbc") {

    # build row bounds from sense + rhs
    row_lb <- rep(-Inf, length(model$rhs))
    row_ub <- rep( Inf, length(model$rhs))

    ii_le <- which(model$sense == "<=")
    ii_ge <- which(model$sense == ">=")
    ii_eq <- which(model$sense == "==" | model$sense == "=")

    row_ub[ii_le] <- model$rhs[ii_le]
    row_lb[ii_ge] <- model$rhs[ii_ge]
    row_lb[ii_eq] <- model$rhs[ii_eq]
    row_ub[ii_eq] <- model$rhs[ii_eq]

    cbc_args <- list(
      threads = as.character(cores),
      log = as.integer(verbose),
      verbose = 15,
      ratio = as.character(gap_limit),
      sec = as.character(time_limit),
      timem = "elapsed",
      heuristicsOnOff = "on"
    )
    cbc_args <- utils::modifyList(cbc_args, solver_params_user)

    rt <- system.time({
      sol_cbc <- rcbc::cbc_solve(
        obj = model$obj,
        mat = model$A,
        is_integer = ifelse(model$vtype == "B", TRUE, FALSE),
        row_ub = row_ub,
        row_lb = row_lb,
        col_lb = model$bounds$lower$val,
        col_ub = model$bounds$upper$val,
        max = ifelse(model$modelsense == "min", FALSE, TRUE),
        cbc_args = cbc_args
      )
    })

    runtime <- rt[[3]]

    status_cbc <- rcbc::solution_status(sol_cbc)
    status_code <- dplyr::case_when(
      status_cbc == "optimal" ~ 0L,
      status_cbc == "infeasible" ~ 1L,
      (status_cbc == "timelimit" && !is.null(sol_cbc$objective_value)) ~ 2L,
      (status_cbc == "timelimit" && is.null(sol_cbc$objective_value)) ~ 3L,
      TRUE ~ 999L
    )

    objval <- sol_cbc$objective_value %||% NA_real_
    solvec <- sol_cbc$column_solution %||% numeric(0)
    gap    <- if (isTRUE(status_code == 0L)) gap_limit else NA_real_


  } else if (solver == "cplex") {

    model$lb <- model$bounds$lower$val
    model$ub <- model$bounds$upper$val

    sense <- model$sense
    sense[sense == ">="] <- "G"
    sense[sense == "=="] <- "E"
    sense[sense == "<="] <- "L"

    params <- list(
      trace = as.integer(verbose),
      epgap = gap_limit,
      tilim = time_limit
    )
    params <- utils::modifyList(params, solver_params_user)

    if (any(solution_limit, output_file)) {
      warning(
        "Options not available with cplex solver via this interface: solution_limit, output_file",
        call. = FALSE, immediate. = TRUE
      )
    }

    rt <- system.time({
      sol_cplex <- Rcplex::Rcplex(
        cvec = model$obj,
        Amat = model$A,
        bvec = model$rhs,
        lb = model$lb,
        ub = model$ub,
        objsense = model$modelsense,
        sense = sense,
        vtype = model$vtype,
        control = params
      )
    })

    runtime <- rt[[3]]

    status_code <- dplyr::case_when(
      sol_cplex$status %in% c(101L, 1L) ~ 0L,
      sol_cplex$status %in% c(2L, 3L, 103L, 118L) ~ 1L,
      sol_cplex$status == 107L ~ 2L,
      sol_cplex$status == 108L ~ 3L,
      sol_cplex$status == 232L ~ 4L,
      TRUE ~ 999L
    )

    objval <- sol_cplex$obj %||% NA_real_
    solvec <- sol_cplex$xopt %||% numeric(0)
    gap    <- 0

  } else if (solver == "symphony") {

    if (isTRUE(output_file)) {
      warning("It is not possible to export a solver log using symphony solver.", call. = FALSE, immediate. = TRUE)
    }

    max_flag <- ifelse(model$modelsense == "min", FALSE, TRUE)
    verbosity <- as.integer(verbose) - 2

    rt <- system.time({
      sol_sym <- Rsymphony::Rsymphony_solve_LP(
        obj = model$obj,
        mat = model$A,
        dir = model$sense,
        rhs = model$rhs,
        bounds = model$bounds,
        types = model$vtype,
        max = max_flag,
        gap_limit = gap_limit,
        time_limit = time_limit,
        verbosity = verbosity,
        first_feasible = solution_limit
      )
    })

    runtime <- rt[[3]]

    status_code <- dplyr::case_when(
      sol_sym$status %in% c(0L, 231L) ~ 0L,
      sol_sym$status %in% c(226L, 237L) ~ 1L,
      sol_sym$status %in% c(235L, 228L) ~ 2L,
      sol_sym$status == 232L ~ 4L,
      TRUE ~ 999L
    )

    objval <- sol_sym$objval %||% NA_real_
    solvec <- sol_sym$solution %||% numeric(0)
    gap    <- if (isTRUE(status_code == 0L)) gap_limit else NA_real_

  } else {
    stop("Internal error: unknown solver '", solver, "'.", call. = FALSE)
  }

  # ------------------------------------------------------------
  # Hard fail if solver returned no usable solution
  # ------------------------------------------------------------
  if (length(solvec) == 0L || all(is.na(solvec))) {
    msg <- paste0(
      "Solver returned an empty solution vector.\n",
      "solver: ", solver, "\n",
      "status_code: ", status_code, "\n",
      "objval: ", as.character(objval), "\n",
      "Possible causes: infeasible model, time limit before first feasible, or solver error.\n",
      "Tip: try increasing time_limit, relaxing constraints, or inspect solver logs (if available)."
    )
    stop(msg, call. = FALSE)
  }

  # Ensure solution length is consistent with model offsets
  n_pu_chk <- as.integer(model$n_pu %||% 0L)
  n_x_chk  <- as.integer(model$n_x  %||% 0L)
  n_z_chk  <- as.integer(model$n_z  %||% 0L)
  w0_chk <- as.integer(model$w_offset %||% 0L)
  x0_chk <- as.integer(model$x_offset %||% 0L)
  z0_chk <- as.integer(model$z_offset %||% 0L)

  needed_len <- max(
    w0_chk + n_pu_chk,
    x0_chk + n_x_chk,
    z0_chk + n_z_chk,
    0L
  )

  if (length(solvec) < needed_len) {
    stop(
      "Solver returned a solution vector of length ", length(solvec),
      " but the model requires at least ", needed_len, " variables ",
      "(based on offsets + variable counts). ",
      "This indicates a mismatch between the built model and the decoded metadata.",
      call. = FALSE
    )
  }

  # ------------------------------------------------------------
  # 2) Minimal decoding using offsets (0-based in C++; +1 in R)
  # ------------------------------------------------------------
  # n_pu <- as.integer(model$n_pu %||% 0L)
  # n_x  <- as.integer(model$n_x  %||% 0L)
  # n_z  <- as.integer(model$n_z  %||% 0L)
  #
  # w0 <- as.integer(model$w_offset %||% 0L)
  # x0 <- as.integer(model$x_offset %||% 0L)
  # z0 <- as.integer(model$z_offset %||% 0L)
  #
  # sol_monitoring <- if (n_pu > 0L && length(solvec) >= (w0 + n_pu)) {
  #   base::round(solvec[(w0 + 1L):(w0 + n_pu)])
  # } else numeric(0)
  #
  # sol_actions <- if (n_x > 0L && length(solvec) >= (x0 + n_x)) {
  #   base::round(solvec[(x0 + 1L):(x0 + n_x)])
  # } else numeric(0)
  #
  # sol_conservation <- if (n_z > 0L && length(solvec) >= (z0 + n_z)) {
  #   solvec[(z0 + 1L):(z0 + n_z)]
  # } else numeric(0)

  # ------------------------------------------------------------
  # 3) Build human-readable tables (you implement this helper)
  # ------------------------------------------------------------

  has_feasible_solution <- status_code %in% c(0L, 2L, 4L)

  if (!has_feasible_solution) {
    stop(
      "Solver did not return a feasible solution.\n",
      "solver: ", solver, "\n",
      "status_code: ", status_code, "\n",
      "status: ", status_cbc %||% "unknown",
      call. = FALSE
    )
  }

  summary <- .pa_extract_solution_tables(x, solvec)

  x$data$runtime_updates <- NULL

  objs <- x$data$objectives %||% list()

  alias_values <- NULL
  if (is.list(objs) && length(objs) == 1L) {
    alias_single <- names(objs)[1]
    alias_values <- stats::setNames(objval, alias_single)
  }


  s <- pproto(
    NULL, Solution,
    problem = x,
    solution = list(
      objective = objval,
      vector = solvec,
      alias_values = alias_values
    ),
    summary = summary,
    diagnostics = list(
      status_code = as.integer(status_code),
      gap = gap,
      runtime = runtime,
      solver = solve_args$solver,
      cores = solve_args$cores,
      timelimit = solve_args$timelimit,
      verbose = solve_args$verbose,
      solution_limit = solve_args$solution_limit,
      name_output_file = solve_args$name_output_file,
      output_file = solve_args$output_file,
      solver_args = solve_args
    ),
    method = list(
      type = "single"
    ),
    meta = list()
  )

  s
}


.validate_split_effects <- function(tbl, context = "effects") {
  if (!("benefit" %in% names(tbl)) || !("loss" %in% names(tbl))) {
    stop("Internal error: .validate_split_effects() requires 'benefit' and 'loss' columns.", call. = FALSE)
  }

  tbl$benefit <- as.numeric(tbl$benefit)
  tbl$loss    <- as.numeric(tbl$loss)

  #if (na_to_zero) {
  tbl$benefit[is.na(tbl$benefit)] <- 0
  tbl$loss[is.na(tbl$loss)] <- 0
  #}

  if (any(tbl$benefit < 0, na.rm = TRUE) || any(tbl$loss < 0, na.rm = TRUE)) {
    stop(
      context,
      ": 'benefit' and 'loss' must be non-negative.",
      call. = FALSE
    )
  }

  bad <- which(tbl$benefit > 0 & tbl$loss > 0)
  if (length(bad) > 0) {
    ex <- tbl[bad[1], intersect(c("pu", "action", "feature", "benefit", "loss"), names(tbl)), drop = FALSE]
    msg <- paste0(
      context,
      ": a single (pu, action, feature) effect cannot have both positive 'benefit' and positive 'loss'. ",
      "This is checked after aggregation by (pu, action, feature), so the same triple cannot contain both gains and damages."
    )
    if (nrow(ex) == 1) {
      msg <- paste0(
        msg,
        " Example offending row -> pu=", ex$pu,
        ", action='", ex$action,
        "', feature=", ex$feature,
        ", benefit=", ex$benefit,
        ", loss=", ex$loss, "."
      )
    }
    stop(msg, call. = FALSE)
  }

  tbl
}



.pa_model_is_current <- function(x) {
  stopifnot(inherits(x, "Problem"))

  has_ptr <- !is.null(x$data$model_ptr)
  has_list <- !is.null(x$data$model_list) && is.list(x$data$model_list)
  dirty <- isTRUE(x$data$meta$model_dirty)

  isTRUE(has_ptr && has_list && !dirty)
}

.pa_compile_problem <- function(x, ...) {
  .pa_build_model(x)
}


.pa_store_area_constraints <- function(x, area_df) {

  stopifnot(inherits(x, "Problem"))
  stopifnot(is.data.frame(area_df))
  stopifnot(nrow(area_df) >= 1)

  required_cols <- c(
    "type", "sense", "value", "tolerance",
    "unit", "area_col", "actions", "name"
  )

  miss <- setdiff(required_cols, names(area_df))
  if (length(miss) > 0) {
    stop(
      "Missing required columns in `area_df`: ",
      paste(miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  if (any(area_df$type != "area")) {
    stop("All rows in `area_df` must have `type = 'area'`.", call. = FALSE)
  }

  if (any(is.na(area_df$sense)) ||
      any(!area_df$sense %in% c("min", "max", "equal"))) {
    stop(
      "`area_df$sense` must contain only 'min', 'max', or 'equal'.",
      call. = FALSE
    )
  }

  if (any(!is.finite(area_df$value)) || any(area_df$value < 0)) {
    stop(
      "`area_df$value` must contain only finite values >= 0.",
      call. = FALSE
    )
  }

  if (any(!is.finite(area_df$tolerance)) || any(area_df$tolerance < 0)) {
    stop(
      "`area_df$tolerance` must contain only finite values >= 0.",
      call. = FALSE
    )
  }

  if (any(is.na(area_df$unit)) ||
      any(!area_df$unit %in% c("m2", "ha", "km2"))) {
    stop(
      "`area_df$unit` must contain only 'm2', 'ha', or 'km2'.",
      call. = FALSE
    )
  }

  if (any(is.na(area_df$name)) || any(!nzchar(area_df$name))) {
    stop(
      "`area_df$name` must contain only non-empty character strings.",
      call. = FALSE
    )
  }

  area_df$actions <- as.character(area_df$actions)
  area_df$area_col <- as.character(area_df$area_col)
  area_df$name <- as.character(area_df$name)
  area_df$type <- as.character(area_df$type)
  area_df$sense <- as.character(area_df$sense)
  area_df$unit <- as.character(area_df$unit)

  new_actions_key <- ifelse(is.na(area_df$actions), "__ALL__", area_df$actions)
  new_key <- paste(new_actions_key, area_df$sense, sep = "||")

  if (anyDuplicated(new_key)) {
    dup <- unique(new_key[duplicated(new_key)])[1]
    parts <- strsplit(dup, "\\|\\|")[[1]]
    act_key <- parts[1]
    sense_key <- parts[2]

    subset_label <- if (identical(act_key, "__ALL__")) {
      "all actions"
    } else {
      paste0("actions = {", act_key, "}")
    }

    stop(
      "Duplicated area constraints detected within `area_df` for ",
      subset_label,
      " and sense = '", sense_key, "'.",
      call. = FALSE
    )
  }

  cons <- x$data$constraints %||% list()
  old <- cons$area

  if (is.null(old)) {
    cons$area <- area_df[, required_cols, drop = FALSE]
    rownames(cons$area) <- NULL
    x$data$constraints <- cons
    return(x)
  }

  if (!is.data.frame(old)) {
    stop(
      "Stored area constraints must be a data.frame. Please rebuild the problem object.",
      call. = FALSE
    )
  }

  old_miss <- setdiff(required_cols, names(old))
  if (length(old_miss) > 0) {
    stop(
      "Stored area constraints are malformed. Missing columns: ",
      paste(old_miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  old$actions <- as.character(old$actions)
  old$sense <- as.character(old$sense)

  old_actions_key <- ifelse(is.na(old$actions), "__ALL__", old$actions)
  old_key <- paste(old_actions_key, old$sense, sep = "||")

  overlap <- intersect(old_key, new_key)
  if (length(overlap) > 0) {
    dup <- overlap[1]
    parts <- strsplit(dup, "\\|\\|")[[1]]
    act_key <- parts[1]
    sense_key <- parts[2]

    subset_label <- if (identical(act_key, "__ALL__")) {
      "all actions"
    } else {
      paste0("actions = {", act_key, "}")
    }

    stop(
      "An area constraint already exists for the same subset of actions (",
      subset_label,
      ") and sense = '", sense_key, "'.",
      call. = FALSE
    )
  }

  out <- rbind(
    old[, required_cols, drop = FALSE],
    area_df[, required_cols, drop = FALSE]
  )
  rownames(out) <- NULL

  cons$area <- out
  x$data$constraints <- cons

  x
}


.pa_store_budget_constraints <- function(x, budget_df) {

  stopifnot(inherits(x, "Problem"))
  stopifnot(is.data.frame(budget_df))
  stopifnot(nrow(budget_df) >= 1)

  required_cols <- c(
    "type", "sense", "value", "tolerance",
    "actions", "include_pu_cost", "include_action_cost", "name"
  )

  miss <- setdiff(required_cols, names(budget_df))
  if (length(miss) > 0) {
    stop(
      "Missing required columns in `budget_df`: ",
      paste(miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  if (any(budget_df$type != "budget")) {
    stop("All rows in `budget_df` must have `type = 'budget'.", call. = FALSE)
  }

  if (any(is.na(budget_df$sense)) ||
      any(!budget_df$sense %in% c("min", "max", "equal"))) {
    stop(
      "`budget_df$sense` must contain only 'min', 'max', or 'equal'.",
      call. = FALSE
    )
  }

  if (any(!is.finite(budget_df$value)) || any(budget_df$value < 0)) {
    stop(
      "`budget_df$value` must contain only finite values >= 0.",
      call. = FALSE
    )
  }

  if (any(!is.finite(budget_df$tolerance)) || any(budget_df$tolerance < 0)) {
    stop(
      "`budget_df$tolerance` must contain only finite values >= 0.",
      call. = FALSE
    )
  }

  if (!is.logical(budget_df$include_pu_cost) ||
      any(is.na(budget_df$include_pu_cost))) {
    stop(
      "`budget_df$include_pu_cost` must contain only TRUE/FALSE values.",
      call. = FALSE
    )
  }

  if (!is.logical(budget_df$include_action_cost) ||
      any(is.na(budget_df$include_action_cost))) {
    stop(
      "`budget_df$include_action_cost` must contain only TRUE/FALSE values.",
      call. = FALSE
    )
  }

  if (any(!budget_df$include_pu_cost & !budget_df$include_action_cost)) {
    stop(
      "Each row in `budget_df` must have at least one of ",
      "`include_pu_cost` or `include_action_cost` set to TRUE.",
      call. = FALSE
    )
  }

  actions_chr <- as.character(budget_df$actions)
  bad_action_pu_mix <- !is.na(actions_chr) & nzchar(actions_chr) & budget_df$include_pu_cost
  if (any(bad_action_pu_mix)) {
    stop(
      "`include_pu_cost = TRUE` is only supported when `actions` is NA/NULL, ",
      "because planning-unit costs are not action-specific.",
      call. = FALSE
    )
  }

  if (any(is.na(budget_df$name)) || any(!nzchar(budget_df$name))) {
    stop(
      "`budget_df$name` must contain only non-empty character strings.",
      call. = FALSE
    )
  }

  budget_df$type <- as.character(budget_df$type)
  budget_df$sense <- as.character(budget_df$sense)
  budget_df$actions <- actions_chr
  budget_df$name <- as.character(budget_df$name)

  new_actions_key <- ifelse(is.na(budget_df$actions), "__ALL__", budget_df$actions)
  new_key <- paste(new_actions_key, budget_df$sense, sep = "||")

  if (anyDuplicated(new_key)) {
    dup <- unique(new_key[duplicated(new_key)])[1]
    parts <- strsplit(dup, "\\|\\|")[[1]]
    act_key <- parts[1]
    sense_key <- parts[2]

    subset_label <- if (identical(act_key, "__ALL__")) {
      "all actions"
    } else {
      paste0("actions = {", act_key, "}")
    }

    stop(
      "Duplicated budget constraints detected within `budget_df` for ",
      subset_label,
      " and sense = '", sense_key, "'.",
      call. = FALSE
    )
  }

  cons <- x$data$constraints %||% list()
  old <- cons$budget

  if (is.null(old)) {
    cons$budget <- budget_df[, required_cols, drop = FALSE]
    rownames(cons$budget) <- NULL
    x$data$constraints <- cons
    return(x)
  }

  if (!is.data.frame(old)) {
    stop(
      "Stored budget constraints must be a data.frame. Please rebuild the problem object.",
      call. = FALSE
    )
  }

  old_miss <- setdiff(required_cols, names(old))
  if (length(old_miss) > 0) {
    stop(
      "Stored budget constraints are malformed. Missing columns: ",
      paste(old_miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  old$type <- as.character(old$type)
  old$sense <- as.character(old$sense)
  old$actions <- as.character(old$actions)
  old$name <- as.character(old$name)

  if (any(old$type != "budget")) {
    stop(
      "Stored budget constraints are malformed: `type` must be 'budget'.",
      call. = FALSE
    )
  }

  if (any(is.na(old$sense)) || any(!old$sense %in% c("min", "max", "equal"))) {
    stop(
      "Stored budget constraints are malformed: invalid values in `sense`.",
      call. = FALSE
    )
  }

  if (!is.logical(old$include_pu_cost) || any(is.na(old$include_pu_cost))) {
    stop(
      "Stored budget constraints are malformed: `include_pu_cost` must contain only TRUE/FALSE values.",
      call. = FALSE
    )
  }

  if (!is.logical(old$include_action_cost) || any(is.na(old$include_action_cost))) {
    stop(
      "Stored budget constraints are malformed: `include_action_cost` must contain only TRUE/FALSE values.",
      call. = FALSE
    )
  }

  old_bad_action_pu_mix <- !is.na(old$actions) & nzchar(old$actions) & old$include_pu_cost
  if (any(old_bad_action_pu_mix)) {
    stop(
      "Stored budget constraints are malformed: action-specific rows cannot have `include_pu_cost = TRUE`.",
      call. = FALSE
    )
  }

  old_actions_key <- ifelse(is.na(old$actions), "__ALL__", old$actions)
  old_key <- paste(old_actions_key, old$sense, sep = "||")

  overlap <- intersect(old_key, new_key)
  if (length(overlap) > 0) {
    dup <- overlap[1]
    parts <- strsplit(dup, "\\|\\|")[[1]]
    act_key <- parts[1]
    sense_key <- parts[2]

    subset_label <- if (identical(act_key, "__ALL__")) {
      "all actions"
    } else {
      paste0("actions = {", act_key, "}")
    }

    stop(
      "A budget constraint already exists for the same subset of actions (",
      subset_label,
      ") and sense = '", sense_key, "'.",
      call. = FALSE
    )
  }

  out <- rbind(
    old[, required_cols, drop = FALSE],
    budget_df[, required_cols, drop = FALSE]
  )
  rownames(out) <- NULL

  cons$budget <- out
  x$data$constraints <- cons

  x
}


.pa_apply_budget_constraints_if_present <- function(x) {
  stopifnot(inherits(x, "Problem"))

  cons <- x$data$constraints %||% list()
  if (length(cons) == 0L) return(x)

  specs <- cons$budget %||% NULL
  if (is.null(specs)) return(x)

  if (is.null(x$data$model_ptr)) {
    stop("Model pointer is missing while applying budget constraints.", call. = FALSE)
  }

  x <- .pa_refresh_model_snapshot(x)

  ml <- x$data$model_list
  if (is.null(ml)) {
    stop("Model snapshot is missing while applying budget constraints.", call. = FALSE)
  }

  if (!is.data.frame(specs)) {
    stop("Stored budget constraints must be a data.frame.", call. = FALSE)
  }

  required_cols <- c(
    "type", "sense", "value", "tolerance",
    "actions", "include_pu_cost", "include_action_cost", "name"
  )

  miss <- setdiff(required_cols, names(specs))
  if (length(miss) > 0) {
    stop(
      "Stored budget constraints are malformed. Missing columns: ",
      paste(miss, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  n_pu <- as.integer(ml$n_pu %||% 0L)
  if (n_pu <= 0L) {
    stop("Model has n_pu=0; cannot apply budget constraints.", call. = FALSE)
  }

  da <- x$data$dist_actions_model %||% NULL
  if (is.null(da) || !is.data.frame(da) || nrow(da) == 0L) {
    stop(
      "Budget constraints require `x$data$dist_actions_model`.",
      call. = FALSE
    )
  }

  if (!("internal_row" %in% names(da))) {
    stop(
      "`dist_actions_model` must contain column `internal_row`.",
      call. = FALSE
    )
  }

  if (!("cost" %in% names(da))) {
    stop(
      "`dist_actions_model` must contain column `cost`.",
      call. = FALSE
    )
  }

  x0 <- as.integer(ml$x_offset %||% 0L)
  w0 <- as.integer(ml$w_offset %||% 0L)

  for (k in seq_len(nrow(specs))) {
    spec <- specs[k, , drop = FALSE]

    sense <- as.character(spec$sense)[1]
    rhs <- as.numeric(spec$value)[1]
    tol <- as.numeric(spec$tolerance)[1]
    nm <- as.character(spec$name)[1]
    actions_txt <- as.character(spec$actions)[1]
    include_pu_cost <- isTRUE(spec$include_pu_cost[[1]])
    include_action_cost <- isTRUE(spec$include_action_cost[[1]])

    if (!sense %in% c("min", "max", "equal")) {
      stop("Unknown budget constraint sense: ", sense, call. = FALSE)
    }
    if (!is.finite(rhs) || is.na(rhs) || rhs < 0) {
      stop("Invalid budget constraint value in stored budget constraints.", call. = FALSE)
    }
    if (!is.finite(tol) || is.na(tol) || tol < 0) {
      stop("Invalid budget constraint tolerance in stored budget constraints.", call. = FALSE)
    }
    if (is.na(nm) || !nzchar(nm)) {
      stop("Stored budget constraint has invalid `name`.", call. = FALSE)
    }
    if (!include_pu_cost && !include_action_cost) {
      stop(
        "Stored budget constraint `", nm,
        "` has both `include_pu_cost` and `include_action_cost` set to FALSE.",
        call. = FALSE
      )
    }

    var_index <- integer(0)
    coeff <- numeric(0)

    # ------------------------------------------------------------
    # planning-unit cost component (only for global budget)
    # ------------------------------------------------------------
    if (include_pu_cost) {
      if (!is.na(actions_txt) && nzchar(actions_txt)) {
        stop(
          "Stored budget constraint `", nm,
          "` has `include_pu_cost = TRUE` with an action-specific subset, which is not supported.",
          call. = FALSE
        )
      }

      pu_cost <- .pa_get_cost_vec(x$data$pu)

      if (length(pu_cost) != n_pu) {
        stop(
          "Planning-unit cost vector length (", length(pu_cost), ") != n_pu (", n_pu,
          ") while applying budget constraint `", nm, "`.",
          call. = FALSE
        )
      }

      j0_w <- w0 + (0:(n_pu - 1L))
      var_index <- c(var_index, j0_w)
      coeff <- c(coeff, as.numeric(pu_cost))
    }

    # ------------------------------------------------------------
    # action cost component
    # ------------------------------------------------------------
    if (include_action_cost) {
      keep <- rep(TRUE, nrow(da))

      if (!is.na(actions_txt) && nzchar(actions_txt)) {
        actions_chr <- strsplit(actions_txt, "\\|", fixed = FALSE)[[1]]

        act_subset <- .pa_resolve_action_subset(x, subset = actions_chr)
        if (!is.data.frame(act_subset) || nrow(act_subset) == 0L) {
          stop(
            "Stored budget constraint `", nm,
            "` refers to an invalid or empty action subset.",
            call. = FALSE
          )
        }

        keep <- da$action %in% act_subset$id
      }

      if (!any(keep)) {
        stop(
          "Stored budget constraint `", nm,
          "` does not match any rows in `dist_actions_model`.",
          call. = FALSE
        )
      }

      var_index_x <- x0 + (as.integer(da$internal_row[keep]) - 1L)
      coeff_x <- as.numeric(da$cost[keep])

      if (length(var_index_x) != length(coeff_x) || length(coeff_x) == 0L) {
        stop(
          "Failed to assemble action-cost coefficients for budget constraint `", nm, "`.",
          call. = FALSE
        )
      }

      var_index <- c(var_index, var_index_x)
      coeff <- c(coeff, coeff_x)
    }

    if (length(var_index) != length(coeff) || length(coeff) == 0L) {
      stop(
        "Failed to assemble coefficients for budget constraint `", nm, "`.",
        call. = FALSE
      )
    }

    if (identical(sense, "min")) {

      x <- .pa_add_linear_constraint(
        x,
        var_index_0based = var_index,
        coeff = coeff,
        sense = ">=",
        rhs = rhs,
        name = nm
      )

    } else if (identical(sense, "max")) {

      x <- .pa_add_linear_constraint(
        x,
        var_index_0based = var_index,
        coeff = coeff,
        sense = "<=",
        rhs = rhs,
        name = nm
      )

    } else if (identical(sense, "equal")) {

      if (tol == 0) {

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = "==",
          rhs = rhs,
          name = nm
        )

      } else {

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = ">=",
          rhs = rhs - tol,
          name = paste0(nm, "_lower")
        )

        x <- .pa_add_linear_constraint(
          x,
          var_index_0based = var_index,
          coeff = coeff,
          sense = "<=",
          rhs = rhs + tol,
          name = paste0(nm, "_upper")
        )
      }
    }
  }

  x
}



.pa_active_objective_alias <- function(x) {
  objs <- x$data$objectives

  if (is.null(objs) || !is.list(objs) || length(objs) == 0) {
    return(NULL)
  }

  args <- x$data$model_args %||% list()
  mtype <- args$model_type %||% NULL
  oid   <- args$objective_id %||% NULL

  for (nm in names(objs)) {
    ob <- objs[[nm]]

    same_mtype <- !is.null(mtype) &&
      identical(as.character(ob$model_type)[1], as.character(mtype)[1])

    same_oid <- !is.null(oid) &&
      identical(as.character(ob$objective_id)[1], as.character(oid)[1])

    if (isTRUE(same_mtype) || isTRUE(same_oid)) {
      return(ob$alias %||% nm)
    }
  }

  NULL
}


.pa_feature_names_from_internal_ids <- function(x, ids) {
  ids <- as.integer(ids)

  f <- x$data$features

  if (is.null(f) || !inherits(f, "data.frame") || nrow(f) == 0) {
    return(as.character(ids))
  }

  label_col <- if ("name" %in% names(f)) {
    "name"
  } else if ("id" %in% names(f)) {
    "id"
  } else {
    NULL
  }

  if (is.null(label_col) || !("internal_id" %in% names(f))) {
    return(as.character(ids))
  }

  out <- as.character(f[[label_col]][match(ids, f$internal_id)])
  out[is.na(out)] <- as.character(ids[is.na(out)])

  out
}


.pa_action_names_from_internal_ids <- function(x, ids) {
  ids <- as.integer(ids)

  a <- x$data$actions

  if (is.null(a) || !inherits(a, "data.frame") || nrow(a) == 0) {
    return(as.character(ids))
  }

  label_col <- if ("name" %in% names(a)) {
    "name"
  } else if ("id" %in% names(a)) {
    "id"
  } else {
    NULL
  }

  if (is.null(label_col) || !("internal_id" %in% names(a))) {
    return(as.character(ids))
  }

  out <- as.character(a[[label_col]][match(ids, a$internal_id)])
  out[is.na(out)] <- as.character(ids[is.na(out)])

  out
}

#' Skip if no commercial solvers
#'
#' Skip a test if no commercial solvers are installed.
#'
#' @param packages `character` vector containing the package dependencies
#'   for commercial solvers. Defaults to `"gurobi"` and `"Rcplex"`.
#'
#' @return `logical` indicating success.
skip_if_no_commercial_solvers_installed <- function(packages = c("gurobi", "Rcplex")) {
  # assert arguments are valid
  assertthat::assert_that(
    is.character(packages), assertthat::noNA(packages))
  # check if any dependencies present
  result <- vapply(packages, requireNamespace, logical(1), quietly = TRUE)
  # skip if none installed
  if (any(result)) {
    return(invisible(TRUE))
  }
  testthat::skip("No commercial solvers installed")
}


skip_if_no_cbc <- function() {
  testthat::skip_if_not(
    multiscape:::available_to_solve("cbc"),
    message = "CBC solver is not available."
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

rcpp_new_optimization_problem <- function(nrow = 1000000L, ncol = 1000000L, ncell = 100000L) {
    .Call(`_multiscape_rcpp_new_optimization_problem`, nrow, ncol, ncell)
}

rcpp_optimization_problem_as_list <- function(x) {
    .Call(`_multiscape_rcpp_optimization_problem_as_list`, x)
}

rcpp_get_optimization_problem_ncol <- function(x) {
    .Call(`_multiscape_rcpp_get_optimization_problem_ncol`, x)
}

rcpp_get_optimization_problem_nrow <- function(x) {
    .Call(`_multiscape_rcpp_get_optimization_problem_nrow`, x)
}

rcpp_get_optimization_problem_ncell <- function(x) {
    .Call(`_multiscape_rcpp_get_optimization_problem_ncell`, x)
}

rcpp_get_optimization_problem_A <- function(x) {
    .Call(`_multiscape_rcpp_get_optimization_problem_A`, x)
}

rcpp_add_action_locks <- function(x, dist_actions_data) {
    .Call(`_multiscape_rcpp_add_action_locks`, x, dist_actions_data)
}

rcpp_add_action_max_per_pu <- function(x, dist_actions_data, max_per_pu, internal_pu_ids, internal_action_ids) {
    .Call(`_multiscape_rcpp_add_action_max_per_pu`, x, dist_actions_data, max_per_pu, internal_pu_ids, internal_action_ids)
}

rcpp_add_base_variables <- function(x, pu_data, dist_actions_data, dist_features_data, add_z = TRUE) {
    .Call(`_multiscape_rcpp_add_base_variables`, x, pu_data, dist_actions_data, dist_features_data, add_z)
}

rcpp_add_linear_constraint <- function(model_ptr, j0, x, sense, rhs, name = "", block_name = "linear_constraint", tag = "") {
    .Call(`_multiscape_rcpp_add_linear_constraint`, model_ptr, j0, x, sense, rhs, name, block_name, tag)
}

rcpp_add_linking_w_le_sum_x <- function(x, dist_actions_data, block_name = "linking_w_le_sum_x", tag = "") {
    .Call(`_multiscape_rcpp_add_linking_w_le_sum_x`, x, dist_actions_data, block_name, tag)
}

rcpp_add_linking_x_le_w <- function(x, dist_actions_data, block_name = "linking_x_le_w", tag = "") {
    .Call(`_multiscape_rcpp_add_linking_x_le_w`, x, dist_actions_data, block_name, tag)
}

rcpp_add_linking_z_le_w <- function(x, dist_features_data, block_name = "linking_z_le_w", tag = "") {
    .Call(`_multiscape_rcpp_add_linking_z_le_w`, x, dist_features_data, block_name, tag)
}

rcpp_add_objective_max_benefit <- function(x, coef_x, weight = 1.0, weight_multiplier = 1.0, block_name = "objective_add_max_benefit", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_max_benefit`, x, coef_x, weight, weight_multiplier, block_name, tag)
}

rcpp_add_objective_max_net_profit <- function(x, pu_data, dist_actions_data, dist_profit_data, profit_col = "profit", include_pu_cost = TRUE, include_action_cost = TRUE, weight = 1.0, block_name = "objective_max_net_profit", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_max_net_profit`, x, pu_data, dist_actions_data, dist_profit_data, profit_col, include_pu_cost, include_action_cost, weight, block_name, tag)
}

rcpp_add_objective_max_profit <- function(x, dist_actions_data, dist_profit_data, profit_col = "profit", weight = 1.0, block_name = "objective_max_profit", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_max_profit`, x, dist_actions_data, dist_profit_data, profit_col, weight, block_name, tag)
}

rcpp_add_objective_min_cost <- function(x, pu_data, dist_actions_data, weight = 1.0, include_pu_cost = TRUE, include_action_cost = TRUE, block_name = "objective_add_min_cost", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_min_cost`, x, pu_data, dist_actions_data, weight, include_pu_cost, include_action_cost, block_name, tag)
}

rcpp_add_objective_min_fragmentation <- function(x, relation_data, weight = 1.0, weight_multiplier = 1.0, block_name = "objective_add_min_fragmentation", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_min_fragmentation`, x, relation_data, weight, weight_multiplier, block_name, tag)
}

rcpp_add_objective_min_fragmentation_actions <- function(x, dist_actions_data, relation_data, actions_to_use = NULL, action_weights = NULL, weight = 1.0, weight_multiplier = 1.0, block_name = "objective_add_min_fragmentation_actions_by_action", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_min_fragmentation_actions`, x, dist_actions_data, relation_data, actions_to_use, action_weights, weight, weight_multiplier, block_name, tag)
}

rcpp_add_objective_min_intervention_impact <- function(x, pu_data, dist_actions_data, dist_features_data, subset_key, impact_col, features_to_use, actions_to_use, internal_feature_col, weight = 1.0, weight_multiplier = 1.0, block_name = "objective_add_min_intervention_impact", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_min_intervention_impact`, x, pu_data, dist_actions_data, dist_features_data, subset_key, impact_col, features_to_use, actions_to_use, internal_feature_col, weight, weight_multiplier, block_name, tag)
}

rcpp_add_objective_min_loss <- function(x, coef_x, weight = 1.0, weight_multiplier = 1.0, block_name = "objective_add_min_loss", tag = "") {
    .Call(`_multiscape_rcpp_add_objective_min_loss`, x, coef_x, weight, weight_multiplier, block_name, tag)
}

rcpp_add_pu_locks <- function(x, pu_data, block_name = "pu_locks", tag = "") {
    .Call(`_multiscape_rcpp_add_pu_locks`, x, pu_data, block_name, tag)
}

rcpp_add_target_recovery <- function(x, features_data, dist_actions_data, dist_benefit_data, target_col, tol = 1e-12) {
    .Call(`_multiscape_rcpp_add_target_recovery`, x, features_data, dist_actions_data, dist_benefit_data, target_col, tol)
}

rcpp_fix_z_ineligible_by_positive_delta <- function(x, dist_features_data, dist_benefit_data, fix_lb_too = TRUE, eps = 1e-12) {
    .Call(`_multiscape_rcpp_fix_z_ineligible_by_positive_delta`, x, dist_features_data, dist_benefit_data, fix_lb_too, eps)
}

rcpp_model_add_columns <- function(x, obj, lb, ub, vtype, names = NULL, block_name = "augmecon_slacks", tag = "") {
    .Call(`_multiscape_rcpp_model_add_columns`, x, obj, lb, ub, vtype, names, block_name, tag)
}

rcpp_model_set_objective_vector <- function(x, obj, model_sense = "min") {
    invisible(.Call(`_multiscape_rcpp_model_set_objective_vector`, x, obj, model_sense))
}

rcpp_reset_objective <- function(x, modelsense = "") {
    .Call(`_multiscape_rcpp_reset_objective`, x, modelsense)
}

rcpp_add_to_objective <- function(x, ind, val) {
    invisible(.Call(`_multiscape_rcpp_add_to_objective`, x, ind, val))
}

rcpp_add_to_objective_scalar <- function(x, ind1, val) {
    invisible(.Call(`_multiscape_rcpp_add_to_objective_scalar`, x, ind1, val))
}

rcpp_prepare_objective_max_benefit <- function(x, dist_actions_data, dist_effects_data, block_name = "prepare_max_benefit", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_max_benefit`, x, dist_actions_data, dist_effects_data, block_name, tag)
}

rcpp_prepare_objective_max_net_profit <- function(x, pu_data, dist_actions_data, dist_profit_data, profit_col = "profit", include_pu_cost = TRUE, include_action_cost = TRUE, block_name = "objective_max_net_profit", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_max_net_profit`, x, pu_data, dist_actions_data, dist_profit_data, profit_col, include_pu_cost, include_action_cost, block_name, tag)
}

rcpp_prepare_objective_max_profit <- function(x, dist_actions_data, dist_profit_data, profit_col = "profit", block_name = "objective_max_profit", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_max_profit`, x, dist_actions_data, dist_profit_data, profit_col, block_name, tag)
}

rcpp_prepare_objective_min_cost <- function(x, pu_data, dist_actions_data, include_pu_cost = TRUE, include_action_cost = TRUE, block_name = "objective_min_cost", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_min_cost`, x, pu_data, dist_actions_data, include_pu_cost, include_action_cost, block_name, tag)
}

rcpp_prepare_objective_min_fragmentation <- function(x, relation_data, block_name = "fragmentation_pu", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_min_fragmentation`, x, relation_data, block_name, tag)
}

rcpp_prepare_objective_min_fragmentation_actions <- function(x, dist_actions_data, relation_data, actions_to_use = NULL, block_name = "fragmentation_actions_by_action", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_min_fragmentation_actions`, x, dist_actions_data, relation_data, actions_to_use, block_name, tag)
}

rcpp_prepare_objective_min_intervention_impact <- function(x, pu_data, dist_actions_data, dist_features_data, subset_key, impact_col, features_to_use, actions_to_use, internal_feature_col = "internal_feature", block_name = "objective_min_intervention_impact", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_min_intervention_impact`, x, pu_data, dist_actions_data, dist_features_data, subset_key, impact_col, features_to_use, actions_to_use, internal_feature_col, block_name, tag)
}

rcpp_prepare_objective_min_loss <- function(x, dist_actions_data, dist_effects_data, block_name = "prepare_min_loss", tag = "") {
    .Call(`_multiscape_rcpp_prepare_objective_min_loss`, x, dist_actions_data, dist_effects_data, block_name, tag)
}

#' @include internal.R
#' @include problem-class.R
#' @include solution-class.R
#' @include solutionset-class.R
NULL

#' Show
#'
#' Display information about an object.
#'
#' @param x Any object.
#'
#' @seealso [methods::show()].
#'
#' @name show
#'
#' @aliases show,Problem-method show,Solution-method show,SolutionSet-method show
#' @keywords internal
#'
NULL

#' @name show
#'
#' @rdname show
#'
#' @usage \S4method{show}{Problem}(x)
#'
methods::setMethod("show", "Problem",
                   function(object) object$show())

#' @name show
#'
#' @rdname show
#'
#' @usage \S4method{show}{SolutionSet}(x)
#'
methods::setMethod("show", "SolutionSet",
                   function(object) object$show())

#' @include internalMO.R
NULL

# -------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------

.pa_plot_spatial_resolve_solutions <- function(x, runs = NULL) {
  if (inherits(x, "Solution")) {
    if (!is.null(runs)) {
      runs <- as.integer(runs)
      if (length(runs) != 1L || is.na(runs[1]) || runs[1] != 1L) {
        stop("For a Solution object, `runs` must be NULL or 1.", call. = FALSE)
      }
    }
    runs <- 1L
    solutions <- list(x)
    names(solutions) <- "1"
    return(solutions)
  }

  if (inherits(x, "SolutionSet")) {
    all_solutions <- x$solution$solutions %||% NULL
    if (is.null(all_solutions) || !is.list(all_solutions) || length(all_solutions) == 0L) {
      stop("SolutionSet has no solutions in x$solution$solutions.", call. = FALSE)
    }

    if (is.null(runs)) {
      runs <- 1L
    } else {
      runs <- as.integer(runs)
      if (any(!is.finite(runs)) || any(is.na(runs)) || any(runs < 1L)) {
        stop("`runs` must contain positive integers.", call. = FALSE)
      }
      if (any(runs > length(all_solutions))) {
        stop(
          "`runs` contains values larger than the number of available solutions (",
          length(all_solutions), ").",
          call. = FALSE
        )
      }
      runs <- unique(runs)
    }

    solutions <- all_solutions[runs]
    names(solutions) <- as.character(runs)
    return(solutions)
  }

  stop("x must be a Solution or SolutionSet.", call. = FALSE)
}

.pa_plot_spatial_get_geometry <- function(solutions) {
  sol0 <- solutions[[1]]
  pr <- sol0$problem %||% NULL
  if (is.null(pr) || !inherits(pr, "Problem")) {
    stop("Solution does not contain a valid problem object.", call. = FALSE)
  }

  pu_sf <- pr$data$pu_sf %||% NULL
  if (is.null(pu_sf) || !inherits(pu_sf, "sf")) {
    stop("No PU geometry found in solution$problem$data$pu_sf.", call. = FALSE)
  }
  if (!("id" %in% names(pu_sf))) {
    stop("PU geometry must contain an 'id' column.", call. = FALSE)
  }

  pu_sf$id <- as.integer(pu_sf$id)
  pu_sf[, "id", drop = FALSE]
}

.pa_plot_spatial_make_base_plot <- function(
    pu_sf_min,
    show_base = TRUE,
    base_fill = "grey92",
    base_color = NA,
    base_alpha = 0.10
) {
  p <- ggplot2::ggplot()
  if (isTRUE(show_base)) {
    p <- p +
      ggplot2::geom_sf(
        data = pu_sf_min,
        fill = base_fill,
        color = base_color,
        alpha = base_alpha
      )
  }
  p + ggplot2::theme_minimal()
}

#' @title Plot spatial outputs from a solution or solution set
#'
#' @description
#' Convenience wrapper to plot spatial outputs from a \code{Solution} or
#' \code{SolutionSet}.
#'
#' Depending on \code{what}, this function dispatches to one of:
#' \itemize{
#'   \item \code{\link{plot_spatial_pu}},
#'   \item \code{\link{plot_spatial_actions}},
#'   \item \code{\link{plot_spatial_features}}.
#' }
#'
#' This wrapper is useful as a compact entry point, while the specialised
#' plotting functions provide a cleaner and more explicit user interface for
#' each spatial output type.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object.
#' @param what Character string indicating what to plot. Must be one of
#'   \code{"pu"}, \code{"actions"}, or \code{"features"}.
#' @param runs Optional integer vector of run ids. If \code{NULL}, a
#'   \code{Solution} is used directly and a \code{SolutionSet} defaults to the
#'   first run.
#' @param actions Optional action subset used when \code{what = "actions"}.
#' @param features Optional feature subset used when \code{what = "features"}.
#' @param value Character string used only when \code{what = "features"}.
#'   Must be one of \code{"final"}, \code{"baseline"}, or \code{"benefit"}.
#' @param layout Character string controlling the layout. Must be one of
#'   \code{"single"} or \code{"facet"}. If \code{NULL}, the default is
#'   \code{"single"} for planning units and actions, and \code{"facet"} for
#'   features.
#' @param max_facets Maximum number of facets shown when faceting without an
#'   explicit action or feature subset.
#' @param ... Additional arguments passed to the specialised plotting function.
#' @param base_alpha Numeric value in \eqn{[0,1]} giving the alpha of the base
#'   planning-unit layer.
#' @param selected_alpha Numeric value in \eqn{[0,1]} giving the alpha of the
#'   highlighted layer.
#' @param base_fill Fill colour for the base planning-unit layer.
#' @param base_color Border colour for the base planning-unit layer.
#' @param selected_color Border colour for highlighted layers.
#' @param draw_borders Logical. If \code{FALSE}, borders are not drawn.
#' @param show_base Logical. If \code{TRUE}, draw the base planning-unit layer
#'   underneath the highlighted output.
#' @param fill_values Optional named vector of colours for discrete maps.
#' @param fill_na Fill colour for missing values.
#' @param use_viridis Logical. If \code{TRUE} and the \pkg{viridis} package is
#'   available, use viridis scales.
#'
#' @return Invisibly returns a \code{ggplot} object.
#'
#' @examples
#' if (requireNamespace("sf", quietly = TRUE) &&
#'     requireNamespace("ggplot2", quietly = TRUE)) {
#'   data("sim_pu_sf", package = "multiscape")
#'
#'   problem <- structure(
#'     list(
#'       data = list(
#'         pu_sf = sim_pu_sf
#'       )
#'     ),
#'     class = "Problem"
#'   )
#'
#'   sol <- structure(
#'     list(
#'       problem = problem,
#'       summary = list(
#'         pu = data.frame(
#'           id = sim_pu_sf$id,
#'           selected = as.integer(seq_len(nrow(sim_pu_sf)) %% 2 == 1)
#'         )
#'       )
#'     ),
#'     class = "Solution"
#'   )
#'
#'   plot_spatial(sol, what = "pu")
#' }
#'
#' @seealso
#' \code{\link{plot_spatial_pu}},
#' \code{\link{plot_spatial_actions}},
#' \code{\link{plot_spatial_features}}
#'
#' @export
plot_spatial <- function(
    x,
    what = c("pu", "actions", "features"),
    runs = NULL,
    actions = NULL,
    features = NULL,
    value = c("final", "baseline", "benefit"),
    layout = NULL,
    max_facets = 4L,
    ...,
    base_alpha = 0.10,
    selected_alpha = 0.90,
    base_fill = "grey92",
    base_color = NA,
    selected_color = NA,
    draw_borders = FALSE,
    show_base = TRUE,
    fill_values = NULL,
    fill_na = "grey80",
    use_viridis = TRUE
) {
  what <- match.arg(what)
  value <- match.arg(value)

  if (identical(what, "pu")) {
    return(plot_spatial_pu(
      x = x,
      runs = runs,
      ...,
      base_alpha = base_alpha,
      selected_alpha = selected_alpha,
      base_fill = base_fill,
      base_color = base_color,
      selected_color = selected_color,
      draw_borders = draw_borders,
      show_base = show_base
    ))
  }

  if (identical(what, "actions")) {
    return(plot_spatial_actions(
      x = x,
      runs = runs,
      actions = actions,
      layout = layout,
      max_facets = max_facets,
      ...,
      base_alpha = base_alpha,
      selected_alpha = selected_alpha,
      base_fill = base_fill,
      base_color = base_color,
      selected_color = selected_color,
      draw_borders = draw_borders,
      show_base = show_base,
      fill_values = fill_values,
      fill_na = fill_na,
      use_viridis = use_viridis
    ))
  }

  plot_spatial_features(
    x = x,
    runs = runs,
    features = features,
    value = value,
    layout = layout,
    max_facets = max_facets,
    ...,
    base_alpha = base_alpha,
    selected_alpha = selected_alpha,
    base_fill = base_fill,
    base_color = base_color,
    selected_color = selected_color,
    draw_borders = draw_borders,
    show_base = show_base,
    fill_na = fill_na,
    use_viridis = use_viridis
  )
}

#' @title Plot selected planning units in space
#'
#' @description
#' Plot the spatial distribution of selected planning units from a
#' \code{Solution} or \code{SolutionSet}.
#'
#' This function maps the planning-unit selection summary returned by
#' \code{\link{get_pu}} onto the planning-unit geometry stored in the associated
#' \code{Problem} object.
#'
#' @details
#' Let \eqn{w_i \in \{0,1\}} denote the planning-unit selection variable for
#' planning unit \eqn{i}. This function plots the user-facing
#' \code{selected == 1} representation of \eqn{w_i}.
#'
#' If several runs are requested, the output is faceted by \code{run_id}.
#'
#' Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object.
#' @param runs Optional integer vector of run ids. If \code{NULL}, a
#'   \code{Solution} is used directly and a \code{SolutionSet} defaults to the
#'   first run.
#' @param ... Reserved for future extensions.
#' @param base_alpha Numeric value in \eqn{[0,1]} giving the alpha of the base
#'   planning-unit layer.
#' @param selected_alpha Numeric value in \eqn{[0,1]} giving the alpha of the
#'   selected planning-unit layer.
#' @param base_fill Fill colour for the base planning-unit layer.
#' @param base_color Border colour for the base planning-unit layer.
#' @param selected_color Border colour for selected planning units.
#' @param draw_borders Logical. If \code{FALSE}, borders are not drawn.
#' @param show_base Logical. If \code{TRUE}, draw the base planning-unit layer
#'   underneath the selected units.
#'
#' @return Invisibly returns a \code{ggplot} object.
#'
#' @examples
#' if (requireNamespace("sf", quietly = TRUE) &&
#'     requireNamespace("ggplot2", quietly = TRUE)) {
#'   data("sim_pu_sf", package = "multiscape")
#'
#'   problem <- structure(
#'     list(
#'       data = list(
#'         pu_sf = sim_pu_sf
#'       )
#'     ),
#'     class = "Problem"
#'   )
#'
#'   sol <- structure(
#'     list(
#'       problem = problem,
#'       summary = list(
#'         pu = data.frame(
#'           id = sim_pu_sf$id,
#'           selected = as.integer(seq_len(nrow(sim_pu_sf)) %% 2 == 1)
#'         )
#'       )
#'     ),
#'     class = "Solution"
#'   )
#'
#'   plot_spatial_pu(sol)
#' }
#'
#' @seealso
#' \code{\link{get_pu}},
#' \code{\link{plot_spatial}},
#' \code{\link{plot_spatial_actions}},
#' \code{\link{plot_spatial_features}}
#'
#' @export
plot_spatial_pu <- function(
    x,
    runs = NULL,
    ...,
    base_alpha = 0.10,
    selected_alpha = 0.90,
    base_fill = "grey92",
    base_color = NA,
    selected_color = NA,
    draw_borders = FALSE,
    show_base = TRUE
) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("plot_spatial_pu() requires the 'sf' package.", call. = FALSE)
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("plot_spatial_pu() requires the 'ggplot2' package.", call. = FALSE)
  }

  if (!isTRUE(draw_borders)) {
    base_color <- NA
    selected_color <- NA
  }

  solutions <- .pa_plot_spatial_resolve_solutions(x, runs = runs)
  multi_runs <- length(solutions) > 1L
  pu_sf_min <- .pa_plot_spatial_get_geometry(solutions)

  pu_list <- vector("list", length(solutions))
  for (i in seq_along(solutions)) {
    sol_i <- solutions[[i]]
    run_i <- as.integer(names(solutions)[i])

    pu_tbl <- get_pu(sol_i, only_selected = FALSE)
    if (!all(c("id", "selected") %in% names(pu_tbl))) {
      stop("PU summary must contain 'id' and 'selected'.", call. = FALSE)
    }

    pu_tbl$id <- as.integer(pu_tbl$id)
    g_i <- merge(pu_sf_min, pu_tbl[, c("id", "selected")], by = "id", all.x = TRUE)
    g_i$selected[is.na(g_i$selected)] <- 0L
    g_i$run_id <- run_i
    pu_list[[i]] <- g_i
  }

  g <- do.call(rbind, pu_list)
  g <- sf::st_as_sf(g)
  g_sel <- g[g$selected %in% 1L, , drop = FALSE]

  if (nrow(g_sel) == 0L) {
    stop("No selected planning units to plot.", call. = FALSE)
  }

  p <- .pa_plot_spatial_make_base_plot(
    pu_sf_min = pu_sf_min,
    show_base = show_base,
    base_fill = base_fill,
    base_color = base_color,
    base_alpha = base_alpha
  ) +
    ggplot2::geom_sf(
      data = g_sel,
      fill = "#2C7FB8",
      color = selected_color,
      alpha = selected_alpha
    ) +
    ggplot2::labs(title = "Selected planning units")

  if (isTRUE(multi_runs)) {
    p <- p + ggplot2::facet_wrap(~run_id)
  }

  print(p)
  invisible(p)
}

#' @title Plot selected actions in space
#'
#' @description
#' Plot the spatial distribution of selected actions from a
#' \code{Solution} or \code{SolutionSet}.
#'
#' This function maps the selected planning unit--action pairs returned by
#' \code{\link{get_actions}(only_selected = TRUE)} onto the planning-unit
#' geometry stored in the associated \code{Problem} object.
#'
#' @details
#' Let \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected in
#' planning unit \eqn{i}. This function plots the selected
#' \code{(pu, action)} pairs in geographic space.
#'
#' If \code{layout = "facet"} and only one run is plotted, one panel is drawn
#' per action.
#'
#' If \code{layout = "single"}, all selected actions are drawn in a single map
#' using discrete fills. If more than one action is selected in the same
#' planning unit, the action labels are collapsed using \code{"+"}.
#'
#' When plotting multiple runs, only \code{layout = "single"} is supported.
#'
#' Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object.
#' @param runs Optional integer vector of run ids. If \code{NULL}, a
#'   \code{Solution} is used directly and a \code{SolutionSet} defaults to the
#'   first run.
#' @param actions Optional action subset to display. Entries may match action
#'   ids or action-set labels.
#' @param layout Character string controlling the layout. Must be one of
#'   \code{"single"} or \code{"facet"}. If \code{NULL}, the default is
#'   \code{"single"}.
#' @param max_facets Maximum number of action facets shown when \code{actions}
#'   is \code{NULL} and faceting would otherwise create many panels.
#' @param ... Reserved for future extensions.
#' @param base_alpha Numeric value in \eqn{[0,1]} giving the alpha of the base
#'   planning-unit layer.
#' @param selected_alpha Numeric value in \eqn{[0,1]} giving the alpha of the
#'   highlighted action layer.
#' @param base_fill Fill colour for the base planning-unit layer.
#' @param base_color Border colour for the base planning-unit layer.
#' @param selected_color Border colour for highlighted layers.
#' @param draw_borders Logical. If \code{FALSE}, borders are not drawn.
#' @param show_base Logical. If \code{TRUE}, draw the base planning-unit layer
#'   underneath the highlighted output.
#' @param fill_values Optional named vector of colours for discrete action maps.
#' @param fill_na Fill colour for missing values.
#' @param use_viridis Logical. If \code{TRUE} and the \pkg{viridis} package is
#'   available, use viridis discrete scales.
#'
#' @return Invisibly returns a \code{ggplot} object.
#'
#' @examples
#' if (requireNamespace("sf", quietly = TRUE) &&
#'     requireNamespace("ggplot2", quietly = TRUE)) {
#'   data("sim_pu_sf", package = "multiscape")
#'
#'   actions_df <- data.frame(
#'     id = c("conservation", "restoration"),
#'     name = c("conservation", "restoration")
#'   )
#'
#'   problem <- structure(
#'     list(
#'       data = list(
#'         pu_sf = sim_pu_sf,
#'         actions = actions_df
#'       )
#'     ),
#'     class = "Problem"
#'   )
#'
#'   ids <- sim_pu_sf$id[seq_len(min(6, nrow(sim_pu_sf)))]
#'   sol <- structure(
#'     list(
#'       problem = problem,
#'       summary = list(
#'         actions = data.frame(
#'           pu = c(ids[1:3], ids[4:6]),
#'           action = c(rep("conservation", 3), rep("restoration", 3)),
#'           selected = 1L
#'         )
#'       )
#'     ),
#'     class = "Solution"
#'   )
#'
#'   plot_spatial_actions(sol, layout = "facet")
#' }
#'
#' @seealso
#' \code{\link{get_actions}},
#' \code{\link{plot_spatial}},
#' \code{\link{plot_spatial_pu}},
#' \code{\link{plot_spatial_features}}
#'
#' @export
plot_spatial_actions <- function(
    x,
    runs = NULL,
    actions = NULL,
    layout = NULL,
    max_facets = 4L,
    ...,
    base_alpha = 0.08,
    selected_alpha = 0.95,
    base_fill = "grey95",
    base_color = NA,
    selected_color = NA,
    draw_borders = FALSE,
    show_base = TRUE,
    fill_values = NULL,
    fill_na = "grey80",
    use_viridis = TRUE
) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("plot_spatial_actions() requires the 'sf' package.", call. = FALSE)
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("plot_spatial_actions() requires the 'ggplot2' package.", call. = FALSE)
  }
  has_viridis <- requireNamespace("viridis", quietly = TRUE)

  if (is.null(layout)) layout <- "single"
  layout <- match.arg(layout, c("single", "facet"))

  if (!isTRUE(draw_borders)) {
    base_color <- NA
    selected_color <- NA
  }

  solutions <- .pa_plot_spatial_resolve_solutions(x, runs = runs)
  multi_runs <- length(solutions) > 1L

  if (isTRUE(multi_runs) && identical(layout, "facet")) {
    stop(
      "When plotting multiple runs for actions, use layout = 'single'. ",
      "Faceting by both run and action is not supported.",
      call. = FALSE
    )
  }

  pu_sf_min <- .pa_plot_spatial_get_geometry(solutions)

  act_list <- vector("list", length(solutions))
  for (i in seq_along(solutions)) {
    sol_i <- solutions[[i]]
    run_i <- as.integer(names(solutions)[i])

    act_tbl <- get_actions(sol_i, only_selected = TRUE)
    if (!all(c("pu", "action", "selected") %in% names(act_tbl))) {
      stop("Action summary must contain 'pu', 'action', and 'selected'.", call. = FALSE)
    }

    act_tbl$pu <- as.integer(act_tbl$pu)
    act_tbl$action <- as.character(act_tbl$action)

    if (!is.null(actions)) {
      keep <- .pa_resolve_action_subset(sol_i$problem, subset = actions)
      keep_ids <- as.character(keep$id)
      act_tbl <- act_tbl[act_tbl$action %in% keep_ids, , drop = FALSE]
    }

    if (nrow(act_tbl) == 0L) next
    act_tbl$run_id <- run_i
    act_list[[i]] <- act_tbl
  }

  act_list <- Filter(Negate(is.null), act_list)
  if (length(act_list) == 0L) {
    stop("No selected actions to plot.", call. = FALSE)
  }

  act_tbl <- do.call(rbind, act_list)
  acts <- unique(act_tbl$action)

  base_theme <- ggplot2::theme_minimal(base_size = 11) +
    ggplot2::theme(
      panel.grid.minor = ggplot2::element_blank(),
      panel.grid.major = ggplot2::element_line(colour = "grey90", linewidth = 0.2),
      legend.title = ggplot2::element_blank(),
      strip.text = ggplot2::element_text(face = "bold"),
      strip.background = ggplot2::element_rect(fill = "grey96", colour = "grey85"),
      plot.title = ggplot2::element_text(face = "bold")
    )

  # -------------------------------------------------------------------
  # one run + action facets
  # -------------------------------------------------------------------
  if (!isTRUE(multi_runs) && identical(layout, "facet")) {
    if (is.null(actions) && length(acts) > max_facets) {
      warning(
        "Showing only the first ", max_facets,
        " actions. Use actions=... or increase max_facets.",
        call. = FALSE
      )
      keep_acts <- acts[seq_len(max_facets)]
      act_tbl <- act_tbl[act_tbl$action %in% keep_acts, , drop = FALSE]
    }

    names(act_tbl)[names(act_tbl) == "pu"] <- "id"
    g <- merge(pu_sf_min, act_tbl[, c("id", "action")], by = "id", all.x = FALSE)
    g <- sf::st_as_sf(g)

    p <- .pa_plot_spatial_make_base_plot(
      pu_sf_min = pu_sf_min,
      show_base = show_base,
      base_fill = base_fill,
      base_color = base_color,
      base_alpha = base_alpha
    ) +
      ggplot2::geom_sf(
        data = g,
        ggplot2::aes(fill = action),
        color = selected_color,
        alpha = selected_alpha
      ) +
      ggplot2::facet_wrap(~action) +
      ggplot2::labs(title = "Selected actions", fill = "") +
      ggplot2::coord_sf(datum = NA) +
      base_theme

    if (!is.null(fill_values)) {
      p <- p + ggplot2::scale_fill_manual(values = fill_values, na.value = fill_na)
    } else if (isTRUE(use_viridis) && has_viridis) {
      p <- p + viridis::scale_fill_viridis(discrete = TRUE, option = "C")
    }

    print(p)
    return(invisible(p))
  }

  # -------------------------------------------------------------------
  # single layout (possibly multiple runs)
  # -------------------------------------------------------------------
  lab_list <- split(act_tbl, act_tbl$run_id)
  warned_multi_action <- FALSE

  lab_out <- lapply(names(lab_list), function(rr) {
    dd <- lab_list[[rr]]

    tmp <- stats::aggregate(
      action ~ pu,
      data = dd,
      FUN = function(z) {
        z <- unique(as.character(z))
        if (length(z) > 1L) warned_multi_action <<- TRUE
        paste(sort(z), collapse = "+")
      }
    )

    names(tmp)[names(tmp) == "pu"] <- "id"
    tmp$id <- as.integer(tmp$id)
    tmp$action <- as.character(tmp$action)
    tmp$run_id <- as.integer(rr)
    tmp
  })

  if (isTRUE(warned_multi_action)) {
    warning(
      "More than one action detected in at least one PU. Labels were collapsed using '+'.",
      call. = FALSE
    )
  }

  lab_act <- do.call(rbind, lab_out)
  g <- merge(pu_sf_min, lab_act, by = "id", all.x = FALSE)
  g <- sf::st_as_sf(g)

  if (nrow(g) == 0L) {
    stop("No geometry matched the plotted action labels.", call. = FALSE)
  }

  p <- .pa_plot_spatial_make_base_plot(
    pu_sf_min = pu_sf_min,
    show_base = show_base,
    base_fill = base_fill,
    base_color = base_color,
    base_alpha = base_alpha
  ) +
    ggplot2::geom_sf(
      data = g,
      ggplot2::aes(fill = action),
      color = selected_color,
      alpha = selected_alpha
    ) +
    ggplot2::labs(title = "Selected actions", fill = "") +
    ggplot2::coord_sf(datum = NA) +
    base_theme

  if (isTRUE(multi_runs)) {
    p <- p +
      ggplot2::facet_wrap(
        ~run_id,
        labeller = ggplot2::labeller(run_id = function(x) paste("Run", x))
      ) +
      ggplot2::theme(
        axis.text = ggplot2::element_blank(),
        axis.title = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_blank()
      )
  }

  if (!is.null(fill_values)) {
    p <- p + ggplot2::scale_fill_manual(values = fill_values, na.value = fill_na)
  } else if (isTRUE(use_viridis) && has_viridis) {
    p <- p + viridis::scale_fill_viridis(discrete = TRUE, option = "C")
  }

  print(p)
  invisible(p)
}

#' @title Plot spatial feature values from a solution
#'
#' @description
#' Plot feature values in space from a \code{Solution} or \code{SolutionSet}.
#'
#' This function combines baseline feature amounts from
#' \code{x$problem$data$dist_features} with positive effects induced by selected
#' actions to produce planning-unit-level feature maps.
#'
#' @details
#' For each planning unit \eqn{i} and feature \eqn{f}, the plotted quantities
#' are:
#' \deqn{
#' \mathrm{baseline}_{if},
#' }
#' \deqn{
#' \mathrm{benefit}_{if},
#' }
#' \deqn{
#' \mathrm{final}_{if} = \mathrm{baseline}_{if} + \mathrm{benefit}_{if}.
#' }
#'
#' In the current implementation:
#' \itemize{
#'   \item \code{baseline} is the summed baseline amount from
#'   \code{x$problem$data$dist_features},
#'   \item \code{benefit} is the summed positive effect from selected actions,
#'   \item \code{final} is \code{baseline + benefit}.
#' }
#'
#' Negative effects are not subtracted in this plotting method. Therefore,
#' \code{value = "final"} should be interpreted as baseline plus selected
#' positive effects under the current plotting logic.
#'
#' If \code{layout = "facet"} and only one run is plotted, one panel is drawn
#' per feature.
#'
#' If multiple runs are plotted, exactly one feature must be requested, and
#' faceting is done by run.
#'
#' Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object.
#' @param runs Optional integer vector of run ids. If \code{NULL}, a
#'   \code{Solution} is used directly and a \code{SolutionSet} defaults to the
#'   first run.
#' @param features Optional feature subset to display. Matching is attempted
#'   against both feature ids and feature names.
#' @param value Character string indicating which feature quantity to plot. Must
#'   be one of \code{"final"}, \code{"baseline"}, or \code{"benefit"}.
#' @param layout Character string controlling the layout. Must be one of
#'   \code{"single"} or \code{"facet"}. If \code{NULL}, the default is
#'   \code{"facet"}.
#' @param max_facets Maximum number of feature facets shown when
#'   \code{features = NULL} and faceting would otherwise create many panels.
#' @param ... Reserved for future extensions.
#' @param base_alpha Unused in the current feature view, kept for interface
#'   consistency.
#' @param selected_alpha Unused in the current feature view, kept for interface
#'   consistency.
#' @param base_fill Unused in the current feature view, kept for interface
#'   consistency.
#' @param base_color Unused in the current feature view, kept for interface
#'   consistency.
#' @param selected_color Border colour for filled feature polygons.
#' @param draw_borders Logical. If \code{FALSE}, borders are not drawn.
#' @param show_base Unused in the current feature view, kept for interface
#'   consistency.
#' @param fill_na Fill colour for missing values.
#' @param use_viridis Logical. If \code{TRUE} and the \pkg{viridis} package is
#'   available, use a continuous viridis scale.
#'
#' @return Invisibly returns a \code{ggplot} object.
#'
#' @examples
#' if (requireNamespace("sf", quietly = TRUE) &&
#'     requireNamespace("ggplot2", quietly = TRUE)) {
#'   data("sim_pu_sf", package = "multiscape")
#'
#'   n <- min(6, nrow(sim_pu_sf))
#'   ids <- sim_pu_sf$id[seq_len(n)]
#'
#'   features_df <- data.frame(
#'     id = c(1, 2),
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_features_df <- data.frame(
#'     pu = rep(ids, times = 2),
#'     feature = rep(c(1, 2), each = n),
#'     amount = c(seq_len(n), rev(seq_len(n)))
#'   )
#'
#'   dist_effects_df <- data.frame(
#'     pu = ids,
#'     action = "conservation",
#'     feature = rep(c(1, 2), length.out = n),
#'     benefit = rep(1, n)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   problem <- structure(
#'     list(
#'       data = list(
#'         pu_sf = sim_pu_sf,
#'         features = features_df,
#'         dist_features = dist_features_df,
#'         dist_effects = dist_effects_df,
#'         actions = actions_df
#'       )
#'     ),
#'     class = "Problem"
#'   )
#'
#'   sol <- structure(
#'     list(
#'       problem = problem,
#'       summary = list(
#'         actions = data.frame(
#'           pu = ids,
#'           action = "conservation",
#'           selected = 1L
#'         )
#'       )
#'     ),
#'     class = "Solution"
#'   )
#'
#'   plot_spatial_features(sol, features = "feature_1", value = "final")
#' }
#'
#' @seealso
#' \code{\link{get_features}},
#' \code{\link{plot_spatial}},
#' \code{\link{plot_spatial_pu}},
#' \code{\link{plot_spatial_actions}}
#'
#' @export
plot_spatial_features <- function(
    x,
    runs = NULL,
    features = NULL,
    value = c("final", "baseline", "benefit"),
    layout = NULL,
    max_facets = 4L,
    ...,
    base_alpha = 0.10,
    selected_alpha = 0.90,
    base_fill = "grey92",
    base_color = NA,
    selected_color = NA,
    draw_borders = FALSE,
    show_base = TRUE,
    fill_na = "grey80",
    use_viridis = TRUE
) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("plot_spatial_features() requires the 'sf' package.", call. = FALSE)
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("plot_spatial_features() requires the 'ggplot2' package.", call. = FALSE)
  }
  has_viridis <- requireNamespace("viridis", quietly = TRUE)

  value <- match.arg(value)
  if (is.null(layout)) layout <- "facet"
  layout <- match.arg(layout, c("single", "facet"))

  if (!isTRUE(draw_borders)) {
    selected_color <- NA
  }

  solutions <- .pa_plot_spatial_resolve_solutions(x, runs = runs)
  multi_runs <- length(solutions) > 1L

  if (isTRUE(multi_runs)) {
    if (is.null(features) || length(features) != 1L) {
      stop(
        "When plotting multiple runs for features, `features` must specify exactly one feature.",
        call. = FALSE
      )
    }
  }

  pu_sf_min <- .pa_plot_spatial_get_geometry(solutions)
  pr <- solutions[[1]]$problem

  distf <- pr$data$dist_features %||% NULL
  feats <- pr$data$features %||% NULL

  if (is.null(distf) || !inherits(distf, "data.frame")) {
    stop("Missing problem$data$dist_features.", call. = FALSE)
  }
  if (is.null(feats) || !inherits(feats, "data.frame")) {
    stop("Missing problem$data$features.", call. = FALSE)
  }
  if (!all(c("pu", "feature", "amount") %in% names(distf))) {
    stop("dist_features must contain 'pu', 'feature', and 'amount'.", call. = FALSE)
  }

  feat_map <- feats[, intersect(c("id", "name"), names(feats)), drop = FALSE]
  if (!("name" %in% names(feat_map))) feat_map$name <- as.character(feat_map$id)

  feature_frames <- vector("list", length(solutions))
  for (i in seq_along(solutions)) {
    sol_i <- solutions[[i]]
    run_i <- as.integer(names(solutions)[i])

    base_tbl <- stats::aggregate(
      amount ~ pu + feature,
      data = distf,
      FUN = sum
    )
    names(base_tbl)[names(base_tbl) == "amount"] <- "baseline"

    eff <- sol_i$problem$data$dist_effects_model %||% sol_i$problem$data$dist_effects %||% NULL
    ben_tbl <- NULL

    if (!is.null(eff) &&
        inherits(eff, "data.frame") &&
        all(c("pu", "feature", "action", "benefit") %in% names(eff))) {

      act_sel <- get_actions(sol_i, only_selected = TRUE)
      if (all(c("pu", "action") %in% names(act_sel))) {
        key_sel <- paste(act_sel$pu, act_sel$action, sep = "||")
        key_eff <- paste(eff$pu, eff$action, sep = "||")
        eff2 <- eff[key_eff %in% key_sel, , drop = FALSE]

        if (nrow(eff2) > 0L) {
          ben_tbl <- stats::aggregate(
            benefit ~ pu + feature,
            data = eff2,
            FUN = sum
          )
          names(ben_tbl)[names(ben_tbl) == "benefit"] <- "benefit"
        }
      }
    }

    if (is.null(ben_tbl)) {
      ben_tbl <- data.frame(
        pu = integer(0),
        feature = numeric(0),
        benefit = numeric(0)
      )
    }

    ff <- merge(base_tbl, ben_tbl, by = c("pu", "feature"), all = TRUE)
    ff$baseline[is.na(ff$baseline)] <- 0
    ff$benefit[is.na(ff$benefit)] <- 0
    ff$final <- ff$baseline + ff$benefit

    ff <- merge(ff, feat_map, by.x = "feature", by.y = "id", all.x = TRUE)
    ff$feature_label <- as.character(ff$name)
    ff$run_id <- run_i

    if (!is.null(features)) {
      features_chr <- as.character(features)
      ff <- ff[
        ff$feature_label %in% features_chr |
          as.character(ff$feature) %in% features_chr,
        ,
        drop = FALSE
      ]
    }

    feature_frames[[i]] <- ff
  }

  ff <- do.call(rbind, feature_frames)
  if (nrow(ff) == 0L) {
    stop("No features available to plot for the requested subset.", call. = FALSE)
  }

  if (!isTRUE(multi_runs)) {
    feat_levels <- unique(ff$feature_label)
    if (identical(layout, "facet") && is.null(features) && length(feat_levels) > max_facets) {
      warning(
        "Showing only the first ", max_facets,
        " features. Use features=... or increase max_facets.",
        call. = FALSE
      )
      keep_feats <- feat_levels[seq_len(max_facets)]
      ff <- ff[ff$feature_label %in% keep_feats, , drop = FALSE]
    }
  }

  names(ff)[names(ff) == "pu"] <- "id"
  g <- merge(pu_sf_min, ff[, c("id", "feature_label", "run_id", value)], by = "id", all.y = TRUE)
  g <- sf::st_as_sf(g)

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(
      data = g,
      ggplot2::aes(fill = .data[[value]]),
      color = selected_color,
      alpha = 1
    ) +
    ggplot2::theme_minimal() +
    ggplot2::labs(title = paste("Spatial", value, "by feature"), fill = value)

  if (isTRUE(multi_runs)) {
    p <- p + ggplot2::facet_wrap(~run_id)
  } else if (identical(layout, "facet")) {
    p <- p + ggplot2::facet_wrap(~feature_label)
  }

  if (isTRUE(use_viridis) && has_viridis) {
    p <- p + viridis::scale_fill_viridis(option = "C", na.value = fill_na)
  }

  print(p)
  invisible(p)
}


#' @title Plot trade-offs from a multi-objective solution set
#'
#' @description
#' Plot pairwise trade-offs among objective values stored in a
#' \code{SolutionSet}.
#'
#' This function is intended for multi-objective workflows in which the solution
#' set contains one row per run and one or more objective value columns of the
#' form \code{value_*}.
#'
#' If exactly two objectives are selected, the function returns a single
#' scatterplot. If three or more objectives are selected, all pairwise
#' combinations are plotted using facets.
#'
#' @details
#' This function reads the run-level table stored in \code{x$solution$runs}. It
#' expects objective values to be stored in columns whose names begin with
#' \code{"value_"}.
#'
#' If the available objective columns are, for example,
#' \code{value_cost}, \code{value_benefit}, and \code{value_frag}, then the
#' corresponding objective aliases are \code{"cost"}, \code{"benefit"}, and
#' \code{"frag"}.
#'
#' Let \eqn{f_k(r)} denote the value of objective \eqn{k} in run \eqn{r}. This
#' function visualizes pairwise projections of the run table of the form:
#' \deqn{
#' \left(f_k(r), f_\ell(r)\right)
#' }
#' for selected pairs of objectives \eqn{k,\ell}.
#'
#' If exactly two objectives are selected, a single panel is produced.
#'
#' If three or more objectives are selected, all pairwise combinations are
#' generated:
#' \deqn{
#' \{(k,\ell): k < \ell,\; k,\ell \in \mathcal{O}\},
#' }
#' where \eqn{\mathcal{O}} is the selected set of objective aliases.
#'
#' By default, plotting more than four objectives is not allowed unless
#' \code{all_pairs = TRUE}, because the number of panels grows quadratically in
#' the number of objectives.
#'
#' \strong{Colouring}
#'
#' If \code{color_by} is supplied, points are coloured by either:
#' \itemize{
#'   \item one of the selected objective aliases, in which case the
#'   corresponding \code{value_*} column is used,
#'   \item or one of the run-level columns \code{run_id}, \code{status},
#'   \code{runtime}, or \code{gap}.
#' }
#'
#' \strong{Connecting runs}
#'
#' If \code{connect = TRUE}, runs are connected in their current table order
#' within each panel. This can be useful when runs correspond to an ordered scan
#' of weights, \eqn{\epsilon}-levels, or frontier points, but it should be used
#' with care when run order has no substantive meaning.
#'
#' \strong{Run labels}
#'
#' If \code{label_runs = TRUE}, each point is labelled by its \code{run_id}. If
#' the \pkg{ggrepel} package is available, repelled labels are used.
#'
#' @param x A \code{SolutionSet} object.
#' @param objectives Optional character vector of objective aliases to display.
#'   These must match the suffixes of the \code{value_*} columns in
#'   \code{x$solution$runs}. If \code{NULL}, all available objective columns are
#'   used.
#' @param color_by Optional character scalar used to colour points. This may be
#'   either one of the selected objective aliases or one of the run-level
#'   columns \code{"run_id"}, \code{"status"}, \code{"runtime"}, or
#'   \code{"gap"}.
#' @param all_pairs Logical. If \code{TRUE}, allow plotting all pairwise
#'   combinations even when more than four objectives are selected. If
#'   \code{NULL}, it is treated as \code{FALSE}.
#' @param connect Logical. If \code{TRUE}, connect points by run order within
#'   each panel.
#' @param label_runs Logical. If \code{TRUE}, add run labels to points.
#' @param point_size Numeric point size.
#' @param line_alpha Numeric alpha value for connecting lines.
#' @param text_size Numeric size for run labels.
#' @param ... Reserved for future extensions.
#'
#' @return Invisibly returns a \code{ggplot} object.
#'
#' @examples
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   solset <- structure(
#'     list(
#'       solution = list(
#'         runs = data.frame(
#'           run_id = 1:5,
#'           status = rep("optimal", 5),
#'           runtime = c(1.2, 1.1, 1.4, 1.3, 1.5),
#'           gap = c(0, 0, 0, 0, 0),
#'           value_cost = c(10, 12, 14, 16, 18),
#'           value_benefit = c(5, 7, 8, 9, 10),
#'           value_loss = c(4, 3, 3, 2, 2)
#'         )
#'       )
#'     ),
#'     class = "SolutionSet"
#'   )
#'
#'   # Plot all pairwise trade-offs
#'   plot_tradeoff(solset)
#'
#'   # Plot two selected objectives
#'   plot_tradeoff(solset, objectives = c("cost", "benefit"))
#'
#'   # Colour points by one objective and label runs
#'   plot_tradeoff(solset, color_by = "loss", label_runs = TRUE)
#' }
#'
#' @seealso
#' \code{\link{solve}},
#' \code{\link{get_solution_vector}}
#'
#' @export
plot_tradeoff <- function(
    x,
    objectives = NULL,
    color_by = NULL,
    all_pairs = NULL,
    connect = FALSE,
    label_runs = FALSE,
    point_size = 3,
    line_alpha = 0.5,
    text_size = 3,
    ...
) {
  if (!inherits(x, "SolutionSet")) {
    stop("plot_tradeoff() requires a SolutionSet object.", call. = FALSE)
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("plot_tradeoff() requires the 'ggplot2' package.", call. = FALSE)
  }

  runs <- x$solution$runs %||% NULL
  if (is.null(runs) || !inherits(runs, "data.frame") || nrow(runs) == 0L) {
    stop("No run table found in x$solution$runs.", call. = FALSE)
  }

  value_cols <- grep("^value_", names(runs), value = TRUE)
  if (length(value_cols) < 2L) {
    stop("At least two objective value columns ('value_*') are required.", call. = FALSE)
  }

  available_obj <- sub("^value_", "", value_cols)

  if (is.null(objectives)) {
    objectives <- available_obj
  } else {
    objectives <- unique(as.character(objectives))
    objectives <- objectives[!is.na(objectives) & nzchar(objectives)]
    if (length(objectives) < 2L) {
      stop("`objectives` must contain at least two objective names.", call. = FALSE)
    }
    bad <- setdiff(objectives, available_obj)
    if (length(bad) > 0L) {
      stop(
        "Unknown objective(s): ", paste(bad, collapse = ", "),
        ". Available objectives are: ", paste(available_obj, collapse = ", "),
        ".",
        call. = FALSE
      )
    }
  }

  all_pairs <- isTRUE(all_pairs)

  if (length(objectives) > 4L && !all_pairs) {
    stop(
      "More than four objectives were selected (", length(objectives), "). ",
      "Please specify a smaller set in `objectives = ...` or use `all_pairs = TRUE`.",
      call. = FALSE
    )
  }

  if (!is.null(color_by)) {
    color_by <- as.character(color_by)[1]
    valid_color_vars <- c(objectives, "run_id", "status", "runtime", "gap")
    if (is.na(color_by) || !nzchar(color_by) || !color_by %in% valid_color_vars) {
      stop(
        "`color_by` must be one of: ",
        paste(valid_color_vars, collapse = ", "),
        ".",
        call. = FALSE
      )
    }
  }

  pair_mat <- utils::combn(objectives, 2)
  pair_df <- vector("list", ncol(pair_mat))

  for (i in seq_len(ncol(pair_mat))) {
    ox <- pair_mat[1, i]
    oy <- pair_mat[2, i]

    dd <- data.frame(
      run_id = runs$run_id %||% seq_len(nrow(runs)),
      run_label = paste("run", runs$run_id %||% seq_len(nrow(runs))),
      status = if ("status" %in% names(runs)) runs$status else NA_character_,
      runtime = if ("runtime" %in% names(runs)) runs$runtime else NA_real_,
      gap = if ("gap" %in% names(runs)) runs$gap else NA_real_,
      obj_x = ox,
      obj_y = oy,
      x = as.numeric(runs[[paste0("value_", ox)]]),
      y = as.numeric(runs[[paste0("value_", oy)]]),
      pair = paste0(ox, " vs ", oy),
      stringsAsFactors = FALSE
    )

    if (!is.null(color_by)) {
      dd$color_value <- if (color_by %in% objectives) {
        runs[[paste0("value_", color_by)]]
      } else {
        runs[[color_by]]
      }
    }

    dd <- dd[order(dd$run_id), , drop = FALSE]
    pair_df[[i]] <- dd
  }

  plot_df <- do.call(rbind, pair_df)

  base_theme <- ggplot2::theme_bw(base_size = 11) +
    ggplot2::theme(
      panel.grid.minor = ggplot2::element_blank(),
      strip.background = ggplot2::element_rect(fill = "grey95", colour = "grey80"),
      strip.text = ggplot2::element_text(face = "bold"),
      legend.position = "right"
    )

  p <- ggplot2::ggplot(plot_df, ggplot2::aes(x = x, y = y))

  if (isTRUE(connect)) {
    p <- p + ggplot2::geom_line(
      ggplot2::aes(group = 1),
      alpha = line_alpha,
      linewidth = 0.5,
      colour = "grey50"
    )
  }

  if (is.null(color_by)) {
    p <- p + ggplot2::geom_point(
      size = point_size,
      shape = 21,
      fill = "#2C7FB8",
      colour = "white",
      stroke = 0.3
    )
  } else {
    p <- p + ggplot2::geom_point(
      ggplot2::aes(fill = color_value),
      size = point_size,
      shape = 21,
      colour = "white",
      stroke = 0.3
    )
  }

  if (isTRUE(label_runs)) {
    if (requireNamespace("ggrepel", quietly = TRUE)) {
      p <- p + ggrepel::geom_text_repel(
        ggplot2::aes(label = .data$run_label),
        size = text_size,
        max.overlaps = Inf,
        show.legend = FALSE
      )
    } else {
      p <- p + ggplot2::geom_text(
        ggplot2::aes(label = .data$run_label),
        size = text_size,
        vjust = -0.6,
        show.legend = FALSE
      )
    }
  }

  if (ncol(pair_mat) == 1L) {
    p <- p +
      ggplot2::labs(
        x = unique(plot_df$obj_x),
        y = unique(plot_df$obj_y),
        fill = if (!is.null(color_by)) color_by else NULL
      ) +
      base_theme
  } else {
    p <- p +
      ggplot2::facet_wrap(~pair, scales = "free") +
      ggplot2::labs(
        x = "Objective value",
        y = "Objective value",
        fill = if (!is.null(color_by)) color_by else NULL
      ) +
      base_theme
  }

  print(p)
  invisible(p)
}

#' @include internal.R
#'
#' @export
if (!methods::isClass("Solution")) methods::setOldClass("Solution")
NULL

#' @name solution-class
#' @aliases Solution
#' @title Solution class
#'
#' @description
#' The \code{Solution} class stores the result of solving a single
#' \code{\link{Problem}} object in \pkg{multiscape}.
#'
#' A \code{Solution} object contains the original problem definition, the core
#' optimization output returned by the solver, user-facing summary tables,
#' diagnostics about the solve process, and metadata describing how the solution
#' was obtained.
#'
#' Objects of this class are typically created by \code{\link{solve}}.
#'
#' @details
#' \strong{Conceptual role}
#'
#' The \code{Solution} class represents the output of one optimization run.
#'
#' It should be understood as the single-run counterpart of the modelling
#' workflow:
#' \preformatted{
#' Problem -> solve() -> Solution
#' }
#'
#' Thus, a \code{Solution} object does not replace the original
#' \code{Problem}; instead, it keeps a reference to it in the \code{problem}
#' field and augments it with optimization results.
#'
#' \strong{Single-run semantics}
#'
#' A \code{Solution} corresponds to one realized solution of one configured
#' optimization problem. This may come from:
#' \itemize{
#'   \item a single-objective solve,
#'   \item one run of a weighted multi-objective workflow,
#'   \item one \eqn{\epsilon}-constraint subproblem,
#'   \item one AUGMECON subproblem,
#'   \item or any other workflow that ultimately produces one optimizer output.
#' }
#'
#' When multiple runs are generated, they are typically collected in a separate
#' \code{SolutionSet} object rather than a single \code{Solution}.
#'
#' \strong{Internal structure}
#'
#' A \code{Solution} object separates results into several layers:
#'
#' \describe{
#'   \item{\code{problem}}{The original \code{Problem} object used to generate
#'   the solution.}
#'
#'   \item{\code{solution}}{A named \code{list} containing the core optimization
#'   output. This may include the objective value, raw model variable vector,
#'   decoded decision vectors, and evaluated objective values by alias.}
#'
#'   \item{\code{summary}}{A named \code{list} of user-facing summary tables,
#'   typically derived from the original problem and the solved decisions. These
#'   tables are intended for inspection, plotting, reporting, and downstream
#'   analysis.}
#'
#'   \item{\code{diagnostics}}{A named \code{list} containing solver diagnostics
#'   such as status, runtime, optimality gap, solver name, and runtime
#'   settings.}
#'
#'   \item{\code{method}}{A named \code{list} describing the optimization method
#'   used to obtain the solution.}
#'
#'   \item{\code{meta}}{A named \code{list} containing additional metadata.}
#'
#'   \item{\code{name}}{A \code{character(1)} identifier for the solution
#'   object.}
#' }
#'
#' \strong{Core optimization output}
#'
#' The \code{solution} field stores the solver-facing result. Typical entries may
#' include:
#' \describe{
#'   \item{\code{objective}}{The scalar objective value returned by the solver
#'   for this run.}
#'   \item{\code{vector}}{The raw internal solution vector in model-variable
#'   order.}
#'   \item{\code{alias_values}}{Objective values evaluated for registered
#'   aliases, when available.}
#' }
#'
#' The raw solution vector may contain not only user-facing decision variables
#' such as planning-unit or action decisions, but also auxiliary variables
#' introduced internally by the model builder.
#'
#' \strong{User-facing summaries}
#'
#' The \code{summary} field stores derived tables intended for interpretation
#' rather than solver interaction. Typical entries include:
#' \describe{
#'   \item{\code{pu}}{Planning-unit summary table.}
#'   \item{\code{actions}}{Planning unit--action allocation summary table.}
#'   \item{\code{features}}{Feature-level outcome summary table.}
#'   \item{\code{targets}}{Target-achievement summary table, when targets were
#'   part of the problem.}
#' }
#'
#' These tables are the main source used by user-facing accessors such as
#' \code{\link{get_pu}}, \code{\link{get_actions}},
#' \code{\link{get_features}}, and \code{\link{get_targets}}.
#'
#' \strong{Diagnostics}
#'
#' The \code{diagnostics} field stores metadata about the optimization process,
#' including solver status and runtime information. Typical entries may include:
#' \itemize{
#'   \item solver name,
#'   \item runtime in seconds,
#'   \item optimality gap,
#'   \item number of cores,
#'   \item time limit,
#'   \item status code or status label.
#' }
#'
#' These values describe how the solution was obtained, not the content of the
#' solution itself.
#'
#' \strong{Printing}
#'
#' The \code{print()} method is intended as a concise diagnostic summary. It
#' reports:
#' \itemize{
#'   \item solver status,
#'   \item objective value,
#'   \item optimality gap,
#'   \item runtime,
#'   \item counts of selected planning units and actions,
#'   \item target fulfillment summary when available,
#'   \item evaluated objective alias values when available,
#'   \item and basic solver information.
#' }
#'
#' This summary is intended for quick inspection. More detailed exploration
#' should use the stored \code{summary}, \code{solution}, and
#' \code{diagnostics} fields directly, or the dedicated accessor functions.
#'
#' @section Fields:
#' \describe{
#'   \item{\code{problem}}{The \code{Problem} object used to generate the
#'   solution.}
#'   \item{\code{solution}}{A named \code{list} containing the core optimization
#'   result.}
#'   \item{\code{summary}}{A named \code{list} of user-facing summary tables.}
#'   \item{\code{diagnostics}}{A named \code{list} of solver diagnostics.}
#'   \item{\code{method}}{A named \code{list} describing the method used.}
#'   \item{\code{meta}}{A named \code{list} of additional metadata.}
#'   \item{\code{name}}{A \code{character(1)} name for the solution object.}
#' }
#'
#' @section Methods:
#' \describe{
#'   \item{\code{print()}}{Print a concise summary of the solution, including
#'   status, objective value, runtime, selection counts, and target
#'   fulfillment.}
#'
#'   \item{\code{show()}}{Alias of \code{print()}.}
#'
#'   \item{\code{repr()}}{Return a short one-line representation of the
#'   solution.}
#' }
#'
#' @return No return value. This page documents the \code{Solution} class.
#'
#' @seealso
#' \code{\link{problem-class}},
#' \code{\link{get_pu}},
#' \code{\link{get_actions}},
#' \code{\link{get_features}},
#' \code{\link{get_targets}},
#' \code{\link{solve}}
NULL

# internal helper
.pa_solution_status_class <- function(status_txt) {
  status_txt <- as.character(status_txt %||% "unknown")[1]

  if (status_txt %in% c("optimal")) {
    return("ok")
  }

  if (status_txt %in% c("time_limit_feasible", "solution_limit")) {
    return("warn")
  }

  if (status_txt %in% c("infeasible_or_unbounded", "time_limit_no_solution", "unknown")) {
    return("bad")
  }

  "muted"
}

# internal helper
.pa_solution_status_inline <- function(status_txt) {
  cls <- .pa_solution_status_class(status_txt)
  paste0("{.", cls, " ", status_txt, "}")
}

# internal helper
.pa_pct_text <- function(x, digits = 3) {
  if (!is.numeric(x) || length(x) == 0 || is.na(x) || !is.finite(x)) {
    return("NA")
  }
  paste0(round(100 * x, digits), "%")
}

# internal helper
.pa_num_text <- function(x, digits = 6) {
  if (!is.numeric(x) || length(x) == 0 || is.na(x) || !is.finite(x)) {
    return("NA")
  }
  format(round(x, digits), trim = TRUE, scientific = FALSE)
}

# internal helper
.pa_n_of_total_text <- function(n, total) {
  if (is.null(n) || length(n) == 0 || is.na(n)) n <- 0L
  if (is.null(total) || length(total) == 0 || is.na(total)) {
    return(as.character(n))
  }
  paste0(n, " of ", total)
}

# internal helper
.pa_solution_alias_values_summary <- function(self, max_show = 6L) {
  av <- self$solution$alias_values %||% NULL

  if (is.null(av) || length(av) == 0) {
    return(NULL)
  }

  av <- unlist(av, use.names = TRUE)
  nm <- names(av) %||% rep("objective", length(av))

  if (length(av) > max_show) {
    keep <- seq_len(max_show)
    txt <- paste0(
      nm[keep], ": ", vapply(av[keep], .pa_num_text, character(1)),
      collapse = ", "
    )
    txt <- paste0(txt, ", ...")
  } else {
    txt <- paste0(
      nm, ": ", vapply(av, .pa_num_text, character(1)),
      collapse = ", "
    )
  }

  txt
}

#' @export
Solution <- pproto(
  "Solution",
  problem = NULL,
  solution = list(),
  summary = list(),
  diagnostics = list(),
  method = list(),
  meta = list(),
  name = "sol",

  print = function(self) {
    ch <- .pa_cli_box_chars()
    div_id <- cli::cli_div(theme = .pa_cli_theme())

    cli::cli_text("A multiscape solution ({.cls Solution})")

    obj <- self$solution$objective %||% NA_real_
    gap <- self$diagnostics$gap %||% NA_real_
    rt  <- self$diagnostics$runtime %||% NA_real_
    st  <- getStatus(self)

    sm <- self$summary %||% list()
    dg <- self$diagnostics %||% list()
    pr <- self$problem %||% NULL

    # ---- totals from problem
    n_pu_total <- if (!is.null(pr) && !is.null(pr$data$pu) && inherits(pr$data$pu, "data.frame")) {
      nrow(pr$data$pu)
    } else {
      NA_integer_
    }

    n_act_total <- if (!is.null(pr) && !is.null(pr$data$dist_actions) && inherits(pr$data$dist_actions, "data.frame")) {
      nrow(pr$data$dist_actions)
    } else {
      NA_integer_
    }

    n_tgt_total <- if (!is.null(sm$targets) && inherits(sm$targets, "data.frame")) {
      nrow(sm$targets)
    } else {
      NA_integer_
    }

    # ---- selected counts
    n_pu_sel <- if (!is.null(sm$pu) && inherits(sm$pu, "data.frame") && "selected" %in% names(sm$pu)) {
      sum(sm$pu$selected %in% 1L, na.rm = TRUE)
    } else {
      NA_integer_
    }

    n_act_sel <- if (!is.null(sm$actions) && inherits(sm$actions, "data.frame") && "selected" %in% names(sm$actions)) {
      sum(sm$actions$selected %in% 1L, na.rm = TRUE)
    } else {
      NA_integer_
    }

    n_tgt_met <- if (!is.null(sm$targets) && inherits(sm$targets, "data.frame") && "gap" %in% names(sm$targets)) {
      sum(sm$targets$gap >= 0, na.rm = TRUE)
    } else {
      NA_integer_
    }

    alias_txt <- .pa_solution_alias_values_summary(self)

    # ---- RESULT
    cli::cli_text("{ch$j}{ch$b}{.h result}", .envir = environment())
    cli::cli_text(
      "{ch$v}{ch$j}{ch$b}status:          {.eval .pa_solution_status_inline(st)}",
      .envir = environment()
    )
    cli::cli_text(
      "{ch$v}{ch$j}{ch$b}objective value: {.strong {.val { .pa_num_text(obj) }}}",
      .envir = environment()
    )
    cli::cli_text(
      "{ch$v}{ch$j}{ch$b}gap:             { .pa_pct_text(gap) }",
      .envir = environment()
    )
    cli::cli_text(
      "{ch$v}{ch$l}{ch$b}runtime:         { .pa_num_text(rt, digits = 3) } sec",
      .envir = environment()
    )

    # ---- SELECTION
    cli::cli_text("{ch$l}{ch$b}{.h selection}", .envir = environment())

    cli::cli_text(
      " {ch$v}{ch$j}{ch$b}planning units:  { .pa_n_of_total_text(n_pu_sel, n_pu_total) } selected",
      .envir = environment()
    )

    cli::cli_text(
      " {ch$v}{ch$j}{ch$b}actions:         { .pa_n_of_total_text(n_act_sel, n_act_total) } selected",
      .envir = environment()
    )

    if (!is.na(n_tgt_total)) {
      cli::cli_text(
        " {ch$v}{ch$l}{ch$b}targets met:     { .pa_n_of_total_text(n_tgt_met, n_tgt_total) }",
        .envir = environment()
      )
    } else {
      cli::cli_text(
        " {ch$v}{ch$l}{ch$b}targets met:     {.muted none}",
        .envir = environment()
      )
    }

    # ---- OBJECTIVE VALUES
    if (!is.null(alias_txt)) {
      cli::cli_text("{ch$l}{ch$b}{.h objective values}", .envir = environment())
      cli::cli_text(
        " {ch$v}{ch$l}{ch$b}{alias_txt}",
        .envir = environment()
      )
    }

    # ---- SOLVER
    cli::cli_text("{ch$l}{ch$b}{.h solver}", .envir = environment())

    solver_txt <- as.character(dg$solver %||% dg$solver_name %||% "unknown")[1]
    cores_txt  <- as.character(dg$cores %||% NA)[1]
    tl_txt     <- as.character(dg$timelimit %||% dg$time_limit %||% NA)[1]

    cli::cli_text(
      " {ch$v}{ch$j}{ch$b}name:            {.code {solver_txt}}",
      .envir = environment()
    )
    cli::cli_text(
      " {ch$v}{ch$j}{ch$b}cores:           {cores_txt}",
      .envir = environment()
    )
    cli::cli_text(
      " {ch$v}{ch$l}{ch$b}time limit:      {tl_txt}",
      .envir = environment()
    )

    info_sym <- cli::symbol$info
    if (is.function(info_sym)) info_sym <- info_sym()
    cli::cli_text(
      cli::col_grey(
        paste0("# ", info_sym, " Use {.code x$summary} to inspect user-facing solution summaries.")
      )
    )

    cli::cli_end(div_id)
    invisible(TRUE)
  },

  show = function(self) self$print(),

  repr = function(self) {
    st <- tryCatch(getStatus(self), error = function(e) "unknown")
    obj <- self$solution$objective %||% NA_real_
    paste0("<Solution> status=", st, ", objective=", .pa_num_text(obj))
  }
)

#' @include internal.R
#'
#' @title Solve a planning problem
#'
#' @description
#' Solve a planning problem stored in a \code{Problem} object.
#'
#' This is the main execution step of the \pkg{multiscape} workflow. It reads
#' the problem specification stored in a \code{Problem} object, builds the
#' corresponding optimization model when needed, applies the configured solver
#' settings, and returns either a \code{\link{solution-class}} or a
#' \code{\link{solutionset-class}} depending on whether the workflow is
#' single-objective or multi-objective.
#'
#' @details
#' \strong{Role of \code{solve()}}
#'
#' The typical \pkg{multiscape} workflow is:
#' \preformatted{
#' x <- create_problem(...)
#' x <- add_...(x, ...)
#' x <- set_...(x, ...)
#' res <- solve(x)
#' }
#'
#' Thus, \code{solve()} is the stage at which the stored problem specification
#' is turned into one or more optimization runs.
#'
#' For most users, \code{solve()} is the standard execution entry point.
#' Explicit compilation with \code{compile_model()} is optional and is mainly
#' useful for advanced inspection or debugging workflows.
#'
#' \strong{What \code{solve()} uses from the problem object}
#'
#' The function uses the information already stored in the \code{Problem}
#' object, including:
#' \itemize{
#'   \item baseline planning data,
#'   \item actions, effects, profit, and spatial relations,
#'   \item targets and constraints,
#'   \item registered objectives,
#'   \item an optional multi-objective method configuration,
#'   \item solver settings.
#' }
#'
#' If a model has not yet been built, it is built internally during the solve
#' process. If a model snapshot or pointer already exists, the solving layer may
#' reuse or refresh it depending on the internal model state.
#'
#' \strong{Single-objective vs multi-objective behaviour}
#'
#' The behaviour of \code{solve()} depends on the problem configuration.
#'
#' \itemize{
#'   \item \strong{Single-objective case.} If exactly one objective is active
#'   and no multi-objective method is configured, \code{solve()} runs a single
#'   optimization problem and returns a \code{\link{solution-class}} object.
#'
#'   \item \strong{Multi-objective case.} If a multi-objective method is
#'   configured, \code{solve()} dispatches internally according to the stored
#'   method name and returns a \code{\link{solutionset-class}} object.
#' }
#'
#' Currently supported multi-objective method names are:
#' \itemize{
#'   \item \code{"weighted"},
#'   \item \code{"epsilon_constraint"},
#'   \item \code{"augmecon"}.
#' }
#'
#' \strong{Consistency rule}
#'
#' If multiple objectives are registered but no multi-objective method has been
#' selected, \code{solve()} stops with an error. In practical terms:
#'
#' \itemize{
#'   \item one objective and no multi-objective method
#'   \eqn{\Rightarrow} single-objective solve,
#'   \item multiple objectives and a valid multi-objective method
#'   \eqn{\Rightarrow} multi-objective solve,
#'   \item multiple objectives and no multi-objective method
#'   \eqn{\Rightarrow} error.
#' }
#'
#' \strong{Solver settings}
#'
#' Solver configuration is read from the \code{Problem} object, typically after
#' calling \code{\link{set_solver}} or one of its convenience wrappers such as
#' \code{\link{set_solver_gurobi}} or \code{\link{set_solver_cbc}}.
#'
#' These settings may include:
#' \itemize{
#'   \item the selected backend,
#'   \item time limits,
#'   \item optimality-gap settings,
#'   \item CPU cores,
#'   \item verbosity options,
#'   \item backend-specific solver parameters.
#' }
#'
#'
#' \strong{Method dispatch}
#'
#' \code{solve()} is an S3 generic. The public method documented here is
#' \code{solve.Problem()}, which operates on \code{Problem} objects.
#'
#' @param x A \code{Problem} object created with \code{\link{create_problem}} and
#'   optionally enriched with actions, effects, targets, constraints,
#'   objectives, spatial relations, method settings, and solver settings.
#' @param ... Additional arguments reserved for internal or legacy solver
#'   handling. These are not part of the main recommended user interface.
#'
#' @return
#' Either:
#' \itemize{
#'   \item a \code{\link{solution-class}} object when solving a single-objective
#'   problem, or
#'   \item a \code{\link{solutionset-class}} object when solving a configured
#'   multi-objective problem.
#' }
#'
#' A \code{Solution} represents one optimization run. A \code{SolutionSet}
#' represents multiple runs together with their run table, design information,
#' and individual \code{Solution} objects.
#'
#' After solving, users will typically inspect or visualize the returned results
#' through methods associated with \code{Solution} and
#' \code{SolutionSet} objects.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Minimal single-objective example
#' # ------------------------------------------------------------
#' pu <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' features <- data.frame(
#'   id = 1:2,
#'   name = c("sp1", "sp2")
#' )
#'
#' dist_features <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' actions <- data.frame(
#'   id = c("conservation", "restoration")
#' )
#'
#' effects <- data.frame(
#'   action = rep(c("conservation", "restoration"), each = 2),
#'   feature = rep(features$id, times = 2),
#'   multiplier = c(0.10, 0.10, 0.50, 0.50)
#' )
#'
#' x <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' ) |>
#'   add_actions(
#'     actions = actions,
#'     cost = c(conservation = 1, restoration = 2)
#'   ) |>
#'   add_effects(
#'     effects = effects,
#'     effect_type = "after"
#'   ) |>
#'   add_constraint_targets_relative(0.05) |>
#'   add_objective_min_cost(alias = "cost")
#'
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   x <- set_solver_cbc(x, verbose = FALSE)
#'   sol <- solve(x)
#'   print(sol)
#' }
#'
#' # ------------------------------------------------------------
#' # Minimal multi-objective example
#' # ------------------------------------------------------------
#' x_mo <- create_problem(
#'   pu = pu,
#'   features = features,
#'   dist_features = dist_features
#' ) |>
#'   add_actions(
#'     actions = actions,
#'     cost = c(conservation = 1, restoration = 2)
#'   ) |>
#'   add_effects(
#'     effects = effects,
#'     effect_type = "after"
#'   ) |>
#'   add_constraint_targets_relative(0.05) |>
#'   add_objective_min_cost(alias = "cost") |>
#'   add_objective_max_benefit(alias = "benefit") |>
#'   set_method_weighted_sum(
#'     aliases = c("cost", "benefit"),
#'     weights = c(0.5, 0.5),
#'     normalize_weights = TRUE
#'   )
#'
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   x_mo <- set_solver_cbc(x_mo, verbose = FALSE)
#'   solset <- solve(x_mo)
#'   print(solset)
#' }
#'
#' @seealso
#' \code{\link{problem-class}},
#' \code{\link{solution-class}},
#' \code{\link{solutionset-class}},
#' \code{\link{compile_model}},
#' \code{\link{set_solver}},
#' \code{\link{set_solver_cbc}},
#' \code{\link{set_solver_gurobi}},
#' \code{\link{set_method_weighted_sum}},
#' \code{\link{set_method_epsilon_constraint}},
#' \code{\link{set_method_augmecon}}
#'
#' @export
solve <- function(x, ...) {
  UseMethod("solve")
}

#' @rdname solve
#' @export
solve.Problem <- function(x, ...) {

  assertthat::assert_that(inherits(x, "Problem"))

  objs <- x$data$objectives %||% list()
  n_obj <- if (is.list(objs)) length(objs) else 0L

  method <- x$data$method %||% NULL
  has_method <- is.list(method) && length(method) > 0L

  if (has_method) {
    method_name <- as.character(method$type %||% method$name %||% NA_character_)[1]

    if (is.na(method_name) || !nzchar(method_name)) {
      stop("Invalid multi-objective method configuration: missing method name.", call. = FALSE)
    }

    .pamo_validate_objectives(x)
    #x <- compile_model(x)

    res <- switch(
      method_name,
      weighted = .pamo_solve_weighted(x, ...),
      epsilon_constraint = .pamo_solve_epsilon_constraint(x, ...),
      augmecon = .pamo_solve_augmecon(x, ...),
      stop("Unknown/unsupported multi-objective method: '", method_name, "'.", call. = FALSE)
    )

    if (!inherits(res, "SolutionSet")) {
      stop(
        "Internal error: multi-objective solve did not return a SolutionSet object.\n",
        "Returned class: ", paste(class(res), collapse = ", "),
        call. = FALSE
      )
    }

    return(res)
  }

  if (n_obj > 1L) {
    stop(
      "Multiple objectives are registered but no multi-objective method was selected.\n",
      "Use set_method_weighted_sum(), set_method_epsilon_constraint(), etc.",
      call. = FALSE
    )
  }

  x <- compile_model(x)
  res <- .pa_solve_single_problem(x, ...)

  if (!inherits(res, "Solution")) {
    stop(
      "Internal error: single-objective solve did not return a Solution object.\n",
      "Returned class: ", paste(class(res), collapse = ", "),
      call. = FALSE
    )
  }

  res
}

#' @include internalMO.R
#'
#' @title Get planning-unit results from a solution
#'
#' @description
#' Extract the planning-unit summary table from a \code{Solution} or
#' \code{SolutionSet} object returned by \code{\link{solve}}.
#'
#' The returned table summarizes solution values at the planning-unit level and
#' typically includes a \code{selected} indicator showing whether each planning
#' unit is selected in the solution.
#'
#' @details
#' This function reads the planning-unit summary stored in
#' \code{x$summary$pu}. It does not reconstruct the table from the raw decision
#' vector; it simply returns the stored summary after optional filtering.
#'
#' Let \eqn{w_i} denote the planning-unit selection variable for planning unit
#' \eqn{i}. In standard \code{multiscape} workflows, the \code{selected} column is the
#' user-facing representation of that planning-unit decision, typically coded as
#' \code{0} or \code{1}.
#'
#' If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
#' belonging to that run are returned. This requires the summary table to
#' contain a \code{run_id} column.
#'
#' If \code{only_selected = TRUE}, only rows with \code{selected == 1} are
#' returned. This requires the summary table to contain a \code{selected}
#' column.
#'
#' This function is intended for user-facing inspection of planning-unit results.
#' For the raw model variable vector, use \code{\link{get_solution_vector}}.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object returned by
#'   \code{\link{solve}}.
#' @param only_selected Logical. If \code{TRUE}, return only rows where
#'   \code{selected == 1}. Default is \code{FALSE}.
#' @param run Optional positive integer giving the run index to extract from a
#'   \code{SolutionSet}. If \code{NULL}, all runs are returned when available.
#'
#' @return A \code{data.frame} containing the stored planning-unit summary.
#'   Typical columns include planning-unit identifiers, optional labels, and a
#'   \code{selected} indicator.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   pu_tbl <- data.frame(
#'     id = 1:4,
#'     cost = c(1, 2, 3, 4)
#'   )
#'
#'   feat_tbl <- data.frame(
#'     id = 1:2,
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_feat_tbl <- data.frame(
#'     pu = c(1, 1, 2, 3, 4),
#'     feature = c(1, 2, 2, 1, 2),
#'     amount = c(5, 2, 3, 4, 1)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   effects_df <- data.frame(
#'     pu = c(1, 2, 3, 4),
#'     action = "conservation",
#'     feature = c(1, 1, 2, 2),
#'     benefit = c(2, 1, 1, 2),
#'     loss = c(0, 0, 0, 0)
#'   )
#'
#'   p <- create_problem(
#'     pu = pu_tbl,
#'     features = feat_tbl,
#'     dist_features = dist_feat_tbl,
#'     cost = "cost"
#'   ) |>
#'     add_actions(actions_df, cost = 0) |>
#'     add_effects(effects_df) |>
#'     add_constraint_targets_relative(0.2) |>
#'     add_objective_min_cost() |>
#'     set_solver_cbc(time_limit = 10)
#'
#'   sol <- solve(p)
#'
#'   get_pu(sol)
#'   get_pu(sol, only_selected = TRUE)
#' }
#' }
#' @seealso
#' \code{\link{get_actions}},
#' \code{\link{get_features}},
#' \code{\link{get_targets}},
#' \code{\link{get_solution_vector}}
#'
#' @export
get_pu <- function(x, only_selected = FALSE, run = NULL) {

  if (!inherits(x, c("Solution", "SolutionSet"))) {
    stop("x must be a Solution or SolutionSet.", call. = FALSE)
  }

  pu <- x$summary$pu %||% NULL
  if (is.null(pu)) {
    stop("No PU summary found (x$summary$pu is NULL).", call. = FALSE)
  }

  if (inherits(x, "SolutionSet") && !is.null(run)) {
    run <- as.integer(run)[1]
    if (!is.finite(run) || is.na(run) || run < 1L) {
      stop("run must be a positive integer (1-based).", call. = FALSE)
    }
    if (!("run_id" %in% names(pu))) {
      stop("PU summary has no 'run_id' column.", call. = FALSE)
    }
    pu <- pu[pu$run_id == run, , drop = FALSE]
  }

  if (isTRUE(only_selected)) {
    if (!("selected" %in% names(pu))) {
      stop("PU summary has no 'selected' column.", call. = FALSE)
    }
    pu <- pu[pu$selected == 1L, , drop = FALSE]
  }

  pu
}

#' @title Get action results from a solution
#'
#' @description
#' Extract the action-allocation summary table from a \code{Solution} or
#' \code{SolutionSet} object returned by \code{\link{solve}}.
#'
#' The returned table summarizes solution values at the
#' planning unit--action level and typically includes a \code{selected}
#' indicator showing whether each feasible \code{(pu, action)} pair is selected
#' in the solution.
#'
#' @details
#' This function reads the action summary stored in \code{x$summary$actions}. It
#' does not reconstruct the table from the raw decision vector; it simply
#' returns the stored summary after optional filtering.
#'
#' Let \eqn{x_{ia}} denote the decision variable associated with selecting
#' action \eqn{a} in planning unit \eqn{i}. In standard \code{multiscape} workflows,
#' the \code{selected} column is the user-facing representation of that
#' decision, typically coded as \code{0} or \code{1}.
#'
#' If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
#' belonging to that run are returned. This requires the summary table to
#' contain a \code{run_id} column.
#'
#' If \code{only_selected = TRUE}, only rows with \code{selected == 1} are
#' returned. This requires the summary table to contain a \code{selected}
#' column.
#'
#' This function is intended for user-facing inspection of action allocations.
#' For the raw model variable vector, use \code{\link{get_solution_vector}}.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object returned by
#'   \code{\link{solve}}.
#' @param only_selected Logical. If \code{TRUE}, return only rows where
#'   \code{selected == 1}. Default is \code{FALSE}.
#' @param run Optional positive integer giving the run index to extract from a
#'   \code{SolutionSet}. If \code{NULL}, all runs are returned when available.
#'
#' @return A \code{data.frame} containing the stored action-allocation summary.
#'   Typical columns include planning-unit ids, action ids, optional labels, and
#'   a \code{selected} indicator.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   pu_tbl <- data.frame(
#'     id = 1:4,
#'     cost = c(1, 2, 3, 4)
#'   )
#'
#'   feat_tbl <- data.frame(
#'     id = 1:2,
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_feat_tbl <- data.frame(
#'     pu = c(1, 1, 2, 3, 4),
#'     feature = c(1, 2, 2, 1, 2),
#'     amount = c(5, 2, 3, 4, 1)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   effects_df <- data.frame(
#'     pu = c(1, 2, 3, 4),
#'     action = "conservation",
#'     feature = c(1, 1, 2, 2),
#'     benefit = c(2, 1, 1, 2),
#'     loss = c(0, 0, 0, 0)
#'   )
#'
#'   p <- create_problem(
#'     pu = pu_tbl,
#'     features = feat_tbl,
#'     dist_features = dist_feat_tbl,
#'     cost = "cost"
#'   ) |>
#'     add_actions(actions_df, cost = 0) |>
#'     add_effects(effects_df) |>
#'     add_constraint_targets_relative(0.2) |>
#'     add_objective_min_cost() |>
#'     set_solver_cbc(time_limit = 10)
#'
#'   sol <- solve(p)
#'
#'   get_actions(sol)
#'   get_actions(sol, only_selected = TRUE)
#' }
#' }
#'
#' @seealso
#' \code{\link{get_pu}},
#' \code{\link{get_features}},
#' \code{\link{get_targets}},
#' \code{\link{get_solution_vector}}
#'
#' @export
get_actions <- function(x, only_selected = FALSE, run = NULL) {

  if (!inherits(x, c("Solution", "SolutionSet"))) {
    stop("x must be a Solution or SolutionSet.", call. = FALSE)
  }

  a <- x$summary$actions %||% NULL
  if (is.null(a)) {
    stop("No actions summary found (x$summary$actions is NULL).", call. = FALSE)
  }

  if (inherits(x, "SolutionSet") && !is.null(run)) {
    run <- as.integer(run)[1]
    if (!is.finite(run) || is.na(run) || run < 1L) {
      stop("run must be a positive integer (1-based).", call. = FALSE)
    }
    if (!("run_id" %in% names(a))) {
      stop("Actions summary has no 'run_id' column.", call. = FALSE)
    }
    a <- a[a$run_id == run, , drop = FALSE]
  }

  if (isTRUE(only_selected)) {
    if (!("selected" %in% names(a))) {
      stop("Actions summary has no 'selected' column.", call. = FALSE)
    }
    a <- a[a$selected == 1L, , drop = FALSE]
  }

  a <- a[, setdiff(names(a), c("internal_pu", "internal_action", "internal_row")), drop = FALSE]

  a
}



#' @title Get feature summary from a solution
#'
#' @description
#' Extract the per-feature summary table from a \code{Solution} or
#' \code{SolutionSet} object returned by \code{\link{solve}}.
#'
#' The returned table summarizes, for each feature, the total amount available
#' in the landscape together with the positive and negative contributions induced
#' by the selected actions in the solution.
#'
#' @details
#' This function reads the feature summary stored in \code{x$summary$features}.
#' It errors if that table is missing.
#'
#' Let \eqn{B_f} denote the total baseline amount available in the landscape for
#' feature \eqn{f}. Let \eqn{G_f} denote the total positive contribution induced
#' by selected actions, and let \eqn{L_f} denote the total negative
#' contribution. Then the returned table is intended to summarize quantities of
#' the form:
#'
#' \deqn{
#' \mathrm{net}_f = G_f - L_f,
#' }
#'
#' \deqn{
#' \mathrm{total}_f = B_f + \mathrm{net}_f.
#' }
#'
#' In the stored summary, these quantities are typically represented by the
#' columns:
#' \itemize{
#'   \item \code{total_available}, corresponding to \eqn{B_f},
#'   \item \code{benefit}, corresponding to \eqn{G_f},
#'   \item \code{loss}, corresponding to \eqn{L_f},
#'   \item \code{net}, corresponding to \eqn{G_f - L_f},
#'   \item \code{total}, corresponding to \eqn{B_f + G_f - L_f}.
#' }
#'
#' If any of the core numeric columns \code{total_available}, \code{benefit}, or
#' \code{loss} are missing, they are created and filled with zero. If
#' \code{net} is missing, it is computed as \code{benefit - loss}. If
#' \code{total} is missing, it is computed as \code{total_available + net}.
#'
#' Thus, this function guarantees that the returned table contains the columns
#' \code{total_available}, \code{benefit}, \code{loss}, \code{net}, and
#' \code{total}, even if some of them were absent from the stored summary.
#'
#' If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
#' belonging to that run are returned. If the result contains a \code{run_id}
#' column but only a single run is present and \code{run} was not requested
#' explicitly, the \code{run_id} column is removed for convenience.
#'
#' This function summarizes feature outcomes in the solution. It is different
#' from \code{\link{get_targets}}, which focuses on target achievement rather
#' than total feature balance.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object returned by
#'   \code{\link{solve}}.
#' @param run Optional positive integer giving the run index to extract from a
#'   \code{SolutionSet}. If \code{NULL}, all runs are returned when available.
#'
#' @return A \code{data.frame} with one row per feature, or one row per
#'   feature--run combination when multiple runs are present. The returned table
#'   always includes the columns \code{total_available}, \code{benefit},
#'   \code{loss}, \code{net}, and \code{total}.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   pu_tbl <- data.frame(
#'     id = 1:4,
#'     cost = c(1, 2, 3, 4)
#'   )
#'
#'   feat_tbl <- data.frame(
#'     id = 1:2,
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_feat_tbl <- data.frame(
#'     pu = c(1, 1, 2, 3, 4),
#'     feature = c(1, 2, 2, 1, 2),
#'     amount = c(5, 2, 3, 4, 1)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   effects_df <- data.frame(
#'     pu = c(1, 2, 3, 4),
#'     action = "conservation",
#'     feature = c(1, 1, 2, 2),
#'     benefit = c(2, 1, 1, 2),
#'     loss = c(0, 0, 0, 0)
#'   )
#'
#'   p <- create_problem(
#'     pu = pu_tbl,
#'     features = feat_tbl,
#'     dist_features = dist_feat_tbl,
#'     cost = "cost"
#'   ) |>
#'     add_actions(actions_df, cost = 0) |>
#'     add_effects(effects_df) |>
#'     add_constraint_targets_relative(0.2) |>
#'     add_objective_min_cost() |>
#'     set_solver_cbc(time_limit = 10)
#'
#'   sol <- solve(p)
#'
#'   get_features(sol)
#' }
#' }
#'
#' @seealso
#' \code{\link{get_pu}},
#' \code{\link{get_actions}},
#' \code{\link{get_targets}}
#'
#' @export
get_features <- function(x, run = NULL) {

  if (!inherits(x, c("Solution", "SolutionSet"))) {
    stop("x must be a Solution or SolutionSet.", call. = FALSE)
  }

  f <- x$summary$features %||% NULL
  if (is.null(f)) {
    stop("No features summary found (x$summary$features is NULL).", call. = FALSE)
  }

  if (inherits(x, "SolutionSet") && !is.null(run)) {
    run <- as.integer(run)[1]
    if (!is.finite(run) || is.na(run) || run < 1L) {
      stop("run must be a positive integer (1-based).", call. = FALSE)
    }
    if (!("run_id" %in% names(f))) {
      stop("Features summary has no 'run_id' column.", call. = FALSE)
    }

    runs_avail <- sort(unique(f$run_id))
    if (!(run %in% runs_avail)) {
      stop(
        "run=", run, " is out of range. Available runs: ",
        paste(runs_avail, collapse = ", "),
        ".",
        call. = FALSE
      )
    }

    f <- f[f$run_id == run, , drop = FALSE]
  }

  out <- f

  # ensure core numeric columns exist
  if (!("total_available" %in% names(out))) out$total_available <- 0
  if (!("benefit" %in% names(out))) out$benefit <- 0
  if (!("loss" %in% names(out))) out$loss <- 0

  out$total_available <- as.numeric(out$total_available)
  out$benefit <- as.numeric(out$benefit)
  out$loss <- as.numeric(out$loss)

  if (!("net" %in% names(out))) {
    out$net <- out$benefit - out$loss
  } else {
    out$net <- as.numeric(out$net)
  }

  if (!("total" %in% names(out))) {
    out$total <- out$total_available + out$net
  } else {
    out$total <- as.numeric(out$total)
  }

  keep_first <- c(
    "run_id",
    "feature",
    "feature_name",
    "total_available",
    "benefit",
    "loss",
    "net",
    "total"
  )

  keep_first <- intersect(keep_first, names(out))
  keep_rest <- setdiff(names(out), keep_first)
  out <- out[, c(keep_first, keep_rest), drop = FALSE]

  if ("run_id" %in% names(out) && length(unique(out$run_id)) <= 1L && is.null(run)) {
    out$run_id <- NULL
  }

  out
}




#' @title Get target achievement summary from a solution
#'
#' @description
#' Extract a user-facing target-achievement table from a \code{Solution} or
#' \code{SolutionSet} object returned by \code{\link{solve}}.
#'
#' The returned table summarizes, for each stored target, the target level, the
#' achieved value in the solution, the gap between achieved and required values,
#' and whether the target was met.
#'
#' @details
#' Targets are optional in \code{multiscape}. If the solution object does not contain
#' a targets summary table at \code{x$summary$targets}, this function returns
#' \code{NULL} without error.
#'
#' This function reads the stored targets summary and returns a simplified
#' user-facing table. If the summary contains \code{achieved} and
#' \code{target_value}, target satisfaction is evaluated as follows.
#'
#' For lower-bound targets:
#' \deqn{
#' \mathrm{met} = (\mathrm{achieved} \ge \mathrm{target}),
#' }
#'
#' and for upper-bound targets:
#' \deqn{
#' \mathrm{met} = (\mathrm{achieved} \le \mathrm{target}).
#' }
#'
#' The interpretation of the target direction is taken from the \code{sense}
#' column when available:
#' \itemize{
#'   \item \code{"ge"}, \code{">="}, or \code{"min"} are treated as lower-bound
#'   targets,
#'   \item \code{"le"}, \code{"<="}, or \code{"max"} are treated as upper-bound
#'   targets,
#'   \item if \code{sense} is missing, the target is treated as a lower bound by
#'   default.
#' }
#'
#' The returned table is simplified and renames some internal fields for
#' readability:
#' \itemize{
#'   \item \code{target_raw} is returned as \code{target_level},
#'   \item \code{basis_total} is returned as \code{total_available},
#'   \item \code{target_value} is returned as \code{target}.
#' }
#'
#' If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
#' belonging to that run are returned. If the result contains a \code{run_id}
#' column but only a single run is present and \code{run} was not requested
#' explicitly, the \code{run_id} column is removed for convenience.
#'
#' The \code{gap} column is expected to be part of the stored summary. When
#' present, it typically represents:
#' \deqn{
#' \mathrm{gap} = \mathrm{achieved} - \mathrm{target}.
#' }
#'
#' @param x A \code{Solution} or \code{SolutionSet} object returned by
#'   \code{\link{solve}}.
#' @param run Optional positive integer giving the run index to extract from a
#'   \code{SolutionSet}. If \code{NULL}, all runs are returned when available.
#'
#' @return A simplified \code{data.frame} target summary, or \code{NULL} if the
#'   solution does not contain targets. Typical columns include
#'   \code{feature}, \code{feature_name}, \code{target_level},
#'   \code{total_available}, \code{target}, \code{achieved}, \code{gap}, and
#'   \code{met}.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   pu_tbl <- data.frame(
#'     id = 1:4,
#'     cost = c(1, 2, 3, 4)
#'   )
#'
#'   feat_tbl <- data.frame(
#'     id = 1:2,
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_feat_tbl <- data.frame(
#'     pu = c(1, 1, 2, 3, 4),
#'     feature = c(1, 2, 2, 1, 2),
#'     amount = c(5, 2, 3, 4, 1)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   effects_df <- data.frame(
#'     pu = c(1, 2, 3, 4),
#'     action = "conservation",
#'     feature = c(1, 1, 2, 2),
#'     benefit = c(2, 1, 1, 2),
#'     loss = c(0, 0, 0, 0)
#'   )
#'
#'   p <- create_problem(
#'     pu = pu_tbl,
#'     features = feat_tbl,
#'     dist_features = dist_feat_tbl,
#'     cost = "cost"
#'   ) |>
#'     add_actions(actions_df, cost = 0) |>
#'     add_effects(effects_df) |>
#'     add_constraint_targets_relative(0.2) |>
#'     add_objective_min_cost() |>
#'     set_solver_cbc(time_limit = 10)
#'
#'   sol <- solve(p)
#'
#'   get_targets(sol)
#' }
#' }
#'
#' @seealso
#' \code{\link{get_pu}},
#' \code{\link{get_actions}},
#' \code{\link{get_features}}
#'
#' @export
get_targets <- function(x, run = NULL) {

  if (!inherits(x, c("Solution", "SolutionSet"))) {
    stop("x must be a Solution or SolutionSet.", call. = FALSE)
  }

  t <- x$summary$targets %||% NULL
  if (is.null(t)) {
    return(NULL)
  }

  if (inherits(x, "SolutionSet") && !is.null(run)) {
    run <- as.integer(run)[1]
    if (!is.finite(run) || is.na(run) || run < 1L) {
      stop("run must be a positive integer (1-based).", call. = FALSE)
    }
    if (!("run_id" %in% names(t))) {
      stop("Targets summary has no 'run_id' column.", call. = FALSE)
    }
    t <- t[t$run_id == run, , drop = FALSE]
  }

  out <- t

  if (all(c("achieved", "target_value") %in% names(out))) {
    if ("sense" %in% names(out)) {
      out$met <- ifelse(
        out$sense %in% c("ge", ">=", "min"),
        out$achieved >= out$target_value,
        ifelse(
          out$sense %in% c("le", "<=", "max"),
          out$achieved <= out$target_value,
          NA
        )
      )
    } else {
      out$met <- out$achieved >= out$target_value
    }
  }

  keep <- c(
    "run_id",
    "feature",
    "feature_name",
    "target_raw",
    "basis_total",
    "target_value",
    "achieved",
    "gap",
    "met"
  )
  keep <- intersect(keep, names(out))
  out <- out[, keep, drop = FALSE]

  names(out)[names(out) == "target_raw"] <- "target_level"
  names(out)[names(out) == "basis_total"] <- "total_available"
  names(out)[names(out) == "target_value"] <- "target"

  if ("run_id" %in% names(out) && length(unique(out$run_id)) <= 1L && is.null(run)) {
    out$run_id <- NULL
  }

  out
}

#' @title Get raw solution vector from a solution
#'
#' @description
#' Return the raw decision-variable vector produced by the solver, in the
#' internal model-variable order used by the optimization backend.
#'
#' @details
#' This function extracts the raw solution vector stored at
#' \code{x$solution$vector} for a \code{Solution} or for a selected run of a
#' \code{SolutionSet}.
#'
#' The returned vector is in the internal variable order of the optimization
#' model. Depending on the problem formulation, it may include:
#' \itemize{
#'   \item planning-unit selection variables,
#'   \item action-allocation variables,
#'   \item auxiliary variables introduced for targets, budgets, fragmentation,
#'   or other constraints/objectives,
#'   \item and potentially additional blocks created internally by the model
#'   builder.
#' }
#'
#' Therefore, this vector is primarily intended for advanced users, debugging,
#' diagnostics, or internal verification. It is not a user-facing allocation
#' table.
#'
#' To inspect selected planning units or selected actions in a more interpretable
#' form, use \code{\link{get_pu}} or \code{\link{get_actions}} instead.
#'
#' For a single solution, the returned vector corresponds to that solution.
#' For a \code{SolutionSet}, the \code{run} argument selects which run to
#' extract.
#'
#' @param x A \code{Solution} or \code{SolutionSet} object returned by
#'   \code{\link{solve}}.
#' @param run Positive integer giving the run index to extract when \code{x} is
#'   a \code{SolutionSet}. Default is \code{1L}.
#'
#' @return A numeric vector with one value per internal model variable.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("rcbc", quietly = TRUE)) {
#'   pu_tbl <- data.frame(
#'     id = 1:4,
#'     cost = c(1, 2, 3, 4)
#'   )
#'
#'   feat_tbl <- data.frame(
#'     id = 1:2,
#'     name = c("feature_1", "feature_2")
#'   )
#'
#'   dist_feat_tbl <- data.frame(
#'     pu = c(1, 1, 2, 3, 4),
#'     feature = c(1, 2, 2, 1, 2),
#'     amount = c(5, 2, 3, 4, 1)
#'   )
#'
#'   actions_df <- data.frame(
#'     id = "conservation",
#'     name = "conservation"
#'   )
#'
#'   effects_df <- data.frame(
#'     pu = c(1, 2, 3, 4),
#'     action = "conservation",
#'     feature = c(1, 1, 2, 2),
#'     benefit = c(2, 1, 1, 2),
#'     loss = c(0, 0, 0, 0)
#'   )
#'
#'   p <- create_problem(
#'     pu = pu_tbl,
#'     features = feat_tbl,
#'     dist_features = dist_feat_tbl,
#'     cost = "cost"
#'   ) |>
#'     add_actions(actions_df, cost = 0) |>
#'     add_effects(effects_df) |>
#'     add_constraint_targets_relative(0.2) |>
#'     add_objective_min_cost() |>
#'     set_solver_cbc(time_limit = 10)
#'
#'   sol <- solve(p)
#'
#'   v <- get_solution_vector(sol)
#'   v
#'   length(v)
#' }
#' }
#'
#' @seealso
#' \code{\link{get_pu}},
#' \code{\link{get_actions}},
#' \code{\link{get_features}},
#' \code{\link{get_targets}}
#'
#' @export
get_solution_vector <- function(x, run = 1L) {
  sol <- .mo_get_solution_from(x, run = run)

  v <- sol$solution$vector %||% NULL
  if (is.null(v)) {
    stop("No raw solution vector found (x$solution$vector is NULL).", call. = FALSE)
  }

  as.numeric(v)
}

#' @title Configure solver settings
#'
#' @description
#' Store solver configuration inside a \code{Problem} object so that
#' \code{\link{solve}} can later run using the stored backend and runtime
#' options.
#'
#' This function does not build or solve the optimization model. It only updates
#' the solver configuration stored in \code{x$data$solve_args}.
#'
#' @details
#' \strong{Purpose}
#'
#' The \code{multiscape} workflow separates problem specification from solver
#' configuration. Problem data, actions, effects, targets, objectives, and
#' methods are stored in the \code{Problem} object, and solver settings are
#' stored separately in \code{x$data$solve_args}.
#'
#' This function allows solver options to be configured once and reused later
#' through \code{\link{solve}(x)} without repeating the same arguments each time.
#'
#' \strong{Stored fields}
#'
#' The solver configuration is stored in \code{x$data$solve_args}. Typical
#' entries include:
#' \itemize{
#'   \item \code{solver},
#'   \item \code{gap_limit},
#'   \item \code{time_limit},
#'   \item \code{solution_limit},
#'   \item \code{cores},
#'   \item \code{verbose},
#'   \item \code{write_log},
#'   \item \code{log_file},
#'   \item \code{solver_params}.
#' }
#'
#' \strong{Incremental update semantics}
#'
#' This function updates solver settings incrementally.
#'
#' If an argument is supplied as \code{NULL}, the previously stored value is
#' kept unchanged. Therefore, repeated calls can be used to modify only selected
#' components of the solver configuration.
#'
#' For example, a user may first configure the solver backend and time limit,
#' and later update only the optimality gap or only a backend-specific
#' parameter.
#'
#' \strong{Gap limit}
#'
#' The argument \code{gap_limit} is interpreted as a relative optimality gap for
#' mixed-integer optimization. It must lie in \eqn{[0,1]}.
#'
#' If the solver stops with incumbent value \eqn{z^{\mathrm{inc}}} and best
#' bound \eqn{z^{\mathrm{bd}}}, then the exact stopping rule depends on the
#' solver backend, but conceptually \code{gap_limit} controls the maximum
#' accepted relative difference between the incumbent and the bound.
#'
#' \strong{Time limit}
#'
#' The argument \code{time_limit} is interpreted as a maximum wall-clock time in
#' seconds allowed for the solver.
#'
#' \strong{Solution limit}
#'
#' The argument \code{solution_limit} is stored as a logical flag. Its exact
#' meaning depends on the backend-specific solving layer, but conceptually it
#' requests early termination after finding a feasible solution according to the
#' behaviour supported by the chosen solver.
#'
#' \strong{Cores}
#'
#' The argument \code{cores} specifies the number of CPU cores to use. If the
#' requested number exceeds the number of detected cores, it is capped to the
#' detected maximum with a warning.
#'
#' \strong{Verbose output and log files}
#'
#' The arguments \code{verbose}, \code{write_log}, and \code{log_file}
#' control how solver logging is handled. These options are stored and later
#' interpreted by the solving layer for the selected backend.
#'
#' \strong{Solver-specific parameters}
#'
#' Additional backend-specific parameters can be passed in two ways:
#' \itemize{
#'   \item through the named list \code{solver_params},
#'   \item through additional named arguments in \code{...}.
#' }
#'
#' These two sources are merged, and the result is then merged with any
#' previously stored \code{solver_params}. Existing parameters are therefore
#' preserved unless explicitly overwritten.
#'
#' This is particularly useful for backend-specific controls such as node
#' selection, emphasis parameters, tolerances, or heuristics.
#'
#' \strong{Supported backends}
#'
#' The \code{solver} argument selects the backend to be used later by
#' \code{\link{solve}}. Supported values are:
#' \itemize{
#'   \item \code{"auto"}: let the solving layer choose an available backend,
#'   \item \code{"gurobi"},
#'   \item \code{"cplex"},
#'   \item \code{"cbc"},
#'   \item \code{"symphony"}.
#' }
#'
#' This function only stores the requested backend. Availability of the backend
#' is checked later when solving.
#'
#' @param x A \code{Problem} object.
#' @param solver Character string indicating the solver backend to use. Must be
#'   one of \code{"auto"}, \code{"gurobi"}, \code{"cplex"},
#'   \code{"cbc"}, or \code{"symphony"}.
#' @param gap_limit Optional numeric value in \eqn{[0,1]} giving the relative
#'   optimality gap for mixed-integer optimization. If \code{NULL}, the
#'   previously stored value is kept unchanged.
#' @param time_limit Optional non-negative numeric value giving the maximum
#'   solving time in seconds. If \code{NULL}, the previously stored value is
#'   kept unchanged.
#' @param solution_limit Optional logical flag controlling backend-specific
#'   early stopping after feasible solution discovery. If \code{NULL}, the
#'   previously stored value is kept unchanged.
#' @param cores Optional positive integer giving the number of CPU cores to use.
#'   If \code{NULL}, the previously stored value is kept unchanged.
#' @param verbose Optional logical flag indicating whether the solver should
#'   print log output. If \code{NULL}, the previously stored value is kept
#'   unchanged.
#' @param log_file Optional character string giving the name of the
#'   solver log file. If \code{NULL}, the previously stored value is kept
#'   unchanged.
#' @param write_log Optional logical flag indicating whether solver output
#'   should be written to a file. If \code{NULL}, the previously stored value is
#'   kept unchanged.
#' @param solver_params Named list of solver-specific parameters. These are
#'   merged with previously stored backend-specific parameters rather than
#'   replacing them completely.
#' @param ... Additional named solver-specific parameters. These are merged into
#'   \code{solver_params}. For example, \code{MIPFocus = 1} for Gurobi.
#'
#' @return An updated \code{Problem} object with modified solver settings stored
#'   in \code{x$data$solve_args}.
#'
#' @seealso
#' \code{\link{solve}},
#' \code{\link{set_solver_gurobi}},
#' \code{\link{set_solver_cplex}},
#' \code{\link{set_solver_cbc}},
#' \code{\link{set_solver_symphony}}
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' x1 <- set_solver(
#'   x,
#'   solver = "cbc",
#'   gap_limit = 0.01,
#'   time_limit = 300,
#'   cores = 2,
#'   verbose = TRUE
#' )
#'
#' x1$data$solve_args
#'
#' # Update only selected settings
#' x2 <- set_solver(
#'   x1,
#'   gap_limit = 0.05,
#'   solver_params = list(randomSeed = 123)
#' )
#'
#' x2$data$solve_args
#'
#' @export
set_solver <- function(x,
                       solver = c("auto", "gurobi", "cplex", "cbc", "symphony"),
                       gap_limit = NULL,
                       time_limit = NULL,
                       solution_limit = NULL,
                       cores = NULL,
                       verbose = FALSE,
                       log_file = NULL,
                       write_log = NULL,
                       solver_params = list(),
                       ...) {

  stopifnot(inherits(x, "Problem"))
  solver <- match.arg(solver)

  dots <- list(...)
  if (length(dots) > 0) {
    solver_params <- utils::modifyList(solver_params %||% list(), dots)
  }
  if (!is.list(solver_params)) stop("solver_params must be a list.", call. = FALSE)

  if (is.null(x$data$solve_args) || !is.list(x$data$solve_args)) x$data$solve_args <- list()

  # start from stored (so we can keep values when args are NULL)
  out <- x$data$solve_args

  # always set solver if explicitly provided
  out$solver <- solver

  # set numeric/logical args only if not NULL
  if (!is.null(gap_limit)) {
    assertthat::assert_that(assertthat::is.scalar(gap_limit), is.finite(gap_limit), gap_limit >= 0, gap_limit <= 1)
    out$gap_limit <- base::round(as.numeric(gap_limit), 3)
  }
  if (!is.null(time_limit)) {
    assertthat::assert_that(assertthat::is.scalar(time_limit), is.finite(time_limit), time_limit >= 0)
    out$time_limit <- base::round(as.numeric(time_limit), 3)
  }
  if (!is.null(solution_limit)) {
    assertthat::assert_that(assertthat::is.flag(solution_limit))
    out$solution_limit <- isTRUE(solution_limit)
  }
  if (!is.null(cores)) {
    assertthat::assert_that(assertthat::is.count(cores))
    cores <- as.integer(cores)
    max_cores <- parallel::detectCores(TRUE)
    if (is.finite(max_cores) && cores > max_cores) {
      warning("cores is larger than detected cores; capping to detected cores.", call. = FALSE, immediate. = TRUE)
      cores <- as.integer(max_cores)
    }
    out$cores <- cores
  }
  if (!is.null(verbose)) {
    assertthat::assert_that(assertthat::is.flag(verbose))
    out$verbose <- isTRUE(verbose)
  }
  if (!is.null(write_log)) {
    assertthat::assert_that(assertthat::is.flag(write_log))
    out$output_file <- isTRUE(write_log)
  }
  if (!is.null(log_file)) {
    assertthat::assert_that(assertthat::is.string(log_file))
    out$name_output_file <- as.character(log_file)[1]
  }

  # merge solver_params with stored solver_params (do not drop existing ones)
  out$solver_params <- utils::modifyList(out$solver_params %||% list(), solver_params)

  x$data$solve_args <- out
  x
}

#' @title Configure Gurobi solver settings
#'
#' @description
#' Convenience wrapper around \code{\link{set_solver}} that stores
#' \code{solver = "gurobi"} in the problem object.
#'
#' This function does not solve the model. It only updates the stored solver
#' configuration.
#'
#' @inheritParams set_solver
#'
#' @return An updated \code{Problem} object with Gurobi solver settings stored
#'   in \code{x$data$solve_args}.
#'
#' @seealso
#' \code{\link{set_solver}},
#' \code{\link{solve}}
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' x <- set_solver_gurobi(
#'   x,
#'   gap_limit = 0.01,
#'   time_limit = 600,
#'   cores = 2,
#'   MIPFocus = 1
#' )
#'
#' x$data$solve_args
#'
#' @export
set_solver_gurobi <- function(x, ..., solver_params = list(), gap_limit = NULL, time_limit = NULL,
                              solution_limit = NULL, cores = NULL, verbose = FALSE,
                              log_file = NULL, write_log = NULL) {
  set_solver(
    x,
    solver = "gurobi",
    gap_limit = gap_limit,
    time_limit = time_limit,
    solution_limit = solution_limit,
    cores = cores,
    verbose = verbose,
    log_file = log_file,
    write_log = write_log,
    solver_params = solver_params,
    ...
  )
}

#' @title Configure CPLEX solver settings
#'
#' @description
#' Convenience wrapper around \code{\link{set_solver}} that stores
#' \code{solver = "cplex"} in the problem object.
#'
#' This function does not solve the model. It only updates the stored solver
#' configuration.
#'
#' @inheritParams set_solver
#'
#' @return An updated \code{Problem} object with CPLEX solver settings stored in
#'   \code{x$data$solve_args}.
#'
#' @seealso
#' \code{\link{set_solver}},
#' \code{\link{solve}}
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' x <- set_solver_cbc(
#'   x,
#'   gap_limit = 0.01,
#'   time_limit = 300,
#'   cores = 2
#' )
#'
#' x$data$solve_args
#'
#' @export
set_solver_cbc <- function(x, ..., solver_params = list(), gap_limit = NULL, time_limit = NULL,
                           solution_limit = NULL, cores = NULL, verbose = FALSE,
                           log_file = NULL, write_log = NULL) {
  set_solver(
    x,
    solver = "cbc",
    gap_limit = gap_limit,
    time_limit = time_limit,
    solution_limit = solution_limit,
    cores = cores,
    verbose = verbose,
    log_file = log_file,
    write_log = write_log,
    solver_params = solver_params,
    ...
  )
}

#' Configure CPLEX solver settings
#'
#' @description
#' Convenience wrapper around \code{\link{set_solver}} that sets
#' \code{solver = "cplex"}.
#'
#' @inheritParams set_solver
#'
#' @return
#' An updated \code{Problem} object with CPLEX solver settings.
#'
#' @seealso
#' \code{\link{set_solver}},
#' \code{\link{solve}}
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' x <- set_solver_cplex(
#'   x,
#'   gap_limit = 0.001,
#'   time_limit = 1200,
#'   cores = 2
#' )
#'
#' x$data$solve_args
#'
#' @export
set_solver_cplex <- function(x, ..., solver_params = list(), gap_limit = NULL, time_limit = NULL,
                             solution_limit = NULL, cores = NULL, verbose = FALSE,
                             log_file = NULL, write_log = NULL) {
  set_solver(
    x,
    solver = "cplex",
    gap_limit = gap_limit,
    time_limit = time_limit,
    solution_limit = solution_limit,
    cores = cores,
    verbose = verbose,
    log_file = log_file,
    write_log = write_log,
    solver_params = solver_params,
    ...
  )
}

#' @title Configure SYMPHONY solver settings
#'
#' @description
#' Convenience wrapper around \code{\link{set_solver}} that stores
#' \code{solver = "symphony"} in the problem object.
#'
#' This function does not solve the model. It only updates the stored solver
#' configuration.
#'
#' @inheritParams set_solver
#'
#' @return An updated \code{Problem} object with SYMPHONY solver settings stored
#'   in \code{x$data$solve_args}.
#'
#' @seealso
#' \code{\link{set_solver}},
#' \code{\link{solve}}
#'
#' @examples
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' x <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl,
#'   cost = "cost"
#' )
#'
#' x <- set_solver_symphony(
#'   x,
#'   gap_limit = 0.05,
#'   time_limit = 300
#' )
#'
#' x$data$solve_args
#'
#' @export
set_solver_symphony <- function(x, ..., solver_params = list(), gap_limit = NULL, time_limit = NULL,
                                solution_limit = NULL, cores = NULL, verbose = FALSE,
                                log_file = NULL, write_log = NULL) {
  set_solver(
    x,
    solver = "symphony",
    gap_limit = gap_limit,
    time_limit = time_limit,
    solution_limit = solution_limit,
    cores = cores,
    verbose = verbose,
    log_file = log_file,
    write_log = write_log,
    solver_params = solver_params,
    ...
  )
}

#' Simulated planning units
#'
#' Example planning units as an `sf` object for package examples and tests.
#'
#' @format An object of class `sf`.
#' @usage data(sim_pu_sf)
"sim_pu_sf"

#' Simulated planning units
#'
#' Example planning units as an `sf` object for package examples and tests.
#'
#' @format An object of class `sf`.
#' @usage data(sim_pu)
"sim_pu"

#' Simulated features
#'
#' Example feature table for package examples and tests.
#'
#' @format A data frame with feature identifiers and names.
#' @usage data(sim_features)
"sim_features"

#' Simulated feature distribution
#'
#' Example distribution of feature amounts across planning units.
#'
#' @format A data frame linking planning units and features with an `amount` column.
#' @usage data(sim_dist_features)
"sim_dist_features"

#' @include internal.R

# -------------------------------------------------------------------------
# Utilities
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
# Promote Problem -> Problem (internal; never exposed to user)
# -------------------------------------------------------------------------

# .pamo_as_mo <- function(x) {
#   if (inherits(x, "Problem")) return(x)
#
#   if (inherits(x, "Problem")) {
#     obj <- pproto(NULL, Problem, base = x)
#
#     # fuerza clase S3 para que UseMethod("solve") encuentre solve.Problem
#     cls <- class(obj)
#     if (is.null(cls)) cls <- character()
#     class(obj) <- unique(c("Problem", cls))
#
#     return(obj)
#   }
#
#   stop("Expected a Problem or a Problem.", call. = FALSE)
# }


# -------------------------------------------------------------------------
# Atomic objective registry accessors (stored in multiscape::Problem)
# Registry: x$data$objectives[[alias]] = list(...)
# -------------------------------------------------------------------------

.pamo_get_specs <- function(x) {
  stopifnot(inherits(x, "Problem"))
  specs <- x$data$objectives %||% list()
  if (!is.list(specs)) specs <- list()
  specs
}

# internal: get ONE atomic objective spec by alias
.pamo_get_objective_spec <- function(x, alias) {
  stopifnot(inherits(x, "Problem"))
  alias <- as.character(alias)[1]

  if (is.na(alias) || !nzchar(alias)) {
    stop("alias must be a non-empty string.", call. = FALSE)
  }

  specs <- .pamo_get_specs(x)
  sp <- specs[[alias]]

  if (is.null(sp)) {
    stop(
      "Objective alias not found: '", alias, "'.\n",
      "Tip: register objectives via add_objective_* (alias='...') before calling set_method_*().",
      call. = FALSE
    )
  }

  # defensive minimal validation (clearer errors later)
  if (is.null(sp$objective_id)) stop("Objective '", alias, "' has no objective_id.", call. = FALSE)
  if (is.null(sp$sense)) stop("Objective '", alias, "' has no sense.", call. = FALSE)

  sp
}

# internal: get MANY objective specs in order (and validate duplicates)
.pamo_get_objective_specs <- function(x, aliases) {
  stopifnot(inherits(x, "Problem"))

  aliases <- as.character(aliases)
  if (length(aliases) == 0L) stop("aliases must have length > 0.", call. = FALSE)

  if (anyNA(aliases) || any(!nzchar(aliases))) {
    stop("aliases must be non-empty strings.", call. = FALSE)
  }

  if (anyDuplicated(aliases) != 0L) {
    dups <- unique(aliases[duplicated(aliases)])
    stop("Duplicated objective aliases: ", paste(dups, collapse = ", "), call. = FALSE)
  }

  specs <- lapply(aliases, function(a) .pamo_get_objective_spec(x, a))
  names(specs) <- aliases
  specs
}

# internal: validate registry sanity
.pamo_validate_objectives <- function(x) {
  stopifnot(inherits(x, "Problem"))

  specs <- .pamo_get_specs(x)
  if (length(specs) == 0) {
    stop("No objectives registered. Use add_objective_* (with alias=...) first.", call. = FALSE)
  }

  al <- names(specs)
  if (anyNA(al) || any(!nzchar(al))) stop("Objective registry has invalid aliases.", call. = FALSE)
  if (anyDuplicated(al)) stop("Duplicated objective aliases are not allowed.", call. = FALSE)

  invisible(TRUE)
}

# -------------------------------------------------------------------------
# Objective IR (intermediate representation)
# Each IR carries:
# - sense: "min"|"max"
# - terms: list of atomic terms (still indivisible objectives)
# - objective_id, objective_args: to be able to re-activate the objective in a single-objective model
# -------------------------------------------------------------------------
.pamo_objvec_action_boundary_cut <- function(base_superset, term) {
  stopifnot(inherits(base_superset, "Problem"))

  op <- base_superset$data$model_ptr
  if (is.null(op)) {
    stop("Superset model is not built (missing model_ptr).", call. = FALSE)
  }

  ml <- base_superset$data$model_list %||% NULL
  if (is.null(ml) || !is.list(ml)) {
    base_superset <- .pa_refresh_model_snapshot(base_superset)
    ml <- base_superset$data$model_list %||% NULL
  }
  if (is.null(ml) || !is.list(ml)) {
    stop("Model snapshot is missing in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }

  m0 <- .pa_model_from_ptr(
    op,
    args = base_superset$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  n_col <- as.integer(
    ml$ncol %||%
      ml$n_col %||%
      length(m0$obj)
  )

  n_x <- as.integer(ml$n_x %||% NA_integer_)
  x_offset <- as.integer(ml$x_offset %||% NA_integer_)
  y0 <- as.integer(ml$y_action_offset %||% NA_integer_)
  n_y <- as.integer(ml$n_y_action %||% NA_integer_)

  if (!is.finite(n_col) || is.na(n_col) || n_col <= 0L) {
    stop("Invalid model dimension (ncol) in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  if (!is.finite(n_x) || is.na(n_x) || n_x <= 0L) {
    stop("Invalid n_x in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  if (!is.finite(x_offset) || is.na(x_offset) || x_offset < 0L) {
    stop("Invalid x_offset in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  if (!is.finite(y0) || is.na(y0) || y0 < 0L) {
    stop("Invalid y_action_offset in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  if (!is.finite(n_y) || is.na(n_y) || n_y <= 0L) {
    stop("Invalid n_y_action in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }

  v <- numeric(n_col)

  # ---- relation
  rel_name <- as.character(term$relation_name %||% "boundary")[1]
  rel_model <- base_superset$data$spatial_relations_model[[rel_name]] %||%
    base_superset$data$spatial_relations[[rel_name]] %||%
    NULL

  if (is.null(rel_model) || !inherits(rel_model, "data.frame")) {
    stop("Missing relation '", rel_name, "' in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  for (nm in c("internal_pu1", "internal_pu2", "weight")) {
    if (!nm %in% names(rel_model)) {
      stop("Relation '", rel_name, "' must contain column '", nm, "'.", call. = FALSE)
    }
  }

  # ---- dist_actions model-ready
  da <- base_superset$data$dist_actions_model %||% NULL
  if (is.null(da) || !inherits(da, "data.frame")) {
    stop("dist_actions_model is missing in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  for (nm in c("internal_row", "internal_pu", "internal_action")) {
    if (!nm %in% names(da)) {
      stop("dist_actions_model must contain column '", nm, "'.", call. = FALSE)
    }
  }

  # ---- infer global number of actions
  all_actions <- sort(unique(as.integer(da$internal_action)))
  all_actions <- all_actions[is.finite(all_actions) & !is.na(all_actions)]
  if (length(all_actions) == 0L) {
    stop("Could not infer internal actions in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }
  n_actions_global <- max(all_actions)

  # ---- selected actions
  act_raw <- term$actions %||% NULL
  if (is.null(act_raw) || length(act_raw) == 0L) {
    act_int <- all_actions
  } else {
    act_int <- as.integer(act_raw)
    act_int <- sort(unique(act_int[is.finite(act_int) & !is.na(act_int)]))
  }

  if (length(act_int) == 0L) {
    return(v)
  }

  # ---- action weights
  aw <- term$action_weights %||% NULL
  if (is.null(aw)) {
    aw_map <- stats::setNames(rep(1, length(act_int)), act_int)
  } else {
    aw <- as.numeric(aw)
    if (length(aw) == length(all_actions)) {
      tmp_map <- stats::setNames(aw, all_actions)
      aw_map <- tmp_map[as.character(act_int)]
    } else if (length(aw) == length(act_int)) {
      aw_map <- stats::setNames(aw, act_int)
    } else {
      stop(
        "action_weights length must match either all actions or the selected action subset.",
        call. = FALSE
      )
    }
  }

  weight_multiplier <- as.numeric(term$weight_multiplier %||% 1)[1]
  if (!is.finite(weight_multiplier)) weight_multiplier <- 1

  n_pu <- nrow(base_superset$data$pu)
  if (!is.finite(n_pu) || is.na(n_pu) || n_pu <= 0L) {
    stop("Invalid n_pu in .pamo_objvec_action_boundary_cut().", call. = FALSE)
  }

  # ---- canonicalize relation exactly like C++
  ip1 <- as.integer(rel_model$internal_pu1)
  ip2 <- as.integer(rel_model$internal_pu2)
  ww  <- as.numeric(rel_model$weight)

  ok <- is.finite(ip1) & is.finite(ip2) & is.finite(ww)
  ip1 <- ip1[ok]
  ip2 <- ip2[ok]
  ww  <- ww[ok]

  self_w <- numeric(n_pu)
  edge_map <- new.env(parent = emptyenv(), hash = TRUE)

  make_key <- function(a, b) paste0(a, "::", b)

  for (r in seq_along(ip1)) {
    i <- ip1[r]
    j <- ip2[r]
    wij <- ww[r]

    if (i == j) {
      self_w[i] <- self_w[i] + wij
    } else {
      a <- min(i, j)
      b <- max(i, j)
      k <- make_key(a, b)

      old <- if (exists(k, envir = edge_map, inherits = FALSE)) {
        get(k, envir = edge_map, inherits = FALSE)
      } else {
        NA_real_
      }

      if (is.na(old) || wij > old) {
        assign(k, wij, envir = edge_map)
      }
    }
  }

  edge_keys <- ls(edge_map, all.names = TRUE)
  if (length(edge_keys) == 0L) {
    return(v)
  }

  edge_df <- data.frame(
    a = integer(length(edge_keys)),
    b = integer(length(edge_keys)),
    weight = numeric(length(edge_keys))
  )

  for (k in seq_along(edge_keys)) {
    parts <- strsplit(edge_keys[k], "::", fixed = TRUE)[[1]]
    edge_df$a[k] <- as.integer(parts[1])
    edge_df$b[k] <- as.integer(parts[2])
    edge_df$weight[k] <- get(edge_keys[k], envir = edge_map, inherits = FALSE)
  }

  edge_df <- edge_df[order(edge_df$a, edge_df$b), , drop = FALSE]
  k_edges <- nrow(edge_df)

  expected_n_y <- k_edges * n_actions_global
  if (expected_n_y != n_y) {
    stop(
      "Mismatch in y_action layout in .pamo_objvec_action_boundary_cut().\n",
      "expected n_y_action = k_edges * n_actions_global = ", expected_n_y,
      ", but model snapshot reports n_y_action = ", n_y, ".",
      call. = FALSE
    )
  }

  incident_w <- numeric(n_pu)
  for (r in seq_len(k_edges)) {
    incident_w[edge_df$a[r]] <- incident_w[edge_df$a[r]] + edge_df$weight[r]
    incident_w[edge_df$b[r]] <- incident_w[edge_df$b[r]] + edge_df$weight[r]
  }

  # ---- sparse map (pu, action) -> x column
  key2 <- function(i, a) paste0(i, "::", a)

  da_key <- key2(da$internal_pu, da$internal_action)
  x_cols <- x_offset + as.integer(da$internal_row) - 1L
  x_map <- stats::setNames(as.integer(x_cols), da_key)

  # ---- linear part on x
  for (act in act_int) {
    awi <- as.numeric(aw_map[[as.character(act)]])
    if (!is.finite(awi) || awi == 0) next

    for (i in seq_len(n_pu)) {
      k <- key2(i, act)

      # use [ ] instead of [[ ]] so missing keys return NA instead of error
      col_x <- unname(x_map[k])[1]

      if (!is.finite(col_x) || is.na(col_x)) next

      coef <- weight_multiplier * awi * (incident_w[i] + self_w[i])
      if (coef != 0) {
        v[col_x + 1L] <- v[col_x + 1L] + coef
      }
    }
  }

  # ---- edge part on global y_action block
  for (e in seq_len(k_edges)) {
    we <- edge_df$weight[e]

    for (act in act_int) {
      awi <- as.numeric(aw_map[[as.character(act)]])
      if (!is.finite(awi) || awi == 0) next

      # global layout: e is 0-based in the formula
      bcol0 <- y0 + (e - 1L) * n_actions_global + (act - 1L)
      coef <- weight_multiplier * (-2) * we * awi

      if (coef != 0) {
        v[bcol0 + 1L] <- v[bcol0 + 1L] + coef
      }
    }
  }

  v
}

.pamo_objective_to_ir <- function(x, spec) {
  stopifnot(inherits(x, "Problem"))
  stopifnot(is.list(spec), !is.null(spec$objective_id))

  id <- as.character(spec$objective_id)[1]
  a  <- spec$objective_args %||% list()

  sense <- as.character(spec$sense %||% NA_character_)[1]
  if (is.na(sense) || !nzchar(sense)) sense <- NA_character_

  .c1 <- function(z, default = NULL) {
    if (is.null(z)) return(default)
    as.character(z)[1]
  }

  .n1 <- function(z, default = NULL) {
    if (is.null(z)) return(default)
    as.numeric(z)[1]
  }

  .l1 <- function(z, default = FALSE) {
    if (is.null(z)) return(default)
    isTRUE(z)
  }

  .chr <- function(z) {
    if (is.null(z)) return(NULL)
    z <- as.character(z)
    z <- unique(z[!is.na(z) & nzchar(z)])
    if (length(z) == 0) return(NULL)
    z
  }

  # ------------------------------------------------------------------
  # min_cost
  # ------------------------------------------------------------------
  if (identical(id, "min_cost")) {

    inc_pu  <- .l1(a$include_pu_cost, TRUE)
    inc_act <- .l1(a$include_action_cost, TRUE)
    actions <- .chr(a$actions)
    feats   <- .chr(a$features)

    terms <- list()
    if (inc_pu) {
      terms <- c(terms, list(list(
        type = "pu_cost",
        features = feats
      )))
    }
    if (inc_act) {
      terms <- c(terms, list(list(
        type = "action_cost",
        actions = actions
      )))
    }

    a$include_pu_cost     <- inc_pu
    a$include_action_cost <- inc_act
    a$actions             <- actions
    a$features            <- feats

    return(list(
      sense = "min",
      terms = terms,
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # max_benefit
  # ------------------------------------------------------------------
  if (identical(id, "max_benefit")) {
    bcol    <- .c1(a$benefit_col, "benefit")
    actions <- .chr(a$actions)
    feats   <- .chr(a$features)

    a$benefit_col <- bcol
    a$actions     <- actions
    a$features    <- feats

    return(list(
      sense = "max",
      terms = list(list(
        type = "benefit",
        benefit_col = bcol,
        actions = actions,
        features = feats
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # max_profit
  # ------------------------------------------------------------------
  if (identical(id, "max_profit")) {
    pcol    <- .c1(a$profit_col, "profit")
    actions <- .chr(a$actions)

    a$profit_col <- pcol
    a$actions    <- actions

    return(list(
      sense = "max",
      terms = list(list(
        type = "profit",
        profit_col = pcol,
        actions = actions
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # max_net_profit
  # ------------------------------------------------------------------
  if (identical(id, "max_net_profit")) {
    pcol    <- .c1(a$profit_col, "profit")
    inc_pu  <- .l1(a$include_pu_cost, TRUE)
    inc_act <- .l1(a$include_action_cost, TRUE)
    actions <- .chr(a$actions)
    feats   <- .chr(a$features)

    a$profit_col          <- pcol
    a$include_pu_cost     <- inc_pu
    a$include_action_cost <- inc_act
    a$actions             <- actions
    a$features            <- feats

    return(list(
      sense = "max",
      terms = list(list(
        type = "net_profit",
        profit_col = pcol,
        include_pu_cost = inc_pu,
        include_action_cost = inc_act,
        actions = actions,
        features = feats
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # min_fragmentation
  # ------------------------------------------------------------------
  if (identical(id, "min_fragmentation")) {
    rel   <- .c1(a$relation_name, "boundary")
    mul   <- .n1(a$weight_multiplier, 1)
    feats <- .chr(a$features)

    a$relation_name     <- rel
    a$weight_multiplier <- mul
    a$features          <- feats

    return(list(
      sense = "min",
      terms = list(list(
        type = "boundary_cut",
        relation_name = rel,
        weight_multiplier = mul,
        features = feats
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # min_action_fragmentation
  # ------------------------------------------------------------------
  if (identical(id, "min_action_fragmentation")) {
    rel     <- .c1(a$relation_name, "boundary")
    mul     <- .n1(a$weight_multiplier, 1)
    actions <- .chr(a$actions)

    aw <- a$action_weights %||% NULL
    if (!is.null(aw)) aw <- as.numeric(aw)

    a$relation_name     <- rel
    a$weight_multiplier <- mul
    a$actions           <- actions
    a$action_weights    <- aw

    return(list(
      sense = "min",
      terms = list(list(
        type = "action_boundary_cut",
        relation_name = rel,
        weight_multiplier = mul,
        actions = actions,
        action_weights = aw
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # min_loss
  # ------------------------------------------------------------------
  if (identical(id, "min_loss")) {
    lcol    <- .c1(a$loss_col, "loss")
    actions <- .chr(a$actions)
    feats   <- .chr(a$features)

    a$loss_col <- lcol
    a$actions  <- actions
    a$features <- feats

    return(list(
      sense = "min",
      terms = list(list(
        type = "loss",
        actions = actions,
        features = feats
      )),
      objective_id = id,
      objective_args = a
    ))
  }

  # ------------------------------------------------------------------
  # min_intervention_impact
  # ------------------------------------------------------------------
  if (identical(id, "min_intervention_impact")) {
    icol    <- .c1(a$impact_col, "amount")
    feats   <- .chr(a$features)
    actions <- .chr(a$actions)

    a$impact_col <- icol
    a$features   <- feats
    a$actions    <- actions

    return(list(
      sense = "min",
      terms = list(list(
        type = "intervention_impact",
        impact_col = icol,
        features = feats,
        actions = actions
      )),
      objective_id = id,
      objective_args = a
    ))
  }


  stop("Unknown objective_id in .pamo_objective_to_ir(): ", id, call. = FALSE)
}

# -------------------------------------------------------------------------
# Cloning base Problem safely for MO runs
# (Avoid copying externalptr / built model pointer)
# -------------------------------------------------------------------------

.pamo_clone_base <- function(base) {
  stopifnot(inherits(base, "Problem"))

  b <- pproto(NULL, base)

  # copia superficial del contenedor
  b$data <- base$data

  # independizar solo lo mutable
  b$data$model_args <- unserialize(serialize(base$data$model_args %||% list(), NULL))
  b$data$meta <- unserialize(serialize(base$data$meta %||% list(), NULL))

  # limpiar estado de modelo para este run
  b$data$model_ptr <- NULL
  b$data$has_model <- FALSE
  b$data$model_list <- NULL
  b$data$model_index <- NULL
  b$data$runtime_updates <- NULL
  b$data$mo_cache <- NULL

  b$data$meta$model_dirty <- TRUE

  b
}

# -------------------------------------------------------------------------
# Activate an IR as a single-objective config in multiscape::Problem
# (Used by the "rebuild + pad" objective vector strategy)
# -------------------------------------------------------------------------

.pamo_activate_ir_as_single_objective <- function(x, ir) {
  stopifnot(inherits(x, "Problem"))
  stopifnot(is.list(ir), !is.null(ir$objective_id))

  id <- as.character(ir$objective_id)[1]
  a  <- ir$objective_args %||% list()

  map <- list(
    min_cost = "minimizeCosts",
    max_benefit = "maximizeBenefits",
    max_profit = "maximizeProfit",
    min_loss = "minimizeLosses",
    max_net_profit = "maximizeNetProfit",
    min_fragmentation = "minimizeFragmentation",
    min_action_fragmentation = "minimizeActionFragmentation",
    min_intervention_impact = "minimizeInterventionImpact",
    custom = "custom"
  )

  mt <- map[[id]]
  if (is.null(mt)) stop("No model_type mapping for objective_id: ", id, call. = FALSE)

  if (is.null(x$data$model_args) || !is.list(x$data$model_args)) x$data$model_args <- list()
  x$data$model_args$model_type <- mt
  x$data$model_args$objective_id <- id
  x$data$model_args$objective_args <- a

  if (is.null(x$data$meta) || !is.list(x$data$meta)) x$data$meta <- list()
  x$data$meta$model_dirty <- TRUE

  x
}

# -------------------------------------------------------------------------
# Superset model selection + objective-vector extraction
# Strategy:
# - Build ONE "superset" model using the objective that likely introduces most aux vars
# - For each objective, build its own model and extract model$obj
# - Pad with zeros to match superset length
# This avoids dealing with offsets/variable mapping for now.
# -------------------------------------------------------------------------
.pamo_prepare_relation_model <- function(rel) {
  stopifnot(is.data.frame(rel), nrow(rel) > 0)

  need <- c("internal_pu1", "internal_pu2", "weight")
  if (!all(need %in% names(rel))) {
    stop(
      "Spatial relation must contain columns: ",
      paste(need, collapse = ", "),
      call. = FALSE
    )
  }

  rel <- rel[, c("internal_pu1","internal_pu2","weight",
                 intersect(names(rel), c("distance","source","relation_name"))),
             drop = FALSE]

  rel$internal_pu1 <- as.integer(rel$internal_pu1)
  rel$internal_pu2 <- as.integer(rel$internal_pu2)
  rel$weight <- as.numeric(rel$weight)

  if (anyNA(rel$internal_pu1) || anyNA(rel$internal_pu2)) {
    stop("Relation has NA in internal_pu1/internal_pu2.", call. = FALSE)
  }
  if (any(!is.finite(rel$weight)) || any(rel$weight < 0)) {
    stop("Relation has non-finite or negative weights.", call. = FALSE)
  }

  rel <- rel[order(rel$internal_pu1, rel$internal_pu2), , drop = FALSE]
  rel$internal_edge <- seq_len(nrow(rel))
  rel
}

.pamo_prepare_superset_model <- function(base, ir_list) {
  stopifnot(inherits(base, "Problem"))
  stopifnot(is.list(ir_list), length(ir_list) > 0)

  spec <- .pamo_compile_superset_spec(base, ir_list)

  # elegir un objetivo "semilla" válido para que el core del modelo se materialice.
  # preferimos min_cost si existe; si no, el primero.
  pick_seed_ir <- function(ir_list) {
    idx_cost <- which(vapply(
      ir_list,
      function(ir) identical(ir$objective_id %||% "", "min_cost"),
      logical(1)
    ))
    if (length(idx_cost) > 0L) return(ir_list[[idx_cost[1]]])
    ir_list[[1]]
  }

  ir_seed <- pick_seed_ir(ir_list)

  b <- .pamo_clone_base(base)
  b <- .pamo_activate_ir_as_single_objective(b, ir_seed)

  b$data$model_args <- b$data$model_args %||% list()
  b$data$model_args$mo_mode <- TRUE

  # la clave del superset: needs estructurales, independientes del método
  b$data$model_args$needs <- utils::modifyList(
    b$data$model_args$needs %||% list(),
    spec$needs
  )

  if (!is.null(spec$relation_name)) {
    b$data$model_args$needs$relation_name <- spec$relation_name
  }

  b <- .pa_build_model(b)

  op <- b$data$model_ptr
  if (is.null(op)) {
    stop("Superset build failed: model_ptr is NULL.", call. = FALSE)
  }

  # sanity check mínimo: si hay objetivos de representación, el modelo debe tener z
  if (isTRUE(spec$needs$z)) {
    op_list <- .pa_model_from_ptr(
      op,
      args = b$data$model_args %||% list(),
      drop_triplets = TRUE
    )

    n_z <- as.integer(op_list$n_z %||% 0L)
    if (n_z <= 0L) {
      stop(
        "MO superset requires z variables, but the built model has n_z = 0.",
        call. = FALSE
      )
    }
  }

  # sanity check PU fragmentation
  if (isTRUE(spec$needs$y_pu)) {
    op_list <- .pa_model_from_ptr(
      op,
      args = b$data$model_args %||% list(),
      drop_triplets = TRUE
    )

    n_y_pu <- as.integer(op_list$n_y_pu %||% 0L)
    if (n_y_pu <= 0L) {
      stop(
        "MO superset requires PU fragmentation auxiliaries, but the built model has n_y_pu = 0.",
        call. = FALSE
      )
    }
  }

  b
}


.pamo_model_obj_length <- function(x) {
  stopifnot(inherits(x, "Problem"))
  if (is.null(x$data$model_ptr)) stop("Problem has no model_ptr; build model first.", call. = FALSE)

  m <- .pa_model_from_ptr(
    x$data$model_ptr,
    args = x$data$model_args %||% list(),
    drop_triplets = TRUE
  )
  length(m$obj)
}


# -------------------------------------------------------------------------
# Optional normalization helpers (for obj vectors)
# NOTE: this is normalization of objective *vectors* (not weights).
# If you don't want it now, keep but don't use.
# -------------------------------------------------------------------------

.pamo_normalize_vec <- function(v, method = c("max", "l1", "l2")) {
  method <- match.arg(method)
  v <- as.numeric(v)
  if (!any(is.finite(v))) return(v)

  if (method == "max") {
    s <- max(abs(v), na.rm = TRUE)
    if (is.finite(s) && s > 0) return(v / s)
    return(v)
  }

  if (method == "l1") {
    s <- sum(abs(v), na.rm = TRUE)
    if (is.finite(s) && s > 0) return(v / s)
    return(v)
  }

  if (method == "l2") {
    s <- sqrt(sum(v^2, na.rm = TRUE))
    if (is.finite(s) && s > 0) return(v / s)
    return(v)
  }

  v
}

# ---------------------------------------------------------
# Internal: solve weighted
# ---------------------------------------------------------
.pamo_solve_weighted <- function(x, ...) {
  stopifnot(inherits(x, "Problem"))

  data <- x$data %||% list()
  method <- x$data$method %||% list()
  aliases <- as.character(method$aliases %||% character(0))
  w <- as.numeric(method$weights %||% numeric(0))
  normalize_weights <- isTRUE(method$normalize_weights)
  objective_scaling <- isTRUE(method$objective_scaling)

  dots <- list(...)
  verbose <- .pamo_cli_enabled(x, dots)

  # pass-through solver controls
  gap_limit  <- dots$gap_limit %||% NULL
  time_limit <- dots$time_limit %||% NULL

  .pamo_cli_header("Weighted method", verbose = verbose)
  .pamo_cli_step("Preparing weighted design.", verbose = verbose)

  if (length(aliases) == 0L) {
    stop("Weighted method: missing aliases.", call. = FALSE)
  }

  if (length(w) == 0L) {
    stop("Weighted method: missing weights. Provide them in set_method_weighted().", call. = FALSE)
  }

  if (length(w) != length(aliases)) {
    stop(
      "Weighted method: length(weights) must match length(aliases).\n",
      "length(weights) = ", length(w), ", length(aliases) = ", length(aliases), ".",
      call. = FALSE
    )
  }

  # ---- design: one row per run, only design parameters
  design_df <- data.frame(run_id = 1L, stringsAsFactors = FALSE)
  for (i in seq_along(aliases)) {
    design_df[[paste0("weight_", aliases[i])]] <- as.numeric(w[i])
  }

  n_runs <- nrow(design_df)

  .pamo_cli_step(
    "Design contains {.val {n_runs}} run{?s}.",
    verbose = verbose
  )

  scales <- NULL
  if (objective_scaling) {
    .pamo_cli_step(
      "Computing payoff ranges for {.field objective_scaling = TRUE}.",
      verbose = verbose
    )

    scales <- .pamo_compute_weighted_ranges(
      x = x,
      aliases = aliases,
      verbose = verbose,
      gap_limit = gap_limit,
      time_limit = time_limit
    )
  }

  solutions <- vector("list", n_runs)

  value_mat <- matrix(NA_real_, n_runs, length(aliases))
  colnames(value_mat) <- paste0("value_", aliases)

  status  <- character(n_runs)
  runtime <- numeric(n_runs)
  gap     <- numeric(n_runs)

  for (r in seq_len(n_runs)) {
    w_r <- as.numeric(design_df[r, paste0("weight_", aliases), drop = TRUE])

    one <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = aliases,
        weights = w_r,
        normalize_weights = normalize_weights,
        objective_scaling = objective_scaling,
        scales = scales,
        gap_limit = gap_limit,
        time_limit = time_limit
      )
    )

    solutions[[r]] <- one$solution
    status[r]  <- as.character(one$status %||% NA_character_)
    runtime[r] <- as.numeric(one$runtime %||% NA_real_)
    gap[r]     <- as.numeric(one$gap %||% NA_real_)

    # evaluate all registered aliases on the obtained solution
    alias_values <- stats::setNames(
      vapply(
        aliases,
        function(a) .pamo_eval_alias_on_solution(x, one$solution, a),
        numeric(1)
      ),
      aliases
    )

    value_mat[r, ] <- unname(as.numeric(alias_values))

    # store metadata inside each individual Solution
    if (!is.null(solutions[[r]]) && inherits(solutions[[r]], "Solution")) {
      solutions[[r]]$solution$alias_values <- alias_values
      solutions[[r]]$meta$run_id <- design_df$run_id[r]
      solutions[[r]]$method$type <- "weighted"
      solutions[[r]]$method$weights <- stats::setNames(as.numeric(w_r), aliases)
    }
  }

  # ---- runs: one row per run, only outputs / summaries
  runs <- data.frame(
    run_id = design_df$run_id,
    status = status,
    runtime = runtime,
    gap = gap,
    stringsAsFactors = FALSE
  )

  runs <- cbind(
    runs,
    as.data.frame(value_mat, stringsAsFactors = FALSE)
  )

  summary_set <- .pamo_bind_solution_summaries(
    solutions = solutions,
    run_ids = design_df$run_id
  )

  .pamo_cli_done("Weighted solve finished.", verbose = verbose)

  pproto(
    NULL, SolutionSet,
    problem = x,
    solution = list(
      design = design_df,
      runs = runs,
      solutions = solutions
    ),
    summary = summary_set,
    diagnostics = list(
      n_design = nrow(design_df),
      n_runs = nrow(runs),
      n_solutions = length(solutions),
      status_summary = .pa_solutionset_status_summary(runs),
      runtime_range = if ("runtime" %in% names(runs)) .pa_solutionset_range_text(runs$runtime, digits = 3) else "none",
      gap_range = if ("gap" %in% names(runs)) .pa_solutionset_range_text(runs$gap, digits = 4) else "none"
    ),
    method = method,
    meta = list(
      call = match.call()
    ),
    name = "solset"
  )
}


.pamo_solve_epsilon_constraint <- function(x, ...) {

  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% NULL
  if (is.null(method) || !is.list(method)) {
    stop("epsilon_constraint: x$data$method is missing or invalid.", call. = FALSE)
  }

  data <- x$data %||% list()
  primary     <- as.character(method$primary %||% NA_character_)[1]
  aliases     <- as.character(method$aliases %||% character(0))
  constrained <- as.character(method$constrained %||% character(0))
  mode        <- as.character(method$mode %||% "manual")[1]

  # pass-through solver controls / progress / verbosity
  dots <- list(...)
  gap_limit  <- dots$gap_limit %||% NULL
  time_limit <- dots$time_limit %||% NULL
  verbose <- .pamo_cli_enabled(x, dots)

  # progress control
  progress <- dots$progress %||% dots$verbose %||% interactive()
  progress <- isTRUE(progress)

  .pamo_cli_header("Epsilon-constraint method", verbose = verbose)

  if (is.na(primary) || !nzchar(primary)) {
    stop("epsilon_constraint: missing primary objective.", call. = FALSE)
  }
  if (length(aliases) == 0L) {
    stop("epsilon_constraint: aliases are missing.", call. = FALSE)
  }
  if (length(constrained) == 0L) {
    stop("epsilon_constraint: no constrained objectives were defined.", call. = FALSE)
  }
  if (!mode %in% c("manual", "auto")) {
    stop("epsilon_constraint: unknown mode '", mode, "'.", call. = FALSE)
  }

  .pamo_cli_step(
    "Primary objective: {.val {primary}}. Secondary objectives: {.val {paste(constrained, collapse = ', ')}}.",
    verbose = verbose
  )

  # ---------------------------------------------------------
  # Build or recover epsilon design
  # ---------------------------------------------------------
  if (identical(mode, "manual")) {
    .pamo_cli_step("Using user-supplied epsilon grid.", verbose = verbose)

    design_df <- method$runs

    if (is.null(design_df) || !inherits(design_df, "data.frame") || nrow(design_df) == 0L) {
      stop("epsilon_constraint (manual): x$data$method$runs is missing/empty.", call. = FALSE)
    }

  } else {
    .pamo_cli_step("Building automatic epsilon grid.", verbose = verbose)

    if (!exists(".pamo_build_auto_epsilon_runs", mode = "function")) {
      stop(
        "epsilon_constraint (auto): missing internal helper .pamo_build_auto_epsilon_runs().",
        call. = FALSE
      )
    }

    design_df <- method$runs %||% NULL
    if (is.null(design_df)) {
      design_df <- .pamo_build_auto_epsilon_runs(
        x,
        gap_limit = gap_limit,
        time_limit = time_limit,
        verbose = verbose
      )
    }

    if (is.null(design_df) || !inherits(design_df, "data.frame") || nrow(design_df) == 0L) {
      stop("epsilon_constraint (auto): generated epsilon runs are empty.", call. = FALSE)
    }

    x$data$method$runs <- design_df
  }

  if (!("run_id" %in% names(design_df))) {
    design_df$run_id <- seq_len(nrow(design_df))
  }

  eps_cols <- paste0("eps_", constrained)
  miss_cols <- setdiff(eps_cols, names(design_df))
  if (length(miss_cols) > 0L) {
    stop(
      "epsilon_constraint: design is missing epsilon columns: ",
      paste(miss_cols, collapse = ", "),
      call. = FALSE
    )
  }

  .pamo_cli_step(
    "Grid ready with {.val {nrow(design_df)}} run{?s}.",
    verbose = verbose
  )

  # ---------------------------------------------------------
  # Solve runs
  # ---------------------------------------------------------
  n_runs <- nrow(design_df)

  solutions <- vector("list", n_runs)

  value_mat <- matrix(NA_real_, n_runs, length(aliases))
  colnames(value_mat) <- paste0("value_", aliases)

  status  <- character(n_runs)
  runtime <- numeric(n_runs)
  gap     <- numeric(n_runs)

  progress_id <- NULL
  if (progress && n_runs > 1L) {
    progress_id <- NULL
    if (progress && n_runs > 1L) {
      progress_id <- cli::cli_progress_bar(
        name = "Solving runs",
        total = n_runs,
        clear = FALSE,
        format = "{cli::pb_name}{cli::pb_bar} {cli::pb_current}/{cli::pb_total} | ETA: {cli::pb_eta}"
      )
    }
  }

  for (r in seq_len(n_runs)) {

    eps_r <- as.list(design_df[r, eps_cols, drop = FALSE])
    names(eps_r) <- constrained

    one <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "epsilon_constraint",
        primary = primary,
        eps = eps_r,
        gap_limit = gap_limit,
        time_limit = time_limit
      )
    )

    solutions[[r]] <- one$solution
    status[r]  <- as.character(one$status %||% NA_character_)
    runtime[r] <- as.numeric(one$runtime %||% NA_real_)
    gap[r]     <- as.numeric(one$gap %||% NA_real_)

    alias_values <- stats::setNames(
      vapply(
        aliases,
        function(a) .pamo_eval_alias_on_solution(x, one$solution, a),
        numeric(1)
      ),
      aliases
    )

    value_mat[r, ] <- unname(as.numeric(alias_values))

    if (!is.null(solutions[[r]]) && inherits(solutions[[r]], "Solution")) {
      solutions[[r]]$solution$alias_values <- alias_values
      solutions[[r]]$meta$run_id <- design_df$run_id[r]
      solutions[[r]]$method$type <- "epsilon_constraint"
      solutions[[r]]$method$primary_alias <- primary
      solutions[[r]]$method$eps <- stats::setNames(
        as.numeric(design_df[r, eps_cols, drop = TRUE]),
        constrained
      )
    }

    if (!is.null(progress_id)) {
      cli::cli_progress_update(id = progress_id, set = r)
    }
  }

  if (!is.null(progress_id)) {
    cli::cli_progress_done(id = progress_id)
  }

  # ---------------------------------------------------------
  # Runs summary: outputs only
  # ---------------------------------------------------------
  runs <- data.frame(
    run_id = design_df$run_id,
    status = status,
    runtime = runtime,
    gap = gap,
    stringsAsFactors = FALSE
  )

  runs <- cbind(
    runs,
    as.data.frame(value_mat, stringsAsFactors = FALSE)
  )

  # ---------------------------------------------------------
  # Extras: method-specific metadata
  # ---------------------------------------------------------
  extras <- list()
  extras$epsilon_grid <- design_df

  ext <- attr(design_df, "extremes", exact = TRUE)
  if (!is.null(ext)) {
    extras$extremes <- ext
  }

  if (!is.null(method$meta)) {
    extras$method_meta <- method$meta
  }

  summary_set <- .pamo_bind_solution_summaries(
    solutions = solutions,
    run_ids = design_df$run_id
  )

  .pamo_cli_done("Epsilon-constraint solve finished.", verbose = verbose)

  ext <- attr(design_df, "extremes", exact = TRUE)

  ranges_df <- NULL
  if (!is.null(ext)) {
    primary_vals <- c(
      as.numeric(ext$primary_at_primary %||% NA_real_),
      as.numeric(ext$primary_at_secondary %||% NA_real_)
    )
    primary_vals <- primary_vals[is.finite(primary_vals)]

    secondary_vals <- c(
      as.numeric(ext$secondary_min %||% NA_real_),
      as.numeric(ext$secondary_max %||% NA_real_)
    )
    secondary_vals <- secondary_vals[is.finite(secondary_vals)]

    ranges_df <- data.frame(
      alias = c(primary, constrained[1]),
      role = c("primary", "secondary"),
      min = c(
        if (length(primary_vals)) min(primary_vals) else NA_real_,
        if (length(secondary_vals)) min(secondary_vals) else NA_real_
      ),
      max = c(
        if (length(primary_vals)) max(primary_vals) else NA_real_,
        if (length(secondary_vals)) max(secondary_vals) else NA_real_
      ),
      source = c("anchors", "auto_grid"),
      stringsAsFactors = FALSE
    )
  }

  pproto(
    NULL, SolutionSet,
    problem = x,
    solution = list(
      design = design_df,
      runs = runs,
      solutions = solutions,
      frontier = list(
        primary = primary,
        secondary = constrained,
        ranges = ranges_df,
        extremes = ext
      )
    ),
    summary = summary_set,
    diagnostics = list(
      n_design = nrow(design_df),
      n_runs = nrow(runs),
      n_solutions = length(solutions),
      n_optimal = sum(runs$status == "optimal", na.rm = TRUE),
      n_infeasible = sum(runs$status %in% c("infeasible", "infeasible_or_unbounded"), na.rm = TRUE),
      status_summary = .pa_solutionset_status_summary(runs),
      runtime_range = if ("runtime" %in% names(runs)) .pa_solutionset_range_text(runs$runtime, digits = 3) else "none",
      gap_range = if ("gap" %in% names(runs)) .pa_solutionset_range_text(runs$gap, digits = 4) else "none"
    ),
    method = method,
    meta = list(
      call = match.call()
    ),
    name = "solset"
  )
}


.pamo_build_auto_epsilon_runs <- function(x,
                                          gap_limit = NULL,
                                          time_limit = NULL,
                                          verbose = FALSE) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% list()

  primary          <- as.character(method$primary %||% NA_character_)[1]
  aliases          <- as.character(method$aliases %||% character(0))
  constrained      <- as.character(method$constrained %||% character(0))
  n_points         <- as.integer(method$n_points %||% NA_integer_)[1]
  include_extremes <- isTRUE(method$include_extremes)

  if (is.na(primary) || !nzchar(primary)) {
    stop("Auto epsilon mode: missing primary objective.", call. = FALSE)
  }

  if (length(aliases) == 0L) {
    stop("Auto epsilon mode: aliases are missing.", call. = FALSE)
  }

  if (!primary %in% aliases) {
    stop("Auto epsilon mode: primary must be included in aliases.", call. = FALSE)
  }

  if (length(constrained) == 0L) {
    stop("Auto epsilon mode requires at least one constrained objective.", call. = FALSE)
  }

  if (!is.finite(n_points) || is.na(n_points) || n_points < 2L) {
    stop("Auto epsilon mode requires n_points >= 2.", call. = FALSE)
  }

  if (length(aliases) != 2L) {
    stop(
      "Auto epsilon mode currently supports exactly 2 objectives.\n",
      "You provided ", length(aliases), " objective(s): ",
      paste(aliases, collapse = ", "), ".\n",
      "Use mode='manual' for 3+ objectives.",
      call. = FALSE
    )
  }

  if (length(constrained) != 1L) {
    stop(
      "Auto epsilon mode currently expects exactly 1 constrained objective.",
      call. = FALSE
    )
  }

  secondary <- constrained[1]

  .pamo_cli_step(
    "Computing automatic epsilon bounds for {.val {secondary}} against {.val {primary}}.",
    verbose = verbose
  )

  ext <- .pamo_compute_epsilon_extremes_2obj(
    x = x,
    primary = primary,
    secondary = secondary,
    gap_limit = gap_limit,
    time_limit = time_limit
  )

  sec_min <- as.numeric(ext$secondary_min)[1]
  sec_max <- as.numeric(ext$secondary_max)[1]

  if (!is.finite(sec_min) || !is.finite(sec_max)) {
    stop("Auto epsilon mode: computed epsilon bounds are not finite.", call. = FALSE)
  }

  if (sec_min > sec_max) {
    tmp <- sec_min
    sec_min <- sec_max
    sec_max <- tmp
  }

  .pamo_cli_step(
    "Secondary range: [{.val {format(sec_min, digits = 6)}}, {.val {format(sec_max, digits = 6)}}].",
    verbose = verbose
  )

  eps_vals <- seq(from = sec_min, to = sec_max, length.out = n_points)

  if (!isTRUE(include_extremes)) {
    if (length(eps_vals) <= 2L) {
      stop(
        "include_extremes=FALSE requires n_points >= 3.",
        call. = FALSE
      )
    }
    eps_vals <- eps_vals[2:(length(eps_vals) - 1L)]
  }

  if (length(eps_vals) == 0L) {
    stop("Auto epsilon mode produced an empty epsilon grid.", call. = FALSE)
  }

  eps_col <- paste0("eps_", secondary)

  design_df <- data.frame(
    run_id = seq_along(eps_vals),
    stringsAsFactors = FALSE
  )
  design_df[[eps_col]] <- as.numeric(eps_vals)

  attr(design_df, "extremes") <- ext
  attr(design_df, "primary_alias") <- primary
  attr(design_df, "secondary_alias") <- secondary
  attr(design_df, "include_extremes") <- include_extremes
  attr(design_df, "n_points_requested") <- n_points

  .pamo_cli_done(
    "Automatic epsilon grid built with {.val {nrow(design_df)}} run{?s}.",
    verbose = verbose
  )

  design_df
}


# -------------------------------------------------------------------------
# "Custom objectives" (DEV/advanced) constructor + registry add
# (This keeps the path you started, but stores in base$data$objectives)
# -------------------------------------------------------------------------

.pamo_objective <- function(alias, sense = c("min", "max"), build, eval, meta = list()) {
  sense <- match.arg(sense)
  if (!is.character(alias) || length(alias) != 1 || !nzchar(alias)) {
    stop("objective `alias` must be a non-empty string.", call. = FALSE)
  }
  if (!is.function(build)) stop("objective `build` must be a function.", call. = FALSE)
  if (!is.function(eval))  stop("objective `eval` must be a function.", call. = FALSE)

  structure(
    list(alias = alias, sense = sense, build = build, eval = eval, meta = meta),
    class = "pa_objective"
  )
}

add_objective <- function(x, objective) {
  #x <- .pamo_as_mo(x)
  x <- .pa_clone_data(x)

  if (!inherits(objective, "pa_objective")) {
    stop("add_objective() expects an objective of class 'pa_objective'.", call. = FALSE)
  }

  # store it as an atomic spec in the multiscape registry
  if (is.null(x$data$objectives) || !is.list(x$data$objectives)) {
    x$data$objectives <- list()
  }
  if (!is.null(x$data$objectives[[objective$alias]])) {
    stop("Objective alias already exists: '", objective$alias, "'.", call. = FALSE)
  }

  x$data$objectives[[objective$alias]] <- list(
    alias = objective$alias,
    objective_id = "custom",
    model_type = "custom",
    objective_args = list(
      build = objective$build,
      eval  = objective$eval,
      meta  = objective$meta %||% list()
    ),
    sense = objective$sense
  )

  x
}



.pamo_objvec_from_ir <- function(base_superset, ir) {
  stopifnot(inherits(base_superset, "Problem"))

  op <- base_superset$data$model_ptr
  if (is.null(op)) {
    stop("Superset model is not built (missing model_ptr).", call. = FALSE)
  }

  # ------------------------------------------------------------
  # local helpers
  # ------------------------------------------------------------
  .chr <- function(z) {
    if (is.null(z)) return(NULL)
    z <- as.character(z)
    z <- unique(z[!is.na(z) & nzchar(z)])
    if (length(z) == 0) return(NULL)
    z
  }

  .resolve_feature_internal <- function(features) {
    if (is.null(features)) return(integer(0))

    feat_df <- base_superset$data$features
    if (is.null(feat_df) || !inherits(feat_df, "data.frame") || nrow(feat_df) == 0) {
      stop("Cannot resolve feature subset: x$data$features is missing/empty.", call. = FALSE)
    }
    if (!all(c("id", "internal_id") %in% names(feat_df))) {
      stop("x$data$features must contain 'id' and 'internal_id'.", call. = FALSE)
    }

    vals_chr <- .chr(features)
    vals_int <- suppressWarnings(as.integer(vals_chr))

    # 1) try internal ids first
    if (length(vals_int) > 0 && all(!is.na(vals_int)) &&
        all(vals_int %in% feat_df$internal_id)) {
      return(as.integer(unique(vals_int)))
    }

    # 2) then external ids
    idx <- match(vals_chr, as.character(feat_df$id))
    if (all(!is.na(idx))) {
      return(as.integer(feat_df$internal_id[idx]))
    }

    # 3) then names if available
    if ("name" %in% names(feat_df)) {
      idx2 <- match(vals_chr, as.character(feat_df$name))
      if (all(!is.na(idx2))) {
        return(as.integer(feat_df$internal_id[idx2]))
      }
    }

    bad <- vals_chr[
      !(vals_chr %in% as.character(feat_df$id)) &
        !(vals_chr %in% as.character(feat_df$internal_id)) &
        !("name" %in% names(feat_df) && vals_chr %in% as.character(feat_df$name))
    ]

    stop(
      "Unknown features in objective subset: ",
      paste(unique(bad), collapse = ", "),
      call. = FALSE
    )
  }

  .resolve_action_internal <- function(actions) {
    if (is.null(actions)) return(integer(0))

    act_df <- base_superset$data$actions
    if (is.null(act_df) || !inherits(act_df, "data.frame") || nrow(act_df) == 0) {
      stop("Cannot resolve action subset: x$data$actions is missing/empty.", call. = FALSE)
    }
    if (!all(c("id", "internal_id") %in% names(act_df))) {
      stop("x$data$actions must contain 'id' and 'internal_id'.", call. = FALSE)
    }

    vals_chr <- .chr(actions)
    vals_int <- suppressWarnings(as.integer(vals_chr))

    # 1) try internal ids first
    if (length(vals_int) > 0 && all(!is.na(vals_int)) &&
        all(vals_int %in% act_df$internal_id)) {
      return(as.integer(unique(vals_int)))
    }

    # 2) otherwise fall back to the standard resolver (ids / action_set / etc.)
    out <- .pa_resolve_action_subset(base_superset, subset = vals_chr)
    as.integer(out$internal_id)
  }

  .subset_dist_actions <- function(actions = NULL) {
    da <- base_superset$data$dist_actions_model
    if (is.null(da) || !inherits(da, "data.frame")) {
      stop("dist_actions_model is missing.", call. = FALSE)
    }

    act_int <- .resolve_action_internal(actions)
    if (length(act_int) == 0) return(da)

    da[da$internal_action %in% act_int, , drop = FALSE]
  }

  .subset_dist_effects <- function(df, actions = NULL, features = NULL) {
    if (is.null(df) || !inherits(df, "data.frame")) return(df)

    act_int  <- .resolve_action_internal(actions)
    feat_int <- .resolve_feature_internal(features)

    out <- df

    if (length(act_int) > 0 && "internal_action" %in% names(out)) {
      out <- out[out$internal_action %in% act_int, , drop = FALSE]
    }

    if (length(feat_int) > 0 && "internal_feature" %in% names(out)) {
      out <- out[out$internal_feature %in% feat_int, , drop = FALSE]
    }

    out
  }

  .subset_dist_profit <- function(df, actions = NULL) {
    if (is.null(df) || !inherits(df, "data.frame")) return(df)

    act_int <- .resolve_action_internal(actions)
    if (length(act_int) == 0) return(df)

    df[df$internal_action %in% act_int, , drop = FALSE]
  }

  .subset_dist_features <- function(df, features = NULL) {
    if (is.null(df) || !inherits(df, "data.frame")) return(df)

    feat_int <- .resolve_feature_internal(features)
    if (length(feat_int) == 0) return(df)

    df[df$internal_feature %in% feat_int, , drop = FALSE]
  }

  # Convención MO: motor en MIN; luego en R se invierte signo para objetivos max
  rcpp_reset_objective(op, "min")

  terms <- ir$terms %||% list()

  for (t in terms) {
    type <- t$type

    if (identical(type, "pu_cost")) {

      feats <- .chr(t$features)
      if (!is.null(feats)) {
        stop(
          "Subset by features is not currently supported for 'pu_cost'. ",
          "PU costs are attached to w variables, not directly to feature-specific contributions.",
          call. = FALSE
        )
      }

      rcpp_add_objective_min_cost(
        op,
        pu_data = base_superset$data$pu,
        dist_actions_data = base_superset$data$dist_actions_model,
        include_pu_cost = TRUE,
        include_action_cost = FALSE,
        weight = 1.0
      )

    } else if (identical(type, "action_cost")) {

      da_sub <- .subset_dist_actions(actions = t$actions)

      rcpp_add_objective_min_cost(
        op,
        pu_data = base_superset$data$pu,
        dist_actions_data = da_sub,
        include_pu_cost = FALSE,
        include_action_cost = TRUE,
        weight = 1.0
      )

    } else if (identical(type, "boundary_cut")) {

      feats <- .chr(t$features)
      if (!is.null(feats)) {
        stop(
          "Subset by features is not currently supported for 'boundary_cut'. ",
          "PU fragmentation is defined over selected units, not over feature-specific selections.",
          call. = FALSE
        )
      }

      rel_name <- as.character(t$relation_name %||% "boundary")[1]
      rel_model <- base_superset$data$spatial_relations_model[[rel_name]] %||%
        base_superset$data$spatial_relations[[rel_name]]

      if (is.null(rel_model)) {
        stop("Missing relation '", rel_name, "'.", call. = FALSE)
      }

      rcpp_prepare_objective_min_fragmentation(op, rel_model)

      rcpp_add_objective_min_fragmentation(
        op,
        relation_data = rel_model,
        weight_multiplier = as.numeric(t$weight_multiplier %||% 1)[1]
      )

    } else if (identical(type, "benefit")) {

      de_sub <- .subset_dist_effects(
        df = base_superset$data$dist_effects_model,
        actions = t$actions,
        features = t$features
      )

      da_sub <- .subset_dist_actions(actions = t$actions)

      if (is.null(de_sub) || !inherits(de_sub, "data.frame")) {
        stop("dist_effects_model is missing for benefit objective.", call. = FALSE)
      }
      for (nm in c("internal_pu", "internal_action", "benefit")) {
        if (!(nm %in% names(de_sub))) {
          stop("dist_effects_model must contain column '", nm, "' for benefit objective.", call. = FALSE)
        }
      }

      prep <- rcpp_prepare_objective_max_benefit(
        x = op,
        dist_actions_data = da_sub,
        dist_effects_data = de_sub,
        block_name = "prepare_max_benefit",
        tag = "from=.pamo_objvec_from_ir"
      )

      rcpp_add_objective_max_benefit(
        x = op,
        coef_x = as.numeric(prep$coef_x),
        weight = 1.0,
        block_name = "objective_add_max_benefit",
        tag = "from=.pamo_objvec_from_ir"
      )

    } else if (identical(type, "loss")) {

      de_sub <- .subset_dist_effects(
        df = base_superset$data$dist_effects_model,
        actions = t$actions,
        features = t$features
      )

      da_sub <- .subset_dist_actions(actions = t$actions)

      if (is.null(de_sub) || !inherits(de_sub, "data.frame")) {
        stop("dist_effects_model is missing for loss objective.", call. = FALSE)
      }
      for (nm in c("internal_pu", "internal_action", "loss")) {
        if (!(nm %in% names(de_sub))) {
          stop("dist_effects_model must contain column '", nm, "' for loss objective.", call. = FALSE)
        }
      }

      prep <- rcpp_prepare_objective_min_loss(
        x = op,
        dist_actions_data = da_sub,
        dist_effects_data = de_sub,
        block_name = "prepare_min_loss",
        tag = "from=.pamo_objvec_from_ir"
      )

      rcpp_add_objective_min_loss(
        x = op,
        coef_x = as.numeric(prep$coef_x),
        weight = 1.0,
        block_name = "objective_add_min_loss",
        tag = "from=.pamo_objvec_from_ir"

      )

    } else if (identical(type, "profit")) {

      pcol <- as.character(t$profit_col %||% "profit")[1]

      dp_sub <- .subset_dist_profit(
        df = base_superset$data$dist_profit_model,
        actions = t$actions
      )

      da_sub <- .subset_dist_actions(actions = t$actions)

      rcpp_add_objective_max_profit(
        op,
        dist_actions_data = da_sub,
        dist_profit_data  = dp_sub,
        profit_col = pcol,
        weight = 1.0
      )

    } else if (identical(type, "net_profit")) {

      pcol    <- as.character(t$profit_col %||% "profit")[1]
      inc_pu  <- isTRUE(t$include_pu_cost %||% TRUE)
      inc_act <- isTRUE(t$include_action_cost %||% TRUE)

      feats <- .chr(t$features)
      if (!is.null(feats) && isTRUE(inc_pu)) {
        stop(
          "Subset by features is not currently supported for 'net_profit' when include_pu_cost=TRUE. ",
          "PU costs are attached to w variables and cannot yet be attributed to feature subsets.",
          call. = FALSE
        )
      }

      da_sub <- .subset_dist_actions(actions = t$actions)
      dp_sub <- .subset_dist_profit(
        df = base_superset$data$dist_profit_model,
        actions = t$actions
      )

      rcpp_add_objective_max_net_profit(
        op,
        pu_data = base_superset$data$pu,
        dist_actions_data = da_sub,
        dist_profit_data  = dp_sub,
        profit_col = pcol,
        include_pu_cost = inc_pu,
        include_action_cost = inc_act,
        weight = 1.0
      )

    } else if (identical(type, "action_boundary_cut")) {

      rel_name <- as.character(t$relation_name %||% "boundary")[1]
      rel_model <- base_superset$data$spatial_relations_model[[rel_name]] %||%
        base_superset$data$spatial_relations[[rel_name]]

      if (is.null(rel_model)) {
        stop("Missing relation '", rel_name, "'.", call. = FALSE)
      }

      act_int <- .resolve_action_internal(t$actions)

      rcpp_prepare_objective_min_fragmentation_actions(
        op,
        dist_actions_data = base_superset$data$dist_actions_model,
        relation_data = rel_model,
        actions_to_use = if (length(act_int) > 0) act_int else NULL,
        block_name = "fragmentation_actions",
        tag = "from=.pamo_objvec_from_ir"
      )

      term2 <- t
      term2$actions <- if (length(act_int) > 0) as.integer(act_int) else NULL

      v_term <- .pamo_objvec_action_boundary_cut(base_superset, term2)

      if (length(v_term) == 0L) {
        stop("Empty vector returned by .pamo_objvec_action_boundary_cut().", call. = FALSE)
      }

      if (!exists("rcpp_model_set_objective_vector", mode = "function")) {
        m0 <- .pa_model_from_ptr(
          op,
          args = base_superset$data$model_args %||% list(),
          drop_triplets = TRUE
        )
        obj0 <- as.numeric(m0$obj)
        if (length(obj0) != length(v_term)) {
          stop("Current model objective length does not match action_boundary_cut term vector.", call. = FALSE)
        }
        rcpp_model_set_objective_vector(
          x = op,
          obj = obj0 + v_term,
          model_sense = "min"
        )
      } else {
        m0 <- .pa_model_from_ptr(
          op,
          args = base_superset$data$model_args %||% list(),
          drop_triplets = TRUE
        )
        obj0 <- as.numeric(m0$obj)
        if (length(obj0) != length(v_term)) {
          stop("Current model objective length does not match action_boundary_cut term vector.", call. = FALSE)
        }
        rcpp_model_set_objective_vector(
          x = op,
          obj = obj0 + v_term,
          model_sense = "min"
        )
      }

    } else if (identical(type, "intervention_impact")) {

      icol <- as.character(t$impact_col %||% "amount")[1]

      df_sub <- .subset_dist_features(
        df = base_superset$data$dist_features,
        features = t$features
      )

      feat_int <- .resolve_feature_internal(t$features)
      act_int  <- .resolve_action_internal(t$actions)

      if (length(act_int) == 0) {
        stop(
          "intervention_impact requires a non-empty action subset.",
          call. = FALSE
        )
      }

      subset_key <- .pa_subset_to_string(as.character(t$actions))

      rcpp_add_objective_min_intervention_impact(
        op,
        pu_data = base_superset$data$pu,
        dist_features_data = df_sub,
        dist_actions_data = base_superset$data$dist_actions_model,
        subset_key = subset_key,
        impact_col = icol,
        features_to_use = as.integer(feat_int),
        actions_to_use = as.integer(act_int),
        internal_feature_col = "internal_feature",
        weight = 1.0,
        weight_multiplier = as.numeric(t$weight_multiplier %||% 1)[1],
        block_name = "objective_add_min_intervention_impact",
        tag = "from=.pamo_objvec_from_ir"
      )

    } else if (identical(type, "custom")) {
      stop("custom objectives are not supported in weighted yet.", call. = FALSE)

    } else {
      stop("Unknown term type in .pamo_objvec_from_ir(): ", type, call. = FALSE)
    }
  }

  m <- .pa_model_from_ptr(
    op,
    args = base_superset$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  as.numeric(m$obj)
}



.pamo_apply_weighted_objective <- function(base,
                                           ir_list,
                                           weights,
                                           objective_scaling = FALSE,
                                           scales = NULL) {
  stopifnot(inherits(base, "Problem"))
  stopifnot(is.list(ir_list), length(ir_list) > 0)
  weights <- as.numeric(weights)
  if (length(weights) != length(ir_list)) {
    stop("weights length must match ir_list length.", call. = FALSE)
  }

  # 1) construir superset UNA vez (incluye prepares necesarios)
  base2 <- .pamo_prepare_superset_model(base, ir_list)

  # 2) construir objvec para cada IR sobre el MISMO superset (sin rebuild)
  objvecs <- vector("list", length(ir_list))
  for (i in seq_along(ir_list)) {
    v <- .pamo_objvec_from_ir(base2, ir_list[[i]])

    # convención: todo se resuelve como MIN en el motor;
    # si IR es "max", flipa signo acá.
    if (identical(ir_list[[i]]$sense, "max")) v <- -v

    objvecs[[i]] <- v
  }

  # 3) objective scaling by payoff range
  if (isTRUE(objective_scaling)) {
    if (is.null(scales) || length(scales) != length(objvecs)) {
      stop(
        "objective_scaling = TRUE requires one positive scale per objective.",
        call. = FALSE
      )
    }

    objvecs <- Map(
      function(v, s) {
        s <- as.numeric(s)[1]
        if (!is.finite(s) || s <= 0) {
          warning("Encountered non-positive scale in weighted objective scaling; leaving objective unscaled.",
                  call. = FALSE)
          return(v)
        }
        v / s
      },
      objvecs,
      as.list(scales)
    )
  }

  # 4) combinar pesos -> objetivo final
  obj_w <- Reduce(`+`, Map(`*`, objvecs, as.list(weights)))

  # 5) IMPORTANTÍSIMO: inyectar runtime update para que multiscape lo use en solve()
  base2$data$runtime_updates <- list(
    obj = obj_w,
    modelsense = "min"
  )

  # 6) asegurar que multiscape NO reconstruya el modelo y no te borre runtime_updates
  base2$data$meta$model_dirty <- FALSE
  base2$data$has_model <- TRUE

  # 7) guarda cache para evaluación posterior (opcional pero muy útil)
  base2$data$mo_cache <- list(
    ir_list = ir_list,
    weights = weights,
    objective_scaling = objective_scaling,
    scales = scales,
    objvecs = objvecs,
    obj_weighted = obj_w
  )

  base2
}

# ---------------------------------------------------------
# Internal: solve a single run using multiscape as engine
# ---------------------------------------------------------
.pamo_solve_one <- function(x, spec) {

  if (!inherits(x, "Problem")) stop("Problem$base must be a Problem.", call. = FALSE)

  base <- .pamo_clone_base(x)

  if (identical(spec$type, "weighted")) {

    aliases <- as.character(spec$aliases)
    weights <- as.numeric(spec$weights)
    objective_scaling <- isTRUE(spec$objective_scaling)
    scales <- spec$scales %||% NULL

    specs <- .pamo_get_objective_specs(x, aliases)
    ir_list <- lapply(specs, function(sp) .pamo_objective_to_ir(base, sp))

    base <- .pamo_apply_weighted_objective(
      base = base,
      ir_list = ir_list,
      weights = weights,
      objective_scaling = objective_scaling,
      scales = scales
    )

  } else if (identical(spec$type, "epsilon_constraint")) {

    primary     <- as.character(spec$primary)[1]
    eps_list    <- spec$eps %||% list()
    eps_tol     <- spec$eps_tol %||% list()
    sec_aliases <- names(eps_list)

    if (is.na(primary) || !nzchar(primary)) stop("epsilon_constraint: primary is invalid.", call. = FALSE)
    if (length(sec_aliases) == 0) stop("epsilon_constraint: eps list is empty.", call. = FALSE)

    # ---- build IRs for (primary + constrained) so we can build ONE superset
    all_aliases <- unique(c(primary, sec_aliases))
    specs_all <- .pamo_get_objective_specs(x, all_aliases)
    ir_all <- lapply(specs_all, function(sp) .pamo_objective_to_ir(base, sp))

    # ---- 0) build superset once (includes prepares)
    base <- .pamo_prepare_superset_model(base, ir_all)

    # ---- 1) add epsilon constraints for secondaries
    for (a in sec_aliases) {
      sp_sec <- specs_all[[a]]
      ir_sec <- ir_all[[a]]

      eps_val <- as.numeric(eps_list[[a]])[1]
      if (!is.finite(eps_val)) {
        stop("epsilon_constraint: eps for '", a, "' must be finite.", call. = FALSE)
      }

      tol_a <- as.numeric(eps_tol[[a]] %||% 0)[1]
      if (!is.finite(tol_a) || tol_a < 0) {
        stop("epsilon_constraint: eps_tol for '", a, "' must be finite and >= 0.", call. = FALSE)
      }

      base <- .pamo_apply_epsilon_constraint(
        base = base,
        ir = ir_sec,
        eps = eps_val + tol_a,
        sense = as.character(sp_sec$sense %||% "min")[1],
        name = paste0("eps_", a),
        block_name = "epsilon_constraint"
      )
    }

    # ---- 2) set primary objective (as runtime update, like weighted)
    sp_primary <- specs_all[[primary]]
    ir_primary <- ir_all[[primary]]

    base <- .pamo_apply_single_objective(
      base = base,
      ir = ir_primary,
      sense = as.character(sp_primary$sense %||% "min")[1]
    )

  } else if (identical(spec$type, "augmecon")) {

    primary      <- as.character(spec$primary)[1]
    eps_list     <- spec$eps %||% list()
    eps_tol      <- spec$eps_tol %||% list()
    sec_aliases  <- names(eps_list)
    sec_ranges   <- as.numeric(spec$secondary_ranges %||% numeric(0))
    augmentation <- as.numeric(spec$augmentation %||% NA_real_)[1]
    slack_upper_bound <- as.numeric((x$data$method %||% list())$slack_upper_bound %||% 1e6)[1]

    if (is.na(primary) || !nzchar(primary)) {
      stop("augmecon: primary is invalid.", call. = FALSE)
    }
    if (length(sec_aliases) == 0L) {
      stop("augmecon: eps list is empty.", call. = FALSE)
    }
    if (!is.finite(augmentation) || augmentation <= 0) {
      stop("augmecon: augmentation must be a finite positive number.", call. = FALSE)
    }
    if (!is.finite(slack_upper_bound) || slack_upper_bound <= 0) {
      stop("augmecon: slack_upper_bound must be a finite positive number.", call. = FALSE)
    }
    if (length(sec_ranges) != length(sec_aliases)) {
      stop(
        "augmecon: secondary_ranges must have one value per secondary objective.\n",
        "length(secondary_ranges) = ", length(sec_ranges),
        ", length(sec_aliases) = ", length(sec_aliases), ".",
        call. = FALSE
      )
    }
    if (any(!is.finite(sec_ranges)) || any(sec_ranges <= 0)) {
      stop("augmecon: secondary_ranges must be finite and > 0.", call. = FALSE)
    }
    names(sec_ranges) <- sec_aliases

    # ---- build IRs for primary + secondaries
    all_aliases <- unique(c(primary, sec_aliases))
    specs_all <- .pamo_get_objective_specs(x, all_aliases)
    ir_all <- lapply(specs_all, function(sp) .pamo_objective_to_ir(base, sp))

    # ---- build one superset
    base <- .pamo_prepare_superset_model(base, ir_all)

    # ---- add explicit slack variables (one per secondary objective)
    base <- .pamo_add_augmecon_slacks(
      base = base,
      secondary_aliases = sec_aliases,
      ub = slack_upper_bound
    )

    slack_cols <- base$data$mo_cache$augmecon$slack_cols_0based[sec_aliases]
    if (length(slack_cols) != length(sec_aliases)) {
      stop("augmecon: failed to recover slack columns for all secondary objectives.", call. = FALSE)
    }

    # ---- add equality constraints with slack: g_j(x) + s_j = eps_j
    for (j in seq_along(sec_aliases)) {
      a <- sec_aliases[j]
      sp_sec <- specs_all[[a]]
      ir_sec <- ir_all[[a]]

      eps_val <- as.numeric(eps_list[[a]])[1]
      if (!is.finite(eps_val)) {
        stop("augmecon: eps for '", a, "' must be finite.", call. = FALSE)
      }

      tol_a <- as.numeric(eps_tol[[a]] %||% 0)[1]
      if (!is.finite(tol_a) || tol_a < 0) {
        stop("augmecon: eps_tol for '", a, "' must be finite and >= 0.", call. = FALSE)
      }

      base <- .pamo_apply_augmecon_constraint(
        base = base,
        ir = ir_sec,
        eps = eps_val + tol_a,
        slack_col = as.integer(slack_cols[[a]]),
        sense = as.character(sp_sec$sense %||% "min")[1],
        name = paste0("augmecon_", a),
        block_name = "augmecon_constraint"
      )
    }

    # ---- primary objective + normalized slack rewards
    ir_primary <- ir_all[[primary]]
    slack_weights <- augmentation / sec_ranges

    base <- .pamo_apply_augmecon_objective(
      base = base,
      ir_primary = ir_primary,
      slack_cols = as.integer(unname(slack_cols)),
      slack_weights = as.numeric(slack_weights)
    )
  } else {
    stop("Unsupported spec$type in .pamo_solve_one: ", spec$type, call. = FALSE)
  }

  gap_limit  <- spec$gap_limit %||% NULL
  time_limit <- spec$time_limit %||% NULL

  # base$data$method <- NULL
  # base$data$results <- NULL
  # base$data$runtime_updates <- NULL

  out <- .pa_solve_single_problem(
    base,
    gap_limit = gap_limit,
    time_limit = time_limit
  )

  .pamo_extract_solution(out)
}

.pamo_apply_epsilon_constraint <- function(base, ir, eps, sense = c("min","max"),
                                           name = "", block_name = "epsilon_constraint", tag = "") {
  stopifnot(inherits(base, "Problem"))
  sense <- match.arg(sense)

  if (is.null(base$data$model_ptr)) stop("Model not built (model_ptr is NULL).", call. = FALSE)

  eps <- as.numeric(eps)[1]
  if (!is.finite(eps)) stop("eps must be finite.", call. = FALSE)

  v <- .pamo_objvec_from_ir(base, ir)
  if (length(v) == 0) stop("epsilon objvec is empty.", call. = FALSE)

  # canonical: <=
  if (identical(sense, "max")) {
    v <- -v
    eps <- -eps
  }

  idx <- which(v != 0)
  if (length(idx) == 0) stop("epsilon objvec has no non-zero coefficients.", call. = FALSE)

  # add row: sum(v[j]*x[j]) <= eps

  rcpp_add_linear_constraint(
    base$data$model_ptr,
    j0 = as.integer(idx - 1L),
    x  = as.numeric(v[idx]),
    sense = "<=",
    rhs = as.numeric(eps),
    name = as.character(name %||% ""),
    block_name = block_name,
    tag = tag
  )

  base$data$meta$model_dirty <- TRUE
  base <- .pa_refresh_model_snapshot(base)
  base
}

# ---------------------------------------------------------
# Internal: extract (minimal) results from multiscape Solution
# ---------------------------------------------------------
.pamo_extract_solution <- function(out) {
  objval <- tryCatch(out$solution$objective, error = function(e) NA_real_)
  gap <- tryCatch(out$diagnostics$gap, error = function(e) NA_real_)
  runtime <- tryCatch(out$diagnostics$runtime, error = function(e) NA_real_)
  status <- tryCatch(getStatus(out), error = function(e) "unknown")

  list(
    solution = out,
    status = status,
    runtime = runtime,
    gap = gap,
    objval = objval
  )
}

.pamo_solver_error_status <- function(e) {
  msg <- conditionMessage(e)

  if (grepl(
    "status:\\s*infeasible_or_unbounded|\\binfeasible_or_unbounded\\b",
    msg,
    ignore.case = TRUE
  )) {
    return("infeasible_or_unbounded")
  }

  if (grepl(
    "status:\\s*infeasible|\\binfeasible\\b",
    msg,
    ignore.case = TRUE
  )) {
    return("infeasible")
  }

  if (grepl(
    "empty solution vector|no solution vector|solution vector is empty|objval:\\s*NA|time limit before first feasible",
    msg,
    ignore.case = TRUE
  )) {
    return("no_solution")
  }

  NA_character_
}


.pamo_failed_run <- function(status = "failed", message = NULL) {
  list(
    solution = NULL,
    status = status,
    runtime = NA_real_,
    gap = NA_real_,
    objval = NA_real_,
    message = as.character(message %||% NA_character_)
  )
}


.pamo_solve_one_or_keep_failed <- function(
    x,
    spec,
    stop_on_infeasible = FALSE,
    stop_on_no_solution = FALSE,
    stop_on_error = TRUE
) {
  tryCatch(
    .pamo_solve_one(x = x, spec = spec),
    error = function(e) {
      msg <- conditionMessage(e)
      status <- .pamo_solver_error_status(e)

      if (identical(status, "infeasible") ||
          identical(status, "infeasible_or_unbounded")) {

        if (isTRUE(stop_on_infeasible)) {
          stop(e)
        }

        return(.pamo_failed_run(
          status = status,
          message = msg
        ))
      }

      if (identical(status, "no_solution")) {

        if (isTRUE(stop_on_no_solution)) {
          stop(e)
        }

        return(.pamo_failed_run(
          status = status,
          message = msg
        ))
      }

      if (isTRUE(stop_on_error)) {
        stop(e)
      }

      .pamo_failed_run(
        status = "failed",
        message = msg
      )
    }
  )
}


#' @keywords internal
.mo_abort <- function(...) stop(paste0(...), call. = FALSE)

#' @keywords internal
.mo_get_solution_from <- function(x, run = 1L) {

  run <- as.integer(run)[1]
  if (!is.finite(run) || is.na(run) || run < 1L) {
    .mo_abort("run must be a positive integer (1-based).")
  }

  # Case 1: already a Solution
  if (inherits(x, "Solution")) {
    return(x)
  }

  # Case 2: SolutionSet -> x$solution$solutions[[run]]
  if (inherits(x, "SolutionSet")) {
    sols <- x$solution$solutions %||% NULL

    if (is.null(sols)) {
      .mo_abort("SolutionSet has no solutions (x$solution$solutions is NULL).")
    }
    if (!is.list(sols) || length(sols) == 0L) {
      .mo_abort("SolutionSet contains an empty solutions list.")
    }
    if (run > length(sols)) {
      .mo_abort("run=", run, " out of range. There are only ", length(sols), " solutions.")
    }

    sol <- sols[[run]]
    if (!inherits(sol, "Solution")) {
      .mo_abort("x$solution$solutions[[run]] is not a Solution.")
    }

    return(sol)
  }

  # Case 3: plain list of Solution (optional internal convenience)
  if (is.list(x) && length(x) > 0L && inherits(x[[1]], "Solution")) {
    if (run > length(x)) {
      .mo_abort("run=", run, " out of range. There are only ", length(x), " solutions.")
    }
    return(x[[run]])
  }

  .mo_abort(
    "Unsupported object. Expected a Solution or SolutionSet."
  )
}



# -------------------------------------------------------------------------
# Internal: extract raw decision vector from a multiscape Solution
# -------------------------------------------------------------------------
.pamo_get_solution_vector <- function(sol) {

  if (!is.null(sol$solution$vector) && is.numeric(sol$solution$vector) && length(sol$solution$vector) > 0) {
    return(as.numeric(sol$solution$vector))
  }

  # fallback defensivo por si entra una lista rara interna
  candidates <- c("vector", "solution", "sol", "x", "best_solution", "solution_vector")

  d <- sol$solution %||% sol

  for (nm in candidates) {
    v <- d[[nm]] %||% NULL
    if (is.numeric(v) && length(v) > 0) {
      return(as.numeric(v))
    }
  }

  stop(
    "Could not extract the raw decision vector from the solution object.\n",
    "Expected x$solution$vector or an equivalent numeric field.",
    call. = FALSE
  )
}




.pamo_eval_boundary_cut_on_solution <- function(x, solution, term) {
  stopifnot(inherits(x, "Problem"))

  rel_name <- as.character(term$relation_name %||% "boundary")[1]
  rel <- x$data$spatial_relations[[rel_name]] %||% NULL

  if (is.null(rel) || !inherits(rel, "data.frame")) {
    stop("Missing spatial relation '", rel_name, "' while evaluating boundary_cut.", call. = FALSE)
  }

  for (nm in c("internal_pu1", "internal_pu2", "weight")) {
    if (!nm %in% names(rel)) {
      stop("Spatial relation '", rel_name, "' must contain column '", nm, "'.", call. = FALSE)
    }
  }

  sol_vec <- .pamo_get_solution_vector(solution)

  n_pu <- nrow(x$data$pu)
  if (!is.finite(n_pu) || is.na(n_pu) || n_pu <= 0L) {
    stop("Could not determine n_pu while evaluating boundary_cut.", call. = FALSE)
  }

  if (length(sol_vec) < n_pu) {
    stop(
      "Solution vector is shorter than n_pu while evaluating boundary_cut.\n",
      "length(sol_vec) = ", length(sol_vec), ", n_pu = ", n_pu, ".",
      call. = FALSE
    )
  }

  # Phase 1/2A assumption: w occupies the first n_pu decision variables
  w <- as.numeric(sol_vec[seq_len(n_pu)])
  w_bin <- as.numeric(w > 0.5)

  ip1 <- as.integer(rel$internal_pu1)
  ip2 <- as.integer(rel$internal_pu2)
  ww  <- as.numeric(rel$weight)

  ok <- is.finite(ip1) & is.finite(ip2) & is.finite(ww)
  ip1 <- ip1[ok]
  ip2 <- ip2[ok]
  ww  <- ww[ok]

  if (length(ip1) == 0L) return(0)

  # self weights (diagonal rows) and canonical off-diagonal edges
  self_w <- numeric(n_pu)
  edge_map <- new.env(parent = emptyenv(), hash = TRUE)

  make_key <- function(a, b) paste0(a, "::", b)

  for (r in seq_along(ip1)) {
    i <- ip1[r]
    j <- ip2[r]
    wij <- ww[r]

    if (i == j) {
      self_w[i] <- self_w[i] + wij
    } else {
      a <- min(i, j)
      b <- max(i, j)
      k <- make_key(a, b)
      old <- if (exists(k, envir = edge_map, inherits = FALSE)) {
        get(k, envir = edge_map, inherits = FALSE)
      } else {
        NA_real_
      }
      if (is.na(old) || wij > old) {
        assign(k, wij, envir = edge_map)
      }
    }
  }

  edge_keys <- ls(edge_map, all.names = TRUE)

  incident <- numeric(n_pu)
  edge_sum <- 0

  for (k in edge_keys) {
    parts <- strsplit(k, "::", fixed = TRUE)[[1]]
    a <- as.integer(parts[1])
    b <- as.integer(parts[2])
    wij <- get(k, envir = edge_map, inherits = FALSE)

    incident[a] <- incident[a] + wij
    incident[b] <- incident[b] + wij

    edge_sum <- edge_sum + wij * w_bin[a] * w_bin[b]
  }

  linear_sum <- sum((incident + self_w) * w_bin)
  val <- linear_sum - 2 * edge_sum

  mult <- as.numeric(term$weight_multiplier %||% 1)[1]
  if (!is.finite(mult)) mult <- 1

  as.numeric(mult * val)
}



# -------------------------------------------------------------------------
# Internal: evaluate one registered objective alias on a solved solution
# Uses the same IR/objective-vector machinery, so evaluation is consistent
# with the MO engine.
# -------------------------------------------------------------------------

.pamo_get_alias_value_from_solution <- function(x, solution, alias) {
  stopifnot(inherits(x, "Problem"))
  .pamo_eval_alias_on_solution(x, solution, alias)
}

.pamo_eval_alias_on_solution <- function(x, solution, alias) {
  stopifnot(inherits(x, "Problem"))

  alias <- as.character(alias)[1]
  if (is.na(alias) || !nzchar(alias)) {
    stop("alias must be a non-empty string.", call. = FALSE)
  }

  spec <- .pamo_get_objective_spec(x, alias)
  ir   <- .pamo_objective_to_ir(x, spec)

  terms <- ir$terms %||% list()

  # Fase 1:
  # si el objetivo es exactamente min_fragmentation (PU boundary cut),
  # lo evaluamos directamente desde w sin reconstruir auxiliares y_pu.
  if (length(terms) == 1L && identical(terms[[1]]$type %||% "", "boundary_cut")) {
    val <- .pamo_eval_boundary_cut_on_solution(x, solution, terms[[1]])
    return(as.numeric(val))
  }

  if (length(terms) == 1L && identical(terms[[1]]$type %||% "", "intervention_impact")) {
    val <- .pamo_eval_intervention_impact_on_solution(x, solution, terms[[1]])
    return(as.numeric(val))
  }

  # if (length(terms) == 1L && identical(terms[[1]]$type %||% "", "action_boundary_cut")) {
  #   val <- .pamo_eval_action_boundary_cut_on_solution(x, solution, terms[[1]])
  #   return(as.numeric(val))
  # }

  # resto de objetivos: evaluación por objvec
  base_eval <- .pamo_prepare_superset_model(x, list(ir))

  obj_vec <- .pamo_objvec_from_ir(base_eval, ir)
  sol_vec <- .pamo_get_solution_vector(solution)

  n_obj <- length(obj_vec)
  n_sol <- length(sol_vec)

  if (n_obj > n_sol) {
    stop(
      "Objective vector length does not match solution vector length when evaluating alias '", alias, "'.\n",
      "length(obj_vec) = ", n_obj, ", length(sol_vec) = ", n_sol, ".",
      call. = FALSE
    )
  }

  # permitir evaluar objetivos 'base' sobre soluciones de un superset
  sol_use <- as.numeric(sol_vec[seq_len(n_obj)])

  val_engine <- sum(as.numeric(obj_vec) * sol_use, na.rm = TRUE)

  as.numeric(val_engine)
}

.pamo_add_alias_upper_bound_constraint <- function(base_eval, x, alias, rhs, tol = 0, name = NULL) {
  stopifnot(inherits(base_eval, "Problem"))
  stopifnot(inherits(x, "Problem"))

  alias <- as.character(alias)[1]
  rhs   <- as.numeric(rhs)[1]
  tol   <- as.numeric(tol)[1]

  if (!is.finite(rhs)) {
    stop("rhs must be finite in .pamo_add_alias_upper_bound_constraint().", call. = FALSE)
  }
  if (!is.finite(tol) || tol < 0) {
    stop("tol must be a finite non-negative number.", call. = FALSE)
  }

  spec <- .pamo_get_objective_spec(x, alias)
  ir   <- .pamo_objective_to_ir(x, spec)

  obj_vec <- .pamo_objvec_from_ir(base_eval, ir)

  op_list <- .pa_model_from_ptr(
    base_eval$data$model_ptr,
    args = base_eval$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  n_model <- as.integer(op_list$ncol %||% length(obj_vec))
  if (length(obj_vec) != n_model) {
    stop(
      "Cannot add alias bound for '", alias, "': objective vector length does not match model dimension.\n",
      "length(obj_vec) = ", length(obj_vec), ", model dimension = ", n_model, ".",
      call. = FALSE
    )
  }

  nz <- which(abs(obj_vec) > 0)
  if (length(nz) == 0L) {
    stop("Alias '", alias, "' produced an empty objective vector.", call. = FALSE)
  }

  if (is.null(name)) {
    name <- paste0("eps_bound_", alias)
  }

  base_eval <- .pa_add_linear_constraint(
    base_eval,
    var_index_0based = as.integer(nz - 1L),
    coeff = as.numeric(obj_vec[nz]),
    sense = "<=",
    rhs = as.numeric(rhs + tol),
    name = name
  )

  base_eval
}

# -------------------------------------------------------------------------
# Internal: compute 2-objective extreme points for auto epsilon mode
#
# Returns:
# - primary_at_primary
# - secondary_at_primary
# - primary_at_secondary
# - secondary_at_secondary
# - secondary_min
# - secondary_max
# -------------------------------------------------------------------------
.pamo_compute_epsilon_extremes_2obj <- function(x, primary, secondary, gap_limit = NULL, time_limit = NULL) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% list()
  do_lexi <- isTRUE(method$lexicographic)
  lexi_tol <- as.numeric(method$lexicographic_tol %||% 0)[1]
  if (!is.finite(lexi_tol) || lexi_tol < 0) lexi_tol <- 0

  specs_all <- .pamo_get_specs(x)
  obj_alias <- names(specs_all)

  if (!primary %in% obj_alias) {
    stop("primary alias not found: '", primary, "'.", call. = FALSE)
  }
  if (!secondary %in% obj_alias) {
    stop("secondary alias not found: '", secondary, "'.", call. = FALSE)
  }

  # ------------------------------------------------------------
  # Caso simple (no lexicográfico)
  # ------------------------------------------------------------
  if (!isTRUE(do_lexi)) {

    # Extreme 1: optimize primary
    sol_primary <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = primary,
        weights = 1,
        objective_scaling = FALSE,
        scales = NULL
      )
    )

    primary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary$solution, primary)
    secondary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary$solution, secondary)

    # Extreme 2: optimize secondary
    sol_secondary <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = secondary,
        weights = 1,
        normalize = FALSE,
        gap_limit = gap_limit,
        time_limit = time_limit
      )
    )

    primary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary$solution, primary)
    secondary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary$solution, secondary)

    vals_secondary <- c(secondary_at_primary, secondary_at_secondary)

    if (any(!is.finite(vals_secondary))) {
      stop(
        "Could not compute finite extreme values for the secondary objective.\n",
        "secondary_at_primary = ", secondary_at_primary, "\n",
        "secondary_at_secondary = ", secondary_at_secondary,
        call. = FALSE
      )
    }

    return(list(
      primary = primary,
      secondary = secondary,

      primary_at_primary = as.numeric(primary_at_primary),
      secondary_at_primary = as.numeric(secondary_at_primary),

      primary_at_secondary = as.numeric(primary_at_secondary),
      secondary_at_secondary = as.numeric(secondary_at_secondary),

      secondary_min = min(vals_secondary),
      secondary_max = max(vals_secondary),

      solution_primary = sol_primary$solution,
      solution_secondary = sol_secondary$solution
    ))
  }

  # ------------------------------------------------------------
  # Caso lexicográfico
  # ------------------------------------------------------------

  # 1) resolver primary puro
  sol_primary_stage1 <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "weighted",
      aliases = primary,
      weights = 1,
      normalize = FALSE,
      gap_limit = gap_limit,
      time_limit = time_limit
    )
  )

  primary_star <- .pamo_eval_alias_on_solution(x, sol_primary_stage1$solution, primary)

  # 2) resolver secondary sujeto a primary <= primary_star + tol
  sol_primary_lexi <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "epsilon_constraint",
      primary = secondary,
      eps = stats::setNames(list(primary_star), primary),
      gap_limit = gap_limit,
      time_limit = time_limit,
      eps_tol = stats::setNames(list(lexi_tol), primary)
    )
  )

  primary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary_lexi$solution, primary)
  secondary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary_lexi$solution, secondary)

  # 3) resolver secondary puro
  sol_secondary_stage1 <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "weighted",
      aliases = secondary,
      weights = 1,
      normalize = FALSE,
      gap_limit = gap_limit,
      time_limit = time_limit
    )
  )

  secondary_star <- .pamo_eval_alias_on_solution(x, sol_secondary_stage1$solution, secondary)

  # 4) resolver primary sujeto a secondary <= secondary_star + tol
  sol_secondary_lexi <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "epsilon_constraint",
      primary = primary,
      eps = stats::setNames(list(secondary_star), secondary),
      gap_limit = gap_limit,
      time_limit = time_limit,
      eps_tol = stats::setNames(list(lexi_tol), secondary)
    )
  )

  primary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary_lexi$solution, primary)
  secondary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary_lexi$solution, secondary)

  vals_secondary <- c(secondary_at_primary, secondary_at_secondary)

  if (any(!is.finite(vals_secondary))) {
    stop(
      "Could not compute finite extreme values for the secondary objective.\n",
      "secondary_at_primary = ", secondary_at_primary, "\n",
      "secondary_at_secondary = ", secondary_at_secondary,
      call. = FALSE
    )
  }

  list(
    primary = primary,
    secondary = secondary,

    primary_at_primary = as.numeric(primary_at_primary),
    secondary_at_primary = as.numeric(secondary_at_primary),

    primary_at_secondary = as.numeric(primary_at_secondary),
    secondary_at_secondary = as.numeric(secondary_at_secondary),

    secondary_min = min(vals_secondary),
    secondary_max = max(vals_secondary),
    vals_secondary = vals_secondary,

    solution_primary = sol_primary_lexi$solution,
    solution_secondary = sol_secondary_lexi$solution,

    meta = list(
      lexicographic = TRUE,
      lexicographic_tol = lexi_tol,
      stage1_primary_solution = sol_primary_stage1$solution,
      stage1_secondary_solution = sol_secondary_stage1$solution
    )
  )
}


.pamo_compile_superset_spec <- function(base, ir_list) {
  stopifnot(inherits(base, "Problem"))
  stopifnot(is.list(ir_list), length(ir_list) > 0)

  all_terms <- unlist(
    lapply(ir_list, function(ir) ir$terms %||% list()),
    recursive = FALSE
  )

  all_types <- unique(vapply(all_terms, `[[`, character(1), "type"))

  need_z     <- "representation" %in% all_types
  need_y_pu  <- "boundary_cut" %in% all_types
  need_y_act <- "action_boundary_cut" %in% all_types
  need_y_int <- "intervention_boundary_cut" %in% all_types
  need_u_int <- "intervention_impact" %in% all_types

  if (isTRUE(need_y_int)) {
    stop(
      "Phase 1 superset does not yet support 'intervention_boundary_cut' objectives.\n",
      "Please remove intervention fragmentation from the MO method for now.",
      call. = FALSE
    )
  }

  # ------------------------------------------------------------
  # Spatial fragmentation terms: at most one relation_name in phase 1
  # ------------------------------------------------------------
  spatial_terms <- Filter(
    function(t) t$type %in% c("boundary_cut", "action_boundary_cut", "intervention_boundary_cut"),
    all_terms
  )

  rel_names <- unique(vapply(
    spatial_terms,
    function(t) as.character(t$relation_name %||% "boundary")[1],
    character(1)
  ))

  rel_names <- rel_names[!is.na(rel_names) & nzchar(rel_names)]

  if (length(rel_names) > 1L) {
    stop(
      "Phase 1 superset supports only one relation_name across fragmentation objectives.\n",
      "Found: ", paste(rel_names, collapse = ", "),
      call. = FALSE
    )
  }

  # ------------------------------------------------------------
  # intervention_impact: phase 1 supported ONLY if all terms
  # share the same action subset
  # ------------------------------------------------------------
  u_int_actions <- NULL

  if (isTRUE(need_u_int)) {
    int_terms <- Filter(function(t) identical(t$type, "intervention_impact"), all_terms)

    norm_actions <- function(x) {
      x <- as.character(x %||% character(0))
      x <- unique(x[!is.na(x) & nzchar(x)])
      sort(x)
    }

    action_keys <- vapply(
      int_terms,
      function(t) paste(norm_actions(t$actions), collapse = "||"),
      character(1)
    )

    uniq_keys <- unique(action_keys)

    if (length(uniq_keys) != 1L) {
      pretty <- vapply(
        int_terms,
        function(t) {
          aa <- norm_actions(t$actions)
          if (length(aa) == 0L) "<empty>" else paste(aa, collapse = ", ")
        },
        character(1)
      )

      stop(
        "Phase 1 superset supports 'intervention_impact' only when all such objectives ",
        "share the same action subset.\n",
        "Found subsets: ", paste(unique(pretty), collapse = " | "),
        call. = FALSE
      )
    }

    u_int_actions <- norm_actions(int_terms[[1]]$actions)

    if (length(u_int_actions) == 0L) {
      stop(
        "Phase 1 superset requires a non-empty action subset for 'intervention_impact'.",
        call. = FALSE
      )
    }
  }

  list(
    needs = list(
      z = isTRUE(need_z),
      y_pu = isTRUE(need_y_pu),
      y_action = isTRUE(need_y_act),
      y_intervention = FALSE,
      u_intervention = isTRUE(need_u_int),
      u_intervention_actions = u_int_actions
    ),
    relation_name = if (length(rel_names) == 1L) rel_names[1] else NULL
  )
}


.pamo_apply_single_objective <- function(base, ir, sense = c("min", "max")) {
  stopifnot(inherits(base, "Problem"))
  sense <- match.arg(sense)

  if (is.null(base$data$model_ptr)) {
    stop("Model not built (model_ptr is NULL).", call. = FALSE)
  }

  v <- .pamo_objvec_from_ir(base, ir)

  # El motor trabaja siempre en minimización
  if (identical(sense, "max")) {
    v <- -as.numeric(v)
  } else {
    v <- as.numeric(v)
  }

  # Si existe un setter C++ del objetivo, úsalo
  if (exists("rcpp_model_set_objective_vector", mode = "function")) {
    rcpp_model_set_objective_vector(
      x = base$data$model_ptr,
      obj = v,
      model_sense = "min"
    )
  } else {
    # fallback temporal: deja la actualización pendiente
    base$data$runtime_updates <- list(
      obj = v,
      modelsense = "min"
    )
  }

  base$data$meta <- base$data$meta %||% list()
  base$data$meta$model_dirty <- FALSE
  base$data$has_model <- TRUE

  base
}


.pamo_eval_action_boundary_cut_on_solution <- function(x, solution, term) {
  stopifnot(inherits(x, "Problem"))

  rel_name <- as.character(term$relation_name %||% "boundary")[1]
  sol_problem <- solution$problem %||% NULL

  rel <- sol_problem$data$spatial_relations_model[[rel_name]] %||%
    sol_problem$data$spatial_relations[[rel_name]] %||%
    x$data$spatial_relations_model[[rel_name]] %||%
    x$data$spatial_relations[[rel_name]] %||%
    NULL

  if (is.null(rel) || !inherits(rel, "data.frame")) {
    stop("Missing spatial relation '", rel_name, "' while evaluating action_boundary_cut.", call. = FALSE)
  }

  for (nm in c("internal_pu1", "internal_pu2", "weight")) {
    if (!nm %in% names(rel)) {
      stop("Spatial relation '", rel_name, "' must contain column '", nm, "'.", call. = FALSE)
    }
  }

  da <- sol_problem$data$dist_actions_model %||%
    sol_problem$data$dist_actions %||%
    x$data$dist_actions_model %||%
    x$data$dist_actions %||%
    NULL

  if (is.null(da) || !inherits(da, "data.frame")) {
    stop("Missing dist_actions data while evaluating action_boundary_cut.", call. = FALSE)
  }

  for (nm in c("internal_pu", "internal_action")) {
    if (!nm %in% names(da)) {
      stop("dist_actions data must contain column '", nm, "'.", call. = FALSE)
    }
  }

  sol_vec <- .pamo_get_solution_vector(solution)

  ml <- sol_problem$data$model_list %||% x$data$model_list %||% NULL

  n_pu <- as.integer(ml$n_pu %||% nrow(x$data$pu))
  if (!is.finite(n_pu) || is.na(n_pu) || n_pu <= 0L) {
    stop("Could not determine n_pu while evaluating action_boundary_cut.", call. = FALSE)
  }

  n_x <- as.integer(ml$n_x %||% nrow(da))
  if (!is.finite(n_x) || is.na(n_x) || n_x <= 0L) {
    stop("Could not determine n_x while evaluating action_boundary_cut.", call. = FALSE)
  }

  # IMPORTANT:
  # fallback must follow the structural layout w | x | ...
  x0 <- as.integer(ml$x_offset %||% n_pu)
  if (!is.finite(x0) || is.na(x0) || x0 < 0L) {
    stop("Could not determine x_offset while evaluating action_boundary_cut.", call. = FALSE)
  }

  if (!"internal_row" %in% names(da)) {
    if (nrow(da) != n_x) {
      stop(
        "dist_actions data does not contain 'internal_row', and nrow(dist_actions) != n_x.\n",
        "nrow(dist_actions) = ", nrow(da), ", n_x = ", n_x, ".",
        call. = FALSE
      )
    }
    da$internal_row <- seq_len(nrow(da))
  }

  if (any(!is.finite(da$internal_row)) || any(is.na(da$internal_row))) {
    stop("dist_actions$internal_row contains NA/non-finite values.", call. = FALSE)
  }
  if (any(da$internal_row < 1L) || any(da$internal_row > n_x)) {
    stop("dist_actions$internal_row is out of valid range 1..n_x.", call. = FALSE)
  }

  if (length(sol_vec) < (x0 + n_x)) {
    stop(
      "Solution vector is too short for action boundary evaluation.\n",
      "length(sol_vec) = ", length(sol_vec),
      ", required >= ", x0 + n_x, ".",
      call. = FALSE
    )
  }

  all_actions <- sort(unique(as.integer(da$internal_action)))
  all_actions <- all_actions[is.finite(all_actions) & !is.na(all_actions)]

  act_raw <- term$actions %||% NULL
  if (is.null(act_raw) || length(act_raw) == 0L) {
    act_int <- all_actions
  } else {
    act_int <- as.integer(act_raw)
    act_int <- sort(unique(act_int))
  }

  if (length(act_int) == 0L) return(0)

  aw <- term$action_weights %||% NULL
  if (is.null(aw)) {
    aw_map <- stats::setNames(rep(1, length(act_int)), act_int)
  } else {
    aw <- as.numeric(aw)
    if (length(aw) == length(all_actions)) {
      tmp_map <- stats::setNames(aw, all_actions)
      aw_map <- tmp_map[as.character(act_int)]
    } else if (length(aw) == length(act_int)) {
      aw_map <- stats::setNames(aw, act_int)
    } else {
      stop(
        "action_weights length must match either all actions or the selected action subset.",
        call. = FALSE
      )
    }
  }

  da_sub <- da[, c("internal_row", "internal_pu", "internal_action"), drop = FALSE]
  da_sub <- da_sub[da_sub$internal_action %in% act_int, , drop = FALSE]
  if (nrow(da_sub) == 0L) return(0)

  x_vals <- sol_vec[x0 + da_sub$internal_row]
  x_bin  <- as.numeric(x_vals > 0.5)

  ip1 <- as.integer(rel$internal_pu1)
  ip2 <- as.integer(rel$internal_pu2)
  ww  <- as.numeric(rel$weight)

  ok <- is.finite(ip1) & is.finite(ip2) & is.finite(ww)
  ip1 <- ip1[ok]
  ip2 <- ip2[ok]
  ww  <- ww[ok]

  self_w <- numeric(n_pu)
  edge_map <- new.env(parent = emptyenv(), hash = TRUE)

  make_key <- function(a, b) paste0(a, "::", b)

  for (r in seq_along(ip1)) {
    i <- ip1[r]
    j <- ip2[r]
    wij <- ww[r]

    if (i == j) {
      self_w[i] <- self_w[i] + wij
    } else {
      a <- min(i, j)
      b <- max(i, j)
      k <- make_key(a, b)
      old <- if (exists(k, envir = edge_map, inherits = FALSE)) {
        get(k, envir = edge_map, inherits = FALSE)
      } else {
        NA_real_
      }
      if (is.na(old) || wij > old) {
        assign(k, wij, envir = edge_map)
      }
    }
  }

  edge_keys <- ls(edge_map, all.names = TRUE)
  if (length(edge_keys) == 0L) return(0)

  edge_df <- data.frame(
    a = integer(length(edge_keys)),
    b = integer(length(edge_keys)),
    weight = numeric(length(edge_keys))
  )

  for (k in seq_along(edge_keys)) {
    parts <- strsplit(edge_keys[k], "::", fixed = TRUE)[[1]]
    edge_df$a[k] <- as.integer(parts[1])
    edge_df$b[k] <- as.integer(parts[2])
    edge_df$weight[k] <- get(edge_keys[k], envir = edge_map, inherits = FALSE)
  }

  edge_df <- edge_df[order(edge_df$a, edge_df$b), , drop = FALSE]

  incident_w <- numeric(n_pu)
  for (r in seq_len(nrow(edge_df))) {
    incident_w[edge_df$a[r]] <- incident_w[edge_df$a[r]] + edge_df$weight[r]
    incident_w[edge_df$b[r]] <- incident_w[edge_df$b[r]] + edge_df$weight[r]
  }

  val <- 0

  for (act in act_int) {
    awi <- as.numeric(aw_map[[as.character(act)]])
    if (!is.finite(awi) || awi == 0) next

    x_i <- numeric(n_pu)
    idx_act <- da_sub$internal_action == act
    if (any(idx_act)) {
      pu_act <- da_sub$internal_pu[idx_act]
      xv_act <- x_bin[idx_act]
      x_i[pu_act] <- xv_act
    }

    linear_sum <- sum((incident_w + self_w) * x_i)

    edge_sum <- 0
    for (r in seq_len(nrow(edge_df))) {
      edge_sum <- edge_sum + edge_df$weight[r] * x_i[edge_df$a[r]] * x_i[edge_df$b[r]]
    }

    val <- val + awi * (linear_sum - 2 * edge_sum)
  }

  mult <- as.numeric(term$weight_multiplier %||% 1)[1]
  if (!is.finite(mult)) mult <- 1

  as.numeric(mult * val)
}


.pamo_eval_intervention_impact_on_solution <- function(x, solution, term) {
  stopifnot(inherits(x, "Problem"))

  icol <- as.character(term$impact_col %||% "amount")[1]
  feats <- as.character(term$features %||% character(0))
  acts  <- as.character(term$actions %||% character(0))

  distf <- x$data$dist_features %||% NULL
  if (is.null(distf) || !inherits(distf, "data.frame") || nrow(distf) == 0) {
    stop("dist_features is missing or empty while evaluating intervention_impact.", call. = FALSE)
  }

  if (!(icol %in% names(distf))) {
    stop("impact column '", icol, "' not found in dist_features.", call. = FALSE)
  }

  # filtrar features si corresponde
  if (length(feats) > 0) {
    feat_subset <- .pa_resolve_feature_subset(x, features = feats)
    feat_ids <- unique(as.character(feat_subset$id))
    distf <- distf[as.character(distf$feature) %in% feat_ids, , drop = FALSE]
  }

  if (nrow(distf) == 0) return(0)

  # resolver subset de acciones
  act_subset <- .pa_resolve_action_subset(x, subset = acts)
  if (!inherits(act_subset, "data.frame") || nrow(act_subset) == 0) {
    stop("intervention_impact resolved to zero actions.", call. = FALSE)
  }

  act_ids <- unique(as.character(act_subset$id))

  # tabla de acciones de la solución
  act_tbl <- solution$summary$actions %||% NULL
  if (is.null(act_tbl) || !inherits(act_tbl, "data.frame")) {
    stop("No actions table found in solution while evaluating intervention_impact.", call. = FALSE)
  }

  if (!("selected" %in% names(act_tbl))) {
    stop("Solution actions table has no 'selected' column.", call. = FALSE)
  }

  # columnas esperadas: pu, action
  pu_col <- if ("pu" %in% names(act_tbl)) "pu" else if ("internal_pu" %in% names(act_tbl)) "internal_pu" else NULL
  ac_col <- if ("action" %in% names(act_tbl)) "action" else if ("id" %in% names(act_tbl)) "id" else NULL

  if (is.null(pu_col) || is.null(ac_col)) {
    stop(
      "Could not identify PU/action columns in solution$summary$actions while evaluating intervention_impact.",
      call. = FALSE
    )
  }

  act_sel <- act_tbl[
    act_tbl$selected > 0.5 & as.character(act_tbl[[ac_col]]) %in% act_ids,
    ,
    drop = FALSE
  ]

  if (nrow(act_sel) == 0) return(0)

  pu_hit <- unique(act_sel[[pu_col]])

  # sumar impacto por PU intervenida
  val <- sum(
    distf[distf$pu %in% pu_hit, icol, drop = TRUE],
    na.rm = TRUE
  )

  as.numeric(val)
}


.pamo_bind_solution_summaries <- function(solutions, run_ids = NULL) {
  if (is.null(solutions) || length(solutions) == 0L) {
    return(list(
      pu = NULL,
      actions = NULL,
      features = NULL,
      targets = NULL
    ))
  }

  if (is.null(run_ids)) {
    run_ids <- seq_along(solutions)
  }

  if (length(run_ids) != length(solutions)) {
    stop("run_ids length must match solutions length.", call. = FALSE)
  }

  bind_one <- function(name) {
    out <- vector("list", length(solutions))

    for (i in seq_along(solutions)) {
      sol <- solutions[[i]]
      if (!inherits(sol, "Solution")) next

      tb <- sol$summary[[name]] %||% NULL
      if (is.null(tb) || !inherits(tb, "data.frame") || nrow(tb) == 0L) next

      tb$run_id <- run_ids[i]
      out[[i]] <- tb
    }

    out <- Filter(Negate(is.null), out)
    if (length(out) == 0L) return(NULL)

    # usar tu helper de fill si las columnas no coinciden exactamente
    acc <- out[[1]]
    if (length(out) > 1L) {
      for (j in 2:length(out)) {
        acc <- .pa_rbind_fill(acc, out[[j]])
      }
    }
    acc
  }

  list(
    pu = bind_one("pu"),
    actions = bind_one("actions"),
    features = bind_one("features"),
    targets = bind_one("targets")
  )
}


.pamo_eval_alias_canonical_on_solution <- function(x, solution, alias) {
  stopifnot(inherits(x, "Problem"))

  alias <- as.character(alias)[1]
  if (is.na(alias) || !nzchar(alias)) {
    stop("alias must be a non-empty string.", call. = FALSE)
  }

  val <- .pamo_eval_alias_on_solution(x, solution, alias)
  spec <- .pamo_get_objective_spec(x, alias)

  sense <- as.character(spec$sense %||% NA_character_)[1]
  if (!sense %in% c("min", "max")) {
    stop("Objective alias '", alias, "' has invalid sense.", call. = FALSE)
  }

  if (identical(sense, "max")) {
    return(as.numeric(-val))
  }

  as.numeric(val)
}


.pamo_compute_weighted_ranges <- function(x,
                                          aliases,
                                          tol = 1e-9,
                                          verbose = FALSE,
                                          gap_limit = NULL,
                                          time_limit = NULL) {
  stopifnot(inherits(x, "Problem"))

  aliases <- as.character(aliases)
  if (length(aliases) == 0L) {
    stop("aliases must be non-empty in .pamo_compute_weighted_ranges().", call. = FALSE)
  }

  .pamo_cli_step(
    "Computing payoff table for {.val {length(aliases)}} objective{?s}.",
    verbose = verbose
  )

  payoff <- matrix(
    NA_real_,
    nrow = length(aliases),
    ncol = length(aliases),
    dimnames = list(aliases, aliases)
  )

  for (i in seq_along(aliases)) {
    a_i <- aliases[i]

    .pamo_cli_step(
      "Anchor {.val {i}}/{.val {length(aliases)}}: optimizing {.val {a_i}}.",
      verbose = verbose
    )

    one <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = a_i,
        weights = 1,
        objective_scaling = FALSE,
        scales = NULL,
        gap_limit = gap_limit,
        time_limit = time_limit
      )
    )

    sol_i <- one$solution
    if (is.null(sol_i) || !inherits(sol_i, "Solution")) {
      stop(
        "Failed to compute scaling range anchor for objective '", a_i, "'.",
        call. = FALSE
      )
    }

    vals_i <- vapply(
      aliases,
      function(a_j) .pamo_eval_alias_canonical_on_solution(x, sol_i, a_j),
      numeric(1)
    )

    payoff[i, ] <- vals_i
  }

  ideal <- apply(payoff, 2, min, na.rm = TRUE)
  nadir <- apply(payoff, 2, max, na.rm = TRUE)
  ranges <- nadir - ideal
  names(ranges) <- aliases

  bad <- !is.finite(ranges) | (ranges <= tol)
  if (any(bad)) {
    .pamo_cli_warn(
      "Some objectives had near-zero payoff range: {.val {paste(names(ranges)[bad], collapse = ', ')}}. Using scale = 1.",
      verbose = verbose
    )
    ranges[bad] <- 1
  }

  .pamo_cli_done("Payoff ranges computed.", verbose = verbose)

  ranges
}


################################################################################
######## AUGMECON ##############################################################
################################################################################

.pamo_apply_augmecon_objective <- function(base,
                                           ir_primary,
                                           slack_cols,
                                           slack_weights) {
  stopifnot(inherits(base, "Problem"))

  slack_cols <- as.integer(slack_cols)
  slack_weights <- as.numeric(slack_weights)

  if (length(slack_cols) == 0L) {
    stop("slack_cols must contain at least one slack column.", call. = FALSE)
  }

  if (length(slack_cols) != length(slack_weights)) {
    stop(
      "slack_cols and slack_weights must have the same length.",
      call. = FALSE
    )
  }

  if (any(!is.finite(slack_cols)) || any(is.na(slack_cols)) || any(slack_cols < 0L)) {
    stop("slack_cols must contain valid 0-based column indices.", call. = FALSE)
  }

  if (any(!is.finite(slack_weights)) || any(is.na(slack_weights)) || any(slack_weights <= 0)) {
    stop("slack_weights must contain finite positive values.", call. = FALSE)
  }

  v_primary <- .pamo_objvec_from_ir(base, ir_primary)

  # canonical minimization form
  if (identical(ir_primary$sense, "max")) {
    v_primary <- -v_primary
  }

  obj <- as.numeric(v_primary)

  # add AUGMECON slack contribution:
  # minimize primary - sum_j weight_j * slack_j
  for (i in seq_along(slack_cols)) {
    j <- slack_cols[i] + 1L
    if (j > length(obj)) {
      stop(
        "Slack column ", slack_cols[i], " is outside the model dimension (length(obj) = ",
        length(obj), ").",
        call. = FALSE
      )
    }
    obj[j] <- obj[j] - slack_weights[i]
  }

  base$data$runtime_updates <- list(
    obj = obj,
    modelsense = "min"
  )

  base$data$meta <- base$data$meta %||% list()
  base$data$meta$model_dirty <- FALSE
  base$data$has_model <- TRUE

  base$data$mo_cache <- base$data$mo_cache %||% list()
  base$data$mo_cache$augmecon <- utils::modifyList(
    base$data$mo_cache$augmecon %||% list(),
    list(
      ir_primary = ir_primary,
      slack_cols_0based = slack_cols,
      slack_weights = slack_weights,
      obj_augmented = obj
    )
  )

  base
}



.pamo_solve_augmecon <- function(x, ...) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% NULL
  if (is.null(method) || !is.list(method)) {
    stop("augmecon: x$data$method is missing or invalid.", call. = FALSE)
  }

  primary      <- as.character(method$primary %||% NA_character_)[1]
  aliases      <- as.character(method$aliases %||% character(0))
  secondary    <- as.character(method$secondary %||% character(0))
  grid         <- method$grid %||% NULL
  n_points     <- as.integer(method$n_points %||% NA_integer_)[1]
  augmentation <- as.numeric(method$augmentation %||% NA_real_)[1]

  dots <- list(...)
  gap_limit  <- dots$gap_limit %||% NULL
  time_limit <- dots$time_limit %||% NULL
  verbose <- .pamo_cli_enabled(x, dots)

  progress <- dots$progress %||% dots$verbose %||% interactive()
  progress <- isTRUE(progress)

  .pamo_cli_header("AUGMECON method", verbose = verbose)

  if (is.na(primary) || !nzchar(primary)) {
    stop("augmecon: missing primary objective.", call. = FALSE)
  }
  if (length(aliases) == 0L) {
    stop("augmecon: aliases are missing.", call. = FALSE)
  }
  if (length(secondary) == 0L) {
    stop("augmecon: secondary objectives are missing.", call. = FALSE)
  }
  if (!is.finite(augmentation) || is.na(augmentation) || augmentation <= 0) {
    stop("augmecon: augmentation must be a finite positive number.", call. = FALSE)
  }

  secondary_ranges <- vapply(
    secondary,
    function(a) {
      rr <- .pamo_compute_secondary_range_against_primary(
        x = x,
        primary = primary,
        secondary = a,
        gap_limit = gap_limit,
        time_limit = time_limit
      )
      rng <- as.numeric(rr$secondary_max)[1] - as.numeric(rr$secondary_min)[1]
      if (!is.finite(rng) || rng <= 0) rng <- 1
      rng
    },
    numeric(1)
  )

  .pamo_cli_step(
    "Primary objective: {.val {primary}}. Secondary objectives: {.val {paste(secondary, collapse = ', ')}}.",
    verbose = verbose
  )

  # design
  design_df <- method$runs %||% NULL
  if (is.null(design_df)) {
    if (is.null(grid)) {
      .pamo_cli_step("Building automatic epsilon grid.", verbose = verbose)

      design_df <- .pamo_build_auto_augmecon_runs(
        x = x,
        gap_limit = gap_limit,
        time_limit = time_limit,
        verbose = verbose
      )
    } else {
      .pamo_cli_step("Using user-supplied epsilon grid.", verbose = verbose)

      design_df <- .pamo_build_manual_augmecon_runs(
        x = x,
        verbose = verbose
      )
    }
    x$data$method$runs <- design_df
  }

  if (!("run_id" %in% names(design_df))) {
    design_df$run_id <- seq_len(nrow(design_df))
  }

  eps_cols <- paste0("eps_", secondary)
  miss_cols <- setdiff(eps_cols, names(design_df))
  if (length(miss_cols) > 0L) {
    stop(
      "augmecon: design is missing epsilon columns: ",
      paste(miss_cols, collapse = ", "),
      call. = FALSE
    )
  }

  .pamo_cli_step(
    "Design ready with {.val {nrow(design_df)}} run{?s}. Augmentation = {.val {augmentation}}.",
    verbose = verbose
  )

  n_runs <- nrow(design_df)
  solutions <- vector("list", n_runs)

  value_mat <- matrix(NA_real_, n_runs, length(aliases))
  colnames(value_mat) <- paste0("value_", aliases)

  status  <- character(n_runs)
  runtime <- numeric(n_runs)
  gap     <- numeric(n_runs)
  messages <- character(n_runs)

  stop_on_infeasible <- isTRUE(method$stop_on_infeasible %||% FALSE)
  stop_on_no_solution <- isTRUE(method$stop_on_no_solution %||% FALSE)
  stop_on_error <- isTRUE(method$stop_on_error %||% TRUE)

  progress_id <- NULL
  if (progress && n_runs > 1L) {
    progress_id <- NULL
    if (progress && n_runs > 1L) {
      progress_id <- cli::cli_progress_bar(
        name = "Solving runs",
        total = n_runs,
        clear = FALSE,
        format = "{cli::pb_name}{cli::pb_bar} {cli::pb_current}/{cli::pb_total} | ETA: {cli::pb_eta}"
      )
    }
  }


  for (r in seq_len(n_runs)) {
    eps_r <- as.list(design_df[r, eps_cols, drop = FALSE])
    names(eps_r) <- secondary

    one <- .pamo_solve_one_or_keep_failed(
      x = x,
      spec = list(
        type = "augmecon",
        primary = primary,
        eps = eps_r,
        eps_tol = stats::setNames(as.list(rep(0, length(secondary))), secondary),
        augmentation = augmentation,
        secondary_ranges = secondary_ranges,
        gap_limit = gap_limit,
        time_limit = time_limit
      ),
      stop_on_infeasible = stop_on_infeasible,
      stop_on_no_solution = stop_on_no_solution,
      stop_on_error = stop_on_error
    )

    solutions[r] <- list(one$solution)
    status[r]  <- as.character(one$status %||% NA_character_)
    runtime[r] <- as.numeric(one$runtime %||% NA_real_)
    gap[r]     <- as.numeric(one$gap %||% NA_real_)
    messages[r] <- as.character(one$message %||% NA_character_)

    if (!is.null(solutions[[r]]) && inherits(solutions[[r]], "Solution")) {

      alias_values <- stats::setNames(
        vapply(
          aliases,
          function(a) .pamo_eval_alias_on_solution(x, one$solution, a),
          numeric(1)
        ),
        aliases
      )

      solutions[[r]]$solution$alias_values <- alias_values
      solutions[[r]]$meta$run_id <- design_df$run_id[r]
      solutions[[r]]$method$type <- "augmecon"
      solutions[[r]]$method$primary_alias <- primary
      solutions[[r]]$method$eps <- stats::setNames(
        as.numeric(design_df[r, eps_cols, drop = TRUE]),
        secondary
      )
      solutions[[r]]$method$augmentation <- augmentation

      value_mat[r, ] <- unname(as.numeric(alias_values))

    } else {

      value_mat[r, ] <- NA_real_
    }

    if (!is.null(progress_id)) {
      cli::cli_progress_update(id = progress_id, set = r)
    }
  }

  if (!is.null(progress_id)) {
    cli::cli_progress_done(id = progress_id)
  }

  runs <- data.frame(
    run_id = design_df$run_id,
    status = status,
    runtime = runtime,
    gap = gap,
    message = messages,
    stringsAsFactors = FALSE
  )

  runs <- cbind(runs, as.data.frame(value_mat, stringsAsFactors = FALSE))

  summary_set <- .pamo_bind_solution_summaries(
    solutions = solutions,
    run_ids = design_df$run_id
  )

  .pamo_cli_done("AUGMECON solve finished.", verbose = verbose)

  ext <- attr(design_df, "extremes", exact = TRUE)

  ranges_list <- list()

  # rango del primario a partir de los anchors de todos los secundarios
  primary_vals <- numeric(0)
  if (!is.null(ext) && length(ext) > 0L) {
    primary_vals <- unlist(
      lapply(ext, function(e) {
        c(
          as.numeric(e$primary_at_primary %||% NA_real_),
          as.numeric(e$primary_at_secondary %||% NA_real_)
        )
      }),
      use.names = FALSE
    )
    primary_vals <- primary_vals[is.finite(primary_vals)]
  }

  ranges_list[[1]] <- data.frame(
    alias = primary,
    role = "primary",
    min = if (length(primary_vals)) min(primary_vals) else NA_real_,
    max = if (length(primary_vals)) max(primary_vals) else NA_real_,
    source = "anchors",
    stringsAsFactors = FALSE
  )

  if (!is.null(ext) && length(ext) > 0L) {
    for (s in secondary) {
      e <- ext[[s]]
      if (is.null(e)) next

      ranges_list[[length(ranges_list) + 1L]] <- data.frame(
        alias = s,
        role = "secondary",
        min = as.numeric(e$secondary_min %||% NA_real_)[1],
        max = as.numeric(e$secondary_max %||% NA_real_)[1],
        source = "auto_grid",
        stringsAsFactors = FALSE
      )
    }
  }

  ranges_df <- do.call(rbind, ranges_list)
  rownames(ranges_df) <- NULL

  n_infeasible <- sum(
    runs$status %in% c("infeasible", "infeasible_or_unbounded"),
    na.rm = TRUE
  )

  n_no_solution <- sum(
    runs$status %in% c("no_solution"),
    na.rm = TRUE
  )

  n_failed <- sum(
    runs$status %in% c("failed"),
    na.rm = TRUE
  )

  if (n_infeasible > 0L) {
    warning(
      n_infeasible, " of ", nrow(runs),
      " AUGMECON runs were infeasible and were kept in the SolutionSet with missing objective values. ",
      "Feasible runs were retained.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  if (n_no_solution > 0L) {
    warning(
      n_no_solution, " of ", nrow(runs),
      " AUGMECON runs did not return a solution vector and were kept in the SolutionSet with missing objective values. ",
      "This may indicate infeasibility, numerical issues, or solver termination before finding a feasible solution.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  if (n_failed > 0L) {
    warning(
      n_failed, " of ", nrow(runs),
      " AUGMECON runs failed for reasons other than infeasibility/no-solution.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  pproto(
    NULL, SolutionSet,
    problem = x,
    solution = list(
      design = design_df,
      runs = runs,
      solutions = solutions,
      frontier = list(
        primary = primary,
        secondary = secondary,
        ranges = ranges_df,
        extremes = ext
      )
    ),
    summary = summary_set,
    diagnostics = list(
      n_design = nrow(design_df),
      n_runs = nrow(runs),
      n_solutions = sum(vapply(solutions, inherits, logical(1), "Solution")),
      n_optimal = sum(runs$status == "optimal", na.rm = TRUE),
      n_infeasible = sum(runs$status %in% c("infeasible", "infeasible_or_unbounded"), na.rm = TRUE),
      n_no_solution = sum(runs$status == "no_solution", na.rm = TRUE),
      n_failed = sum(runs$status == "failed", na.rm = TRUE),
      status_summary = .pa_solutionset_status_summary(runs),
      runtime_range = if ("runtime" %in% names(runs)) .pa_solutionset_range_text(runs$runtime, digits = 3) else "none",
      gap_range = if ("gap" %in% names(runs)) .pa_solutionset_range_text(runs$gap, digits = 4) else "none"
    ),
    method = method,
    meta = list(
      call = match.call()
    ),
    name = "solset"
  )
}


.pamo_compute_secondary_range_against_primary <- function(
    x,
    primary,
    secondary,
    gap_limit = NULL,
    time_limit = NULL
) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% list()
  do_lexi <- isTRUE(method$lexicographic)
  lexi_tol <- as.numeric(method$lexicographic_tol %||% 0)[1]
  if (!is.finite(lexi_tol) || lexi_tol < 0) lexi_tol <- 0

  primary <- as.character(primary)[1]
  secondary <- as.character(secondary)[1]

  if (is.na(primary) || !nzchar(primary)) {
    stop("`primary` must be a non-empty string.", call. = FALSE)
  }
  if (is.na(secondary) || !nzchar(secondary)) {
    stop("`secondary` must be a non-empty string.", call. = FALSE)
  }
  if (identical(primary, secondary)) {
    stop("`primary` and `secondary` must be different aliases.", call. = FALSE)
  }

  specs_all <- .pamo_get_specs(x)
  obj_alias <- names(specs_all)

  if (!primary %in% obj_alias) {
    stop("Primary alias not found: '", primary, "'.", call. = FALSE)
  }
  if (!secondary %in% obj_alias) {
    stop("Secondary alias not found: '", secondary, "'.", call. = FALSE)
  }

  # ------------------------------------------------------------
  # Case 1: non-lexicographic anchors
  # ------------------------------------------------------------
  if (!isTRUE(do_lexi)) {

    # Anchor A: optimize primary
    sol_primary <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = primary,
        weights = 1,
        objective_scaling = FALSE,
        scales = NULL
      )
    )

    if (is.null(sol_primary$solution) || !inherits(sol_primary$solution, "Solution")) {
      stop("Failed to compute primary anchor solution for '", primary, "'.", call. = FALSE)
    }

    primary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary$solution, primary)
    secondary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary$solution, secondary)

    # Anchor B: optimize secondary
    sol_secondary <- .pamo_solve_one(
      x = x,
      spec = list(
        type = "weighted",
        aliases = secondary,
        weights = 1,
        objective_scaling = FALSE,
        scales = NULL
      )
    )

    if (is.null(sol_secondary$solution) || !inherits(sol_secondary$solution, "Solution")) {
      stop("Failed to compute secondary anchor solution for '", secondary, "'.", call. = FALSE)
    }

    primary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary$solution, primary)
    secondary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary$solution, secondary)

    vals_secondary <- c(secondary_at_primary, secondary_at_secondary)

    if (any(!is.finite(vals_secondary))) {
      stop(
        "Could not compute finite anchor values for secondary objective '", secondary, "'.\n",
        "secondary_at_primary = ", secondary_at_primary, "\n",
        "secondary_at_secondary = ", secondary_at_secondary,
        call. = FALSE
      )
    }

    return(list(
      primary = primary,
      secondary = secondary,

      primary_at_primary = as.numeric(primary_at_primary),
      secondary_at_primary = as.numeric(secondary_at_primary),

      primary_at_secondary = as.numeric(primary_at_secondary),
      secondary_at_secondary = as.numeric(secondary_at_secondary),

      secondary_min = min(vals_secondary),
      secondary_max = max(vals_secondary),

      solution_primary = sol_primary$solution,
      solution_secondary = sol_secondary$solution,

      meta = list(
        lexicographic = FALSE,
        lexicographic_tol = 0
      )
    ))
  }

  # ------------------------------------------------------------
  # Case 2: lexicographic anchors
  # ------------------------------------------------------------

  # Stage 1A: optimize primary
  sol_primary_stage1 <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "weighted",
      aliases = primary,
      weights = 1,
      objective_scaling = FALSE,
      scales = NULL
    )
  )

  if (is.null(sol_primary_stage1$solution) || !inherits(sol_primary_stage1$solution, "Solution")) {
    stop("Failed to compute stage-1 primary solution for '", primary, "'.", call. = FALSE)
  }

  primary_star <- .pamo_eval_alias_on_solution(x, sol_primary_stage1$solution, primary)

  # Stage 2A: optimize secondary subject to primary <= primary_star + tol
  sol_primary_lexi <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "epsilon_constraint",
      primary = secondary,
      eps = stats::setNames(list(primary_star), primary),
      eps_tol = stats::setNames(list(lexi_tol), primary)
    )
  )

  if (is.null(sol_primary_lexi$solution) || !inherits(sol_primary_lexi$solution, "Solution")) {
    stop("Failed to compute lexicographic primary anchor for '", primary, "'.", call. = FALSE)
  }

  primary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary_lexi$solution, primary)
  secondary_at_primary <- .pamo_eval_alias_on_solution(x, sol_primary_lexi$solution, secondary)

  # Stage 1B: optimize secondary
  sol_secondary_stage1 <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "weighted",
      aliases = secondary,
      weights = 1,
      objective_scaling = FALSE,
      scales = NULL
    )
  )

  if (is.null(sol_secondary_stage1$solution) || !inherits(sol_secondary_stage1$solution, "Solution")) {
    stop("Failed to compute stage-1 secondary solution for '", secondary, "'.", call. = FALSE)
  }

  secondary_star <- .pamo_eval_alias_on_solution(x, sol_secondary_stage1$solution, secondary)

  # Stage 2B: optimize primary subject to secondary <= secondary_star + tol
  sol_secondary_lexi <- .pamo_solve_one(
    x = x,
    spec = list(
      type = "epsilon_constraint",
      primary = primary,
      eps = stats::setNames(list(secondary_star), secondary),
      eps_tol = stats::setNames(list(lexi_tol), secondary)
    )
  )

  if (is.null(sol_secondary_lexi$solution) || !inherits(sol_secondary_lexi$solution, "Solution")) {
    stop("Failed to compute lexicographic secondary anchor for '", secondary, "'.", call. = FALSE)
  }

  primary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary_lexi$solution, primary)
  secondary_at_secondary <- .pamo_eval_alias_on_solution(x, sol_secondary_lexi$solution, secondary)

  vals_secondary <- c(secondary_at_primary, secondary_at_secondary)

  if (any(!is.finite(vals_secondary))) {
    stop(
      "Could not compute finite anchor values for secondary objective '", secondary, "'.\n",
      "secondary_at_primary = ", secondary_at_primary, "\n",
      "secondary_at_secondary = ", secondary_at_secondary,
      call. = FALSE
    )
  }

  list(
    primary = primary,
    secondary = secondary,

    primary_at_primary = as.numeric(primary_at_primary),
    secondary_at_primary = as.numeric(secondary_at_primary),

    primary_at_secondary = as.numeric(primary_at_secondary),
    secondary_at_secondary = as.numeric(secondary_at_secondary),

    secondary_min = min(vals_secondary),
    secondary_max = max(vals_secondary),

    solution_primary = sol_primary_lexi$solution,
    solution_secondary = sol_secondary_lexi$solution,

    meta = list(
      lexicographic = TRUE,
      lexicographic_tol = lexi_tol,
      stage1_primary_solution = sol_primary_stage1$solution,
      stage1_secondary_solution = sol_secondary_stage1$solution
    )
  )
}



.pamo_build_auto_augmecon_runs <- function(x,
                                           gap_limit = NULL,
                                           time_limit = NULL,
                                           verbose = FALSE) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% list()

  primary          <- as.character(method$primary %||% NA_character_)[1]
  aliases          <- as.character(method$aliases %||% character(0))
  secondary        <- as.character(method$secondary %||% character(0))
  n_points         <- as.integer(method$n_points %||% NA_integer_)[1]
  include_extremes <- isTRUE(method$include_extremes)

  if (is.na(primary) || !nzchar(primary)) {
    stop("Auto AUGMECON mode: missing primary objective.", call. = FALSE)
  }

  if (length(aliases) == 0L) {
    stop("Auto AUGMECON mode: aliases are missing.", call. = FALSE)
  }

  if (!primary %in% aliases) {
    stop("Auto AUGMECON mode: primary must be included in aliases.", call. = FALSE)
  }

  if (length(secondary) == 0L) {
    stop("Auto AUGMECON mode requires at least one secondary objective.", call. = FALSE)
  }

  if (!all(secondary %in% aliases)) {
    stop(
      "Auto AUGMECON mode: all secondary objectives must be included in aliases.\n",
      "Missing: ",
      paste(setdiff(secondary, aliases), collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.finite(n_points) || is.na(n_points) || n_points < 2L) {
    stop("Auto AUGMECON mode requires n_points >= 2.", call. = FALSE)
  }

  .pamo_cli_step(
    "Computing automatic ranges for {.val {length(secondary)}} secondary objective{?s}.",
    verbose = verbose
  )

  # ------------------------------------------------------------
  # build one epsilon sequence per secondary objective
  # ------------------------------------------------------------
  grid_list <- vector("list", length(secondary))
  names(grid_list) <- secondary

  extremes <- vector("list", length(secondary))
  names(extremes) <- secondary

  for (i in seq_along(secondary)) {
    s <- secondary[i]

    .pamo_cli_step(
      "Secondary {.val {i}}/{.val {length(secondary)}}: {.val {s}}.",
      verbose = verbose
    )

    ext_s <- .pamo_compute_secondary_range_against_primary(
      x = x,
      primary = primary,
      secondary = s,
      gap_limit = gap_limit,
      time_limit = time_limit
    )

    s_min <- as.numeric(ext_s$secondary_min)[1]
    s_max <- as.numeric(ext_s$secondary_max)[1]

    if (!is.finite(s_min) || !is.finite(s_max)) {
      stop(
        "Auto AUGMECON mode: computed non-finite bounds for secondary objective '", s, "'.",
        call. = FALSE
      )
    }

    if (s_min > s_max) {
      tmp <- s_min
      s_min <- s_max
      s_max <- tmp
    }

    .pamo_cli_step(
      "Range for {.val {s}}: [{.val {format(s_min, digits = 6)}}, {.val {format(s_max, digits = 6)}}].",
      verbose = verbose
    )

    vals <- seq(from = s_min, to = s_max, length.out = n_points)

    if (!isTRUE(include_extremes)) {
      if (length(vals) <= 2L) {
        stop(
          "include_extremes = FALSE requires n_points >= 3.",
          call. = FALSE
        )
      }
      vals <- vals[2:(length(vals) - 1L)]
    }

    if (length(vals) == 0L) {
      stop(
        "Auto AUGMECON mode produced an empty grid for secondary objective '", s, "'.",
        call. = FALSE
      )
    }

    grid_list[[s]] <- as.numeric(vals)
    extremes[[s]] <- ext_s
  }

  # ------------------------------------------------------------
  # Cartesian product of epsilon levels
  # ------------------------------------------------------------
  design_df <- expand.grid(
    grid_list,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )

  names(design_df) <- paste0("eps_", names(grid_list))
  design_df$run_id <- seq_len(nrow(design_df))

  # reorder columns: run_id first
  design_df <- design_df[, c("run_id", setdiff(names(design_df), "run_id")), drop = FALSE]

  attr(design_df, "extremes") <- extremes
  attr(design_df, "primary_alias") <- primary
  attr(design_df, "secondary_aliases") <- secondary
  attr(design_df, "include_extremes") <- include_extremes
  attr(design_df, "n_points_requested") <- n_points

  .pamo_cli_done(
    "Automatic AUGMECON grid built with {.val {nrow(design_df)}} run{?s}.",
    verbose = verbose
  )

  design_df
}



.pamo_build_manual_augmecon_runs <- function(x, verbose = FALSE) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% list()
  grid <- method$grid %||% NULL
  secondary <- as.character(method$secondary %||% character(0))

  if (is.null(grid) || !is.list(grid) || length(grid) == 0L) {
    stop("Manual AUGMECON mode requires a non-empty named grid list.", call. = FALSE)
  }

  if (length(secondary) == 0L) {
    stop("Manual AUGMECON mode: secondary objectives are missing.", call. = FALSE)
  }

  if (!all(secondary %in% names(grid))) {
    stop(
      "Manual AUGMECON mode: grid is missing secondary objectives: ",
      paste(setdiff(secondary, names(grid)), collapse = ", "),
      call. = FALSE
    )
  }

  .pamo_cli_step(
    "Building manual AUGMECON grid for {.val {length(secondary)}} secondary objective{?s}.",
    verbose = verbose
  )

  grid <- grid[secondary]

  for (i in seq_along(secondary)) {
    s <- secondary[i]
    vals <- grid[[s]]

    .pamo_cli_step(
      "Secondary {.val {i}}/{.val {length(secondary)}}: {.val {s}} with {.val {length(vals)}} epsilon level{?s}.",
      verbose = verbose
    )
  }

  design_df <- expand.grid(
    grid,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )

  names(design_df) <- paste0("eps_", secondary)
  design_df$run_id <- seq_len(nrow(design_df))
  design_df <- design_df[, c("run_id", setdiff(names(design_df), "run_id")), drop = FALSE]

  attr(design_df, "primary_alias") <- as.character(method$primary %||% NA_character_)[1]
  attr(design_df, "secondary_aliases") <- secondary
  attr(design_df, "manual_grid") <- TRUE

  .pamo_cli_done(
    "Manual AUGMECON grid built with {.val {nrow(design_df)}} run{?s}.",
    verbose = verbose
  )

  design_df
}




.pamo_cli_enabled <- function(x, dots = list()) {
  # prioridad:
  # 1) verbose pasado a solve(...)
  # 2) solver verbose si existe
  # 3) interactive()
  v <- dots$verbose %||%
    x$data$solver$verbose %||%
    x$data$solver_args$verbose %||%
    interactive()

  isTRUE(v)
}

.pamo_cli_header <- function(..., verbose = TRUE) {
  if (isTRUE(verbose)) {
    cli::cli_h2(..., .envir = parent.frame())
  }
}

.pamo_cli_step <- function(..., verbose = TRUE) {
  if (isTRUE(verbose)) {
    cli::cli_alert_info(..., .envir = parent.frame())
  }
}

.pamo_cli_done <- function(..., verbose = TRUE) {
  if (isTRUE(verbose)) {
    cli::cli_alert_success(..., .envir = parent.frame())
  }
}

.pamo_cli_warn <- function(..., verbose = TRUE) {
  if (isTRUE(verbose)) {
    cli::cli_alert_warning(..., .envir = parent.frame())
  }
}



.pamo_add_augmecon_slacks <- function(base,
                                      secondary_aliases,
                                      ub = 1e6,
                                      prefix = "aug_slack") {
  stopifnot(inherits(base, "Problem"))

  secondary_aliases <- as.character(secondary_aliases)
  secondary_aliases <- secondary_aliases[!is.na(secondary_aliases) & nzchar(secondary_aliases)]

  if (length(secondary_aliases) == 0L) {
    stop("secondary_aliases must contain at least one objective alias.", call. = FALSE)
  }

  ub <- as.numeric(ub)[1]
  if (!is.finite(ub) || ub <= 0) {
    stop("`ub` must be a single finite positive number.", call. = FALSE)
  }

  if (is.null(base$data$model_ptr)) {
    stop("Model not built (model_ptr is NULL).", call. = FALSE)
  }

  # snapshot actual del modelo
  m0 <- .pa_model_from_ptr(
    base$data$model_ptr,
    args = base$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  n_old <- length(m0$obj)
  if (!is.finite(n_old) || n_old <= 0L) {
    stop("Could not determine current number of model columns.", call. = FALSE)
  }

  k <- length(secondary_aliases)

  # necesitamos una helper de bajo nivel para añadir columnas al modelo ya construido
  if (!exists("rcpp_model_add_columns", mode = "function")) {
    stop(
      "AUGMECON with explicit slacks requires a low-level helper `rcpp_model_add_columns()`.\n",
      "This helper is not available yet, so slack variables cannot be appended to the built model.",
      call. = FALSE
    )
  }

  slack_names <- paste0(prefix, "_", secondary_aliases)

  # Añadir k columnas continuas:
  # obj = 0 por ahora (el objetivo aumentado se define después)
  # lb = 0
  # ub = ub
  # vtype = "C"
  #
  # Se asume que rcpp_model_add_columns():
  # - modifica el modelo en sitio
  # - acepta vectores de obj/lb/ub/vtype/names
  # - deja el modelo consistente para luego añadir restricciones
  rcpp_model_add_columns(
    x = base$data$model_ptr,
    obj = rep(0, k),
    lb = rep(0, k),
    ub = rep(ub, k),
    vtype = rep("C", k),
    names = slack_names
  )

  # refrescar snapshot/model_list tras añadir columnas
  base <- .pa_refresh_model_snapshot(base)

  m1 <- .pa_model_from_ptr(
    base$data$model_ptr,
    args = base$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  n_new <- length(m1$obj)
  if (!identical(n_new, n_old + k)) {
    stop(
      "Failed to append AUGMECON slack variables.\n",
      "Expected ", n_old + k, " columns after insertion, got ", n_new, ".",
      call. = FALSE
    )
  }

  # columnas nuevas en índice 0-based
  slack_cols_0based <- seq.int(from = n_old, length.out = k)
  names(slack_cols_0based) <- secondary_aliases

  # guardar metadata útil
  base$data$mo_cache <- base$data$mo_cache %||% list()
  base$data$mo_cache$augmecon <- utils::modifyList(
    base$data$mo_cache$augmecon %||% list(),
    list(
      secondary_aliases = secondary_aliases,
      slack_names = slack_names,
      slack_cols_0based = slack_cols_0based,
      slack_cols_1based = slack_cols_0based + 1L,
      slack_upper_bound = ub
    )
  )

  base$data$meta <- base$data$meta %||% list()
  base$data$meta$model_dirty <- FALSE
  base$data$has_model <- TRUE

  base
}



.pamo_apply_augmecon_constraint <- function(base, ir, eps, slack_col, sense = c("min", "max"),
                                            name = "", block_name = "augmecon_constraint", tag = "") {
  stopifnot(inherits(base, "Problem"))
  sense <- match.arg(sense)

  if (is.null(base$data$model_ptr)) {
    stop("Model not built (model_ptr is NULL).", call. = FALSE)
  }

  eps <- as.numeric(eps)[1]
  if (!is.finite(eps)) {
    stop("eps must be finite.", call. = FALSE)
  }

  slack_col <- as.integer(slack_col)[1]
  if (!is.finite(slack_col) || is.na(slack_col) || slack_col < 0L) {
    stop("slack_col must be a valid 0-based column index.", call. = FALSE)
  }

  v <- .pamo_objvec_from_ir(base, ir)
  if (length(v) == 0L) {
    stop("AUGMECON objective vector is empty.", call. = FALSE)
  }

  # canonical minimization form
  if (identical(sense, "max")) {
    v <- -v
    eps <- -eps
  }

  # build equality: g(x) + s = eps
  v2 <- as.numeric(v)
  m <- .pa_model_from_ptr(
    base$data$model_ptr,
    args = base$data$model_args %||% list(),
    drop_triplets = TRUE
  )

  n_model <- length(m$obj)
  if (length(v2) != n_model) {
    stop(
      "AUGMECON constraint vector length does not match model dimension.\n",
      "length(v2) = ", length(v2), ", model dimension = ", n_model, ".",
      call. = FALSE
    )
  }
  if (slack_col + 1L > n_model) {
    stop("slack_col is outside model dimension.", call. = FALSE)
  }

  v2[slack_col + 1L] <- v2[slack_col + 1L] + 1

  idx <- which(abs(v2) > 1e-12)
  if (length(idx) == 0L) {
    stop("AUGMECON constraint vector has no non-zero coefficients.", call. = FALSE)
  }

  rcpp_add_linear_constraint(
    base$data$model_ptr,
    j0 = as.integer(idx - 1L),
    x = as.numeric(v2[idx]),
    sense = "=",
    rhs = as.numeric(eps),
    name = as.character(name %||% ""),
    block_name = block_name,
    tag = tag
  )

  base$data$meta$model_dirty <- TRUE
  base <- .pa_refresh_model_snapshot(base)
  base
}


.pamo_compile_problem <- function(x, method_name = NULL, ...) {
  stopifnot(inherits(x, "Problem"))

  method <- x$data$method %||% NULL
  if (is.null(method) || !is.list(method) || length(method) == 0L) {
    stop(
      "Multi-objective compilation requires a valid x$data$method configuration.",
      call. = FALSE
    )
  }

  if (is.null(method_name)) {
    method_name <- as.character(method$type %||% method$name %||% NA_character_)[1]
  }

  if (is.na(method_name) || !nzchar(method_name)) {
    stop(
      "Invalid multi-objective method configuration: missing method name.",
      call. = FALSE
    )
  }

  aliases <- switch(
    method_name,
    weighted = as.character(method$aliases %||% character(0)),
    epsilon_constraint = unique(c(
      as.character(method$primary %||% character(0)),
      as.character(method$constrained %||% character(0))
    )),
    augmecon = unique(c(
      as.character(method$primary %||% character(0)),
      as.character(method$secondary %||% character(0))
    )),
    stop(
      "Unknown/unsupported multi-objective method: '", method_name, "'.",
      call. = FALSE
    )
  )

  aliases <- aliases[!is.na(aliases) & nzchar(aliases)]

  if (length(aliases) == 0L) {
    stop(
      "Could not determine the active objective aliases for multi-objective compilation.",
      call. = FALSE
    )
  }

  specs <- .pamo_get_objective_specs(x, aliases)
  ir_list <- lapply(specs, function(sp) .pamo_objective_to_ir(x, sp))

  out <- .pamo_prepare_superset_model(x, ir_list)

  out$data$meta <- out$data$meta %||% list()
  out$data$meta$model_dirty <- FALSE
  out$data$has_model <- TRUE

  out$data$mo_cache <- out$data$mo_cache %||% list()
  out$data$mo_cache$compiled_method <- method_name
  out$data$mo_cache$compiled_aliases <- aliases

  out
}

#' @include internal.R
NULL

#' @name create_problem
#' @title Create a planning problem input object
#'
#' @description
#' Create a \code{Problem} object from tabular or spatial inputs.
#'
#' \code{create_problem()} is the main entry point to the \code{multiscape}
#' workflow. Its role is to standardize heterogeneous planning inputs into a
#' common internal representation that can later be used by actions, effects,
#' targets, constraints, spatial relations, objectives, and solvers.
#'
#' In all supported workflows, the result is a \code{Problem} object with a
#' canonical tabular core, optionally enriched with aligned spatial metadata
#' such as coordinates, geometry, raster references, or raw planning-unit
#' attributes.
#'
#' @details
#'
#' A \code{Problem} object created by \code{create_problem()} is the basic input
#' structure used throughout downstream \code{multiscape} workflows.
#'
#' The returned object always contains a canonical tabular planning core and may
#' additionally store aligned spatial metadata depending on the input workflow.
#' This is the internal representation on which later functions such as
#' \code{add_actions()}, \code{add_effects()},
#' \code{add_constraint_targets()}, \code{add_spatial_relations()}, and
#' \code{solve()} operate.
#'
#' \strong{Conceptual role of \code{create_problem()}}
#'
#' Regardless of the input workflow, \code{create_problem()} aims to produce a
#' consistent internal representation containing at least:
#' \itemize{
#'   \item a planning-unit table,
#'   \item a feature table,
#'   \item a feature-distribution table.
#' }
#'
#' Internally, the problem is always reduced to a tabular core of the form:
#' \itemize{
#'   \item planning units \eqn{i \in \mathcal{I}},
#'   \item features \eqn{f \in \mathcal{F}},
#'   \item non-negative baseline amounts \eqn{a_{if} \ge 0},
#'   \item and, when available, spatial metadata associated with planning units.
#' }
#'
#' The feature-distribution table provides the canonical sparse representation of
#' these baseline amounts. Thus, after \code{create_problem()} has run, the
#' landscape is internally represented by baseline feature amounts
#' \eqn{a_{if}}, where \eqn{a_{if}} denotes the amount of feature \eqn{f} in
#' planning unit \eqn{i}. These baseline amounts are later combined with
#' actions, effects, targets, objectives, and constraints to build the
#' optimization model.
#'
#' \strong{Which input mode should I use?}
#'
#' \code{create_problem()} supports four main workflows. The best choice depends
#' on the form of your data and on whether you want to preserve geometry.
#'
#' \itemize{
#'   \item \strong{Tabular mode.} Use this when \code{pu}, \code{features}, and
#'   \code{dist_features} are already available as \code{data.frame} objects and
#'   no spatial derivation is needed. This is the simplest workflow: all problem
#'   components are already tabular, so \code{create_problem()} only
#'   standardizes them into the canonical internal representation.
#'
#'   \item \strong{Vector-PU spatial mode.} Use this when planning units are
#'   polygons and feature information is stored in one or more raster layers. In
#'   this mode, \code{dist_features} is derived by aggregating raster values
#'   over planning-unit polygons.
#'
#'   \item \strong{Raster-cell fast mode.} Use this when both \code{pu} and
#'   \code{features} are rasters and each valid raster cell should become one
#'   planning unit. This mode avoids raster-to-polygon conversion, treats
#'   \code{NA} feature values as zero before building \code{dist_features},
#'   keeps only strictly positive amounts in the stored distribution table, and
#'   is generally the preferred option for large regular grids.
#'
#'   \item \strong{Hybrid sf + tabular mode.} Use this when you already have a
#'   curated tabular \code{dist_features} table but still want to preserve
#'   planning-unit geometry and attributes for later plotting, spatial
#'   relations, or feasibility specifications. Here, the feature distribution is
#'   already tabular, but geometry and raw attributes are preserved from the
#'   \code{sf} planning-unit object for later spatial operations.
#' }
#'
#' \strong{How cost is interpreted across modes}
#'
#' \code{cost} handling depends on the selected workflow.
#'
#' \itemize{
#'   \item \strong{Tabular mode.} In purely tabular mode, \code{cost} is not
#'   used by this generic method. Planning-unit costs are expected to already be
#'   present in the \code{pu} table supplied to the internal tabular builder.
#'
#'   \item \strong{Vector-PU spatial mode.} In vector-PU mode, \code{cost} is
#'   required and may be either:
#'   \itemize{
#'     \item the name of a numeric attribute column in the planning-unit layer,
#'     \item or a cost raster to be aggregated over polygons using the
#'     \code{cost_aggregation} argument.
#'   }
#'
#'   \item \strong{Raster-cell fast mode.} In raster-cell mode, \code{cost}
#'   must be a single-layer raster aligned with \code{pu} and \code{features}.
#'   A raster cell becomes a planning unit only if the mask cell is not missing
#'   and the corresponding cost value is finite and strictly positive. In other
#'   words, if \eqn{m_i} denotes the mask value of cell \eqn{i} and \eqn{c_i}
#'   its cost value, then cell \eqn{i} is retained only when the mask is
#'   observed and \eqn{c_i > 0}.
#'
#'   \item \strong{Hybrid sf + tabular mode.} In hybrid mode, \code{cost} must
#'   be either:
#'   \itemize{
#'     \item the name of a numeric attribute column in the \code{sf} layer,
#'     \item or omitted if the \code{sf} attributes already contain a column
#'     literally named \code{cost}.
#'   }
#' }
#'
#' \strong{Feature identifiers}
#'
#' Features are internally standardized to an \code{id}-based representation.
#' In spatial modes where raster layers are used, one feature is created per
#' raster layer, with:
#' \itemize{
#'   \item \code{id = 1, 2, \dots, nlyr(features)},
#'   \item \code{name} equal to the raster layer name, if available,
#'   \item otherwise a generated name of the form \code{"feature.1"},
#'   \code{"feature.2"}, and so on.
#' }
#'
#' \strong{Planning-unit identifiers}
#'
#' In vector and hybrid modes, planning-unit ids are taken from the column named
#' by \code{pu_id_col}. If that column is missing and \code{pu_id_col = "id"},
#' sequential ids are created with a warning.
#'
#' In raster-cell mode, planning-unit ids are always created sequentially from
#' the valid raster cells retained after masking and cost filtering.
#'
#' \strong{Ordering and alignment}
#'
#' In spatial modes, planning units, derived coordinates, raw attributes,
#' geometry, and extracted feature amounts are aligned to the same planning-unit
#' id order before the final \code{Problem} object is created.
#'
#' This alignment is critical because later functions assume that all stored
#' planning-unit components refer to the same ordered set of planning units.
#'
#' After \code{create_problem()}, typical next steps include adding actions,
#' spatial relations, targets or other constraints, objectives, and then solving
#' the resulting problem.
#'
#' @param pu Planning-units input. Depending on the selected method, this may be:
#'   \itemize{
#'     \item a \code{data.frame} with planning-unit information,
#'     \item an \code{sf} object,
#'     \item a \code{terra::SpatVector},
#'     \item a vector file path readable by \pkg{terra},
#'     \item a \code{terra::SpatRaster},
#'     \item or a raster file path.
#'   }
#' @param features Feature input. Depending on the selected workflow, this may be:
#'   \itemize{
#'     \item a \code{data.frame} with at least an \code{id} column,
#'     \item a \code{terra::SpatRaster} with one layer per feature,
#'     \item or a raster file path.
#'   }
#' @param dist_features Feature-distribution input. In tabular and hybrid
#'   workflows, this must be a \code{data.frame} with columns \code{pu},
#'   \code{feature}, and \code{amount}. In spatial modes it may be omitted or
#'   set to \code{NULL}, in which case it is derived automatically.
#' @param cost In spatial modes, planning-unit cost information. Depending on the
#'   input mode, this may be:
#'   \itemize{
#'     \item a column name in the planning-unit attribute table,
#'     \item a raster object,
#'     \item or a raster file path.
#'   }
#'   In raster-cell mode, \code{cost} is required and valid planning units are
#'   defined only where cost is finite and strictly positive.
#' @param pu_id_col Character string giving the name of the planning-unit id
#'   column in vector or \code{sf} inputs. Ignored in purely tabular mode and in
#'   raster-cell mode.
#' @param cost_aggregation Character string used in vector-PU mode when
#'   \code{cost} is a raster and must be aggregated over polygons. Must be one
#'   of \code{"mean"} or \code{"sum"}.
#' @param ... Additional arguments forwarded to internal builders.
#'
#' @return A \code{Problem} object for downstream \code{multiscape} workflows.
#'   The returned object always contains a canonical tabular core with planning
#'   units, features, and feature-distribution data, and may additionally
#'   contain aligned spatial metadata depending on the input mode.
#'
#' @examples
#' # ------------------------------------------------------
#' # 1) Tabular mode
#' # ------------------------------------------------------
#' pu_tbl <- data.frame(
#'   id = 1:4,
#'   cost = c(1, 2, 3, 4)
#' )
#'
#' feat_tbl <- data.frame(
#'   id = 1:2,
#'   name = c("feature_1", "feature_2")
#' )
#'
#' dist_feat_tbl <- data.frame(
#'   pu = c(1, 1, 2, 3, 4),
#'   feature = c(1, 2, 2, 1, 2),
#'   amount = c(5, 2, 3, 4, 1)
#' )
#'
#' p1 <- create_problem(
#'   pu = pu_tbl,
#'   features = feat_tbl,
#'   dist_features = dist_feat_tbl
#' )
#'
#' print(p1)
#'
#' # ------------------------------------------------------
#' # 2) Hybrid sf + tabular mode using package data
#' # ------------------------------------------------------
#'
#' p2 <- create_problem(
#'   pu = sim_pu,
#'   features = sim_features,
#'   dist_features = sim_dist_features,
#'   cost = "cost"
#' )
#'
#' print(p2)
#'
#' @seealso
#' \code{\link{add_actions}},
#' \code{\link{add_constraint_locked_pu}},
#' \code{\link{add_spatial_boundary}},
#' \code{\link{add_spatial_knn}},
#' \code{\link{solve}}
NULL

#' @export
#' @rdname create_problem
methods::setGeneric(
  "create_problem",
  signature = methods::signature("pu", "features", "dist_features"),
  function(
    pu,
    features,
    dist_features,
    cost = NULL,
    pu_id_col = "id",
    cost_aggregation = c("mean", "sum"),
    ...
  ) {
    standardGeneric("create_problem")
  }
)

# =========================================================
# Internal helpers (minimal, self-contained for spatial paths)
# =========================================================

.pa_is_raster_path <- function(x) {
  is.character(x) && length(x) == 1L && !is.na(x) && file.exists(x) &&
    grepl("\\.(tif|tiff|grd|img|nc|asc|sdat)$", tolower(x))
}

.pa_is_vector_path <- function(x) {
  is.character(x) && length(x) == 1L && !is.na(x) && file.exists(x) &&
    grepl("\\.(gpkg|shp|geojson|json|gml|kml)$", tolower(x))
}

.pa_read_rast <- function(x) {
  if (!requireNamespace("terra", quietly = TRUE)) return(NULL)
  if (inherits(x, "SpatRaster")) return(x)
  if (.pa_is_raster_path(x)) {
    return(tryCatch(terra::rast(x), error = function(e) NULL))
  }
  NULL
}

.pa_read_vect <- function(x) {
  if (!requireNamespace("terra", quietly = TRUE)) return(NULL)
  if (inherits(x, "SpatVector")) return(x)
  if (inherits(x, "sf")) return(tryCatch(terra::vect(x), error = function(e) NULL))
  if (.pa_is_vector_path(x)) {
    return(tryCatch(terra::vect(x), error = function(e) NULL))
  }
  NULL
}

.pa_fun_from_name <- function(x) {
  x <- match.arg(x, c("mean", "sum"))
  if (identical(x, "mean")) return(function(v) mean(v, na.rm = TRUE))
  function(v) sum(v, na.rm = TRUE)
}

# =========================================================
# Internal: fast raster-cell implementation (1 PU per valid cell)
# =========================================================
.pa_create_problem_raster_cells_impl <- function(
    pu,
    features,
    cost,
    ...
) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Raster-cell mode requires the 'terra' package.", call. = FALSE)
  }

  pu_r   <- .pa_read_rast(pu)
  feat_r <- .pa_read_rast(features)
  cost_r <- .pa_read_rast(cost)

  if (is.null(pu_r) || is.null(feat_r) || is.null(cost_r)) {
    stop(
      "Raster-cell mode requires `pu`, `features`, and `cost` to be terra::SpatRaster ",
      "objects or valid raster file paths.",
      call. = FALSE
    )
  }
  if (terra::nlyr(cost_r) != 1) stop("`cost` raster must have exactly 1 layer.", call. = FALSE)

  if (!terra::compareGeom(pu_r, cost_r, stopOnError = FALSE) ||
      !terra::compareGeom(pu_r, feat_r, stopOnError = FALSE)) {
    stop("`pu`, `features`, and `cost` rasters must share extent/resolution/CRS.", call. = FALSE)
  }

  cvals <- terra::values(cost_r, mat = FALSE)
  ok <- !is.na(terra::values(pu_r, mat = FALSE)) &
    !is.na(cvals) & is.finite(cvals) & (cvals > 0)

  idx_cells <- which(ok)
  if (!length(idx_cells)) {
    stop("No valid cells found after applying mask and cost rules (cost must be finite and > 0).", call. = FALSE)
  }

  pu_df <- data.frame(
    id = seq_along(idx_cells),
    cost = cvals[idx_cells],
    locked_in = FALSE,
    locked_out = FALSE,
    stringsAsFactors = FALSE
  )

  xy <- terra::xyFromCell(cost_r, idx_cells)
  pu_coords <- data.frame(id = pu_df$id, x = xy[, 1], y = xy[, 2], stringsAsFactors = FALSE)

  feat_names <- names(feat_r)
  if (is.null(feat_names) || any(feat_names == "")) {
    feat_names <- paste0("feature.", seq_len(terra::nlyr(feat_r)))
  }

  features_df <- data.frame(
    id = seq_len(terra::nlyr(feat_r)),
    name = feat_names,
    stringsAsFactors = FALSE
  )

  M <- terra::values(feat_r, mat = TRUE)
  M <- M[idx_cells, , drop = FALSE]
  M[is.na(M)] <- 0

  w <- which(M > 0, arr.ind = TRUE)
  dist_features_df <- if (!nrow(w)) {
    data.frame(pu = integer(0), feature = integer(0), amount = numeric(0))
  } else {
    data.frame(
      pu = w[, 1],
      feature = w[, 2],
      amount = M[w],
      stringsAsFactors = FALSE
    )
  }

  x <- .pa_create_problem_tabular_impl(
    pu = pu_df,
    features = features_df,
    dist_features = dist_features_df,
    boundary = NULL,
    ...
  )

  x$data$pu_coords <- pu_coords
  x$data$cell_index <- idx_cells
  x$data$pu_raster_mask <- pu_r
  x$data$cost_raster <- cost_r
  x$data$features_raster <- feat_r

  x
}

# =========================================================
# Method: TABULAR inputs
# =========================================================
#' @export
#' @rdname create_problem
methods::setMethod(
  "create_problem",
  methods::signature(pu = "data.frame", features = "data.frame", dist_features = "data.frame"),
  function(
    pu,
    features,
    dist_features,
    cost = NULL,
    pu_id_col = "id",
    cost_aggregation = c("mean", "sum"),
    ...
  ) {
    x <- .pa_create_problem_tabular_impl(
      pu = pu,
      features = features,
      dist_features = dist_features,
      boundary = NULL,
      ...
    )

    # preserve raw PU attributes for downstream helpers such as add_locked_pu()
    x$data$pu_data_raw <- pu

    x
  }
)

# =========================================================
# Method: SPATIAL inputs (dist_features missing)
# =========================================================
#' @export
#' @rdname create_problem
methods::setMethod(
  "create_problem",
  methods::signature(pu = "ANY", features = "ANY", dist_features = "missing"),
  function(
    pu,
    features,
    dist_features,
    cost = NULL,
    pu_id_col = "id",
    cost_aggregation = c("mean", "sum"),
    ...
  ) {
    cost_aggregation <- match.arg(cost_aggregation)

    if (is.null(cost)) {
      stop(
        "Spatial mode: `dist_features` is missing, so you must provide `cost` ",
        "(either a PU column name for vector PUs, or a raster/file path for spatial rasters).",
        call. = FALSE
      )
    }
    if (!requireNamespace("terra", quietly = TRUE)) {
      stop("Spatial mode requires the 'terra' package. Please install it.", call. = FALSE)
    }

    pu_is_r <- inherits(pu, "SpatRaster") || .pa_is_raster_path(pu)
    ft_is_r <- inherits(features, "SpatRaster") || .pa_is_raster_path(features)

    if (pu_is_r && ft_is_r) {
      return(.pa_create_problem_raster_cells_impl(
        pu = pu,
        features = features,
        cost = cost,
        ...
      ))
    }

    fun_cost <- .pa_fun_from_name(cost_aggregation)

    pu_r <- NULL
    pu_v <- NULL

    if (inherits(pu, "SpatRaster") || .pa_is_raster_path(pu)) {
      pu_r <- .pa_read_rast(pu)
      pu_r <- terra::round(pu_r)
      pu_v <- terra::as.polygons(pu_r, dissolve = TRUE, values = TRUE, na.rm = TRUE)
      names(pu_v) <- pu_id_col
    } else {
      pu_v <- .pa_read_vect(pu)
      if (is.null(pu_v)) stop("Could not read `pu` as a vector layer.", call. = FALSE)

      if (!(pu_id_col %in% names(pu_v))) {
        if (identical(pu_id_col, "id")) {
          warning(
            "Planning unit layer has no 'id' column. Creating sequential ids (1..n). ",
            "If you want to use an existing field, set pu_id_col to its name.",
            call. = FALSE, immediate. = TRUE
          )
          pu_v$id <- seq_len(nrow(pu_v))
          pu_id_col <- "id"
        } else {
          stop(
            "Planning unit vector is missing the id column: '", pu_id_col, "'. ",
            "Either provide that column or set pu_id_col to an existing column name.",
            call. = FALSE
          )
        }
      }
    }

    pu_df_raw <- terra::as.data.frame(pu_v)
    pu_df_raw$row_id <- seq_len(nrow(pu_df_raw))
    names(pu_df_raw)[names(pu_df_raw) == pu_id_col] <- "id"
    pu_df_raw$id <- as.integer(round(pu_df_raw$id))

    pu_df <- pu_df_raw

    if (is.character(cost) && (cost %in% names(pu_df))) {
      pu_df$cost <- pu_df[[cost]]
    } else {
      cost_r <- .pa_read_rast(cost)
      if (is.null(cost_r)) {
        stop("You must provide `cost` either as a PU column name (vector) or as a raster/file path.", call. = FALSE)
      }
      cost_ex <- terra::extract(cost_r, pu_v, fun = fun_cost, na.rm = TRUE)
      idx <- match(pu_df$row_id, cost_ex[[1]])
      pu_df$cost <- cost_ex[[2]][idx]
    }

    ctr <- terra::centroids(pu_v)
    xy  <- terra::crds(ctr)
    pu_coords <- data.frame(
      id = pu_df$id,
      x  = as.numeric(xy[, 1]),
      y  = as.numeric(xy[, 2]),
      stringsAsFactors = FALSE
    )

    feat_r <- .pa_read_rast(features)
    if (is.null(feat_r)) {
      stop(
        "Spatial mode: `features` must be a terra::SpatRaster (or raster file path) with one layer per feature.\n",
        "If you intended tabular mode, provide `dist_features` as a data.frame.",
        call. = FALSE
      )
    }

    feat_names <- names(feat_r)
    if (is.null(feat_names) || any(feat_names == "")) {
      feat_names <- paste0("feature.", seq_len(terra::nlyr(feat_r)))
      names(feat_r) <- feat_names
    }

    features_df <- data.frame(
      id = seq_len(terra::nlyr(feat_r)),
      name = feat_names,
      stringsAsFactors = FALSE
    )

    feat_mat <- .pa_fast_extract(feat_r, pu_v, fun = "sum")

    ord <- order(pu_df$id)
    pu_df     <- pu_df[ord, , drop = FALSE]
    pu_df_raw <- pu_df_raw[ord, , drop = FALSE]
    feat_mat  <- feat_mat[ord, , drop = FALSE]
    pu_coords <- pu_coords[ord, , drop = FALSE]

    pu_df_out <- pu_df[, c("id", "cost"), drop = FALSE]
    pu_df_out$locked_in <- FALSE
    pu_df_out$locked_out <- FALSE

    dist_features_df <- data.frame(
      pu      = rep(pu_df_out$id, times = terra::nlyr(feat_r)),
      feature = rep(features_df$id, each = nrow(pu_df_out)),
      amount  = as.vector(feat_mat),
      stringsAsFactors = FALSE
    )
    dist_features_df <- dist_features_df[
      is.finite(dist_features_df$amount) & dist_features_df$amount > 0,
      , drop = FALSE
    ]

    x <- .pa_create_problem_tabular_impl(
      pu = pu_df_out,
      features = features_df,
      dist_features = dist_features_df,
      boundary = NULL,
      ...
    )

    x$data$pu_coords <- pu_coords
    x$data$pu_data_raw <- pu_df_raw

    if (is.null(x$data$spatial_relations) || !is.list(x$data$spatial_relations)) {
      x$data$spatial_relations <- list()
    }
    if (!is.null(pu_r)) x$data$pu_raster_id <- pu_r

    if (requireNamespace("sf", quietly = TRUE)) {
      pu_sf_store <- tryCatch(sf::st_as_sf(pu_v), error = function(e) NULL)
      if (!is.null(pu_sf_store)) {
        if (pu_id_col %in% names(pu_sf_store)) names(pu_sf_store)[names(pu_sf_store) == pu_id_col] <- "id"
        if (!("id" %in% names(pu_sf_store))) pu_sf_store$id <- seq_len(nrow(pu_sf_store))

        pu_sf_store <- pu_sf_store[, "id", drop = FALSE]
        ord2 <- match(x$data$pu$id, pu_sf_store$id)

        if (any(is.na(ord2))) {
          warning(
            "Could not safely match PU geometry to PU ids; 'pu_sf' will not be stored in the problem object.",
            call. = FALSE, immediate. = TRUE
          )
        } else {
          x$data$pu_sf <- pu_sf_store[ord2, , drop = FALSE]
        }
      }
    }

    x
  }
)

# =========================================================
# Method: SPATIAL inputs (dist_features explicitly NULL)
# =========================================================
#' @export
#' @rdname create_problem
methods::setMethod(
  "create_problem",
  methods::signature(pu = "ANY", features = "ANY", dist_features = "NULL"),
  function(
    pu,
    features,
    dist_features,
    cost = NULL,
    pu_id_col = "id",
    cost_aggregation = c("mean", "sum"),
    ...
  ) {
    methods::selectMethod("create_problem", signature = c("ANY", "ANY", "missing"))(
      pu = pu,
      features = features,
      cost = cost,
      pu_id_col = pu_id_col,
      cost_aggregation = cost_aggregation,
      ...
    )
  }
)


# =========================================================
# Method: HYBRID sf + tabular features/dist_features
# =========================================================
#' @export
#' @rdname create_problem
methods::setMethod(
  "create_problem",
  methods::signature(pu = "ANY", features = "data.frame", dist_features = "data.frame"),
  function(
    pu,
    features,
    dist_features,
    cost = NULL,
    pu_id_col = "id",
    cost_aggregation = c("mean", "sum"),
    ...
  ) {

    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("This method requires the 'sf' package.", call. = FALSE)
    }

    pu_sf <- pu

    # ---- ensure id column exists
    if (!(pu_id_col %in% names(pu_sf))) {
      if (identical(pu_id_col, "id")) {
        warning(
          "Planning unit sf object has no 'id' column. Creating sequential ids (1..n). ",
          "If you want to use an existing field, set pu_id_col to its name.",
          call. = FALSE,
          immediate. = TRUE
        )
        pu_sf$id <- seq_len(nrow(pu_sf))
        pu_id_col <- "id"
      } else {
        stop(
          "Planning unit sf object is missing the id column: '", pu_id_col, "'. ",
          "Either provide that column or set pu_id_col to an existing column name.",
          call. = FALSE
        )
      }
    }

    # ---- raw PU attributes (without geometry)
    pu_df_raw <- sf::st_drop_geometry(pu_sf)
    names(pu_df_raw)[names(pu_df_raw) == pu_id_col] <- "id"
    pu_df_raw$id <- as.integer(round(pu_df_raw$id))

    # ---- cost must come from sf attribute table
    if (is.null(cost)) {
      if (!("cost" %in% names(pu_df_raw))) {
        stop(
          "For pu='sf' + features='data.frame' + dist_features='data.frame', ",
          "you must provide `cost` as a column name in the sf attribute table, ",
          "or include a column literally named 'cost'.",
          call. = FALSE
        )
      }
      pu_df_raw$cost <- as.numeric(pu_df_raw$cost)
    } else if (is.character(cost) && length(cost) == 1L && (cost %in% names(pu_df_raw))) {
      pu_df_raw$cost <- as.numeric(pu_df_raw[[cost]])
    } else {
      stop(
        "For pu='sf' + features='data.frame' + dist_features='data.frame', ",
        "`cost` must be the name of a column in the sf attribute table.",
        call. = FALSE
      )
    }

    if (any(!is.finite(pu_df_raw$cost))) {
      stop("PU cost column contains non-finite values.", call. = FALSE)
    }

    # ---- features checks
    if (!inherits(features, "data.frame")) {
      stop("`features` must be a data.frame.", call. = FALSE)
    }
    if (!("id" %in% names(features))) {
      stop("`features` must contain an 'id' column.", call. = FALSE)
    }
    features$id <- as.integer(features$id)

    if ("name" %in% names(features)) {
      if (anyNA(features$name) || any(!nzchar(as.character(features$name)))) {
        stop("`features$name` must not contain NA or empty strings.", call. = FALSE)
      }
      if (anyDuplicated(as.character(features$name)) != 0L) {
        dups <- unique(as.character(features$name)[duplicated(as.character(features$name))])
        stop(
          "`features$name` must be unique when feature names are used in `dist_features$feature`. ",
          "Duplicated names: ", paste(dups, collapse = ", "),
          call. = FALSE
        )
      }
    }

    # ---- dist_features checks
    if (!inherits(dist_features, "data.frame")) {
      stop("`dist_features` must be a data.frame.", call. = FALSE)
    }

    req_df_cols <- c("pu", "feature", "amount")
    miss_df_cols <- setdiff(req_df_cols, names(dist_features))
    if (length(miss_df_cols) > 0) {
      stop(
        "`dist_features` must contain columns: pu, feature, amount. Missing: ",
        paste(miss_df_cols, collapse = ", "),
        call. = FALSE
      )
    }

    dist_features$pu <- as.integer(dist_features$pu)
    dist_features$amount <- as.numeric(dist_features$amount)

    # ---- resolve dist_features$feature against features$id or features$name
    feature_raw <- dist_features$feature

    # case 1: already numeric ids
    feature_as_int <- suppressWarnings(as.integer(feature_raw))
    can_use_ids <- !anyNA(feature_as_int) &&
      all(feature_as_int %in% features$id)

    if (can_use_ids) {
      dist_features$feature <- feature_as_int

    } else {
      # case 2: try matching against features$name
      if (!("name" %in% names(features))) {
        stop(
          "`dist_features$feature` could not be matched to `features$id`, ",
          "and `features` has no `name` column to resolve feature names.",
          call. = FALSE
        )
      }

      feature_chr <- as.character(feature_raw)
      idx <- match(feature_chr, as.character(features$name))

      if (anyNA(idx)) {
        bad <- unique(feature_chr[is.na(idx)])
        stop(
          "`dist_features$feature` must contain either valid feature ids ",
          "from `features$id` or valid feature names from `features$name`.\n",
          "Unmatched values: ",
          paste(utils::head(bad, 20), collapse = ", "),
          if (length(bad) > 20) " ..." else "",
          call. = FALSE
        )
      }

      dist_features$feature <- as.integer(features$id[idx])
    }

    # ---- build PU table for tabular backend
    pu_df_out <- pu_df_raw[, c("id", "cost"), drop = FALSE]
    pu_df_out$locked_in <- FALSE
    pu_df_out$locked_out <- FALSE

    # ---- consistency checks
    bad_pu <- setdiff(unique(dist_features$pu), pu_df_out$id)
    if (length(bad_pu) > 0) {
      stop(
        "Some `dist_features$pu` ids are not present in `pu`: ",
        paste(utils::head(bad_pu, 20), collapse = ", "),
        if (length(bad_pu) > 20) " ..." else "",
        call. = FALSE
      )
    }

    bad_feat <- setdiff(unique(dist_features$feature), features$id)
    if (length(bad_feat) > 0) {
      stop(
        "Some `dist_features$feature` ids are not present in `features`: ",
        paste(utils::head(bad_feat, 20), collapse = ", "),
        if (length(bad_feat) > 20) " ..." else "",
        call. = FALSE
      )
    }

    # ---- build tabular problem
    x <- .pa_create_problem_tabular_impl(
      pu = pu_df_out,
      features = features,
      dist_features = dist_features,
      boundary = NULL,
      ...
    )

    # ---- preserve spatial information
    ctr <- suppressWarnings(sf::st_centroid(sf::st_geometry(pu_sf)))
    xy <- sf::st_coordinates(ctr)

    pu_coords <- data.frame(
      id = pu_df_out$id,
      x = as.numeric(xy[, 1]),
      y = as.numeric(xy[, 2]),
      stringsAsFactors = FALSE
    )

    pu_sf_store <- pu_sf
    if (pu_id_col %in% names(pu_sf_store)) {
      names(pu_sf_store)[names(pu_sf_store) == pu_id_col] <- "id"
    }
    pu_sf_store$id <- as.integer(round(pu_sf_store$id))

    ord <- match(x$data$pu$id, pu_sf_store$id)
    if (any(is.na(ord))) {
      warning(
        "Could not safely match PU geometry to PU ids; 'pu_sf' will not be stored in the problem object.",
        call. = FALSE,
        immediate. = TRUE
      )
    } else {
      x$data$pu_sf <- pu_sf_store[ord, , drop = FALSE]
    }

    x$data$pu_coords <- pu_coords
    x$data$pu_data_raw <- pu_df_raw

    if (is.null(x$data$spatial_relations) || !is.list(x$data$spatial_relations)) {
      x$data$spatial_relations <- list()
    }

    x
  }
)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_min_loss}
\alias{add_objective_min_loss}
\title{Add objective: minimize loss}
\usage{
add_objective_min_loss(x, actions = NULL, features = NULL, alias = NULL)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{actions}{Optional subset of actions to include in the objective. Values
may match \code{x$data$actions$id} and, if present,
\code{x$data$actions$action_set}. If \code{NULL}, all actions are included.}

\item{features}{Optional subset of features to include in the objective.
Values may match \code{x$data$features$id} and, if present,
\code{x$data$features$name}. If \code{NULL}, all features are included.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that minimizes the total negative effects generated by
selected actions on selected features.

This objective is based on the canonical effects table and uses only the
non-negative \code{loss} component.
}
\details{
Use this function when harmful ecological effects should be minimized
explicitly, without offsetting them against beneficial effects.

Let \eqn{\ell_{iaf} \ge 0} denote the stored loss associated with planning
unit \eqn{i}, action \eqn{a}, and feature \eqn{f}.

If no subsets are supplied, the objective can be written as:

\deqn{
\min \sum_{(i,a,f) \in \mathcal{R}} \ell_{iaf} \, x_{ia}.
}

where \eqn{\mathcal{R}} denotes the set of stored loss rows and
\eqn{x_{ia} \in \{0,1\}} indicates whether action \eqn{a} is selected in
planning unit \eqn{i}.

If \code{actions} is provided, only rows whose action belongs to the selected
subset contribute to the objective.

If \code{features} is provided, only rows whose feature belongs to the
selected subset contribute to the objective.

More generally, letting \eqn{\mathcal{R}^{\star}} be the subset induced by
the selected actions and features, the objective is:

\deqn{
\min \sum_{(i,a,f) \in \mathcal{R}^{\star}} \ell_{iaf} \, x_{ia}.
}

This objective minimizes harmful effects only. It does not offset losses
against benefits unless benefits are handled elsewhere through additional
objectives or constraints.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)
feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)
dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)
actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)
effects_df <- data.frame(
  pu = c(1, 2, 3, 4, 1, 2, 3, 4),
  action = c("conservation", "conservation", "conservation", "conservation",
             "restoration", "restoration", "restoration", "restoration"),
  feature = c(1, 1, 1, 1, 2, 2, 2, 2),
  benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
  loss = c(0, 0, 1, 0, 0, 1, 0, 0)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
  add_effects(effects_df)

p1 <- add_objective_min_loss(p)
p1$data$model_args

p2 <- add_objective_min_loss(
  p,
  actions = "restoration"
)
p2$data$model_args

p3 <- add_objective_min_loss(
  p,
  features = 1
)
p3$data$model_args

}
\seealso{
\code{\link{add_objective_max_benefit}},
\code{\link{add_effects}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{add_spatial_queen}
\alias{add_spatial_queen}
\title{Add queen adjacency from polygons}
\usage{
add_spatial_queen(x, geometry = NULL, name = "queen", weight = 1)
}
\arguments{
\item{x}{A \code{Problem} object created with
\code{\link{create_problem}} or another object containing aligned
planning-unit polygons.}

\item{geometry}{Optional \code{sf} object with planning-unit polygons and an
\code{id} column. If \code{NULL}, \code{x$data$pu_sf} is used.}

\item{name}{Character string giving the key under which the relation is
stored.}

\item{weight}{Numeric scalar giving the edge weight assigned to each queen
adjacency.}
}
\value{
An updated \code{Problem} object.
}
\description{
Build and register a queen adjacency relation from planning-unit polygons.

Two planning units are queen-adjacent if their boundaries touch, either along
a shared edge or at a shared vertex.
}
\details{
Use this function when neighbourhood should include both shared edges and
corner-touching polygon contacts.

This constructor derives an adjacency graph from polygon geometry using a
queen criterion. If planning units \eqn{i} and \eqn{j} touch at any boundary
point, then an edge \eqn{(i,j)} is added to the relation.

Let \eqn{G = (\mathcal{I}, E)} denote the resulting graph. Then:
\deqn{
(i,j) \in E \quad \Longleftrightarrow \quad
\partial i \cap \partial j \neq \varnothing.
}

Thus, queen adjacency includes all rook neighbours plus corner-touching
neighbours.

All edges receive the same user-supplied weight.

The resulting relation is stored as an undirected spatial relation.
}
\examples{
\donttest{
library(terra)

data("sim_pu_sf", package = "multiscape")
sim_features <- load_sim_features_raster()

p <- create_problem(
  pu = sim_pu_sf,
  features = sim_features,
  cost = "cost"
)

p <- add_spatial_queen(
  x = p,
  geometry = sim_pu_sf,
  name = "queen",
  weight = 1
)

head(p$data$spatial_relations$queen)
}

}
\seealso{
\code{\link{add_spatial_rook}},
\code{\link{add_spatial_boundary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_locked_pu.R
\name{add_constraint_locked_pu}
\alias{add_constraint_locked_pu}
\title{Add locked planning units to a problem}
\usage{
add_constraint_locked_pu(x, locked_in = NULL, locked_out = NULL)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}.}

\item{locked_in}{Optional locked-in specification. It may be \code{NULL}, a
column name in the raw planning-unit data, a logical vector, or a vector of
planning-unit ids.}

\item{locked_out}{Optional locked-out specification. It may be \code{NULL}, a
column name in the raw planning-unit data, a logical vector, or a vector of
planning-unit ids.}
}
\value{
An updated \code{Problem} object in which the planning-unit table
contains logical columns \code{locked_in} and \code{locked_out}.
}
\description{
Define planning units that must be included in, or excluded from, the
optimization problem.

This function updates the planning-unit table stored in the \code{Problem}
object by creating or replacing the logical columns \code{locked_in} and
\code{locked_out}. These columns are later used by the model builder when
translating the problem into optimization constraints.

Lock information may be supplied either directly as logical vectors, as
vectors of planning-unit ids, or by referencing columns in the raw
planning-unit data originally passed to \code{\link{create_problem}}.
}
\details{
Use this function when whole planning units must be forced into or out of
the solution, regardless of which action may later be selected in them.

Let \eqn{\mathcal{I}} denote the set of planning units and let
\eqn{w_i \in \{0,1\}} denote the binary variable indicating whether planning
unit \eqn{i \in \mathcal{I}} is selected by the model.

This function defines two subsets:
\itemize{
\item \eqn{\mathcal{I}^{in} \subseteq \mathcal{I}}, the planning units that
must be included,
\item \eqn{\mathcal{I}^{out} \subseteq \mathcal{I}}, the planning units
that must be excluded.
}

Conceptually, these sets correspond to the following conditions:
\itemize{
\item if \eqn{i \in \mathcal{I}^{in}}, then \eqn{w_i = 1},
\item if \eqn{i \in \mathcal{I}^{out}}, then \eqn{w_i = 0}.
}

These constraints are not imposed immediately by this function; instead, they
are stored in the planning-unit table and enforced later when building the
optimization model.

\strong{Philosophy}

The role of \code{\link{create_problem}} is to construct and normalize the basic
inputs of the planning problem. Locking planning units is treated as a
separate modelling step so that users can define or revise selection
restrictions after the \code{Problem} object has already been created.

In contrast, \code{\link{add_constraint_locked_actions}} is used to fix
specific feasible \code{(pu, action)} decisions rather than whole planning
units.

\strong{Supported input formats}

For both \code{locked_in} and \code{locked_out}, the function accepts:
\itemize{
\item \code{NULL}, meaning that no planning units are locked on that side,
\item a single character string, interpreted as a column name in the raw
planning-unit data,
\item a logical vector of length \code{nrow(pu)},
\item a vector of planning-unit ids.
}

When a column name is supplied, the referenced column is coerced to logical.
Numeric values are interpreted as non-zero = \code{TRUE}; character and
factor values are interpreted using common logical strings such as
\code{"true"}, \code{"t"}, \code{"1"}, \code{"yes"}, and \code{"y"}.
Missing values are treated as \code{FALSE}.

\strong{Replacement behaviour}

Each call to \code{add_constraint_locked_pu()} replaces any existing
\code{locked_in} and \code{locked_out} columns in the planning-unit table. In
other words, the function defines the complete current set of locked planning
units; it does not merge new values with previous ones.

\strong{Consistency checks}

The function checks that no planning unit is simultaneously assigned to both
\code{locked_in} and \code{locked_out}. If such conflicts are found, an error
is raised.
}
\examples{
pu <- data.frame(
  id = 1:5,
  cost = c(2, 3, 1, 4, 2),
  lock_col = c(TRUE, FALSE, FALSE, TRUE, FALSE),
  out_col = c(FALSE, FALSE, FALSE, FALSE, TRUE)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

# 1) Lock by planning-unit ids
p1 <- add_constraint_locked_pu(
  x = p,
  locked_in = c(1, 3),
  locked_out = c(5)
)

p1$data$pu[, c("id", "locked_in", "locked_out")]

# 2) Read lock information from raw PU data columns
p2 <- add_constraint_locked_pu(
  x = p,
  locked_in = "lock_col",
  locked_out = "out_col"
)

p2$data$pu[, c("id", "locked_in", "locked_out")]

# 3) Use logical vectors
p3 <- add_constraint_locked_pu(
  x = p,
  locked_in = c(TRUE, FALSE, TRUE, FALSE, FALSE),
  locked_out = c(FALSE, FALSE, FALSE, TRUE, FALSE)
)

p3$data$pu[, c("id", "locked_in", "locked_out")]

}
\seealso{
\code{\link{create_problem}},
\code{\link{add_actions}},
\code{\link{add_constraint_locked_actions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solve.R
\name{solve}
\alias{solve}
\alias{solve.Problem}
\title{Solve a planning problem}
\usage{
solve(x, ...)

\method{solve}{Problem}(x, ...)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}} and
optionally enriched with actions, effects, targets, constraints,
objectives, spatial relations, method settings, and solver settings.}

\item{...}{Additional arguments reserved for internal or legacy solver
handling. These are not part of the main recommended user interface.}
}
\value{
Either:
\itemize{
\item a \code{\link{solution-class}} object when solving a single-objective
problem, or
\item a \code{\link{solutionset-class}} object when solving a configured
multi-objective problem.
}

A \code{Solution} represents one optimization run. A \code{SolutionSet}
represents multiple runs together with their run table, design information,
and individual \code{Solution} objects.

After solving, users will typically inspect or visualize the returned results
through methods associated with \code{Solution} and
\code{SolutionSet} objects.
}
\description{
Solve a planning problem stored in a \code{Problem} object.

This is the main execution step of the \pkg{multiscape} workflow. It reads
the problem specification stored in a \code{Problem} object, builds the
corresponding optimization model when needed, applies the configured solver
settings, and returns either a \code{\link{solution-class}} or a
\code{\link{solutionset-class}} depending on whether the workflow is
single-objective or multi-objective.
}
\details{
\strong{Role of \code{solve()}}

The typical \pkg{multiscape} workflow is:
\preformatted{
x <- create_problem(...)
x <- add_...(x, ...)
x <- set_...(x, ...)
res <- solve(x)
}

Thus, \code{solve()} is the stage at which the stored problem specification
is turned into one or more optimization runs.

For most users, \code{solve()} is the standard execution entry point.
Explicit compilation with \code{compile_model()} is optional and is mainly
useful for advanced inspection or debugging workflows.

\strong{What \code{solve()} uses from the problem object}

The function uses the information already stored in the \code{Problem}
object, including:
\itemize{
\item baseline planning data,
\item actions, effects, profit, and spatial relations,
\item targets and constraints,
\item registered objectives,
\item an optional multi-objective method configuration,
\item solver settings.
}

If a model has not yet been built, it is built internally during the solve
process. If a model snapshot or pointer already exists, the solving layer may
reuse or refresh it depending on the internal model state.

\strong{Single-objective vs multi-objective behaviour}

The behaviour of \code{solve()} depends on the problem configuration.

\itemize{
\item \strong{Single-objective case.} If exactly one objective is active
and no multi-objective method is configured, \code{solve()} runs a single
optimization problem and returns a \code{\link{solution-class}} object.

\item \strong{Multi-objective case.} If a multi-objective method is
configured, \code{solve()} dispatches internally according to the stored
method name and returns a \code{\link{solutionset-class}} object.
}

Currently supported multi-objective method names are:
\itemize{
\item \code{"weighted"},
\item \code{"epsilon_constraint"},
\item \code{"augmecon"}.
}

\strong{Consistency rule}

If multiple objectives are registered but no multi-objective method has been
selected, \code{solve()} stops with an error. In practical terms:

\itemize{
\item one objective and no multi-objective method
\eqn{\Rightarrow} single-objective solve,
\item multiple objectives and a valid multi-objective method
\eqn{\Rightarrow} multi-objective solve,
\item multiple objectives and no multi-objective method
\eqn{\Rightarrow} error.
}

\strong{Solver settings}

Solver configuration is read from the \code{Problem} object, typically after
calling \code{\link{set_solver}} or one of its convenience wrappers such as
\code{\link{set_solver_gurobi}} or \code{\link{set_solver_cbc}}.

These settings may include:
\itemize{
\item the selected backend,
\item time limits,
\item optimality-gap settings,
\item CPU cores,
\item verbosity options,
\item backend-specific solver parameters.
}

\strong{Method dispatch}

\code{solve()} is an S3 generic. The public method documented here is
\code{solve.Problem()}, which operates on \code{Problem} objects.
}
\examples{
# ------------------------------------------------------------
# Minimal single-objective example
# ------------------------------------------------------------
pu <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions <- data.frame(
  id = c("conservation", "restoration")
)

effects <- data.frame(
  action = rep(c("conservation", "restoration"), each = 2),
  feature = rep(features$id, times = 2),
  multiplier = c(0.10, 0.10, 0.50, 0.50)
)

x <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
) |>
  add_actions(
    actions = actions,
    cost = c(conservation = 1, restoration = 2)
  ) |>
  add_effects(
    effects = effects,
    effect_type = "after"
  ) |>
  add_constraint_targets_relative(0.05) |>
  add_objective_min_cost(alias = "cost")

if (requireNamespace("rcbc", quietly = TRUE)) {
  x <- set_solver_cbc(x, verbose = FALSE)
  sol <- solve(x)
  print(sol)
}

# ------------------------------------------------------------
# Minimal multi-objective example
# ------------------------------------------------------------
x_mo <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
) |>
  add_actions(
    actions = actions,
    cost = c(conservation = 1, restoration = 2)
  ) |>
  add_effects(
    effects = effects,
    effect_type = "after"
  ) |>
  add_constraint_targets_relative(0.05) |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_max_benefit(alias = "benefit") |>
  set_method_weighted_sum(
    aliases = c("cost", "benefit"),
    weights = c(0.5, 0.5),
    normalize_weights = TRUE
  )

if (requireNamespace("rcbc", quietly = TRUE)) {
  x_mo <- set_solver_cbc(x_mo, verbose = FALSE)
  solset <- solve(x_mo)
  print(solset)
}

}
\seealso{
\code{\link{problem-class}},
\code{\link{solution-class}},
\code{\link{solutionset-class}},
\code{\link{compile_model}},
\code{\link{set_solver}},
\code{\link{set_solver_cbc}},
\code{\link{set_solver_gurobi}},
\code{\link{set_method_weighted_sum}},
\code{\link{set_method_epsilon_constraint}},
\code{\link{set_method_augmecon}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_effects.R
\name{add_effects}
\alias{add_effects}
\title{Add action effects to a planning problem}
\usage{
add_effects(
  x,
  effects = NULL,
  effect_type = c("delta", "after"),
  effect_aggregation = c("sum", "mean"),
  component = c("any", "benefit", "loss")
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}. It
must already contain feasible actions; run \code{\link{add_actions}} first.}

\item{effects}{Effect specification. One of:
\itemize{
\item \code{NULL}, to store an empty effects table,
\item a \code{data.frame(action, feature, multiplier)},
\item a \code{data.frame(pu, action, feature, ...)} with explicit effects,
\item a named list of \code{terra::SpatRaster} objects, one per action.
}}

\item{effect_type}{Character string indicating how supplied effect values are
interpreted. Must be one of:
\itemize{
\item \code{"delta"}: values represent signed net changes,
\item \code{"after"}: values represent after-action amounts and are
converted to net changes relative to baseline feature amounts.
}}

\item{effect_aggregation}{Character string giving the aggregation used when
converting raster values to planning-unit level. Must be one of
\code{"sum"} or \code{"mean"}.}

\item{component}{Character string controlling which component of the
canonical effects table is retained. Must be one of:
\itemize{
\item \code{"any"}: keep all non-zero rows,
\item \code{"benefit"}: keep only rows with \code{benefit > 0},
\item \code{"loss"}: keep only rows with \code{loss > 0}.
}}
}
\value{
An updated \code{Problem} object containing:
\describe{
\item{\code{dist_effects}}{A canonical effects table with columns
\code{pu}, \code{action}, \code{feature}, \code{benefit}, \code{loss},
\code{internal_pu}, \code{internal_action}, \code{internal_feature}, and
optional labels such as \code{feature_name} and \code{action_name}.}
\item{\code{effects_meta}}{Metadata describing how effects were
interpreted and stored.}
}
}
\description{
Define the effects of management actions on features across planning units.

Effects are stored in a canonical representation in a effects table, with one
row per \code{(pu, action, feature)} triple and two non-negative columns:
\itemize{
\item \code{benefit}: the positive component of the effect,
\item \code{loss}: the magnitude of the negative component of the effect.
}

The net effect is therefore interpreted as
\deqn{
\Delta_{iaf} = \mathrm{benefit}_{iaf} - \mathrm{loss}_{iaf},
}
where \eqn{i} indexes planning units, \eqn{a} indexes actions, and
\eqn{f} indexes features.

Under the semantics adopted by this package, each
\code{(pu, action, feature)} triple represents a single net effect.
Consequently, after validation and aggregation, a stored row cannot have both
\code{benefit > 0} and \code{loss > 0} at the same time.
}
\details{
\strong{When to use \code{add_effects()}.}

Use this function when you want to specify what feasible actions do to
features. It is the stage at which an action-based decision space is linked
to feature-level ecological or functional consequences.

This function provides a unified interface for specifying action effects from
several input formats while enforcing a single internal representation.
Regardless of how the user supplies the effects, the stored output always
follows the same canonical structure based on non-negative
\code{benefit}/\code{loss} components.

Let \eqn{i \in \mathcal{I}} index planning units,
\eqn{a \in \mathcal{A}} index actions, and
\eqn{f \in \mathcal{F}} index features.
Let \eqn{b_{if}} denote the baseline amount of feature \eqn{f} in planning
unit \eqn{i}, as given by the feature-distribution table. Let
\eqn{\Delta_{iaf}} denote the net change caused by applying action
\eqn{a} in planning unit \eqn{i} to feature \eqn{f}. The canonical stored
representation is:

\deqn{
\mathrm{benefit}_{iaf} = \max(\Delta_{iaf}, 0),
}

\deqn{
\mathrm{loss}_{iaf} = \max(-\Delta_{iaf}, 0).
}

Hence:
\itemize{
\item if \eqn{\Delta_{iaf} > 0}, then \code{benefit > 0} and \code{loss = 0},
\item if \eqn{\Delta_{iaf} < 0}, then \code{benefit = 0} and \code{loss > 0},
\item if \eqn{\Delta_{iaf} = 0}, then both are zero.
}

\strong{Why split effects into benefit and loss?}

This representation avoids ambiguity in downstream optimization models. It
allows the package to support, for example, objectives that maximize
beneficial effects, minimize damages, impose no-net-loss conditions, or
combine both components differently in multi-objective formulations.

\strong{Supported effect specifications}

The \code{effects} argument may be provided in one of the following forms:

\enumerate{
\item \code{NULL}. An empty effects table is stored.

\item A \code{data.frame(action, feature, multiplier)}. In this case,
effects are constructed by multiplying baseline feature amounts by the
supplied multiplier:
\deqn{
  \Delta_{iaf} = b_{if} \times m_{af},
  }
where \eqn{m_{af}} is the multiplier associated with action \eqn{a} and
feature \eqn{f}. This specification is expanded over all feasible
\code{(pu, action)} pairs.

\item A \code{data.frame(pu, action, feature, ...)} giving explicit effects
for individual triples. The table may contain:
\itemize{
\item \code{delta} or \code{effect}: interpreted as signed net changes,
\item \code{after}: interpreted as after-action amounts if
\code{effect_type = "after"},
\item \code{benefit} and/or \code{loss}: explicit non-negative split
components,
\item legacy signed \code{benefit} without \code{loss}: interpreted as a
signed net effect for backwards compatibility.
}

\item A named list of \code{terra::SpatRaster} objects, one per action. In
this case, names must match action ids, and each raster must contain one
layer per feature. Raster values are aggregated to planning-unit level
using \code{effect_aggregation}.
}

\strong{Interpretation of \code{effect_type}}

If \code{effect_type = "delta"}, supplied values are interpreted as net
changes directly.

If \code{effect_type = "after"}, supplied values are interpreted as
after-action amounts and converted internally to net effects using:

\deqn{
\Delta_{iaf} = \mathrm{after}_{iaf} - b_{if}.
}

Missing baseline values are treated as zero.

\strong{Feasibility and locked-out decisions}

Effects are only retained for feasible \code{(pu, action)} pairs. Thus,
\code{add_actions()} must be called first. Pairs marked as locked out
(\code{status == 3}) are removed before storing the final effects table.

This function does not define the action-decision layer itself; it builds on
the feasible \code{(pu, action)} pairs already stored in the problem.

\strong{Duplicate rows and semantic validation}

If multiple rows are supplied for the same \code{(pu, action, feature)}
triple, they are aggregated by summing \code{benefit} and \code{loss}
separately. The resulting triple must still respect the package semantics,
namely that both components cannot be strictly positive simultaneously.
Inputs violating this rule are rejected.

\strong{Component filtering}

After canonicalization and validation, rows can be restricted to:
\itemize{
\item \code{component = "any"}: keep all non-zero effects,
\item \code{component = "benefit"}: keep only rows with \code{benefit > 0},
\item \code{component = "loss"}: keep only rows with \code{loss > 0}.
}

Rows with \code{benefit = 0} and \code{loss = 0} are always removed.

\strong{Raster handling}

When effects are supplied as rasters, they are automatically aligned to the
planning-unit raster or geometry when needed before extraction or zonal
aggregation.

\strong{Stored output}

The resulting effects table contains user-facing ids, internal integer ids,
and optional labels for actions and features. Metadata describing the stored
representation and input interpretation are written to an effects metadata
field.

After defining effects, typical next steps include adding objectives that use
beneficial or harmful effects, and then solving the configured problem.
}
\examples{
# ------------------------------------------------------
# Minimal problem with actions
# ------------------------------------------------------
pu <- data.frame(
  id = 1:3,
  cost = c(1, 2, 3)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3),
  feature = c(1, 2, 1, 2),
  amount = c(10, 5, 8, 4)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

actions <- data.frame(
  id = c("conservation", "restoration"),
  name = c("Conservation", "Restoration")
)

p <- add_actions(
  x = p,
  actions = actions,
  cost = c(conservation = 2, restoration = 4)
)

print(p)

# ------------------------------------------------------
# 1) Empty effects
# ------------------------------------------------------
p0 <- add_effects(p, effects = NULL)
p0$data$dist_effects

# ------------------------------------------------------
# 2) Multipliers by action and feature
# delta = baseline amount * multiplier
# ------------------------------------------------------
mult <- data.frame(
  action = c("conservation", "conservation",
             "restoration", "restoration"),
  feature = c("sp1", "sp2", "sp1", "sp2"),
  multiplier = c(0.10, 0.00, -0.20, 0.25)
)

p1 <- add_effects(
  x = p,
  effects = mult,
  effect_type = "delta"
)

p1$data$dist_effects

# ------------------------------------------------------
# 3) Explicit net effects using signed deltas
# ------------------------------------------------------
eff_delta <- data.frame(
  pu = c(1, 2, 3, 3),
  action = c("conservation", "restoration", "restoration", "conservation"),
  feature = c(1, 1, 2, 2),
  delta = c(2, -3, 1, 0.5)
)

p2 <- add_effects(
  x = p,
  effects = eff_delta
)

p2$data$dist_effects

# ------------------------------------------------------
# 4) Explicit after-action amounts
# delta = after - baseline
# ------------------------------------------------------
eff_after <- data.frame(
  pu = c(1, 2, 3),
  action = c("conservation", "restoration", "restoration"),
  feature = c(1, 1, 2),
  after = c(12, 5, 6)
)

p3 <- add_effects(
  x = p,
  effects = eff_after,
  effect_type = "after"
)

p3$data$dist_effects

# ------------------------------------------------------
# 5) Keep only beneficial effects
# ------------------------------------------------------
p4 <- add_effects(
  x = p,
  effects = eff_delta,
  component = "benefit"
)

p4$data$dist_effects

}
\seealso{
\code{\link{add_actions}},
\code{\link{add_benefits}},
\code{\link{add_losses}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_solution.R
\name{get_targets}
\alias{get_targets}
\title{Get target achievement summary from a solution}
\usage{
get_targets(x, run = NULL)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object returned by
\code{\link{solve}}.}

\item{run}{Optional positive integer giving the run index to extract from a
\code{SolutionSet}. If \code{NULL}, all runs are returned when available.}
}
\value{
A simplified \code{data.frame} target summary, or \code{NULL} if the
solution does not contain targets. Typical columns include
\code{feature}, \code{feature_name}, \code{target_level},
\code{total_available}, \code{target}, \code{achieved}, \code{gap}, and
\code{met}.
}
\description{
Extract a user-facing target-achievement table from a \code{Solution} or
\code{SolutionSet} object returned by \code{\link{solve}}.

The returned table summarizes, for each stored target, the target level, the
achieved value in the solution, the gap between achieved and required values,
and whether the target was met.
}
\details{
Targets are optional in \code{multiscape}. If the solution object does not contain
a targets summary table at \code{x$summary$targets}, this function returns
\code{NULL} without error.

This function reads the stored targets summary and returns a simplified
user-facing table. If the summary contains \code{achieved} and
\code{target_value}, target satisfaction is evaluated as follows.

For lower-bound targets:
\deqn{
\mathrm{met} = (\mathrm{achieved} \ge \mathrm{target}),
}

and for upper-bound targets:
\deqn{
\mathrm{met} = (\mathrm{achieved} \le \mathrm{target}).
}

The interpretation of the target direction is taken from the \code{sense}
column when available:
\itemize{
\item \code{"ge"}, \code{">="}, or \code{"min"} are treated as lower-bound
targets,
\item \code{"le"}, \code{"<="}, or \code{"max"} are treated as upper-bound
targets,
\item if \code{sense} is missing, the target is treated as a lower bound by
default.
}

The returned table is simplified and renames some internal fields for
readability:
\itemize{
\item \code{target_raw} is returned as \code{target_level},
\item \code{basis_total} is returned as \code{total_available},
\item \code{target_value} is returned as \code{target}.
}

If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
belonging to that run are returned. If the result contains a \code{run_id}
column but only a single run is present and \code{run} was not requested
explicitly, the \code{run_id} column is removed for convenience.

The \code{gap} column is expected to be part of the stored summary. When
present, it typically represents:
\deqn{
\mathrm{gap} = \mathrm{achieved} - \mathrm{target}.
}
}
\examples{
\donttest{
if (requireNamespace("rcbc", quietly = TRUE)) {
  pu_tbl <- data.frame(
    id = 1:4,
    cost = c(1, 2, 3, 4)
  )

  feat_tbl <- data.frame(
    id = 1:2,
    name = c("feature_1", "feature_2")
  )

  dist_feat_tbl <- data.frame(
    pu = c(1, 1, 2, 3, 4),
    feature = c(1, 2, 2, 1, 2),
    amount = c(5, 2, 3, 4, 1)
  )

  actions_df <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  effects_df <- data.frame(
    pu = c(1, 2, 3, 4),
    action = "conservation",
    feature = c(1, 1, 2, 2),
    benefit = c(2, 1, 1, 2),
    loss = c(0, 0, 0, 0)
  )

  p <- create_problem(
    pu = pu_tbl,
    features = feat_tbl,
    dist_features = dist_feat_tbl,
    cost = "cost"
  ) |>
    add_actions(actions_df, cost = 0) |>
    add_effects(effects_df) |>
    add_constraint_targets_relative(0.2) |>
    add_objective_min_cost() |>
    set_solver_cbc(time_limit = 10)

  sol <- solve(p)

  get_targets(sol)
}
}

}
\seealso{
\code{\link{get_pu}},
\code{\link{get_actions}},
\code{\link{get_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/problem-class.R
\name{problem-class}
\alias{problem-class}
\alias{Problem}
\title{Problem class}
\value{
No return value. This page documents the \code{Problem} class.
}
\description{
The \code{Problem} class is the central container used by \pkg{multiscape} to
represent a planning problem before, during, and after model construction.

A \code{Problem} object stores the full problem specification in a modular
way. This includes the baseline planning data, optional spatial metadata,
action definitions, effects, profit, targets, constraints, objective
registrations, solver settings, and, when available, a built optimization
model or model snapshot.

In other words, \code{Problem} is the persistent object that connects the full
\code{multiscape} workflow:
\preformatted{
create_problem()
-> add_*() / set_*()
-> solve()
}
}
\details{
\strong{Conceptual role}

The \code{Problem} class is designed for a data-first and modular workflow.
User-facing functions do not usually modify a solver object directly. Instead,
they enrich the \code{Problem} object by storing new specifications in its
internal \code{data} field.

Thus, a \code{Problem} object should be understood as a structured container
for the mathematical planning problem, not necessarily as a built
optimization model.

Before \code{\link{solve}} is called, the object may contain only input data
and user specifications. During or after solving, it may additionally contain
a built model pointer, model metadata, and solver-related information.

\strong{Core mathematical interpretation}

At a high level, the \code{Problem} object stores the ingredients required to
define an optimization problem over:
\itemize{
\item planning units \eqn{i \in \mathcal{P}},
\item features \eqn{f \in \mathcal{F}},
\item actions \eqn{a \in \mathcal{A}},
\item optional spatial relations over planning units,
\item and user-defined objectives and constraints.
}

The baseline ecological state is typically stored through a planning
unit--feature table of amounts:
\deqn{
a_{if} \ge 0,
}
where \eqn{a_{if}} is the baseline amount of feature \eqn{f} in planning unit
\eqn{i}.

Subsequent functions then add action feasibility, effects, profit, targets,
spatial relations, and optimization settings to this baseline representation.

\strong{How objects are created}

\code{Problem} objects are usually created by \code{\link{create_problem}}.

After creation, downstream functions such as \code{\link{add_actions}},
\code{\link{add_effects}}, \code{\link{add_profit}},
\code{\link{add_constraint_targets_absolute}}, \code{\link{add_constraint_targets_relative}},
spatial relation constructors, objective setters, and solver setters extend
the internal \code{data} list.

\strong{Internal storage}

The class contains a single field:
\describe{
\item{\code{data}}{A named \code{list} storing the full problem
specification, metadata, and, when available, built-model information.}
}

Common entries of \code{data} include:
\describe{
\item{\code{pu}}{Planning-unit table.}
\item{\code{features}}{Feature table.}
\item{\code{actions}}{Action catalog.}
\item{\code{dist_features}}{Planning unit--feature baseline amounts.}
\item{\code{dist_actions}}{Feasible planning unit--action pairs.}
\item{\code{dist_effects}}{Action effects by planning unit, action, and
feature.}
\item{\code{dist_profit}}{Profit by planning unit--action pair.}
\item{\code{pu_sf}}{Planning-unit geometry when available.}
\item{\code{pu_coords}}{Planning-unit coordinates when available.}
\item{\code{spatial_relations}}{Registered spatial relations.}
\item{\code{targets}}{Stored target specifications.}
\item{\code{constraints}}{Stored user-defined constraints.}
\item{\code{objectives}}{Registered atomic objectives for single- or
multi-objective workflows.}
\item{\code{method}}{Stored multi-objective method configuration, when
applicable.}
\item{\code{solve_args}}{Stored solver settings.}
\item{\code{model_ptr}}{Pointer to a built optimization model, when
available.}
\item{\code{model_args}}{Metadata describing model construction.}
\item{\code{model_list}}{Optional exported model snapshot or representation.}
\item{\code{meta}}{Auxiliary metadata, including model-dirty flags and other
bookkeeping fields.}
}

Not every \code{Problem} object contains all of these entries. The content of
\code{data} depends on how far the workflow has progressed.

\strong{Lifecycle}

A \code{Problem} object typically moves through the following stages:
\enumerate{
\item input stage: baseline planning units, features, and feature
distributions are stored,
\item specification stage: actions, effects, targets, objectives,
constraints, and spatial relations are added,
\item model stage: an optimization model is built from the stored
specification,
\item solve stage: the model is solved and results are returned in a
separate \code{Solution} or \code{SolutionSet} object.
}

The \code{Problem} object itself is not the solution. It is the structured
problem definition from which a solution can be obtained.
}
\section{Methods}{

\describe{
\item{\code{print()}}{Print a structured summary of the stored problem
specification, including data, actions and effects, spatial inputs, targets
and constraints, and model status. If a model has already been built,
additional dimensions and auxiliary-variable information are displayed.}

\item{\code{show()}}{Alias of \code{print()}.}

\item{\code{repr()}}{Return a short one-line representation of the object.}

\item{\code{getData(name)}}{Return a named entry from \code{self$data}.}

\item{\code{getPlanningUnitsAmount()}}{Return the number of planning units
stored in \code{x$data$pu}.}

\item{\code{getMonitoringCosts()}}{Return the planning-unit cost vector,
typically taken from \code{x$data$pu$cost}.}

\item{\code{getFeatureAmount()}}{Return the number of stored features.}

\item{\code{getFeatureNames()}}{Return feature names from
\code{x$data$features$name}, or feature ids if names are unavailable.}

\item{\code{getActionCosts()}}{Return action-level costs from
\code{x$data$dist_actions$cost} when available.}

\item{\code{getActionsAmount()}}{Return the number of stored actions.}
}
}

\section{Printing and diagnostics}{

The \code{print()} method is intended as a quick diagnostic summary. It helps
users understand:
\itemize{
\item what data have already been loaded,
\item whether actions, effects, spatial relations, targets, and constraints
have been registered,
\item whether objectives and methods have been configured,
\item whether a model has already been built,
\item and whether the object appears ready to be solved.
}

In particular, the model section of the printed output summarizes whether the
current problem specification is incomplete, ready, or already materialized as
a built optimization model.
}

\seealso{
\code{\link{create_problem}},
\code{\link{add_actions}},
\code{\link{add_effects}},
\code{\link{add_constraint_targets_absolute}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-sim.R
\docType{data}
\name{sim_dist_features}
\alias{sim_dist_features}
\title{Simulated feature distribution}
\format{
A data frame linking planning units and features with an \code{amount} column.
}
\usage{
data(sim_dist_features)
}
\description{
Example distribution of feature amounts across planning units.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_action_max_per_pu.R
\name{add_action_max_per_pu}
\alias{add_action_max_per_pu}
\title{Limit the number of actions per planning unit (maximum)}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}. Must already contain actions (i.e., run
\code{\link{add_actions}} first).}

\item{max}{Integer scalar \eqn{\ge 0}. Maximum number of actions allowed per PU.
Default is \code{1L}.}

\item{pu}{Optional integer vector of planning unit ids (external ids, i.e.,
\code{x$data$pu$id}) to which the constraint is applied. If \code{NULL} (default),
the constraint applies to all PUs.}

\item{actions}{Optional character vector of action ids (i.e., \code{x$data$actions$id})
to include in the sum. If \code{NULL} (default), all actions are included.}

\item{overwrite}{Logical. If \code{TRUE}, replace any existing \code{action_max_per_pu}
constraint stored in \code{x$data$constraints}. If \code{FALSE} (default) and a constraint
already exists, an error is raised.}
}
\value{
The updated \code{Problem} object with \code{x$data$constraints$action_max_per_pu}
set to a list containing the constraint specification.
}
\description{
Store a constraint that limits the number of actions that can be selected within each
planning unit (PU). The constraint has the form:
\deqn{\sum_{a \in A} x_{pu,a} \le \mathrm{max}}
for each PU in the selected set.

This function is \strong{data-only}: it records the constraint specification in the
\code{Problem} object but does not build or modify the optimization model. The constraint
is later translated into linear constraints by the model builder (e.g.,
\code{.pa_build_model_apply_constraints()}).
}
\details{
The constraint can be applied to:
\itemize{
\item all planning units and all actions (default), or
\item a subset of planning units via \code{pu}, and/or
\item a subset of actions via \code{actions}.
}

Note that the subset is applied over the existing feasible \code{(pu, action)} pairs
in \code{x$data$dist_actions}. If the requested \code{pu}/\code{actions} subset yields
zero feasible pairs, the function stops with an informative error.

If \code{x$data$model_ptr} is already present (a model has been built previously),
the function marks the model as dirty by setting \code{x$data$meta$model_dirty <- TRUE},
signalling that the model should be rebuilt before solving.
}
\seealso{
\code{\link{add_actions}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_spatial}
\alias{plot_spatial}
\title{Plot spatial outputs from a solution or solution set}
\usage{
plot_spatial(
  x,
  what = c("pu", "actions", "features"),
  runs = NULL,
  actions = NULL,
  features = NULL,
  value = c("final", "baseline", "benefit"),
  layout = NULL,
  max_facets = 4L,
  ...,
  base_alpha = 0.1,
  selected_alpha = 0.9,
  base_fill = "grey92",
  base_color = NA,
  selected_color = NA,
  draw_borders = FALSE,
  show_base = TRUE,
  fill_values = NULL,
  fill_na = "grey80",
  use_viridis = TRUE
)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object.}

\item{what}{Character string indicating what to plot. Must be one of
\code{"pu"}, \code{"actions"}, or \code{"features"}.}

\item{runs}{Optional integer vector of run ids. If \code{NULL}, a
\code{Solution} is used directly and a \code{SolutionSet} defaults to the
first run.}

\item{actions}{Optional action subset used when \code{what = "actions"}.}

\item{features}{Optional feature subset used when \code{what = "features"}.}

\item{value}{Character string used only when \code{what = "features"}.
Must be one of \code{"final"}, \code{"baseline"}, or \code{"benefit"}.}

\item{layout}{Character string controlling the layout. Must be one of
\code{"single"} or \code{"facet"}. If \code{NULL}, the default is
\code{"single"} for planning units and actions, and \code{"facet"} for
features.}

\item{max_facets}{Maximum number of facets shown when faceting without an
explicit action or feature subset.}

\item{...}{Additional arguments passed to the specialised plotting function.}

\item{base_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the base
planning-unit layer.}

\item{selected_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the
highlighted layer.}

\item{base_fill}{Fill colour for the base planning-unit layer.}

\item{base_color}{Border colour for the base planning-unit layer.}

\item{selected_color}{Border colour for highlighted layers.}

\item{draw_borders}{Logical. If \code{FALSE}, borders are not drawn.}

\item{show_base}{Logical. If \code{TRUE}, draw the base planning-unit layer
underneath the highlighted output.}

\item{fill_values}{Optional named vector of colours for discrete maps.}

\item{fill_na}{Fill colour for missing values.}

\item{use_viridis}{Logical. If \code{TRUE} and the \pkg{viridis} package is
available, use viridis scales.}
}
\value{
Invisibly returns a \code{ggplot} object.
}
\description{
Convenience wrapper to plot spatial outputs from a \code{Solution} or
\code{SolutionSet}.

Depending on \code{what}, this function dispatches to one of:
\itemize{
\item \code{\link{plot_spatial_pu}},
\item \code{\link{plot_spatial_actions}},
\item \code{\link{plot_spatial_features}}.
}

This wrapper is useful as a compact entry point, while the specialised
plotting functions provide a cleaner and more explicit user interface for
each spatial output type.
}
\examples{
if (requireNamespace("sf", quietly = TRUE) &&
    requireNamespace("ggplot2", quietly = TRUE)) {
  data("sim_pu_sf", package = "multiscape")

  problem <- structure(
    list(
      data = list(
        pu_sf = sim_pu_sf
      )
    ),
    class = "Problem"
  )

  sol <- structure(
    list(
      problem = problem,
      summary = list(
        pu = data.frame(
          id = sim_pu_sf$id,
          selected = as.integer(seq_len(nrow(sim_pu_sf)) \%\% 2 == 1)
        )
      )
    ),
    class = "Solution"
  )

  plot_spatial(sol, what = "pu")
}

}
\seealso{
\code{\link{plot_spatial_pu}},
\code{\link{plot_spatial_actions}},
\code{\link{plot_spatial_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_min_fragmentation_action}
\alias{add_objective_min_fragmentation_action}
\title{Add objective: minimize action fragmentation}
\usage{
add_objective_min_fragmentation_action(
  x,
  relation_name = "boundary",
  weight_multiplier = 1,
  action_weights = NULL,
  actions = NULL,
  alias = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{relation_name}{Character string giving the name of the spatial relation
to use. The relation must already exist in
\code{x$data$spatial_relations}.}

\item{weight_multiplier}{Numeric scalar greater than or equal to zero. Global
multiplier applied to the relation weights when the objective is built.}

\item{action_weights}{Optional action weights. Either a named numeric vector
with names equal to action ids, or a \code{data.frame} with columns
\code{action} and \code{weight}. These weights scale the contribution of
each action to the final objective.}

\item{actions}{Optional subset of actions to include. Values may match
\code{x$data$actions$id} and, if present,
\code{x$data$actions$action_set}. If \code{NULL}, all actions are included.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that minimizes fragmentation at the action level over a
stored spatial relation.

Unlike \code{\link{add_objective_min_fragmentation_pu}}, which acts on the
selected planning-unit set through \eqn{w_i}, this objective acts on the
spatial arrangement of individual action decisions through the action
variables \eqn{x_{ia}}.
}
\details{
Use this function when spatial cohesion should be encouraged separately for
each selected action pattern.

Let \eqn{\mathcal{I}} denote the set of planning units and let
\eqn{\mathcal{A}} denote the set of actions.

Let \eqn{x_{ia} \in \{0,1\}} indicate whether action \eqn{a \in \mathcal{A}}
is selected in planning unit \eqn{i \in \mathcal{I}}.

Let the chosen spatial relation define weighted pairs with
weights \eqn{\omega_{ij} \ge 0}, and let
\eqn{\lambda =} \code{weight_multiplier} be the global scaling factor applied
to these weights.

If \code{actions} is supplied, only the selected subset
\eqn{\mathcal{A}^{\star} \subseteq \mathcal{A}} contributes to the final
objective. If \code{actions = NULL}, all actions are included.

The internal preparation step constructs one auxiliary variable
\eqn{b_{ija} \in [0,1]} for each unique non-diagonal undirected edge
\eqn{(i,j)} with \eqn{i < j} and for each action \eqn{a}. The intended
semantics is:
\deqn{
b_{ija} = x_{ia} \land x_{ja}.
}

Whenever both decision variables \eqn{x_{ia}} and \eqn{x_{ja}} exist in the
model, this conjunction is enforced by the linearization:
\deqn{
b_{ija} \le x_{ia},
}
\deqn{
b_{ija} \le x_{ja},
}
\deqn{
b_{ija} \ge x_{ia} + x_{ja} - 1.
}

If one of the two action variables does not exist because the corresponding
\code{(pu, action)} pair is not feasible, the auxiliary variable is forced to
zero.

Therefore, \eqn{b_{ija}=1} if and only if action \eqn{a} is selected in both
adjacent planning units \eqn{i} and \eqn{j}; otherwise \eqn{b_{ija}=0}.

The exact objective coefficients are assembled later by the model builder
from:
\itemize{
\item the action decision variables \eqn{x_{ia}},
\item the edge-conjunction variables \eqn{b_{ija}},
\item the relation weights \eqn{\omega_{ij}},
\item the multiplier \eqn{\lambda},
\item and, if supplied, the action-specific weights.
}

If action-specific weights are provided, let \eqn{\alpha_a \ge 0} denote the
weight associated with action \eqn{a}. Then the resulting objective can be
interpreted as an action-wise compactness or fragmentation functional of the
form:
\deqn{
\min \sum_{a \in \mathcal{A}^{\star}} \alpha_a \,
F_a(x_{\cdot a}, b_{\cdot\cdot a}; \lambda \omega),
}
where \eqn{F_a} is the fragmentation expression induced by the selected
relation and the internal coefficient construction for action \eqn{a}.

In practical terms, this objective penalizes solutions in which the same
action is spatially scattered or broken into separate patches, while allowing
different actions to form different spatial patterns.

This differs from planning-unit fragmentation:
\itemize{
\item \code{add_objective_min_fragmentation_pu()} encourages cohesion of the
union of selected planning units,
\item \code{add_objective_min_fragmentation_action()} encourages cohesion of
each selected action pattern separately.
}

Setting \code{weight_multiplier = 0} removes the contribution of the spatial
relation from the objective after scaling.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

bound_df <- data.frame(
  id1 = c(1, 1, 2, 1, 2, 3, 4),
  id2 = c(1, 2, 2, 3, 4, 4, 4),
  boundary = c(4, 1, 4, 1, 1, 1, 4)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2))

p <- add_spatial_boundary(
  x = p,
  boundary = bound_df,
  name = "boundary",
  include_self = TRUE,
  edge_factor = 1
)

p <- add_objective_min_fragmentation_action(
  p,
  relation_name = "boundary",
  actions = "restoration",
  weight_multiplier = 1
)

p$data$model_args

}
\seealso{
\code{\link{add_objective_min_fragmentation_pu}},
\code{\link{add_spatial_boundary}},
\code{\link{add_spatial_relations}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compile_model.R
\name{compile_model}
\alias{compile_model}
\alias{compile_model.Problem}
\title{Compile the optimization model stored in a Problem}
\usage{
compile_model(x, force = FALSE, ...)

\method{compile_model}{Problem}(x, force = FALSE, ...)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{force}{Logical. If \code{TRUE}, rebuild the model even if a current
compiled model already exists.}

\item{...}{Reserved for future extensions.}
}
\value{
A \code{Problem} object with compiled model structures stored internally.
}
\description{
Materializes the optimization model represented by a \code{Problem} object
without solving it. This is an advanced function mainly intended for
debugging, inspection, and explicit model preparation.

In standard workflows, users normally do not need to call this function,
because \code{\link{solve}} compiles the model automatically when needed.
}
\details{
Use this function when you want to prepare the optimization model explicitly
before solving, inspect compiled model structures, or verify that the problem
compiles successfully.

Conceptually, a \code{Problem} object stores a declarative optimization
specification: planning data, actions, effects, targets, constraints,
objectives, spatial relations, and optional method or solver settings.
\code{compile_model()} transforms that stored specification into an internal
compiled model representation that can later be reused by the solving layer.

The exact compiled representation is implementation-specific, but it may
include indexed variables, prepared constraint blocks, objective structures,
and internal model snapshots or pointers.

Compilation does not solve the optimization problem. Therefore, a problem may
compile successfully and still later be infeasible, numerically difficult, or
otherwise fail during solver execution.
}
\examples{
\dontrun{
x <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

x <- x |>
  add_constraint_targets_relative(0.3) |>
  add_objective_min_cost(alias = "cost")

x <- compile_model(x)

# Force recompilation
x <- compile_model(x, force = TRUE)
}

}
\seealso{
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_model.R
\name{.pa_build_model}
\alias{.pa_build_model}
\title{Build optimization model from Problem}
\usage{
.pa_build_model(x)
}
\arguments{
\item{x}{Problem object (class "Problem") created with create_problem().}
}
\value{
Updated \code{Problem} object with model pointer and model snapshot.
}
\description{
Materializes (builds) the optimization model using the current state of the \code{Problem} object:
prepared data tables, stored objective settings, and stored constraints (e.g., targets).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_method_epsilon_constraint.R
\name{set_method_epsilon_constraint}
\alias{set_method_epsilon_constraint}
\title{Set the epsilon-constraint multi-objective method}
\usage{
set_method_epsilon_constraint(
  x,
  primary,
  eps = NULL,
  aliases = NULL,
  mode = c("manual", "auto"),
  n_points = 10,
  include_extremes = TRUE,
  lexicographic = TRUE,
  lexicographic_tol = 1e-08
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{primary}{Character string giving the alias of the primary objective to
optimize directly.}

\item{eps}{Optional epsilon specification used only in \code{mode = "manual"}.
It may be:
\itemize{
\item a named numeric vector, defining epsilon values for a single run,
\item or a named list of numeric vectors, defining epsilon values for a
grid of runs.
}
Names must correspond exactly to the constrained objective aliases.}

\item{aliases}{Optional character vector of objective aliases to include.
By default, all registered objective aliases are used. The value of
\code{primary} must be included in \code{aliases}.}

\item{mode}{Character string. Must be either \code{"manual"} or
\code{"auto"}.}

\item{n_points}{Integer scalar used only in \code{mode = "auto"}. Number of
epsilon points to generate automatically for the constrained objective.
Must be at least 2.}

\item{include_extremes}{Logical scalar used only in \code{mode = "auto"}. If
\code{TRUE}, include extreme epsilon values in the automatically generated
grid.}

\item{lexicographic}{Logical scalar used only in \code{mode = "auto"}. If
\code{TRUE}, compute extreme points lexicographically.}

\item{lexicographic_tol}{Numeric scalar \eqn{\ge 0}. Tolerance used in
lexicographic extreme-point computation.}
}
\value{
An updated \code{Problem} object with the epsilon-constraint method
configuration stored in \code{x$data$method}.

In manual mode, \code{x$data$method$runs} contains the explicit
epsilon-design grid.

In automatic mode, \code{x$data$method$runs} is \code{NULL} until the grid is
generated later during \code{\link{solve}}.
}
\description{
Configure a \code{Problem} object to be solved with the
epsilon-constraint multi-objective method.

In this method, one objective is designated as the \emph{primary} objective
and is optimized directly, while the remaining objectives are transformed
into \eqn{\varepsilon}-constraints.

Two operating modes are supported:
\itemize{
\item \strong{manual mode}: the user supplies the \eqn{\varepsilon}-levels
explicitly,
\item \strong{automatic mode}: the \eqn{\varepsilon}-levels are generated
later during \code{\link{solve}} from extreme-point or payoff information.
}

This function does not solve the problem. It stores the method configuration
in \code{x$data$method}, to be used later by \code{\link{solve}}.
}
\details{
Use this method when one objective should be optimized directly while the
remaining objectives are controlled through explicit performance thresholds.

\strong{General idea}

Suppose that \eqn{m \ge 2} objective functions have already been registered
in the problem:
\deqn{
f_1(x), f_2(x), \dots, f_m(x).
}

The epsilon-constraint method selects one of them as the primary objective,
say \eqn{f_p(x)}, and treats the remaining objectives as constrained
objectives.

For a fixed vector of epsilon levels, the method solves subproblems in which
the primary objective is optimized directly and the remaining objectives are
imposed through epsilon constraints.

A representative formulation is:

\deqn{
\max \; f_p(x)
}

subject to

\deqn{
f_k(x) \ge \varepsilon_k, \qquad k \in \mathcal{C},
}

together with all original feasibility constraints of the planning problem,
where \eqn{\mathcal{C}} is the set of constrained objectives.

Depending on objective sense, the internal implementation may transform
minimization objectives into equivalent constrained forms, but the method
always follows the same principle:
\itemize{
\item one objective is optimized directly,
\item all remaining objectives are imposed through
\eqn{\varepsilon}-constraints.
}

By solving the problem repeatedly for different epsilon levels, the method
generates a set of efficient trade-off solutions.

\strong{Atomic objectives requirement}

The epsilon-constraint method can only be used with atomic objectives that
have already been registered under aliases. These aliases are typically
created by calling objective setters with an \code{alias} argument, for
example:
\preformatted{
x <- x |>
  add_objective_max_benefit(alias = "benefit") |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_min_fragmentation(alias = "frag")
}

The \code{primary} argument selects which registered objective is optimized
directly. The remaining aliases are treated as constrained objectives.

\strong{Manual mode}

In \code{mode = "manual"}, the user must provide \code{eps}.

The \code{eps} argument can be supplied as:
\itemize{
\item a named numeric vector, defining a single run,
\item or a named list of numeric vectors, defining a grid of runs.
}

The names of \code{eps} must correspond exactly to the constrained objective
aliases, that is, to all aliases in \code{aliases} except \code{primary}.

If the constrained objectives are
\eqn{\mathcal{C} = \{c_1, \dots, c_q\}}, then manual mode creates a design
grid containing all combinations of the supplied epsilon levels for the
constrained objectives.

Each row of this grid defines one subproblem to be solved later.

\strong{Important:} manual mode supports \strong{two or more objectives}.
In particular, it can be used with:
\itemize{
\item 2 objectives: 1 primary + 1 constrained objective,
\item 3 or more objectives: 1 primary + multiple constrained objectives.
}

Thus, manual mode is the general way to use the epsilon-constraint method
when more than two objectives are involved.

In manual mode, the generated design grid is stored immediately in
\code{x$data$method$runs}. Its epsilon columns are named
\code{eps_<alias>}, for example \code{eps_frag}.

\strong{Automatic mode}

In \code{mode = "auto"}, the user omits \code{eps} and instead supplies
\code{n_points}.

In this case, the epsilon grid is not built immediately. Instead, it is
constructed later during \code{\link{solve}} using extreme-point or
payoff-table information.

In the current implementation, automatic mode supports
\strong{exactly two objectives only}:
\itemize{
\item one primary objective,
\item one constrained objective.
}

Therefore, if \code{mode = "auto"}, then \code{aliases} must contain exactly
two objective aliases. Problems with three or more objectives must use
\code{mode = "manual"}.

If \code{include_extremes = TRUE}, the automatically generated grid includes
the extreme values of the constrained objective. Otherwise, only interior
values are used.

If \code{lexicographic = TRUE}, the extreme points used to generate the grid
are computed lexicographically. In that case, one objective is optimized
first, and then the second objective is optimized while constraining the
first to remain within \code{lexicographic_tol} of its optimum.

\strong{Stored configuration}

The configured method stores:
\itemize{
\item \code{name = "epsilon_constraint"},
\item \code{mode},
\item \code{primary},
\item \code{aliases},
\item \code{constrained},
\item epsilon design information,
\item lexicographic configuration.
}

In manual mode, \code{x$data$method$runs} contains the explicit design grid.

In automatic mode, \code{x$data$method$runs} is initially \code{NULL} and is
generated later during \code{\link{solve}}.

For more than two objectives, automatic grid generation is currently
unavailable because the number of epsilon combinations grows rapidly and
requires explicit user control.
}
\examples{
# Small toy problem
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

effects_df <- data.frame(
  pu = c(1, 2, 3, 4, 1, 2, 3, 4),
  action = c("conservation", "conservation", "conservation", "conservation",
             "restoration", "restoration", "restoration", "restoration"),
  feature = c(1, 1, 1, 1, 2, 2, 2, 2),
  benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
  loss = c(0, 0, 1, 0, 0, 1, 0, 0)
)

x <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
  add_effects(effects_df) |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_max_benefit(alias = "benefit") |>
  add_objective_min_loss(alias = "loss")

# Manual mode with one constrained objective
x1 <- set_method_epsilon_constraint(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost"),
  mode = "manual",
  eps = c(cost = 5)
)

x1$data$method

# Manual mode with multiple epsilon values
x2 <- set_method_epsilon_constraint(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost"),
  mode = "manual",
  eps = list(cost = c(4, 6, 8))
)

x2$data$method$runs

# Manual mode with more than two objectives
x3 <- set_method_epsilon_constraint(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost", "loss"),
  mode = "manual",
  eps = list(
    cost = c(4, 6),
    loss = c(0, 1)
  )
)

x3$data$method$runs

# Automatic mode currently supports exactly two objectives
x4 <- set_method_epsilon_constraint(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost"),
  mode = "auto",
  n_points = 5,
  include_extremes = TRUE,
  lexicographic = TRUE,
  lexicographic_tol = 1e-8
)

x4$data$method

}
\seealso{
\code{\link{set_method_augmecon}},
\code{\link{set_method_weighted_sum}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_min_fragmentation_pu}
\alias{add_objective_min_fragmentation_pu}
\title{Add objective: minimize fragmentation}
\usage{
add_objective_min_fragmentation_pu(
  x,
  relation_name = "boundary",
  weight_multiplier = 1,
  alias = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{relation_name}{Character string giving the name of the spatial relation
to use. The relation must already exist in
\code{x$data$spatial_relations}.}

\item{weight_multiplier}{Numeric scalar greater than or equal to zero. Global
multiplier applied to the relation weights when the objective is built.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that minimizes planning-unit fragmentation over a stored
spatial relation.

This objective acts on the planning-unit selection pattern through the binary
planning-unit variables \eqn{w_i}. It is therefore appropriate when spatial
cohesion is to be encouraged at the level of the selected planning-unit set
as a whole.
}
\details{
Use this function when spatial cohesion should be encouraged at the level of
the selected planning-unit set as a whole.

Let \eqn{\mathcal{I}} denote the set of planning units and let
\eqn{w_i \in \{0,1\}} indicate whether planning unit \eqn{i \in \mathcal{I}}
is selected.

Let the chosen spatial relation define a set of weighted pairs with weights
\eqn{\omega_{ij} \ge 0}. These relation weights are interpreted by the model
builder after scaling by \eqn{\lambda =} \code{weight_multiplier}.

The internal preparation step constructs one auxiliary variable
\eqn{y_{ij} \in [0,1]} for each unique non-diagonal undirected edge
\eqn{(i,j)} with \eqn{i < j}. The intended semantics is:
\deqn{
y_{ij} = w_i \land w_j.
}

This is enforced by the standard linearization:
\deqn{
y_{ij} \le w_i,
}
\deqn{
y_{ij} \le w_j,
}
\deqn{
y_{ij} \ge w_i + w_j - 1.
}

Thus, \eqn{y_{ij}=1} if and only if both planning units \eqn{i} and
\eqn{j} are selected, and \eqn{y_{ij}=0} otherwise.

The exact objective coefficients are assembled later by the model builder
from:
\itemize{
\item the planning-unit variables \eqn{w_i},
\item the edge-conjunction variables \eqn{y_{ij}},
\item the stored relation weights \eqn{\omega_{ij}},
\item and the multiplier \eqn{\lambda}.
}

Conceptually, the resulting objective is a boundary- or relation-based
compactness functional that penalizes exposed or weakly connected selected
patterns while rewarding adjacency among selected planning units.

In the common case where \code{relation_name = "boundary"} and the relation
was built with \code{\link{add_spatial_boundary}}, the objective corresponds
to a boundary-length-style fragmentation penalty.

Setting \code{weight_multiplier = 0} removes the contribution of the spatial
relation from the objective after scaling.

This objective does not distinguish between different actions within the same
planning unit. If action-specific spatial cohesion is required, use
\code{\link{add_objective_min_fragmentation_action}} instead.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

bound_df <- data.frame(
  id1 = c(1, 1, 2, 1, 2, 3, 4),
  id2 = c(1, 2, 2, 3, 4, 4, 4),
  boundary = c(4, 1, 4, 1, 1, 1, 4)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2))

p <- add_spatial_boundary(
  x = p,
  boundary = bound_df,
  name = "boundary",
  include_self = TRUE,
  edge_factor = 1
)

p <- add_objective_min_fragmentation_pu(
  p,
  relation_name = "boundary"
)

p$data$model_args

}
\seealso{
\code{\link{add_spatial_boundary}},
\code{\link{add_spatial_relations}},
\code{\link{add_objective_min_fragmentation_action}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show.R
\name{show}
\alias{show}
\alias{show,Problem-method}
\alias{show,Solution-method}
\alias{show,SolutionSet-method}
\title{Show}
\usage{
\S4method{show}{Problem}(x)

\S4method{show}{SolutionSet}(x)
}
\arguments{
\item{x}{Any object.}
}
\description{
Display information about an object.
}
\seealso{
\code{\link[methods:show]{methods::show()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_solution.R
\name{get_pu}
\alias{get_pu}
\title{Get planning-unit results from a solution}
\usage{
get_pu(x, only_selected = FALSE, run = NULL)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object returned by
\code{\link{solve}}.}

\item{only_selected}{Logical. If \code{TRUE}, return only rows where
\code{selected == 1}. Default is \code{FALSE}.}

\item{run}{Optional positive integer giving the run index to extract from a
\code{SolutionSet}. If \code{NULL}, all runs are returned when available.}
}
\value{
A \code{data.frame} containing the stored planning-unit summary.
Typical columns include planning-unit identifiers, optional labels, and a
\code{selected} indicator.
}
\description{
Extract the planning-unit summary table from a \code{Solution} or
\code{SolutionSet} object returned by \code{\link{solve}}.

The returned table summarizes solution values at the planning-unit level and
typically includes a \code{selected} indicator showing whether each planning
unit is selected in the solution.
}
\details{
This function reads the planning-unit summary stored in
\code{x$summary$pu}. It does not reconstruct the table from the raw decision
vector; it simply returns the stored summary after optional filtering.

Let \eqn{w_i} denote the planning-unit selection variable for planning unit
\eqn{i}. In standard \code{multiscape} workflows, the \code{selected} column is the
user-facing representation of that planning-unit decision, typically coded as
\code{0} or \code{1}.

If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
belonging to that run are returned. This requires the summary table to
contain a \code{run_id} column.

If \code{only_selected = TRUE}, only rows with \code{selected == 1} are
returned. This requires the summary table to contain a \code{selected}
column.

This function is intended for user-facing inspection of planning-unit results.
For the raw model variable vector, use \code{\link{get_solution_vector}}.
}
\examples{
\donttest{
if (requireNamespace("rcbc", quietly = TRUE)) {
  pu_tbl <- data.frame(
    id = 1:4,
    cost = c(1, 2, 3, 4)
  )

  feat_tbl <- data.frame(
    id = 1:2,
    name = c("feature_1", "feature_2")
  )

  dist_feat_tbl <- data.frame(
    pu = c(1, 1, 2, 3, 4),
    feature = c(1, 2, 2, 1, 2),
    amount = c(5, 2, 3, 4, 1)
  )

  actions_df <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  effects_df <- data.frame(
    pu = c(1, 2, 3, 4),
    action = "conservation",
    feature = c(1, 1, 2, 2),
    benefit = c(2, 1, 1, 2),
    loss = c(0, 0, 0, 0)
  )

  p <- create_problem(
    pu = pu_tbl,
    features = feat_tbl,
    dist_features = dist_feat_tbl,
    cost = "cost"
  ) |>
    add_actions(actions_df, cost = 0) |>
    add_effects(effects_df) |>
    add_constraint_targets_relative(0.2) |>
    add_objective_min_cost() |>
    set_solver_cbc(time_limit = 10)

  sol <- solve(p)

  get_pu(sol)
  get_pu(sol, only_selected = TRUE)
}
}
}
\seealso{
\code{\link{get_actions}},
\code{\link{get_features}},
\code{\link{get_targets}},
\code{\link{get_solution_vector}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_method_augmecon.R
\name{set_method_augmecon}
\alias{set_method_augmecon}
\title{Set the AUGMECON multi-objective method}
\usage{
set_method_augmecon(
  x,
  primary,
  aliases = NULL,
  grid = NULL,
  n_points = 10,
  include_extremes = TRUE,
  lexicographic = TRUE,
  lexicographic_tol = 1e-09,
  augmentation = 0.001,
  slack_upper_bound = 1e+06
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{primary}{Character string giving the alias of the primary objective,
that is, the objective optimized directly in the AUGMECON formulation.}

\item{aliases}{Optional character vector of objective aliases to include in
the method. If \code{NULL}, all registered objective aliases are used. The
value of \code{primary} must be included in \code{aliases}.}

\item{grid}{Optional named list defining manual epsilon levels for the
secondary objectives. Each name must correspond to a secondary objective
alias, and each element must be a non-empty numeric vector of finite
values. If \code{NULL}, the grid is generated automatically.}

\item{n_points}{Number of automatically generated epsilon levels per
secondary objective when \code{grid = NULL}. Must be at least 2. Ignored
when \code{grid} is supplied.}

\item{include_extremes}{Logical. If \code{TRUE}, automatically generated
grids include extreme values of each secondary objective.}

\item{lexicographic}{Logical. If \code{TRUE}, use lexicographic anchoring
when computing extreme points for automatic grid construction.}

\item{lexicographic_tol}{Non-negative numeric tolerance used in
lexicographic anchoring.}

\item{augmentation}{Positive numeric augmentation coefficient
\eqn{\rho}. The effective coefficient of each secondary slack is computed
internally as \eqn{\rho / R_k}, where \eqn{R_k} is the payoff-table range
of the corresponding secondary objective.}

\item{slack_upper_bound}{Positive numeric upper bound imposed on the explicit
non-negative slack variables associated with the secondary objectives.}
}
\value{
The updated \code{Problem} object with the AUGMECON method
configuration stored in \code{x$data$method}.
}
\description{
Configure a \code{Problem} object to be solved with the augmented
epsilon-constraint method (AUGMECON).

AUGMECON is an exact multi-objective optimization method in which one
objective is treated as the primary objective and the remaining objectives
are converted into \eqn{\varepsilon}-constraints. In the augmented formulation,
each secondary objective is associated with a non-negative slack variable,
and the primary objective is augmented with a small reward term based on the
normalized slacks. This augmentation is used to avoid weakly efficient
solutions, following Mavrotas (2009).

This function does not solve the problem directly. It stores the AUGMECON
configuration in \code{x$data$method}, to be used later by
\code{\link{solve}}.
}
\details{
Use this method when one objective should be optimized directly, the
remaining objectives should be controlled through epsilon levels, and weakly
efficient solutions should be reduced through the augmented formulation.

\strong{General idea}

Suppose that \eqn{m \ge 2} objective functions have already been registered
in the problem:
\deqn{
f_1(x), f_2(x), \dots, f_m(x).
}

AUGMECON selects one of them as the primary objective, say
\eqn{f_p(x)}, and treats the remaining \eqn{m - 1} objectives as secondary
objectives.

For a fixed combination of epsilon levels, the method solves a sequence of
single-objective subproblems of the form:

\deqn{
\max \; f_p(x) + \rho \sum_{k \in \mathcal{S}} \frac{s_k}{R_k}
}

subject to

\deqn{
f_k(x) - s_k = \varepsilon_k, \qquad k \in \mathcal{S},
}

\deqn{
s_k \ge 0, \qquad k \in \mathcal{S},
}

together with all original feasibility constraints of the planning problem.

Here:
\itemize{
\item \eqn{f_p(x)} is the primary objective,
\item \eqn{\mathcal{S}} is the set of secondary objectives,
\item \eqn{\varepsilon_k} is the imposed level for secondary objective
\eqn{k},
\item \eqn{s_k} is a non-negative slack variable,
\item \eqn{R_k} is the payoff-table range used to normalize objective
\eqn{k},
\item \eqn{\rho > 0} is a small augmentation coefficient.
}

In the original AUGMECON formulation of Mavrotas (2009), the augmentation
term ensures that, among solutions with the same primary objective value, the
solver prefers those with larger normalized slack, thereby avoiding weakly
efficient points and improving Pareto-front generation.

\strong{Secondary-objective equalities and slacks}

The key difference between standard epsilon-constraint and AUGMECON is that
the secondary objectives are written as equalities with slacks rather than as
simple inequalities. For a maximization-type secondary objective, this takes
the form:

\deqn{
f_k(x) - s_k = \varepsilon_k, \qquad s_k \ge 0.
}

This implies:
\deqn{
f_k(x) \ge \varepsilon_k,
}

while explicitly measuring the excess above the imposed epsilon level through
\eqn{s_k}. The augmentation term then rewards such excess in normalized form.

In implementation terms, the exact sign convention for each objective depends
on whether it is internally treated as a minimization or maximization
objective, but the method always preserves the same AUGMECON principle:
\itemize{
\item one objective is optimized directly,
\item all others are turned into constrained objectives,
\item non-negative slacks measure controlled deviation from the imposed
epsilon levels,
\item the primary objective is augmented with a small slack-based reward.
}

\strong{Manual and automatic epsilon grids}

AUGMECON requires a grid of epsilon levels for each secondary objective.

If \code{grid} is supplied, it must be a named list with one numeric vector
per secondary objective. Each vector defines the exact epsilon levels to be
explored for that objective.

If \code{grid = NULL}, the grid is generated automatically later during
\code{\link{solve}}. In that case, the method first computes extreme points
and payoff-table ranges for the secondary objectives, and then generates
\code{n_points} levels for each one.

If \code{include_extremes = TRUE}, the automatic grid includes the extreme
values of each secondary objective.

If \code{lexicographic = TRUE}, extreme points are computed using
lexicographic anchoring, which can improve payoff-table quality when
objectives are tightly competing. The tolerance used for lexicographic
anchoring is controlled by \code{lexicographic_tol}.

\strong{Normalization and augmentation}

The augmentation term is scaled using the payoff-table ranges of the
secondary objectives. If \eqn{R_k} denotes the range of secondary objective
\eqn{k}, then the effective coefficient applied to the slack is:

\deqn{
\frac{\rho}{R_k},
}

where \eqn{\rho = \code{augmentation}}.

This normalization is important because different objectives may be measured
on very different numerical scales. Without normalization, a slack belonging
to a large-scale objective could dominate the augmentation term simply due to
units.

In this implementation, the user supplies \code{augmentation} as the base
coefficient \eqn{\rho}, while the normalized slack coefficients are computed
internally at solve time using the corresponding payoff-table ranges.

\strong{Stored configuration}

This function stores the method definition in \code{x$data$method} with:
\itemize{
\item \code{name = "augmecon"},
\item the primary objective alias,
\item the full set of participating aliases,
\item the set of secondary aliases,
\item either a manual grid or the information required to generate one
automatically,
\item augmentation and slack-bound parameters.
}

The actual payoff table, grid construction, and subproblem solution loop are
performed later by \code{\link{solve}}.
}
\examples{
# Small toy problem
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

effects_df <- data.frame(
  pu = c(1, 2, 3, 4, 1, 2, 3, 4),
  action = c("conservation", "conservation", "conservation", "conservation",
             "restoration", "restoration", "restoration", "restoration"),
  feature = c(1, 1, 1, 1, 2, 2, 2, 2),
  benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
  loss = c(0, 0, 1, 0, 0, 1, 0, 0)
)

x <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
  add_effects(effects_df) |>
  add_objective_max_benefit(alias = "benefit") |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_min_loss(alias = "loss")

# Automatic epsilon grids generated later during solve()
x1 <- set_method_augmecon(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost"),
  n_points = 5,
  include_extremes = TRUE,
  lexicographic = TRUE,
  augmentation = 1e-3
)

x1$data$method

# Manual epsilon grids for two secondary objectives
x2 <- set_method_augmecon(
  x,
  primary = "benefit",
  aliases = c("benefit", "cost", "loss"),
  grid = list(
    cost = c(4, 6, 8),
    loss = c(0, 1)
  ),
  augmentation = 1e-3,
  slack_upper_bound = 1e6
)

x2$data$method

}
\references{
Mavrotas, G. (2009). Effective implementation of the
\eqn{\varepsilon}-constraint method in multi-objective mathematical programming
problems. \emph{Applied Mathematics and Computation}, 213(2), 455--465.
}
\seealso{
\code{\link{set_method_epsilon_constraint}},
\code{\link{set_method_weighted_sum}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_targets.R
\name{add_constraint_targets_relative}
\alias{add_constraint_targets_relative}
\title{Add relative targets}
\usage{
add_constraint_targets_relative(
  x,
  targets,
  features = NULL,
  actions = NULL,
  label = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{targets}{Target specification as proportions in \eqn{[0,1]}. It may be
a scalar, vector, named vector, or \code{data.frame}. See Details.}

\item{features}{Optional feature specification indicating which features the
supplied target values refer to when \code{targets} does not identify
features explicitly. If \code{NULL}, all features are targeted.}

\item{actions}{Optional character vector indicating which actions count toward
target achievement. Entries may match action ids, \code{action_set} labels,
or both. If \code{NULL}, all actions count.}

\item{label}{Optional character string stored with the targets for reporting
and bookkeeping.}
}
\value{
An updated \code{Problem} object with relative targets appended to
the stored target table.
}
\description{
Add feature-level relative targets to a planning problem.

These targets are stored in the problem object and later translated into
linear constraints when the optimization model is built. Relative targets are
supplied as proportions in \eqn{[0,1]} and are converted internally into
absolute thresholds using the current total amount of each
feature in the landscape.

Each call appends one or more target definitions to the problem. This makes
it possible to combine multiple target rules, including targets associated
with different action subsets.
}
\details{
Use this function when target requirements are naturally expressed as
proportions of current baseline feature totals rather than in original
feature units.

Let \eqn{\mathcal{F}} denote the set of features. For each targeted feature
\eqn{f \in \mathcal{F}}, let \eqn{B_f} denote the current baseline total
amount of that feature in the landscape, as computed by
\code{.pa_feature_totals()}.

If the user supplies a relative target \eqn{r_f \in [0,1]}, then this
function converts it to an absolute threshold:
\deqn{
T_f = r_f \times B_f.
}

The absolute threshold \eqn{T_f} is stored in \code{target_value}, while:
\itemize{
\item the original user-supplied proportion \eqn{r_f} is stored in
\code{target_raw},
\item the baseline total \eqn{B_f} is stored in \code{basis_total}.
}

When the optimization model is built, the resulting target is interpreted as:
\deqn{
\sum_{(i,a) \in \mathcal{D}_f^{\star}} c_{iaf} x_{ia} \ge T_f,
}
where:
\itemize{
\item \eqn{i \in \mathcal{I}} indexes planning units,
\item \eqn{a \in \mathcal{A}} indexes actions,
\item \eqn{x_{ia}} indicates whether action \eqn{a} is selected in planning
unit \eqn{i},
\item \eqn{c_{iaf}} is the contribution of that action to feature
\eqn{f},
\item \eqn{\mathcal{D}_f^{\star}} is the subset of planning unit--action
pairs allowed to count toward the target for feature \eqn{f}.
}

The \code{actions} argument restricts which actions may contribute toward
target achievement, but it does not affect the baseline amount \eqn{B_f} used
to compute the threshold. In other words, relative targets are always scaled
against the current full landscape baseline.

Therefore, \code{actions} changes who may satisfy the target, but not how the
threshold itself is scaled.

The \code{targets} argument is parsed by \code{.pa_parse_targets()} and may be
supplied in several equivalent forms, including:
\itemize{
\item a single numeric value recycled to all selected features,
\item a numeric vector aligned to \code{features},
\item a named numeric vector where names identify features,
\item a \code{data.frame} with \code{feature} and \code{target} columns.
}

If \code{targets} does not explicitly identify features:
\itemize{
\item if \code{features = NULL}, the target is applied to all features,
\item if \code{features} is supplied, the target values are interpreted with
respect to that feature set.
}

Relative targets must lie in \eqn{[0,1]}.

Repeated calls append new target rules rather than replacing previous ones.
This allows cumulative target modelling, including multiple rules on the same
feature with different contributing action subsets.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(data.frame(id = "conservation", name = "conservation"), cost = 0)

# Require 30\% of the baseline total for all features
p1 <- add_constraint_targets_relative(p, 0.3)
p1$data$targets

# Require 20\% for one selected feature
p2 <- add_constraint_targets_relative(
  p,
  0.2,
  features = 1
)
p2$data$targets

# Restrict which actions count toward target achievement
p3 <- add_constraint_targets_relative(
  p,
  0.2,
  actions = "conservation"
)
p3$data$targets

}
\seealso{
\code{\link{add_constraint_targets_absolute}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{add_spatial_knn}
\alias{add_spatial_knn}
\title{Add k-nearest-neighbours spatial relations}
\usage{
add_spatial_knn(
  x,
  coords = NULL,
  k = 8,
  name = "knn",
  weight_mode = c("constant", "inverse", "inverse_sq"),
  distance_eps = 1e-09
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}.}

\item{coords}{Optional coordinates specification. This may be:
\itemize{
\item a \code{data.frame(id, x, y)}, or
\item a numeric matrix with two columns \code{(x, y)} aligned to the
order of planning units.
}
If \code{NULL}, coordinates are taken from \code{x$data$pu_coords} or from
columns \code{x$data$pu$x} and \code{x$data$pu$y}.}

\item{k}{Integer giving the number of neighbours per planning unit. Must be
at least 1 and strictly less than the number of planning units.}

\item{name}{Character string giving the key under which the relation is
stored.}

\item{weight_mode}{Character string indicating how distance is converted to
weight. Must be one of \code{"constant"}, \code{"inverse"}, or
\code{"inverse_sq"}.}

\item{distance_eps}{Small positive numeric constant used to avoid division by
zero in inverse-distance weighting.}
}
\value{
An updated \code{Problem} object.
}
\description{
Build and register a k-nearest-neighbours graph between planning units using
coordinates.

This constructor does not require polygon geometry. It uses planning-unit
coordinates supplied explicitly or stored in the \code{Problem} object.
}
\details{
Use this function when neighbourhood should be defined by a fixed number of
nearby planning units rather than by polygon topology or a fixed distance
threshold.

Let \eqn{s_i = (x_i, y_i)} denote the coordinates of planning unit
\eqn{i}. For each planning unit, this function identifies the \code{k}
nearest distinct planning units under Euclidean distance.

If \eqn{d_{ij}} denotes the Euclidean distance between units \eqn{i} and
\eqn{j}, then the k-nearest-neighbours relation is constructed by adding an
edge from \eqn{i} to each of its \code{k} nearest neighbours.

Edge weights are then assigned according to \code{weight_mode}:
\itemize{
\item \code{"constant"}:
\deqn{\omega_{ij} = 1,}
\item \code{"inverse"}:
\deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)},}
\item \code{"inverse_sq"}:
\deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)^2},}
}
where \eqn{\varepsilon} = \code{distance_eps} is a small constant to avoid
division by zero.

The raw k-nearest-neighbours structure is directional by construction, but
the stored relation is registered as undirected by default through
\code{\link{add_spatial_relations}}, which collapses duplicate unordered
pairs.

If the \pkg{RANN} package is available, it is used for efficient nearest
neighbour search. Otherwise, a full distance matrix is computed.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4),
  x = c(0, 1, 0, 1),
  y = c(0, 0, 1, 1)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
)

p <- add_spatial_knn(
  x = p,
  k = 2,
  name = "knn2",
  weight_mode = "constant"
)

p$data$spatial_relations$knn2

}
\seealso{
\code{\link{add_spatial_distance}},
\code{\link{add_spatial_relations}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_effects.R
\name{add_benefits}
\alias{add_benefits}
\title{Add benefits}
\usage{
add_benefits(
  x,
  benefits = NULL,
  effect_type = c("delta", "after"),
  effect_aggregation = c("sum", "mean")
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}. It
must already contain feasible actions; run \code{\link{add_actions}} first.}

\item{benefits}{Alias of \code{effects}, kept for backwards compatibility.}

\item{effect_type}{Character string indicating how supplied effect values are
interpreted. Must be one of:
\itemize{
\item \code{"delta"}: values represent signed net changes,
\item \code{"after"}: values represent after-action amounts and are
converted to net changes relative to baseline feature amounts.
}}

\item{effect_aggregation}{Character string giving the aggregation used when
converting raster values to planning-unit level. Must be one of
\code{"sum"} or \code{"mean"}.}
}
\value{
An updated \code{Problem} object containing:
\describe{
\item{\code{dist_effects}}{The canonical filtered effects table, containing
only rows with \code{benefit > 0}.}
\item{\code{dist_benefit}}{A backwards-compatible mirror table containing
only the benefit component.}
}
}
\description{
Convenience wrapper around \code{\link{add_effects}} that keeps only positive
effects, that is, rows with \code{benefit > 0}.

This function is useful when the workflow is focused only on beneficial
consequences of actions and a gain-only wrapper is more convenient than
calling \code{add_effects(..., component = "benefit")} directly.  Internally,
it calls \code{add_effects()} with \code{component = "benefit"}.
The canonical stored result therefore contains only rows whose effect
satisfies \eqn{\mathrm{benefit}_{iaf} > 0}.

For backwards compatibility, a mirror table containing only the benefit
component is also written to \code{x$data$dist_benefit}. The canonical
filtered representation remains the one stored by \code{add_effects()}.
}
\examples{
pu <- data.frame(id = 1:2, cost = c(1, 2))
features <- data.frame(id = 1, name = "sp1")
dist_features <- data.frame(pu = 1:2, feature = 1, amount = c(5, 10))

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  p,
  data.frame(id = "restoration")
)

eff <- data.frame(
  pu = c(1, 2),
  action = c("restoration", "restoration"),
  feature = c(1, 1),
  delta = c(2, -1)
)

p <- add_benefits(p, benefits = eff)
p$data$dist_benefit

}
\seealso{
\code{\link{add_effects}},
\code{\link{add_losses}},
\code{\link{add_objective_max_benefit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_solver.R
\name{set_solver_gurobi}
\alias{set_solver_gurobi}
\title{Configure Gurobi solver settings}
\usage{
set_solver_gurobi(
  x,
  ...,
  solver_params = list(),
  gap_limit = NULL,
  time_limit = NULL,
  solution_limit = NULL,
  cores = NULL,
  verbose = FALSE,
  log_file = NULL,
  write_log = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{...}{Additional named solver-specific parameters. These are merged into
\code{solver_params}. For example, \code{MIPFocus = 1} for Gurobi.}

\item{solver_params}{Named list of solver-specific parameters. These are
merged with previously stored backend-specific parameters rather than
replacing them completely.}

\item{gap_limit}{Optional numeric value in \eqn{[0,1]} giving the relative
optimality gap for mixed-integer optimization. If \code{NULL}, the
previously stored value is kept unchanged.}

\item{time_limit}{Optional non-negative numeric value giving the maximum
solving time in seconds. If \code{NULL}, the previously stored value is
kept unchanged.}

\item{solution_limit}{Optional logical flag controlling backend-specific
early stopping after feasible solution discovery. If \code{NULL}, the
previously stored value is kept unchanged.}

\item{cores}{Optional positive integer giving the number of CPU cores to use.
If \code{NULL}, the previously stored value is kept unchanged.}

\item{verbose}{Optional logical flag indicating whether the solver should
print log output. If \code{NULL}, the previously stored value is kept
unchanged.}

\item{log_file}{Optional character string giving the name of the
solver log file. If \code{NULL}, the previously stored value is kept
unchanged.}

\item{write_log}{Optional logical flag indicating whether solver output
should be written to a file. If \code{NULL}, the previously stored value is
kept unchanged.}
}
\value{
An updated \code{Problem} object with Gurobi solver settings stored
in \code{x$data$solve_args}.
}
\description{
Convenience wrapper around \code{\link{set_solver}} that stores
\code{solver = "gurobi"} in the problem object.

This function does not solve the model. It only updates the stored solver
configuration.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

x <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
)

x <- set_solver_gurobi(
  x,
  gap_limit = 0.01,
  time_limit = 600,
  cores = 2,
  MIPFocus = 1
)

x$data$solve_args

}
\seealso{
\code{\link{set_solver}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_sim.R
\name{load_sim_features_raster}
\alias{load_sim_features_raster}
\title{Example feature raster}
\usage{
load_sim_features_raster()
}
\value{
A \code{terra::SpatRaster}.
}
\description{
Load the example feature raster shipped with the package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_min_intervention_impact}
\alias{add_objective_min_intervention_impact}
\title{Add objective: minimize intervention impact}
\usage{
add_objective_min_intervention_impact(
  x,
  impact_col = "amount",
  features = NULL,
  actions = NULL,
  alias = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{impact_col}{Character string giving the column in the
feature-distribution table that contains the per-\code{(pu, feature)}
impact amount. The default is \code{"amount"}.}

\item{features}{Optional subset of features to include. Values may match
\code{x$data$features$id} and, if present, \code{x$data$features$name}.}

\item{actions}{Optional subset of actions used to define the intervention
context. Values may match \code{x$data$actions$id} and, if present,
\code{x$data$actions$action_set}.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that minimizes the impact associated with selecting
planning units for intervention.

This objective uses planning-unit selection variables rather than summing the
same impact repeatedly over multiple actions. As a result, each planning unit
contributes at most once to the objective, regardless of how many feasible
actions exist in that unit.
}
\details{
Use this function when intervention itself has a baseline ecological, social,
or operational burden that should be minimized independently of the detailed
effects of particular actions.

Let \eqn{w_i \in \{0,1\}} denote whether planning unit \eqn{i} is selected
for intervention. Let \eqn{q_{if}} denote the impact amount associated with
planning unit \eqn{i} and feature \eqn{f}, taken from column
\code{impact_col} in the feature-distribution table.

If all selected features are included, the objective can be interpreted as:

\deqn{
\min \sum_{i \in \mathcal{I}} \left( \sum_{f \in \mathcal{F}^{\star}} q_{if}
\right) w_i,
}

where \eqn{\mathcal{F}^{\star}} denotes the selected subset of features.

Thus, the coefficient attached to \eqn{w_i} is the aggregated impact of the
selected features in planning unit \eqn{i}.

The role of \code{actions} in this objective is not to make impact additive
over actions, but to restrict the notion of intervention to planning units
that are relevant for the selected subset of actions in downstream model
construction. Even when multiple feasible actions exist in a planning unit,
the planning unit contributes at most once through \eqn{w_i}.

This objective is useful when intervention itself has a baseline ecological,
social, or operational impact that should be minimized independently of the
detailed gain or loss generated by particular actions.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)
feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)
dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)
actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2))

p1 <- add_objective_min_intervention_impact(p)
p1$data$model_args

p2 <- add_objective_min_intervention_impact(
  p,
  features = 1
)
p2$data$model_args

p3 <- add_objective_min_intervention_impact(
  p,
  actions = "restoration"
)
p3$data$model_args

}
\seealso{
\code{\link{add_objective_max_benefit}},
\code{\link{add_objective_min_loss}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_locked_actions.R
\name{add_constraint_locked_actions}
\alias{add_constraint_locked_actions}
\title{Add locked action decisions to a planning problem}
\usage{
add_constraint_locked_actions(x, locked_in = NULL, locked_out = NULL)
}
\arguments{
\item{x}{A \code{Problem} object with action feasibility already defined via
\code{\link{add_actions}}.}

\item{locked_in}{Optional specification of feasible \code{(pu, action)} pairs
that must be selected. It may be \code{NULL}, a \code{data.frame}, or a
named list.}

\item{locked_out}{Optional specification of feasible \code{(pu, action)}
pairs that must not be selected. It may be \code{NULL}, a
\code{data.frame}, or a named list.}
}
\value{
An updated \code{Problem} object in which the status column of the
feasible action table has been modified to reflect locked-in and locked-out
decisions.
}
\description{
Fix feasible planning unit--action decisions to be selected or excluded.

This function modifies the status of existing feasible
\code{(pu, action)} pairs stored in the feasible action table. It does not
create new feasible action pairs and therefore must be used only after
\code{\link{add_actions}} has been called.
}
\details{
Use this function when only specific feasible \code{(pu, action)} decisions
must be forced in or out of the solution, rather than whole planning units.

Let \eqn{\mathcal{I}} denote the set of planning units and
\eqn{\mathcal{A}} the set of actions. Let
\eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
feasible planning unit--action pairs already defined in the problem.

This function allows the user to define two subsets:
\itemize{
\item \eqn{\mathcal{D}^{in} \subseteq \mathcal{D}}, the set of feasible
pairs that must be selected,
\item \eqn{\mathcal{D}^{out} \subseteq \mathcal{D}}, the set of feasible
pairs that must not be selected.
}

These sets are encoded by updating the \code{status} column of the feasible
action table. The function validates that all requested locked-in and
locked-out pairs are already feasible. Therefore, it cannot be used to
introduce new planning unit--action combinations into the problem.

In optimization terms, if \eqn{x_{ia}} denotes the decision variable
associated with planning unit \eqn{i} and action \eqn{a}, then:
\itemize{
\item locked-in pairs conceptually impose \eqn{x_{ia} = 1},
\item locked-out pairs conceptually impose \eqn{x_{ia} = 0}.
}

The exact translation into solver-side constraints occurs later when the
model is built.

In contrast, \code{\link{add_constraint_locked_pu}} fixes whole planning
units through the unit-selection variables, whereas this function fixes only
specific feasible \code{(pu, action)} decisions.

\strong{Accepted formats}

Both \code{locked_in} and \code{locked_out} accept the same formats:
\itemize{
\item \code{NULL},
\item a \code{data.frame} with columns \code{pu} and \code{action},
optionally including a \code{feasible} column used as a filter,
\item a named list whose names are action ids and whose elements are either
vectors of planning unit ids or \code{sf} objects.
}

If a \code{feasible} column is supplied in a \code{data.frame}, only rows
with \code{feasible = TRUE} are used. Missing values in \code{feasible} are
treated as \code{FALSE}.

If an \code{sf} specification is supplied, the problem object must contain
planning-unit geometry, and planning units are matched spatially using
\code{sf::st_intersects()}.

\strong{Conflict checking}

A given \code{(pu, action)} pair cannot be simultaneously requested in both
\code{locked_in} and \code{locked_out}. Such overlaps are rejected.

In addition, if a planning unit is already marked as locked out at the
planning-unit level, then all feasible actions in that planning unit are
forced to \code{status = 3}. Any attempt to lock in an action within such a
planning unit raises an error.

\strong{Order of precedence}

User-supplied locked-in and locked-out action requests are first applied to
the feasible action table. Afterwards, any planning-unit-level
\code{locked_out} flag is enforced, overriding action-level status and
ensuring consistency with planning-unit exclusions.
}
\examples{
pu <- data.frame(
  id = 1:4,
  cost = c(2, 3, 1, 4)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  x = p,
  actions = data.frame(id = c("conservation", "restoration")),
  cost = c(conservation = 3, restoration = 8)
)

# Lock a few feasible decisions
p <- add_constraint_locked_actions(
  x = p,
  locked_in = data.frame(
    pu = c(1, 2),
    action = c("conservation", "restoration")
  ),
  locked_out = data.frame(
    pu = c(4),
    action = c("conservation")
  )
)

p$data$dist_actions

# Named-list interface
p2 <- add_constraint_locked_actions(
  x = p,
  locked_in = list(
    conservation = c(1, 3)
  ),
  locked_out = list(
    restoration = c(2)
  )
)

p2$data$dist_actions

}
\seealso{
\code{\link{add_actions}},
\code{\link{add_constraint_locked_pu}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_min_cost}
\alias{add_objective_min_cost}
\title{Add objective: minimize cost}
\usage{
add_objective_min_cost(
  x,
  include_pu_cost = TRUE,
  include_action_cost = TRUE,
  actions = NULL,
  alias = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{include_pu_cost}{Logical. If \code{TRUE}, include planning-unit costs
in the objective.}

\item{include_action_cost}{Logical. If \code{TRUE}, include action costs in
the objective.}

\item{actions}{Optional subset of actions to include in the action-cost
component. Values may match \code{x$data$actions$id} and, if present,
\code{x$data$actions$action_set}. If \code{NULL}, all feasible actions are
included in the action-cost term.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that minimizes the total cost of the solution.

Depending on the function arguments, the objective may include planning-unit
costs, action costs, or both. Action costs can optionally be restricted to a
subset of actions.
}
\details{
Use this function when the planning problem is framed primarily as a
cost-minimization problem, with costs arising from planning-unit selection,
action implementation, or both.

Let \eqn{\mathcal{I}} be the set of planning units and let
\eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
feasible planning unit--action decisions.

Let:
\itemize{
\item \eqn{w_i \in \{0,1\}} denote whether planning unit \eqn{i} is
selected,
\item \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected
in planning unit \eqn{i},
\item \eqn{c_i^{PU} \ge 0} denote the planning-unit cost of unit \eqn{i},
\item \eqn{c_{ia}^{A} \ge 0} denote the cost of selecting action
\eqn{a} in planning unit \eqn{i}.
}

The most general form of this objective is:

\deqn{
\min \left(
\sum_{i \in \mathcal{I}} c_i^{PU} w_i
+
\sum_{(i,a) \in \mathcal{D}^{\star}} c_{ia}^{A} x_{ia}
\right),
}

where \eqn{\mathcal{D}^{\star}} denotes the subset of feasible decisions
whose action contributes to the action-cost term.

If \code{include_pu_cost = FALSE}, the planning-unit cost term is omitted.

If \code{include_action_cost = FALSE}, the action-cost term is omitted.

If \code{actions = NULL}, all feasible actions contribute to the action-cost
term. If \code{actions} is supplied, only the selected subset contributes to
that term. Planning-unit costs are never subset by \code{actions}; they are
always global whenever \code{include_pu_cost = TRUE}.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)
feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)
dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)
actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2))

p1 <- add_objective_min_cost(p)
p1$data$model_args

p2 <- add_objective_min_cost(
  p,
  include_pu_cost = FALSE,
  include_action_cost = TRUE
)
p2$data$model_args

p3 <- add_objective_min_cost(
  p,
  actions = "restoration"
)
p3$data$model_args

}
\seealso{
\code{\link{add_objective_max_profit}},
\code{\link{add_objective_max_net_profit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{add_objective_max_profit}
\alias{add_objective_max_profit}
\title{Add objective: maximize profit}
\usage{
add_objective_max_profit(
  x,
  profit_col = "profit",
  actions = NULL,
  alias = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{profit_col}{Character string giving the profit column in the stored
profit table.}

\item{actions}{Optional subset of actions to include. Values may match
\code{x$data$actions$id} and, if present,
\code{x$data$actions$action_set}. If \code{NULL}, all actions are included.}

\item{alias}{Optional identifier used to register this objective for
multi-objective workflows.}
}
\value{
An updated \code{Problem} object.
}
\description{
Define an objective that maximizes total profit from selected planning
unit--action decisions.
}
\details{
Use this function when the objective is to maximize gross economic return,
without subtracting planning-unit or action costs.

Let \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected in
planning unit \eqn{i}, and let \eqn{\pi_{ia}} denote the profit associated
with that decision, as taken from column \code{profit_col} in the stored
profit table.

If all actions are included, the objective is:

\deqn{
\max \sum_{(i,a) \in \mathcal{D}} \pi_{ia} x_{ia},
}

where \eqn{\mathcal{D}} denotes the set of feasible planning unit--action
decisions.

If \code{actions} is provided, only the selected subset contributes to the
objective. Letting \eqn{\mathcal{D}^{\star}} denote the feasible decisions
whose action belongs to the selected subset, the objective becomes:

\deqn{
\max \sum_{(i,a) \in \mathcal{D}^{\star}} \pi_{ia} x_{ia}.
}

This objective considers profit only. It does not subtract planning-unit
costs or action costs. For a net-profit formulation, use
\code{\link{add_objective_max_net_profit}}.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)
feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)
dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)
actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)
profit_df <- data.frame(
  pu = c(1, 2, 3, 4, 1, 2, 3, 4),
  action = c("conservation", "conservation", "conservation", "conservation",
             "restoration", "restoration", "restoration", "restoration"),
  profit = c(5, 4, 3, 2, 8, 7, 6, 5)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
  add_profit(profit_df)

p1 <- add_objective_max_profit(p)
p1$data$model_args

p2 <- add_objective_max_profit(
  p,
  actions = "restoration"
)
p2$data$model_args

}
\seealso{
\code{\link{add_objective_min_cost}},
\code{\link{add_objective_max_net_profit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{spatial_relations}
\alias{spatial_relations}
\title{Spatial relations infrastructure}
\description{
Define a unified internal representation for spatial relations between
planning units and provide constructors for building such relations from
common spatial inputs.

Spatial relations in \code{multiscape} are graph-like structures defined over
the set of planning units. They are used to represent adjacency, shared
boundary length, proximity, neighbourhood graphs, or other pairwise spatial
relationships that may later be used by objectives, constraints, or
diagnostics.

Relations are stored inside the \code{Problem} object as
\code{x$data$spatial_relations[[name]]}.
}
\details{
\strong{Internal representation}

Each spatial relation is stored as a \code{data.frame} containing, at a
minimum, the columns:
\describe{
\item{\code{internal_pu1}}{Integer index of the first planning unit.}
\item{\code{internal_pu2}}{Integer index of the second planning unit.}
\item{\code{weight}}{Non-negative edge weight. Its interpretation depends
on the relation type.}
}

Additional columns may also be present, such as \code{pu1}, \code{pu2},
\code{distance}, \code{source}, or other diagnostic fields. These extra
columns are preserved when possible.

Thus, a spatial relation can be interpreted as a weighted graph
\eqn{G = (\mathcal{I}, E, \omega)}, where:
\itemize{
\item \eqn{\mathcal{I}} is the set of planning units,
\item \eqn{E} is the set of stored edges,
\item \eqn{\omega_{ij} \ge 0} is the weight associated with edge
\eqn{(i,j)}.
}

\strong{Directed and undirected relations}

Some relation constructors create undirected relations by default. In that
case, each unordered pair is typically stored once, usually with
\eqn{i < j}, unless self-edges are explicitly allowed.

\strong{Boundary relations and diagonal terms}

The function \code{\link{add_spatial_boundary}} supports boundary-length
relations derived either from a boundary table or from polygon geometry.

If \code{include_self = TRUE}, diagonal entries \eqn{(i,i)} may be added to
represent exposed boundary. These diagonal weights are intended for
boundary-based compactness or fragmentation formulations, where perimeter
exposed to the outside of the selected solution should contribute to the
objective.

\strong{Geometry safety}

All \code{sf}-based constructors operate only through spatial predicates,
topological relations, or boundary-length calculations. They never subdivide,
cut, or alter planning-unit geometries. Planning units are aligned to the
internal planning-unit order before spatial relations are computed.

\strong{Available constructors}

The following functions build and register spatial relations:
\itemize{
\item \code{\link{add_spatial_relations}}: register a precomputed relation,
\item \code{\link{add_spatial_boundary}}: shared boundary length relation,
\item \code{\link{add_spatial_rook}}: rook adjacency from polygons,
\item \code{\link{add_spatial_queen}}: queen adjacency from polygons,
\item \code{\link{add_spatial_knn}}: k-nearest-neighbours graph from
coordinates,
\item \code{\link{add_spatial_distance}}: distance-threshold graph from
coordinates.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_problem.R
\name{create_problem}
\alias{create_problem}
\alias{create_problem,data.frame,data.frame,data.frame-method}
\alias{create_problem,ANY,ANY,missing-method}
\alias{create_problem,ANY,ANY,NULL-method}
\alias{create_problem,ANY,data.frame,data.frame-method}
\title{Create a planning problem input object}
\usage{
create_problem(
  pu,
  features,
  dist_features,
  cost = NULL,
  pu_id_col = "id",
  cost_aggregation = c("mean", "sum"),
  ...
)

\S4method{create_problem}{data.frame,data.frame,data.frame}(
  pu,
  features,
  dist_features,
  cost = NULL,
  pu_id_col = "id",
  cost_aggregation = c("mean", "sum"),
  ...
)

\S4method{create_problem}{ANY,ANY,missing}(
  pu,
  features,
  dist_features,
  cost = NULL,
  pu_id_col = "id",
  cost_aggregation = c("mean", "sum"),
  ...
)

\S4method{create_problem}{ANY,ANY,NULL}(
  pu,
  features,
  dist_features,
  cost = NULL,
  pu_id_col = "id",
  cost_aggregation = c("mean", "sum"),
  ...
)

\S4method{create_problem}{ANY,data.frame,data.frame}(
  pu,
  features,
  dist_features,
  cost = NULL,
  pu_id_col = "id",
  cost_aggregation = c("mean", "sum"),
  ...
)
}
\arguments{
\item{pu}{Planning-units input. Depending on the selected method, this may be:
\itemize{
\item a \code{data.frame} with planning-unit information,
\item an \code{sf} object,
\item a \code{terra::SpatVector},
\item a vector file path readable by \pkg{terra},
\item a \code{terra::SpatRaster},
\item or a raster file path.
}}

\item{features}{Feature input. Depending on the selected workflow, this may be:
\itemize{
\item a \code{data.frame} with at least an \code{id} column,
\item a \code{terra::SpatRaster} with one layer per feature,
\item or a raster file path.
}}

\item{dist_features}{Feature-distribution input. In tabular and hybrid
workflows, this must be a \code{data.frame} with columns \code{pu},
\code{feature}, and \code{amount}. In spatial modes it may be omitted or
set to \code{NULL}, in which case it is derived automatically.}

\item{cost}{In spatial modes, planning-unit cost information. Depending on the
input mode, this may be:
\itemize{
\item a column name in the planning-unit attribute table,
\item a raster object,
\item or a raster file path.
}
In raster-cell mode, \code{cost} is required and valid planning units are
defined only where cost is finite and strictly positive.}

\item{pu_id_col}{Character string giving the name of the planning-unit id
column in vector or \code{sf} inputs. Ignored in purely tabular mode and in
raster-cell mode.}

\item{cost_aggregation}{Character string used in vector-PU mode when
\code{cost} is a raster and must be aggregated over polygons. Must be one
of \code{"mean"} or \code{"sum"}.}

\item{...}{Additional arguments forwarded to internal builders.}
}
\value{
A \code{Problem} object for downstream \code{multiscape} workflows.
The returned object always contains a canonical tabular core with planning
units, features, and feature-distribution data, and may additionally
contain aligned spatial metadata depending on the input mode.
}
\description{
Create a \code{Problem} object from tabular or spatial inputs.

\code{create_problem()} is the main entry point to the \code{multiscape}
workflow. Its role is to standardize heterogeneous planning inputs into a
common internal representation that can later be used by actions, effects,
targets, constraints, spatial relations, objectives, and solvers.

In all supported workflows, the result is a \code{Problem} object with a
canonical tabular core, optionally enriched with aligned spatial metadata
such as coordinates, geometry, raster references, or raw planning-unit
attributes.
}
\details{
A \code{Problem} object created by \code{create_problem()} is the basic input
structure used throughout downstream \code{multiscape} workflows.

The returned object always contains a canonical tabular planning core and may
additionally store aligned spatial metadata depending on the input workflow.
This is the internal representation on which later functions such as
\code{add_actions()}, \code{add_effects()},
\code{add_constraint_targets()}, \code{add_spatial_relations()}, and
\code{solve()} operate.

\strong{Conceptual role of \code{create_problem()}}

Regardless of the input workflow, \code{create_problem()} aims to produce a
consistent internal representation containing at least:
\itemize{
\item a planning-unit table,
\item a feature table,
\item a feature-distribution table.
}

Internally, the problem is always reduced to a tabular core of the form:
\itemize{
\item planning units \eqn{i \in \mathcal{I}},
\item features \eqn{f \in \mathcal{F}},
\item non-negative baseline amounts \eqn{a_{if} \ge 0},
\item and, when available, spatial metadata associated with planning units.
}

The feature-distribution table provides the canonical sparse representation of
these baseline amounts. Thus, after \code{create_problem()} has run, the
landscape is internally represented by baseline feature amounts
\eqn{a_{if}}, where \eqn{a_{if}} denotes the amount of feature \eqn{f} in
planning unit \eqn{i}. These baseline amounts are later combined with
actions, effects, targets, objectives, and constraints to build the
optimization model.

\strong{Which input mode should I use?}

\code{create_problem()} supports four main workflows. The best choice depends
on the form of your data and on whether you want to preserve geometry.

\itemize{
\item \strong{Tabular mode.} Use this when \code{pu}, \code{features}, and
\code{dist_features} are already available as \code{data.frame} objects and
no spatial derivation is needed. This is the simplest workflow: all problem
components are already tabular, so \code{create_problem()} only
standardizes them into the canonical internal representation.

\item \strong{Vector-PU spatial mode.} Use this when planning units are
polygons and feature information is stored in one or more raster layers. In
this mode, \code{dist_features} is derived by aggregating raster values
over planning-unit polygons.

\item \strong{Raster-cell fast mode.} Use this when both \code{pu} and
\code{features} are rasters and each valid raster cell should become one
planning unit. This mode avoids raster-to-polygon conversion, treats
\code{NA} feature values as zero before building \code{dist_features},
keeps only strictly positive amounts in the stored distribution table, and
is generally the preferred option for large regular grids.

\item \strong{Hybrid sf + tabular mode.} Use this when you already have a
curated tabular \code{dist_features} table but still want to preserve
planning-unit geometry and attributes for later plotting, spatial
relations, or feasibility specifications. Here, the feature distribution is
already tabular, but geometry and raw attributes are preserved from the
\code{sf} planning-unit object for later spatial operations.
}

\strong{How cost is interpreted across modes}

\code{cost} handling depends on the selected workflow.

\itemize{
\item \strong{Tabular mode.} In purely tabular mode, \code{cost} is not
used by this generic method. Planning-unit costs are expected to already be
present in the \code{pu} table supplied to the internal tabular builder.

\item \strong{Vector-PU spatial mode.} In vector-PU mode, \code{cost} is
required and may be either:
\itemize{
\item the name of a numeric attribute column in the planning-unit layer,
\item or a cost raster to be aggregated over polygons using the
\code{cost_aggregation} argument.
}

\item \strong{Raster-cell fast mode.} In raster-cell mode, \code{cost}
must be a single-layer raster aligned with \code{pu} and \code{features}.
A raster cell becomes a planning unit only if the mask cell is not missing
and the corresponding cost value is finite and strictly positive. In other
words, if \eqn{m_i} denotes the mask value of cell \eqn{i} and \eqn{c_i}
its cost value, then cell \eqn{i} is retained only when the mask is
observed and \eqn{c_i > 0}.

\item \strong{Hybrid sf + tabular mode.} In hybrid mode, \code{cost} must
be either:
\itemize{
\item the name of a numeric attribute column in the \code{sf} layer,
\item or omitted if the \code{sf} attributes already contain a column
literally named \code{cost}.
}
}

\strong{Feature identifiers}

Features are internally standardized to an \code{id}-based representation.
In spatial modes where raster layers are used, one feature is created per
raster layer, with:
\itemize{
\item \code{id = 1, 2, \dots, nlyr(features)},
\item \code{name} equal to the raster layer name, if available,
\item otherwise a generated name of the form \code{"feature.1"},
\code{"feature.2"}, and so on.
}

\strong{Planning-unit identifiers}

In vector and hybrid modes, planning-unit ids are taken from the column named
by \code{pu_id_col}. If that column is missing and \code{pu_id_col = "id"},
sequential ids are created with a warning.

In raster-cell mode, planning-unit ids are always created sequentially from
the valid raster cells retained after masking and cost filtering.

\strong{Ordering and alignment}

In spatial modes, planning units, derived coordinates, raw attributes,
geometry, and extracted feature amounts are aligned to the same planning-unit
id order before the final \code{Problem} object is created.

This alignment is critical because later functions assume that all stored
planning-unit components refer to the same ordered set of planning units.

After \code{create_problem()}, typical next steps include adding actions,
spatial relations, targets or other constraints, objectives, and then solving
the resulting problem.
}
\examples{
# ------------------------------------------------------
# 1) Tabular mode
# ------------------------------------------------------
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

p1 <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl
)

print(p1)

# ------------------------------------------------------
# 2) Hybrid sf + tabular mode using package data
# ------------------------------------------------------

p2 <- create_problem(
  pu = sim_pu,
  features = sim_features,
  dist_features = sim_dist_features,
  cost = "cost"
)

print(p2)

}
\seealso{
\code{\link{add_actions}},
\code{\link{add_constraint_locked_pu}},
\code{\link{add_spatial_boundary}},
\code{\link{add_spatial_knn}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solutionset-class.R
\name{solutionset-class}
\alias{solutionset-class}
\alias{SolutionSet}
\title{SolutionSet class}
\value{
No return value. This page documents the \code{SolutionSet} class.
}
\description{
The \code{SolutionSet} class stores the result of solving a
\code{\link{Problem}} object when multiple runs are produced.

A \code{SolutionSet} object is the multi-run counterpart of
\code{\link{solution-class}}. It contains the original problem, run-level
design and outcome tables, the list of individual \code{Solution} objects,
diagnostics summarizing the full set of runs, method metadata, and additional
metadata.

Objects of this class are typically created by \code{\link{solve}} in
workflows that generate more than one optimization run, such as weighted-sum
scans, \eqn{\epsilon}-constraint designs, or AUGMECON grids.
}
\details{
\strong{Conceptual role}

The \code{SolutionSet} class represents the result of a multi-run solving
workflow:
\preformatted{
Problem -> solve() -> SolutionSet
}

Each run corresponds to one specific optimization subproblem, parameter
setting, or trade-off configuration. The \code{SolutionSet} object keeps all
these runs together in a structured form.

Thus, a \code{SolutionSet} is not a single solution with multiple labels, but
a collection of distinct \code{Solution} objects linked to a shared problem
and a shared multi-run design.

\strong{Typical use cases}

A \code{SolutionSet} is typically returned when the chosen solution method
generates multiple runs, for example:
\itemize{
\item several weighted-sum configurations,
\item a grid of \eqn{\epsilon}-constraint runs,
\item an AUGMECON exploration of multiple
\eqn{\epsilon}-combinations,
\item or any other workflow producing more than one optimizer call.
}

\strong{Internal structure}

A \code{SolutionSet} object separates information into several layers:

\describe{
\item{\code{problem}}{The original \code{Problem} object shared by all
runs.}

\item{\code{solution}}{A named \code{list} containing the core multi-run
optimization outputs. This typically includes the run design, the run
summary table, and the list of individual \code{Solution} objects.}

\item{\code{summary}}{A named \code{list} containing user-facing summaries
aggregated across runs when such summaries are available.}

\item{\code{diagnostics}}{A named \code{list} containing diagnostics about
the solution set as a whole.}

\item{\code{method}}{A named \code{list} describing the multi-run
optimization method used.}

\item{\code{meta}}{A named \code{list} containing additional metadata.}

\item{\code{name}}{A \code{character(1)} identifier for the solution set.}
}

\strong{Run-level content}

The \code{solution} field is the main entry point for run-level information.
Typical entries include:
\describe{
\item{\code{design}}{A \code{data.frame} describing the experimental or
optimization design, for example weights or \eqn{\epsilon}-levels.}

\item{\code{runs}}{A \code{data.frame} summarizing the outcome of each run.
This typically includes run identifiers, solver status, runtime, gap, and
objective values.}

\item{\code{solutions}}{A list of individual \code{Solution} objects, one
per run.}
}

In many workflows, the \code{runs} table is the most important compact
representation of the solution set, while \code{solutions[[i]]} provides the
full detailed output for run \eqn{i}.

\strong{Objective values and run tables}

In multi-objective workflows, the run table often stores objective values in
columns named \code{value_<alias>}, where \code{<alias>} is the alias of a
registered objective. For example:
\itemize{
\item \code{value_cost},
\item \code{value_frag},
\item \code{value_benefit}.
}

This naming convention is used by downstream functions such as
\code{\link{plot_tradeoff}}.

Depending on the solving method, the run table may also contain design
columns such as:
\itemize{
\item \code{weight_<alias>} for weighted-sum runs,
\item \code{eps_<alias>} for \eqn{\epsilon}-constraint or AUGMECON runs.
}

\strong{Relationship with \code{Solution}}

A \code{SolutionSet} is conceptually a collection of \code{Solution} objects.
The individual runs are typically stored in:
\deqn{
\code{x$solution$solutions[[i]]}
}

where each element is itself a full \code{Solution} object.

Therefore:
\itemize{
\item use \code{SolutionSet} when working with the full set of runs,
\item use an individual \code{Solution} when inspecting one particular run
in detail.
}

\strong{Diagnostics}

The \code{diagnostics} field stores metadata about the multi-run solve
process. Depending on the implementation, it may summarize:
\itemize{
\item number of runs,
\item status frequencies,
\item runtime ranges,
\item gap ranges,
\item and other aggregate information about the set of runs.
}

\strong{Printing}

The \code{print()} method provides a concise summary of the full solution set.
It reports:
\itemize{
\item the optimization method name,
\item the participating objective aliases,
\item the number of design rows, runs, and stored solutions,
\item run-level status summaries,
\item runtime and gap ranges,
\item and the names of design and objective-value columns when available.
}

This printed output is intended as a quick overview. Detailed inspection
should use:
\itemize{
\item \code{x$solution$runs},
\item \code{x$solution$design},
\item \code{x$solution$solutions[[i]]},
\item or the accessor methods documented below.
}
}
\section{Fields}{

\describe{
\item{\code{problem}}{The \code{Problem} object used to generate the full
solution set.}
\item{\code{solution}}{A named \code{list} containing the core multi-run
outputs, typically including \code{design}, \code{runs}, and
\code{solutions}.}
\item{\code{summary}}{A named \code{list} containing user-facing summaries
associated with the solution set.}
\item{\code{diagnostics}}{A named \code{list} containing diagnostics about
the solution set as a whole.}
\item{\code{method}}{A named \code{list} describing the optimization method
used.}
\item{\code{meta}}{A named \code{list} of additional metadata.}
\item{\code{name}}{A \code{character(1)} name for the solution set
object.}
}
}

\section{Methods}{

\describe{
\item{\code{print()}}{Print a concise summary of the solution set,
including method name, number of runs, and run-level diagnostics.}

\item{\code{show()}}{Alias of \code{print()}.}

\item{\code{repr()}}{Return a short one-line representation of the
solution set.}

\item{\code{getMethod()}}{Return the method specification stored in
\code{self$method}.}

\item{\code{getDesign()}}{Return the design table stored in
\code{self$solution$design}.}

\item{\code{getRuns()}}{Return the run summary table stored in
\code{self$solution$runs}.}

\item{\code{getSolutions()}}{Return the list of individual
\code{Solution} objects stored in \code{self$solution$solutions}.}
}
}

\seealso{
\code{\link{solution-class}},
\code{\link{plot_tradeoff}},
\code{\link{get_pu}},
\code{\link{get_actions}},
\code{\link{get_features}},
\code{\link{get_targets}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_method_weighted_sum.R
\name{set_method_weighted_sum}
\alias{set_method_weighted_sum}
\title{Set the weighted-sum multi-objective method}
\usage{
set_method_weighted_sum(
  x,
  aliases,
  weights,
  normalize_weights = FALSE,
  objective_scaling = FALSE
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{aliases}{Character vector of objective aliases to combine. Each alias
must correspond to a previously registered atomic objective.}

\item{weights}{Numeric vector of weights, with the same length and order as
\code{aliases}. Weights must be finite. If \code{normalize_weights = TRUE},
they are rescaled to sum to 1 before being stored.}

\item{normalize_weights}{Logical. If \code{TRUE}, normalize the supplied
weights to sum to 1 before storing them.}

\item{objective_scaling}{Logical. If \code{TRUE}, request scaling of the
participating objectives before weighted aggregation in the solving layer.}
}
\value{
The updated \code{Problem} object with the weighted-sum method
configuration stored in \code{x$data$method}.
}
\description{
Configure a \code{Problem} object to be solved with a weighted-sum
multi-objective method.

In the weighted-sum method, several registered atomic objectives are combined
into a single scalar objective using a weighted linear combination. This
function stores that configuration in \code{x$data$method} so that it can be
used later by \code{\link{solve}}.
}
\details{
Use this method when several registered objectives should be combined into a
single scalar optimization problem through explicit preference weights.

\strong{General idea}

Suppose that a set of atomic objectives has already been registered in the
problem under aliases \eqn{k \in \mathcal{K}}. Let \eqn{f_k(x)} denote the
scalar value of objective \eqn{k}, and let \eqn{w_k} denote its user-supplied
weight.

The weighted-sum method combines them into a single scalar objective of the
form:

\deqn{
\sum_{k \in \mathcal{K}} w_k \, f_k(x).
}

In practice, the exact sign convention used internally depends on the sense
of each registered atomic objective, for example whether it is a
minimization-type or maximization-type objective. The solving layer is
therefore responsible for constructing a solver-ready scalar objective from
the stored objective specifications and the requested weights.

\strong{Atomic objectives requirement}

The weighted-sum method can only be used with atomic objectives that have
already been registered under aliases. These aliases are typically created by
calling objective setters with an \code{alias} argument, for example:
\preformatted{
x <- x |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_min_fragmentation(alias = "frag")
}

Internally, each atomic objective is stored in
\code{x$data$objectives[[alias]]} together with its metadata, such as:
\itemize{
\item \code{objective_id},
\item \code{model_type},
\item \code{sense},
\item \code{objective_args}.
}

The \code{aliases} argument passed to this function selects which of those
registered atomic objectives are included in the weighted combination.

\strong{Weight normalization}

If \code{normalize_weights = TRUE}, the supplied weights are rescaled to sum
to 1:

\deqn{
\tilde{w}_k = \frac{w_k}{\sum_{j \in \mathcal{K}} w_j}.
}

This normalization does not change the optimizer's solution in a pure
weighted-sum formulation as long as all weights are multiplied by the same
positive constant, but it can improve interpretability and numerical
conditioning.

If \code{normalize_weights = FALSE}, the supplied weights are stored exactly
as provided.

\strong{Objective scaling}

If \code{objective_scaling = TRUE}, the solving layer is expected to scale the
participating objectives before combining them. The purpose of scaling is to
reduce distortions caused by objectives being measured on very different
numerical ranges.

Conceptually, if \eqn{R_k} denotes a scale or range associated with objective
\eqn{k}, then a scaled weighted sum may be interpreted as:

\deqn{
\sum_{k \in \mathcal{K}} w_k \, \frac{f_k(x)}{R_k}.
}

The exact scaling rule is implemented later in the solving layer. This
function only stores whether objective scaling has been requested.

\strong{Mixed objective senses}

Weighted sums are straightforward when all participating objectives have the
same optimization sense. When minimization and maximization objectives are
mixed, the solving layer must standardize them internally before building the
scalar objective.

This function validates that the requested aliases exist, but it does not
itself resolve objective-sense compatibility. That logic is delegated to the
downstream solving layer.

\strong{Theoretical limitation}

The weighted-sum method typically recovers only \emph{supported} efficient
solutions, that is, solutions lying on the convex hull of the Pareto front in
objective space. In non-convex multi-objective problems, especially mixed
integer problems, some efficient solutions cannot be obtained by any weighted
combination. In such cases, methods such as
\code{\link{set_method_epsilon_constraint}} or
\code{\link{set_method_augmecon}} may be preferable.

\strong{Stored configuration}

This function stores the method definition in \code{x$data$method} with:
\itemize{
\item \code{name = "weighted"},
\item \code{aliases},
\item \code{weights},
\item \code{normalize_weights},
\item \code{objective_scaling}.
}

The actual scalarization is performed later by \code{\link{solve}}.
}
\examples{
# Small toy problem
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

actions_df <- data.frame(
  id = c("conservation", "restoration"),
  name = c("conservation", "restoration")
)

effects_df <- data.frame(
  pu = c(1, 2, 3, 4, 1, 2, 3, 4),
  action = c("conservation", "conservation", "conservation", "conservation",
             "restoration", "restoration", "restoration", "restoration"),
  feature = c(1, 1, 1, 1, 2, 2, 2, 2),
  benefit = c(2, 1, 0, 1, 3, 0, 1, 2),
  loss = c(0, 0, 1, 0, 0, 1, 0, 0)
)

x <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(actions_df, cost = c(conservation = 1, restoration = 2)) |>
  add_effects(effects_df) |>
  add_objective_min_cost(alias = "cost") |>
  add_objective_max_benefit(alias = "benefit")

x <- set_method_weighted_sum(
  x,
  aliases = c("cost", "benefit"),
  weights = c(0.4, 0.6),
  normalize_weights = FALSE
)

x$data$method

# Normalize weights before storing
x2 <- set_method_weighted_sum(
  x,
  aliases = c("cost", "benefit"),
  weights = c(2, 3),
  normalize_weights = TRUE
)

x2$data$method

# Request objective scaling
x3 <- set_method_weighted_sum(
  x,
  aliases = c("cost", "benefit"),
  weights = c(0.7, 0.3),
  objective_scaling = TRUE
)

x3$data$method

}
\seealso{
\code{\link{set_method_epsilon_constraint}},
\code{\link{set_method_augmecon}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-sim.R
\docType{data}
\name{sim_features}
\alias{sim_features}
\title{Simulated features}
\format{
A data frame with feature identifiers and names.
}
\usage{
data(sim_features)
}
\description{
Example feature table for package examples and tests.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_profit.R
\name{add_profit}
\alias{add_profit}
\title{Add profit to a planning problem}
\usage{
add_profit(x, profit = NULL)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}. It
must already contain feasible actions and an action catalogue; run
\code{\link{add_actions}} first.}

\item{profit}{Profit specification. One of:
\itemize{
\item \code{NULL}: profit is set to 0 for all feasible
\code{(pu, action)} pairs,
\item a numeric scalar: recycled to all feasible pairs,
\item a named numeric vector: names are action ids and values define
action-level profit,
\item a \code{data.frame(action, profit)} defining action-level profit,
\item a \code{data.frame(pu, action, profit)} defining pair-specific
profit.
}}
}
\value{
An updated \code{Problem} object with a stored profit table created
or replaced. The stored table contains columns \code{pu}, \code{action},
\code{profit}, \code{internal_pu}, and \code{internal_action}, and
includes only rows with non-zero profit.
}
\description{
Define economic profit values for feasible planning unit--action pairs and
store them in a profit table.

Profit is stored separately from ecological effects. In particular,
\code{profit} is not the same as ecological \code{benefit} or
\code{loss} as represented in \code{\link{add_effects}}. This separation
allows the package to distinguish economic returns from ecological
consequences when building objectives, constraints, and reporting summaries.
}
\details{
\strong{When to use \code{add_profit()}.}

Use this function when economic returns, penalties, or other action-specific
financial values are part of the planning problem. Typical downstream uses
include objectives such as \code{\link{add_objective_max_profit}} and
\code{\link{add_objective_max_net_profit}}.

Let \eqn{\mathcal{I}} denote the set of planning units and
\eqn{\mathcal{A}} the set of actions. Let
\eqn{\mathcal{D} \subseteq \mathcal{I} \times \mathcal{A}} denote the set of
feasible planning unit--action pairs currently stored in the problem.

This function assigns to each feasible pair \eqn{(i,a) \in \mathcal{D}} a
numeric profit value \eqn{\pi_{ia} \in \mathbb{R}} and stores the result in a
profit table.

Thus, the stored table can be interpreted as a mapping
\deqn{
\pi : \mathcal{D} \to \mathbb{R},
}
where \eqn{\pi_{ia}} represents the economic return associated with selecting
action \eqn{a} in planning unit \eqn{i}.

Profit values may be positive, zero, or negative. Positive values represent
gains or revenues, zero represents no net profit contribution, and negative
values can be used to encode penalties or net economic losses.

The stored table contains:
\itemize{
\item \code{pu}: external planning-unit id,
\item \code{action}: action id,
\item \code{profit}: numeric profit value,
\item \code{internal_pu}: internal planning-unit index,
\item \code{internal_action}: internal action index.
}

\strong{Supported input formats}

The \code{profit} argument may be specified in several ways:
\itemize{
\item \code{NULL}: assign profit 0 to all feasible \code{(pu, action)}
pairs,
\item a numeric scalar: assign the same profit value to all feasible pairs,
\item a named numeric vector: names are action ids, assigning one global
profit value per action,
\item a \code{data.frame(action, profit)}: assign one global profit value
per action,
\item a \code{data.frame(pu, action, profit)}: assign pair-specific profit
values.
}

When action-level profit is supplied, the same profit value is assigned to
all feasible planning units for that action. When pair-specific profit is
supplied, only the listed \code{(pu, action)} pairs receive explicit values;
unmatched feasible pairs are interpreted as zero-profit pairs.

\strong{Storage behaviour}

This function stores only rows with non-zero profit values. Feasible pairs
whose final profit is zero are omitted from the stored profit table.
Missing values produced during matching or joins are treated as zero before
this filtering step. Therefore, the resulting table is a sparse
representation of economic returns over the feasible decision space.

\strong{Data-only behaviour}

This function is purely data-oriented. It does not build or modify the
optimization model, and it does not change feasibility. It simply assigns
profit values to rows already present in the feasible action table.

In particular:
\itemize{
\item it does not add new feasible \code{(pu, action)} pairs,
\item it does not remove infeasible pairs,
\item it does not apply solver-side filtering such as dropping locked-out
decisions,
\item it does not modify ecological effect tables.
}

Any such filtering is expected to occur later when model-ready tables are
prepared, typically during the build stage invoked by \code{solve()}.

\strong{Use in optimization}

Profit values stored by this function can later be used in objectives such as
\code{\link{add_objective_max_profit}} or
\code{\link{add_objective_max_net_profit}}, in derived budget expressions, or
in reporting and summary functions.

For example, if \eqn{x_{ia} \in \{0,1\}} denotes whether action \eqn{a} is
selected in planning unit \eqn{i}, then a profit-maximization objective
typically takes the form
\deqn{
\max \sum_{(i,a) \in \mathcal{D}} \pi_{ia} x_{ia}.
}
}
\examples{
# Minimal problem
pu <- data.frame(
  id = 1:4,
  cost = c(2, 3, 1, 4)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  x = p,
  actions = data.frame(id = c("harvest", "restoration"))
)

# 1) Constant profit for every feasible (pu, action)
p1 <- add_profit(p, profit = 10)
p1$data$dist_profit

# 2) Profit per action using a named vector
pr <- c(harvest = 50, restoration = -5)
p2 <- add_profit(p, profit = pr)
p2$data$dist_profit

# 3) Profit per action using a data frame
pr_df <- data.frame(
  action = c("harvest", "restoration"),
  profit = c(40, 15)
)
p3 <- add_profit(p, profit = pr_df)
p3$data$dist_profit

# 4) Profit per (pu, action) pair
pr_pair <- data.frame(
  pu = c(1, 2, 3),
  action = c("harvest", "harvest", "restoration"),
  profit = c(100, 80, 30)
)
p4 <- add_profit(p, profit = pr_pair)
p4$data$dist_profit

}
\seealso{
\code{\link{add_actions}},
\code{\link{add_objective_max_profit}},
\code{\link{add_objective_max_net_profit}},
\code{\link{add_effects}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_area.R
\name{add_constraint_area}
\alias{add_constraint_area}
\title{Add area constraint}
\usage{
add_constraint_area(
  x,
  area,
  sense,
  tolerance = 0,
  area_col = NULL,
  area_unit = c("m2", "ha", "km2"),
  actions = NULL,
  name = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{area}{Numeric scalar greater than or equal to zero. Target value for
the constrained area.}

\item{sense}{Character string indicating the type of area constraint. Must be
one of \code{"min"}, \code{"max"}, or \code{"equal"}.}

\item{tolerance}{Numeric scalar greater than or equal to zero. Only used when
\code{sense = "equal"}. In that case, equality is interpreted as a band
around \code{area} with half-width \code{tolerance}. Ignored otherwise.}

\item{area_col}{Optional character string giving the name of the area column
in \code{x$data$pu}. If \code{NULL}, the area source is resolved later by
the model builder.}

\item{area_unit}{Character string indicating the unit of \code{area} and
\code{tolerance}. Must be one of \code{"m2"}, \code{"ha"}, or
\code{"km2"}.}

\item{actions}{Optional subset of actions to which the constraint applies.
If \code{NULL}, the constraint applies to the total selected area in the
problem through the planning-unit selection variables. Otherwise, it applies
to the selected decision variables associated with the specified subset of
actions. This argument is resolved using the package's standard action
subset parser.}

\item{name}{Optional character string used as the label of the stored linear
constraint when it is later added to the optimization model. If
\code{NULL}, a default name is generated.}
}
\value{
An updated \code{Problem} object with the new area-constraint
specification appended to \code{x$data$constraints$area}.
}
\description{
Add an area constraint to a planning problem.

This function stores one area-constraint specification in the
\code{Problem} object so that it can later be incorporated when the
optimization model is assembled. Multiple area constraints can be added by
calling this function repeatedly, provided that no duplicated combination of
action subset and constraint sense is introduced.
}
\details{
Use this function when area requirements must be imposed either on the total
selected landscape or on the subset of selected decisions associated with
specific actions.

Let \eqn{\mathcal{I}} denote the set of planning units and let
\eqn{a_i \ge 0} be the area associated with planning unit
\eqn{i \in \mathcal{I}}

When \code{actions = NULL}, the constraint refers to the total selected area
in the problem. In that case, let \eqn{w_i \in \{0,1\}} denote the binary
variable indicating whether planning unit \eqn{i} is selected by at least one
decision in the model.

Depending on \code{sense}, this function stores one of the following
constraints:

If \code{sense = "min"}:
\deqn{
\sum_{i \in \mathcal{I}} a_i w_i \ge A
}

If \code{sense = "max"}:
\deqn{
\sum_{i \in \mathcal{I}} a_i w_i \le A
}

If \code{sense = "equal"} and \code{tolerance = 0}:
\deqn{
\sum_{i \in \mathcal{I}} a_i w_i = A
}

If \code{sense = "equal"} and \code{tolerance > 0}, the equality is stored as
a two-sided band:
\deqn{
A - \tau \le \sum_{i \in \mathcal{I}} a_i w_i \le A + \tau
}
where \eqn{\tau} is the value supplied through \code{tolerance}.

When \code{actions} is not \code{NULL}, the constraint is applied only to the
selected decisions associated with the specified subset of actions. Let
\eqn{\mathcal{A}^\star \subseteq \mathcal{A}} denote that subset and let
\eqn{x_{ia} \in \{0,1\}} denote the binary variable indicating whether action
\eqn{a \in \mathcal{A}^\star} is selected in planning unit
\eqn{i \in \mathcal{I}}. In that case, the constrained quantity is
\deqn{
\sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}^\star} a_i x_{ia}.
}

Under formulations where at most one action can be selected per planning
unit, this coincides with the area allocated to that subset of actions.

Areas are obtained from the planning-unit table. If \code{area_col} is
provided, that column is used. Otherwise, the model builder later determines
the default area source according to the internal rules of the package. The
value of \code{area_unit} indicates the unit in which \code{area} and
\code{tolerance} are expressed and therefore how the stored threshold should
be interpreted.

This function only stores the constraint specification; it does not validate
the feasibility of the threshold against the available planning units at this
stage.

Multiple area constraints can be stored in a \code{Problem} object. However,
at most one can be stored for the same combination of action subset and
constraint sense. Attempting to add a duplicated
\code{actions}--\code{sense} combination results in an error.
}
\examples{
pu <- data.frame(
  id = 1:4,
  cost = c(2, 3, 1, 4),
  area_ha = c(10, 15, 8, 20)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

actions <- data.frame(
  id = c("conservation", "restoration")
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  p,
  actions = actions,
  cost = c(conservation = 1, restoration = 2)
)

p <- add_constraint_area(
  x = p,
  area = 25,
  sense = "min",
  area_col = "area_ha",
  area_unit = "ha"
)

p <- add_constraint_area(
  x = p,
  area = 15,
  sense = "max",
  area_col = "area_ha",
  area_unit = "ha",
  actions = "restoration"
)

p <- add_constraint_area(
  x = p,
  area = 5,
  sense = "min",
  area_col = "area_ha",
  area_unit = "ha",
  actions = "restoration"
)

p$data$constraints$area

}
\seealso{
\code{\link{create_problem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_solution.R
\name{get_solution_vector}
\alias{get_solution_vector}
\title{Get raw solution vector from a solution}
\usage{
get_solution_vector(x, run = 1L)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object returned by
\code{\link{solve}}.}

\item{run}{Positive integer giving the run index to extract when \code{x} is
a \code{SolutionSet}. Default is \code{1L}.}
}
\value{
A numeric vector with one value per internal model variable.
}
\description{
Return the raw decision-variable vector produced by the solver, in the
internal model-variable order used by the optimization backend.
}
\details{
This function extracts the raw solution vector stored at
\code{x$solution$vector} for a \code{Solution} or for a selected run of a
\code{SolutionSet}.

The returned vector is in the internal variable order of the optimization
model. Depending on the problem formulation, it may include:
\itemize{
\item planning-unit selection variables,
\item action-allocation variables,
\item auxiliary variables introduced for targets, budgets, fragmentation,
or other constraints/objectives,
\item and potentially additional blocks created internally by the model
builder.
}

Therefore, this vector is primarily intended for advanced users, debugging,
diagnostics, or internal verification. It is not a user-facing allocation
table.

To inspect selected planning units or selected actions in a more interpretable
form, use \code{\link{get_pu}} or \code{\link{get_actions}} instead.

For a single solution, the returned vector corresponds to that solution.
For a \code{SolutionSet}, the \code{run} argument selects which run to
extract.
}
\examples{
\donttest{
if (requireNamespace("rcbc", quietly = TRUE)) {
  pu_tbl <- data.frame(
    id = 1:4,
    cost = c(1, 2, 3, 4)
  )

  feat_tbl <- data.frame(
    id = 1:2,
    name = c("feature_1", "feature_2")
  )

  dist_feat_tbl <- data.frame(
    pu = c(1, 1, 2, 3, 4),
    feature = c(1, 2, 2, 1, 2),
    amount = c(5, 2, 3, 4, 1)
  )

  actions_df <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  effects_df <- data.frame(
    pu = c(1, 2, 3, 4),
    action = "conservation",
    feature = c(1, 1, 2, 2),
    benefit = c(2, 1, 1, 2),
    loss = c(0, 0, 0, 0)
  )

  p <- create_problem(
    pu = pu_tbl,
    features = feat_tbl,
    dist_features = dist_feat_tbl,
    cost = "cost"
  ) |>
    add_actions(actions_df, cost = 0) |>
    add_effects(effects_df) |>
    add_constraint_targets_relative(0.2) |>
    add_objective_min_cost() |>
    set_solver_cbc(time_limit = 10)

  sol <- solve(p)

  v <- get_solution_vector(sol)
  v
  length(v)
}
}

}
\seealso{
\code{\link{get_pu}},
\code{\link{get_actions}},
\code{\link{get_features}},
\code{\link{get_targets}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solution-class.R
\name{solution-class}
\alias{solution-class}
\alias{Solution}
\title{Solution class}
\value{
No return value. This page documents the \code{Solution} class.
}
\description{
The \code{Solution} class stores the result of solving a single
\code{\link{Problem}} object in \pkg{multiscape}.

A \code{Solution} object contains the original problem definition, the core
optimization output returned by the solver, user-facing summary tables,
diagnostics about the solve process, and metadata describing how the solution
was obtained.

Objects of this class are typically created by \code{\link{solve}}.
}
\details{
\strong{Conceptual role}

The \code{Solution} class represents the output of one optimization run.

It should be understood as the single-run counterpart of the modelling
workflow:
\preformatted{
Problem -> solve() -> Solution
}

Thus, a \code{Solution} object does not replace the original
\code{Problem}; instead, it keeps a reference to it in the \code{problem}
field and augments it with optimization results.

\strong{Single-run semantics}

A \code{Solution} corresponds to one realized solution of one configured
optimization problem. This may come from:
\itemize{
\item a single-objective solve,
\item one run of a weighted multi-objective workflow,
\item one \eqn{\epsilon}-constraint subproblem,
\item one AUGMECON subproblem,
\item or any other workflow that ultimately produces one optimizer output.
}

When multiple runs are generated, they are typically collected in a separate
\code{SolutionSet} object rather than a single \code{Solution}.

\strong{Internal structure}

A \code{Solution} object separates results into several layers:

\describe{
\item{\code{problem}}{The original \code{Problem} object used to generate
the solution.}

\item{\code{solution}}{A named \code{list} containing the core optimization
output. This may include the objective value, raw model variable vector,
decoded decision vectors, and evaluated objective values by alias.}

\item{\code{summary}}{A named \code{list} of user-facing summary tables,
typically derived from the original problem and the solved decisions. These
tables are intended for inspection, plotting, reporting, and downstream
analysis.}

\item{\code{diagnostics}}{A named \code{list} containing solver diagnostics
such as status, runtime, optimality gap, solver name, and runtime
settings.}

\item{\code{method}}{A named \code{list} describing the optimization method
used to obtain the solution.}

\item{\code{meta}}{A named \code{list} containing additional metadata.}

\item{\code{name}}{A \code{character(1)} identifier for the solution
object.}
}

\strong{Core optimization output}

The \code{solution} field stores the solver-facing result. Typical entries may
include:
\describe{
\item{\code{objective}}{The scalar objective value returned by the solver
for this run.}
\item{\code{vector}}{The raw internal solution vector in model-variable
order.}
\item{\code{alias_values}}{Objective values evaluated for registered
aliases, when available.}
}

The raw solution vector may contain not only user-facing decision variables
such as planning-unit or action decisions, but also auxiliary variables
introduced internally by the model builder.

\strong{User-facing summaries}

The \code{summary} field stores derived tables intended for interpretation
rather than solver interaction. Typical entries include:
\describe{
\item{\code{pu}}{Planning-unit summary table.}
\item{\code{actions}}{Planning unit--action allocation summary table.}
\item{\code{features}}{Feature-level outcome summary table.}
\item{\code{targets}}{Target-achievement summary table, when targets were
part of the problem.}
}

These tables are the main source used by user-facing accessors such as
\code{\link{get_pu}}, \code{\link{get_actions}},
\code{\link{get_features}}, and \code{\link{get_targets}}.

\strong{Diagnostics}

The \code{diagnostics} field stores metadata about the optimization process,
including solver status and runtime information. Typical entries may include:
\itemize{
\item solver name,
\item runtime in seconds,
\item optimality gap,
\item number of cores,
\item time limit,
\item status code or status label.
}

These values describe how the solution was obtained, not the content of the
solution itself.

\strong{Printing}

The \code{print()} method is intended as a concise diagnostic summary. It
reports:
\itemize{
\item solver status,
\item objective value,
\item optimality gap,
\item runtime,
\item counts of selected planning units and actions,
\item target fulfillment summary when available,
\item evaluated objective alias values when available,
\item and basic solver information.
}

This summary is intended for quick inspection. More detailed exploration
should use the stored \code{summary}, \code{solution}, and
\code{diagnostics} fields directly, or the dedicated accessor functions.
}
\section{Fields}{

\describe{
\item{\code{problem}}{The \code{Problem} object used to generate the
solution.}
\item{\code{solution}}{A named \code{list} containing the core optimization
result.}
\item{\code{summary}}{A named \code{list} of user-facing summary tables.}
\item{\code{diagnostics}}{A named \code{list} of solver diagnostics.}
\item{\code{method}}{A named \code{list} describing the method used.}
\item{\code{meta}}{A named \code{list} of additional metadata.}
\item{\code{name}}{A \code{character(1)} name for the solution object.}
}
}

\section{Methods}{

\describe{
\item{\code{print()}}{Print a concise summary of the solution, including
status, objective value, runtime, selection counts, and target
fulfillment.}

\item{\code{show()}}{Alias of \code{print()}.}

\item{\code{repr()}}{Return a short one-line representation of the
solution.}
}
}

\seealso{
\code{\link{problem-class}},
\code{\link{get_pu}},
\code{\link{get_actions}},
\code{\link{get_features}},
\code{\link{get_targets}},
\code{\link{solve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_solution.R
\name{get_features}
\alias{get_features}
\title{Get feature summary from a solution}
\usage{
get_features(x, run = NULL)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object returned by
\code{\link{solve}}.}

\item{run}{Optional positive integer giving the run index to extract from a
\code{SolutionSet}. If \code{NULL}, all runs are returned when available.}
}
\value{
A \code{data.frame} with one row per feature, or one row per
feature--run combination when multiple runs are present. The returned table
always includes the columns \code{total_available}, \code{benefit},
\code{loss}, \code{net}, and \code{total}.
}
\description{
Extract the per-feature summary table from a \code{Solution} or
\code{SolutionSet} object returned by \code{\link{solve}}.

The returned table summarizes, for each feature, the total amount available
in the landscape together with the positive and negative contributions induced
by the selected actions in the solution.
}
\details{
This function reads the feature summary stored in \code{x$summary$features}.
It errors if that table is missing.

Let \eqn{B_f} denote the total baseline amount available in the landscape for
feature \eqn{f}. Let \eqn{G_f} denote the total positive contribution induced
by selected actions, and let \eqn{L_f} denote the total negative
contribution. Then the returned table is intended to summarize quantities of
the form:

\deqn{
\mathrm{net}_f = G_f - L_f,
}

\deqn{
\mathrm{total}_f = B_f + \mathrm{net}_f.
}

In the stored summary, these quantities are typically represented by the
columns:
\itemize{
\item \code{total_available}, corresponding to \eqn{B_f},
\item \code{benefit}, corresponding to \eqn{G_f},
\item \code{loss}, corresponding to \eqn{L_f},
\item \code{net}, corresponding to \eqn{G_f - L_f},
\item \code{total}, corresponding to \eqn{B_f + G_f - L_f}.
}

If any of the core numeric columns \code{total_available}, \code{benefit}, or
\code{loss} are missing, they are created and filled with zero. If
\code{net} is missing, it is computed as \code{benefit - loss}. If
\code{total} is missing, it is computed as \code{total_available + net}.

Thus, this function guarantees that the returned table contains the columns
\code{total_available}, \code{benefit}, \code{loss}, \code{net}, and
\code{total}, even if some of them were absent from the stored summary.

If \code{x} is a \code{SolutionSet} and \code{run} is provided, only rows
belonging to that run are returned. If the result contains a \code{run_id}
column but only a single run is present and \code{run} was not requested
explicitly, the \code{run_id} column is removed for convenience.

This function summarizes feature outcomes in the solution. It is different
from \code{\link{get_targets}}, which focuses on target achievement rather
than total feature balance.
}
\examples{
\donttest{
if (requireNamespace("rcbc", quietly = TRUE)) {
  pu_tbl <- data.frame(
    id = 1:4,
    cost = c(1, 2, 3, 4)
  )

  feat_tbl <- data.frame(
    id = 1:2,
    name = c("feature_1", "feature_2")
  )

  dist_feat_tbl <- data.frame(
    pu = c(1, 1, 2, 3, 4),
    feature = c(1, 2, 2, 1, 2),
    amount = c(5, 2, 3, 4, 1)
  )

  actions_df <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  effects_df <- data.frame(
    pu = c(1, 2, 3, 4),
    action = "conservation",
    feature = c(1, 1, 2, 2),
    benefit = c(2, 1, 1, 2),
    loss = c(0, 0, 0, 0)
  )

  p <- create_problem(
    pu = pu_tbl,
    features = feat_tbl,
    dist_features = dist_feat_tbl,
    cost = "cost"
  ) |>
    add_actions(actions_df, cost = 0) |>
    add_effects(effects_df) |>
    add_constraint_targets_relative(0.2) |>
    add_objective_min_cost() |>
    set_solver_cbc(time_limit = 10)

  sol <- solve(p)

  get_features(sol)
}
}

}
\seealso{
\code{\link{get_pu}},
\code{\link{get_actions}},
\code{\link{get_targets}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{add_spatial_boundary}
\alias{add_spatial_boundary}
\title{Add spatial boundary-length relations}
\usage{
add_spatial_boundary(
  x,
  boundary = NULL,
  geometry = NULL,
  name = "boundary",
  weight_col = NULL,
  weight_multiplier = 1,
  include_self = TRUE,
  edge_factor = 1
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{boundary}{Optional \code{data.frame} describing boundary lengths.
Accepted formats are:
\itemize{
\item \code{(id1, id2, boundary)}, or
\item \code{(pu1, pu2, weight)}.
}}

\item{geometry}{Optional \code{sf} object with planning-unit polygons and an
\code{id} column. If \code{NULL}, \code{x$data$pu_sf} is used.}

\item{name}{Character string giving the key under which the relation is
stored.}

\item{weight_col}{Optional character string giving the name of the weight
column in \code{boundary}. If \code{NULL}, the function tries to infer it
from \code{"boundary"} or \code{"weight"}.}

\item{weight_multiplier}{Positive numeric scalar applied to all boundary
weights.}

\item{include_self}{Logical. If \code{TRUE}, include diagonal entries
representing exposed boundary.}

\item{edge_factor}{Numeric scalar greater than or equal to zero. Multiplier
applied to exposed boundary when constructing diagonal entries.}
}
\value{
An updated \code{Problem} object with the stored relation in
\code{x$data$spatial_relations[[name]]}.
}
\description{
Build and register a boundary-length spatial relation between planning units.

Boundary relations represent shared edge length between adjacent polygons. In
contrast to queen adjacency, they only account for boundary segments of
positive length and ignore point-only contacts.
}
\details{
Use this function when spatial structure should be represented through shared
boundary length rather than binary contiguity or coordinate-based proximity.

Two input modes are supported:
\enumerate{
\item \strong{Boundary-table mode.} If \code{boundary} is supplied, it is
interpreted as a boundary table, for example a Marxan-style
\code{bound.dat}.

\item \strong{Geometry mode.} If \code{boundary = NULL}, boundary lengths
are derived from polygon geometry using \code{geometry} or
\code{x$data$pu_sf}.
}

Let \eqn{\omega_{ij} \ge 0} denote the shared boundary length between
planning units \eqn{i} and \eqn{j}, multiplied by
\code{weight_multiplier}.

For off-diagonal entries \eqn{i \neq j}, the stored weight is:
\deqn{
\omega_{ij} = \mathrm{\gamma} \times b_{ij},
}
where \eqn{b_{ij}} is the shared boundary length and \eqn{\gamma} is
the user-supplied \code{weight_multiplier}.

If \code{include_self = TRUE}, diagonal entries are also created. These are
not geometric self-neighbours in the graph sense; instead, they represent the
effective boundary exposed to the outside of the solution.

Let \eqn{p_i} be the total perimeter of planning unit \eqn{i}, and let
\eqn{\sum_{j \neq i} \omega_{ij}} be the total incident shared boundary
recorded for that planning unit. Then the exposed boundary is represented by
a diagonal term derived from the difference between total perimeter and shared
boundary, scaled by \code{edge_factor}.

These diagonal terms are useful in boundary-based compactness or
fragmentation objectives, because they encode the portion of each planning
unit's perimeter that would remain exposed if the unit were selected.

\strong{Boundary-table mode}

If \code{boundary} is provided, accepted formats are:
\itemize{
\item \code{(id1, id2, boundary)}, or
\item \code{(pu1, pu2, weight)}.
}

If the table contains diagonal rows \eqn{(i,i)}, these are interpreted as
total perimeter values in boundary-table mode.

\strong{Geometry mode}

If \code{boundary = NULL}, shared boundary lengths are derived directly from
polygon geometry. Only positive-length intersections are retained. Point
touches are ignored.

\strong{Storage}

The final relation is stored through \code{\link{add_spatial_relations}},
typically as an undirected relation with optional diagonal entries.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

bound_df <- data.frame(
  id1 = c(1, 1, 2, 1, 2, 3, 4),
  id2 = c(1, 2, 2, 3, 4, 4, 4),
  boundary = c(4, 1, 4, 1, 1, 1, 4)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
)

p <- add_spatial_boundary(
  x = p,
  boundary = bound_df,
  name = "boundary",
  include_self = TRUE,
  edge_factor = 1
)

p$data$spatial_relations$boundary

}
\seealso{
\code{\link{add_spatial_relations}},
\code{\link{add_objective_min_fragmentation_pu}},
\code{\link{add_objective_min_fragmentation_action}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print}
\alias{print}
\alias{print.Problem}
\alias{print.Solution}
\alias{print.SolutionSet}
\title{Print}
\usage{
\method{print}{Problem}(x, ...)

\method{print}{Solution}(x, ...)

\method{print}{SolutionSet}(x, ...)
}
\arguments{
\item{x}{Any object.}

\item{...}{Not used.}
}
\value{
None.
}
\description{
Displays information about an object.
}
\seealso{
\code{\link[base:print]{base::print()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_spatial_pu}
\alias{plot_spatial_pu}
\title{Plot selected planning units in space}
\usage{
plot_spatial_pu(
  x,
  runs = NULL,
  ...,
  base_alpha = 0.1,
  selected_alpha = 0.9,
  base_fill = "grey92",
  base_color = NA,
  selected_color = NA,
  draw_borders = FALSE,
  show_base = TRUE
)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object.}

\item{runs}{Optional integer vector of run ids. If \code{NULL}, a
\code{Solution} is used directly and a \code{SolutionSet} defaults to the
first run.}

\item{...}{Reserved for future extensions.}

\item{base_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the base
planning-unit layer.}

\item{selected_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the
selected planning-unit layer.}

\item{base_fill}{Fill colour for the base planning-unit layer.}

\item{base_color}{Border colour for the base planning-unit layer.}

\item{selected_color}{Border colour for selected planning units.}

\item{draw_borders}{Logical. If \code{FALSE}, borders are not drawn.}

\item{show_base}{Logical. If \code{TRUE}, draw the base planning-unit layer
underneath the selected units.}
}
\value{
Invisibly returns a \code{ggplot} object.
}
\description{
Plot the spatial distribution of selected planning units from a
\code{Solution} or \code{SolutionSet}.

This function maps the planning-unit selection summary returned by
\code{\link{get_pu}} onto the planning-unit geometry stored in the associated
\code{Problem} object.
}
\details{
Let \eqn{w_i \in \{0,1\}} denote the planning-unit selection variable for
planning unit \eqn{i}. This function plots the user-facing
\code{selected == 1} representation of \eqn{w_i}.

If several runs are requested, the output is faceted by \code{run_id}.

Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
}
\examples{
if (requireNamespace("sf", quietly = TRUE) &&
    requireNamespace("ggplot2", quietly = TRUE)) {
  data("sim_pu_sf", package = "multiscape")

  problem <- structure(
    list(
      data = list(
        pu_sf = sim_pu_sf
      )
    ),
    class = "Problem"
  )

  sol <- structure(
    list(
      problem = problem,
      summary = list(
        pu = data.frame(
          id = sim_pu_sf$id,
          selected = as.integer(seq_len(nrow(sim_pu_sf)) \%\% 2 == 1)
        )
      )
    ),
    class = "Solution"
  )

  plot_spatial_pu(sol)
}

}
\seealso{
\code{\link{get_pu}},
\code{\link{plot_spatial}},
\code{\link{plot_spatial_actions}},
\code{\link{plot_spatial_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_objectives.R
\name{register_atomic_objective_internal}
\alias{register_atomic_objective_internal}
\title{Register an atomic objective (internal)}
\arguments{
\item{x}{A \code{Problem} object.}

\item{alias}{Optional character scalar used to register the objective as an
atomic objective. If \code{NULL}, no registration entry is created.}

\item{objective_id}{Character string giving the stable internal identifier of
the objective, for example \code{"min_cost"} or \code{"max_benefit"}.}

\item{model_type}{Character string giving the model-builder label associated
with this objective, for example \code{"minimizeCosts"}.}

\item{objective_args}{A list of objective-specific arguments to be stored with
the objective definition.}

\item{sense}{Character string giving the optimization direction. Must be
either \code{"min"} or \code{"max"}.}
}
\value{
An updated \code{Problem} object.
}
\description{
Internal helper used by objective setter functions to define the active
single-objective configuration of a \code{Problem} object and, optionally,
register that objective as an atomic objective for multi-objective workflows.
}
\details{
In \code{multiscape}, an \emph{atomic objective} is a fully specified
objective definition that can later be reused by a multi-objective method
such as a weighted-sum formulation, an \eqn{\epsilon}-constraint method,
AUGMECON, or other objective-orchestration procedures.

Each atomic objective is identified by:
\itemize{
\item a stable internal identifier \code{objective_id},
\item a solver-facing model label \code{model_type},
\item a list of objective-specific arguments stored in
\code{objective_args},
\item an optimization sense, either \code{"min"} or \code{"max"},
\item and, optionally, a user-facing identifier \code{alias}.
}

If \code{alias} is not \code{NULL}, the objective is stored in
\code{x$data$objectives[[alias]]}. This makes it possible to refer to the
same objective later by a stable user-facing name. For example, a user may
register objectives under aliases such as \code{"cost"},
\code{"benefit"}, or \code{"frag"} and then pass those aliases to a
multi-objective method.

If \code{alias} is \code{NULL}, no atomic-objective entry is created in
\code{x$data$objectives}. In that case, the calling function still defines
the currently active single-objective configuration through
\code{x$data$model_args}, but no reusable multi-objective registration is
created.

Thus, this helper supports two complementary modes:
\itemize{
\item \strong{single-objective mode}: only the active objective is stored,
\item \strong{multi-objective-ready mode}: the active objective is stored
and also registered under an alias for later reuse.
}

Conceptually, if an objective function is denoted by \eqn{f(x)}, this helper
does not itself define the mathematical form of \eqn{f}; rather, it stores
the metadata required so that downstream code can reconstruct the correct
objective expression, its direction of optimization, and its user-visible
identity.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_spatial_features}
\alias{plot_spatial_features}
\title{Plot spatial feature values from a solution}
\usage{
plot_spatial_features(
  x,
  runs = NULL,
  features = NULL,
  value = c("final", "baseline", "benefit"),
  layout = NULL,
  max_facets = 4L,
  ...,
  base_alpha = 0.1,
  selected_alpha = 0.9,
  base_fill = "grey92",
  base_color = NA,
  selected_color = NA,
  draw_borders = FALSE,
  show_base = TRUE,
  fill_na = "grey80",
  use_viridis = TRUE
)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object.}

\item{runs}{Optional integer vector of run ids. If \code{NULL}, a
\code{Solution} is used directly and a \code{SolutionSet} defaults to the
first run.}

\item{features}{Optional feature subset to display. Matching is attempted
against both feature ids and feature names.}

\item{value}{Character string indicating which feature quantity to plot. Must
be one of \code{"final"}, \code{"baseline"}, or \code{"benefit"}.}

\item{layout}{Character string controlling the layout. Must be one of
\code{"single"} or \code{"facet"}. If \code{NULL}, the default is
\code{"facet"}.}

\item{max_facets}{Maximum number of feature facets shown when
\code{features = NULL} and faceting would otherwise create many panels.}

\item{...}{Reserved for future extensions.}

\item{base_alpha}{Unused in the current feature view, kept for interface
consistency.}

\item{selected_alpha}{Unused in the current feature view, kept for interface
consistency.}

\item{base_fill}{Unused in the current feature view, kept for interface
consistency.}

\item{base_color}{Unused in the current feature view, kept for interface
consistency.}

\item{selected_color}{Border colour for filled feature polygons.}

\item{draw_borders}{Logical. If \code{FALSE}, borders are not drawn.}

\item{show_base}{Unused in the current feature view, kept for interface
consistency.}

\item{fill_na}{Fill colour for missing values.}

\item{use_viridis}{Logical. If \code{TRUE} and the \pkg{viridis} package is
available, use a continuous viridis scale.}
}
\value{
Invisibly returns a \code{ggplot} object.
}
\description{
Plot feature values in space from a \code{Solution} or \code{SolutionSet}.

This function combines baseline feature amounts from
\code{x$problem$data$dist_features} with positive effects induced by selected
actions to produce planning-unit-level feature maps.
}
\details{
For each planning unit \eqn{i} and feature \eqn{f}, the plotted quantities
are:
\deqn{
\mathrm{baseline}_{if},
}
\deqn{
\mathrm{benefit}_{if},
}
\deqn{
\mathrm{final}_{if} = \mathrm{baseline}_{if} + \mathrm{benefit}_{if}.
}

In the current implementation:
\itemize{
\item \code{baseline} is the summed baseline amount from
\code{x$problem$data$dist_features},
\item \code{benefit} is the summed positive effect from selected actions,
\item \code{final} is \code{baseline + benefit}.
}

Negative effects are not subtracted in this plotting method. Therefore,
\code{value = "final"} should be interpreted as baseline plus selected
positive effects under the current plotting logic.

If \code{layout = "facet"} and only one run is plotted, one panel is drawn
per feature.

If multiple runs are plotted, exactly one feature must be requested, and
faceting is done by run.

Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
}
\examples{
if (requireNamespace("sf", quietly = TRUE) &&
    requireNamespace("ggplot2", quietly = TRUE)) {
  data("sim_pu_sf", package = "multiscape")

  n <- min(6, nrow(sim_pu_sf))
  ids <- sim_pu_sf$id[seq_len(n)]

  features_df <- data.frame(
    id = c(1, 2),
    name = c("feature_1", "feature_2")
  )

  dist_features_df <- data.frame(
    pu = rep(ids, times = 2),
    feature = rep(c(1, 2), each = n),
    amount = c(seq_len(n), rev(seq_len(n)))
  )

  dist_effects_df <- data.frame(
    pu = ids,
    action = "conservation",
    feature = rep(c(1, 2), length.out = n),
    benefit = rep(1, n)
  )

  actions_df <- data.frame(
    id = "conservation",
    name = "conservation"
  )

  problem <- structure(
    list(
      data = list(
        pu_sf = sim_pu_sf,
        features = features_df,
        dist_features = dist_features_df,
        dist_effects = dist_effects_df,
        actions = actions_df
      )
    ),
    class = "Problem"
  )

  sol <- structure(
    list(
      problem = problem,
      summary = list(
        actions = data.frame(
          pu = ids,
          action = "conservation",
          selected = 1L
        )
      )
    ),
    class = "Solution"
  )

  plot_spatial_features(sol, features = "feature_1", value = "final")
}

}
\seealso{
\code{\link{get_features}},
\code{\link{plot_spatial}},
\code{\link{plot_spatial_pu}},
\code{\link{plot_spatial_actions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_effects.R
\name{add_losses}
\alias{add_losses}
\title{Add losses}
\usage{
add_losses(
  x,
  losses = NULL,
  effect_type = c("delta", "after"),
  effect_aggregation = c("sum", "mean")
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}. It
must already contain feasible actions; run \code{\link{add_actions}} first.}

\item{losses}{Alias of \code{effects}, used for symmetry with
\code{add_benefits()}.}

\item{effect_type}{Character string indicating how supplied effect values are
interpreted. Must be one of:
\itemize{
\item \code{"delta"}: values represent signed net changes,
\item \code{"after"}: values represent after-action amounts and are
converted to net changes relative to baseline feature amounts.
}}

\item{effect_aggregation}{Character string giving the aggregation used when
converting raster values to planning-unit level. Must be one of
\code{"sum"} or \code{"mean"}.}
}
\value{
An updated \code{Problem} object containing:
\describe{
\item{\code{dist_effects}}{The canonical filtered effects table,
containing only rows with \code{loss > 0}.}
\item{\code{dist_loss}}{A convenience table containing only the loss
component.}
\item{\code{losses_meta}}{Metadata for the stored loss table.}
}
}
\description{
Convenience wrapper around \code{\link{add_effects}} that keeps only negative
effects, represented by rows with \code{loss > 0}.

This function is useful when the workflow is focused only on damaging
consequences of actions and a loss-only wrapper is more convenient than
calling \code{add_effects(..., component = "loss")} directly. Internally,
it calls \code{add_effects()} with \code{component = "loss"}.
The canonical stored result therefore contains only rows whose effect
satisfies \eqn{\mathrm{loss}_{iaf} > 0}.

In addition, a mirror table containing only the loss component is stored in
\code{x$data$dist_loss}.
}
\examples{
pu <- data.frame(id = 1:2, cost = c(1, 2))
features <- data.frame(id = 1, name = "sp1")
dist_features <- data.frame(pu = 1:2, feature = 1, amount = c(5, 10))

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  p,
  data.frame(id = "harvest")
)

eff <- data.frame(
  pu = c(1, 2),
  action = c("harvest", "harvest"),
  feature = c(1, 1),
  delta = c(2, -4)
)

p <- add_losses(p, losses = eff)
p$data$dist_loss

}
\seealso{
\code{\link{add_effects}},
\code{\link{add_benefits}},
\code{\link{add_objective_min_loss}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_budget.R
\name{add_constraint_budget}
\alias{add_constraint_budget}
\title{Add budget constraint}
\usage{
add_constraint_budget(
  x,
  budget,
  sense,
  tolerance = 0,
  actions = NULL,
  include_pu_cost = TRUE,
  include_action_cost = TRUE,
  name = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{budget}{Numeric scalar greater than or equal to zero. Target value for
the constrained budget.}

\item{sense}{Character string indicating the type of budget constraint. Must
be one of \code{"min"}, \code{"max"}, or \code{"equal"}.}

\item{tolerance}{Numeric scalar greater than or equal to zero. Only used when
\code{sense = "equal"}. In that case, equality is interpreted as a band
around \code{budget} with half-width \code{tolerance}. Ignored otherwise.}

\item{actions}{Optional subset of actions to which the constraint applies.
If \code{NULL}, the constraint applies to the whole problem. Otherwise, it
applies only to the selected decision variables associated with the
specified subset of actions. This argument is resolved using the package's
standard action subset parser.}

\item{include_pu_cost}{Logical scalar indicating whether planning-unit costs
should be included in the constrained budget. This is only supported when
\code{actions = NULL}.}

\item{include_action_cost}{Logical scalar indicating whether action costs
should be included in the constrained budget.}

\item{name}{Optional character string used as the label of the stored linear
constraint when it is later added to the optimization model. If
\code{NULL}, a default name is generated.}
}
\value{
An updated \code{Problem} object with the new budget-constraint
specification appended to the stored budget-constraint table.
}
\description{
Add a budget constraint to a planning problem.

This function stores one budget-constraint specification in the
\code{Problem} object so that it can later be incorporated when the
optimization model is assembled. Multiple budget constraints can be added by
calling this function repeatedly, provided that no duplicated combination of
action subset and constraint sense is introduced.
}
\details{
Use this function when spending limits or minimum spending requirements must
be imposed either on the full problem or on the subset of selected decisions
associated with specific actions.

Let \eqn{\mathcal{I}} denote the set of planning units and let
\eqn{\mathcal{A}} denote the set of actions. Let
\eqn{w_i \in \{0,1\}} denote the binary variable indicating whether planning
unit \eqn{i \in \mathcal{I}} is selected by at least one decision in the
model, and let \eqn{x_{ia} \in \{0,1\}} denote the binary variable
indicating whether action \eqn{a \in \mathcal{A}} is selected in planning
unit \eqn{i \in \mathcal{I}}.

The total constrained budget can include two cost components:

\itemize{
\item planning-unit costs, associated with \eqn{w_i};
\item action costs, associated with \eqn{x_{ia}}.
}

The arguments \code{include_pu_cost} and \code{include_action_cost}
determine which of these components are included in the stored budget
constraint.

When \code{actions = NULL}, the constraint refers to the total budget across
the whole problem. In that case, depending on the values of
\code{include_pu_cost} and \code{include_action_cost}, the constrained
quantity is one of the following:

If only planning-unit costs are included:
\deqn{
\sum_{i \in \mathcal{I}} c_i^{pu} w_i
}

If only action costs are included:
\deqn{
\sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}} c_{ia}^{act} x_{ia}
}

If both components are included:
\deqn{
\sum_{i \in \mathcal{I}} c_i^{pu} w_i +
\sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}} c_{ia}^{act} x_{ia}
}

Depending on \code{sense}, this function stores one of the following
constraints:

If \code{sense = "min"}:
\deqn{
C \ge B
}

If \code{sense = "max"}:
\deqn{
C \le B
}

If \code{sense = "equal"} and \code{tolerance = 0}:
\deqn{
C = B
}

If \code{sense = "equal"} and \code{tolerance > 0}, the equality is stored as
a two-sided band:
\deqn{
B - \tau \le C \le B + \tau
}

where \eqn{C} denotes the selected cost expression and \eqn{\tau} is the
value supplied through \code{tolerance}.

When \code{actions} is not \code{NULL}, only action costs can be included.
In that case, let \eqn{\mathcal{A}^\star \subseteq \mathcal{A}} denote the
selected subset of actions, and the constrained quantity is
\deqn{
\sum_{i \in \mathcal{I}} \sum_{a \in \mathcal{A}^\star} c_{ia}^{act} x_{ia}.
}

Action-specific budget constraints only support action costs. Therefore,
\code{include_pu_cost = TRUE} is only allowed when \code{actions = NULL},
because planning-unit costs are not action-specific.

This function only stores the constraint specification; it does not validate
the feasibility of the threshold against the available cost data at this
stage.

Multiple budget constraints can be stored in a \code{Problem} object.
However, at most one can be stored for the same combination of action subset
and constraint sense. Attempting to add a duplicated
\code{actions}--\code{sense} combination results in an error.
}
\examples{
pu <- data.frame(
  id = 1:4,
  cost = c(2, 3, 1, 4)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

actions <- data.frame(
  id = c("conservation", "restoration")
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

p <- add_actions(
  p,
  actions = actions,
  cost = c(conservation = 1, restoration = 2)
)

p <- add_constraint_budget(
  x = p,
  budget = 10,
  sense = "max",
  include_pu_cost = TRUE,
  include_action_cost = TRUE
)

p <- add_constraint_budget(
  x = p,
  budget = 4,
  sense = "max",
  actions = "restoration",
  include_pu_cost = FALSE,
  include_action_cost = TRUE
)

p <- add_constraint_budget(
  x = p,
  budget = 1,
  sense = "min",
  actions = "restoration",
  include_pu_cost = FALSE,
  include_action_cost = TRUE
)

p$data$constraints$budget

}
\seealso{
\code{\link{create_problem}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{add_spatial_relations}
\alias{add_spatial_relations}
\title{Add spatial relations}
\usage{
add_spatial_relations(x, relations, name, directed = FALSE, allow_self = FALSE)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}.}

\item{relations}{A \code{data.frame} describing relation edges. It must
contain either:
\itemize{
\item \code{pu1}, \code{pu2}, and \code{weight}, using external
planning-unit ids, or
\item \code{internal_pu1}, \code{internal_pu2}, and \code{weight}, using
internal planning-unit indices.
}
Extra columns such as \code{distance} or \code{source} are allowed and are
preserved when possible.}

\item{name}{Character string giving the key under which the relation is
stored.}

\item{directed}{Logical. If \code{FALSE}, treat edges as undirected and
collapse duplicate unordered pairs. If \code{TRUE}, keep edges as directed
ordered pairs.}

\item{allow_self}{Logical. If \code{TRUE}, allow self-edges
\eqn{(i,i)}. Default is \code{FALSE}.}
}
\value{
An updated \code{Problem} object with the relation stored in
\code{x$data$spatial_relations[[name]]}.
}
\description{
Register an externally computed spatial relation inside a \code{Problem}
object using the unified internal representation adopted by \code{multiscape}.

Most users will typically prefer one of the convenience constructors such as
\code{\link{add_spatial_boundary}}, \code{\link{add_spatial_rook}},
\code{\link{add_spatial_queen}}, \code{\link{add_spatial_knn}}, or
\code{\link{add_spatial_distance}}. This function is the advanced low-level
entry point for adding an already computed relation.
}
\details{
Use this function when the spatial relation has already been computed
externally and should be registered directly in the \code{Problem} object.

The input relation may be provided either in terms of external planning-unit
identifiers or in terms of internal planning-unit indices.

Specifically, the input \code{relations} table must contain either:
\itemize{
\item \code{pu1}, \code{pu2}, and \code{weight}, or
\item \code{internal_pu1}, \code{internal_pu2}, and \code{weight}.
}

If external ids are supplied, they are mapped to internal indices using the
planning-unit identifiers stored in the problem.

Let \eqn{G = (\mathcal{I}, E, \omega)} denote the supplied relation, where
\eqn{E} corresponds to the rows of \code{relations}. If
\code{directed = FALSE}, each edge is treated as undirected, so pairs
\eqn{(i,j)} and \eqn{(j,i)} are interpreted as the same edge. In that case,
duplicated undirected edges are collapsed automatically using the maximum
weight observed for each unordered pair.

If \code{directed = TRUE}, edges are preserved as ordered pairs, so
\eqn{(i,j)} and \eqn{(j,i)} are distinct unless the user provides both.

Self-edges \eqn{(i,i)} are permitted only if \code{allow_self = TRUE}.

The final relation is stored in \code{x$data$spatial_relations[[name]]}.

If a relation with the same \code{name} already exists, it is replaced.
}
\examples{
pu <- data.frame(id = 1:3, cost = c(1, 2, 3))

features <- data.frame(
  id = 1,
  name = "sp1"
)

dist_features <- data.frame(
  pu = 1:3,
  feature = 1,
  amount = c(1, 1, 1)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

rel <- data.frame(
  pu1 = c(1, 1, 2),
  pu2 = c(2, 3, 3),
  weight = c(1, 1, 2)
)

p <- add_spatial_relations(
  x = p,
  relations = rel,
  name = "my_relation"
)

p$data$spatial_relations$my_relation

}
\seealso{
\code{\link{add_spatial_boundary}},
\code{\link{add_spatial_rook}},
\code{\link{add_spatial_queen}},
\code{\link{add_spatial_knn}},
\code{\link{add_spatial_distance}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_tradeoff}
\alias{plot_tradeoff}
\title{Plot trade-offs from a multi-objective solution set}
\usage{
plot_tradeoff(
  x,
  objectives = NULL,
  color_by = NULL,
  all_pairs = NULL,
  connect = FALSE,
  label_runs = FALSE,
  point_size = 3,
  line_alpha = 0.5,
  text_size = 3,
  ...
)
}
\arguments{
\item{x}{A \code{SolutionSet} object.}

\item{objectives}{Optional character vector of objective aliases to display.
These must match the suffixes of the \code{value_*} columns in
\code{x$solution$runs}. If \code{NULL}, all available objective columns are
used.}

\item{color_by}{Optional character scalar used to colour points. This may be
either one of the selected objective aliases or one of the run-level
columns \code{"run_id"}, \code{"status"}, \code{"runtime"}, or
\code{"gap"}.}

\item{all_pairs}{Logical. If \code{TRUE}, allow plotting all pairwise
combinations even when more than four objectives are selected. If
\code{NULL}, it is treated as \code{FALSE}.}

\item{connect}{Logical. If \code{TRUE}, connect points by run order within
each panel.}

\item{label_runs}{Logical. If \code{TRUE}, add run labels to points.}

\item{point_size}{Numeric point size.}

\item{line_alpha}{Numeric alpha value for connecting lines.}

\item{text_size}{Numeric size for run labels.}

\item{...}{Reserved for future extensions.}
}
\value{
Invisibly returns a \code{ggplot} object.
}
\description{
Plot pairwise trade-offs among objective values stored in a
\code{SolutionSet}.

This function is intended for multi-objective workflows in which the solution
set contains one row per run and one or more objective value columns of the
form \code{value_*}.

If exactly two objectives are selected, the function returns a single
scatterplot. If three or more objectives are selected, all pairwise
combinations are plotted using facets.
}
\details{
This function reads the run-level table stored in \code{x$solution$runs}. It
expects objective values to be stored in columns whose names begin with
\code{"value_"}.

If the available objective columns are, for example,
\code{value_cost}, \code{value_benefit}, and \code{value_frag}, then the
corresponding objective aliases are \code{"cost"}, \code{"benefit"}, and
\code{"frag"}.

Let \eqn{f_k(r)} denote the value of objective \eqn{k} in run \eqn{r}. This
function visualizes pairwise projections of the run table of the form:
\deqn{
\left(f_k(r), f_\ell(r)\right)
}
for selected pairs of objectives \eqn{k,\ell}.

If exactly two objectives are selected, a single panel is produced.

If three or more objectives are selected, all pairwise combinations are
generated:
\deqn{
\{(k,\ell): k < \ell,\; k,\ell \in \mathcal{O}\},
}
where \eqn{\mathcal{O}} is the selected set of objective aliases.

By default, plotting more than four objectives is not allowed unless
\code{all_pairs = TRUE}, because the number of panels grows quadratically in
the number of objectives.

\strong{Colouring}

If \code{color_by} is supplied, points are coloured by either:
\itemize{
\item one of the selected objective aliases, in which case the
corresponding \code{value_*} column is used,
\item or one of the run-level columns \code{run_id}, \code{status},
\code{runtime}, or \code{gap}.
}

\strong{Connecting runs}

If \code{connect = TRUE}, runs are connected in their current table order
within each panel. This can be useful when runs correspond to an ordered scan
of weights, \eqn{\epsilon}-levels, or frontier points, but it should be used
with care when run order has no substantive meaning.

\strong{Run labels}

If \code{label_runs = TRUE}, each point is labelled by its \code{run_id}. If
the \pkg{ggrepel} package is available, repelled labels are used.
}
\examples{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  solset <- structure(
    list(
      solution = list(
        runs = data.frame(
          run_id = 1:5,
          status = rep("optimal", 5),
          runtime = c(1.2, 1.1, 1.4, 1.3, 1.5),
          gap = c(0, 0, 0, 0, 0),
          value_cost = c(10, 12, 14, 16, 18),
          value_benefit = c(5, 7, 8, 9, 10),
          value_loss = c(4, 3, 3, 2, 2)
        )
      )
    ),
    class = "SolutionSet"
  )

  # Plot all pairwise trade-offs
  plot_tradeoff(solset)

  # Plot two selected objectives
  plot_tradeoff(solset, objectives = c("cost", "benefit"))

  # Colour points by one objective and label runs
  plot_tradeoff(solset, color_by = "loss", label_runs = TRUE)
}

}
\seealso{
\code{\link{solve}},
\code{\link{get_solution_vector}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_constraint_targets.R
\name{add_constraint_targets_absolute}
\alias{add_constraint_targets_absolute}
\title{Add absolute targets}
\usage{
add_constraint_targets_absolute(
  x,
  targets,
  features = NULL,
  actions = NULL,
  label = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{targets}{Target specification. This is interpreted as an absolute target
value in the same units as the modelled feature contributions. It may be a
scalar, vector, named vector, or \code{data.frame}. See Details.}

\item{features}{Optional feature specification indicating which features the
supplied target values refer to when \code{targets} does not identify
features explicitly. If \code{NULL}, all features are targeted.}

\item{actions}{Optional character vector indicating which actions count toward
target achievement. Entries may match action ids, \code{action_set} labels,
or both. If \code{NULL}, all actions count.}

\item{label}{Optional character string stored with the targets for reporting
and bookkeeping.}
}
\value{
An updated \code{Problem} object with absolute targets appended to
the stored target table.
}
\description{
Add feature-level absolute targets to a planning problem.

These targets are stored in the problem object and later translated into
linear constraints when the optimization model is built. Absolute targets
are interpreted directly in the same units as the feature
contributions used by the model. Each call appends one or more target definitions to the problem. This makes
it possible to combine multiple target rules, including targets associated
with different action subsets.
}
\details{
Use this function when target requirements are naturally expressed in the
original units of the modelled feature contributions, rather than as
proportions of current baseline totals.

Let \eqn{\mathcal{F}} denote the set of features. For each targeted feature
\eqn{f \in \mathcal{F}}, this function stores an absolute target threshold
\eqn{T_f \ge 0}.

When the optimization model is built, each such target is interpreted as a
lower-bound constraint of the form:
\deqn{
\sum_{(i,a) \in \mathcal{D}_f^{\star}} c_{iaf} x_{ia} \ge T_f,
}
where:
\itemize{
\item \eqn{i \in \mathcal{I}} indexes planning units,
\item \eqn{a \in \mathcal{A}} indexes actions,
\item \eqn{x_{ia}} indicates whether action \eqn{a} is selected in planning
unit \eqn{i},
\item \eqn{c_{iaf}} is the contribution of that action to feature
\eqn{f},
\item \eqn{\mathcal{D}_f^{\star}} is the subset of planning unit--action
pairs allowed to count toward the target for feature \eqn{f}.
}

In the absolute case, the stored target threshold is simply:
\deqn{
T_f = t_f,
}
where \eqn{t_f} is the user-supplied target value for feature \eqn{f}.

The \code{actions} argument restricts which actions may contribute toward
achievement of the target, but it does not modify the value of \eqn{T_f}
itself.

The \code{targets} argument is parsed by \code{.pa_parse_targets()} and may be
supplied in several equivalent forms, including:
\itemize{
\item a single numeric value recycled to all selected features,
\item a numeric vector aligned to \code{features},
\item a named numeric vector where names identify features,
\item a \code{data.frame} with \code{feature} and \code{target} columns.
}

If \code{targets} does not explicitly identify features:
\itemize{
\item if \code{features = NULL}, the target is applied to all features,
\item if \code{features} is supplied, the target values are interpreted with
respect to that feature set.
}

Repeated calls append new target rules rather than replacing previous ones.
This allows cumulative target modelling, including multiple rules on the same
feature with different contributing action subsets.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
) |>
  add_actions(data.frame(id = "conservation", name = "conservation"), cost = 0)

# Same absolute target for all features
p1 <- add_constraint_targets_absolute(p, 3)
p1$data$targets

# Different targets by feature
p2 <- add_constraint_targets_absolute(
  p,
  c("1" = 4, "2" = 2)
)
p2$data$targets

# Restrict which actions count toward target achievement
p3 <- add_constraint_targets_absolute(
  p,
  2,
  actions = "conservation"
)
p3$data$targets

}
\seealso{
\code{\link{add_constraint_targets_relative}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_spatial_relations.R
\name{add_spatial_distance}
\alias{add_spatial_distance}
\title{Add distance-threshold spatial relations}
\usage{
add_spatial_distance(
  x,
  coords = NULL,
  max_distance,
  name = "distance",
  weight_mode = c("constant", "inverse", "inverse_sq"),
  distance_eps = 1e-09
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}.}

\item{coords}{Optional coordinates specification, following the same rules as
in \code{\link{add_spatial_knn}}.}

\item{max_distance}{Positive numeric scalar giving the maximum distance for
an edge.}

\item{name}{Character string giving the key under which the relation is
stored.}

\item{weight_mode}{Character string indicating how distance is converted to
weight. Must be one of \code{"constant"}, \code{"inverse"}, or
\code{"inverse_sq"}.}

\item{distance_eps}{Small positive numeric constant used to avoid division by
zero in inverse-distance weighting.}
}
\value{
An updated \code{Problem} object.
}
\description{
Build and register a spatial relation connecting planning units whose
Euclidean distance is less than or equal to a user-defined threshold.

This constructor does not require polygon geometry and instead uses
planning-unit coordinates.
}
\details{
Use this function when neighbourhood should be defined by a fixed distance
radius rather than by polygon topology or a fixed number of neighbours.

Let \eqn{s_i = (x_i, y_i)} denote the coordinates of planning unit
\eqn{i}. Let \eqn{d_{ij}} be the Euclidean distance between planning units
\eqn{i} and \eqn{j}.

For a user-supplied threshold \eqn{d_{\max}}, this constructor creates an
edge between \eqn{i} and \eqn{j} whenever:
\deqn{
d_{ij} \le d_{\max}.
}

Edge weights are assigned according to \code{weight_mode}:
\itemize{
\item \code{"constant"}:
\deqn{\omega_{ij} = 1,}
\item \code{"inverse"}:
\deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)},}
\item \code{"inverse_sq"}:
\deqn{\omega_{ij} = \frac{1}{\max(d_{ij}, \varepsilon)^2},}
}
where \eqn{\varepsilon} = \code{distance_eps} is a small constant.

The implementation computes an \eqn{O(n^2)} distance matrix and is therefore
best suited to small or moderate numbers of planning units. For large
problems, \code{\link{add_spatial_knn}} is often more scalable.

The resulting relation is registered as undirected.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4),
  x = c(0, 1, 0, 1),
  y = c(0, 0, 1, 1)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

p <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
)

p <- add_spatial_distance(
  x = p,
  max_distance = 1.01,
  name = "within_1",
  weight_mode = "constant"
)

p$data$spatial_relations$within_1

}
\seealso{
\code{\link{add_spatial_knn}},
\code{\link{add_spatial_relations}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_solver.R
\name{set_solver}
\alias{set_solver}
\title{Configure solver settings}
\usage{
set_solver(
  x,
  solver = c("auto", "gurobi", "cplex", "cbc", "symphony"),
  gap_limit = NULL,
  time_limit = NULL,
  solution_limit = NULL,
  cores = NULL,
  verbose = FALSE,
  log_file = NULL,
  write_log = NULL,
  solver_params = list(),
  ...
)
}
\arguments{
\item{x}{A \code{Problem} object.}

\item{solver}{Character string indicating the solver backend to use. Must be
one of \code{"auto"}, \code{"gurobi"}, \code{"cplex"},
\code{"cbc"}, or \code{"symphony"}.}

\item{gap_limit}{Optional numeric value in \eqn{[0,1]} giving the relative
optimality gap for mixed-integer optimization. If \code{NULL}, the
previously stored value is kept unchanged.}

\item{time_limit}{Optional non-negative numeric value giving the maximum
solving time in seconds. If \code{NULL}, the previously stored value is
kept unchanged.}

\item{solution_limit}{Optional logical flag controlling backend-specific
early stopping after feasible solution discovery. If \code{NULL}, the
previously stored value is kept unchanged.}

\item{cores}{Optional positive integer giving the number of CPU cores to use.
If \code{NULL}, the previously stored value is kept unchanged.}

\item{verbose}{Optional logical flag indicating whether the solver should
print log output. If \code{NULL}, the previously stored value is kept
unchanged.}

\item{log_file}{Optional character string giving the name of the
solver log file. If \code{NULL}, the previously stored value is kept
unchanged.}

\item{write_log}{Optional logical flag indicating whether solver output
should be written to a file. If \code{NULL}, the previously stored value is
kept unchanged.}

\item{solver_params}{Named list of solver-specific parameters. These are
merged with previously stored backend-specific parameters rather than
replacing them completely.}

\item{...}{Additional named solver-specific parameters. These are merged into
\code{solver_params}. For example, \code{MIPFocus = 1} for Gurobi.}
}
\value{
An updated \code{Problem} object with modified solver settings stored
in \code{x$data$solve_args}.
}
\description{
Store solver configuration inside a \code{Problem} object so that
\code{\link{solve}} can later run using the stored backend and runtime
options.

This function does not build or solve the optimization model. It only updates
the solver configuration stored in \code{x$data$solve_args}.
}
\details{
\strong{Purpose}

The \code{multiscape} workflow separates problem specification from solver
configuration. Problem data, actions, effects, targets, objectives, and
methods are stored in the \code{Problem} object, and solver settings are
stored separately in \code{x$data$solve_args}.

This function allows solver options to be configured once and reused later
through \code{\link{solve}(x)} without repeating the same arguments each time.

\strong{Stored fields}

The solver configuration is stored in \code{x$data$solve_args}. Typical
entries include:
\itemize{
\item \code{solver},
\item \code{gap_limit},
\item \code{time_limit},
\item \code{solution_limit},
\item \code{cores},
\item \code{verbose},
\item \code{write_log},
\item \code{log_file},
\item \code{solver_params}.
}

\strong{Incremental update semantics}

This function updates solver settings incrementally.

If an argument is supplied as \code{NULL}, the previously stored value is
kept unchanged. Therefore, repeated calls can be used to modify only selected
components of the solver configuration.

For example, a user may first configure the solver backend and time limit,
and later update only the optimality gap or only a backend-specific
parameter.

\strong{Gap limit}

The argument \code{gap_limit} is interpreted as a relative optimality gap for
mixed-integer optimization. It must lie in \eqn{[0,1]}.

If the solver stops with incumbent value \eqn{z^{\mathrm{inc}}} and best
bound \eqn{z^{\mathrm{bd}}}, then the exact stopping rule depends on the
solver backend, but conceptually \code{gap_limit} controls the maximum
accepted relative difference between the incumbent and the bound.

\strong{Time limit}

The argument \code{time_limit} is interpreted as a maximum wall-clock time in
seconds allowed for the solver.

\strong{Solution limit}

The argument \code{solution_limit} is stored as a logical flag. Its exact
meaning depends on the backend-specific solving layer, but conceptually it
requests early termination after finding a feasible solution according to the
behaviour supported by the chosen solver.

\strong{Cores}

The argument \code{cores} specifies the number of CPU cores to use. If the
requested number exceeds the number of detected cores, it is capped to the
detected maximum with a warning.

\strong{Verbose output and log files}

The arguments \code{verbose}, \code{write_log}, and \code{log_file}
control how solver logging is handled. These options are stored and later
interpreted by the solving layer for the selected backend.

\strong{Solver-specific parameters}

Additional backend-specific parameters can be passed in two ways:
\itemize{
\item through the named list \code{solver_params},
\item through additional named arguments in \code{...}.
}

These two sources are merged, and the result is then merged with any
previously stored \code{solver_params}. Existing parameters are therefore
preserved unless explicitly overwritten.

This is particularly useful for backend-specific controls such as node
selection, emphasis parameters, tolerances, or heuristics.

\strong{Supported backends}

The \code{solver} argument selects the backend to be used later by
\code{\link{solve}}. Supported values are:
\itemize{
\item \code{"auto"}: let the solving layer choose an available backend,
\item \code{"gurobi"},
\item \code{"cplex"},
\item \code{"cbc"},
\item \code{"symphony"}.
}

This function only stores the requested backend. Availability of the backend
is checked later when solving.
}
\examples{
pu_tbl <- data.frame(
  id = 1:4,
  cost = c(1, 2, 3, 4)
)

feat_tbl <- data.frame(
  id = 1:2,
  name = c("feature_1", "feature_2")
)

dist_feat_tbl <- data.frame(
  pu = c(1, 1, 2, 3, 4),
  feature = c(1, 2, 2, 1, 2),
  amount = c(5, 2, 3, 4, 1)
)

x <- create_problem(
  pu = pu_tbl,
  features = feat_tbl,
  dist_features = dist_feat_tbl,
  cost = "cost"
)

x1 <- set_solver(
  x,
  solver = "cbc",
  gap_limit = 0.01,
  time_limit = 300,
  cores = 2,
  verbose = TRUE
)

x1$data$solve_args

# Update only selected settings
x2 <- set_solver(
  x1,
  gap_limit = 0.05,
  solver_params = list(randomSeed = 123)
)

x2$data$solve_args

}
\seealso{
\code{\link{solve}},
\code{\link{set_solver_gurobi}},
\code{\link{set_solver_cplex}},
\code{\link{set_solver_cbc}},
\code{\link{set_solver_symphony}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_actions.R
\name{add_actions}
\alias{add_actions}
\title{Add management actions to a planning problem}
\usage{
add_actions(
  x,
  actions,
  include_pairs = NULL,
  exclude_pairs = NULL,
  cost = NULL
)
}
\arguments{
\item{x}{A \code{Problem} object created with \code{\link{create_problem}}.}

\item{actions}{A \code{data.frame} defining the action catalogue. It must
contain a unique \code{id} column. A column named \code{action} is also
accepted and automatically renamed to \code{id}.}

\item{include_pairs}{Optional specification of feasible \code{(pu, action)}
pairs. It can be \code{NULL}, a \code{data.frame} with columns
\code{pu} and \code{action} (optionally also \code{feasible}), or a named
list whose names are action ids and whose elements are vectors of planning
unit ids or \code{sf} objects.}

\item{exclude_pairs}{Optional specification of infeasible \code{(pu, action)}
pairs. It uses the same formats as \code{include_pairs} and removes
matching pairs from the feasible set.}

\item{cost}{Optional cost specification for feasible pairs. It may be
\code{NULL}, a scalar numeric value, a named numeric vector indexed by
action id, or a \code{data.frame} with columns \code{action, cost} or
\code{pu, action, cost}.}
}
\value{
An updated \code{Problem} object with:
\describe{
\item{\code{actions}}{The action catalogue, including a unique integer
\code{internal_id} for each action.}
\item{\code{dist_actions}}{The feasible planning unit--action table with
columns \code{pu}, \code{action}, \code{cost}, \code{status},
\code{internal_pu}, and \code{internal_action}.}
\item{\code{pu index}}{A mapping from user-supplied planning-unit ids to
internal integer ids.}
\item{\code{action index}}{A mapping from action ids to internal integer
ids.}
}
}
\description{
Define the action catalogue, the set of feasible planning unit--action pairs,
and their implementation costs.

This function adds two core components to a \code{Problem} object. First, it
stores the action catalogue. Second, it creates the feasible planning
unit--action table, including implementation costs, status codes, and
internal indices used by the optimization backend.

Conceptually, if \eqn{\mathcal{I}} is the set of planning units and
\eqn{\mathcal{A}} is the set of actions, this function determines which
pairs \eqn{(i,a) \in \mathcal{I} \times \mathcal{A}} are feasible decisions
and assigns a non-negative implementation cost to each feasible pair.
}
\details{
\strong{When to use \code{add_actions()}.}

Use this function when you want to move from a planning problem defined only
by planning units and features to a problem in which decisions are explicitly
represented as actions applied in planning units.

\strong{Action catalogue.}

The \code{actions} argument must be a \code{data.frame} with a unique
\code{id} column identifying each action. If a column named \code{action} is
supplied instead, it is renamed internally to \code{id}. Additional columns
are preserved. If no \code{name} column is provided, action labels are taken
from \code{id}. If an \code{action_set} column is present, it is also
preserved and can later be used to refer to groups of actions.

Actions are stored sorted by \code{id} to ensure reproducible internal
indexing.

\strong{Feasible planning unit--action pairs.}

Feasibility is controlled through \code{include_pairs} and
\code{exclude_pairs}.

If \code{include_pairs = NULL}, all possible \code{(pu, action)} pairs are
initially considered feasible, that is, all pairs
\eqn{(i,a) \in \mathcal{I} \times \mathcal{A}}.

If \code{include_pairs} is supplied, only those pairs are retained. If
\code{exclude_pairs} is also supplied, matching pairs are removed
afterwards.

More precisely, let \eqn{\mathcal{D}^{\mathrm{inc}}} denote the set of
included planning unit--action pairs and let
\eqn{\mathcal{D}^{\mathrm{exc}}} denote the set of excluded pairs.

If \code{include_pairs = NULL}, the feasible decision set is:
\deqn{
\{(i,a) : i \in \mathcal{I},\ a \in \mathcal{A}\} \setminus \mathcal{D}^{\mathrm{exc}}.
}

If \code{include_pairs} is supplied, the feasible decision set is:
\deqn{
\mathcal{D}^{\mathrm{inc}} \setminus \mathcal{D}^{\mathrm{exc}}.
}

Both \code{include_pairs} and \code{exclude_pairs} can be specified as:
\itemize{
\item \code{NULL},
\item a \code{data.frame} with columns \code{pu} and \code{action},
\item or a named list whose names are action ids.
}

When supplied as a \code{data.frame}, the object must contain columns
\code{pu} and \code{action}. An optional logical-like column
\code{feasible} may also be provided; only rows with \code{feasible = TRUE}
are retained. Missing values in \code{feasible} are treated as
\code{FALSE}.

When supplied as a named list, names must match action ids. Each element may
contain either:
\itemize{
\item a vector of planning-unit ids, or
\item an \code{sf} object defining the spatial zone where the action is
feasible.
}

In the spatial case, feasible planning units are identified using
\code{sf::st_intersects()} against the stored planning-unit geometry.

\strong{Feasibility versus decision fixing.}

This function only determines whether a pair \eqn{(i,a)} exists in the model.
It does not force a feasible action to be selected or forbidden beyond
structural infeasibility. Fixed decisions should instead be imposed later
with \code{\link{add_constraint_locked_actions}}.

\strong{Costs.}

Costs can be supplied in several ways:
\itemize{
\item If \code{cost = NULL}, all feasible pairs receive a default cost of
\code{1}.
\item If \code{cost} is a scalar, that value is assigned to all feasible
pairs.
\item If \code{cost} is a named numeric vector, names must match action ids
and costs are assigned by action.
\item If \code{cost} is a \code{data.frame}, it must define either:
\itemize{
\item action-level costs through columns \code{action} and \code{cost}, or
\item pair-specific costs through columns \code{pu}, \code{action}, and
\code{cost}.
}
}

In all cases, costs must be finite and non-negative.

In practice, a scalar cost is useful when all actions cost the same
everywhere, a named vector is useful when cost depends only on action type,
and a \code{(pu, action, cost)} table is useful when cost varies by both
planning unit and action.

\strong{Status values.}

Internally, all feasible pairs are initialized with \code{status = 0},
meaning that the decision is free. If planning units have already been marked
as locked out, then all feasible actions in those planning units are assigned
\code{status = 3}. This preserves consistency with planning-unit exclusions
already stored in the problem.

\strong{Replacement behaviour.}

Calling \code{add_actions()} replaces any previous action catalogue and
feasible action table stored in the problem object.

After defining actions, typical next steps include adding effects, optional
decision-fixing constraints, objectives, and solver settings before calling
\code{solve()}.
}
\examples{
# ------------------------------------------------------
# Minimal planning problem
# ------------------------------------------------------
pu <- data.frame(
  id = 1:4,
  cost = c(2, 3, 1, 4)
)

features <- data.frame(
  id = 1:2,
  name = c("sp1", "sp2")
)

dist_features <- data.frame(
  pu = c(1, 1, 2, 3, 4, 4),
  feature = c(1, 2, 1, 2, 1, 2),
  amount = c(1, 2, 1, 3, 2, 1)
)

p <- create_problem(
  pu = pu,
  features = features,
  dist_features = dist_features
)

actions <- data.frame(
  id = c("conservation", "restoration"),
  name = c("Conservation", "Restoration")
)

# Example 1: all actions feasible in all planning units
p1 <- add_actions(
  x = p,
  actions = actions,
  cost = c(conservation = 5, restoration = 12)
)

print(p1)
utils::head(p1$data$dist_actions)

# Example 2: specify feasible pairs explicitly
include_df <- data.frame(
  pu = c(1, 2, 3, 4),
  action = c("conservation", "conservation", "restoration", "restoration")
)

p2 <- add_actions(
  x = p,
  actions = actions,
  include_pairs = include_df,
  cost = 10
)

p2$data$dist_actions

# Example 3: remove selected pairs after full expansion
exclude_df <- data.frame(
  pu = c(2, 4),
  action = c("restoration", "conservation")
)

p3 <- add_actions(
  x = p,
  actions = actions,
  exclude_pairs = exclude_df,
  cost = c(conservation = 3, restoration = 8)
)

p3$data$dist_actions

}
\seealso{
\code{\link{create_problem}},
\code{\link{add_constraint_locked_actions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_spatial_actions}
\alias{plot_spatial_actions}
\title{Plot selected actions in space}
\usage{
plot_spatial_actions(
  x,
  runs = NULL,
  actions = NULL,
  layout = NULL,
  max_facets = 4L,
  ...,
  base_alpha = 0.08,
  selected_alpha = 0.95,
  base_fill = "grey95",
  base_color = NA,
  selected_color = NA,
  draw_borders = FALSE,
  show_base = TRUE,
  fill_values = NULL,
  fill_na = "grey80",
  use_viridis = TRUE
)
}
\arguments{
\item{x}{A \code{Solution} or \code{SolutionSet} object.}

\item{runs}{Optional integer vector of run ids. If \code{NULL}, a
\code{Solution} is used directly and a \code{SolutionSet} defaults to the
first run.}

\item{actions}{Optional action subset to display. Entries may match action
ids or action-set labels.}

\item{layout}{Character string controlling the layout. Must be one of
\code{"single"} or \code{"facet"}. If \code{NULL}, the default is
\code{"single"}.}

\item{max_facets}{Maximum number of action facets shown when \code{actions}
is \code{NULL} and faceting would otherwise create many panels.}

\item{...}{Reserved for future extensions.}

\item{base_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the base
planning-unit layer.}

\item{selected_alpha}{Numeric value in \eqn{[0,1]} giving the alpha of the
highlighted action layer.}

\item{base_fill}{Fill colour for the base planning-unit layer.}

\item{base_color}{Border colour for the base planning-unit layer.}

\item{selected_color}{Border colour for highlighted layers.}

\item{draw_borders}{Logical. If \code{FALSE}, borders are not drawn.}

\item{show_base}{Logical. If \code{TRUE}, draw the base planning-unit layer
underneath the highlighted output.}

\item{fill_values}{Optional named vector of colours for discrete action maps.}

\item{fill_na}{Fill colour for missing values.}

\item{use_viridis}{Logical. If \code{TRUE} and the \pkg{viridis} package is
available, use viridis discrete scales.}
}
\value{
Invisibly returns a \code{ggplot} object.
}
\description{
Plot the spatial distribution of selected actions from a
\code{Solution} or \code{SolutionSet}.

This function maps the selected planning unit--action pairs returned by
\code{\link{get_actions}(only_selected = TRUE)} onto the planning-unit
geometry stored in the associated \code{Problem} object.
}
\details{
Let \eqn{x_{ia} \in \{0,1\}} denote whether action \eqn{a} is selected in
planning unit \eqn{i}. This function plots the selected
\code{(pu, action)} pairs in geographic space.

If \code{layout = "facet"} and only one run is plotted, one panel is drawn
per action.

If \code{layout = "single"}, all selected actions are drawn in a single map
using discrete fills. If more than one action is selected in the same
planning unit, the action labels are collapsed using \code{"+"}.

When plotting multiple runs, only \code{layout = "single"} is supported.

Planning-unit geometry must be available in \code{x$problem$data$pu_sf}.
}
\examples{
if (requireNamespace("sf", quietly = TRUE) &&
    requireNamespace("ggplot2", quietly = TRUE)) {
  data("sim_pu_sf", package = "multiscape")

  actions_df <- data.frame(
    id = c("conservation", "restoration"),
    name = c("conservation", "restoration")
  )

  problem <- structure(
    list(
      data = list(
        pu_sf = sim_pu_sf,
        actions = actions_df
      )
    ),
    class = "Problem"
  )

  ids <- sim_pu_sf$id[seq_len(min(6, nrow(sim_pu_sf)))]
  sol <- structure(
    list(
      problem = problem,
      summary = list(
        actions = data.frame(
          pu = c(ids[1:3], ids[4:6]),
          action = c(rep("conservation", 3), rep("restoration", 3)),
          selected = 1L
        )
      )
    ),
    class = "Solution"
  )

  plot_spatial_actions(sol, layout = "facet")
}

}
\seealso{
\code{\link{get_actions}},
\code{\link{plot_spatial}},
\code{\link{plot_spatial_pu}},
\code{\link{plot_spatial_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-sim.R
\docType{data}
\name{sim_pu_sf}
\alias{sim_pu_sf}
\title{Simulated planning units}
\format{
An object of class \code{sf}.
}
\usage{
data(sim_pu_sf)
}
\description{
Example planning units as an \code{sf} object for package examples and tests.
}
\keyword{datasets}
